Require Export DeclarationSyntax.
Require Export DeclarationTyping.
Require Export InfrastructureOperations.
Require Export InfrastructureLemmasTerms.
Set Implicit Arguments.

(******************************************************************************)
(* Context insertions.                                                        *)
(******************************************************************************)

Inductive shift_etvar : nat → Env → Env → Prop :=
  | ib_here {Γ : Env} :
      shift_etvar O Γ (etvar Γ)
  | ib_var {Γ1 : Env} {Γ2 : Env} {c T} :
      shift_etvar c Γ1 Γ2 →
      shift_etvar c (evar Γ1 T) (evar Γ2 (tshiftTy c T))
  | ib_etvar {Γ1 : Env} {Γ2 : Env} {c} :
      shift_etvar c Γ1 Γ2 →
      shift_etvar (S c) (etvar Γ1) (etvar Γ2).
Hint Constructors shift_etvar.

Inductive shift_evar : nat → Env → Env → Prop :=
  | iv_here {Γ : Env} {T} :
      shift_evar O Γ (evar Γ T)
  | iv_var {Γ1 : Env} {Γ2 : Env} {c T} :
      shift_evar c Γ1 Γ2 →
      shift_evar (S c) (evar Γ1 T) (evar Γ2 T)
  | iv_etvar {Γ1 : Env} {Γ2 : Env} {c} :
      shift_evar c Γ1 Γ2 →
      shift_evar c (etvar Γ1) (etvar Γ2).
Hint Constructors shift_evar.

(******************************************************************************)
(* Lemmas about shifting and context lookups.                                 *)
(******************************************************************************)

Lemma shift_etvar_lookup_etvar {Γ1 Γ2 c} (ib : shift_etvar c Γ1 Γ2) :
  ∀ X, lookup_etvar Γ1 X →
       lookup_etvar Γ2 (shiftIndex c X).
Proof.
  induction ib; intros; dependent destruction H; simpl;
    try rewrite tshiftTy_comm; repeat constructor; auto.
Qed.

Lemma shift_etvar_lookup_evar {Γ1 Γ2 c} (ib : shift_etvar c Γ1 Γ2) :
  ∀ x T, lookup_evar Γ1 x T →
         lookup_evar Γ2 x (tshiftTy c T).
Proof.
  induction ib; intros; dependent destruction H; simpl;
    try rewrite tshiftTy_comm; repeat constructor; auto.
Qed.

Lemma shift_evar_lookup_etvar {Γ1 Γ2 c} (iv : shift_evar c Γ1 Γ2) :
  ∀ X, lookup_etvar Γ1 X → lookup_etvar Γ2 X.
Proof.
  induction iv; intros; dependent destruction H; simpl;
    repeat constructor; auto.
Qed.

Lemma shift_evar_lookup_evar {Γ1 Γ2 c} (iv : shift_evar c Γ1 Γ2) :
  ∀ x T, lookup_evar Γ1 x T → lookup_evar Γ2 (shiftIndex c x) T.
Proof.
  induction iv; intros; dependent destruction H; simpl;
    repeat constructor; auto.
Qed.

Lemma shift_etvar_wftyp {Γ1 T} (wT: wfTy Γ1 T) :
  ∀ {c Γ2}, shift_etvar c Γ1 Γ2 → wfTy Γ2 (tshiftTy c T).
Proof.
  induction wT; simpl; econstructor; eauto using shift_etvar_lookup_etvar.
Qed.

Lemma weakening_etvar_wftyp {Γ T} (wT: wfTy Γ T) :
  wfTy (etvar Γ) (tshiftTy O T).
Proof.
  eauto using shift_etvar_wftyp.
Qed.

Lemma shift_evar_wftyp {Γ1 T} (wT: wfTy Γ1 T) :
  ∀ {c Γ2}, shift_evar c Γ1 Γ2 → wfTy Γ2 T.
Proof.
  induction wT; simpl; econstructor; eauto using shift_evar_lookup_etvar.
Qed.

Lemma weakening_var_wftyp {Γ : Env} {S T} :
  wfTy Γ T → wfTy (evar Γ S) T.
Proof.
  eauto using shift_evar_wftyp.
Qed.

(******************************************************************************)
(* Context extensions.                                                        *)
(******************************************************************************)

Lemma extend_append (Γ : Env) (Δ1 Δ2 : Ext) :
  extend Γ (append Δ1 Δ2) = extend (extend Γ Δ1) Δ2.
Proof.
  induction Δ2; simpl; f_equal; auto.
Qed.

Lemma lengthExt_append (Δ1 Δ2 : Ext) :
  lengthExt (append Δ1 Δ2) = lengthExt Δ2 + lengthExt Δ1.
Proof.
  induction Δ2; simpl; try rewrite IHΔ2; induction Δ1; auto.
Qed.

Lemma tshiftExt_append (c : nat) (Δ1 Δ2 : Ext) :
  tshiftExt c (append Δ1 Δ2) = append (tshiftExt c Δ1) (tshiftExt c Δ2).
Proof.
  induction Δ2; simpl; f_equal; auto.
Qed.

Lemma tsubstExt_append (X : nat) (T' : Ty) (Δ1 Δ2 : Ext) :
  tsubstExt X T' (append Δ1 Δ2) = append (tsubstExt X T' Δ1) (tsubstExt X T' Δ2).
Proof.
  induction Δ2; simpl; f_equal; auto.
Qed.

Lemma weakenshift_etvar {Γ1 : Env} {Γ2 : Env} {Δ : Ext} {c} :
  shift_etvar c Γ1 Γ2 →
  shift_etvar c (extend Γ1 Δ) (extend Γ2 (tshiftExt c Δ)).
Proof.
  revert c Γ2; induction Δ; simpl; intros; try constructor; auto.
Qed.

Lemma weakenshift_evar {Γ1 : Env} {Γ2 : Env} (Δ : Ext) {c} :
  shift_evar c Γ1 Γ2 → shift_evar (lengthExt Δ + c) (extend Γ1 Δ) (extend Γ2 Δ).
Proof.
  revert c Γ2; induction Δ; simpl; intros; try constructor; auto.
Qed.

(******************************************************************************)
(* Context etvar substitution                                                 *)
(******************************************************************************)

Inductive subst_etvar : nat → Ty → Env → Env → Prop :=
  | tse_here {Γ : Env} {T} :
      wfTy Γ T →
      subst_etvar O T (etvar Γ) Γ
  | tse_var {Γ1 : Env} {Γ2 : Env} {T' T} {X} :
      subst_etvar X T' Γ1 Γ2 →
      subst_etvar X T' (evar Γ1 T) (evar Γ2 (tsubstTy X T' T))
  | tse_etvar {Γ1 : Env} {Γ2 : Env} {X T'} :
      subst_etvar X T' Γ1 Γ2 →
      subst_etvar (S X) T' (etvar Γ1) (etvar Γ2).
Hint Constructors subst_etvar.

Lemma weaken_subst_etvar {Γ1 Γ2 : Env} {Δ : Ext} {X T'} :
  subst_etvar X T' Γ1 Γ2 →
  subst_etvar X T' (extend Γ1 Δ) (extend Γ2 (tsubstExt X T' Δ)).
Proof.
  revert X T' Γ1 Γ2; induction Δ; simpl; intros; try constructor; auto.
Qed.
Hint Resolve weaken_subst_etvar.

Lemma subst_etvar_lookup_etvar {Γ1 Γ2 : Env} {X T' Y} :
  subst_etvar X T' Γ1 Γ2 → lookup_etvar Γ1 Y →
  wfTy Γ2 (tsubstIndex X T' Y).
Proof.
  intros esub gb; revert Γ2 X esub.
  induction gb; intros; dependent destruction esub; simpl;
    auto using wf_tvar, weakening_etvar_wftyp, weakening_var_wftyp.
Qed.
Hint Resolve subst_etvar_lookup_etvar.


Lemma subst_etvar_wftyp_ind {Γ1 Γ2 : Env} {X T' T} :
  subst_etvar X T' Γ1 Γ2 → wfTy Γ1 T → wfTy Γ2 (tsubstTy X T' T).
Proof.
  intros esub st; revert Γ2 X T' esub.
  induction st; simpl; intros; eauto using wfTy, subst_etvar_lookup_etvar.
Qed.

Lemma subst_etvar_lookup_evar {Γ1 : Env} {Γ2 : Env} {X x T' T} :
  subst_etvar X T' Γ1 Γ2 → lookup_evar Γ1 x T →
  lookup_evar Γ2 x (tsubstTy X T' T).
Proof.
  intros sube gv; revert x T gv.
  induction sube; intros; dependent destruction gv; simpl;
    try rewrite tsubstTy_tshiftTy_reflection;
    try rewrite tsubstTy_tshiftTy_comm; try constructor; auto.
Qed.

(******************************************************************************)
(* Context evar substitution                                                  *)
(******************************************************************************)

Inductive subst_evar : Subst → Tm → Env → Env → Prop :=
  | se_here {Γ : Env} {t T} :
      Typing Γ t T →
      subst_evar sub_here t (evar Γ T) Γ
  | se_var {Γ1 Γ2 : Env} {T t} {x} :
      subst_evar x t Γ1 Γ2 →
      subst_evar (sub_var x) t (evar Γ1 T) (evar Γ2 T)
  | se_etvar {Γ1 Γ2 : Env} {x} {t} :
      subst_evar x t Γ1 Γ2 →
      subst_evar (sub_tvar x) t (etvar Γ1) (etvar Γ2).
Hint Constructors subst_evar.

Lemma weaken_subst_evar {Γ1 Γ2 : Env} {Δ : Ext} {x} {t} :
  subst_evar x t Γ1 Γ2 →
  subst_evar (weakenSubst (lengthExt Δ) x) t (extend Γ1 Δ) (extend Γ2 Δ).
Proof.
  revert x Γ1 Γ2.
  induction Δ; simpl; intros; try constructor; auto.
Qed.

Lemma subst_evar_lookup_etvar {Γ1 Γ2 : Env} {x t Y} :
  subst_evar x t Γ1 Γ2 → lookup_etvar Γ1 Y → lookup_etvar Γ2 Y.
Proof.
  intros sube gv; revert Y gv.
  induction sube; intros; dependent destruction gv; simpl;
    try constructor; auto.
Qed.

Lemma subst_evar_wftyp_ind {Γ1 Γ2 : Env} {x t T} :
  subst_evar x t Γ1 Γ2 → wfTy Γ1 T → wfTy Γ2 T.
Proof.
  intros esub st; revert Γ2 x t esub.
  induction st; simpl; eauto using subst_evar_lookup_etvar, wfTy.
Qed.