Require Export Coq.Unicode.Utf8.
Require Export Coq.Program.Equality.
Require Export Coq.Program.Tactics.

(******************************************************************************)
(* Representation of types.                                                   *)
(******************************************************************************)

Definition Lab := nat.

Inductive Ty : Set :=
  | tvar (X: nat)
  | top
  | tarr (T1 T2: Ty)
  | tall (T1 T2: Ty)
  | trec (Tr : TRec)
with TRec : Set :=
  | tnil
  | tcons (l: Lab) (T: Ty) (Tr: TRec).

Scheme ind_Ty := Induction for Ty Sort Prop
with ind_TRec := Induction for TRec Sort Prop.
Combined Scheme ind_TRec_Ty from ind_Ty, ind_TRec.

(******************************************************************************)
(* Representation of terms.                                                   *)
(******************************************************************************)

Inductive Pat : Set :=
  | pvar (T : Ty)
  | prec (ps : PRec)
with PRec : Set :=
  | pnil
  | pcons (l : Lab) (p : Pat) (ps : PRec).
Scheme ind_Pat := Induction for Pat Sort Prop
with ind_PRec := Induction for PRec Sort Prop.
Combined Scheme ind_Pat_PRec from ind_Pat, ind_PRec.

Inductive Tm : Set :=
  | var (x: nat)
  | abs (T : Ty) (t : Tm)
  | app (t1 : Tm) (t2 : Tm)
  | tabs (T : Ty) (t : Tm)
  | tapp (t : Tm) (T : Ty)
  | rec (ts : Rec)
  | prj (t : Tm) (l : Lab)
  | lett (p : Pat) (t1 : Tm)
      (t2 : Tm)
with Rec : Set :=
  | rnil
  | rcons (l : Lab) (t : Tm)
      (ts : Rec).
Scheme ind_Tm := Induction for Tm Sort Prop
with ind_Rec := Induction for Rec Sort Prop.
Combined Scheme ind_Tm_Rec from ind_Tm, ind_Rec.

(* Calculates the number of cariables bound in a pattern. *)
Fixpoint bindPat (p : Pat) {struct p} : nat :=
  match p with
    | pvar T  => 1
    | prec ps => bindPRec ps
  end
with bindPRec (ps : PRec) {struct ps} : nat :=
  match ps with
    | pnil         => 0
    | pcons l p ps => bindPRec ps + bindPat p
  end.

Inductive Env : Set :=
  | empty
  | evar   (Γ : Env) (T : Ty)
  | etvar  (Γ : Env) (T : Ty).

(* A context extension holding the types of pattern variables. *)
Inductive Ext : Set :=
  | exempty
  | exvar   (Δ : Ext) (T : Ty).
