Set Implicit Arguments.

Require Import Prelude.
Require Import Specification.
Require Import DeBruijn.

Module Interaction (Spec : Specification).

  Module DEBRUIJN := DEBRUIJN Spec.
  Export DEBRUIJN.

  Ltac terminduction :=
    match goal with
      | |- forall s (t : Term s), _ =>
        simpl; intros s t; induction t; nsimpl;
        [ idtac
        | apply (f_equal (@Fix _));
            extensionality z;
            extensionality zΓ;
          match goal with
            | zΓ: ?z ∈ ctorFields ?ct,
              ts: ∀ s: Sort, s ∈ ctorFields ?ct → Term s,
              IH: forall (s: Sort) (sΓ: s ∈ _), _ |- _ =>
                generalize (ctorBindspec _ _ zΓ) as bs; intro;
                specialize (IH z zΓ); revert IH;
                generalize (ts z zΓ) as t; intros
          end
        ]
    end.

  Section LemmaShiftInj.

    Lemma shiftIndex_inj {α} :
      ∀ (c : Cutoff α) (i1 i2 : Index α),
        shiftIndex c i1 = shiftIndex c i2 -> i1 = i2.
    Proof.
      induction c; simpl; intros.
      - inversion H; auto.
      - destruct i1, i2; simpl; try congruence.
        inversion H.
        apply IHc in H1.
        congruence.
    Qed.

    (* Lemma shift_inj {α} : *)
    (*   ∀ (c : Cutoff α) {s} (αIn: α ∈ depsOf s) (t1 t2: Term s), *)
    (*     shift αIn c t1 = shift αIn c t2 -> t1 = t2. *)
    (* Proof. *)
    (* Admitted. *)

  End LemmaShiftInj.

  Arguments Eqh_evalbindspec_shift {Γ} α {cs ts} h {αs} bs.
  Arguments Eqh_evalbindspec_subst {Γ} α {xs tα ts} h {αs} bs.
  Arguments EqcRefl {Γ ts} α {c}.
  Arguments EqcWeakenShift {Γ ts} α c1 c2 h.
  Arguments EqcWeakenShift' {Γ ts} α c1 c2 h.
  Arguments EqcWeakenAppend {Γ ts α} c h1 h2.
  Arguments EqtCongShift {Γ ts α s c1 c2 αIn t1 t2} _ _.
  Arguments EqtCongSubst {Γ ts α s x1 x2 αIn tα1 tα2 t1 t2} _ _ _.
  Arguments EqxWeakenAppend {Γ ts α} x h1 h2.
  Arguments EqtRefl {Γ ts} s {t}.
  Arguments EqxRefl {Γ ts} α {x}.

  Section LemmaShiftShiftComm.

    Lemma shiftIndex_shiftIndex_comm {α} :
      ∀ (c1 c2 : Cutoff α) (i : Index α),
        shiftIndex (shiftCutoff  c1 c2) (shiftIndex c1 i) =
        shiftIndex (shiftCutoff2 c2 c1) (shiftIndex c2 i).
    Proof.
      intro c1; induction c1 as [|c1]; intros c2 i;
      [ reflexivity
      | destruct c2 as [|c2];
        [ reflexivity
        | destruct i;
          [ reflexivity
          | exact (f_equal IS (IHc1 c2 i))
          ]
        ]
      ].
    Qed.

    Lemma shift_shift_comm_ind {α} :
      (∀ s (t : Term s) (αIn: α ∈ depsOf s) (c1 c2 : Cutoff α),
         shift αIn (shiftCutoff c1 c2)  (shift αIn c1 t) =
         shift αIn (shiftCutoff2 c2 c1) (shift αIn c2 t)).
    Proof.
      terminduction.
      - ncrush.
        + apply (f_equal (@Var _)).
          apply shiftIndex_shiftIndex_comm.
        + reflexivity.
      - ncrush.
        + eapply EqTerm_sound.
          eapply EqtTrans.
          eapply EqtCongShift.
          eapply EqcTrans.
          eapply EqcCongWeaken.
          eapply EqcRefl.
          eapply Eqh_evalbindspec_shift.
          eapply EqcWeakenShift.
          eapply EqtRefl.
          eapply EqtTrans.
          eapply EqtAssumption.
          eapply H.
          eapply EqtSym.
          eapply EqtCongShift.
          eapply EqcTrans.
          eapply EqcCongWeaken.
          eapply EqcRefl.
          eapply Eqh_evalbindspec_shift.
          eapply EqcWeakenShift'.
          eapply EqtRefl.
        + reflexivity.
    Defined.

    (* Print shift_shift_comm_ind. *)

    Lemma shift_shift_comm_het_ind {α β} (αβ : α ≢ β) :
      (∀ (s : Sort) (t : Term s)
         (αIn : α ∈ depsOf s) (βIn : β ∈ depsOf s)
         (c1 : Cutoff α) (c2 : Cutoff β),
           shift βIn c2 (shift αIn c1 t) =
           shift αIn c1 (shift βIn c2 t)).
    Proof.
      terminduction.
      - ncrush; reflexivity.
      - ncrush.
        + eapply EqTerm_sound.
          eapply EqtTrans.
          eapply EqtCongShift.
          eapply EqcCongWeaken.
          eapply EqcRefl.
          eapply Eqh_evalbindspec_shift.
          eapply EqtRefl.
          eapply EqtTrans.
          eapply EqtAssumption.
          eapply H.
          eapply EqtSym.
          eapply EqtCongShift.
          eapply EqcCongWeaken.
          eapply EqcRefl.
          eapply Eqh_evalbindspec_shift.
          eapply EqtRefl.
        + eapply EqTerm_sound.
          eapply EqtCongShift.
          eapply EqcCongWeaken.
          eapply EqcRefl.
          eapply Eqh_evalbindspec_shift.
          eapply EqtRefl.
        + eapply EqTerm_sound.
          eapply EqtCongShift.
          eapply EqcSym.
          eapply EqcCongWeaken.
          eapply EqcRefl.
          eapply Eqh_evalbindspec_shift.
          eapply EqtRefl.
        + reflexivity.
    Defined.

    Lemma shiftIndex_shiftIndex0_hom_comm_ind:
      ∀ (α : Namespace) (h : HVL) (c : Cutoff α) (y : Index α),
        shiftIndex (weaken (CS c) h) (shiftIndex (weaken C0 h) y) =
        shiftIndex (weaken C0 h) (shiftIndex (weaken c h) y).
    Proof.
      induction h; intros c y.
      - reflexivity.
      - ncrush.
        + destruct y; simpl.
          reflexivity.
          f_equal; apply IHh.
        + apply IHh.
    Qed.

    Lemma shift_shift0_hom_comm_ind {α} :
      (∀ s (t : Term s) (αIn : α ∈ depsOf s) (h: HVL) (c : Cutoff α),
         shift αIn (weaken (CS c) h) (shift αIn (weaken C0 h) t) =
         shift αIn (weaken C0     h) (shift αIn (weaken c  h) t)).
    Proof.
      terminduction.
      - ncrush.
        + f_equal.
          apply shiftIndex_shiftIndex0_hom_comm_ind.
        + reflexivity.
      - ncrush.
        + eapply EqTerm_sound.
          eapply EqtTrans.
          eapply EqtCongShift.
          eapply EqcTrans.
          eapply EqcCongWeaken.
          eapply EqcRefl.
          eapply Eqh_evalbindspec_shift.
          eapply EqcWeakenAppend.
          eapply EqtCongShift.
          eapply EqcWeakenAppend.
          eapply EqtRefl.
          eapply EqtTrans.
          eapply EqtAssumption.
          eapply H.
          eapply EqtSym.
          eapply EqtCongShift.
          eapply EqcTrans.
          eapply EqcCongWeaken.
          eapply EqcRefl.
          eapply Eqh_evalbindspec_shift.
          eapply EqcWeakenAppend.
          eapply EqtCongShift.
          eapply EqcWeakenAppend.
          eapply EqtRefl.
        + reflexivity.
    Defined.

    (* Print shift_shift0_hom_comm_ind. *)

    Lemma shift_shift0_hom_comm {α} :
      (∀ s (t : Term s) (αIn : α ∈ depsOf s) (c : Cutoff α),
         shift αIn (CS c) (shift αIn C0 t) =
         shift αIn C0     (shift αIn c  t)).
    Proof.
      intros; eapply (shift_shift0_hom_comm_ind _ _ HV0).
    Qed.

    Lemma shift_shift0_het_comm_ind {α β} (αβ: α ≢ β) :
      (∀ s (t : Term s) (αIn : α ∈ depsOf s) (βIn: β ∈ depsOf s)
         (h: HVL) (c : Cutoff α),
         shift αIn (weaken c  h) (shift βIn (weaken C0 h) t) =
         shift βIn (weaken C0 h) (shift αIn (weaken c  h) t)).
    Proof.
      terminduction.
      - ncrush; reflexivity.
      - destruct (α ∈? depsOf z) as [αInz|αNotInz],
                 (β ∈? depsOf z) as [βInz|βNotInz]; nsimpl.
        + eapply EqTerm_sound.
          eapply EqtTrans.
          eapply EqtCongShift.
          eapply EqcTrans.
          eapply EqcCongWeaken.
          eapply EqcRefl.
          eapply Eqh_evalbindspec_shift.
          eapply EqcWeakenAppend.
          eapply EqtCongShift.
          eapply EqcWeakenAppend.
          eapply EqtRefl.
          eapply EqtTrans.
          eapply EqtAssumption.
          eapply H.
          eapply EqtSym.
          eapply EqtCongShift.
          eapply EqcTrans.
          eapply EqcCongWeaken.
          eapply EqcRefl.
          eapply Eqh_evalbindspec_shift.
          eapply EqcWeakenAppend.
          eapply EqtCongShift.
          eapply EqcWeakenAppend.
          eapply EqtRefl.
        + eapply EqTerm_sound.
          eapply EqtCongShift.
          eapply EqcCongWeaken.
          eapply EqcRefl.
          eapply Eqh_evalbindspec_shift.
          eapply EqtRefl.
        + eapply EqTerm_sound.
          eapply EqtCongShift.
          eapply EqcCongWeaken.
          eapply EqcRefl.
          eapply EqhSym.
          eapply Eqh_evalbindspec_shift.
          eapply EqtRefl.
        + reflexivity.
    Defined.

    (* Print shift_shift0_het_comm_ind. *)

    Lemma shift_shift0_het_comm {α β} (αβ: α ≢ β) :
      (∀ s (t : Term s) (αIn : α ∈ depsOf s) (βIn: β ∈ depsOf s)
         (c : Cutoff α),
         shift αIn c  (shift βIn C0 t) =
         shift βIn C0 (shift αIn c  t)).
    Proof.
      intros; apply (shift_shift0_het_comm_ind αβ _ _ _ HV0).
    Qed.

    Lemma weaken_shift_comm {α s} (αIn : α ∈ depsOf s) (t: Term s) (h : HVL) :
      ∀ (c: Cutoff α),
        weaken (shift αIn c t) h =
        shift αIn (weaken c h) (weaken t h).
    Proof.
      induction h; simpl; intros.
      - reflexivity.
      - ncrush.
        + rewrite shift_shift0_hom_comm.
          f_equal; eauto.
        + rewrite shift_shift0_het_comm; auto.
          f_equal; eauto.
        + eauto.
    Qed.

  End LemmaShiftShiftComm.

  Section SubstShiftReflection.

    Lemma substIndex_shiftIndex_reflection_ind {α} (sub : Term (sortOf α)) :
      ∀ hvl y,
        substIndex (weaken T0 hvl) sub (shiftIndex (weakenCutoff C0 hvl) y) =
        Var y.
    Proof.
      intro hvl; induction hvl as [|β hvl]; nsimpl.
      - reflexivity.
      - ncrush.
        + destruct y; simpl.
          reflexivity.
          rewrite IHhvl; nsimpl; auto.
        + rewrite IHhvl; nsimpl; auto.
        + auto.
    Qed.

    Lemma subst_shift_reflection_ind {α} :
      (∀ (s : Sort) (t : Term s) (αIn : α ∈ depsOf s)
         (hvl : HVL) (sub : Term (sortOf α)),
         subst αIn
           (weaken T0 hvl) sub
           (shift αIn (weaken C0 hvl) t) = t).
    Proof.
      terminduction.
      - ncrush.
        + apply substIndex_shiftIndex_reflection_ind.
        + reflexivity.
      - ncrush.
        + eapply EqTerm_sound.
          eapply EqtTrans.
          eapply EqtCongSubst.
          eapply EqxTrans.
          eapply EqxCongWeaken.
          eapply EqxRefl.
          eapply Eqh_evalbindspec_shift.
          eapply EqxWeakenAppend.
          eapply EqtRefl.
          eapply EqtCongShift.
          eapply EqcWeakenAppend.
          eapply EqtRefl.
          eapply EqtAssumption.
          eapply H.
        + reflexivity.
    Qed.

    Lemma subst0_shift0_reflection {α}
          (s : Sort) (αIn : α ∈ depsOf s)
          (t' : Term (sortOf α)) (t : Term s) :
      subst αIn T0 t' (shift αIn C0 t) = t.
    Proof.
      apply (subst_shift_reflection_ind _ _ HV0).
    Qed.

    Lemma substTerms0_shiftTerms0_reflection {α Γ} (tα: Term (sortOf α)) {ts: Terms Γ} :
      substTerms (Xs T0) tα (shiftTerms (Cs (C0 (α := α))) ts) = ts.
    Proof.
      extensionality z; extensionality zΓ.
      unfold shiftTerms, shift', substTerms, subst'.
      ncrush; eauto using subst0_shift0_reflection.
    Qed.

  End SubstShiftReflection.


  Section ShiftSubst.

    Context {α: Namespace}.

    Lemma shift_substIndex0_hom_comm_ind (c : Cutoff α) (tα : Term (sortOf α)) :
      ∀ (hvl : HVL) (y : Index α),
        shift (namespaceDepVar α)
          (weaken c hvl)
          (substIndex (weaken T0 hvl) tα y) =
        substIndex
          (weaken T0 hvl)
          (shift (namespaceDepVar α) c tα)
          (shiftIndex (weaken (CS c) hvl) y).
    Proof.
      intro hvl; induction hvl as [|β hvl]; nsimpl.
      - destruct y; nsimpl; reflexivity.
      - ncrush.
        + destruct y; nsimpl.
          reflexivity.
          rewrite shift_shift0_hom_comm.
          f_equal; auto.
        + rewrite shift_shift0_het_comm; auto.
          f_equal; auto.
        + auto.
    Qed.

    Lemma shift_subst0_hom_comm_ind :
      (∀ (s : Sort) (t : Term s) (αIn : α ∈ depsOf s)
             (hvl : HVL) (c : Cutoff α) (sub : Term (sortOf α)),
        shift αIn (weaken c hvl)
              (subst αIn (weaken T0 hvl) sub t) =
        subst αIn (weaken T0 hvl) (shift (namespaceDepVar α) c sub)
              (shift αIn (weaken (CS c) hvl) t)).
    Proof.
      terminduction.
      - ncrush.
        apply shift_substIndex0_hom_comm_ind.
        reflexivity.
      - ncrush.
        + eapply EqTerm_sound.
          eapply EqtTrans.
          eapply EqtCongShift.
          eapply EqcTrans.
          eapply EqcCongWeaken.
          eapply EqcRefl.
          eapply Eqh_evalbindspec_subst.
          eapply EqcWeakenAppend.
          eapply EqtCongSubst.
          eapply EqxWeakenAppend.
          eapply EqtRefl.
          eapply EqtRefl.
          eapply EqtTrans.
          eapply EqtAssumption.
          eapply H.
          eapply EqtSym.
          eapply EqtCongSubst.
          eapply EqxTrans.
          eapply EqxCongWeaken.
          eapply EqxRefl.
          eapply Eqh_evalbindspec_shift.
          eapply EqxWeakenAppend.
          eapply EqtRefl.
          eapply EqtCongShift.
          eapply EqcWeakenAppend.
          eapply EqtRefl.
        + reflexivity.
    Defined.

    (* Print shift_subst0_hom_comm_ind. *)

    Lemma shift_subst0_hom_comm :
      ∀ (s : Sort) (t : Term s) (αIn : α ∈ depsOf s)
        (c : Cutoff α) (sub : Term (sortOf α)),
        shift αIn c (subst αIn T0 sub t) =
        subst αIn T0 (shift (namespaceDepVar α) c sub)
              (shift αIn (CS c) t).
    Proof.
      intros; apply (shift_subst0_hom_comm_ind _ _ HV0).
    Qed.

    Context {β: Namespace}.
    Variable αβ : α ≢ β.

    Section AlphaInBeta.

      Variable αInβ : α ∈ depsOf (sortOf β).

      Lemma shift_substIndex0_het_comm1_ind
        (cα : Cutoff α) (tβ : Term (sortOf β)) :
        ∀ hvl (y : Index β),
          shift αInβ (weaken cα hvl) (substIndex (weaken T0 hvl) tβ y) =
          substIndex (weaken T0 hvl) (shift αInβ cα tβ) y.
      Proof.
        intro hvl; induction hvl as [|γ hvl]; nsimpl.
        - destruct y; nsimpl; reflexivity.
        - ncrush.
          + rewrite shift_shift0_hom_comm; f_equal; auto.
          + destruct y; ncrush.
            reflexivity.
            rewrite shift_shift0_het_comm; auto.
            f_equal; auto.
          + rewrite shift_shift0_het_comm; auto.
            f_equal; auto.
          + auto.
      Qed.

      Lemma shift_subst0_het_comm1_ind :
        (∀ {s} (t : Term s)
           (αIn : α ∈ depsOf s)
           (βIn : β ∈ depsOf s)
           (hvl : HVL) (cα : Cutoff α) (tβ : Term (sortOf β)),
           shift αIn (weaken cα hvl)
                 (subst βIn (weaken T0 hvl) tβ t) =
           subst βIn (weaken T0 hvl)
                 (shift αInβ cα tβ)
                 (shift αIn (weaken cα hvl) t)).
      Proof.
        terminduction.
        - ncrush.
          apply shift_substIndex0_het_comm1_ind.
          reflexivity.
          reflexivity.
        - ncrush.
          + eapply EqTerm_sound.
            eapply EqtTrans.
            eapply EqtCongShift.
            eapply EqcTrans.
            eapply EqcCongWeaken.
            eapply EqcRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqcWeakenAppend.
            eapply EqtCongSubst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtRefl.
            eapply EqtTrans.
            eapply EqtAssumption.
            eapply H.
            eapply EqtCongSubst.
            eapply EqxSym.
            eapply EqxTrans.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_shift.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtCongShift.
            eapply EqcSym.
            eapply EqcWeakenAppend.
            eapply EqtRefl.
          + eapply EqTerm_sound.
            eapply EqtCongShift.
            eapply EqcCongWeaken.
            eapply EqcRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqtRefl.
          + reflexivity.
      Defined.

      (* Print shift_subst0_het_comm1_ind. *)

      Lemma shift_subst0_het_comm1 :
        (∀ {s} (t : Term s)
           (αIn : α ∈ depsOf s)
           (βIn : β ∈ depsOf s)
           (cα : Cutoff α) (tβ : Term _),
           shift αIn cα (subst βIn T0 tβ t) =
           subst βIn T0 (shift αInβ cα tβ) (shift αIn cα t)).
      Proof.
        intros; apply (shift_subst0_het_comm1_ind _ _ _ HV0).
      Qed.

    End AlphaInBeta.

    Section AlphaNotInBeta.

      Variable αNotInβ : α ∉ depsOf (sortOf β).

      Lemma shift_subst_het_comm2 :
        (∀ {s} (t : Term s)
           (αIn : α ∈ depsOf s)
           (βIn : β ∈ depsOf s)
           (cα : Cutoff α) (xβ : Trace β) (tβ : Term _),
           shift αIn cα (subst βIn xβ tβ t) =
           subst βIn xβ tβ (shift αIn cα t)).
      Proof.
        terminduction.
        - ncrush; reflexivity.
        - ncrush.
          + eapply EqTerm_sound.
            eapply EqtTrans.
            eapply EqtCongShift.
            eapply EqcCongWeaken.
            eapply EqcRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqtRefl.
            eapply EqtTrans.
            eapply EqtAssumption.
            eapply H.
            eapply EqtCongSubst.
            eapply EqxSym.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_shift.
            eapply EqtRefl.
            eapply EqtRefl.
          + eapply EqTerm_sound.
            eapply EqtCongShift.
            eapply EqcCongWeaken.
            eapply EqcRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqtRefl.
          + eapply EqTerm_sound.
            eapply EqtCongSubst.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply EqhSym.
            eapply Eqh_evalbindspec_shift.
            eapply EqtRefl.
            eapply EqtRefl.
          + reflexivity.
      Defined.

      (* Print shift_subst_het_comm2. *)

      Lemma shift_subst0_het_comm2_ind :
        (∀ {s} (t : Term s)
           (αIn : α ∈ depsOf s)
           (βIn : β ∈ depsOf s)
           (hvl : HVL) (cα : Cutoff α) (tβ : Term (sortOf β)),
           shift αIn (weaken cα hvl) (subst βIn (weaken T0 hvl) tβ t) =
           subst βIn (weaken T0 hvl) tβ (shift αIn (weaken cα hvl) t)).
      Proof.
        terminduction.
        - ncrush; reflexivity.
        - ncrush.
          + eapply EqTerm_sound.
            eapply EqtTrans.
            eapply EqtCongShift.
            eapply EqcTrans.
            eapply EqcCongWeaken.
            eapply EqcRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqcWeakenAppend.
            eapply EqtCongSubst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtRefl.
            eapply EqtTrans.
            eapply EqtAssumption.
            eapply H.
            eapply EqtSym.
            eapply EqtTrans.
            eapply EqtCongSubst.
            eapply EqxTrans.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_shift.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtCongShift.
            eapply EqcWeakenAppend.
            eapply EqtRefl.
            eapply EqtRefl.
          + eapply EqTerm_sound.
            eapply EqtCongShift.
            eapply EqcCongWeaken.
            eapply EqcRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqtRefl.
          + eapply EqTerm_sound.
            eapply EqtCongSubst.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply EqhSym.
            eapply Eqh_evalbindspec_shift.
            eapply EqtRefl.
            eapply EqtRefl.
          + reflexivity.
      Defined.

      (* Print shift_subst0_het_comm2_ind. *)

      Lemma shift_subst0_het_comm2 :
        (∀ {s} (t : Term s)
           (αIn : α ∈ depsOf s)
           (βIn : β ∈ depsOf s)
           (cα : Cutoff α) (tβ : Term _),
           shift αIn cα (subst βIn T0 tβ t) =
           subst βIn T0 tβ (shift αIn cα t)).
      Proof.
        intros; apply (shift_subst0_het_comm2_ind _ _ _ HV0).
      Qed.

    End AlphaNotInBeta.

  End ShiftSubst.

  Section SubstShift.

    Lemma substIndex_shiftIndex0_comm_ind {α}
      (x : Trace α) (sub : Term (sortOf α)) :
      ∀ (hvl : HVL) y ,
        substIndex (weaken (TS (namespaceDepVar α) x) hvl) sub
                   (shiftIndex (weaken C0 hvl) y) =
        shift (namespaceDepVar α) (weaken C0 hvl)
              (substIndex (weaken x hvl) sub y).
    Proof.
      intro hvl; induction hvl as [|β hvl]; nsimpl.
      - reflexivity.
      - ncrush.
        + destruct y; nsimpl.
          reflexivity.
          rewrite shift_shift0_hom_comm.
          f_equal; auto.
        + rewrite shift_shift0_het_comm; auto.
          f_equal; auto.
        + auto.
    Qed.

    Lemma subst_shift0_comm_ind {α} :
      (∀ (s : Sort) (t : Term s) (αIn : α ∈ depsOf s)
             (hvl : HVL) (x : Trace α) (sub : Term (sortOf α)),
         subst αIn (weaken (TS (namespaceDepVar α) x) hvl) sub
               (shift αIn (weaken C0 hvl) t) =
         shift αIn (weaken C0 hvl)
               (subst αIn (weaken x hvl) sub t)).
    Proof.
      terminduction.
      - ncrush.
        + apply substIndex_shiftIndex0_comm_ind.
        + reflexivity.
      - ncrush.
        + eapply EqTerm_sound.
          eapply EqtTrans.
          eapply EqtCongSubst.
          eapply EqxTrans.
          eapply EqxCongWeaken.
          eapply EqxRefl.
          eapply Eqh_evalbindspec_shift.
          eapply EqxWeakenAppend.
          eapply EqtRefl.
          eapply EqtCongShift.
          eapply EqcWeakenAppend.
          eapply EqtRefl.
          eapply EqtTrans.
          eapply EqtAssumption.
          eapply H.
          eapply EqtSym.
          eapply EqtCongShift.
          eapply EqcTrans.
          eapply EqcCongWeaken.
          eapply EqcRefl.
          eapply Eqh_evalbindspec_subst.
          eapply EqcWeakenAppend.
          eapply EqtCongSubst.
          eapply EqxWeakenAppend.
          eapply EqtRefl.
          eapply EqtRefl.
        + reflexivity.
    Defined.

    (* Print subst_shift0_comm_ind. *)

    Lemma subst_shift0_comm {α} :
      ∀ (s : Sort) (t : Term s) (αIn : α ∈ depsOf s)
        (x : Trace α) (sub : Term (sortOf α)),
        subst αIn (TS (namespaceDepVar α) x) sub (shift αIn C0 t) =
        shift αIn C0 (subst αIn x sub t).
    Proof.
      intros; apply (subst_shift0_comm_ind _ _ HV0).
    Qed.

  End SubstShift.

  Section SubstShiftDistinct.

    Context {α β : Namespace}.
    Variable αβ : α ≢ β.

    Section AlphaInBeta.

      Variable αInβ : α ∈ depsOf (sortOf β).

      Lemma substIndex_shiftIndex0_het_comm1_ind
        (xβ : Trace β) (tβ : Term (sortOf β)) :
        ∀ (hvl : HVL) (y : Index β),
          substIndex (weaken (TS αInβ xβ) hvl) tβ y =
          shift αInβ (weaken C0 hvl) (substIndex (weaken xβ hvl) tβ y).
      Proof.
        intro hvl; induction hvl as [|γ vl]; nsimpl.
        - reflexivity.
        - ncrush.
          + rewrite shift_shift0_hom_comm.
            f_equal; auto.
          + destruct y; nsimpl.
            * reflexivity.
            * rewrite shift_shift0_het_comm; auto.
              f_equal; auto.
          + rewrite shift_shift0_het_comm; auto.
            f_equal; auto.
          + auto.
      Qed.

      Lemma substIndex_shiftIndex0_het_comm1
        (xβ : Trace β) (tβ : Term (sortOf β)) :
        ∀ (y : Index β),
          substIndex (TS αInβ xβ) tβ y =
          shift αInβ C0 (substIndex xβ tβ y).
      Proof.
        apply (substIndex_shiftIndex0_het_comm1_ind _ _ HV0).
      Qed.

      Lemma subst_shift0_het_comm1_ind :
        (∀ (s : Sort) (t : Term s)
           (αIn : α ∈ depsOf s)
           (βIn : β ∈ depsOf s)
           (hvl : HVL) (xβ : Trace β) (tβ : Term (sortOf β)),
           subst βIn (weaken (TS αInβ xβ) hvl) tβ
                 (shift αIn (weaken C0 hvl) t) =
           shift αIn (weaken C0 hvl)
                 (subst βIn (weaken xβ hvl) tβ t)).
      Proof.
        terminduction.
        - ncrush.
          + apply substIndex_shiftIndex0_het_comm1_ind.
          + reflexivity.
          + reflexivity.
        - ncrush.
          + eapply EqTerm_sound.
            eapply EqtTrans.
            eapply EqtCongSubst.
            eapply EqxTrans.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_shift.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtCongShift.
            eapply EqcWeakenAppend.
            eapply EqtRefl.
            eapply EqtTrans.
            eapply EqtAssumption.
            eapply H.
            eapply EqtSym.
            eapply EqtCongShift.
            eapply EqcTrans.
            eapply EqcCongWeaken.
            eapply EqcRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqcWeakenAppend.
            eapply EqtCongSubst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtRefl.
          + eapply EqTerm_sound.
            eapply EqtCongShift.
            eapply EqcCongWeaken.
            eapply EqcRefl.
            eapply EqhSym.
            eapply Eqh_evalbindspec_subst.
            eapply EqtRefl.
          + reflexivity.
      Qed.

      Lemma subst_shift0_het_comm1 :
        (∀ (s : Sort) (t : Term s)
           (αIn : α ∈ depsOf s)
           (βIn : β ∈ depsOf s)
           (xβ : Trace β) (tβ : Term (sortOf β)),
           subst βIn (TS αInβ xβ) tβ
                 (shift αIn C0 t) =
           shift αIn C0
                 (subst βIn xβ tβ t)).
      Proof.
        intros; apply (subst_shift0_het_comm1_ind _ _ _ HV0).
      Qed.

    End AlphaInBeta.

    Section AlphaNotInBeta.

      Variable αNotInβ : α ∉ depsOf (sortOf β).

      Lemma subst_shift0_het_comm2_ind :
        (∀ (s : Sort) (t : Term s)
           (αIn : α ∈ depsOf s)
           (βIn : β ∈ depsOf s)
           (hvl : HVL) (xβ : Trace β) (tβ : Term (sortOf β)),
           subst βIn (weaken xβ hvl) tβ
                 (shift αIn (weaken C0 hvl) t) =
           shift αIn (weaken C0 hvl)
                 (subst βIn (weaken xβ hvl) tβ t)).
      Proof.
        terminduction.
        - ncrush.
          + reflexivity.
          + reflexivity.
        - ncrush.
          + eapply EqTerm_sound.
            eapply EqtTrans.
            eapply EqtCongSubst.
            eapply EqxTrans.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_shift.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtCongShift.
            eapply EqcWeakenAppend.
            eapply EqtRefl.
            eapply EqtTrans.
            eapply EqtAssumption.
            eapply H.
            eapply EqtSym.
            eapply EqtCongShift.
            eapply EqcTrans.
            eapply EqcCongWeaken.
            eapply EqcRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqcWeakenAppend.
            eapply EqtCongSubst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtRefl.
          + eapply EqTerm_sound.
            eapply EqtCongSubst.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_shift.
            eapply EqtRefl.
            eapply EqtRefl.
          + eapply EqTerm_sound.
            eapply EqtCongShift.
            eapply EqcCongWeaken.
            eapply EqcRefl.
            eapply EqhSym.
            eapply Eqh_evalbindspec_subst.
            eapply EqtRefl.
          + reflexivity.
      Qed.

      Lemma subst_shift0_het_comm2 :
        (∀ (s : Sort) (t : Term s)
           (αIn : α ∈ depsOf s)
           (βIn : β ∈ depsOf s)
           (xβ : Trace β) (tβ : Term (sortOf β)),
           subst βIn xβ tβ
                 (shift αIn C0 t) =
           shift αIn C0
                 (subst βIn xβ tβ t)).
      Proof.
        intros; apply (subst_shift0_het_comm2_ind _ _ _ HV0).
      Qed.

    End AlphaNotInBeta.

  End SubstShiftDistinct.

  Lemma substTerms_shiftTerms0_comm1 {α β Γ} (αIn: α ∈ depsOf (sortOf β))
    (x: Trace β) (tβ: Term (sortOf β)) (ts: Terms Γ) :
    substTerms (Xs (TS αIn x)) tβ (shiftTerms (Cs (C0 (α := α))) ts) =
    shiftTerms (Cs (C0 (α := α))) (substTerms (Xs x) tβ ts).
  Proof.
    extensionality z; extensionality zΓ.
    unfold shiftTerms, shift', substTerms, subst'.
    destruct (α ≟ β); ncrush;
      eauto using subst_shift0_comm, subst_shift0_het_comm1.
  Qed.

  Lemma substTerms_shiftTerms0_comm2 {α β Γ} (αNotIn: α ∉ depsOf (sortOf β))
    (x: Trace β) (tβ: Term (sortOf β)) (ts: Terms Γ) :
    substTerms (Xs x) tβ (shiftTerms (Cs (C0 (α := α))) ts) =
    shiftTerms (Cs (C0 (α := α))) (substTerms (Xs x) tβ ts).
  Proof.
    extensionality z; extensionality zΓ.
    unfold shiftTerms, shift', substTerms, subst'.
    destruct (α ≟ β); ncrush;
      eauto using subst_shift0_comm, subst_shift0_het_comm2.
  Qed.

  Section WeakenSubst.

    Lemma weaken_subst_comm {α s} (αIn : α ∈ depsOf s) (t: Term s) (h : HVL) :
      ∀ (xα: Trace α) (tα: Term (sortOf α)),
        weaken (subst αIn xα tα t) h =
        subst αIn (weaken xα h) tα (weaken t h).
    Proof.
      induction h as [|β hvl]; simpl; intros.
      - reflexivity.
      - destruct (α ≟ β); ncrush.
        + rewrite subst_shift0_comm.
          f_equal; eauto.
        + rewrite subst_shift0_het_comm1; nsimpl.
          f_equal; eauto.
        + rewrite subst_shift0_het_comm2; nsimpl.
          f_equal; eauto.
        + eauto.
    Qed.

  End WeakenSubst.

  Section SubstSubst.

    Context {α : Namespace}.

    Lemma subst_substIndex0_comm_ind
      (x : Trace α) (sub1 sub2 : Term (sortOf α)) :
      ∀ (hvl : HVL) (y : Index α),
        subst (namespaceDepVar α) (weaken x hvl) sub1
              (substIndex (weaken T0 hvl) sub2 y) =
        subst (namespaceDepVar α) (weaken T0 hvl)
              (subst (namespaceDepVar α) x sub1 sub2)
              (substIndex (weaken (TS (namespaceDepVar α) x) hvl) sub1 y).
    Proof.
      intro hvl; induction hvl as [|β hvl]; nsimpl.
      - destruct y; nsimpl.
        + reflexivity.
        + rewrite subst0_shift0_reflection.
          reflexivity.
      - ncrush.
        + destruct y; nsimpl.
          * reflexivity.
          * rewrite subst_shift0_comm.
            rewrite subst_shift0_comm.
            f_equal; auto.
        + destruct (β ∈? depsOf (sortOf α)); nsimpl.
          * rewrite subst_shift0_het_comm1; nsimpl; auto.
            rewrite subst_shift0_het_comm1; nsimpl; auto.
            f_equal; auto.
        + auto.
    Qed.

    Lemma subst_subst0_comm_ind :
      (∀ s (t : Term s) (αIn : α ∈ depsOf s)
         (hvl : HVL) (x : Trace α) (sub1 sub2 : Term (sortOf α)),
         subst αIn (weaken x hvl)  sub1                                    (subst αIn (weaken T0 hvl) sub2 t) =
         subst αIn (weaken T0 hvl) (subst (namespaceDepVar α) x sub1 sub2) (subst αIn (weaken (TS (namespaceDepVar α) x) hvl) sub1 t)).
    Proof.
      terminduction.
      - ncrush.
        + apply subst_substIndex0_comm_ind.
        + reflexivity.
      - ncrush.
        + eapply EqTerm_sound.
          eapply EqtTrans.
          eapply EqtCongSubst.
          eapply EqxTrans.
          eapply EqxCongWeaken.
          eapply EqxRefl.
          eapply Eqh_evalbindspec_subst.
          eapply EqxWeakenAppend.
          eapply EqtRefl.
          eapply EqtCongSubst.
          eapply EqxWeakenAppend.
          eapply EqtRefl.
          eapply EqtRefl.
          eapply EqtTrans.
          eapply EqtAssumption.
          eapply H.
          eapply EqtSym.
          eapply EqtCongSubst.
          eapply EqxTrans.
          eapply EqxCongWeaken.
          eapply EqxRefl.
          eapply Eqh_evalbindspec_subst.
          eapply EqxWeakenAppend.
          eapply EqtRefl.
          eapply EqtCongSubst.
          eapply EqxWeakenAppend.
          eapply EqtRefl.
          eapply EqtRefl.
        + reflexivity.
    Defined.

    Lemma subst_subst0_comm :
      (∀ s (t : Term s) (αIn : α ∈ depsOf s)
         (x : Trace α) (sub1 sub2 : Term (sortOf α)),
         subst αIn x sub1
               (subst αIn T0 sub2 t) =
         subst αIn T0
               (subst (namespaceDepVar α) x sub1 sub2)
               (subst αIn (TS (namespaceDepVar α) x) sub1 t)).
    Proof.
      intros; apply (subst_subst0_comm_ind _ _ HV0).
    Qed.

  End SubstSubst.

  Section SubstSubstDistinct.

    Context {α : Namespace}.
    Context {β : Namespace}.
    Variable αβ : α ≢ β.

    Section Variant1.

      Variable αNotInβ : α ∉ depsOf (sortOf β).
      Variable βNotInα : β ∉ depsOf (sortOf α).

      Lemma subst_subst_distinct_comm :
        (∀ s (t : Term s)
           (αIn : α ∈ depsOf s)
           (βIn : β ∈ depsOf s)
           (hvl : HVL)
           (xα : Trace α) (tα : Term (sortOf α))
           (xβ : Trace β) (tβ : Term (sortOf β)),
           subst αIn (weaken xα hvl) tα (subst βIn (weaken xβ hvl) tβ t) =
           subst βIn (weaken xβ hvl) tβ (subst αIn (weaken xα hvl) tα t)).
      Proof.
        terminduction.
        - ncrush.
          reflexivity.
        - ncrush.
          + eapply EqTerm_sound.
            eapply EqtTrans.
            eapply EqtCongSubst.
            eapply EqxTrans.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtCongSubst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtRefl.
            eapply EqtTrans.
            eapply EqtAssumption.
            eapply H.
            eapply EqtSym.
            eapply EqtCongSubst.
            eapply EqxTrans.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtCongSubst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtRefl.
          + eapply EqTerm_sound.
            eapply EqtCongSubst.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqtRefl.
            eapply EqtRefl.
          + eapply EqTerm_sound.
            eapply EqtSym.
            eapply EqtCongSubst.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqtRefl.
            eapply EqtRefl.
          + reflexivity.
      Defined.

    End Variant1.

    Section Variant2.

      Variable αInβ : α ∈ depsOf (sortOf β).
      Variable βNotInα : β ∉ depsOf (sortOf α).

      Lemma substαInβ_substIndexβ_distinct_comm_ind
        (xα1 xα2 : Trace α) (tα : Term (sortOf α))
        (xβ1 xβ2 : Trace β) (tβ : Term (sortOf β))
        (hyp : ∀ y,
                 subst αInβ xα1 tα (substIndex xβ1 tβ y) =
                 substIndex xβ2 (subst αInβ xα2 tα tβ) y) :
        ∀ hvl y,
          subst αInβ (weaken xα1 hvl) tα (substIndex (weaken xβ1 hvl) tβ y) =
          substIndex (weaken xβ2 hvl) (subst αInβ xα2 tα tβ) y.
      Proof.
        intro hvl; induction hvl as [|γ hvl]; nsimpl; auto.
        - destruct (α ≟ γ) as [αγ|αγ], (β ≟ γ) as [βγ|βγ]; ncrush.
          + rewrite subst_shift0_comm.
            f_equal; auto.
          + destruct y; nsimpl.
            reflexivity.
            rewrite subst_shift0_het_comm2; nsimpl; auto.
            f_equal; auto.
          + rewrite subst_shift0_het_comm1; nsimpl.
            f_equal; auto.
          + rewrite subst_shift0_het_comm2; nsimpl; auto.
            f_equal; auto.
          + auto.
      Qed.

      Lemma substαInβ_substIndexβ0_distinct_comm_ind
        (xα : Trace α) (tα : Term (sortOf α)) (tβ : Term (sortOf β)) :
        ∀ (hvl : HVL) (y : Index β),
          subst αInβ (weaken xα hvl) tα (substIndex (weaken T0 hvl) tβ y) =
          substIndex (weaken T0 hvl) (subst αInβ xα tα tβ) y.
      Proof.
        apply substαInβ_substIndexβ_distinct_comm_ind; destruct y; ncrush; auto.
      Qed.

      Lemma substαInβ_substβ0_distinct_comm_ind :
        (∀ s (t : Term s)
           (αIn : α ∈ depsOf s)
           (βIn : β ∈ depsOf s)
           (hvl : HVL)
           (xα : Trace α) (tα : Term (sortOf α)) (tβ : Term (sortOf β)),
           subst αIn (weaken xα hvl) tα                    (subst βIn (weaken T0 hvl) tβ t) =
           subst βIn (weaken T0 hvl) (subst αInβ xα tα tβ) (subst αIn (weaken xα hvl) tα t)).
      Proof.
        terminduction.
        - ncrush.
          + apply substαInβ_substIndexβ0_distinct_comm_ind; auto.
          + reflexivity.
        - ncrush.
          + eapply EqTerm_sound.
            eapply EqtTrans.
            eapply EqtCongSubst.
            eapply EqxTrans.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtCongSubst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtRefl.
            eapply EqtTrans.
            eapply EqtAssumption.
            eapply H.
            eapply EqtSym.
            eapply EqtCongSubst.
            eapply EqxTrans.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtCongSubst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtRefl.
          + eapply EqTerm_sound.
            eapply EqtCongSubst.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqtRefl.
            eapply EqtRefl.
          + reflexivity.
      Defined.

    End Variant2.

    Section Variant3.

      Variable αNotInβ : α ∉ depsOf (sortOf β).
      Variable βInα : β ∈ depsOf (sortOf α).

      Lemma subst_substIndex0_distinct_comm3_indα
        (xα : Trace α) (tα : Term (sortOf α)) (tβ : Term (sortOf β)) :
        ∀ (hvl : HVL) (y : Index α),
          substIndex (weaken xα hvl) tα y =
          subst βInα (weaken T0 hvl) tβ
                (substIndex (weaken (TS βInα xα) hvl) tα y).
      Proof.
        intro hvl; induction hvl as [|γ hvl]; nsimpl.
        - rewrite subst0_shift0_reflection; auto.
        - destruct (α ≟ γ) as [αγ|αγ], (β ≟ γ) as [βγ|βγ]; ncrush.
          + destruct y; nsimpl.
            reflexivity.
            rewrite subst_shift0_het_comm2; auto.
            f_equal; auto.
          + rewrite subst_shift0_comm.
            f_equal; auto.
          + rewrite subst_shift0_het_comm1; nsimpl.
            f_equal; auto.
          + rewrite subst_shift0_het_comm2; nsimpl.
            f_equal; auto.
          + auto.
      Qed.

      Lemma subst_subst0_distinct_comm3_ind :
        (∀ s (t : Term s)
           (αIn : α ∈ depsOf s)
           (βIn : β ∈ depsOf s)
           (hvl : HVL)
           (xα : Trace α) (tα : Term (sortOf α)) (tβ : Term (sortOf β)),
           subst αIn (weaken xα hvl) tα (subst βIn (weaken T0 hvl) tβ t) =
           subst βIn (weaken T0 hvl) tβ (subst αIn (weaken (TS βInα xα) hvl) tα t)).
      Proof.
        terminduction.
        - ncrush.
          + apply subst_substIndex0_distinct_comm3_indα.
          + reflexivity.
        - ncrush.
          + eapply EqTerm_sound.
            eapply EqtTrans.
            eapply EqtCongSubst.
            eapply EqxTrans.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtCongSubst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtRefl.
            eapply EqtTrans.
            eapply EqtAssumption.
            eapply H.
            eapply EqtSym.
            eapply EqtCongSubst.
            eapply EqxTrans.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtCongSubst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtRefl.
          + eapply EqTerm_sound.
            eapply EqtSym.
            eapply EqtCongSubst.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqtRefl.
            eapply EqtRefl.
          + reflexivity.
      Defined.

      Lemma subst_subst0_distinct_comm3 :
        (∀ s (t : Term s)
           (αIn : α ∈ depsOf s)
           (βIn : β ∈ depsOf s)
           (xα : Trace α) (tα : Term (sortOf α)) (tβ : Term (sortOf β)),
           subst αIn xα tα (subst βIn T0 tβ t) =
           subst βIn T0 tβ (subst αIn (TS βInα xα) tα t)).
      Proof.
        intros; apply (subst_subst0_distinct_comm3_ind _ _ _ HV0).
      Qed.

    End Variant3.

    Section Variant4.

      Variable αInβ : α ∈ depsOf (sortOf β).
      Variable βInα : β ∈ depsOf (sortOf α).

      Lemma subst_substIndex0_distinct_comm4_indβ
        (xα : Trace α) (tα : Term (sortOf α)) (tβ : Term (sortOf β)) :
        ∀ (hvl : HVL) (y : Index β),
          subst αInβ (weaken xα hvl) tα (substIndex (weaken T0 hvl) tβ y) =
          substIndex (weaken T0 hvl) (subst αInβ xα tα tβ) y.
      Proof.
        intro hvl; induction hvl as [|γ hvl]; simpl; intros; auto.
        - destruct y; nsimpl; auto.
        - destruct (α ≟ γ) as [αγ|αγ], (β ≟ γ) as [βγ|βγ]; ncrush.
          + rewrite subst_shift0_comm.
            f_equal; auto.
          + destruct y; nsimpl.
            * reflexivity.
            * rewrite subst_shift0_het_comm1; nsimpl.
              f_equal; auto.
          + rewrite subst_shift0_het_comm1; nsimpl.
            f_equal; auto.
          + auto.
      Qed.

      Lemma subst_substIndex0_distinct_comm4_indα
        (xα : Trace α) (tα : Term (sortOf α)) (tβ : Term (sortOf β)) :
        ∀ (hvl : HVL) (y : Index α),
          substIndex (weaken xα hvl) tα y =
          subst βInα (weaken T0 hvl) (subst αInβ xα tα tβ)
                (substIndex (weaken (TS βInα xα) hvl) tα y).
      Proof.
        intro hvl; induction hvl as [|γ hvl]; nsimpl.
        - rewrite subst0_shift0_reflection; auto.
        - destruct (α ≟ γ) as [αγ|αγ], (β ≟ γ) as [βγ|βγ]; nsimpl.
          + destruct y; nsimpl.
            reflexivity.
            rewrite subst_shift0_het_comm1; nsimpl.
            f_equal; auto.
          + rewrite subst_shift0_comm.
            f_equal; auto.
          + ncrush.
            * rewrite subst_shift0_het_comm1; nsimpl.
              f_equal; auto.
            * auto.
      Qed.

      Lemma subst_subst0_distinct_comm4_ind :
        (∀ s (t : Term s)
           (αIn : α ∈ depsOf s)
           (βIn : β ∈ depsOf s)
           (hvl : HVL)
           (xα : Trace α) (tα : Term (sortOf α)) (tβ : Term (sortOf β)),
           subst αIn (weaken xα hvl) tα                    (subst βIn (weaken T0 hvl) tβ t) =
           subst βIn (weaken T0 hvl) (subst αInβ xα tα tβ) (subst αIn (weaken (TS βInα xα) hvl) tα t)).
      Proof.
        terminduction.
        - ncrush.
          + apply subst_substIndex0_distinct_comm4_indβ.
          + apply subst_substIndex0_distinct_comm4_indα.
          + reflexivity.
        - ncrush.
          + eapply EqTerm_sound.
            eapply EqtTrans.
            eapply EqtCongSubst.
            eapply EqxTrans.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtCongSubst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtRefl.
            eapply EqtTrans.
            eapply EqtAssumption.
            eapply H.
            eapply EqtSym.
            eapply EqtCongSubst.
            eapply EqxTrans.
            eapply EqxCongWeaken.
            eapply EqxRefl.
            eapply Eqh_evalbindspec_subst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtCongSubst.
            eapply EqxWeakenAppend.
            eapply EqtRefl.
            eapply EqtRefl.
          + reflexivity.
      Defined.

      Lemma subst_subst0_distinct_comm4 :
        (∀ s (t : Term s)
           (αIn : α ∈ depsOf s)
           (βIn : β ∈ depsOf s)
           (xα : Trace α) (tα : Term (sortOf α)) (tβ : Term (sortOf β)),
           subst αIn xα tα                    (subst βIn T0 tβ t) =
           subst βIn T0 (subst αInβ xα tα tβ) (subst αIn (TS βInα xα) tα t)).
      Proof.
        intros; apply (subst_subst0_distinct_comm4_ind _ _ _ HV0).
      Qed.

    End Variant4.

  End SubstSubstDistinct.

End Interaction.