Require Import Coq.Program.Equality.
Require Import Coq.Program.Tactics.
Require Export Needle.
Section Namespace.
  Inductive Namespace : Type :=
    | ty .
  Lemma eq_namespace_dec (a : Namespace) (b : Namespace) :
    (sumbool (a =
    b) (not (a =
    b))).
  Proof.
    decide equality .
  Defined.
End Namespace.

Section HVarlist.
  Inductive Hvl : Type :=
    | H0 
    | HS (a : Namespace) (k : Hvl).
  
  Fixpoint appendHvl (k : Hvl) (k0 : Hvl) {struct k0} : Hvl :=
    match k0 with
      | (H0) => k
      | (HS a k0) => (HS a (appendHvl k k0))
    end.
  
  Lemma appendHvl_assoc  :
    (forall (k : Hvl) (k0 : Hvl) (k1 : Hvl) ,
      ((appendHvl (appendHvl k k0) k1) =
      (appendHvl k (appendHvl k0 k1)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End HVarlist.

Section Index.
  Inductive Index (a : Namespace) : Type :=
    | I0 
    | IS : (Index a) -> (Index a).
  
  Global Arguments I0 [a] .
  Global Arguments IS [a] _ .
  
  Lemma eq_index_dec {a : Namespace} (i : (Index a)) (j : (Index a)) :
    (sumbool (i =
    j) (not (i =
    j))).
  Proof.
    decide equality .
  Qed.
  
  Fixpoint weakenIndexty (X : (Index ty)) (k : Hvl) {struct k} : (Index ty) :=
    match k with
      | (H0) => X
      | (HS a k) => match a with
        | (ty) => (IS (weakenIndexty X k))
      end
    end.
  Lemma weakenIndexty_append  :
    (forall (X : (Index ty)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexty (weakenIndexty X k) k0) =
      (weakenIndexty X (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End Index.

Section Terms.
  Inductive Ty : Set :=
    | tvar (X : (Index ty))
    | tarrow (T1 : Ty) (T2 : Ty)
    | tall (T : Ty).
  Scheme ind_Ty := Induction for Ty Sort Prop.
End Terms.

Section Functions.
  
End Functions.

Section Shift.
  Inductive Cutoff (a : Namespace) : Type :=
    | C0 
    | CS : (Cutoff a) -> (Cutoff a).
  
  Global Arguments C0 {a} .
  Global Arguments CS {a} _ .
  Fixpoint weakenCutoffty (c : (Cutoff ty)) (k : Hvl) {struct k} : (Cutoff ty) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (ty) => (CS (weakenCutoffty c k))
      end
    end.
  
  Lemma weakenCutoffty_append  :
    (forall (c : (Cutoff ty)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffty (weakenCutoffty c k) k0) =
      (weakenCutoffty c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint shiftIndex (c : (Cutoff ty)) (X : (Index ty)) {struct c} : (Index ty) :=
    match c with
      | (C0) => (IS X)
      | (CS c) => match X with
        | (I0) => I0
        | (IS X) => (IS (shiftIndex c X))
      end
    end.
  Fixpoint shiftTy (c : (Cutoff ty)) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (tvar X) => (tvar (shiftIndex c X))
      | (tarrow T1 T2) => (tarrow (shiftTy (weakenCutoffty c H0) T1) (shiftTy (weakenCutoffty c H0) T2))
      | (tall T) => (tall (shiftTy (weakenCutoffty c (appendHvl (HS ty H0) H0)) T))
    end.
End Shift.

Section Weaken.
  Fixpoint weakenTy (S0 : Ty) (k : Hvl) {struct k} : Ty :=
    match k with
      | (H0) => S0
      | (HS ty k) => (shiftTy C0 (weakenTy S0 k))
    end.
End Weaken.

Section Subst.
  Inductive Trace (a : Namespace) : Type :=
    | X0 
    | XS (b : Namespace)
        (T : (Trace a)).
  
  Global Arguments X0 [a] .
  Global Arguments XS [a] b _ .
  Fixpoint weakenTrace {a : Namespace} (x : (Trace a)) (k : Hvl) {struct k} : (Trace a) :=
    match k with
      | (H0) => x
      | (HS b k) => (XS b (weakenTrace x k))
    end.
  Lemma weakenTrace_append (a : Namespace) :
    (forall (x : (Trace a)) (k : Hvl) (k0 : Hvl) ,
      ((weakenTrace (weakenTrace x k) k0) =
      (weakenTrace x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint substIndex (d : (Trace ty)) (S0 : Ty) (X : (Index ty)) {struct d} : Ty :=
    match d with
      | (X0) => match X with
        | (I0) => S0
        | (IS X) => (tvar X)
      end
      | (XS ty d) => match X with
        | (I0) => (tvar I0)
        | (IS X) => (shiftTy C0 (substIndex d S0 X))
      end
    end.
  Fixpoint substTy (d : (Trace ty)) (S0 : Ty) (S1 : Ty) {struct S1} : Ty :=
    match S1 with
      | (tvar X) => (substIndex d S0 X)
      | (tarrow T1 T2) => (tarrow (substTy d S0 T1) (substTy d S0 T2))
      | (tall T) => (tall (substTy (XS ty d) S0 T))
    end.
End Subst.

Global Hint Resolve (f_equal (shiftTy C0)) : cong_shift0.
 Hint Rewrite weakenCutoffty_append : interaction.
 Hint Rewrite weakenCutoffty_append : weakencutoff_append.
 Hint Rewrite weakenTrace_append : interaction.
 Hint Rewrite weakenTrace_append : weakentrace_append.
Section SubstShiftReflection.
  Lemma substIndex0_shiftIndex0_reflection_ind (S0 : Ty) :
    (forall (k0 : Hvl) (X2 : (Index ty)) ,
      ((substIndex (weakenTrace X0 k0) S0 (shiftIndex (weakenCutoffty C0 k0) X2)) =
      (tvar X2))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Definition subst0_shift0_Ty_reflection_ind := (ind_Ty (fun (S1 : Ty) =>
    (forall (k1 : Hvl) (S2 : Ty) ,
      ((substTy (weakenTrace X0 k1) S2 (shiftTy (weakenCutoffty C0 k1) S1)) =
      S1))) (fun (X5 : (Index ty)) =>
    (fun (k1 : Hvl) (S1 : Ty) =>
      (substIndex0_shiftIndex0_reflection_ind S1 k1 X5))) (fun (T1 : Ty) IHT5 (T2 : Ty) IHT6 (k1 : Hvl) (S1 : Ty) =>
    (f_equal2 tarrow (IHT5 k1 S1) (IHT6 k1 S1))) (fun (T : Ty) IHT5 (k1 : Hvl) (S1 : Ty) =>
    (f_equal tall (eq_trans (f_equal3 substTy (weakenTrace_append ty X0 k1 (HS ty H0)) eq_refl eq_refl) (IHT5 (HS ty k1) S1))))).
  Definition substTy0_shiftTy0_reflection (S2 : Ty) : (forall (S1 : Ty) ,
    ((substTy X0 S1 (shiftTy C0 S2)) =
    S2)) := (subst0_shift0_Ty_reflection_ind S2 H0).
End SubstShiftReflection.
Section ShiftInteraction.
  Section ShiftIndexCommInd.
    Lemma shiftIndex_shiftIndex0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff ty)) (X : (Index ty)) ,
        ((shiftIndex (weakenCutoffty (CS c) k) (shiftIndex (weakenCutoffty C0 k) X)) =
        (shiftIndex (weakenCutoffty C0 k) (shiftIndex (weakenCutoffty c k) X)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
  End ShiftIndexCommInd.
  Section ShiftCommInd.
    Definition shift_shift0_Ty_comm_ind := (ind_Ty (fun (S0 : Ty) =>
      (forall (k0 : Hvl) (c0 : (Cutoff ty)) ,
        ((shiftTy (weakenCutoffty (CS c0) k0) (shiftTy (weakenCutoffty C0 k0) S0)) =
        (shiftTy (weakenCutoffty C0 k0) (shiftTy (weakenCutoffty c0 k0) S0))))) (fun (X2 : (Index ty)) =>
      (fun (k0 : Hvl) (c0 : (Cutoff ty)) =>
        (f_equal tvar (shiftIndex_shiftIndex0_comm_ind k0 c0 X2)))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k0 : Hvl) (c0 : (Cutoff ty)) =>
      (f_equal2 tarrow (IHT0 k0 c0) (IHT3 k0 c0))) (fun (T : Ty) IHT0 (k0 : Hvl) (c0 : (Cutoff ty)) =>
      (f_equal tall (IHT0 (HS ty k0) c0)))).
  End ShiftCommInd.
  Section ShiftComm.
    Definition shift_shift0_Ty_comm (S0 : Ty) : (forall (c0 : (Cutoff ty)) ,
      ((shiftTy (CS c0) (shiftTy C0 S0)) =
      (shiftTy C0 (shiftTy c0 S0)))) := (shift_shift0_Ty_comm_ind S0 H0).
  End ShiftComm.
End ShiftInteraction.
 Hint Rewrite shift_shift0_Ty_comm : interaction.
 Hint Rewrite shift_shift0_Ty_comm : shift_shift0.
Section WeakenShift.
  Lemma weakenTy_shiftTy  :
    (forall (k6 : Hvl) (c1 : (Cutoff ty)) (S7 : Ty) ,
      ((weakenTy (shiftTy c1 S7) k6) =
      (shiftTy (weakenCutoffty c1 k6) (weakenTy S7 k6)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
End WeakenShift.
Section ShiftSubstInteraction.
  Section ShiftSubstIndexCommInd.
    Lemma shiftTy_substIndex0_comm_ind (c0 : (Cutoff ty)) (S1 : Ty) :
      (forall (k1 : Hvl) (X5 : (Index ty)) ,
        ((shiftTy (weakenCutoffty c0 k1) (substIndex (weakenTrace X0 k1) S1 X5)) =
        (substIndex (weakenTrace X0 k1) (shiftTy c0 S1) (shiftIndex (weakenCutoffty (CS c0) k1) X5)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End ShiftSubstIndexCommInd.
  Section ShiftSubstCommInd.
    Definition shift_subst0_Ty_comm_ind := (ind_Ty (fun (S2 : Ty) =>
      (forall (k2 : Hvl) (c1 : (Cutoff ty)) (S3 : Ty) ,
        ((shiftTy (weakenCutoffty c1 k2) (substTy (weakenTrace X0 k2) S3 S2)) =
        (substTy (weakenTrace X0 k2) (shiftTy c1 S3) (shiftTy (weakenCutoffty (CS c1) k2) S2))))) (fun (X8 : (Index ty)) =>
      (fun (k2 : Hvl) (c1 : (Cutoff ty)) (S2 : Ty) =>
        (shiftTy_substIndex0_comm_ind c1 S2 k2 X8))) (fun (T1 : Ty) IHT8 (T2 : Ty) IHT9 (k2 : Hvl) (c1 : (Cutoff ty)) (S2 : Ty) =>
      (f_equal2 tarrow (IHT8 k2 c1 S2) (IHT9 k2 c1 S2))) (fun (T : Ty) IHT8 (k2 : Hvl) (c1 : (Cutoff ty)) (S2 : Ty) =>
      (f_equal tall (eq_trans (f_equal2 shiftTy eq_refl (f_equal3 substTy (weakenTrace_append ty X0 k2 (HS ty H0)) eq_refl eq_refl)) (eq_trans (IHT8 (HS ty k2) c1 S2) (f_equal3 substTy (eq_sym (weakenTrace_append ty X0 k2 (HS ty H0))) eq_refl eq_refl)))))).
  End ShiftSubstCommInd.
  Section ShiftSubstComm.
    Definition shiftTy_substTy0_comm (S3 : Ty) : (forall (c1 : (Cutoff ty)) (S2 : Ty) ,
      ((shiftTy c1 (substTy X0 S2 S3)) =
      (substTy X0 (shiftTy c1 S2) (shiftTy (CS c1) S3)))) := (shift_subst0_Ty_comm_ind S3 H0).
  End ShiftSubstComm.
  Section SubstShiftIndexCommInd.
    Lemma substIndex_shiftTy0_comm_ind (d : (Trace ty)) (S2 : Ty) :
      (forall (k2 : Hvl) (X8 : (Index ty)) ,
        ((substIndex (weakenTrace (XS ty d) k2) S2 (shiftIndex (weakenCutoffty C0 k2) X8)) =
        (shiftTy (weakenCutoffty C0 k2) (substIndex (weakenTrace d k2) S2 X8)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End SubstShiftIndexCommInd.
  Section SubstShiftCommInd.
    Definition subst_shift0_Ty_comm_ind := (ind_Ty (fun (S5 : Ty) =>
      (forall (k5 : Hvl) (d2 : (Trace ty)) (S6 : Ty) ,
        ((substTy (weakenTrace (XS ty d2) k5) S6 (shiftTy (weakenCutoffty C0 k5) S5)) =
        (shiftTy (weakenCutoffty C0 k5) (substTy (weakenTrace d2 k5) S6 S5))))) (fun (X11 : (Index ty)) =>
      (fun (k5 : Hvl) (d2 : (Trace ty)) (S5 : Ty) =>
        (substIndex_shiftTy0_comm_ind d2 S5 k5 X11))) (fun (T1 : Ty) IHT11 (T2 : Ty) IHT12 (k5 : Hvl) (d2 : (Trace ty)) (S5 : Ty) =>
      (f_equal2 tarrow (IHT11 k5 d2 S5) (IHT12 k5 d2 S5))) (fun (T : Ty) IHT11 (k5 : Hvl) (d2 : (Trace ty)) (S5 : Ty) =>
      (f_equal tall (eq_trans (f_equal3 substTy (weakenTrace_append ty (XS ty d2) k5 (HS ty H0)) eq_refl eq_refl) (eq_trans (IHT11 (HS ty k5) d2 S5) (f_equal2 shiftTy eq_refl (f_equal3 substTy (eq_sym (weakenTrace_append ty d2 k5 (HS ty H0))) eq_refl eq_refl))))))).
  End SubstShiftCommInd.
  Section SubstShiftComm.
    Definition substTy_shiftTy0_comm (S6 : Ty) : (forall (d2 : (Trace ty)) (S5 : Ty) ,
      ((substTy (XS ty d2) S5 (shiftTy C0 S6)) =
      (shiftTy C0 (substTy d2 S5 S6)))) := (subst_shift0_Ty_comm_ind S6 H0).
  End SubstShiftComm.
  Section SubstSubordInd.
    
  End SubstSubordInd.
  Section SubstSubord.
    
  End SubstSubord.
End ShiftSubstInteraction.
 Hint Rewrite substTy0_shiftTy0_reflection : interaction.
 Hint Rewrite substTy0_shiftTy0_reflection : reflection.
 Hint Rewrite substTy_shiftTy0_comm : interaction.
 Hint Rewrite substTy_shiftTy0_comm : subst_shift0.
 Hint Rewrite shiftTy_substTy0_comm : interaction.
 Hint Rewrite shiftTy_substTy0_comm : shift_subst0.
Section SubstSubstInteraction.
  Section SubstSubstIndexCommInd.
    Lemma substTy_substIndex0_commright_ind (d2 : (Trace ty)) (S5 : Ty) (S6 : Ty) :
      (forall (k5 : Hvl) (X11 : (Index ty)) ,
        ((substTy (weakenTrace d2 k5) S5 (substIndex (weakenTrace X0 k5) S6 X11)) =
        (substTy (weakenTrace X0 k5) (substTy d2 S5 S6) (substIndex (weakenTrace (XS ty d2) k5) S5 X11)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
  End SubstSubstIndexCommInd.
  Section SubstSubstCommInd.
    Definition subst_subst0_Ty_comm_ind := (ind_Ty (fun (S7 : Ty) =>
      (forall (k6 : Hvl) (d3 : (Trace ty)) (S8 : Ty) (S9 : Ty) ,
        ((substTy (weakenTrace d3 k6) S8 (substTy (weakenTrace X0 k6) S9 S7)) =
        (substTy (weakenTrace X0 k6) (substTy d3 S8 S9) (substTy (weakenTrace (XS ty d3) k6) S8 S7))))) (fun (X14 : (Index ty)) =>
      (fun (k6 : Hvl) (d3 : (Trace ty)) (S7 : Ty) (S8 : Ty) =>
        (substTy_substIndex0_commright_ind d3 S7 S8 k6 X14))) (fun (T1 : Ty) IHT14 (T2 : Ty) IHT15 (k6 : Hvl) (d3 : (Trace ty)) (S7 : Ty) (S8 : Ty) =>
      (f_equal2 tarrow (IHT14 k6 d3 S7 S8) (IHT15 k6 d3 S7 S8))) (fun (T : Ty) IHT14 (k6 : Hvl) (d3 : (Trace ty)) (S7 : Ty) (S8 : Ty) =>
      (f_equal tall (eq_trans (f_equal3 substTy (weakenTrace_append ty d3 k6 (HS ty H0)) eq_refl (f_equal3 substTy (weakenTrace_append ty X0 k6 (HS ty H0)) eq_refl eq_refl)) (eq_trans (IHT14 (HS ty k6) d3 S7 S8) (f_equal3 substTy (eq_sym (weakenTrace_append ty X0 k6 (HS ty H0))) eq_refl (f_equal3 substTy (eq_sym (weakenTrace_append ty (XS ty d3) k6 (HS ty H0))) eq_refl eq_refl))))))).
  End SubstSubstCommInd.
  Section SubstSubstComm.
    Definition substTy_substTy0_comm (S9 : Ty) : (forall (d3 : (Trace ty)) (S7 : Ty) (S8 : Ty) ,
      ((substTy d3 S7 (substTy X0 S8 S9)) =
      (substTy X0 (substTy d3 S7 S8) (substTy (XS ty d3) S7 S9)))) := (subst_subst0_Ty_comm_ind S9 H0).
  End SubstSubstComm.
  Section WeakenSubst.
    Lemma weakenTy_substTy  :
      (forall (k6 : Hvl) (d3 : (Trace ty)) (S7 : Ty) (S8 : Ty) ,
        ((weakenTy (substTy d3 S7 S8) k6) =
        (substTy (weakenTrace d3 k6) S7 (weakenTy S8 k6)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
  End WeakenSubst.
  Section WeakenAppend.
    Lemma weakenTy_append  :
      (forall (S7 : Ty) (k6 : Hvl) (k7 : Hvl) ,
        ((weakenTy (weakenTy S7 k6) k7) =
        (weakenTy S7 (appendHvl k6 k7)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
  End WeakenAppend.
End SubstSubstInteraction.
 Hint Rewrite substTy_substTy0_comm : interaction.
 Hint Rewrite substTy_substTy0_comm : subst_subst0.
 Hint Rewrite weakenTy_shiftTy : weaken_shift.
 Hint Rewrite weakenTy_substTy : weaken_subst.
Section WellFormed.
  Fixpoint wfindex {a : Namespace} (k6 : Hvl) (i : (Index a)) {struct k6} : Prop :=
    match k6 with
      | (H0) => False
      | (HS b k6) => match (eq_namespace_dec a b) with
        | (left e) => match i with
          | (I0) => True
          | (IS i) => (wfindex k6 i)
        end
        | (right n) => (wfindex k6 i)
      end
    end.
  Inductive wfTy (k6 : Hvl) : Ty -> Prop :=
    | wfTy_tvar {X17 : (Index ty)}
        (wfi : (wfindex k6 X17)) :
        (wfTy k6 (tvar X17))
    | wfTy_tarrow {T17 : Ty}
        (wf : (wfTy k6 T17)) {T18 : Ty}
        (wf0 : (wfTy k6 T18)) :
        (wfTy k6 (tarrow T17 T18))
    | wfTy_tall {T19 : Ty}
        (wf : (wfTy (HS ty k6) T19)) :
        (wfTy k6 (tall T19)).
  Scheme ind_wfTy := Induction for wfTy Sort Prop.
End WellFormed.
Section ShiftWellFormed.
  Inductive shifthvl_ty : (forall (c2 : (Cutoff ty)) (k6 : Hvl) (k7 : Hvl) ,
    Prop) :=
    | shifthvl_ty_here (k6 : Hvl)
        : (shifthvl_ty C0 k6 (HS ty k6))
    | shifthvl_ty_there_ty
        (c2 : (Cutoff ty)) (k6 : Hvl)
        (k7 : Hvl) :
        (shifthvl_ty c2 k6 k7) -> (shifthvl_ty (CS c2) (HS ty k6) (HS ty k7)).
  Lemma weaken_shifthvl_ty  :
    (forall (k6 : Hvl) {c2 : (Cutoff ty)} {k7 : Hvl} {k8 : Hvl} ,
      (shifthvl_ty c2 k7 k8) -> (shifthvl_ty (weakenCutoffty c2 k6) (appendHvl k7 k6) (appendHvl k8 k6))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma shift_wfindex_ty  :
    (forall (c2 : (Cutoff ty)) (k6 : Hvl) (k7 : Hvl) (ins : (shifthvl_ty c2 k6 k7)) (X17 : (Index ty)) ,
      (wfindex k6 X17) -> (wfindex k7 (shiftIndex c2 X17))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Definition shift_wfTy : (forall (k6 : Hvl) ,
    (forall (S7 : Ty) (wf : (wfTy k6 S7)) ,
      (forall (c2 : (Cutoff ty)) (k7 : Hvl) ,
        (shifthvl_ty c2 k6 k7) -> (wfTy k7 (shiftTy c2 S7))))) := (ind_wfTy (fun (k6 : Hvl) (S7 : Ty) (wf : (wfTy k6 S7)) =>
    (forall (c2 : (Cutoff ty)) (k7 : Hvl) ,
      (shifthvl_ty c2 k6 k7) -> (wfTy k7 (shiftTy c2 S7)))) (fun (k6 : Hvl) (X17 : (Index ty)) (wfi : (wfindex k6 X17)) (c2 : (Cutoff ty)) (k7 : Hvl) (ins : (shifthvl_ty c2 k6 k7)) =>
    (wfTy_tvar k7 (shift_wfindex_ty c2 k6 k7 ins X17 wfi))) (fun (k6 : Hvl) (T1 : Ty) (wf : (wfTy k6 T1)) IHT17 (T2 : Ty) (wf0 : (wfTy k6 T2)) IHT18 (c2 : (Cutoff ty)) (k7 : Hvl) (ins : (shifthvl_ty c2 k6 k7)) =>
    (wfTy_tarrow k7 (IHT17 c2 k7 (weaken_shifthvl_ty H0 ins)) (IHT18 c2 k7 (weaken_shifthvl_ty H0 ins)))) (fun (k6 : Hvl) (T : Ty) (wf : (wfTy (HS ty k6) T)) IHT17 (c2 : (Cutoff ty)) (k7 : Hvl) (ins : (shifthvl_ty c2 k6 k7)) =>
    (wfTy_tall k7 (IHT17 (CS c2) (HS ty k7) (weaken_shifthvl_ty (HS ty H0) ins))))).
End ShiftWellFormed.
 Hint Resolve shift_wfindex_ty : infra.
 Hint Resolve shift_wfindex_ty : shift.
 Hint Resolve shift_wfindex_ty : shift_wf.
 Hint Resolve shift_wfindex_ty : wf.
 Hint Resolve shift_wfTy : infra.
 Hint Resolve shift_wfTy : shift.
 Hint Resolve shift_wfTy : shift_wf.
 Hint Resolve shift_wfTy : wf.
 Hint Constructors shifthvl_ty : infra.
 Hint Constructors shifthvl_ty : shift.
 Hint Constructors shifthvl_ty : shift_wf.
 Hint Constructors shifthvl_ty : wf.
 Hint Resolve weaken_shifthvl_ty : infra.
 Hint Resolve weaken_shifthvl_ty : shift.
 Hint Resolve weaken_shifthvl_ty : shift_wf.
 Hint Resolve weaken_shifthvl_ty : weaken.
 Hint Resolve weaken_shifthvl_ty : wf.
Section SubstWellFormed.
  Inductive substhvl_ty (k6 : Hvl) : (forall (d3 : (Trace ty)) (k7 : Hvl) (k8 : Hvl) ,
    Prop) :=
    | substhvl_ty_here :
        (substhvl_ty k6 X0 (HS ty k6) k6)
    | substhvl_ty_there_ty
        {d3 : (Trace ty)} {k7 : Hvl}
        {k8 : Hvl} :
        (substhvl_ty k6 d3 k7 k8) -> (substhvl_ty k6 (XS ty d3) (HS ty k7) (HS ty k8)).
  Lemma weaken_substhvl_ty  :
    (forall {k7 : Hvl} (k6 : Hvl) {d3 : (Trace ty)} {k8 : Hvl} {k9 : Hvl} ,
      (substhvl_ty k7 d3 k8 k9) -> (substhvl_ty k7 (weakenTrace d3 k6) (appendHvl k8 k6) (appendHvl k9 k6))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma substhvl_ty_wfindex_ty {k6 : Hvl} {S7 : Ty} (wft : (wfTy k6 S7)) :
    (forall {d3 : (Trace ty)} {k7 : Hvl} {k8 : Hvl} ,
      (substhvl_ty k6 d3 k7 k8) -> (forall {X17 : (Index ty)} ,
        (wfindex k7 X17) -> (wfTy k8 (substIndex d3 S7 X17)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Definition substhvl_ty_wfTy {k6 : Hvl} {S7 : Ty} (wf : (wfTy k6 S7)) : (forall (k7 : Hvl) ,
    (forall (S8 : Ty) (wf0 : (wfTy k7 S8)) ,
      (forall {d3 : (Trace ty)} {k8 : Hvl} ,
        (substhvl_ty k6 d3 k7 k8) -> (wfTy k8 (substTy d3 S7 S8))))) := (ind_wfTy (fun (k7 : Hvl) (S8 : Ty) (wf0 : (wfTy k7 S8)) =>
    (forall {d3 : (Trace ty)} {k8 : Hvl} ,
      (substhvl_ty k6 d3 k7 k8) -> (wfTy k8 (substTy d3 S7 S8)))) (fun (k7 : Hvl) {X17 : (Index ty)} (wfi : (wfindex k7 X17)) {d3 : (Trace ty)} {k8 : Hvl} (del : (substhvl_ty k6 d3 k7 k8)) =>
    (substhvl_ty_wfindex_ty wf del wfi)) (fun (k7 : Hvl) (T1 : Ty) (wf0 : (wfTy k7 T1)) IHT17 (T2 : Ty) (wf1 : (wfTy k7 T2)) IHT18 {d3 : (Trace ty)} {k8 : Hvl} (del : (substhvl_ty k6 d3 k7 k8)) =>
    (wfTy_tarrow k8 (IHT17 (weakenTrace d3 H0) k8 (weaken_substhvl_ty H0 del)) (IHT18 (weakenTrace d3 H0) k8 (weaken_substhvl_ty H0 del)))) (fun (k7 : Hvl) (T : Ty) (wf0 : (wfTy (HS ty k7) T)) IHT17 {d3 : (Trace ty)} {k8 : Hvl} (del : (substhvl_ty k6 d3 k7 k8)) =>
    (wfTy_tall k8 (IHT17 (weakenTrace d3 (HS ty H0)) (HS ty k8) (weaken_substhvl_ty (HS ty H0) del))))).
End SubstWellFormed.
 Hint Resolve substhvl_ty_wfindex_ty : infra.
 Hint Resolve substhvl_ty_wfindex_ty : subst.
 Hint Resolve substhvl_ty_wfindex_ty : subst_wf.
 Hint Resolve substhvl_ty_wfindex_ty : wf.
 Hint Resolve substhvl_ty_wfTy : infra.
 Hint Resolve substhvl_ty_wfTy : subst.
 Hint Resolve substhvl_ty_wfTy : subst_wf.
 Hint Resolve substhvl_ty_wfTy : wf.
 Hint Constructors substhvl_ty : infra.
 Hint Constructors substhvl_ty : subst.
 Hint Constructors substhvl_ty : subst_wf.
 Hint Constructors substhvl_ty : wf.
Section Context.
  Inductive Env : Type :=
    | empty 
    | etvar (G : Env) (T1 : Ty)
        (T2 : Ty).
  Fixpoint appendEnv (G : Env) (G0 : Env) : Env :=
    match G0 with
      | (empty) => G
      | (etvar G1 T1 T2) => (etvar (appendEnv G G1) T1 T2)
    end.
  Fixpoint domainEnv (G : Env) : Hvl :=
    match G with
      | (empty) => H0
      | (etvar G0 T1 T2) => (HS ty (domainEnv G0))
    end.
  Lemma appendEnv_assoc  :
    (forall (G : Env) (G0 : Env) (G1 : Env) ,
      ((appendEnv G (appendEnv G0 G1)) =
      (appendEnv (appendEnv G G0) G1))).
  Proof.
    needleGenericAppendEnvAssoc .
  Qed.
  Lemma domainEnv_appendEnv  :
    (forall (G : Env) (G0 : Env) ,
      ((domainEnv (appendEnv G G0)) =
      (appendHvl (domainEnv G) (domainEnv G0)))).
  Proof.
    needleGenericDomainEnvAppendEnv .
  Qed.
  Fixpoint shiftEnv (c2 : (Cutoff ty)) (G : Env) : Env :=
    match G with
      | (empty) => empty
      | (etvar G0 T1 T2) => (etvar (shiftEnv c2 G0) (shiftTy (weakenCutoffty c2 (domainEnv G0)) T1) (shiftTy (weakenCutoffty c2 (domainEnv G0)) T2))
    end.
  Fixpoint substEnv (d3 : (Trace ty)) (S7 : Ty) (G : Env) : Env :=
    match G with
      | (empty) => empty
      | (etvar G0 T1 T2) => (etvar (substEnv d3 S7 G0) (substTy (weakenTrace d3 (domainEnv G0)) S7 T1) (substTy (weakenTrace d3 (domainEnv G0)) S7 T2))
    end.
  Lemma domainEnv_shiftEnv  :
    (forall (c2 : (Cutoff ty)) (G : Env) ,
      ((domainEnv (shiftEnv c2 G)) =
      (domainEnv G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainEnv_substEnv  :
    (forall (d3 : (Trace ty)) (S7 : Ty) (G : Env) ,
      ((domainEnv (substEnv d3 S7 G)) =
      (domainEnv G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
End Context.
 Hint Rewrite domainEnv_appendEnv : interaction.
 Hint Rewrite domainEnv_appendEnv : env_domain_append.
Section ContextStuff.
  Section ShiftEnvAppendEnv.
    Lemma shiftEnv_appendEnv  :
      (forall (c2 : (Cutoff ty)) (G : Env) (G0 : Env) ,
        ((shiftEnv c2 (appendEnv G G0)) =
        (appendEnv (shiftEnv c2 G) (shiftEnv (weakenCutoffty c2 (domainEnv G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
  End ShiftEnvAppendEnv.
  Section SubstEnvAppendEnv.
    Lemma substEnv_appendEnv  :
      (forall (d3 : (Trace ty)) (S7 : Ty) (G : Env) (G0 : Env) ,
        ((substEnv d3 S7 (appendEnv G G0)) =
        (appendEnv (substEnv d3 S7 G) (substEnv (weakenTrace d3 (domainEnv G)) S7 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
  End SubstEnvAppendEnv.
  Section Lookups.
    Inductive lookup_etvar : Env -> (Index ty) -> Ty -> Ty -> Prop :=
      | lookup_etvar_here {G : Env}
          {T17 : Ty} {T18 : Ty} :
          (wfTy (domainEnv G) T17) -> (wfTy (domainEnv G) T18) -> (lookup_etvar (etvar G T17 T18) I0 (shiftTy C0 T17) (shiftTy C0 T18))
      | lookup_etvar_there_etvar
          {G : Env} {X17 : (Index ty)}
          {T17 : Ty} {T18 : Ty} {T19 : Ty}
          {T20 : Ty} :
          (lookup_etvar G X17 T17 T18) -> (lookup_etvar (etvar G T19 T20) (IS X17) (shiftTy C0 T17) (shiftTy C0 T18)).
    Lemma lookup_etvar_inversion_here  :
      (forall (G : Env) (S7 : Ty) (S8 : Ty) (S9 : Ty) (S10 : Ty) ,
        (lookup_etvar (etvar G S7 S8) I0 S9 S10) -> ((shiftTy C0 S7) =
        S9) /\
        ((shiftTy C0 S8) =
        S10)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_etvar_functional  :
      (forall {G : Env} {X17 : (Index ty)} ,
        (forall {S7 : Ty} {S8 : Ty} ,
          (lookup_etvar G X17 S7 S8) -> (forall {S9 : Ty} {S10 : Ty} ,
            (lookup_etvar G X17 S9 S10) -> (S7 =
            S9) /\
            (S8 =
            S10)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_etvar_wf  :
      (forall {G : Env} {X17 : (Index ty)} {S7 : Ty} {S8 : Ty} ,
        (lookup_etvar G X17 S7 S8) -> (wfTy (domainEnv G) S7) /\
        (wfTy (domainEnv G) S8)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma weaken_lookup_etvar  :
      (forall {G : Env} (G0 : Env) {X17 : (Index ty)} {T17 : Ty} {T18 : Ty} ,
        (lookup_etvar G X17 T17 T18) -> (lookup_etvar (appendEnv G G0) (weakenIndexty X17 (domainEnv G0)) (weakenTy T17 (domainEnv G0)) (weakenTy T18 (domainEnv G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma lookup_etvar_wfindex  :
      (forall {G : Env} {X17 : (Index ty)} {S7 : Ty} {S8 : Ty} ,
        (lookup_etvar G X17 S7 S8) -> (wfindex (domainEnv G) X17)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
  End Lookups.
  Inductive shift_etvar : (Cutoff ty) -> Env -> Env -> Prop :=
    | shift_etvar_here {G : Env}
        {T17 : Ty} {T18 : Ty} :
        (shift_etvar C0 G (etvar G T17 T18))
    | shiftetvar_there_etvar
        {c2 : (Cutoff ty)} {G : Env}
        {G0 : Env} {T1 : Ty} {T2 : Ty} :
        (shift_etvar c2 G G0) -> (shift_etvar (CS c2) (etvar G T1 T2) (etvar G0 (shiftTy c2 T1) (shiftTy c2 T2))).
  Lemma weaken_shift_etvar  :
    (forall (G : Env) {c2 : (Cutoff ty)} {G0 : Env} {G1 : Env} ,
      (shift_etvar c2 G0 G1) -> (shift_etvar (weakenCutoffty c2 (domainEnv G)) (appendEnv G0 G) (appendEnv G1 (shiftEnv c2 G)))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Lemma shift_etvar_shifthvl_ty  :
    (forall {c2 : (Cutoff ty)} {G : Env} {G0 : Env} ,
      (shift_etvar c2 G G0) -> (shifthvl_ty c2 (domainEnv G) (domainEnv G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Inductive subst_etvar (G : Env) (T17 : Ty) (T18 : Ty) (S7 : Ty) : (Trace ty) -> Env -> Env -> Prop :=
    | subst_etvar_here :
        (subst_etvar G T17 T18 S7 X0 (etvar G T17 T18) G)
    | subst_etvar_there_etvar
        {d3 : (Trace ty)} {G0 : Env}
        {G1 : Env} {T1 : Ty} {T2 : Ty} :
        (subst_etvar G T17 T18 S7 d3 G0 G1) -> (subst_etvar G T17 T18 S7 (XS ty d3) (etvar G0 T1 T2) (etvar G1 (substTy d3 S7 T1) (substTy d3 S7 T2))).
  Lemma weaken_subst_etvar {G : Env} {T17 : Ty} {T18 : Ty} {S7 : Ty} :
    (forall (G0 : Env) {d3 : (Trace ty)} {G1 : Env} {G2 : Env} ,
      (subst_etvar G T17 T18 S7 d3 G1 G2) -> (subst_etvar G T17 T18 S7 (weakenTrace d3 (domainEnv G0)) (appendEnv G1 G0) (appendEnv G2 (substEnv d3 S7 G0)))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Lemma subst_etvar_substhvl_ty {G : Env} {S7 : Ty} {T17 : Ty} {T18 : Ty} :
    (forall {d3 : (Trace ty)} {G0 : Env} {G1 : Env} ,
      (subst_etvar G T17 T18 S7 d3 G0 G1) -> (substhvl_ty (domainEnv G) d3 (domainEnv G0) (domainEnv G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
End ContextStuff.
 Hint Rewrite domainEnv_shiftEnv : interaction.
 Hint Rewrite domainEnv_shiftEnv : env_domain_shift.
 Hint Rewrite domainEnv_substEnv : interaction.
 Hint Rewrite domainEnv_substEnv : env_domain_subst.
 Hint Rewrite shiftEnv_appendEnv : interaction.
 Hint Rewrite shiftEnv_appendEnv : env_shift_append.
 Hint Rewrite substEnv_appendEnv : interaction.
 Hint Rewrite substEnv_appendEnv : env_shift_append.
 Hint Constructors lookup_etvar : infra.
 Hint Constructors lookup_etvar : lookup.
 Hint Constructors lookup_etvar : shift.
 Hint Constructors lookup_etvar : subst.
 Hint Resolve weaken_lookup_etvar : infra.
 Hint Resolve weaken_lookup_etvar : lookup.
 Hint Resolve weaken_lookup_etvar : shift.
 Hint Resolve weaken_lookup_etvar : weaken.
Lemma weaken_lookup_etvar_here  :
  (forall {G : Env} (G0 : Env) {T17 : Ty} (wf : (wfTy (domainEnv G) T17)) {T18 : Ty} (wf0 : (wfTy (domainEnv G) T18)) ,
    (lookup_etvar (appendEnv (etvar G T17 T18) G0) (weakenIndexty I0 (domainEnv G0)) (weakenTy (shiftTy C0 T17) (domainEnv G0)) (weakenTy (shiftTy C0 T18) (domainEnv G0)))).
Proof.
  eauto with lookup .
Qed.
 Hint Constructors wfTy : infra.
 Hint Constructors wfTy : wf.
 Hint Extern 10 ((wfTy _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 2 ((wfTy _ _)) => match goal with
  | H : (wfTy _ (tvar _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (tarrow _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (tall _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Resolve lookup_etvar_wf : infra.
 Hint Resolve lookup_etvar_wf : wf.
 Hint Resolve lookup_etvar_wfindex : infra.
 Hint Resolve lookup_etvar_wfindex : lookup.
 Hint Resolve lookup_etvar_wfindex : wf.
 Hint Constructors shift_etvar : infra.
 Hint Constructors shift_etvar : shift.
 Hint Constructors shift_etvar : subst.
 Hint Resolve weaken_shift_etvar : infra.
 Hint Resolve weaken_shift_etvar : shift.
 Hint Resolve shift_etvar_shifthvl_ty : infra.
 Hint Resolve shift_etvar_shifthvl_ty : shift.
 Hint Resolve shift_etvar_shifthvl_ty : shift_wf.
 Hint Resolve shift_etvar_shifthvl_ty : wf.
 Hint Constructors subst_etvar : infra.
 Hint Constructors subst_etvar : subst.
 Hint Resolve weaken_subst_etvar : infra.
 Hint Resolve weaken_subst_etvar : subst.
 Hint Resolve subst_etvar_substhvl_ty : infra.
 Hint Resolve subst_etvar_substhvl_ty : subst.
 Hint Resolve subst_etvar_substhvl_ty : subst_wf.
 Hint Resolve subst_etvar_substhvl_ty : wf.
 Hint Resolve subst_etvar_substhvl_ty : substenv_substhvl.
Lemma shift_etvar_lookup_etvar  :
  (forall {c2 : (Cutoff ty)} {G : Env} {G0 : Env} (ins : (shift_etvar c2 G G0)) {X17 : (Index ty)} {T17 : Ty} {T18 : Ty} ,
    (lookup_etvar G X17 T17 T18) -> (lookup_etvar G0 (shiftIndex c2 X17) (shiftTy c2 T17) (shiftTy c2 T18))).
Proof.
  needleGenericInsertLookup .
Qed.
 Hint Resolve shift_etvar_lookup_etvar : infra.
 Hint Resolve shift_etvar_lookup_etvar : lookup.
 Hint Resolve shift_etvar_lookup_etvar : shift.
Fixpoint size_Ty (S0 : Ty) {struct S0} : nat :=
  match S0 with
    | (tvar X) => 1
    | (tarrow T1 T2) => (plus 1 (plus (size_Ty T1) (size_Ty T2)))
    | (tall T) => (plus 1 (size_Ty T))
  end.
Lemma shift_size_Ty  :
  (forall (S7 : Ty) (c1 : (Cutoff ty)) ,
    ((size_Ty (shiftTy c1 S7)) =
    (size_Ty S7))).
Proof.
  apply_mutual_ind ind_Ty.
  - intros X15 c1.
  reflexivity.
  - intros T14 IHT14 T15 IHT15.
  intros c2; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT14 c2)).
  + exact ((IHT15 c2)).
  - intros T16 IHT16.
  intros c2; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT16 (CS c2))).
Qed.

 Hint Rewrite shift_size_Ty : interaction.
 Hint Rewrite shift_size_Ty : shift_size.
Lemma weaken_size_Ty  :
  (forall (k6 : Hvl) (S7 : Ty) ,
    ((size_Ty (weakenTy S7 k6)) =
    (size_Ty S7))).
Proof.
  needleGenericWeakenSize .
Qed.
 Hint Rewrite weaken_size_Ty : interaction.
 Hint Rewrite weaken_size_Ty : weaken_size.