Require Import Coq.Program.Equality.
Require Import Coq.Program.Tactics.
Require Export Needle.
Section Namespace.
  Inductive Namespace : Type :=
    | TmVar 
    | TyVar .
  Lemma eq_namespace_dec (a : Namespace) (b : Namespace) :
    (sumbool (a =
    b) (not (a =
    b))).
  Proof.
    decide equality .
  Defined.
End Namespace.

Section HVarlist.
  Inductive Hvl : Type :=
    | H0 
    | HS (a : Namespace) (k : Hvl).
  
  Fixpoint appendHvl (k : Hvl) (k0 : Hvl) {struct k0} : Hvl :=
    match k0 with
      | (H0) => k
      | (HS a k0) => (HS a (appendHvl k k0))
    end.
  
  Lemma appendHvl_assoc  :
    (forall (k : Hvl) (k0 : Hvl) (k1 : Hvl) ,
      ((appendHvl (appendHvl k k0) k1) =
      (appendHvl k (appendHvl k0 k1)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End HVarlist.

Section Index.
  Inductive Index (a : Namespace) : Type :=
    | I0 
    | IS : (Index a) -> (Index a).
  
  Global Arguments I0 [a] .
  Global Arguments IS [a] _ .
  
  Lemma eq_index_dec {a : Namespace} (i : (Index a)) (j : (Index a)) :
    (sumbool (i =
    j) (not (i =
    j))).
  Proof.
    decide equality .
  Qed.
  
  Fixpoint weakenIndexTmVar (x : (Index TmVar)) (k : Hvl) {struct k} : (Index TmVar) :=
    match k with
      | (H0) => x
      | (HS a k) => match a with
        | (TmVar) => (IS (weakenIndexTmVar x k))
        | _ => (weakenIndexTmVar x k)
      end
    end.
  Fixpoint weakenIndexTyVar (X : (Index TyVar)) (k : Hvl) {struct k} : (Index TyVar) :=
    match k with
      | (H0) => X
      | (HS a k) => match a with
        | (TyVar) => (IS (weakenIndexTyVar X k))
        | _ => (weakenIndexTyVar X k)
      end
    end.
  Lemma weakenIndexTmVar_append  :
    (forall (x : (Index TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTmVar (weakenIndexTmVar x k) k0) =
      (weakenIndexTmVar x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenIndexTyVar_append  :
    (forall (X : (Index TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTyVar (weakenIndexTyVar X k) k0) =
      (weakenIndexTyVar X (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End Index.

Section Terms.
  Inductive Kind : Set :=
    | Star 
    | KArr (K1 : Kind) (K2 : Kind).
  Scheme ind_Kind := Induction for Kind Sort Prop.
  
  Inductive Ty : Set :=
    | TVar (X : (Index TyVar))
    | TAbs (K : Kind) (T : Ty)
    | TApp (T1 : Ty) (T2 : Ty)
    | TArr (T1 : Ty) (T2 : Ty).
  Scheme ind_Ty := Induction for Ty Sort Prop.
  
  Inductive Tm : Set :=
    | Var (x : (Index TmVar))
    | Abs (T : Ty) (t : Tm)
    | App (t1 : Tm) (t2 : Tm).
  Scheme ind_Tm := Induction for Tm Sort Prop.
End Terms.

Section Functions.
  
End Functions.

Section Shift.
  Inductive Cutoff (a : Namespace) : Type :=
    | C0 
    | CS : (Cutoff a) -> (Cutoff a).
  
  Global Arguments C0 {a} .
  Global Arguments CS {a} _ .
  Fixpoint weakenCutoffTmVar (c : (Cutoff TmVar)) (k : Hvl) {struct k} : (Cutoff TmVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TmVar) => (CS (weakenCutoffTmVar c k))
        | _ => (weakenCutoffTmVar c k)
      end
    end.
  Fixpoint weakenCutoffTyVar (c : (Cutoff TyVar)) (k : Hvl) {struct k} : (Cutoff TyVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TyVar) => (CS (weakenCutoffTyVar c k))
        | _ => (weakenCutoffTyVar c k)
      end
    end.
  
  Lemma weakenCutoffTmVar_append  :
    (forall (c : (Cutoff TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTmVar (weakenCutoffTmVar c k) k0) =
      (weakenCutoffTmVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenCutoffTyVar_append  :
    (forall (c : (Cutoff TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTyVar (weakenCutoffTyVar c k) k0) =
      (weakenCutoffTyVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint shift_TmVar_Index (c : (Cutoff TmVar)) (x : (Index TmVar)) {struct c} : (Index TmVar) :=
    match c with
      | (C0) => (IS x)
      | (CS c) => match x with
        | (I0) => I0
        | (IS x) => (IS (shift_TmVar_Index c x))
      end
    end.
  Fixpoint shift_TyVar_Index (c : (Cutoff TyVar)) (X : (Index TyVar)) {struct c} : (Index TyVar) :=
    match c with
      | (C0) => (IS X)
      | (CS c) => match X with
        | (I0) => I0
        | (IS X) => (IS (shift_TyVar_Index c X))
      end
    end.
  Fixpoint shift_TyVar_Ty (c : (Cutoff TyVar)) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (TVar X) => (TVar (shift_TyVar_Index c X))
      | (TAbs K T) => (TAbs K (shift_TyVar_Ty (weakenCutoffTyVar c H0) T))
      | (TApp T1 T2) => (TApp (shift_TyVar_Ty (weakenCutoffTyVar c H0) T1) (shift_TyVar_Ty (weakenCutoffTyVar c H0) T2))
      | (TArr T0 T3) => (TArr (shift_TyVar_Ty (weakenCutoffTyVar c H0) T0) (shift_TyVar_Ty (weakenCutoffTyVar c H0) T3))
    end.
  Fixpoint shift_TmVar_Tm (c : (Cutoff TmVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var (shift_TmVar_Index c x))
      | (Abs T t) => (Abs T (shift_TmVar_Tm (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (shift_TmVar_Tm (weakenCutoffTmVar c H0) t1) (shift_TmVar_Tm (weakenCutoffTmVar c H0) t2))
    end.
  Fixpoint shift_TyVar_Tm (c : (Cutoff TyVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (Abs T t) => (Abs (shift_TyVar_Ty (weakenCutoffTyVar c H0) T) (shift_TyVar_Tm (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (shift_TyVar_Tm (weakenCutoffTyVar c H0) t1) (shift_TyVar_Tm (weakenCutoffTyVar c H0) t2))
    end.
End Shift.

Section Weaken.
  Fixpoint weakenKind (K : Kind) (k : Hvl) {struct k} : Kind :=
    match k with
      | (H0) => K
      | (HS TmVar k) => (weakenKind K k)
      | (HS TyVar k) => (weakenKind K k)
    end.
  Fixpoint weakenTy (S0 : Ty) (k : Hvl) {struct k} : Ty :=
    match k with
      | (H0) => S0
      | (HS TmVar k) => (weakenTy S0 k)
      | (HS TyVar k) => (shift_TyVar_Ty C0 (weakenTy S0 k))
    end.
  Fixpoint weakenTm (s : Tm) (k : Hvl) {struct k} : Tm :=
    match k with
      | (H0) => s
      | (HS TmVar k) => (shift_TmVar_Tm C0 (weakenTm s k))
      | (HS TyVar k) => (shift_TyVar_Tm C0 (weakenTm s k))
    end.
End Weaken.

Section Subst.
  Inductive Trace (a : Namespace) : Type :=
    | X0 
    | XS (b : Namespace)
        (T : (Trace a)).
  
  Global Arguments X0 [a] .
  Global Arguments XS [a] b _ .
  Fixpoint weakenTrace {a : Namespace} (x : (Trace a)) (k : Hvl) {struct k} : (Trace a) :=
    match k with
      | (H0) => x
      | (HS b k) => (XS b (weakenTrace x k))
    end.
  Lemma weakenTrace_append (a : Namespace) :
    (forall (x : (Trace a)) (k : Hvl) (k0 : Hvl) ,
      ((weakenTrace (weakenTrace x k) k0) =
      (weakenTrace x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint subst_TmVar_Index (d : (Trace TmVar)) (s : Tm) (x : (Index TmVar)) {struct d} : Tm :=
    match d with
      | (X0) => match x with
        | (I0) => s
        | (IS x) => (Var x)
      end
      | (XS TmVar d) => match x with
        | (I0) => (Var I0)
        | (IS x) => (shift_TmVar_Tm C0 (subst_TmVar_Index d s x))
      end
      | (XS TyVar d) => (shift_TyVar_Tm C0 (subst_TmVar_Index d s x))
    end.
  Fixpoint subst_TyVar_Index (d : (Trace TyVar)) (S0 : Ty) (X : (Index TyVar)) {struct d} : Ty :=
    match d with
      | (X0) => match X with
        | (I0) => S0
        | (IS X) => (TVar X)
      end
      | (XS TmVar d) => (subst_TyVar_Index d S0 X)
      | (XS TyVar d) => match X with
        | (I0) => (TVar I0)
        | (IS X) => (shift_TyVar_Ty C0 (subst_TyVar_Index d S0 X))
      end
    end.
  Fixpoint subst_TyVar_Ty (d : (Trace TyVar)) (S0 : Ty) (S1 : Ty) {struct S1} : Ty :=
    match S1 with
      | (TVar X) => (subst_TyVar_Index d S0 X)
      | (TAbs K T) => (TAbs K (subst_TyVar_Ty d S0 T))
      | (TApp T1 T2) => (TApp (subst_TyVar_Ty d S0 T1) (subst_TyVar_Ty d S0 T2))
      | (TArr T0 T3) => (TArr (subst_TyVar_Ty d S0 T0) (subst_TyVar_Ty d S0 T3))
    end.
  Fixpoint subst_TmVar_Tm (d : (Trace TmVar)) (s : Tm) (s0 : Tm) {struct s0} : Tm :=
    match s0 with
      | (Var x) => (subst_TmVar_Index d s x)
      | (Abs T t) => (Abs T (subst_TmVar_Tm (XS TmVar d) s t))
      | (App t1 t2) => (App (subst_TmVar_Tm d s t1) (subst_TmVar_Tm d s t2))
    end.
  Fixpoint subst_TyVar_Tm (d : (Trace TyVar)) (S0 : Ty) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (Abs T t) => (Abs (subst_TyVar_Ty d S0 T) (subst_TyVar_Tm (XS TmVar d) S0 t))
      | (App t1 t2) => (App (subst_TyVar_Tm d S0 t1) (subst_TyVar_Tm d S0 t2))
    end.
End Subst.

Global Hint Resolve (f_equal (shift_TmVar_Tm C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TyVar_Tm C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TyVar_Ty C0)) : cong_shift0.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : interaction.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : weakencutoff_append.
 Hint Rewrite weakenTrace_append : interaction.
 Hint Rewrite weakenTrace_append : weakentrace_append.
Section SubstShiftReflection.
  Lemma subst_TmVar_Index0_shift_TmVar_Index0_reflection_ind (s : Tm) :
    (forall (k4 : Hvl) (x9 : (Index TmVar)) ,
      ((subst_TmVar_Index (weakenTrace X0 k4) s (shift_TmVar_Index (weakenCutoffTmVar C0 k4) x9)) =
      (Var x9))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Lemma subst_TyVar_Index0_shift_TyVar_Index0_reflection_ind (S0 : Ty) :
    (forall (k4 : Hvl) (X4 : (Index TyVar)) ,
      ((subst_TyVar_Index (weakenTrace X0 k4) S0 (shift_TyVar_Index (weakenCutoffTyVar C0 k4) X4)) =
      (TVar X4))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Definition subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind := (ind_Ty (fun (S2 : Ty) =>
    (forall (k7 : Hvl) (S3 : Ty) ,
      ((subst_TyVar_Ty (weakenTrace X0 k7) S3 (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) S2)) =
      S2))) (fun (X8 : (Index TyVar)) =>
    (fun (k7 : Hvl) (S2 : Ty) =>
      (subst_TyVar_Index0_shift_TyVar_Index0_reflection_ind S2 k7 X8))) (fun (K : Kind) (T : Ty) IHT4 (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 TAbs eq_refl (IHT4 k7 S2))) (fun (T1 : Ty) IHT4 (T2 : Ty) IHT5 (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 TApp (IHT4 k7 S2) (IHT5 k7 S2))) (fun (T1 : Ty) IHT4 (T2 : Ty) IHT5 (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 TArr (IHT4 k7 S2) (IHT5 k7 S2)))).
  Definition subst_TmVar_0_shift_TmVar_0_Tm_reflection_ind := (ind_Tm (fun (s0 : Tm) =>
    (forall (k7 : Hvl) (s1 : Tm) ,
      ((subst_TmVar_Tm (weakenTrace X0 k7) s1 (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) s0)) =
      s0))) (fun (x14 : (Index TmVar)) =>
    (fun (k7 : Hvl) (s0 : Tm) =>
      (subst_TmVar_Index0_shift_TmVar_Index0_reflection_ind s0 k7 x14))) (fun (T : Ty) (t : Tm) IHt17 (k7 : Hvl) (s0 : Tm) =>
    (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k7 (HS TmVar H0)) eq_refl eq_refl) (IHt17 (HS TmVar k7) s0)))) (fun (t1 : Tm) IHt17 (t2 : Tm) IHt18 (k7 : Hvl) (s0 : Tm) =>
    (f_equal2 App (IHt17 k7 s0) (IHt18 k7 s0)))).
  Definition subst_TyVar_0_shift_TyVar_0_Tm_reflection_ind := (ind_Tm (fun (s0 : Tm) =>
    (forall (k7 : Hvl) (S2 : Ty) ,
      ((subst_TyVar_Tm (weakenTrace X0 k7) S2 (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) s0)) =
      s0))) (fun (x14 : (Index TmVar)) =>
    (fun (k7 : Hvl) (S2 : Ty) =>
      eq_refl)) (fun (T : Ty) (t : Tm) IHt17 (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 Abs (let IHT4 := (subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind T) in
    (IHT4 k7 S2)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k7 (HS TmVar H0)) eq_refl eq_refl) (IHt17 (HS TmVar k7) S2)))) (fun (t1 : Tm) IHt17 (t2 : Tm) IHt18 (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 App (IHt17 k7 S2) (IHt18 k7 S2)))).
  Definition subst_TyVar_Ty0_shift_TyVar_Ty0_reflection (S3 : Ty) : (forall (S2 : Ty) ,
    ((subst_TyVar_Ty X0 S2 (shift_TyVar_Ty C0 S3)) =
    S3)) := (subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind S3 H0).
  Definition subst_TmVar_Tm0_shift_TmVar_Tm0_reflection (s1 : Tm) : (forall (s0 : Tm) ,
    ((subst_TmVar_Tm X0 s0 (shift_TmVar_Tm C0 s1)) =
    s1)) := (subst_TmVar_0_shift_TmVar_0_Tm_reflection_ind s1 H0).
  Definition subst_TyVar_Tm0_shift_TyVar_Tm0_reflection (s0 : Tm) : (forall (S2 : Ty) ,
    ((subst_TyVar_Tm X0 S2 (shift_TyVar_Tm C0 s0)) =
    s0)) := (subst_TyVar_0_shift_TyVar_0_Tm_reflection_ind s0 H0).
End SubstShiftReflection.
Section ShiftInteraction.
  Section ShiftIndexCommInd.
    Lemma shift_TmVar_Index_shift_TmVar_Index0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff TmVar)) (x : (Index TmVar)) ,
        ((shift_TmVar_Index (weakenCutoffTmVar (CS c) k) (shift_TmVar_Index (weakenCutoffTmVar C0 k) x)) =
        (shift_TmVar_Index (weakenCutoffTmVar C0 k) (shift_TmVar_Index (weakenCutoffTmVar c k) x)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
    Lemma shift_TyVar_Index_shift_TyVar_Index0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff TyVar)) (X : (Index TyVar)) ,
        ((shift_TyVar_Index (weakenCutoffTyVar (CS c) k) (shift_TyVar_Index (weakenCutoffTyVar C0 k) X)) =
        (shift_TyVar_Index (weakenCutoffTyVar C0 k) (shift_TyVar_Index (weakenCutoffTyVar c k) X)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
  End ShiftIndexCommInd.
  Section ShiftCommInd.
    Definition shift_TyVar__shift_TyVar_0_Ty_comm_ind := (ind_Ty (fun (S0 : Ty) =>
      (forall (k4 : Hvl) (c4 : (Cutoff TyVar)) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar (CS c4) k4) (shift_TyVar_Ty (weakenCutoffTyVar C0 k4) S0)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k4) (shift_TyVar_Ty (weakenCutoffTyVar c4 k4) S0))))) (fun (X4 : (Index TyVar)) =>
      (fun (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
        (f_equal TVar (shift_TyVar_Index_shift_TyVar_Index0_comm_ind k4 c4 X4)))) (fun (K : Kind) (T : Ty) IHT4 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 TAbs eq_refl (IHT4 k4 c4))) (fun (T1 : Ty) IHT4 (T2 : Ty) IHT5 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 TApp (IHT4 k4 c4) (IHT5 k4 c4))) (fun (T1 : Ty) IHT4 (T2 : Ty) IHT5 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 TArr (IHT4 k4 c4) (IHT5 k4 c4)))).
    Definition shift_TmVar__shift_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c4 : (Cutoff TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar (CS c4) k4) (shift_TmVar_Tm (weakenCutoffTmVar C0 k4) s)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k4) (shift_TmVar_Tm (weakenCutoffTmVar c4 k4) s))))) (fun (x9 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
        (f_equal Var (shift_TmVar_Index_shift_TmVar_Index0_comm_ind k4 c4 x9)))) (fun (T : Ty) (t : Tm) IHt11 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 Abs eq_refl (IHt11 (HS TmVar k4) c4))) (fun (t1 : Tm) IHt11 (t2 : Tm) IHt12 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt11 k4 c4) (IHt12 k4 c4)))).
    Definition shift_TmVar__shift_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c4 : (Cutoff TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c4 k4) (shift_TyVar_Tm (weakenCutoffTyVar C0 k4) s)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k4) (shift_TmVar_Tm (weakenCutoffTmVar c4 k4) s))))) (fun (x9 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt11 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 Abs eq_refl (IHt11 (HS TmVar k4) c4))) (fun (t1 : Tm) IHt11 (t2 : Tm) IHt12 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt11 k4 c4) (IHt12 k4 c4)))).
    Definition shift_TyVar__shift_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c4 : (Cutoff TyVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c4 k4) (shift_TmVar_Tm (weakenCutoffTmVar C0 k4) s)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k4) (shift_TyVar_Tm (weakenCutoffTyVar c4 k4) s))))) (fun (x9 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt11 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 Abs eq_refl (IHt11 (HS TmVar k4) c4))) (fun (t1 : Tm) IHt11 (t2 : Tm) IHt12 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt11 k4 c4) (IHt12 k4 c4)))).
    Definition shift_TyVar__shift_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c4 : (Cutoff TyVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar (CS c4) k4) (shift_TyVar_Tm (weakenCutoffTyVar C0 k4) s)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k4) (shift_TyVar_Tm (weakenCutoffTyVar c4 k4) s))))) (fun (x9 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt11 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 Abs (let IHT4 := (shift_TyVar__shift_TyVar_0_Ty_comm_ind T) in
      (IHT4 k4 c4)) (IHt11 (HS TmVar k4) c4))) (fun (t1 : Tm) IHt11 (t2 : Tm) IHt12 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt11 k4 c4) (IHt12 k4 c4)))).
  End ShiftCommInd.
  Section ShiftComm.
    Definition shift_TyVar__shift_TyVar_0_Ty_comm (S0 : Ty) : (forall (c4 : (Cutoff TyVar)) ,
      ((shift_TyVar_Ty (CS c4) (shift_TyVar_Ty C0 S0)) =
      (shift_TyVar_Ty C0 (shift_TyVar_Ty c4 S0)))) := (shift_TyVar__shift_TyVar_0_Ty_comm_ind S0 H0).
    Definition shift_TmVar__shift_TmVar_0_Tm_comm (s : Tm) : (forall (c4 : (Cutoff TmVar)) ,
      ((shift_TmVar_Tm (CS c4) (shift_TmVar_Tm C0 s)) =
      (shift_TmVar_Tm C0 (shift_TmVar_Tm c4 s)))) := (shift_TmVar__shift_TmVar_0_Tm_comm_ind s H0).
    Definition shift_TmVar__shift_TyVar_0_Tm_comm (s : Tm) : (forall (c4 : (Cutoff TmVar)) ,
      ((shift_TmVar_Tm c4 (shift_TyVar_Tm C0 s)) =
      (shift_TyVar_Tm C0 (shift_TmVar_Tm c4 s)))) := (shift_TmVar__shift_TyVar_0_Tm_comm_ind s H0).
    Definition shift_TyVar__shift_TmVar_0_Tm_comm (s : Tm) : (forall (c4 : (Cutoff TyVar)) ,
      ((shift_TyVar_Tm c4 (shift_TmVar_Tm C0 s)) =
      (shift_TmVar_Tm C0 (shift_TyVar_Tm c4 s)))) := (shift_TyVar__shift_TmVar_0_Tm_comm_ind s H0).
    Definition shift_TyVar__shift_TyVar_0_Tm_comm (s : Tm) : (forall (c4 : (Cutoff TyVar)) ,
      ((shift_TyVar_Tm (CS c4) (shift_TyVar_Tm C0 s)) =
      (shift_TyVar_Tm C0 (shift_TyVar_Tm c4 s)))) := (shift_TyVar__shift_TyVar_0_Tm_comm_ind s H0).
  End ShiftComm.
End ShiftInteraction.
 Hint Rewrite shift_TmVar__shift_TmVar_0_Tm_comm shift_TmVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TmVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Ty_comm : interaction.
 Hint Rewrite shift_TmVar__shift_TmVar_0_Tm_comm shift_TmVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TmVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Ty_comm : shift_shift0.
Section WeakenShift.
  Lemma weakenTy_shift_TyVar_Ty  :
    (forall (k38 : Hvl) (c9 : (Cutoff TyVar)) (S26 : Ty) ,
      ((weakenTy (shift_TyVar_Ty c9 S26) k38) =
      (shift_TyVar_Ty (weakenCutoffTyVar c9 k38) (weakenTy S26 k38)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_shift_TmVar_Tm  :
    (forall (k38 : Hvl) (c9 : (Cutoff TmVar)) (s10 : Tm) ,
      ((weakenTm (shift_TmVar_Tm c9 s10) k38) =
      (shift_TmVar_Tm (weakenCutoffTmVar c9 k38) (weakenTm s10 k38)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_shift_TyVar_Tm  :
    (forall (k38 : Hvl) (c9 : (Cutoff TyVar)) (s10 : Tm) ,
      ((weakenTm (shift_TyVar_Tm c9 s10) k38) =
      (shift_TyVar_Tm (weakenCutoffTyVar c9 k38) (weakenTm s10 k38)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
End WeakenShift.
Section ShiftSubstInteraction.
  Section ShiftSubstIndexCommInd.
    Lemma shift_TmVar_Tm_subst_TmVar_Index0_comm_ind (c4 : (Cutoff TmVar)) (s0 : Tm) :
      (forall (k7 : Hvl) (x14 : (Index TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c4 k7) (subst_TmVar_Index (weakenTrace X0 k7) s0 x14)) =
        (subst_TmVar_Index (weakenTrace X0 k7) (shift_TmVar_Tm c4 s0) (shift_TmVar_Index (weakenCutoffTmVar (CS c4) k7) x14)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TyVar_Tm_subst_TmVar_Index0_comm_ind (c4 : (Cutoff TyVar)) (s0 : Tm) :
      (forall (k7 : Hvl) (x14 : (Index TmVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c4 k7) (subst_TmVar_Index (weakenTrace X0 k7) s0 x14)) =
        (subst_TmVar_Index (weakenTrace X0 k7) (shift_TyVar_Tm c4 s0) x14))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TyVar_Ty_subst_TyVar_Index0_comm_ind (c4 : (Cutoff TyVar)) (S2 : Ty) :
      (forall (k7 : Hvl) (X8 : (Index TyVar)) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar c4 k7) (subst_TyVar_Index (weakenTrace X0 k7) S2 X8)) =
        (subst_TyVar_Index (weakenTrace X0 k7) (shift_TyVar_Ty c4 S2) (shift_TyVar_Index (weakenCutoffTyVar (CS c4) k7) X8)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End ShiftSubstIndexCommInd.
  Section ShiftSubstCommInd.
    Definition shift_TyVar__subst_TyVar_0_Ty_comm_ind := (ind_Ty (fun (S5 : Ty) =>
      (forall (k12 : Hvl) (c9 : (Cutoff TyVar)) (S6 : Ty) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar c9 k12) (subst_TyVar_Ty (weakenTrace X0 k12) S6 S5)) =
        (subst_TyVar_Ty (weakenTrace X0 k12) (shift_TyVar_Ty c9 S6) (shift_TyVar_Ty (weakenCutoffTyVar (CS c9) k12) S5))))) (fun (X11 : (Index TyVar)) =>
      (fun (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
        (shift_TyVar_Ty_subst_TyVar_Index0_comm_ind c9 S5 k12 X11))) (fun (K : Kind) (T : Ty) IHT4 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 TAbs eq_refl (IHT4 k12 c9 S5))) (fun (T1 : Ty) IHT4 (T2 : Ty) IHT5 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 TApp (IHT4 k12 c9 S5) (IHT5 k12 c9 S5))) (fun (T1 : Ty) IHT4 (T2 : Ty) IHT5 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 TArr (IHT4 k12 c9 S5) (IHT5 k12 c9 S5)))).
    Definition shift_TmVar__subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c9 : (Cutoff TmVar)) (s3 : Tm) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c9 k12) (subst_TmVar_Tm (weakenTrace X0 k12) s3 s2)) =
        (subst_TmVar_Tm (weakenTrace X0 k12) (shift_TmVar_Tm c9 s3) (shift_TmVar_Tm (weakenCutoffTmVar (CS c9) k12) s2))))) (fun (x26 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c9 : (Cutoff TmVar)) (s2 : Tm) =>
        (shift_TmVar_Tm_subst_TmVar_Index0_comm_ind c9 s2 k12 x26))) (fun (T : Ty) (t : Tm) IHt29 (k12 : Hvl) (c9 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt29 (HS TmVar k12) c9 s2) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt29 (t2 : Tm) IHt30 (k12 : Hvl) (c9 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 App (IHt29 k12 c9 s2) (IHt30 k12 c9 s2)))).
    Definition shift_TmVar__subst_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c9 k12) (subst_TyVar_Tm (weakenTrace X0 k12) S5 s2)) =
        (subst_TyVar_Tm (weakenTrace X0 k12) S5 (shift_TmVar_Tm (weakenCutoffTmVar c9 k12) s2))))) (fun (x26 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt29 (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt29 (HS TmVar k12) c9 S5) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt29 (t2 : Tm) IHt30 (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) =>
      (f_equal2 App (IHt29 k12 c9 S5) (IHt30 k12 c9 S5)))).
    Definition shift_TyVar__subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c9 : (Cutoff TyVar)) (s3 : Tm) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c9 k12) (subst_TmVar_Tm (weakenTrace X0 k12) s3 s2)) =
        (subst_TmVar_Tm (weakenTrace X0 k12) (shift_TyVar_Tm c9 s3) (shift_TyVar_Tm (weakenCutoffTyVar c9 k12) s2))))) (fun (x26 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c9 : (Cutoff TyVar)) (s2 : Tm) =>
        (shift_TyVar_Tm_subst_TmVar_Index0_comm_ind c9 s2 k12 x26))) (fun (T : Ty) (t : Tm) IHt29 (k12 : Hvl) (c9 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt29 (HS TmVar k12) c9 s2) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt29 (t2 : Tm) IHt30 (k12 : Hvl) (c9 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal2 App (IHt29 k12 c9 s2) (IHt30 k12 c9 s2)))).
    Definition shift_TyVar__subst_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c9 k12) (subst_TyVar_Tm (weakenTrace X0 k12) S5 s2)) =
        (subst_TyVar_Tm (weakenTrace X0 k12) (shift_TyVar_Ty c9 S5) (shift_TyVar_Tm (weakenCutoffTyVar (CS c9) k12) s2))))) (fun (x26 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt29 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 Abs (let IHT4 := (shift_TyVar__subst_TyVar_0_Ty_comm_ind T) in
      (IHT4 k12 c9 S5)) (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt29 (HS TmVar k12) c9 S5) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt29 (t2 : Tm) IHt30 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 App (IHt29 k12 c9 S5) (IHt30 k12 c9 S5)))).
  End ShiftSubstCommInd.
  Section ShiftSubstComm.
    Definition shift_TyVar_Ty_subst_TyVar_Ty0_comm (S6 : Ty) : (forall (c9 : (Cutoff TyVar)) (S5 : Ty) ,
      ((shift_TyVar_Ty c9 (subst_TyVar_Ty X0 S5 S6)) =
      (subst_TyVar_Ty X0 (shift_TyVar_Ty c9 S5) (shift_TyVar_Ty (CS c9) S6)))) := (shift_TyVar__subst_TyVar_0_Ty_comm_ind S6 H0).
    Definition shift_TmVar_Tm_subst_TmVar_Tm0_comm (s3 : Tm) : (forall (c9 : (Cutoff TmVar)) (s2 : Tm) ,
      ((shift_TmVar_Tm c9 (subst_TmVar_Tm X0 s2 s3)) =
      (subst_TmVar_Tm X0 (shift_TmVar_Tm c9 s2) (shift_TmVar_Tm (CS c9) s3)))) := (shift_TmVar__subst_TmVar_0_Tm_comm_ind s3 H0).
    Definition shift_TmVar_Tm_subst_TyVar_Tm0_comm (s2 : Tm) : (forall (c9 : (Cutoff TmVar)) (S5 : Ty) ,
      ((shift_TmVar_Tm c9 (subst_TyVar_Tm X0 S5 s2)) =
      (subst_TyVar_Tm X0 S5 (shift_TmVar_Tm c9 s2)))) := (shift_TmVar__subst_TyVar_0_Tm_comm_ind s2 H0).
    Definition shift_TyVar_Tm_subst_TmVar_Tm0_comm (s3 : Tm) : (forall (c9 : (Cutoff TyVar)) (s2 : Tm) ,
      ((shift_TyVar_Tm c9 (subst_TmVar_Tm X0 s2 s3)) =
      (subst_TmVar_Tm X0 (shift_TyVar_Tm c9 s2) (shift_TyVar_Tm c9 s3)))) := (shift_TyVar__subst_TmVar_0_Tm_comm_ind s3 H0).
    Definition shift_TyVar_Tm_subst_TyVar_Tm0_comm (s2 : Tm) : (forall (c9 : (Cutoff TyVar)) (S5 : Ty) ,
      ((shift_TyVar_Tm c9 (subst_TyVar_Tm X0 S5 s2)) =
      (subst_TyVar_Tm X0 (shift_TyVar_Ty c9 S5) (shift_TyVar_Tm (CS c9) s2)))) := (shift_TyVar__subst_TyVar_0_Tm_comm_ind s2 H0).
  End ShiftSubstComm.
  Section SubstShiftIndexCommInd.
    Lemma subst_TmVar_Index_shift_TmVar_Tm0_comm_ind (d : (Trace TmVar)) (s2 : Tm) :
      (forall (k12 : Hvl) (x26 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace (XS TmVar d) k12) s2 (shift_TmVar_Index (weakenCutoffTmVar C0 k12) x26)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k12) (subst_TmVar_Index (weakenTrace d k12) s2 x26)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Index_shift_TyVar_Tm0_comm_ind (d : (Trace TmVar)) (s2 : Tm) :
      (forall (k12 : Hvl) (x26 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace (XS TyVar d) k12) s2 x26) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k12) (subst_TmVar_Index (weakenTrace d k12) s2 x26)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Index_shift_TmVar_Ty0_comm_ind (d : (Trace TyVar)) (S5 : Ty) :
      (forall (k12 : Hvl) (X11 : (Index TyVar)) ,
        ((subst_TyVar_Index (weakenTrace (XS TmVar d) k12) S5 X11) =
        (subst_TyVar_Index (weakenTrace d k12) S5 X11))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Index_shift_TyVar_Ty0_comm_ind (d : (Trace TyVar)) (S5 : Ty) :
      (forall (k12 : Hvl) (X11 : (Index TyVar)) ,
        ((subst_TyVar_Index (weakenTrace (XS TyVar d) k12) S5 (shift_TyVar_Index (weakenCutoffTyVar C0 k12) X11)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k12) (subst_TyVar_Index (weakenTrace d k12) S5 X11)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End SubstShiftIndexCommInd.
  Section SubstShiftCommInd.
    Definition subst_TyVar__shift_TyVar_0_Ty_comm_ind := (ind_Ty (fun (S15 : Ty) =>
      (forall (k28 : Hvl) (d15 : (Trace TyVar)) (S16 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace (XS TyVar d15) k28) S16 (shift_TyVar_Ty (weakenCutoffTyVar C0 k28) S15)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k28) (subst_TyVar_Ty (weakenTrace d15 k28) S16 S15))))) (fun (X14 : (Index TyVar)) =>
      (fun (k28 : Hvl) (d15 : (Trace TyVar)) (S15 : Ty) =>
        (subst_TyVar_Index_shift_TyVar_Ty0_comm_ind d15 S15 k28 X14))) (fun (K : Kind) (T : Ty) IHT4 (k28 : Hvl) (d15 : (Trace TyVar)) (S15 : Ty) =>
      (f_equal2 TAbs eq_refl (IHT4 k28 d15 S15))) (fun (T1 : Ty) IHT4 (T2 : Ty) IHT5 (k28 : Hvl) (d15 : (Trace TyVar)) (S15 : Ty) =>
      (f_equal2 TApp (IHT4 k28 d15 S15) (IHT5 k28 d15 S15))) (fun (T1 : Ty) IHT4 (T2 : Ty) IHT5 (k28 : Hvl) (d15 : (Trace TyVar)) (S15 : Ty) =>
      (f_equal2 TArr (IHT4 k28 d15 S15) (IHT5 k28 d15 S15)))).
    Definition subst_TmVar__shift_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s8 : Tm) =>
      (forall (k28 : Hvl) (d15 : (Trace TmVar)) (s9 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace (XS TmVar d15) k28) s9 (shift_TmVar_Tm (weakenCutoffTmVar C0 k28) s8)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k28) (subst_TmVar_Tm (weakenTrace d15 k28) s9 s8))))) (fun (x38 : (Index TmVar)) =>
      (fun (k28 : Hvl) (d15 : (Trace TmVar)) (s8 : Tm) =>
        (subst_TmVar_Index_shift_TmVar_Tm0_comm_ind d15 s8 k28 x38))) (fun (T : Ty) (t : Tm) IHt41 (k28 : Hvl) (d15 : (Trace TmVar)) (s8 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TmVar d15) k28 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt41 (HS TmVar k28) d15 s8) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d15 k28 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt41 (t2 : Tm) IHt42 (k28 : Hvl) (d15 : (Trace TmVar)) (s8 : Tm) =>
      (f_equal2 App (IHt41 k28 d15 s8) (IHt42 k28 d15 s8)))).
    Definition subst_TmVar__shift_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s8 : Tm) =>
      (forall (k28 : Hvl) (d15 : (Trace TmVar)) (s9 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace (XS TyVar d15) k28) s9 (shift_TyVar_Tm (weakenCutoffTyVar C0 k28) s8)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k28) (subst_TmVar_Tm (weakenTrace d15 k28) s9 s8))))) (fun (x38 : (Index TmVar)) =>
      (fun (k28 : Hvl) (d15 : (Trace TmVar)) (s8 : Tm) =>
        (subst_TmVar_Index_shift_TyVar_Tm0_comm_ind d15 s8 k28 x38))) (fun (T : Ty) (t : Tm) IHt41 (k28 : Hvl) (d15 : (Trace TmVar)) (s8 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TyVar d15) k28 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt41 (HS TmVar k28) d15 s8) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d15 k28 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt41 (t2 : Tm) IHt42 (k28 : Hvl) (d15 : (Trace TmVar)) (s8 : Tm) =>
      (f_equal2 App (IHt41 k28 d15 s8) (IHt42 k28 d15 s8)))).
    Definition subst_TyVar__shift_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s8 : Tm) =>
      (forall (k28 : Hvl) (d15 : (Trace TyVar)) (S15 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace d15 k28) S15 (shift_TmVar_Tm (weakenCutoffTmVar C0 k28) s8)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k28) (subst_TyVar_Tm (weakenTrace d15 k28) S15 s8))))) (fun (x38 : (Index TmVar)) =>
      (fun (k28 : Hvl) (d15 : (Trace TyVar)) (S15 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt41 (k28 : Hvl) (d15 : (Trace TyVar)) (S15 : Ty) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d15 k28 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt41 (HS TmVar k28) d15 S15) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d15 k28 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt41 (t2 : Tm) IHt42 (k28 : Hvl) (d15 : (Trace TyVar)) (S15 : Ty) =>
      (f_equal2 App (IHt41 k28 d15 S15) (IHt42 k28 d15 S15)))).
    Definition subst_TyVar__shift_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s8 : Tm) =>
      (forall (k28 : Hvl) (d15 : (Trace TyVar)) (S15 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace (XS TyVar d15) k28) S15 (shift_TyVar_Tm (weakenCutoffTyVar C0 k28) s8)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k28) (subst_TyVar_Tm (weakenTrace d15 k28) S15 s8))))) (fun (x38 : (Index TmVar)) =>
      (fun (k28 : Hvl) (d15 : (Trace TyVar)) (S15 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt41 (k28 : Hvl) (d15 : (Trace TyVar)) (S15 : Ty) =>
      (f_equal2 Abs (let IHT4 := (subst_TyVar__shift_TyVar_0_Ty_comm_ind T) in
      (IHT4 k28 d15 S15)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TyVar d15) k28 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt41 (HS TmVar k28) d15 S15) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d15 k28 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt41 (t2 : Tm) IHt42 (k28 : Hvl) (d15 : (Trace TyVar)) (S15 : Ty) =>
      (f_equal2 App (IHt41 k28 d15 S15) (IHt42 k28 d15 S15)))).
  End SubstShiftCommInd.
  Section SubstShiftComm.
    Definition subst_TyVar_Ty_shift_TyVar_Ty0_comm (S23 : Ty) : (forall (d22 : (Trace TyVar)) (S22 : Ty) ,
      ((subst_TyVar_Ty (XS TyVar d22) S22 (shift_TyVar_Ty C0 S23)) =
      (shift_TyVar_Ty C0 (subst_TyVar_Ty d22 S22 S23)))) := (subst_TyVar__shift_TyVar_0_Ty_comm_ind S23 H0).
    Definition subst_TmVar_Tm_shift_TmVar_Tm0_comm (s9 : Tm) : (forall (d22 : (Trace TmVar)) (s8 : Tm) ,
      ((subst_TmVar_Tm (XS TmVar d22) s8 (shift_TmVar_Tm C0 s9)) =
      (shift_TmVar_Tm C0 (subst_TmVar_Tm d22 s8 s9)))) := (subst_TmVar__shift_TmVar_0_Tm_comm_ind s9 H0).
    Definition subst_TmVar_Tm_shift_TyVar_Tm0_comm (s9 : Tm) : (forall (d22 : (Trace TmVar)) (s8 : Tm) ,
      ((subst_TmVar_Tm (XS TyVar d22) s8 (shift_TyVar_Tm C0 s9)) =
      (shift_TyVar_Tm C0 (subst_TmVar_Tm d22 s8 s9)))) := (subst_TmVar__shift_TyVar_0_Tm_comm_ind s9 H0).
    Definition subst_TyVar_Tm_shift_TmVar_Tm0_comm (s8 : Tm) : (forall (d22 : (Trace TyVar)) (S22 : Ty) ,
      ((subst_TyVar_Tm d22 S22 (shift_TmVar_Tm C0 s8)) =
      (shift_TmVar_Tm C0 (subst_TyVar_Tm d22 S22 s8)))) := (subst_TyVar__shift_TmVar_0_Tm_comm_ind s8 H0).
    Definition subst_TyVar_Tm_shift_TyVar_Tm0_comm (s8 : Tm) : (forall (d22 : (Trace TyVar)) (S22 : Ty) ,
      ((subst_TyVar_Tm (XS TyVar d22) S22 (shift_TyVar_Tm C0 s8)) =
      (shift_TyVar_Tm C0 (subst_TyVar_Tm d22 S22 s8)))) := (subst_TyVar__shift_TyVar_0_Tm_comm_ind s8 H0).
  End SubstShiftComm.
  Section SubstSubordInd.
    Definition subst_TyVar__TmVar_Ty_ind := (ind_Ty (fun (S22 : Ty) =>
      (forall (k35 : Hvl) (d22 : (Trace TyVar)) (S23 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace (XS TmVar d22) k35) S23 S22) =
        (subst_TyVar_Ty (weakenTrace d22 k35) S23 S22)))) (fun (X17 : (Index TyVar)) =>
      (fun (k35 : Hvl) (d22 : (Trace TyVar)) (S22 : Ty) =>
        (subst_TyVar_Index_shift_TmVar_Ty0_comm_ind d22 S22 k35 X17))) (fun (K : Kind) (T : Ty) IHT4 (k35 : Hvl) (d22 : (Trace TyVar)) (S22 : Ty) =>
      (f_equal2 TAbs eq_refl (IHT4 k35 d22 S22))) (fun (T1 : Ty) IHT4 (T2 : Ty) IHT5 (k35 : Hvl) (d22 : (Trace TyVar)) (S22 : Ty) =>
      (f_equal2 TApp (IHT4 k35 d22 S22) (IHT5 k35 d22 S22))) (fun (T1 : Ty) IHT4 (T2 : Ty) IHT5 (k35 : Hvl) (d22 : (Trace TyVar)) (S22 : Ty) =>
      (f_equal2 TArr (IHT4 k35 d22 S22) (IHT5 k35 d22 S22)))).
    Definition subst_TyVar__TmVar_Tm_ind := (ind_Tm (fun (s8 : Tm) =>
      (forall (k35 : Hvl) (d22 : (Trace TyVar)) (S22 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace (XS TmVar d22) k35) S22 s8) =
        (subst_TyVar_Tm (weakenTrace d22 k35) S22 s8)))) (fun (x41 : (Index TmVar)) =>
      (fun (k35 : Hvl) (d22 : (Trace TyVar)) (S22 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt44 (k35 : Hvl) (d22 : (Trace TyVar)) (S22 : Ty) =>
      (f_equal2 Abs (let IHT4 := (subst_TyVar__TmVar_Ty_ind T) in
      (IHT4 k35 d22 S22)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TmVar d22) k35 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt44 (HS TmVar k35) d22 S22) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d22 k35 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt44 (t2 : Tm) IHt45 (k35 : Hvl) (d22 : (Trace TyVar)) (S22 : Ty) =>
      (f_equal2 App (IHt44 k35 d22 S22) (IHt45 k35 d22 S22)))).
  End SubstSubordInd.
  Section SubstSubord.
    Definition subst_TyVar_Ty_TmVar (S23 : Ty) : (forall (d22 : (Trace TyVar)) (S22 : Ty) ,
      ((subst_TyVar_Ty (XS TmVar d22) S22 S23) =
      (subst_TyVar_Ty d22 S22 S23))) := (subst_TyVar__TmVar_Ty_ind S23 H0).
    Definition subst_TyVar_Tm_TmVar (s8 : Tm) : (forall (d22 : (Trace TyVar)) (S22 : Ty) ,
      ((subst_TyVar_Tm (XS TmVar d22) S22 s8) =
      (subst_TyVar_Tm d22 S22 s8))) := (subst_TyVar__TmVar_Tm_ind s8 H0).
  End SubstSubord.
End ShiftSubstInteraction.
 Hint Rewrite subst_TmVar_Tm0_shift_TmVar_Tm0_reflection subst_TyVar_Tm0_shift_TyVar_Tm0_reflection subst_TyVar_Ty0_shift_TyVar_Ty0_reflection : interaction.
 Hint Rewrite subst_TmVar_Tm0_shift_TmVar_Tm0_reflection subst_TyVar_Tm0_shift_TyVar_Tm0_reflection subst_TyVar_Ty0_shift_TyVar_Ty0_reflection : reflection.
 Hint Rewrite subst_TmVar_Tm_shift_TmVar_Tm0_comm subst_TmVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Tm_shift_TmVar_Tm0_comm subst_TyVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Ty_shift_TyVar_Ty0_comm : interaction.
 Hint Rewrite subst_TmVar_Tm_shift_TmVar_Tm0_comm subst_TmVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Tm_shift_TmVar_Tm0_comm subst_TyVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Ty_shift_TyVar_Ty0_comm : subst_shift0.
 Hint Rewrite subst_TyVar_Tm_TmVar subst_TyVar_Ty_TmVar : interaction.
 Hint Rewrite subst_TyVar_Tm_TmVar subst_TyVar_Ty_TmVar : subst_shift0.
 Hint Rewrite shift_TmVar_Tm_subst_TmVar_Tm0_comm shift_TmVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Tm_subst_TmVar_Tm0_comm shift_TyVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Ty_subst_TyVar_Ty0_comm : interaction.
 Hint Rewrite shift_TmVar_Tm_subst_TmVar_Tm0_comm shift_TmVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Tm_subst_TmVar_Tm0_comm shift_TyVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Ty_subst_TyVar_Ty0_comm : shift_subst0.
Section SubstSubstInteraction.
  Section SubstSubstIndexCommInd.
    Lemma subst_TmVar_Tm_subst_TmVar_Index0_commright_ind (d22 : (Trace TmVar)) (s8 : Tm) (s9 : Tm) :
      (forall (k35 : Hvl) (x41 : (Index TmVar)) ,
        ((subst_TmVar_Tm (weakenTrace d22 k35) s8 (subst_TmVar_Index (weakenTrace X0 k35) s9 x41)) =
        (subst_TmVar_Tm (weakenTrace X0 k35) (subst_TmVar_Tm d22 s8 s9) (subst_TmVar_Index (weakenTrace (XS TmVar d22) k35) s8 x41)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Tm_subst_TmVar_Index0_commright_ind (d22 : (Trace TyVar)) (S22 : Ty) (s8 : Tm) :
      (forall (k35 : Hvl) (x41 : (Index TmVar)) ,
        ((subst_TyVar_Tm (weakenTrace d22 k35) S22 (subst_TmVar_Index (weakenTrace X0 k35) s8 x41)) =
        (subst_TmVar_Index (weakenTrace X0 k35) (subst_TyVar_Tm d22 S22 s8) x41))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Ty_subst_TyVar_Index0_commright_ind (d22 : (Trace TyVar)) (S22 : Ty) (S23 : Ty) :
      (forall (k35 : Hvl) (X17 : (Index TyVar)) ,
        ((subst_TyVar_Ty (weakenTrace d22 k35) S22 (subst_TyVar_Index (weakenTrace X0 k35) S23 X17)) =
        (subst_TyVar_Ty (weakenTrace X0 k35) (subst_TyVar_Ty d22 S22 S23) (subst_TyVar_Index (weakenTrace (XS TyVar d22) k35) S22 X17)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Ty_subst_TyVar_Index0_commleft_ind (d22 : (Trace TmVar)) (s8 : Tm) (S22 : Ty) :
      (forall (k35 : Hvl) (x41 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace d22 k35) s8 x41) =
        (subst_TyVar_Tm (weakenTrace X0 k35) S22 (subst_TmVar_Index (weakenTrace (XS TyVar d22) k35) s8 x41)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
  End SubstSubstIndexCommInd.
  Section SubstSubstCommInd.
    Definition subst_TyVar__subst_TyVar_0_Ty_comm_ind := (ind_Ty (fun (S26 : Ty) =>
      (forall (k38 : Hvl) (d25 : (Trace TyVar)) (S27 : Ty) (S28 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace d25 k38) S27 (subst_TyVar_Ty (weakenTrace X0 k38) S28 S26)) =
        (subst_TyVar_Ty (weakenTrace X0 k38) (subst_TyVar_Ty d25 S27 S28) (subst_TyVar_Ty (weakenTrace (XS TyVar d25) k38) S27 S26))))) (fun (X20 : (Index TyVar)) =>
      (fun (k38 : Hvl) (d25 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) =>
        (subst_TyVar_Ty_subst_TyVar_Index0_commright_ind d25 S26 S27 k38 X20))) (fun (K : Kind) (T : Ty) IHT4 (k38 : Hvl) (d25 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) =>
      (f_equal2 TAbs eq_refl (IHT4 k38 d25 S26 S27))) (fun (T1 : Ty) IHT4 (T2 : Ty) IHT5 (k38 : Hvl) (d25 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) =>
      (f_equal2 TApp (IHT4 k38 d25 S26 S27) (IHT5 k38 d25 S26 S27))) (fun (T1 : Ty) IHT4 (T2 : Ty) IHT5 (k38 : Hvl) (d25 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) =>
      (f_equal2 TArr (IHT4 k38 d25 S26 S27) (IHT5 k38 d25 S26 S27)))).
    Definition subst_TmVar__subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s10 : Tm) =>
      (forall (k38 : Hvl) (d25 : (Trace TmVar)) (s11 : Tm) (s12 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace d25 k38) s11 (subst_TmVar_Tm (weakenTrace X0 k38) s12 s10)) =
        (subst_TmVar_Tm (weakenTrace X0 k38) (subst_TmVar_Tm d25 s11 s12) (subst_TmVar_Tm (weakenTrace (XS TmVar d25) k38) s11 s10))))) (fun (x47 : (Index TmVar)) =>
      (fun (k38 : Hvl) (d25 : (Trace TmVar)) (s10 : Tm) (s11 : Tm) =>
        (subst_TmVar_Tm_subst_TmVar_Index0_commright_ind d25 s10 s11 k38 x47))) (fun (T : Ty) (t : Tm) IHt50 (k38 : Hvl) (d25 : (Trace TmVar)) (s10 : Tm) (s11 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d25 k38 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k38 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt50 (HS TmVar k38) d25 s10 s11) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k38 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TmVar d25) k38 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt50 (t2 : Tm) IHt51 (k38 : Hvl) (d25 : (Trace TmVar)) (s10 : Tm) (s11 : Tm) =>
      (f_equal2 App (IHt50 k38 d25 s10 s11) (IHt51 k38 d25 s10 s11)))).
    Definition subst_TmVar__subst_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s10 : Tm) =>
      (forall (k38 : Hvl) (d25 : (Trace TmVar)) (s11 : Tm) (S26 : Ty) ,
        ((subst_TmVar_Tm (weakenTrace d25 k38) s11 (subst_TyVar_Tm (weakenTrace X0 k38) S26 s10)) =
        (subst_TyVar_Tm (weakenTrace X0 k38) S26 (subst_TmVar_Tm (weakenTrace (XS TyVar d25) k38) s11 s10))))) (fun (x47 : (Index TmVar)) =>
      (fun (k38 : Hvl) (d25 : (Trace TmVar)) (s10 : Tm) (S26 : Ty) =>
        (subst_TmVar_Ty_subst_TyVar_Index0_commleft_ind d25 s10 S26 k38 x47))) (fun (T : Ty) (t : Tm) IHt50 (k38 : Hvl) (d25 : (Trace TmVar)) (s10 : Tm) (S26 : Ty) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d25 k38 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k38 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt50 (HS TmVar k38) d25 s10 S26) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k38 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TyVar d25) k38 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt50 (t2 : Tm) IHt51 (k38 : Hvl) (d25 : (Trace TmVar)) (s10 : Tm) (S26 : Ty) =>
      (f_equal2 App (IHt50 k38 d25 s10 S26) (IHt51 k38 d25 s10 S26)))).
    Definition subst_TyVar__subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s10 : Tm) =>
      (forall (k38 : Hvl) (d25 : (Trace TyVar)) (S26 : Ty) (s11 : Tm) ,
        ((subst_TyVar_Tm (weakenTrace d25 k38) S26 (subst_TmVar_Tm (weakenTrace X0 k38) s11 s10)) =
        (subst_TmVar_Tm (weakenTrace X0 k38) (subst_TyVar_Tm d25 S26 s11) (subst_TyVar_Tm (weakenTrace d25 k38) S26 s10))))) (fun (x47 : (Index TmVar)) =>
      (fun (k38 : Hvl) (d25 : (Trace TyVar)) (S26 : Ty) (s10 : Tm) =>
        (subst_TyVar_Tm_subst_TmVar_Index0_commright_ind d25 S26 s10 k38 x47))) (fun (T : Ty) (t : Tm) IHt50 (k38 : Hvl) (d25 : (Trace TyVar)) (S26 : Ty) (s10 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d25 k38 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k38 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt50 (HS TmVar k38) d25 S26 s10) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k38 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d25 k38 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt50 (t2 : Tm) IHt51 (k38 : Hvl) (d25 : (Trace TyVar)) (S26 : Ty) (s10 : Tm) =>
      (f_equal2 App (IHt50 k38 d25 S26 s10) (IHt51 k38 d25 S26 s10)))).
    Definition subst_TyVar__subst_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s10 : Tm) =>
      (forall (k38 : Hvl) (d25 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace d25 k38) S26 (subst_TyVar_Tm (weakenTrace X0 k38) S27 s10)) =
        (subst_TyVar_Tm (weakenTrace X0 k38) (subst_TyVar_Ty d25 S26 S27) (subst_TyVar_Tm (weakenTrace (XS TyVar d25) k38) S26 s10))))) (fun (x47 : (Index TmVar)) =>
      (fun (k38 : Hvl) (d25 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt50 (k38 : Hvl) (d25 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) =>
      (f_equal2 Abs (let IHT4 := (subst_TyVar__subst_TyVar_0_Ty_comm_ind T) in
      (IHT4 k38 d25 S26 S27)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d25 k38 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k38 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt50 (HS TmVar k38) d25 S26 S27) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k38 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar (XS TyVar d25) k38 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt50 (t2 : Tm) IHt51 (k38 : Hvl) (d25 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) =>
      (f_equal2 App (IHt50 k38 d25 S26 S27) (IHt51 k38 d25 S26 S27)))).
  End SubstSubstCommInd.
  Section SubstSubstComm.
    Definition subst_TyVar_Ty_subst_TyVar_Ty0_comm (S28 : Ty) : (forall (d25 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) ,
      ((subst_TyVar_Ty d25 S26 (subst_TyVar_Ty X0 S27 S28)) =
      (subst_TyVar_Ty X0 (subst_TyVar_Ty d25 S26 S27) (subst_TyVar_Ty (XS TyVar d25) S26 S28)))) := (subst_TyVar__subst_TyVar_0_Ty_comm_ind S28 H0).
    Definition subst_TmVar_Tm_subst_TmVar_Tm0_comm (s12 : Tm) : (forall (d25 : (Trace TmVar)) (s10 : Tm) (s11 : Tm) ,
      ((subst_TmVar_Tm d25 s10 (subst_TmVar_Tm X0 s11 s12)) =
      (subst_TmVar_Tm X0 (subst_TmVar_Tm d25 s10 s11) (subst_TmVar_Tm (XS TmVar d25) s10 s12)))) := (subst_TmVar__subst_TmVar_0_Tm_comm_ind s12 H0).
    Definition subst_TmVar_Tm_subst_TyVar_Tm0_comm (s11 : Tm) : (forall (d25 : (Trace TmVar)) (s10 : Tm) (S26 : Ty) ,
      ((subst_TmVar_Tm d25 s10 (subst_TyVar_Tm X0 S26 s11)) =
      (subst_TyVar_Tm X0 S26 (subst_TmVar_Tm (XS TyVar d25) s10 s11)))) := (subst_TmVar__subst_TyVar_0_Tm_comm_ind s11 H0).
    Definition subst_TyVar_Tm_subst_TmVar_Tm0_comm (s11 : Tm) : (forall (d25 : (Trace TyVar)) (S26 : Ty) (s10 : Tm) ,
      ((subst_TyVar_Tm d25 S26 (subst_TmVar_Tm X0 s10 s11)) =
      (subst_TmVar_Tm X0 (subst_TyVar_Tm d25 S26 s10) (subst_TyVar_Tm d25 S26 s11)))) := (subst_TyVar__subst_TmVar_0_Tm_comm_ind s11 H0).
    Definition subst_TyVar_Tm_subst_TyVar_Tm0_comm (s10 : Tm) : (forall (d25 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) ,
      ((subst_TyVar_Tm d25 S26 (subst_TyVar_Tm X0 S27 s10)) =
      (subst_TyVar_Tm X0 (subst_TyVar_Ty d25 S26 S27) (subst_TyVar_Tm (XS TyVar d25) S26 s10)))) := (subst_TyVar__subst_TyVar_0_Tm_comm_ind s10 H0).
  End SubstSubstComm.
  Section WeakenSubst.
    Lemma weakenTy_subst_TyVar_Ty  :
      (forall (k38 : Hvl) (d25 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) ,
        ((weakenTy (subst_TyVar_Ty d25 S26 S27) k38) =
        (subst_TyVar_Ty (weakenTrace d25 k38) S26 (weakenTy S27 k38)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_subst_TmVar_Tm  :
      (forall (k38 : Hvl) (d25 : (Trace TmVar)) (s10 : Tm) (s11 : Tm) ,
        ((weakenTm (subst_TmVar_Tm d25 s10 s11) k38) =
        (subst_TmVar_Tm (weakenTrace d25 k38) s10 (weakenTm s11 k38)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_subst_TyVar_Tm  :
      (forall (k38 : Hvl) (d25 : (Trace TyVar)) (S26 : Ty) (s10 : Tm) ,
        ((weakenTm (subst_TyVar_Tm d25 S26 s10) k38) =
        (subst_TyVar_Tm (weakenTrace d25 k38) S26 (weakenTm s10 k38)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
  End WeakenSubst.
  Section WeakenAppend.
    Lemma weakenKind_append  :
      (forall (K5 : Kind) (k38 : Hvl) (k39 : Hvl) ,
        ((weakenKind (weakenKind K5 k38) k39) =
        (weakenKind K5 (appendHvl k38 k39)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenTy_append  :
      (forall (S26 : Ty) (k38 : Hvl) (k39 : Hvl) ,
        ((weakenTy (weakenTy S26 k38) k39) =
        (weakenTy S26 (appendHvl k38 k39)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenTm_append  :
      (forall (s10 : Tm) (k38 : Hvl) (k39 : Hvl) ,
        ((weakenTm (weakenTm s10 k38) k39) =
        (weakenTm s10 (appendHvl k38 k39)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
  End WeakenAppend.
End SubstSubstInteraction.
 Hint Rewrite subst_TmVar_Tm_subst_TmVar_Tm0_comm subst_TyVar_Tm_subst_TyVar_Tm0_comm subst_TyVar_Ty_subst_TyVar_Ty0_comm : interaction.
 Hint Rewrite subst_TmVar_Tm_subst_TmVar_Tm0_comm subst_TyVar_Tm_subst_TyVar_Tm0_comm subst_TyVar_Ty_subst_TyVar_Ty0_comm : subst_subst0.
 Hint Rewrite weakenTm_shift_TmVar_Tm weakenTm_shift_TyVar_Tm weakenTy_shift_TyVar_Ty : weaken_shift.
 Hint Rewrite weakenTm_subst_TmVar_Tm weakenTm_subst_TyVar_Tm weakenTy_subst_TyVar_Ty : weaken_subst.
Section WellFormed.
  Fixpoint wfindex {a : Namespace} (k38 : Hvl) (i : (Index a)) {struct k38} : Prop :=
    match k38 with
      | (H0) => False
      | (HS b k38) => match (eq_namespace_dec a b) with
        | (left e) => match i with
          | (I0) => True
          | (IS i) => (wfindex k38 i)
        end
        | (right n) => (wfindex k38 i)
      end
    end.
  Inductive wfKind (k38 : Hvl) : Kind -> Prop :=
    | wfKind_Star :
        (wfKind k38 (Star))
    | wfKind_KArr {K6 : Kind}
        (wf : (wfKind k38 K6))
        {K7 : Kind}
        (wf0 : (wfKind k38 K7)) :
        (wfKind k38 (KArr K6 K7)).
  Inductive wfTy (k38 : Hvl) : Ty -> Prop :=
    | wfTy_TVar
        {X24 : (Index TyVar)}
        (wfi : (wfindex k38 X24)) :
        (wfTy k38 (TVar X24))
    | wfTy_TAbs {K6 : Kind}
        (wf : (wfKind k38 K6))
        {T53 : Ty}
        (wf0 : (wfTy k38 T53)) :
        (wfTy k38 (TAbs K6 T53))
    | wfTy_TApp {T54 : Ty}
        (wf : (wfTy k38 T54)) {T55 : Ty}
        (wf0 : (wfTy k38 T55)) :
        (wfTy k38 (TApp T54 T55))
    | wfTy_TArr {T56 : Ty}
        (wf : (wfTy k38 T56)) {T57 : Ty}
        (wf0 : (wfTy k38 T57)) :
        (wfTy k38 (TArr T56 T57)).
  Inductive wfTm (k38 : Hvl) : Tm -> Prop :=
    | wfTm_Var
        {x52 : (Index TmVar)}
        (wfi : (wfindex k38 x52)) :
        (wfTm k38 (Var x52))
    | wfTm_Abs {T53 : Ty}
        (wf : (wfTy k38 T53)) {t56 : Tm}
        (wf0 : (wfTm (HS TmVar k38) t56))
        : (wfTm k38 (Abs T53 t56))
    | wfTm_App {t57 : Tm}
        (wf : (wfTm k38 t57)) {t58 : Tm}
        (wf0 : (wfTm k38 t58)) :
        (wfTm k38 (App t57 t58)).
  Scheme ind_wfKind := Induction for wfKind Sort Prop.
  Scheme ind_wfTy := Induction for wfTy Sort Prop.
  Scheme ind_wfTm := Induction for wfTm Sort Prop.
End WellFormed.
Section ShiftWellFormed.
  Inductive shifthvl_TmVar : (forall (c12 : (Cutoff TmVar)) (k38 : Hvl) (k39 : Hvl) ,
    Prop) :=
    | shifthvl_TmVar_here
        (k38 : Hvl) :
        (shifthvl_TmVar C0 k38 (HS TmVar k38))
    | shifthvl_TmVar_there_TmVar
        (c12 : (Cutoff TmVar))
        (k38 : Hvl) (k39 : Hvl) :
        (shifthvl_TmVar c12 k38 k39) -> (shifthvl_TmVar (CS c12) (HS TmVar k38) (HS TmVar k39))
    | shifthvl_TmVar_there_TyVar
        (c12 : (Cutoff TmVar))
        (k38 : Hvl) (k39 : Hvl) :
        (shifthvl_TmVar c12 k38 k39) -> (shifthvl_TmVar c12 (HS TyVar k38) (HS TyVar k39)).
  Inductive shifthvl_TyVar : (forall (c12 : (Cutoff TyVar)) (k38 : Hvl) (k39 : Hvl) ,
    Prop) :=
    | shifthvl_TyVar_here
        (k38 : Hvl) :
        (shifthvl_TyVar C0 k38 (HS TyVar k38))
    | shifthvl_TyVar_there_TmVar
        (c12 : (Cutoff TyVar))
        (k38 : Hvl) (k39 : Hvl) :
        (shifthvl_TyVar c12 k38 k39) -> (shifthvl_TyVar c12 (HS TmVar k38) (HS TmVar k39))
    | shifthvl_TyVar_there_TyVar
        (c12 : (Cutoff TyVar))
        (k38 : Hvl) (k39 : Hvl) :
        (shifthvl_TyVar c12 k38 k39) -> (shifthvl_TyVar (CS c12) (HS TyVar k38) (HS TyVar k39)).
  Lemma weaken_shifthvl_TmVar  :
    (forall (k38 : Hvl) {c12 : (Cutoff TmVar)} {k39 : Hvl} {k40 : Hvl} ,
      (shifthvl_TmVar c12 k39 k40) -> (shifthvl_TmVar (weakenCutoffTmVar c12 k38) (appendHvl k39 k38) (appendHvl k40 k38))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma weaken_shifthvl_TyVar  :
    (forall (k38 : Hvl) {c12 : (Cutoff TyVar)} {k39 : Hvl} {k40 : Hvl} ,
      (shifthvl_TyVar c12 k39 k40) -> (shifthvl_TyVar (weakenCutoffTyVar c12 k38) (appendHvl k39 k38) (appendHvl k40 k38))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma shift_TmVar__wfindex_TmVar  :
    (forall (c12 : (Cutoff TmVar)) (k38 : Hvl) (k39 : Hvl) (ins : (shifthvl_TmVar c12 k38 k39)) (x52 : (Index TmVar)) ,
      (wfindex k38 x52) -> (wfindex k39 (shift_TmVar_Index c12 x52))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TmVar__wfindex_TyVar  :
    (forall (c12 : (Cutoff TmVar)) (k38 : Hvl) (k39 : Hvl) (ins : (shifthvl_TmVar c12 k38 k39)) (X24 : (Index TyVar)) ,
      (wfindex k38 X24) -> (wfindex k39 X24)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TyVar__wfindex_TmVar  :
    (forall (c12 : (Cutoff TyVar)) (k38 : Hvl) (k39 : Hvl) (ins : (shifthvl_TyVar c12 k38 k39)) (x52 : (Index TmVar)) ,
      (wfindex k38 x52) -> (wfindex k39 x52)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TyVar__wfindex_TyVar  :
    (forall (c12 : (Cutoff TyVar)) (k38 : Hvl) (k39 : Hvl) (ins : (shifthvl_TyVar c12 k38 k39)) (X24 : (Index TyVar)) ,
      (wfindex k38 X24) -> (wfindex k39 (shift_TyVar_Index c12 X24))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Definition shift_TmVar__wfKind : (forall (k38 : Hvl) ,
    (forall (K6 : Kind) (wf : (wfKind k38 K6)) ,
      (forall (c12 : (Cutoff TmVar)) (k39 : Hvl) ,
        (shifthvl_TmVar c12 k38 k39) -> (wfKind k39 K6)))) := (fun (k38 : Hvl) =>
    (ind_wfKind k38 (fun (K6 : Kind) (wf : (wfKind k38 K6)) =>
      (forall (c12 : (Cutoff TmVar)) (k39 : Hvl) ,
        (shifthvl_TmVar c12 k38 k39) -> (wfKind k39 K6))) (fun (c12 : (Cutoff TmVar)) (k39 : Hvl) (ins : (shifthvl_TmVar c12 k38 k39)) =>
      (wfKind_Star k39)) (fun (K1 : Kind) (wf : (wfKind k38 K1)) IHK1 (K2 : Kind) (wf0 : (wfKind k38 K2)) IHK2 (c12 : (Cutoff TmVar)) (k39 : Hvl) (ins : (shifthvl_TmVar c12 k38 k39)) =>
      (wfKind_KArr k39 (IHK1 c12 k39 (weaken_shifthvl_TmVar H0 ins)) (IHK2 c12 k39 (weaken_shifthvl_TmVar H0 ins)))))).
  Definition shift_TyVar__wfKind : (forall (k38 : Hvl) ,
    (forall (K6 : Kind) (wf : (wfKind k38 K6)) ,
      (forall (c12 : (Cutoff TyVar)) (k39 : Hvl) ,
        (shifthvl_TyVar c12 k38 k39) -> (wfKind k39 K6)))) := (fun (k38 : Hvl) =>
    (ind_wfKind k38 (fun (K6 : Kind) (wf : (wfKind k38 K6)) =>
      (forall (c12 : (Cutoff TyVar)) (k39 : Hvl) ,
        (shifthvl_TyVar c12 k38 k39) -> (wfKind k39 K6))) (fun (c12 : (Cutoff TyVar)) (k39 : Hvl) (ins : (shifthvl_TyVar c12 k38 k39)) =>
      (wfKind_Star k39)) (fun (K1 : Kind) (wf : (wfKind k38 K1)) IHK1 (K2 : Kind) (wf0 : (wfKind k38 K2)) IHK2 (c12 : (Cutoff TyVar)) (k39 : Hvl) (ins : (shifthvl_TyVar c12 k38 k39)) =>
      (wfKind_KArr k39 (IHK1 c12 k39 (weaken_shifthvl_TyVar H0 ins)) (IHK2 c12 k39 (weaken_shifthvl_TyVar H0 ins)))))).
  Definition shift_TmVar__wfTy : (forall (k38 : Hvl) ,
    (forall (S26 : Ty) (wf : (wfTy k38 S26)) ,
      (forall (c12 : (Cutoff TmVar)) (k39 : Hvl) ,
        (shifthvl_TmVar c12 k38 k39) -> (wfTy k39 S26)))) := (fun (k38 : Hvl) =>
    (ind_wfTy k38 (fun (S26 : Ty) (wf : (wfTy k38 S26)) =>
      (forall (c12 : (Cutoff TmVar)) (k39 : Hvl) ,
        (shifthvl_TmVar c12 k38 k39) -> (wfTy k39 S26))) (fun (X24 : (Index TyVar)) (wfi : (wfindex k38 X24)) (c12 : (Cutoff TmVar)) (k39 : Hvl) (ins : (shifthvl_TmVar c12 k38 k39)) =>
      (wfTy_TVar k39 (shift_TmVar__wfindex_TyVar c12 k38 k39 ins X24 wfi))) (fun (K : Kind) (wf : (wfKind k38 K)) (T : Ty) (wf0 : (wfTy k38 T)) IHT4 (c12 : (Cutoff TmVar)) (k39 : Hvl) (ins : (shifthvl_TmVar c12 k38 k39)) =>
      (wfTy_TAbs k39 (shift_TmVar__wfKind k38 K wf c12 k39 (weaken_shifthvl_TmVar H0 ins)) (IHT4 c12 k39 (weaken_shifthvl_TmVar H0 ins)))) (fun (T1 : Ty) (wf : (wfTy k38 T1)) IHT4 (T2 : Ty) (wf0 : (wfTy k38 T2)) IHT5 (c12 : (Cutoff TmVar)) (k39 : Hvl) (ins : (shifthvl_TmVar c12 k38 k39)) =>
      (wfTy_TApp k39 (IHT4 c12 k39 (weaken_shifthvl_TmVar H0 ins)) (IHT5 c12 k39 (weaken_shifthvl_TmVar H0 ins)))) (fun (T1 : Ty) (wf : (wfTy k38 T1)) IHT4 (T2 : Ty) (wf0 : (wfTy k38 T2)) IHT5 (c12 : (Cutoff TmVar)) (k39 : Hvl) (ins : (shifthvl_TmVar c12 k38 k39)) =>
      (wfTy_TArr k39 (IHT4 c12 k39 (weaken_shifthvl_TmVar H0 ins)) (IHT5 c12 k39 (weaken_shifthvl_TmVar H0 ins)))))).
  Definition shift_TyVar__wfTy : (forall (k38 : Hvl) ,
    (forall (S26 : Ty) (wf : (wfTy k38 S26)) ,
      (forall (c12 : (Cutoff TyVar)) (k39 : Hvl) ,
        (shifthvl_TyVar c12 k38 k39) -> (wfTy k39 (shift_TyVar_Ty c12 S26))))) := (fun (k38 : Hvl) =>
    (ind_wfTy k38 (fun (S26 : Ty) (wf : (wfTy k38 S26)) =>
      (forall (c12 : (Cutoff TyVar)) (k39 : Hvl) ,
        (shifthvl_TyVar c12 k38 k39) -> (wfTy k39 (shift_TyVar_Ty c12 S26)))) (fun (X24 : (Index TyVar)) (wfi : (wfindex k38 X24)) (c12 : (Cutoff TyVar)) (k39 : Hvl) (ins : (shifthvl_TyVar c12 k38 k39)) =>
      (wfTy_TVar k39 (shift_TyVar__wfindex_TyVar c12 k38 k39 ins X24 wfi))) (fun (K : Kind) (wf : (wfKind k38 K)) (T : Ty) (wf0 : (wfTy k38 T)) IHT4 (c12 : (Cutoff TyVar)) (k39 : Hvl) (ins : (shifthvl_TyVar c12 k38 k39)) =>
      (wfTy_TAbs k39 (shift_TyVar__wfKind k38 K wf c12 k39 (weaken_shifthvl_TyVar H0 ins)) (IHT4 c12 k39 (weaken_shifthvl_TyVar H0 ins)))) (fun (T1 : Ty) (wf : (wfTy k38 T1)) IHT4 (T2 : Ty) (wf0 : (wfTy k38 T2)) IHT5 (c12 : (Cutoff TyVar)) (k39 : Hvl) (ins : (shifthvl_TyVar c12 k38 k39)) =>
      (wfTy_TApp k39 (IHT4 c12 k39 (weaken_shifthvl_TyVar H0 ins)) (IHT5 c12 k39 (weaken_shifthvl_TyVar H0 ins)))) (fun (T1 : Ty) (wf : (wfTy k38 T1)) IHT4 (T2 : Ty) (wf0 : (wfTy k38 T2)) IHT5 (c12 : (Cutoff TyVar)) (k39 : Hvl) (ins : (shifthvl_TyVar c12 k38 k39)) =>
      (wfTy_TArr k39 (IHT4 c12 k39 (weaken_shifthvl_TyVar H0 ins)) (IHT5 c12 k39 (weaken_shifthvl_TyVar H0 ins)))))).
  Definition shift_TmVar__wfTm : (forall (k38 : Hvl) ,
    (forall (s10 : Tm) (wf : (wfTm k38 s10)) ,
      (forall (c12 : (Cutoff TmVar)) (k39 : Hvl) ,
        (shifthvl_TmVar c12 k38 k39) -> (wfTm k39 (shift_TmVar_Tm c12 s10))))) := (ind_wfTm (fun (k38 : Hvl) (s10 : Tm) (wf : (wfTm k38 s10)) =>
    (forall (c12 : (Cutoff TmVar)) (k39 : Hvl) ,
      (shifthvl_TmVar c12 k38 k39) -> (wfTm k39 (shift_TmVar_Tm c12 s10)))) (fun (k38 : Hvl) (x52 : (Index TmVar)) (wfi : (wfindex k38 x52)) (c12 : (Cutoff TmVar)) (k39 : Hvl) (ins : (shifthvl_TmVar c12 k38 k39)) =>
    (wfTm_Var k39 (shift_TmVar__wfindex_TmVar c12 k38 k39 ins x52 wfi))) (fun (k38 : Hvl) (T : Ty) (wf : (wfTy k38 T)) (t : Tm) (wf0 : (wfTm (HS TmVar k38) t)) IHt56 (c12 : (Cutoff TmVar)) (k39 : Hvl) (ins : (shifthvl_TmVar c12 k38 k39)) =>
    (wfTm_Abs k39 (shift_TmVar__wfTy k38 T wf c12 k39 (weaken_shifthvl_TmVar H0 ins)) (IHt56 (CS c12) (HS TmVar k39) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k38 : Hvl) (t1 : Tm) (wf : (wfTm k38 t1)) IHt56 (t2 : Tm) (wf0 : (wfTm k38 t2)) IHt57 (c12 : (Cutoff TmVar)) (k39 : Hvl) (ins : (shifthvl_TmVar c12 k38 k39)) =>
    (wfTm_App k39 (IHt56 c12 k39 (weaken_shifthvl_TmVar H0 ins)) (IHt57 c12 k39 (weaken_shifthvl_TmVar H0 ins))))).
  Definition shift_TyVar__wfTm : (forall (k38 : Hvl) ,
    (forall (s10 : Tm) (wf : (wfTm k38 s10)) ,
      (forall (c12 : (Cutoff TyVar)) (k39 : Hvl) ,
        (shifthvl_TyVar c12 k38 k39) -> (wfTm k39 (shift_TyVar_Tm c12 s10))))) := (ind_wfTm (fun (k38 : Hvl) (s10 : Tm) (wf : (wfTm k38 s10)) =>
    (forall (c12 : (Cutoff TyVar)) (k39 : Hvl) ,
      (shifthvl_TyVar c12 k38 k39) -> (wfTm k39 (shift_TyVar_Tm c12 s10)))) (fun (k38 : Hvl) (x52 : (Index TmVar)) (wfi : (wfindex k38 x52)) (c12 : (Cutoff TyVar)) (k39 : Hvl) (ins : (shifthvl_TyVar c12 k38 k39)) =>
    (wfTm_Var k39 (shift_TyVar__wfindex_TmVar c12 k38 k39 ins x52 wfi))) (fun (k38 : Hvl) (T : Ty) (wf : (wfTy k38 T)) (t : Tm) (wf0 : (wfTm (HS TmVar k38) t)) IHt56 (c12 : (Cutoff TyVar)) (k39 : Hvl) (ins : (shifthvl_TyVar c12 k38 k39)) =>
    (wfTm_Abs k39 (shift_TyVar__wfTy k38 T wf c12 k39 (weaken_shifthvl_TyVar H0 ins)) (IHt56 c12 (HS TmVar k39) (weaken_shifthvl_TyVar (HS TmVar H0) ins)))) (fun (k38 : Hvl) (t1 : Tm) (wf : (wfTm k38 t1)) IHt56 (t2 : Tm) (wf0 : (wfTm k38 t2)) IHt57 (c12 : (Cutoff TyVar)) (k39 : Hvl) (ins : (shifthvl_TyVar c12 k38 k39)) =>
    (wfTm_App k39 (IHt56 c12 k39 (weaken_shifthvl_TyVar H0 ins)) (IHt57 c12 k39 (weaken_shifthvl_TyVar H0 ins))))).
End ShiftWellFormed.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : infra.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : shift.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : shift_wf.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : wf.
 Hint Resolve shift_TmVar__wfKind shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfKind shift_TyVar__wfTm shift_TyVar__wfTy : infra.
 Hint Resolve shift_TmVar__wfKind shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfKind shift_TyVar__wfTm shift_TyVar__wfTy : shift.
 Hint Resolve shift_TmVar__wfKind shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfKind shift_TyVar__wfTm shift_TyVar__wfTy : shift_wf.
 Hint Resolve shift_TmVar__wfKind shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfKind shift_TyVar__wfTm shift_TyVar__wfTy : wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : infra.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift_wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : infra.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift_wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : weaken.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : wf.
Section SubstWellFormed.
  Inductive substhvl_TmVar (k38 : Hvl) : (forall (d25 : (Trace TmVar)) (k39 : Hvl) (k40 : Hvl) ,
    Prop) :=
    | substhvl_TmVar_here :
        (substhvl_TmVar k38 X0 (HS TmVar k38) k38)
    | substhvl_TmVar_there_TmVar
        {d25 : (Trace TmVar)}
        {k39 : Hvl} {k40 : Hvl} :
        (substhvl_TmVar k38 d25 k39 k40) -> (substhvl_TmVar k38 (XS TmVar d25) (HS TmVar k39) (HS TmVar k40))
    | substhvl_TmVar_there_TyVar
        {d25 : (Trace TmVar)}
        {k39 : Hvl} {k40 : Hvl} :
        (substhvl_TmVar k38 d25 k39 k40) -> (substhvl_TmVar k38 (XS TyVar d25) (HS TyVar k39) (HS TyVar k40)).
  Inductive substhvl_TyVar (k38 : Hvl) : (forall (d25 : (Trace TyVar)) (k39 : Hvl) (k40 : Hvl) ,
    Prop) :=
    | substhvl_TyVar_here :
        (substhvl_TyVar k38 X0 (HS TyVar k38) k38)
    | substhvl_TyVar_there_TmVar
        {d25 : (Trace TyVar)}
        {k39 : Hvl} {k40 : Hvl} :
        (substhvl_TyVar k38 d25 k39 k40) -> (substhvl_TyVar k38 (XS TmVar d25) (HS TmVar k39) (HS TmVar k40))
    | substhvl_TyVar_there_TyVar
        {d25 : (Trace TyVar)}
        {k39 : Hvl} {k40 : Hvl} :
        (substhvl_TyVar k38 d25 k39 k40) -> (substhvl_TyVar k38 (XS TyVar d25) (HS TyVar k39) (HS TyVar k40)).
  Lemma weaken_substhvl_TmVar  :
    (forall {k39 : Hvl} (k38 : Hvl) {d25 : (Trace TmVar)} {k40 : Hvl} {k41 : Hvl} ,
      (substhvl_TmVar k39 d25 k40 k41) -> (substhvl_TmVar k39 (weakenTrace d25 k38) (appendHvl k40 k38) (appendHvl k41 k38))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma weaken_substhvl_TyVar  :
    (forall {k39 : Hvl} (k38 : Hvl) {d25 : (Trace TyVar)} {k40 : Hvl} {k41 : Hvl} ,
      (substhvl_TyVar k39 d25 k40 k41) -> (substhvl_TyVar k39 (weakenTrace d25 k38) (appendHvl k40 k38) (appendHvl k41 k38))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma substhvl_TmVar_wfindex_TmVar {k38 : Hvl} {s10 : Tm} (wft : (wfTm k38 s10)) :
    (forall {d25 : (Trace TmVar)} {k39 : Hvl} {k40 : Hvl} ,
      (substhvl_TmVar k38 d25 k39 k40) -> (forall {x52 : (Index TmVar)} ,
        (wfindex k39 x52) -> (wfTm k40 (subst_TmVar_Index d25 s10 x52)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TyVar_wfindex_TyVar {k38 : Hvl} {S26 : Ty} (wft : (wfTy k38 S26)) :
    (forall {d25 : (Trace TyVar)} {k39 : Hvl} {k40 : Hvl} ,
      (substhvl_TyVar k38 d25 k39 k40) -> (forall {X24 : (Index TyVar)} ,
        (wfindex k39 X24) -> (wfTy k40 (subst_TyVar_Index d25 S26 X24)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TmVar_wfindex_TyVar {k38 : Hvl} :
    (forall {d25 : (Trace TmVar)} {k39 : Hvl} {k40 : Hvl} ,
      (substhvl_TmVar k38 d25 k39 k40) -> (forall {X24 : (Index TyVar)} ,
        (wfindex k39 X24) -> (wfindex k40 X24))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_TyVar_wfindex_TmVar {k38 : Hvl} :
    (forall {d25 : (Trace TyVar)} {k39 : Hvl} {k40 : Hvl} ,
      (substhvl_TyVar k38 d25 k39 k40) -> (forall {x52 : (Index TmVar)} ,
        (wfindex k39 x52) -> (wfindex k40 x52))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Definition substhvl_TmVar_wfKind {k38 : Hvl} : (forall (k39 : Hvl) ,
    (forall (K6 : Kind) (wf0 : (wfKind k39 K6)) ,
      (forall {d25 : (Trace TmVar)} {k40 : Hvl} ,
        (substhvl_TmVar k38 d25 k39 k40) -> (wfKind k40 K6)))) := (fun (k39 : Hvl) =>
    (ind_wfKind k39 (fun (K6 : Kind) (wf0 : (wfKind k39 K6)) =>
      (forall {d25 : (Trace TmVar)} {k40 : Hvl} ,
        (substhvl_TmVar k38 d25 k39 k40) -> (wfKind k40 K6))) (fun {d25 : (Trace TmVar)} {k40 : Hvl} (del : (substhvl_TmVar k38 d25 k39 k40)) =>
      (wfKind_Star k40)) (fun (K1 : Kind) (wf0 : (wfKind k39 K1)) IHK1 (K2 : Kind) (wf1 : (wfKind k39 K2)) IHK2 {d25 : (Trace TmVar)} {k40 : Hvl} (del : (substhvl_TmVar k38 d25 k39 k40)) =>
      (wfKind_KArr k40 (IHK1 (weakenTrace d25 H0) k40 (weaken_substhvl_TmVar H0 del)) (IHK2 (weakenTrace d25 H0) k40 (weaken_substhvl_TmVar H0 del)))))).
  Definition substhvl_TyVar_wfKind {k38 : Hvl} : (forall (k39 : Hvl) ,
    (forall (K6 : Kind) (wf0 : (wfKind k39 K6)) ,
      (forall {d25 : (Trace TyVar)} {k40 : Hvl} ,
        (substhvl_TyVar k38 d25 k39 k40) -> (wfKind k40 K6)))) := (fun (k39 : Hvl) =>
    (ind_wfKind k39 (fun (K6 : Kind) (wf0 : (wfKind k39 K6)) =>
      (forall {d25 : (Trace TyVar)} {k40 : Hvl} ,
        (substhvl_TyVar k38 d25 k39 k40) -> (wfKind k40 K6))) (fun {d25 : (Trace TyVar)} {k40 : Hvl} (del : (substhvl_TyVar k38 d25 k39 k40)) =>
      (wfKind_Star k40)) (fun (K1 : Kind) (wf0 : (wfKind k39 K1)) IHK1 (K2 : Kind) (wf1 : (wfKind k39 K2)) IHK2 {d25 : (Trace TyVar)} {k40 : Hvl} (del : (substhvl_TyVar k38 d25 k39 k40)) =>
      (wfKind_KArr k40 (IHK1 (weakenTrace d25 H0) k40 (weaken_substhvl_TyVar H0 del)) (IHK2 (weakenTrace d25 H0) k40 (weaken_substhvl_TyVar H0 del)))))).
  Definition substhvl_TmVar_wfTy {k38 : Hvl} : (forall (k39 : Hvl) ,
    (forall (S26 : Ty) (wf0 : (wfTy k39 S26)) ,
      (forall {d25 : (Trace TmVar)} {k40 : Hvl} ,
        (substhvl_TmVar k38 d25 k39 k40) -> (wfTy k40 S26)))) := (fun (k39 : Hvl) =>
    (ind_wfTy k39 (fun (S26 : Ty) (wf0 : (wfTy k39 S26)) =>
      (forall {d25 : (Trace TmVar)} {k40 : Hvl} ,
        (substhvl_TmVar k38 d25 k39 k40) -> (wfTy k40 S26))) (fun {X24 : (Index TyVar)} (wfi : (wfindex k39 X24)) {d25 : (Trace TmVar)} {k40 : Hvl} (del : (substhvl_TmVar k38 d25 k39 k40)) =>
      (wfTy_TVar k40 (substhvl_TmVar_wfindex_TyVar del wfi))) (fun (K : Kind) (wf0 : (wfKind k39 K)) (T : Ty) (wf1 : (wfTy k39 T)) IHT4 {d25 : (Trace TmVar)} {k40 : Hvl} (del : (substhvl_TmVar k38 d25 k39 k40)) =>
      (wfTy_TAbs k40 (substhvl_TmVar_wfKind k39 K wf0 (weaken_substhvl_TmVar H0 del)) (IHT4 (weakenTrace d25 H0) k40 (weaken_substhvl_TmVar H0 del)))) (fun (T1 : Ty) (wf0 : (wfTy k39 T1)) IHT4 (T2 : Ty) (wf1 : (wfTy k39 T2)) IHT5 {d25 : (Trace TmVar)} {k40 : Hvl} (del : (substhvl_TmVar k38 d25 k39 k40)) =>
      (wfTy_TApp k40 (IHT4 (weakenTrace d25 H0) k40 (weaken_substhvl_TmVar H0 del)) (IHT5 (weakenTrace d25 H0) k40 (weaken_substhvl_TmVar H0 del)))) (fun (T1 : Ty) (wf0 : (wfTy k39 T1)) IHT4 (T2 : Ty) (wf1 : (wfTy k39 T2)) IHT5 {d25 : (Trace TmVar)} {k40 : Hvl} (del : (substhvl_TmVar k38 d25 k39 k40)) =>
      (wfTy_TArr k40 (IHT4 (weakenTrace d25 H0) k40 (weaken_substhvl_TmVar H0 del)) (IHT5 (weakenTrace d25 H0) k40 (weaken_substhvl_TmVar H0 del)))))).
  Definition substhvl_TyVar_wfTy {k38 : Hvl} {S26 : Ty} (wf : (wfTy k38 S26)) : (forall (k39 : Hvl) ,
    (forall (S27 : Ty) (wf0 : (wfTy k39 S27)) ,
      (forall {d25 : (Trace TyVar)} {k40 : Hvl} ,
        (substhvl_TyVar k38 d25 k39 k40) -> (wfTy k40 (subst_TyVar_Ty d25 S26 S27))))) := (fun (k39 : Hvl) =>
    (ind_wfTy k39 (fun (S27 : Ty) (wf0 : (wfTy k39 S27)) =>
      (forall {d25 : (Trace TyVar)} {k40 : Hvl} ,
        (substhvl_TyVar k38 d25 k39 k40) -> (wfTy k40 (subst_TyVar_Ty d25 S26 S27)))) (fun {X24 : (Index TyVar)} (wfi : (wfindex k39 X24)) {d25 : (Trace TyVar)} {k40 : Hvl} (del : (substhvl_TyVar k38 d25 k39 k40)) =>
      (substhvl_TyVar_wfindex_TyVar wf del wfi)) (fun (K : Kind) (wf0 : (wfKind k39 K)) (T : Ty) (wf1 : (wfTy k39 T)) IHT4 {d25 : (Trace TyVar)} {k40 : Hvl} (del : (substhvl_TyVar k38 d25 k39 k40)) =>
      (wfTy_TAbs k40 (substhvl_TyVar_wfKind k39 K wf0 (weaken_substhvl_TyVar H0 del)) (IHT4 (weakenTrace d25 H0) k40 (weaken_substhvl_TyVar H0 del)))) (fun (T1 : Ty) (wf0 : (wfTy k39 T1)) IHT4 (T2 : Ty) (wf1 : (wfTy k39 T2)) IHT5 {d25 : (Trace TyVar)} {k40 : Hvl} (del : (substhvl_TyVar k38 d25 k39 k40)) =>
      (wfTy_TApp k40 (IHT4 (weakenTrace d25 H0) k40 (weaken_substhvl_TyVar H0 del)) (IHT5 (weakenTrace d25 H0) k40 (weaken_substhvl_TyVar H0 del)))) (fun (T1 : Ty) (wf0 : (wfTy k39 T1)) IHT4 (T2 : Ty) (wf1 : (wfTy k39 T2)) IHT5 {d25 : (Trace TyVar)} {k40 : Hvl} (del : (substhvl_TyVar k38 d25 k39 k40)) =>
      (wfTy_TArr k40 (IHT4 (weakenTrace d25 H0) k40 (weaken_substhvl_TyVar H0 del)) (IHT5 (weakenTrace d25 H0) k40 (weaken_substhvl_TyVar H0 del)))))).
  Definition substhvl_TmVar_wfTm {k38 : Hvl} {s10 : Tm} (wf : (wfTm k38 s10)) : (forall (k39 : Hvl) ,
    (forall (s11 : Tm) (wf0 : (wfTm k39 s11)) ,
      (forall {d25 : (Trace TmVar)} {k40 : Hvl} ,
        (substhvl_TmVar k38 d25 k39 k40) -> (wfTm k40 (subst_TmVar_Tm d25 s10 s11))))) := (ind_wfTm (fun (k39 : Hvl) (s11 : Tm) (wf0 : (wfTm k39 s11)) =>
    (forall {d25 : (Trace TmVar)} {k40 : Hvl} ,
      (substhvl_TmVar k38 d25 k39 k40) -> (wfTm k40 (subst_TmVar_Tm d25 s10 s11)))) (fun (k39 : Hvl) {x52 : (Index TmVar)} (wfi : (wfindex k39 x52)) {d25 : (Trace TmVar)} {k40 : Hvl} (del : (substhvl_TmVar k38 d25 k39 k40)) =>
    (substhvl_TmVar_wfindex_TmVar wf del wfi)) (fun (k39 : Hvl) (T : Ty) (wf0 : (wfTy k39 T)) (t : Tm) (wf1 : (wfTm (HS TmVar k39) t)) IHt56 {d25 : (Trace TmVar)} {k40 : Hvl} (del : (substhvl_TmVar k38 d25 k39 k40)) =>
    (wfTm_Abs k40 (substhvl_TmVar_wfTy k39 T wf0 (weaken_substhvl_TmVar H0 del)) (IHt56 (weakenTrace d25 (HS TmVar H0)) (HS TmVar k40) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k39 : Hvl) (t1 : Tm) (wf0 : (wfTm k39 t1)) IHt56 (t2 : Tm) (wf1 : (wfTm k39 t2)) IHt57 {d25 : (Trace TmVar)} {k40 : Hvl} (del : (substhvl_TmVar k38 d25 k39 k40)) =>
    (wfTm_App k40 (IHt56 (weakenTrace d25 H0) k40 (weaken_substhvl_TmVar H0 del)) (IHt57 (weakenTrace d25 H0) k40 (weaken_substhvl_TmVar H0 del))))).
  Definition substhvl_TyVar_wfTm {k38 : Hvl} {S26 : Ty} (wf : (wfTy k38 S26)) : (forall (k39 : Hvl) ,
    (forall (s10 : Tm) (wf0 : (wfTm k39 s10)) ,
      (forall {d25 : (Trace TyVar)} {k40 : Hvl} ,
        (substhvl_TyVar k38 d25 k39 k40) -> (wfTm k40 (subst_TyVar_Tm d25 S26 s10))))) := (ind_wfTm (fun (k39 : Hvl) (s10 : Tm) (wf0 : (wfTm k39 s10)) =>
    (forall {d25 : (Trace TyVar)} {k40 : Hvl} ,
      (substhvl_TyVar k38 d25 k39 k40) -> (wfTm k40 (subst_TyVar_Tm d25 S26 s10)))) (fun (k39 : Hvl) {x52 : (Index TmVar)} (wfi : (wfindex k39 x52)) {d25 : (Trace TyVar)} {k40 : Hvl} (del : (substhvl_TyVar k38 d25 k39 k40)) =>
    (wfTm_Var k40 (substhvl_TyVar_wfindex_TmVar del wfi))) (fun (k39 : Hvl) (T : Ty) (wf0 : (wfTy k39 T)) (t : Tm) (wf1 : (wfTm (HS TmVar k39) t)) IHt56 {d25 : (Trace TyVar)} {k40 : Hvl} (del : (substhvl_TyVar k38 d25 k39 k40)) =>
    (wfTm_Abs k40 (substhvl_TyVar_wfTy wf k39 T wf0 (weaken_substhvl_TyVar H0 del)) (IHt56 (weakenTrace d25 (HS TmVar H0)) (HS TmVar k40) (weaken_substhvl_TyVar (HS TmVar H0) del)))) (fun (k39 : Hvl) (t1 : Tm) (wf0 : (wfTm k39 t1)) IHt56 (t2 : Tm) (wf1 : (wfTm k39 t2)) IHt57 {d25 : (Trace TyVar)} {k40 : Hvl} (del : (substhvl_TyVar k38 d25 k39 k40)) =>
    (wfTm_App k40 (IHt56 (weakenTrace d25 H0) k40 (weaken_substhvl_TyVar H0 del)) (IHt57 (weakenTrace d25 H0) k40 (weaken_substhvl_TyVar H0 del))))).
End SubstWellFormed.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : infra.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst_wf.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : wf.
 Hint Resolve substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : infra.
 Hint Resolve substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst.
 Hint Resolve substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst_wf.
 Hint Resolve substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : infra.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst_wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : wf.
Section Context.
  Inductive Ctx : Type :=
    | empty 
    | evar (G : Ctx) (T : Ty)
    | etvar (G : Ctx) (K : Kind).
  Fixpoint appendCtx (G : Ctx) (G0 : Ctx) : Ctx :=
    match G0 with
      | (empty) => G
      | (evar G1 T) => (evar (appendCtx G G1) T)
      | (etvar G1 K) => (etvar (appendCtx G G1) K)
    end.
  Fixpoint domainCtx (G : Ctx) : Hvl :=
    match G with
      | (empty) => H0
      | (evar G0 T) => (HS TmVar (domainCtx G0))
      | (etvar G0 K) => (HS TyVar (domainCtx G0))
    end.
  Lemma appendCtx_assoc  :
    (forall (G : Ctx) (G0 : Ctx) (G1 : Ctx) ,
      ((appendCtx G (appendCtx G0 G1)) =
      (appendCtx (appendCtx G G0) G1))).
  Proof.
    needleGenericAppendEnvAssoc .
  Qed.
  Lemma domainCtx_appendCtx  :
    (forall (G : Ctx) (G0 : Ctx) ,
      ((domainCtx (appendCtx G G0)) =
      (appendHvl (domainCtx G) (domainCtx G0)))).
  Proof.
    needleGenericDomainEnvAppendEnv .
  Qed.
  Fixpoint shift_TmVar_Ctx (c12 : (Cutoff TmVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shift_TmVar_Ctx c12 G0) T)
      | (etvar G0 K) => (etvar (shift_TmVar_Ctx c12 G0) K)
    end.
  Fixpoint shift_TyVar_Ctx (c12 : (Cutoff TyVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shift_TyVar_Ctx c12 G0) (shift_TyVar_Ty (weakenCutoffTyVar c12 (domainCtx G0)) T))
      | (etvar G0 K) => (etvar (shift_TyVar_Ctx c12 G0) K)
    end.
  Fixpoint subst_TmVar_Ctx (d25 : (Trace TmVar)) (s10 : Tm) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (subst_TmVar_Ctx d25 s10 G0) T)
      | (etvar G0 K) => (etvar (subst_TmVar_Ctx d25 s10 G0) K)
    end.
  Fixpoint subst_TyVar_Ctx (d25 : (Trace TyVar)) (S26 : Ty) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (subst_TyVar_Ctx d25 S26 G0) (subst_TyVar_Ty (weakenTrace d25 (domainCtx G0)) S26 T))
      | (etvar G0 K) => (etvar (subst_TyVar_Ctx d25 S26 G0) K)
    end.
  Lemma domainCtx_shift_TmVar_Ctx  :
    (forall (c12 : (Cutoff TmVar)) (G : Ctx) ,
      ((domainCtx (shift_TmVar_Ctx c12 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_shift_TyVar_Ctx  :
    (forall (c12 : (Cutoff TyVar)) (G : Ctx) ,
      ((domainCtx (shift_TyVar_Ctx c12 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_subst_TmVar_Ctx  :
    (forall (d25 : (Trace TmVar)) (s10 : Tm) (G : Ctx) ,
      ((domainCtx (subst_TmVar_Ctx d25 s10 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
  Lemma domainCtx_subst_TyVar_Ctx  :
    (forall (d25 : (Trace TyVar)) (S26 : Ty) (G : Ctx) ,
      ((domainCtx (subst_TyVar_Ctx d25 S26 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
End Context.
 Hint Rewrite domainCtx_appendCtx : interaction.
 Hint Rewrite domainCtx_appendCtx : env_domain_append.
Section ContextStuff.
  Section ShiftEnvAppendEnv.
    Lemma shift_TmVar_Ctx_appendCtx  :
      (forall (c12 : (Cutoff TmVar)) (G : Ctx) (G0 : Ctx) ,
        ((shift_TmVar_Ctx c12 (appendCtx G G0)) =
        (appendCtx (shift_TmVar_Ctx c12 G) (shift_TmVar_Ctx (weakenCutoffTmVar c12 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
    Lemma shift_TyVar_Ctx_appendCtx  :
      (forall (c12 : (Cutoff TyVar)) (G : Ctx) (G0 : Ctx) ,
        ((shift_TyVar_Ctx c12 (appendCtx G G0)) =
        (appendCtx (shift_TyVar_Ctx c12 G) (shift_TyVar_Ctx (weakenCutoffTyVar c12 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
  End ShiftEnvAppendEnv.
  Section SubstEnvAppendEnv.
    Lemma subst_TmVar_Ctx_appendCtx  :
      (forall (d25 : (Trace TmVar)) (s10 : Tm) (G : Ctx) (G0 : Ctx) ,
        ((subst_TmVar_Ctx d25 s10 (appendCtx G G0)) =
        (appendCtx (subst_TmVar_Ctx d25 s10 G) (subst_TmVar_Ctx (weakenTrace d25 (domainCtx G)) s10 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
    Lemma subst_TyVar_Ctx_appendCtx  :
      (forall (d25 : (Trace TyVar)) (S26 : Ty) (G : Ctx) (G0 : Ctx) ,
        ((subst_TyVar_Ctx d25 S26 (appendCtx G G0)) =
        (appendCtx (subst_TyVar_Ctx d25 S26 G) (subst_TyVar_Ctx (weakenTrace d25 (domainCtx G)) S26 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
  End SubstEnvAppendEnv.
  Section Lookups.
    Inductive lookup_evar : Ctx -> (Index TmVar) -> Ty -> Prop :=
      | lookup_evar_here {G : Ctx}
          {T53 : Ty} :
          (wfTy (domainCtx G) T53) -> (lookup_evar (evar G T53) I0 T53)
      | lookup_evar_there_evar
          {G : Ctx} {x52 : (Index TmVar)}
          {T53 : Ty} {T54 : Ty} :
          (lookup_evar G x52 T53) -> (lookup_evar (evar G T54) (IS x52) T53)
      | lookup_evar_there_etvar
          {G : Ctx} {x52 : (Index TmVar)}
          {T53 : Ty} {K6 : Kind} :
          (lookup_evar G x52 T53) -> (lookup_evar (etvar G K6) x52 (shift_TyVar_Ty C0 T53)).
    Inductive lookup_etvar : Ctx -> (Index TyVar) -> Kind -> Prop :=
      | lookup_etvar_here {G : Ctx}
          {K6 : Kind} :
          (wfKind (domainCtx G) K6) -> (lookup_etvar (etvar G K6) I0 K6)
      | lookup_etvar_there_evar
          {G : Ctx} {X24 : (Index TyVar)}
          {K6 : Kind} {T53 : Ty} :
          (lookup_etvar G X24 K6) -> (lookup_etvar (evar G T53) X24 K6)
      | lookup_etvar_there_etvar
          {G : Ctx} {X24 : (Index TyVar)}
          {K6 : Kind} {K7 : Kind} :
          (lookup_etvar G X24 K6) -> (lookup_etvar (etvar G K7) (IS X24) K6).
    Lemma lookup_evar_inversion_here  :
      (forall (G : Ctx) (S26 : Ty) (S27 : Ty) ,
        (lookup_evar (evar G S26) I0 S27) -> (S26 =
        S27)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_etvar_inversion_here  :
      (forall (G : Ctx) (K6 : Kind) (K7 : Kind) ,
        (lookup_etvar (etvar G K6) I0 K7) -> (K6 =
        K7)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_evar_functional  :
      (forall {G : Ctx} {x52 : (Index TmVar)} ,
        (forall {S26 : Ty} ,
          (lookup_evar G x52 S26) -> (forall {S27 : Ty} ,
            (lookup_evar G x52 S27) -> (S26 =
            S27)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_etvar_functional  :
      (forall {G : Ctx} {X24 : (Index TyVar)} ,
        (forall {K6 : Kind} ,
          (lookup_etvar G X24 K6) -> (forall {K7 : Kind} ,
            (lookup_etvar G X24 K7) -> (K6 =
            K7)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_evar_wf  :
      (forall {G : Ctx} {x52 : (Index TmVar)} {S26 : Ty} ,
        (lookup_evar G x52 S26) -> (wfTy (domainCtx G) S26)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma lookup_etvar_wf  :
      (forall {G : Ctx} {X24 : (Index TyVar)} {K6 : Kind} ,
        (lookup_etvar G X24 K6) -> (wfKind (domainCtx G) K6)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma weaken_lookup_evar  :
      (forall {G : Ctx} (G0 : Ctx) {x52 : (Index TmVar)} {T53 : Ty} ,
        (lookup_evar G x52 T53) -> (lookup_evar (appendCtx G G0) (weakenIndexTmVar x52 (domainCtx G0)) (weakenTy T53 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma weaken_lookup_etvar  :
      (forall {G : Ctx} (G0 : Ctx) {X24 : (Index TyVar)} {K6 : Kind} ,
        (lookup_etvar G X24 K6) -> (lookup_etvar (appendCtx G G0) (weakenIndexTyVar X24 (domainCtx G0)) (weakenKind K6 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma lookup_evar_wfindex  :
      (forall {G : Ctx} {x52 : (Index TmVar)} {S26 : Ty} ,
        (lookup_evar G x52 S26) -> (wfindex (domainCtx G) x52)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
    Lemma lookup_etvar_wfindex  :
      (forall {G : Ctx} {X24 : (Index TyVar)} {K7 : Kind} ,
        (lookup_etvar G X24 K7) -> (wfindex (domainCtx G) X24)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
  End Lookups.
  Inductive shift_evar : (Cutoff TmVar) -> Ctx -> Ctx -> Prop :=
    | shift_evar_here {G : Ctx}
        {T53 : Ty} :
        (shift_evar C0 G (evar G T53))
    | shiftevar_there_evar
        {c12 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_evar c12 G G0) -> (shift_evar (CS c12) (evar G T) (evar G0 T))
    | shiftevar_there_etvar
        {c12 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} {K : Kind} :
        (shift_evar c12 G G0) -> (shift_evar c12 (etvar G K) (etvar G0 K)).
  Lemma weaken_shift_evar  :
    (forall (G : Ctx) {c12 : (Cutoff TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_evar c12 G0 G1) -> (shift_evar (weakenCutoffTmVar c12 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 G))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Inductive shift_etvar : (Cutoff TyVar) -> Ctx -> Ctx -> Prop :=
    | shift_etvar_here {G : Ctx}
        {K6 : Kind} :
        (shift_etvar C0 G (etvar G K6))
    | shiftetvar_there_evar
        {c12 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_etvar c12 G G0) -> (shift_etvar c12 (evar G T) (evar G0 (shift_TyVar_Ty c12 T)))
    | shiftetvar_there_etvar
        {c12 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} {K : Kind} :
        (shift_etvar c12 G G0) -> (shift_etvar (CS c12) (etvar G K) (etvar G0 K)).
  Lemma weaken_shift_etvar  :
    (forall (G : Ctx) {c12 : (Cutoff TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_etvar c12 G0 G1) -> (shift_etvar (weakenCutoffTyVar c12 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 (shift_TyVar_Ctx c12 G)))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Lemma shift_evar_shifthvl_TmVar  :
    (forall {c12 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_evar c12 G G0) -> (shifthvl_TmVar c12 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Lemma shift_etvar_shifthvl_TyVar  :
    (forall {c12 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_etvar c12 G G0) -> (shifthvl_TyVar c12 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Inductive subst_evar (G : Ctx) (T53 : Ty) (s10 : Tm) : (Trace TmVar) -> Ctx -> Ctx -> Prop :=
    | subst_evar_here :
        (subst_evar G T53 s10 X0 (evar G T53) G)
    | subst_evar_there_evar
        {d25 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_evar G T53 s10 d25 G0 G1) -> (subst_evar G T53 s10 (XS TmVar d25) (evar G0 T) (evar G1 T))
    | subst_evar_there_etvar
        {d25 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} {K : Kind} :
        (subst_evar G T53 s10 d25 G0 G1) -> (subst_evar G T53 s10 (XS TyVar d25) (etvar G0 K) (etvar G1 K)).
  Lemma weaken_subst_evar {G : Ctx} {T53 : Ty} {s10 : Tm} :
    (forall (G0 : Ctx) {d25 : (Trace TmVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_evar G T53 s10 d25 G1 G2) -> (subst_evar G T53 s10 (weakenTrace d25 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 G0))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Inductive subst_etvar (G : Ctx) (K6 : Kind) (S26 : Ty) : (Trace TyVar) -> Ctx -> Ctx -> Prop :=
    | subst_etvar_here :
        (subst_etvar G K6 S26 X0 (etvar G K6) G)
    | subst_etvar_there_evar
        {d25 : (Trace TyVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_etvar G K6 S26 d25 G0 G1) -> (subst_etvar G K6 S26 (XS TmVar d25) (evar G0 T) (evar G1 (subst_TyVar_Ty d25 S26 T)))
    | subst_etvar_there_etvar
        {d25 : (Trace TyVar)} {G0 : Ctx}
        {G1 : Ctx} {K : Kind} :
        (subst_etvar G K6 S26 d25 G0 G1) -> (subst_etvar G K6 S26 (XS TyVar d25) (etvar G0 K) (etvar G1 K)).
  Lemma weaken_subst_etvar {G : Ctx} {K6 : Kind} {S26 : Ty} :
    (forall (G0 : Ctx) {d25 : (Trace TyVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_etvar G K6 S26 d25 G1 G2) -> (subst_etvar G K6 S26 (weakenTrace d25 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 (subst_TyVar_Ctx d25 S26 G0)))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Lemma subst_evar_substhvl_TmVar {G : Ctx} {s10 : Tm} {T53 : Ty} :
    (forall {d25 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_evar G T53 s10 d25 G0 G1) -> (substhvl_TmVar (domainCtx G) d25 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
  Lemma subst_etvar_substhvl_TyVar {G : Ctx} {S26 : Ty} {K6 : Kind} :
    (forall {d25 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_etvar G K6 S26 d25 G0 G1) -> (substhvl_TyVar (domainCtx G) d25 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
End ContextStuff.
 Hint Rewrite domainCtx_shift_TyVar_Ctx : interaction.
 Hint Rewrite domainCtx_shift_TyVar_Ctx : env_domain_shift.
 Hint Rewrite domainCtx_subst_TyVar_Ctx : interaction.
 Hint Rewrite domainCtx_subst_TyVar_Ctx : env_domain_subst.
 Hint Rewrite shift_TyVar_Ctx_appendCtx : interaction.
 Hint Rewrite shift_TyVar_Ctx_appendCtx : env_shift_append.
 Hint Rewrite subst_TyVar_Ctx_appendCtx : interaction.
 Hint Rewrite subst_TyVar_Ctx_appendCtx : env_shift_append.
 Hint Constructors lookup_evar lookup_etvar : infra.
 Hint Constructors lookup_evar lookup_etvar : lookup.
 Hint Constructors lookup_evar lookup_etvar : shift.
 Hint Constructors lookup_evar lookup_etvar : subst.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : infra.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : lookup.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : shift.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : weaken.
Lemma weaken_lookup_evar_here  :
  (forall {G : Ctx} (G0 : Ctx) {T53 : Ty} (wf : (wfTy (domainCtx G) T53)) ,
    (lookup_evar (appendCtx (evar G T53) G0) (weakenIndexTmVar I0 (domainCtx G0)) (weakenTy T53 (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
Lemma weaken_lookup_etvar_here  :
  (forall {G : Ctx} (G0 : Ctx) {K6 : Kind} (wf : (wfKind (domainCtx G) K6)) ,
    (lookup_etvar (appendCtx (etvar G K6) G0) (weakenIndexTyVar I0 (domainCtx G0)) (weakenKind K6 (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
 Hint Constructors wfKind wfTm wfTy : infra.
 Hint Constructors wfKind wfTm wfTy : wf.
 Hint Extern 10 ((wfKind _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTm _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTy _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 2 ((wfKind _ _)) => match goal with
  | H : (wfKind _ (Star)) |- _ => inversion H; subst; clear H
  | H : (wfKind _ (KArr _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfTy _ _)) => match goal with
  | H : (wfTy _ (TVar _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TAbs _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TApp _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TArr _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfTm _ _)) => match goal with
  | H : (wfTm _ (Var _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Abs _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (App _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Resolve lookup_evar_wf lookup_etvar_wf : infra.
 Hint Resolve lookup_evar_wf lookup_etvar_wf : wf.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : infra.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : lookup.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : wf.
 Hint Constructors shift_evar shift_etvar : infra.
 Hint Constructors shift_evar shift_etvar : shift.
 Hint Constructors shift_evar shift_etvar : subst.
 Hint Resolve weaken_shift_evar weaken_shift_etvar : infra.
 Hint Resolve weaken_shift_evar weaken_shift_etvar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : infra.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : shift_wf.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : wf.
 Hint Constructors subst_evar subst_etvar : infra.
 Hint Constructors subst_evar subst_etvar : subst.
 Hint Resolve weaken_subst_evar weaken_subst_etvar : infra.
 Hint Resolve weaken_subst_evar weaken_subst_etvar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : infra.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : subst_wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : substenv_substhvl.
Lemma shift_evar_lookup_evar  :
  (forall {c12 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c12 G G0)) {x52 : (Index TmVar)} {T53 : Ty} ,
    (lookup_evar G x52 T53) -> (lookup_evar G0 (shift_TmVar_Index c12 x52) T53)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_evar_lookup_etvar  :
  (forall {c12 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c12 G G0)) {X24 : (Index TyVar)} {K6 : Kind} ,
    (lookup_etvar G X24 K6) -> (lookup_etvar G0 X24 K6)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_evar  :
  (forall {c12 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c12 G G0)) {x52 : (Index TmVar)} {T53 : Ty} ,
    (lookup_evar G x52 T53) -> (lookup_evar G0 x52 (shift_TyVar_Ty c12 T53))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_etvar  :
  (forall {c12 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c12 G G0)) {X24 : (Index TyVar)} {K6 : Kind} ,
    (lookup_etvar G X24 K6) -> (lookup_etvar G0 (shift_TyVar_Index c12 X24) K6)).
Proof.
  needleGenericInsertLookup .
Qed.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : infra.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : lookup.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : shift.
Lemma subst_evar_lookup_etvar {G : Ctx} {T54 : Ty} {s10 : Tm} :
  (forall {d25 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_evar G T54 s10 d25 G0 G1)) {X24 : (Index TyVar)} {K7 : Kind} ,
    (lookup_etvar G0 X24 K7) -> (lookup_etvar G1 X24 K7)).
Proof.
  needleGenericSubstEnvLookup .
Qed.
Lemma subst_etvar_lookup_evar {G : Ctx} {K7 : Kind} {S26 : Ty} (wf : (wfTy (domainCtx G) S26)) :
  (forall {d25 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_etvar G K7 S26 d25 G0 G1)) {x52 : (Index TmVar)} {T54 : Ty} ,
    (lookup_evar G0 x52 T54) -> (lookup_evar G1 x52 (subst_TyVar_Ty d25 S26 T54))).
Proof.
  needleGenericSubstEnvLookup .
Qed.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : infra.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : lookup.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : subst.
Fixpoint size_Kind (K : Kind) {struct K} : nat :=
  match K with
    | (Star) => 1
    | (KArr K1 K2) => (plus 1 (plus (size_Kind K1) (size_Kind K2)))
  end.
Fixpoint size_Ty (S0 : Ty) {struct S0} : nat :=
  match S0 with
    | (TVar X) => 1
    | (TAbs K T) => (plus 1 (plus (size_Kind K) (size_Ty T)))
    | (TApp T1 T2) => (plus 1 (plus (size_Ty T1) (size_Ty T2)))
    | (TArr T0 T3) => (plus 1 (plus (size_Ty T0) (size_Ty T3)))
  end.
Fixpoint size_Tm (s : Tm) {struct s} : nat :=
  match s with
    | (Var x) => 1
    | (Abs T t) => (plus 1 (plus (size_Ty T) (size_Tm t)))
    | (App t1 t2) => (plus 1 (plus (size_Tm t1) (size_Tm t2)))
  end.
Lemma shift_TyVar__size_Ty  :
  (forall (S26 : Ty) (c9 : (Cutoff TyVar)) ,
    ((size_Ty (shift_TyVar_Ty c9 S26)) =
    (size_Ty S26))).
Proof.
  apply_mutual_ind ind_Ty.
  - intros X21 c9.
  reflexivity.
  - intros K5 T46 IHT46.
  intros c10; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + reflexivity.
  + exact ((IHT46 c10)).
  - intros T47 IHT47 T48 IHT48.
  intros c10; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT47 c10)).
  + exact ((IHT48 c10)).
  - intros T49 IHT49 T50 IHT50.
  intros c10; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT49 c10)).
  + exact ((IHT50 c10)).
Qed.

Lemma shift_TmVar__size_Tm  :
  (forall (s10 : Tm) (c10 : (Cutoff TmVar)) ,
    ((size_Tm (shift_TmVar_Tm c10 s10)) =
    (size_Tm s10))).
Proof.
  apply_mutual_ind ind_Tm.
  - intros x48 c10.
  reflexivity.
  - intros T51 t50 IHt50.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + reflexivity.
  + exact ((IHt50 (CS c11))).
  - intros t51 IHt51 t52 IHt52.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt51 c11)).
  + exact ((IHt52 c11)).
Qed.

Lemma shift_TyVar__size_Tm  :
  (forall (s10 : Tm) (c11 : (Cutoff TyVar)) ,
    ((size_Tm (shift_TyVar_Tm c11 s10)) =
    (size_Tm s10))).
Proof.
  apply_mutual_ind ind_Tm.
  - intros X23 c11.
  reflexivity.
  - intros T52 t53 IHt53.
  pose proof ((shift_TyVar__size_Ty T52)) as IHT52.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT52 c12)).
  + exact ((IHt53 c12)).
  - intros t54 IHt54 t55 IHt55.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt54 c12)).
  + exact ((IHt55 c12)).
Qed.

 Hint Rewrite shift_TmVar__size_Tm shift_TyVar__size_Tm shift_TyVar__size_Ty : interaction.
 Hint Rewrite shift_TmVar__size_Tm shift_TyVar__size_Tm shift_TyVar__size_Ty : shift_size.
Lemma weaken_size_Kind  :
  (forall (k38 : Hvl) (K6 : Kind) ,
    ((size_Kind (weakenKind K6 k38)) =
    (size_Kind K6))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Tm  :
  (forall (k38 : Hvl) (s10 : Tm) ,
    ((size_Tm (weakenTm s10 k38)) =
    (size_Tm s10))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Ty  :
  (forall (k38 : Hvl) (S26 : Ty) ,
    ((size_Ty (weakenTy S26 k38)) =
    (size_Ty S26))).
Proof.
  needleGenericWeakenSize .
Qed.
 Hint Rewrite weaken_size_Kind weaken_size_Tm weaken_size_Ty : interaction.
 Hint Rewrite weaken_size_Kind weaken_size_Tm weaken_size_Ty : weaken_size.