Require Import Coq.Program.Equality.
Require Import Coq.Program.Tactics.
Require Export Needle.
Section Namespace.
  Inductive Namespace : Type :=
    | TmVar .
  Lemma eq_namespace_dec (a : Namespace) (b : Namespace) :
    (sumbool (a =
    b) (not (a =
    b))).
  Proof.
    decide equality .
  Defined.
End Namespace.

Section HVarlist.
  Inductive Hvl : Type :=
    | H0 
    | HS (a : Namespace) (k : Hvl).
  
  Fixpoint appendHvl (k : Hvl) (k0 : Hvl) {struct k0} : Hvl :=
    match k0 with
      | (H0) => k
      | (HS a k0) => (HS a (appendHvl k k0))
    end.
  
  Lemma appendHvl_assoc  :
    (forall (k : Hvl) (k0 : Hvl) (k1 : Hvl) ,
      ((appendHvl (appendHvl k k0) k1) =
      (appendHvl k (appendHvl k0 k1)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End HVarlist.

Section Index.
  Inductive Index (a : Namespace) : Type :=
    | I0 
    | IS : (Index a) -> (Index a).
  
  Global Arguments I0 [a] .
  Global Arguments IS [a] _ .
  
  Lemma eq_index_dec {a : Namespace} (i : (Index a)) (j : (Index a)) :
    (sumbool (i =
    j) (not (i =
    j))).
  Proof.
    decide equality .
  Qed.
  
  Fixpoint weakenIndexTmVar (x : (Index TmVar)) (k : Hvl) {struct k} : (Index TmVar) :=
    match k with
      | (H0) => x
      | (HS a k) => match a with
        | (TmVar) => (IS (weakenIndexTmVar x k))
      end
    end.
  Lemma weakenIndexTmVar_append  :
    (forall (x : (Index TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTmVar (weakenIndexTmVar x k) k0) =
      (weakenIndexTmVar x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End Index.

Section Terms.
  Inductive Tele : Set :=
    | TNil 
    | TCons (T : Tm) (d : Tele)
  with Tm : Set :=
    | Var (x : (Index TmVar))
    | Abs (d : Tele) (t : Tm)
    | Pi (d : Tele) (T : Tm)
    | App (t : Tm) (ts : Terms)
  with Terms : Set :=
    | Nil 
    | Cons (t : Tm) (ts : Terms).
  Scheme ind_Tele := Induction for Tele Sort Prop
  with ind_Tm := Induction for Tm Sort Prop
  with ind_Terms := Induction for Terms Sort Prop.
  Combined Scheme ind_Tele_Tm_Terms from ind_Tele, ind_Tm, ind_Terms.
End Terms.

Section Functions.
  Fixpoint bind (d : Tele) {struct d} : Hvl :=
    match d with
      | (TNil) => H0
      | (TCons t d) => (appendHvl (HS TmVar H0) (appendHvl (bind d) H0))
    end.
  Scheme ind_bind_Tele := Induction for Tele Sort Prop.
End Functions.

Section Shift.
  Inductive Cutoff (a : Namespace) : Type :=
    | C0 
    | CS : (Cutoff a) -> (Cutoff a).
  
  Global Arguments C0 {a} .
  Global Arguments CS {a} _ .
  Fixpoint weakenCutoffTmVar (c : (Cutoff TmVar)) (k : Hvl) {struct k} : (Cutoff TmVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TmVar) => (CS (weakenCutoffTmVar c k))
      end
    end.
  
  Lemma weakenCutoffTmVar_append  :
    (forall (c : (Cutoff TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTmVar (weakenCutoffTmVar c k) k0) =
      (weakenCutoffTmVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint shift_TmVar_Index (c : (Cutoff TmVar)) (x : (Index TmVar)) {struct c} : (Index TmVar) :=
    match c with
      | (C0) => (IS x)
      | (CS c) => match x with
        | (I0) => I0
        | (IS x) => (IS (shift_TmVar_Index c x))
      end
    end.
  Fixpoint shift_TmVar_Tele (c : (Cutoff TmVar)) (d : Tele) {struct d} : Tele :=
    match d with
      | (TNil) => (TNil)
      | (TCons T d0) => (TCons (shift_TmVar_Tm (weakenCutoffTmVar c H0) T) (shift_TmVar_Tele (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) d0))
    end
  with shift_TmVar_Tm (c : (Cutoff TmVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var (shift_TmVar_Index c x))
      | (Abs d t) => (Abs (shift_TmVar_Tele (weakenCutoffTmVar c H0) d) (shift_TmVar_Tm (weakenCutoffTmVar c (appendHvl (bind d) H0)) t))
      | (Pi d0 T) => (Pi (shift_TmVar_Tele (weakenCutoffTmVar c H0) d0) (shift_TmVar_Tm (weakenCutoffTmVar c (appendHvl (bind d0) H0)) T))
      | (App t0 ts) => (App (shift_TmVar_Tm (weakenCutoffTmVar c H0) t0) (shift_TmVar_Terms (weakenCutoffTmVar c H0) ts))
    end
  with shift_TmVar_Terms (c : (Cutoff TmVar)) (ts : Terms) {struct ts} : Terms :=
    match ts with
      | (Nil) => (Nil)
      | (Cons t ts0) => (Cons (shift_TmVar_Tm (weakenCutoffTmVar c H0) t) (shift_TmVar_Terms (weakenCutoffTmVar c H0) ts0))
    end.
End Shift.

Section Weaken.
  Fixpoint weakenTele (d : Tele) (k : Hvl) {struct k} : Tele :=
    match k with
      | (H0) => d
      | (HS TmVar k) => (shift_TmVar_Tele C0 (weakenTele d k))
    end.
  Fixpoint weakenTm (s : Tm) (k : Hvl) {struct k} : Tm :=
    match k with
      | (H0) => s
      | (HS TmVar k) => (shift_TmVar_Tm C0 (weakenTm s k))
    end.
  Fixpoint weakenTerms (ts : Terms) (k : Hvl) {struct k} : Terms :=
    match k with
      | (H0) => ts
      | (HS TmVar k) => (shift_TmVar_Terms C0 (weakenTerms ts k))
    end.
End Weaken.

Section Subst.
  Inductive Trace (a : Namespace) : Type :=
    | X0 
    | XS (b : Namespace)
        (T : (Trace a)).
  
  Global Arguments X0 [a] .
  Global Arguments XS [a] b _ .
  Fixpoint weakenTrace {a : Namespace} (x : (Trace a)) (k : Hvl) {struct k} : (Trace a) :=
    match k with
      | (H0) => x
      | (HS b k) => (XS b (weakenTrace x k))
    end.
  Lemma weakenTrace_append (a : Namespace) :
    (forall (x : (Trace a)) (k : Hvl) (k0 : Hvl) ,
      ((weakenTrace (weakenTrace x k) k0) =
      (weakenTrace x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint subst_TmVar_Index (d : (Trace TmVar)) (s : Tm) (x : (Index TmVar)) {struct d} : Tm :=
    match d with
      | (X0) => match x with
        | (I0) => s
        | (IS x) => (Var x)
      end
      | (XS TmVar d) => match x with
        | (I0) => (Var I0)
        | (IS x) => (shift_TmVar_Tm C0 (subst_TmVar_Index d s x))
      end
    end.
  Fixpoint subst_TmVar_Tele (d : (Trace TmVar)) (s : Tm) (d0 : Tele) {struct d0} : Tele :=
    match d0 with
      | (TNil) => (TNil)
      | (TCons T d1) => (TCons (subst_TmVar_Tm d s T) (subst_TmVar_Tele (XS TmVar d) s d1))
    end
  with subst_TmVar_Tm (d : (Trace TmVar)) (s : Tm) (s0 : Tm) {struct s0} : Tm :=
    match s0 with
      | (Var x) => (subst_TmVar_Index d s x)
      | (Abs d0 t) => (Abs (subst_TmVar_Tele d s d0) (subst_TmVar_Tm (weakenTrace d (bind d0)) s t))
      | (Pi d1 T) => (Pi (subst_TmVar_Tele d s d1) (subst_TmVar_Tm (weakenTrace d (bind d1)) s T))
      | (App t0 ts) => (App (subst_TmVar_Tm d s t0) (subst_TmVar_Terms d s ts))
    end
  with subst_TmVar_Terms (d : (Trace TmVar)) (s : Tm) (ts : Terms) {struct ts} : Terms :=
    match ts with
      | (Nil) => (Nil)
      | (Cons t ts0) => (Cons (subst_TmVar_Tm d s t) (subst_TmVar_Terms d s ts0))
    end.
End Subst.

Global Hint Resolve (f_equal (shift_TmVar_Tele C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TmVar_Terms C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TmVar_Tm C0)) : cong_shift0.
 Hint Rewrite weakenCutoffTmVar_append : interaction.
 Hint Rewrite weakenCutoffTmVar_append : weakencutoff_append.
 Hint Rewrite weakenTrace_append : interaction.
 Hint Rewrite weakenTrace_append : weakentrace_append.
Lemma stability_shift_TmVar__bind  :
  (forall (d : Tele) ,
    (forall (c : (Cutoff TmVar)) ,
      ((bind (shift_TmVar_Tele c d)) =
      (bind d)))).
Proof.
  apply_mutual_ind (ind_bind_Tele); simpl; intros; congruence .
Qed.
 Hint Rewrite stability_shift_TmVar__bind : interaction.
 Hint Rewrite stability_shift_TmVar__bind : stability_shift.
Lemma stability_weaken_bind  :
  (forall (k : Hvl) (d0 : Tele) ,
    ((bind (weakenTele d0 k)) =
    (bind d0))).
Proof.
  needleGenericStabilityWeaken .
Qed.
 Hint Rewrite stability_weaken_bind : interaction.
 Hint Rewrite stability_weaken_bind : stability_weaken.
Lemma stability_subst_TmVar__bind  :
  (forall (d0 : Tele) ,
    (forall (d1 : (Trace TmVar)) (s : Tm) ,
      ((bind (subst_TmVar_Tele d1 s d0)) =
      (bind d0)))).
Proof.
  apply_mutual_ind (ind_bind_Tele); simpl; intros; congruence .
Qed.
 Hint Rewrite stability_subst_TmVar__bind : interaction.
 Hint Rewrite stability_subst_TmVar__bind : stability_subst.
Section SubstShiftReflection.
  Lemma subst_TmVar_Index0_shift_TmVar_Index0_reflection_ind (s0 : Tm) :
    (forall (k0 : Hvl) (x2 : (Index TmVar)) ,
      ((subst_TmVar_Index (weakenTrace X0 k0) s0 (shift_TmVar_Index (weakenCutoffTmVar C0 k0) x2)) =
      (Var x2))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Definition subst_TmVar_0_shift_TmVar_0_Tele_Tm_Terms_reflection_ind := (ind_Tele_Tm_Terms (fun (d8 : Tele) =>
    (forall (k1 : Hvl) (s1 : Tm) ,
      ((subst_TmVar_Tele (weakenTrace X0 k1) s1 (shift_TmVar_Tele (weakenCutoffTmVar C0 k1) d8)) =
      d8))) (fun (s1 : Tm) =>
    (forall (k1 : Hvl) (s2 : Tm) ,
      ((subst_TmVar_Tm (weakenTrace X0 k1) s2 (shift_TmVar_Tm (weakenCutoffTmVar C0 k1) s1)) =
      s1))) (fun (ts3 : Terms) =>
    (forall (k1 : Hvl) (s1 : Tm) ,
      ((subst_TmVar_Terms (weakenTrace X0 k1) s1 (shift_TmVar_Terms (weakenCutoffTmVar C0 k1) ts3)) =
      ts3))) (fun (k1 : Hvl) (s1 : Tm) =>
    eq_refl) (fun (T : Tm) IHT3 (d : Tele) IHd (k1 : Hvl) (s1 : Tm) =>
    (f_equal2 TCons (IHT3 k1 s1) (eq_trans (f_equal3 subst_TmVar_Tele (weakenTrace_append TmVar X0 k1 (HS TmVar H0)) eq_refl eq_refl) (IHd (HS TmVar k1) s1)))) (fun (x5 : (Index TmVar)) =>
    (fun (k1 : Hvl) (s1 : Tm) =>
      (subst_TmVar_Index0_shift_TmVar_Index0_reflection_ind s1 k1 x5))) (fun (d : Tele) IHd (t : Tm) IHt5 (k1 : Hvl) (s1 : Tm) =>
    (f_equal2 Abs (IHd k1 s1) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenTrace_append TmVar X0 k1 (bind d))) eq_refl (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k1 (bind d)) eq_refl)) (IHt5 (appendHvl k1 (bind d)) s1)))) (fun (d : Tele) IHd (T : Tm) IHT3 (k1 : Hvl) (s1 : Tm) =>
    (f_equal2 Pi (IHd k1 s1) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenTrace_append TmVar X0 k1 (bind d))) eq_refl (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k1 (bind d)) eq_refl)) (IHT3 (appendHvl k1 (bind d)) s1)))) (fun (t : Tm) IHt5 (ts : Terms) IHts3 (k1 : Hvl) (s1 : Tm) =>
    (f_equal2 App (IHt5 k1 s1) (IHts3 k1 s1))) (fun (k1 : Hvl) (s1 : Tm) =>
    eq_refl) (fun (t : Tm) IHt5 (ts : Terms) IHts3 (k1 : Hvl) (s1 : Tm) =>
    (f_equal2 Cons (IHt5 k1 s1) (IHts3 k1 s1)))).
  Lemma subst_TmVar_0_shift_TmVar_0_Tele_reflection_ind  :
    (forall (d8 : Tele) ,
      (forall (k1 : Hvl) (s1 : Tm) ,
        ((subst_TmVar_Tele (weakenTrace X0 k1) s1 (shift_TmVar_Tele (weakenCutoffTmVar C0 k1) d8)) =
        d8))).
  Proof.
    pose proof (subst_TmVar_0_shift_TmVar_0_Tele_Tm_Terms_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_TmVar_0_shift_TmVar_0_Tm_reflection_ind  :
    (forall (s1 : Tm) ,
      (forall (k1 : Hvl) (s2 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace X0 k1) s2 (shift_TmVar_Tm (weakenCutoffTmVar C0 k1) s1)) =
        s1))).
  Proof.
    pose proof (subst_TmVar_0_shift_TmVar_0_Tele_Tm_Terms_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_TmVar_0_shift_TmVar_0_Terms_reflection_ind  :
    (forall (ts3 : Terms) ,
      (forall (k1 : Hvl) (s2 : Tm) ,
        ((subst_TmVar_Terms (weakenTrace X0 k1) s2 (shift_TmVar_Terms (weakenCutoffTmVar C0 k1) ts3)) =
        ts3))).
  Proof.
    pose proof (subst_TmVar_0_shift_TmVar_0_Tele_Tm_Terms_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Definition subst_TmVar_Tele0_shift_TmVar_Tele0_reflection (d8 : Tele) : (forall (s1 : Tm) ,
    ((subst_TmVar_Tele X0 s1 (shift_TmVar_Tele C0 d8)) =
    d8)) := (subst_TmVar_0_shift_TmVar_0_Tele_reflection_ind d8 H0).
  Definition subst_TmVar_Tm0_shift_TmVar_Tm0_reflection (s2 : Tm) : (forall (s1 : Tm) ,
    ((subst_TmVar_Tm X0 s1 (shift_TmVar_Tm C0 s2)) =
    s2)) := (subst_TmVar_0_shift_TmVar_0_Tm_reflection_ind s2 H0).
  Definition subst_TmVar_Terms0_shift_TmVar_Terms0_reflection (ts3 : Terms) : (forall (s1 : Tm) ,
    ((subst_TmVar_Terms X0 s1 (shift_TmVar_Terms C0 ts3)) =
    ts3)) := (subst_TmVar_0_shift_TmVar_0_Terms_reflection_ind ts3 H0).
End SubstShiftReflection.
Section ShiftInteraction.
  Section ShiftIndexCommInd.
    Lemma shift_TmVar_Index_shift_TmVar_Index0_comm_ind  :
      (forall (k : Hvl) (c0 : (Cutoff TmVar)) (x : (Index TmVar)) ,
        ((shift_TmVar_Index (weakenCutoffTmVar (CS c0) k) (shift_TmVar_Index (weakenCutoffTmVar C0 k) x)) =
        (shift_TmVar_Index (weakenCutoffTmVar C0 k) (shift_TmVar_Index (weakenCutoffTmVar c0 k) x)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
  End ShiftIndexCommInd.
  Section ShiftCommInd.
    Definition shift_TmVar__shift_TmVar_0_Tele_Tm_Terms_comm_ind := (ind_Tele_Tm_Terms (fun (d5 : Tele) =>
      (forall (k0 : Hvl) (c1 : (Cutoff TmVar)) ,
        ((shift_TmVar_Tele (weakenCutoffTmVar (CS c1) k0) (shift_TmVar_Tele (weakenCutoffTmVar C0 k0) d5)) =
        (shift_TmVar_Tele (weakenCutoffTmVar C0 k0) (shift_TmVar_Tele (weakenCutoffTmVar c1 k0) d5))))) (fun (s0 : Tm) =>
      (forall (k0 : Hvl) (c1 : (Cutoff TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar (CS c1) k0) (shift_TmVar_Tm (weakenCutoffTmVar C0 k0) s0)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k0) (shift_TmVar_Tm (weakenCutoffTmVar c1 k0) s0))))) (fun (ts1 : Terms) =>
      (forall (k0 : Hvl) (c1 : (Cutoff TmVar)) ,
        ((shift_TmVar_Terms (weakenCutoffTmVar (CS c1) k0) (shift_TmVar_Terms (weakenCutoffTmVar C0 k0) ts1)) =
        (shift_TmVar_Terms (weakenCutoffTmVar C0 k0) (shift_TmVar_Terms (weakenCutoffTmVar c1 k0) ts1))))) (fun (k0 : Hvl) (c1 : (Cutoff TmVar)) =>
      eq_refl) (fun (T : Tm) IHT1 (d : Tele) IHd (k0 : Hvl) (c1 : (Cutoff TmVar)) =>
      (f_equal2 TCons (IHT1 k0 c1) (IHd (HS TmVar k0) c1))) (fun (x2 : (Index TmVar)) =>
      (fun (k0 : Hvl) (c1 : (Cutoff TmVar)) =>
        (f_equal Var (shift_TmVar_Index_shift_TmVar_Index0_comm_ind k0 c1 x2)))) (fun (d : Tele) IHd (t : Tm) IHt2 (k0 : Hvl) (c1 : (Cutoff TmVar)) =>
      (f_equal2 Abs (IHd k0 c1) (eq_trans (f_equal2 shift_TmVar_Tm (eq_trans (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenCutoffTmVar_append (CS c1) k0 (bind d))) (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k0 (bind d)) eq_refl)) (eq_trans (IHt2 (appendHvl k0 (bind d)) c1) (f_equal2 shift_TmVar_Tm (eq_trans (eq_sym (weakenCutoffTmVar_append C0 k0 (bind d))) (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0)))) (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append c1 k0 (bind d))) eq_refl)))))) (fun (d : Tele) IHd (T : Tm) IHT1 (k0 : Hvl) (c1 : (Cutoff TmVar)) =>
      (f_equal2 Pi (IHd k0 c1) (eq_trans (f_equal2 shift_TmVar_Tm (eq_trans (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenCutoffTmVar_append (CS c1) k0 (bind d))) (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k0 (bind d)) eq_refl)) (eq_trans (IHT1 (appendHvl k0 (bind d)) c1) (f_equal2 shift_TmVar_Tm (eq_trans (eq_sym (weakenCutoffTmVar_append C0 k0 (bind d))) (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0)))) (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append c1 k0 (bind d))) eq_refl)))))) (fun (t : Tm) IHt2 (ts : Terms) IHts1 (k0 : Hvl) (c1 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt2 k0 c1) (IHts1 k0 c1))) (fun (k0 : Hvl) (c1 : (Cutoff TmVar)) =>
      eq_refl) (fun (t : Tm) IHt2 (ts : Terms) IHts1 (k0 : Hvl) (c1 : (Cutoff TmVar)) =>
      (f_equal2 Cons (IHt2 k0 c1) (IHts1 k0 c1)))).
    Lemma shift_TmVar__shift_TmVar_0_Tele_comm_ind  :
      (forall (d5 : Tele) ,
        (forall (k0 : Hvl) (c1 : (Cutoff TmVar)) ,
          ((shift_TmVar_Tele (weakenCutoffTmVar (CS c1) k0) (shift_TmVar_Tele (weakenCutoffTmVar C0 k0) d5)) =
          (shift_TmVar_Tele (weakenCutoffTmVar C0 k0) (shift_TmVar_Tele (weakenCutoffTmVar c1 k0) d5))))).
    Proof.
      pose proof (shift_TmVar__shift_TmVar_0_Tele_Tm_Terms_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__shift_TmVar_0_Tm_comm_ind  :
      (forall (s0 : Tm) ,
        (forall (k0 : Hvl) (c1 : (Cutoff TmVar)) ,
          ((shift_TmVar_Tm (weakenCutoffTmVar (CS c1) k0) (shift_TmVar_Tm (weakenCutoffTmVar C0 k0) s0)) =
          (shift_TmVar_Tm (weakenCutoffTmVar C0 k0) (shift_TmVar_Tm (weakenCutoffTmVar c1 k0) s0))))).
    Proof.
      pose proof (shift_TmVar__shift_TmVar_0_Tele_Tm_Terms_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__shift_TmVar_0_Terms_comm_ind  :
      (forall (ts1 : Terms) ,
        (forall (k0 : Hvl) (c1 : (Cutoff TmVar)) ,
          ((shift_TmVar_Terms (weakenCutoffTmVar (CS c1) k0) (shift_TmVar_Terms (weakenCutoffTmVar C0 k0) ts1)) =
          (shift_TmVar_Terms (weakenCutoffTmVar C0 k0) (shift_TmVar_Terms (weakenCutoffTmVar c1 k0) ts1))))).
    Proof.
      pose proof (shift_TmVar__shift_TmVar_0_Tele_Tm_Terms_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End ShiftCommInd.
  Section ShiftComm.
    Definition shift_TmVar__shift_TmVar_0_Tele_comm (d5 : Tele) : (forall (c1 : (Cutoff TmVar)) ,
      ((shift_TmVar_Tele (CS c1) (shift_TmVar_Tele C0 d5)) =
      (shift_TmVar_Tele C0 (shift_TmVar_Tele c1 d5)))) := (shift_TmVar__shift_TmVar_0_Tele_comm_ind d5 H0).
    Definition shift_TmVar__shift_TmVar_0_Tm_comm (s0 : Tm) : (forall (c1 : (Cutoff TmVar)) ,
      ((shift_TmVar_Tm (CS c1) (shift_TmVar_Tm C0 s0)) =
      (shift_TmVar_Tm C0 (shift_TmVar_Tm c1 s0)))) := (shift_TmVar__shift_TmVar_0_Tm_comm_ind s0 H0).
    Definition shift_TmVar__shift_TmVar_0_Terms_comm (ts1 : Terms) : (forall (c1 : (Cutoff TmVar)) ,
      ((shift_TmVar_Terms (CS c1) (shift_TmVar_Terms C0 ts1)) =
      (shift_TmVar_Terms C0 (shift_TmVar_Terms c1 ts1)))) := (shift_TmVar__shift_TmVar_0_Terms_comm_ind ts1 H0).
  End ShiftComm.
End ShiftInteraction.
 Hint Rewrite shift_TmVar__shift_TmVar_0_Tele_comm shift_TmVar__shift_TmVar_0_Terms_comm shift_TmVar__shift_TmVar_0_Tm_comm : interaction.
 Hint Rewrite shift_TmVar__shift_TmVar_0_Tele_comm shift_TmVar__shift_TmVar_0_Terms_comm shift_TmVar__shift_TmVar_0_Tm_comm : shift_shift0.
Section WeakenShift.
  Lemma weakenTele_shift_TmVar_Tele  :
    (forall (k11 : Hvl) (c2 : (Cutoff TmVar)) (d26 : Tele) ,
      ((weakenTele (shift_TmVar_Tele c2 d26) k11) =
      (shift_TmVar_Tele (weakenCutoffTmVar c2 k11) (weakenTele d26 k11)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_shift_TmVar_Tm  :
    (forall (k11 : Hvl) (c2 : (Cutoff TmVar)) (s12 : Tm) ,
      ((weakenTm (shift_TmVar_Tm c2 s12) k11) =
      (shift_TmVar_Tm (weakenCutoffTmVar c2 k11) (weakenTm s12 k11)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTerms_shift_TmVar_Terms  :
    (forall (k11 : Hvl) (c2 : (Cutoff TmVar)) (ts9 : Terms) ,
      ((weakenTerms (shift_TmVar_Terms c2 ts9) k11) =
      (shift_TmVar_Terms (weakenCutoffTmVar c2 k11) (weakenTerms ts9 k11)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
End WeakenShift.
Section ShiftSubstInteraction.
  Section ShiftSubstIndexCommInd.
    Lemma shift_TmVar_Tm_subst_TmVar_Index0_comm_ind (c1 : (Cutoff TmVar)) (s1 : Tm) :
      (forall (k1 : Hvl) (x5 : (Index TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c1 k1) (subst_TmVar_Index (weakenTrace X0 k1) s1 x5)) =
        (subst_TmVar_Index (weakenTrace X0 k1) (shift_TmVar_Tm c1 s1) (shift_TmVar_Index (weakenCutoffTmVar (CS c1) k1) x5)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End ShiftSubstIndexCommInd.
  Section ShiftSubstCommInd.
    Definition shift_TmVar__subst_TmVar_0_Tele_Tm_Terms_comm_ind := (ind_Tele_Tm_Terms (fun (d11 : Tele) =>
      (forall (k2 : Hvl) (c2 : (Cutoff TmVar)) (s2 : Tm) ,
        ((shift_TmVar_Tele (weakenCutoffTmVar c2 k2) (subst_TmVar_Tele (weakenTrace X0 k2) s2 d11)) =
        (subst_TmVar_Tele (weakenTrace X0 k2) (shift_TmVar_Tm c2 s2) (shift_TmVar_Tele (weakenCutoffTmVar (CS c2) k2) d11))))) (fun (s2 : Tm) =>
      (forall (k2 : Hvl) (c2 : (Cutoff TmVar)) (s3 : Tm) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c2 k2) (subst_TmVar_Tm (weakenTrace X0 k2) s3 s2)) =
        (subst_TmVar_Tm (weakenTrace X0 k2) (shift_TmVar_Tm c2 s3) (shift_TmVar_Tm (weakenCutoffTmVar (CS c2) k2) s2))))) (fun (ts5 : Terms) =>
      (forall (k2 : Hvl) (c2 : (Cutoff TmVar)) (s2 : Tm) ,
        ((shift_TmVar_Terms (weakenCutoffTmVar c2 k2) (subst_TmVar_Terms (weakenTrace X0 k2) s2 ts5)) =
        (subst_TmVar_Terms (weakenTrace X0 k2) (shift_TmVar_Tm c2 s2) (shift_TmVar_Terms (weakenCutoffTmVar (CS c2) k2) ts5))))) (fun (k2 : Hvl) (c2 : (Cutoff TmVar)) (s2 : Tm) =>
      eq_refl) (fun (T : Tm) IHT5 (d : Tele) IHd (k2 : Hvl) (c2 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 TCons (IHT5 k2 c2 s2) (eq_trans (f_equal2 shift_TmVar_Tele eq_refl (f_equal3 subst_TmVar_Tele (weakenTrace_append TmVar X0 k2 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k2) c2 s2) (f_equal3 subst_TmVar_Tele (eq_sym (weakenTrace_append TmVar X0 k2 (HS TmVar H0))) eq_refl eq_refl))))) (fun (x8 : (Index TmVar)) =>
      (fun (k2 : Hvl) (c2 : (Cutoff TmVar)) (s2 : Tm) =>
        (shift_TmVar_Tm_subst_TmVar_Index0_comm_ind c2 s2 k2 x8))) (fun (d : Tele) IHd (t : Tm) IHt8 (k2 : Hvl) (c2 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 Abs (IHd k2 c2 s2) (eq_trans (f_equal2 shift_TmVar_Tm (eq_trans (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (stability_subst_TmVar__bind _ _ _) (eq_refl H0))) (weakenCutoffTmVar_append c2 k2 (bind d))) (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k2 (bind d)) eq_refl eq_refl)) (eq_trans (IHt8 (appendHvl k2 (bind d)) c2 s2) (f_equal3 subst_TmVar_Tm (eq_trans (eq_sym (weakenTrace_append TmVar X0 k2 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0)))) eq_refl (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append (CS c2) k2 (bind d))) eq_refl)))))) (fun (d : Tele) IHd (T : Tm) IHT5 (k2 : Hvl) (c2 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 Pi (IHd k2 c2 s2) (eq_trans (f_equal2 shift_TmVar_Tm (eq_trans (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (stability_subst_TmVar__bind _ _ _) (eq_refl H0))) (weakenCutoffTmVar_append c2 k2 (bind d))) (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k2 (bind d)) eq_refl eq_refl)) (eq_trans (IHT5 (appendHvl k2 (bind d)) c2 s2) (f_equal3 subst_TmVar_Tm (eq_trans (eq_sym (weakenTrace_append TmVar X0 k2 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0)))) eq_refl (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append (CS c2) k2 (bind d))) eq_refl)))))) (fun (t : Tm) IHt8 (ts : Terms) IHts5 (k2 : Hvl) (c2 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 App (IHt8 k2 c2 s2) (IHts5 k2 c2 s2))) (fun (k2 : Hvl) (c2 : (Cutoff TmVar)) (s2 : Tm) =>
      eq_refl) (fun (t : Tm) IHt8 (ts : Terms) IHts5 (k2 : Hvl) (c2 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 Cons (IHt8 k2 c2 s2) (IHts5 k2 c2 s2)))).
    Lemma shift_TmVar__subst_TmVar_0_Tele_comm_ind  :
      (forall (d11 : Tele) ,
        (forall (k2 : Hvl) (c2 : (Cutoff TmVar)) (s2 : Tm) ,
          ((shift_TmVar_Tele (weakenCutoffTmVar c2 k2) (subst_TmVar_Tele (weakenTrace X0 k2) s2 d11)) =
          (subst_TmVar_Tele (weakenTrace X0 k2) (shift_TmVar_Tm c2 s2) (shift_TmVar_Tele (weakenCutoffTmVar (CS c2) k2) d11))))).
    Proof.
      pose proof (shift_TmVar__subst_TmVar_0_Tele_Tm_Terms_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__subst_TmVar_0_Tm_comm_ind  :
      (forall (s2 : Tm) ,
        (forall (k2 : Hvl) (c2 : (Cutoff TmVar)) (s3 : Tm) ,
          ((shift_TmVar_Tm (weakenCutoffTmVar c2 k2) (subst_TmVar_Tm (weakenTrace X0 k2) s3 s2)) =
          (subst_TmVar_Tm (weakenTrace X0 k2) (shift_TmVar_Tm c2 s3) (shift_TmVar_Tm (weakenCutoffTmVar (CS c2) k2) s2))))).
    Proof.
      pose proof (shift_TmVar__subst_TmVar_0_Tele_Tm_Terms_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__subst_TmVar_0_Terms_comm_ind  :
      (forall (ts5 : Terms) ,
        (forall (k2 : Hvl) (c2 : (Cutoff TmVar)) (s3 : Tm) ,
          ((shift_TmVar_Terms (weakenCutoffTmVar c2 k2) (subst_TmVar_Terms (weakenTrace X0 k2) s3 ts5)) =
          (subst_TmVar_Terms (weakenTrace X0 k2) (shift_TmVar_Tm c2 s3) (shift_TmVar_Terms (weakenCutoffTmVar (CS c2) k2) ts5))))).
    Proof.
      pose proof (shift_TmVar__subst_TmVar_0_Tele_Tm_Terms_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End ShiftSubstCommInd.
  Section ShiftSubstComm.
    Definition shift_TmVar_Tele_subst_TmVar_Tele0_comm (d11 : Tele) : (forall (c2 : (Cutoff TmVar)) (s2 : Tm) ,
      ((shift_TmVar_Tele c2 (subst_TmVar_Tele X0 s2 d11)) =
      (subst_TmVar_Tele X0 (shift_TmVar_Tm c2 s2) (shift_TmVar_Tele (CS c2) d11)))) := (shift_TmVar__subst_TmVar_0_Tele_comm_ind d11 H0).
    Definition shift_TmVar_Tm_subst_TmVar_Tm0_comm (s3 : Tm) : (forall (c2 : (Cutoff TmVar)) (s2 : Tm) ,
      ((shift_TmVar_Tm c2 (subst_TmVar_Tm X0 s2 s3)) =
      (subst_TmVar_Tm X0 (shift_TmVar_Tm c2 s2) (shift_TmVar_Tm (CS c2) s3)))) := (shift_TmVar__subst_TmVar_0_Tm_comm_ind s3 H0).
    Definition shift_TmVar_Terms_subst_TmVar_Terms0_comm (ts5 : Terms) : (forall (c2 : (Cutoff TmVar)) (s2 : Tm) ,
      ((shift_TmVar_Terms c2 (subst_TmVar_Terms X0 s2 ts5)) =
      (subst_TmVar_Terms X0 (shift_TmVar_Tm c2 s2) (shift_TmVar_Terms (CS c2) ts5)))) := (shift_TmVar__subst_TmVar_0_Terms_comm_ind ts5 H0).
  End ShiftSubstComm.
  Section SubstShiftIndexCommInd.
    Lemma subst_TmVar_Index_shift_TmVar_Tm0_comm_ind (d11 : (Trace TmVar)) (s2 : Tm) :
      (forall (k2 : Hvl) (x8 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace (XS TmVar d11) k2) s2 (shift_TmVar_Index (weakenCutoffTmVar C0 k2) x8)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k2) (subst_TmVar_Index (weakenTrace d11 k2) s2 x8)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End SubstShiftIndexCommInd.
  Section SubstShiftCommInd.
    Definition subst_TmVar__shift_TmVar_0_Tele_Tm_Terms_comm_ind := (ind_Tele_Tm_Terms (fun (d22 : Tele) =>
      (forall (k10 : Hvl) (d23 : (Trace TmVar)) (s10 : Tm) ,
        ((subst_TmVar_Tele (weakenTrace (XS TmVar d23) k10) s10 (shift_TmVar_Tele (weakenCutoffTmVar C0 k10) d22)) =
        (shift_TmVar_Tele (weakenCutoffTmVar C0 k10) (subst_TmVar_Tele (weakenTrace d23 k10) s10 d22))))) (fun (s10 : Tm) =>
      (forall (k10 : Hvl) (d22 : (Trace TmVar)) (s11 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace (XS TmVar d22) k10) s11 (shift_TmVar_Tm (weakenCutoffTmVar C0 k10) s10)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k10) (subst_TmVar_Tm (weakenTrace d22 k10) s11 s10))))) (fun (ts7 : Terms) =>
      (forall (k10 : Hvl) (d22 : (Trace TmVar)) (s10 : Tm) ,
        ((subst_TmVar_Terms (weakenTrace (XS TmVar d22) k10) s10 (shift_TmVar_Terms (weakenCutoffTmVar C0 k10) ts7)) =
        (shift_TmVar_Terms (weakenCutoffTmVar C0 k10) (subst_TmVar_Terms (weakenTrace d22 k10) s10 ts7))))) (fun (k10 : Hvl) (d22 : (Trace TmVar)) (s10 : Tm) =>
      eq_refl) (fun (T : Tm) IHT7 (d : Tele) IHd (k10 : Hvl) (d22 : (Trace TmVar)) (s10 : Tm) =>
      (f_equal2 TCons (IHT7 k10 d22 s10) (eq_trans (f_equal3 subst_TmVar_Tele (weakenTrace_append TmVar (XS TmVar d22) k10 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHd (HS TmVar k10) d22 s10) (f_equal2 shift_TmVar_Tele eq_refl (f_equal3 subst_TmVar_Tele (eq_sym (weakenTrace_append TmVar d22 k10 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (x11 : (Index TmVar)) =>
      (fun (k10 : Hvl) (d22 : (Trace TmVar)) (s10 : Tm) =>
        (subst_TmVar_Index_shift_TmVar_Tm0_comm_ind d22 s10 k10 x11))) (fun (d : Tele) IHd (t : Tm) IHt11 (k10 : Hvl) (d22 : (Trace TmVar)) (s10 : Tm) =>
      (f_equal2 Abs (IHd k10 d22 s10) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenTrace_append TmVar (XS TmVar d22) k10 (bind d))) eq_refl (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k10 (bind d)) eq_refl)) (eq_trans (IHt11 (appendHvl k10 (bind d)) d22 s10) (f_equal2 shift_TmVar_Tm (eq_trans (eq_sym (weakenCutoffTmVar_append C0 k10 (bind d))) (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0)))) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d22 k10 (bind d))) eq_refl eq_refl)))))) (fun (d : Tele) IHd (T : Tm) IHT7 (k10 : Hvl) (d22 : (Trace TmVar)) (s10 : Tm) =>
      (f_equal2 Pi (IHd k10 d22 s10) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenTrace_append TmVar (XS TmVar d22) k10 (bind d))) eq_refl (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k10 (bind d)) eq_refl)) (eq_trans (IHT7 (appendHvl k10 (bind d)) d22 s10) (f_equal2 shift_TmVar_Tm (eq_trans (eq_sym (weakenCutoffTmVar_append C0 k10 (bind d))) (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0)))) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d22 k10 (bind d))) eq_refl eq_refl)))))) (fun (t : Tm) IHt11 (ts : Terms) IHts7 (k10 : Hvl) (d22 : (Trace TmVar)) (s10 : Tm) =>
      (f_equal2 App (IHt11 k10 d22 s10) (IHts7 k10 d22 s10))) (fun (k10 : Hvl) (d22 : (Trace TmVar)) (s10 : Tm) =>
      eq_refl) (fun (t : Tm) IHt11 (ts : Terms) IHts7 (k10 : Hvl) (d22 : (Trace TmVar)) (s10 : Tm) =>
      (f_equal2 Cons (IHt11 k10 d22 s10) (IHts7 k10 d22 s10)))).
    Lemma subst_TmVar__shift_TmVar_0_Tele_comm_ind  :
      (forall (d22 : Tele) ,
        (forall (k10 : Hvl) (d23 : (Trace TmVar)) (s10 : Tm) ,
          ((subst_TmVar_Tele (weakenTrace (XS TmVar d23) k10) s10 (shift_TmVar_Tele (weakenCutoffTmVar C0 k10) d22)) =
          (shift_TmVar_Tele (weakenCutoffTmVar C0 k10) (subst_TmVar_Tele (weakenTrace d23 k10) s10 d22))))).
    Proof.
      pose proof (subst_TmVar__shift_TmVar_0_Tele_Tm_Terms_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__shift_TmVar_0_Tm_comm_ind  :
      (forall (s10 : Tm) ,
        (forall (k10 : Hvl) (d23 : (Trace TmVar)) (s11 : Tm) ,
          ((subst_TmVar_Tm (weakenTrace (XS TmVar d23) k10) s11 (shift_TmVar_Tm (weakenCutoffTmVar C0 k10) s10)) =
          (shift_TmVar_Tm (weakenCutoffTmVar C0 k10) (subst_TmVar_Tm (weakenTrace d23 k10) s11 s10))))).
    Proof.
      pose proof (subst_TmVar__shift_TmVar_0_Tele_Tm_Terms_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__shift_TmVar_0_Terms_comm_ind  :
      (forall (ts7 : Terms) ,
        (forall (k10 : Hvl) (d23 : (Trace TmVar)) (s11 : Tm) ,
          ((subst_TmVar_Terms (weakenTrace (XS TmVar d23) k10) s11 (shift_TmVar_Terms (weakenCutoffTmVar C0 k10) ts7)) =
          (shift_TmVar_Terms (weakenCutoffTmVar C0 k10) (subst_TmVar_Terms (weakenTrace d23 k10) s11 ts7))))).
    Proof.
      pose proof (subst_TmVar__shift_TmVar_0_Tele_Tm_Terms_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End SubstShiftCommInd.
  Section SubstShiftComm.
    Definition subst_TmVar_Tele_shift_TmVar_Tele0_comm (d23 : Tele) : (forall (d22 : (Trace TmVar)) (s10 : Tm) ,
      ((subst_TmVar_Tele (XS TmVar d22) s10 (shift_TmVar_Tele C0 d23)) =
      (shift_TmVar_Tele C0 (subst_TmVar_Tele d22 s10 d23)))) := (subst_TmVar__shift_TmVar_0_Tele_comm_ind d23 H0).
    Definition subst_TmVar_Tm_shift_TmVar_Tm0_comm (s11 : Tm) : (forall (d22 : (Trace TmVar)) (s10 : Tm) ,
      ((subst_TmVar_Tm (XS TmVar d22) s10 (shift_TmVar_Tm C0 s11)) =
      (shift_TmVar_Tm C0 (subst_TmVar_Tm d22 s10 s11)))) := (subst_TmVar__shift_TmVar_0_Tm_comm_ind s11 H0).
    Definition subst_TmVar_Terms_shift_TmVar_Terms0_comm (ts7 : Terms) : (forall (d22 : (Trace TmVar)) (s10 : Tm) ,
      ((subst_TmVar_Terms (XS TmVar d22) s10 (shift_TmVar_Terms C0 ts7)) =
      (shift_TmVar_Terms C0 (subst_TmVar_Terms d22 s10 ts7)))) := (subst_TmVar__shift_TmVar_0_Terms_comm_ind ts7 H0).
  End SubstShiftComm.
  Section SubstSubordInd.
    
  End SubstSubordInd.
  Section SubstSubord.
    
  End SubstSubord.
End ShiftSubstInteraction.
 Hint Rewrite subst_TmVar_Tele0_shift_TmVar_Tele0_reflection subst_TmVar_Terms0_shift_TmVar_Terms0_reflection subst_TmVar_Tm0_shift_TmVar_Tm0_reflection : interaction.
 Hint Rewrite subst_TmVar_Tele0_shift_TmVar_Tele0_reflection subst_TmVar_Terms0_shift_TmVar_Terms0_reflection subst_TmVar_Tm0_shift_TmVar_Tm0_reflection : reflection.
 Hint Rewrite subst_TmVar_Tele_shift_TmVar_Tele0_comm subst_TmVar_Terms_shift_TmVar_Terms0_comm subst_TmVar_Tm_shift_TmVar_Tm0_comm : interaction.
 Hint Rewrite subst_TmVar_Tele_shift_TmVar_Tele0_comm subst_TmVar_Terms_shift_TmVar_Terms0_comm subst_TmVar_Tm_shift_TmVar_Tm0_comm : subst_shift0.
 Hint Rewrite shift_TmVar_Tele_subst_TmVar_Tele0_comm shift_TmVar_Terms_subst_TmVar_Terms0_comm shift_TmVar_Tm_subst_TmVar_Tm0_comm : interaction.
 Hint Rewrite shift_TmVar_Tele_subst_TmVar_Tele0_comm shift_TmVar_Terms_subst_TmVar_Terms0_comm shift_TmVar_Tm_subst_TmVar_Tm0_comm : shift_subst0.
Section SubstSubstInteraction.
  Section SubstSubstIndexCommInd.
    Lemma subst_TmVar_Tm_subst_TmVar_Index0_commright_ind (d22 : (Trace TmVar)) (s10 : Tm) (s11 : Tm) :
      (forall (k10 : Hvl) (x11 : (Index TmVar)) ,
        ((subst_TmVar_Tm (weakenTrace d22 k10) s10 (subst_TmVar_Index (weakenTrace X0 k10) s11 x11)) =
        (subst_TmVar_Tm (weakenTrace X0 k10) (subst_TmVar_Tm d22 s10 s11) (subst_TmVar_Index (weakenTrace (XS TmVar d22) k10) s10 x11)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
  End SubstSubstIndexCommInd.
  Section SubstSubstCommInd.
    Definition subst_TmVar__subst_TmVar_0_Tele_Tm_Terms_comm_ind := (ind_Tele_Tm_Terms (fun (d26 : Tele) =>
      (forall (k11 : Hvl) (d27 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) ,
        ((subst_TmVar_Tele (weakenTrace d27 k11) s12 (subst_TmVar_Tele (weakenTrace X0 k11) s13 d26)) =
        (subst_TmVar_Tele (weakenTrace X0 k11) (subst_TmVar_Tm d27 s12 s13) (subst_TmVar_Tele (weakenTrace (XS TmVar d27) k11) s12 d26))))) (fun (s12 : Tm) =>
      (forall (k11 : Hvl) (d26 : (Trace TmVar)) (s13 : Tm) (s14 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace d26 k11) s13 (subst_TmVar_Tm (weakenTrace X0 k11) s14 s12)) =
        (subst_TmVar_Tm (weakenTrace X0 k11) (subst_TmVar_Tm d26 s13 s14) (subst_TmVar_Tm (weakenTrace (XS TmVar d26) k11) s13 s12))))) (fun (ts9 : Terms) =>
      (forall (k11 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) ,
        ((subst_TmVar_Terms (weakenTrace d26 k11) s12 (subst_TmVar_Terms (weakenTrace X0 k11) s13 ts9)) =
        (subst_TmVar_Terms (weakenTrace X0 k11) (subst_TmVar_Tm d26 s12 s13) (subst_TmVar_Terms (weakenTrace (XS TmVar d26) k11) s12 ts9))))) (fun (k11 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) =>
      eq_refl) (fun (T : Tm) IHT9 (d : Tele) IHd (k11 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) =>
      (f_equal2 TCons (IHT9 k11 d26 s12 s13) (eq_trans (f_equal3 subst_TmVar_Tele (weakenTrace_append TmVar d26 k11 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Tele (weakenTrace_append TmVar X0 k11 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k11) d26 s12 s13) (f_equal3 subst_TmVar_Tele (eq_sym (weakenTrace_append TmVar X0 k11 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Tele (eq_sym (weakenTrace_append TmVar (XS TmVar d26) k11 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (x14 : (Index TmVar)) =>
      (fun (k11 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) =>
        (subst_TmVar_Tm_subst_TmVar_Index0_commright_ind d26 s12 s13 k11 x14))) (fun (d : Tele) IHd (t : Tm) IHt14 (k11 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) =>
      (f_equal2 Abs (IHd k11 d26 s12 s13) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_subst_TmVar__bind _ _ _) (eq_refl H0))) (weakenTrace_append TmVar d26 k11 (bind d))) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k11 (bind d)) eq_refl eq_refl)) (eq_trans (IHt14 (appendHvl k11 (bind d)) d26 s12 s13) (f_equal3 subst_TmVar_Tm (eq_trans (eq_sym (weakenTrace_append TmVar X0 k11 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0)))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TmVar d26) k11 (bind d))) eq_refl eq_refl)))))) (fun (d : Tele) IHd (T : Tm) IHT9 (k11 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) =>
      (f_equal2 Pi (IHd k11 d26 s12 s13) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_subst_TmVar__bind _ _ _) (eq_refl H0))) (weakenTrace_append TmVar d26 k11 (bind d))) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k11 (bind d)) eq_refl eq_refl)) (eq_trans (IHT9 (appendHvl k11 (bind d)) d26 s12 s13) (f_equal3 subst_TmVar_Tm (eq_trans (eq_sym (weakenTrace_append TmVar X0 k11 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0)))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TmVar d26) k11 (bind d))) eq_refl eq_refl)))))) (fun (t : Tm) IHt14 (ts : Terms) IHts9 (k11 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) =>
      (f_equal2 App (IHt14 k11 d26 s12 s13) (IHts9 k11 d26 s12 s13))) (fun (k11 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) =>
      eq_refl) (fun (t : Tm) IHt14 (ts : Terms) IHts9 (k11 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) =>
      (f_equal2 Cons (IHt14 k11 d26 s12 s13) (IHts9 k11 d26 s12 s13)))).
    Lemma subst_TmVar__subst_TmVar_0_Tele_comm_ind  :
      (forall (d26 : Tele) ,
        (forall (k11 : Hvl) (d27 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) ,
          ((subst_TmVar_Tele (weakenTrace d27 k11) s12 (subst_TmVar_Tele (weakenTrace X0 k11) s13 d26)) =
          (subst_TmVar_Tele (weakenTrace X0 k11) (subst_TmVar_Tm d27 s12 s13) (subst_TmVar_Tele (weakenTrace (XS TmVar d27) k11) s12 d26))))).
    Proof.
      pose proof (subst_TmVar__subst_TmVar_0_Tele_Tm_Terms_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__subst_TmVar_0_Tm_comm_ind  :
      (forall (s12 : Tm) ,
        (forall (k11 : Hvl) (d27 : (Trace TmVar)) (s13 : Tm) (s14 : Tm) ,
          ((subst_TmVar_Tm (weakenTrace d27 k11) s13 (subst_TmVar_Tm (weakenTrace X0 k11) s14 s12)) =
          (subst_TmVar_Tm (weakenTrace X0 k11) (subst_TmVar_Tm d27 s13 s14) (subst_TmVar_Tm (weakenTrace (XS TmVar d27) k11) s13 s12))))).
    Proof.
      pose proof (subst_TmVar__subst_TmVar_0_Tele_Tm_Terms_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__subst_TmVar_0_Terms_comm_ind  :
      (forall (ts9 : Terms) ,
        (forall (k11 : Hvl) (d27 : (Trace TmVar)) (s13 : Tm) (s14 : Tm) ,
          ((subst_TmVar_Terms (weakenTrace d27 k11) s13 (subst_TmVar_Terms (weakenTrace X0 k11) s14 ts9)) =
          (subst_TmVar_Terms (weakenTrace X0 k11) (subst_TmVar_Tm d27 s13 s14) (subst_TmVar_Terms (weakenTrace (XS TmVar d27) k11) s13 ts9))))).
    Proof.
      pose proof (subst_TmVar__subst_TmVar_0_Tele_Tm_Terms_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End SubstSubstCommInd.
  Section SubstSubstComm.
    Definition subst_TmVar_Tele_subst_TmVar_Tele0_comm (d27 : Tele) : (forall (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) ,
      ((subst_TmVar_Tele d26 s12 (subst_TmVar_Tele X0 s13 d27)) =
      (subst_TmVar_Tele X0 (subst_TmVar_Tm d26 s12 s13) (subst_TmVar_Tele (XS TmVar d26) s12 d27)))) := (subst_TmVar__subst_TmVar_0_Tele_comm_ind d27 H0).
    Definition subst_TmVar_Tm_subst_TmVar_Tm0_comm (s14 : Tm) : (forall (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) ,
      ((subst_TmVar_Tm d26 s12 (subst_TmVar_Tm X0 s13 s14)) =
      (subst_TmVar_Tm X0 (subst_TmVar_Tm d26 s12 s13) (subst_TmVar_Tm (XS TmVar d26) s12 s14)))) := (subst_TmVar__subst_TmVar_0_Tm_comm_ind s14 H0).
    Definition subst_TmVar_Terms_subst_TmVar_Terms0_comm (ts9 : Terms) : (forall (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) ,
      ((subst_TmVar_Terms d26 s12 (subst_TmVar_Terms X0 s13 ts9)) =
      (subst_TmVar_Terms X0 (subst_TmVar_Tm d26 s12 s13) (subst_TmVar_Terms (XS TmVar d26) s12 ts9)))) := (subst_TmVar__subst_TmVar_0_Terms_comm_ind ts9 H0).
  End SubstSubstComm.
  Section WeakenSubst.
    Lemma weakenTele_subst_TmVar_Tele  :
      (forall (k11 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (d27 : Tele) ,
        ((weakenTele (subst_TmVar_Tele d26 s12 d27) k11) =
        (subst_TmVar_Tele (weakenTrace d26 k11) s12 (weakenTele d27 k11)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_subst_TmVar_Tm  :
      (forall (k11 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (s13 : Tm) ,
        ((weakenTm (subst_TmVar_Tm d26 s12 s13) k11) =
        (subst_TmVar_Tm (weakenTrace d26 k11) s12 (weakenTm s13 k11)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTerms_subst_TmVar_Terms  :
      (forall (k11 : Hvl) (d26 : (Trace TmVar)) (s12 : Tm) (ts9 : Terms) ,
        ((weakenTerms (subst_TmVar_Terms d26 s12 ts9) k11) =
        (subst_TmVar_Terms (weakenTrace d26 k11) s12 (weakenTerms ts9 k11)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
  End WeakenSubst.
  Section WeakenAppend.
    Lemma weakenTele_append  :
      (forall (d26 : Tele) (k11 : Hvl) (k12 : Hvl) ,
        ((weakenTele (weakenTele d26 k11) k12) =
        (weakenTele d26 (appendHvl k11 k12)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenTm_append  :
      (forall (s12 : Tm) (k11 : Hvl) (k12 : Hvl) ,
        ((weakenTm (weakenTm s12 k11) k12) =
        (weakenTm s12 (appendHvl k11 k12)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenTerms_append  :
      (forall (ts9 : Terms) (k11 : Hvl) (k12 : Hvl) ,
        ((weakenTerms (weakenTerms ts9 k11) k12) =
        (weakenTerms ts9 (appendHvl k11 k12)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
  End WeakenAppend.
End SubstSubstInteraction.
 Hint Rewrite subst_TmVar_Tele_subst_TmVar_Tele0_comm subst_TmVar_Terms_subst_TmVar_Terms0_comm subst_TmVar_Tm_subst_TmVar_Tm0_comm : interaction.
 Hint Rewrite subst_TmVar_Tele_subst_TmVar_Tele0_comm subst_TmVar_Terms_subst_TmVar_Terms0_comm subst_TmVar_Tm_subst_TmVar_Tm0_comm : subst_subst0.
 Hint Rewrite weakenTele_shift_TmVar_Tele weakenTerms_shift_TmVar_Terms weakenTm_shift_TmVar_Tm : weaken_shift.
 Hint Rewrite weakenTele_subst_TmVar_Tele weakenTerms_subst_TmVar_Terms weakenTm_subst_TmVar_Tm : weaken_subst.
Section WellFormed.
  Fixpoint wfindex {a : Namespace} (k11 : Hvl) (i : (Index a)) {struct k11} : Prop :=
    match k11 with
      | (H0) => False
      | (HS b k11) => match (eq_namespace_dec a b) with
        | (left e) => match i with
          | (I0) => True
          | (IS i) => (wfindex k11 i)
        end
        | (right n) => (wfindex k11 i)
      end
    end.
  Inductive wfTele (k11 : Hvl) : Tele -> Prop :=
    | wfTele_TNil :
        (wfTele k11 (TNil))
    | wfTele_TCons {T11 : Tm}
        (wf : (wfTm k11 T11))
        {d29 : Tele}
        (wf0 : (wfTele (HS TmVar k11) d29))
        : (wfTele k11 (TCons T11 d29))
  with wfTm (k11 : Hvl) : Tm -> Prop :=
    | wfTm_Var
        {x17 : (Index TmVar)}
        (wfi : (wfindex k11 x17)) :
        (wfTm k11 (Var x17))
    | wfTm_Abs {d29 : Tele}
        (wf : (wfTele k11 d29))
        {t17 : Tm}
        (wf0 : (wfTm (appendHvl k11 (bind d29)) t17))
        : (wfTm k11 (Abs d29 t17))
    | wfTm_Pi {d30 : Tele}
        (wf : (wfTele k11 d30))
        {T11 : Tm}
        (wf0 : (wfTm (appendHvl k11 (bind d30)) T11))
        : (wfTm k11 (Pi d30 T11))
    | wfTm_App {t18 : Tm}
        (wf : (wfTm k11 t18))
        {ts11 : Terms}
        (wf0 : (wfTerms k11 ts11)) :
        (wfTm k11 (App t18 ts11))
  with wfTerms (k11 : Hvl) : Terms -> Prop :=
    | wfTerms_Nil :
        (wfTerms k11 (Nil))
    | wfTerms_Cons {t17 : Tm}
        (wf : (wfTm k11 t17))
        {ts11 : Terms}
        (wf0 : (wfTerms k11 ts11)) :
        (wfTerms k11 (Cons t17 ts11)).
  Scheme ind_wfTele := Induction for wfTele Sort Prop
  with ind_wfTm := Induction for wfTm Sort Prop
  with ind_wfTerms := Induction for wfTerms Sort Prop.
  Combined Scheme ind_wfTele_Tm_Terms from ind_wfTele, ind_wfTm, ind_wfTerms.
End WellFormed.
Section ShiftWellFormed.
  Inductive shifthvl_TmVar : (forall (c3 : (Cutoff TmVar)) (k11 : Hvl) (k12 : Hvl) ,
    Prop) :=
    | shifthvl_TmVar_here
        (k11 : Hvl) :
        (shifthvl_TmVar C0 k11 (HS TmVar k11))
    | shifthvl_TmVar_there_TmVar
        (c3 : (Cutoff TmVar))
        (k11 : Hvl) (k12 : Hvl) :
        (shifthvl_TmVar c3 k11 k12) -> (shifthvl_TmVar (CS c3) (HS TmVar k11) (HS TmVar k12)).
  Lemma weaken_shifthvl_TmVar  :
    (forall (k11 : Hvl) {c3 : (Cutoff TmVar)} {k12 : Hvl} {k13 : Hvl} ,
      (shifthvl_TmVar c3 k12 k13) -> (shifthvl_TmVar (weakenCutoffTmVar c3 k11) (appendHvl k12 k11) (appendHvl k13 k11))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma shift_TmVar__wfindex_TmVar  :
    (forall (c3 : (Cutoff TmVar)) (k11 : Hvl) (k12 : Hvl) (ins : (shifthvl_TmVar c3 k11 k12)) (x17 : (Index TmVar)) ,
      (wfindex k11 x17) -> (wfindex k12 (shift_TmVar_Index c3 x17))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Definition shift_TmVar__wfTele_Tm_Terms : (forall (k11 : Hvl) ,
    (forall (d29 : Tele) (wf : (wfTele k11 d29)) ,
      (forall (c3 : (Cutoff TmVar)) (k12 : Hvl) ,
        (shifthvl_TmVar c3 k11 k12) -> (wfTele k12 (shift_TmVar_Tele c3 d29)))) /\
    (forall (s12 : Tm) (wf : (wfTm k11 s12)) ,
      (forall (c3 : (Cutoff TmVar)) (k12 : Hvl) ,
        (shifthvl_TmVar c3 k11 k12) -> (wfTm k12 (shift_TmVar_Tm c3 s12)))) /\
    (forall (ts11 : Terms) (wf : (wfTerms k11 ts11)) ,
      (forall (c3 : (Cutoff TmVar)) (k12 : Hvl) ,
        (shifthvl_TmVar c3 k11 k12) -> (wfTerms k12 (shift_TmVar_Terms c3 ts11))))) := (ind_wfTele_Tm_Terms (fun (k11 : Hvl) (d29 : Tele) (wf : (wfTele k11 d29)) =>
    (forall (c3 : (Cutoff TmVar)) (k12 : Hvl) ,
      (shifthvl_TmVar c3 k11 k12) -> (wfTele k12 (shift_TmVar_Tele c3 d29)))) (fun (k11 : Hvl) (s12 : Tm) (wf : (wfTm k11 s12)) =>
    (forall (c3 : (Cutoff TmVar)) (k12 : Hvl) ,
      (shifthvl_TmVar c3 k11 k12) -> (wfTm k12 (shift_TmVar_Tm c3 s12)))) (fun (k11 : Hvl) (ts11 : Terms) (wf : (wfTerms k11 ts11)) =>
    (forall (c3 : (Cutoff TmVar)) (k12 : Hvl) ,
      (shifthvl_TmVar c3 k11 k12) -> (wfTerms k12 (shift_TmVar_Terms c3 ts11)))) (fun (k11 : Hvl) (c3 : (Cutoff TmVar)) (k12 : Hvl) (ins : (shifthvl_TmVar c3 k11 k12)) =>
    (wfTele_TNil k12)) (fun (k11 : Hvl) (T : Tm) (wf : (wfTm k11 T)) IHT11 (d : Tele) (wf0 : (wfTele (HS TmVar k11) d)) IHd (c3 : (Cutoff TmVar)) (k12 : Hvl) (ins : (shifthvl_TmVar c3 k11 k12)) =>
    (wfTele_TCons k12 (IHT11 c3 k12 (weaken_shifthvl_TmVar H0 ins)) (IHd (CS c3) (HS TmVar k12) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k11 : Hvl) (x17 : (Index TmVar)) (wfi : (wfindex k11 x17)) (c3 : (Cutoff TmVar)) (k12 : Hvl) (ins : (shifthvl_TmVar c3 k11 k12)) =>
    (wfTm_Var k12 (shift_TmVar__wfindex_TmVar c3 k11 k12 ins x17 wfi))) (fun (k11 : Hvl) (d : Tele) (wf : (wfTele k11 d)) IHd (t : Tm) (wf0 : (wfTm (appendHvl k11 (bind d)) t)) IHt17 (c3 : (Cutoff TmVar)) (k12 : Hvl) (ins : (shifthvl_TmVar c3 k11 k12)) =>
    (wfTm_Abs k12 (IHd c3 k12 (weaken_shifthvl_TmVar H0 ins)) (eq_ind2 wfTm (f_equal2 appendHvl (eq_refl k12) (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0))) eq_refl (IHt17 (weakenCutoffTmVar c3 (bind d)) (appendHvl k12 (bind d)) (weaken_shifthvl_TmVar (bind d) ins))))) (fun (k11 : Hvl) (d : Tele) (wf : (wfTele k11 d)) IHd (T : Tm) (wf0 : (wfTm (appendHvl k11 (bind d)) T)) IHT11 (c3 : (Cutoff TmVar)) (k12 : Hvl) (ins : (shifthvl_TmVar c3 k11 k12)) =>
    (wfTm_Pi k12 (IHd c3 k12 (weaken_shifthvl_TmVar H0 ins)) (eq_ind2 wfTm (f_equal2 appendHvl (eq_refl k12) (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0))) eq_refl (IHT11 (weakenCutoffTmVar c3 (bind d)) (appendHvl k12 (bind d)) (weaken_shifthvl_TmVar (bind d) ins))))) (fun (k11 : Hvl) (t : Tm) (wf : (wfTm k11 t)) IHt17 (ts : Terms) (wf0 : (wfTerms k11 ts)) IHts11 (c3 : (Cutoff TmVar)) (k12 : Hvl) (ins : (shifthvl_TmVar c3 k11 k12)) =>
    (wfTm_App k12 (IHt17 c3 k12 (weaken_shifthvl_TmVar H0 ins)) (IHts11 c3 k12 (weaken_shifthvl_TmVar H0 ins)))) (fun (k11 : Hvl) (c3 : (Cutoff TmVar)) (k12 : Hvl) (ins : (shifthvl_TmVar c3 k11 k12)) =>
    (wfTerms_Nil k12)) (fun (k11 : Hvl) (t : Tm) (wf : (wfTm k11 t)) IHt17 (ts : Terms) (wf0 : (wfTerms k11 ts)) IHts11 (c3 : (Cutoff TmVar)) (k12 : Hvl) (ins : (shifthvl_TmVar c3 k11 k12)) =>
    (wfTerms_Cons k12 (IHt17 c3 k12 (weaken_shifthvl_TmVar H0 ins)) (IHts11 c3 k12 (weaken_shifthvl_TmVar H0 ins))))).
  Lemma shift_TmVar__wfTele (k11 : Hvl) :
    (forall (d29 : Tele) (wf : (wfTele k11 d29)) ,
      (forall (c3 : (Cutoff TmVar)) (k12 : Hvl) ,
        (shifthvl_TmVar c3 k11 k12) -> (wfTele k12 (shift_TmVar_Tele c3 d29)))).
  Proof.
    pose proof ((shift_TmVar__wfTele_Tm_Terms k11)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_TmVar__wfTm (k11 : Hvl) :
    (forall (s12 : Tm) (wf0 : (wfTm k11 s12)) ,
      (forall (c4 : (Cutoff TmVar)) (k13 : Hvl) ,
        (shifthvl_TmVar c4 k11 k13) -> (wfTm k13 (shift_TmVar_Tm c4 s12)))).
  Proof.
    pose proof ((shift_TmVar__wfTele_Tm_Terms k11)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_TmVar__wfTerms (k11 : Hvl) :
    (forall (ts11 : Terms) (wf1 : (wfTerms k11 ts11)) ,
      (forall (c5 : (Cutoff TmVar)) (k14 : Hvl) ,
        (shifthvl_TmVar c5 k11 k14) -> (wfTerms k14 (shift_TmVar_Terms c5 ts11)))).
  Proof.
    pose proof ((shift_TmVar__wfTele_Tm_Terms k11)).
    destruct_conjs .
    easy .
  Qed.
End ShiftWellFormed.
 Hint Resolve shift_TmVar__wfindex_TmVar : infra.
 Hint Resolve shift_TmVar__wfindex_TmVar : shift.
 Hint Resolve shift_TmVar__wfindex_TmVar : shift_wf.
 Hint Resolve shift_TmVar__wfindex_TmVar : wf.
 Hint Resolve shift_TmVar__wfTele shift_TmVar__wfTerms shift_TmVar__wfTm : infra.
 Hint Resolve shift_TmVar__wfTele shift_TmVar__wfTerms shift_TmVar__wfTm : shift.
 Hint Resolve shift_TmVar__wfTele shift_TmVar__wfTerms shift_TmVar__wfTm : shift_wf.
 Hint Resolve shift_TmVar__wfTele shift_TmVar__wfTerms shift_TmVar__wfTm : wf.
 Hint Constructors shifthvl_TmVar : infra.
 Hint Constructors shifthvl_TmVar : shift.
 Hint Constructors shifthvl_TmVar : shift_wf.
 Hint Constructors shifthvl_TmVar : wf.
 Hint Resolve weaken_shifthvl_TmVar : infra.
 Hint Resolve weaken_shifthvl_TmVar : shift.
 Hint Resolve weaken_shifthvl_TmVar : shift_wf.
 Hint Resolve weaken_shifthvl_TmVar : weaken.
 Hint Resolve weaken_shifthvl_TmVar : wf.
Section SubstWellFormed.
  Inductive substhvl_TmVar (k11 : Hvl) : (forall (d29 : (Trace TmVar)) (k12 : Hvl) (k13 : Hvl) ,
    Prop) :=
    | substhvl_TmVar_here :
        (substhvl_TmVar k11 X0 (HS TmVar k11) k11)
    | substhvl_TmVar_there_TmVar
        {d29 : (Trace TmVar)}
        {k12 : Hvl} {k13 : Hvl} :
        (substhvl_TmVar k11 d29 k12 k13) -> (substhvl_TmVar k11 (XS TmVar d29) (HS TmVar k12) (HS TmVar k13)).
  Lemma weaken_substhvl_TmVar  :
    (forall {k12 : Hvl} (k11 : Hvl) {d29 : (Trace TmVar)} {k13 : Hvl} {k14 : Hvl} ,
      (substhvl_TmVar k12 d29 k13 k14) -> (substhvl_TmVar k12 (weakenTrace d29 k11) (appendHvl k13 k11) (appendHvl k14 k11))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma substhvl_TmVar_wfindex_TmVar {k11 : Hvl} {s12 : Tm} (wft : (wfTm k11 s12)) :
    (forall {d29 : (Trace TmVar)} {k12 : Hvl} {k13 : Hvl} ,
      (substhvl_TmVar k11 d29 k12 k13) -> (forall {x17 : (Index TmVar)} ,
        (wfindex k12 x17) -> (wfTm k13 (subst_TmVar_Index d29 s12 x17)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Definition substhvl_TmVar_wfTele_Tm_Terms {k11 : Hvl} {s12 : Tm} (wf : (wfTm k11 s12)) : (forall (k12 : Hvl) ,
    (forall (d29 : Tele) (wf0 : (wfTele k12 d29)) ,
      (forall {d30 : (Trace TmVar)} {k13 : Hvl} ,
        (substhvl_TmVar k11 d30 k12 k13) -> (wfTele k13 (subst_TmVar_Tele d30 s12 d29)))) /\
    (forall (s13 : Tm) (wf0 : (wfTm k12 s13)) ,
      (forall {d29 : (Trace TmVar)} {k13 : Hvl} ,
        (substhvl_TmVar k11 d29 k12 k13) -> (wfTm k13 (subst_TmVar_Tm d29 s12 s13)))) /\
    (forall (ts11 : Terms) (wf0 : (wfTerms k12 ts11)) ,
      (forall {d29 : (Trace TmVar)} {k13 : Hvl} ,
        (substhvl_TmVar k11 d29 k12 k13) -> (wfTerms k13 (subst_TmVar_Terms d29 s12 ts11))))) := (ind_wfTele_Tm_Terms (fun (k12 : Hvl) (d29 : Tele) (wf0 : (wfTele k12 d29)) =>
    (forall {d30 : (Trace TmVar)} {k13 : Hvl} ,
      (substhvl_TmVar k11 d30 k12 k13) -> (wfTele k13 (subst_TmVar_Tele d30 s12 d29)))) (fun (k12 : Hvl) (s13 : Tm) (wf0 : (wfTm k12 s13)) =>
    (forall {d29 : (Trace TmVar)} {k13 : Hvl} ,
      (substhvl_TmVar k11 d29 k12 k13) -> (wfTm k13 (subst_TmVar_Tm d29 s12 s13)))) (fun (k12 : Hvl) (ts11 : Terms) (wf0 : (wfTerms k12 ts11)) =>
    (forall {d29 : (Trace TmVar)} {k13 : Hvl} ,
      (substhvl_TmVar k11 d29 k12 k13) -> (wfTerms k13 (subst_TmVar_Terms d29 s12 ts11)))) (fun (k12 : Hvl) {d29 : (Trace TmVar)} {k13 : Hvl} (del : (substhvl_TmVar k11 d29 k12 k13)) =>
    (wfTele_TNil k13)) (fun (k12 : Hvl) (T : Tm) (wf0 : (wfTm k12 T)) IHT11 (d : Tele) (wf1 : (wfTele (HS TmVar k12) d)) IHd {d29 : (Trace TmVar)} {k13 : Hvl} (del : (substhvl_TmVar k11 d29 k12 k13)) =>
    (wfTele_TCons k13 (IHT11 (weakenTrace d29 H0) k13 (weaken_substhvl_TmVar H0 del)) (IHd (weakenTrace d29 (HS TmVar H0)) (HS TmVar k13) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k12 : Hvl) {x17 : (Index TmVar)} (wfi : (wfindex k12 x17)) {d29 : (Trace TmVar)} {k13 : Hvl} (del : (substhvl_TmVar k11 d29 k12 k13)) =>
    (substhvl_TmVar_wfindex_TmVar wf del wfi)) (fun (k12 : Hvl) (d : Tele) (wf0 : (wfTele k12 d)) IHd (t : Tm) (wf1 : (wfTm (appendHvl k12 (bind d)) t)) IHt17 {d29 : (Trace TmVar)} {k13 : Hvl} (del : (substhvl_TmVar k11 d29 k12 k13)) =>
    (wfTm_Abs k13 (IHd (weakenTrace d29 H0) k13 (weaken_substhvl_TmVar H0 del)) (eq_ind2 wfTm (f_equal2 appendHvl (eq_refl k13) (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0))) eq_refl (IHt17 (weakenTrace d29 (bind d)) (appendHvl k13 (bind d)) (weaken_substhvl_TmVar (bind d) del))))) (fun (k12 : Hvl) (d : Tele) (wf0 : (wfTele k12 d)) IHd (T : Tm) (wf1 : (wfTm (appendHvl k12 (bind d)) T)) IHT11 {d29 : (Trace TmVar)} {k13 : Hvl} (del : (substhvl_TmVar k11 d29 k12 k13)) =>
    (wfTm_Pi k13 (IHd (weakenTrace d29 H0) k13 (weaken_substhvl_TmVar H0 del)) (eq_ind2 wfTm (f_equal2 appendHvl (eq_refl k13) (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0))) eq_refl (IHT11 (weakenTrace d29 (bind d)) (appendHvl k13 (bind d)) (weaken_substhvl_TmVar (bind d) del))))) (fun (k12 : Hvl) (t : Tm) (wf0 : (wfTm k12 t)) IHt17 (ts : Terms) (wf1 : (wfTerms k12 ts)) IHts11 {d29 : (Trace TmVar)} {k13 : Hvl} (del : (substhvl_TmVar k11 d29 k12 k13)) =>
    (wfTm_App k13 (IHt17 (weakenTrace d29 H0) k13 (weaken_substhvl_TmVar H0 del)) (IHts11 (weakenTrace d29 H0) k13 (weaken_substhvl_TmVar H0 del)))) (fun (k12 : Hvl) {d29 : (Trace TmVar)} {k13 : Hvl} (del : (substhvl_TmVar k11 d29 k12 k13)) =>
    (wfTerms_Nil k13)) (fun (k12 : Hvl) (t : Tm) (wf0 : (wfTm k12 t)) IHt17 (ts : Terms) (wf1 : (wfTerms k12 ts)) IHts11 {d29 : (Trace TmVar)} {k13 : Hvl} (del : (substhvl_TmVar k11 d29 k12 k13)) =>
    (wfTerms_Cons k13 (IHt17 (weakenTrace d29 H0) k13 (weaken_substhvl_TmVar H0 del)) (IHts11 (weakenTrace d29 H0) k13 (weaken_substhvl_TmVar H0 del))))).
  Lemma substhvl_TmVar_wfTele {k11 : Hvl} {s12 : Tm} (wf : (wfTm k11 s12)) (k12 : Hvl) (d29 : Tele) (wf0 : (wfTele k12 d29)) :
    (forall {d30 : (Trace TmVar)} {k13 : Hvl} ,
      (substhvl_TmVar k11 d30 k12 k13) -> (wfTele k13 (subst_TmVar_Tele d30 s12 d29))).
  Proof.
    apply ((substhvl_TmVar_wfTele_Tm_Terms wf k12)).
    auto .
  Qed.
  Lemma substhvl_TmVar_wfTm {k11 : Hvl} {s12 : Tm} (wf : (wfTm k11 s12)) (k12 : Hvl) (s13 : Tm) (wf1 : (wfTm k12 s13)) :
    (forall {d31 : (Trace TmVar)} {k14 : Hvl} ,
      (substhvl_TmVar k11 d31 k12 k14) -> (wfTm k14 (subst_TmVar_Tm d31 s12 s13))).
  Proof.
    apply ((substhvl_TmVar_wfTele_Tm_Terms wf k12)).
    auto .
  Qed.
  Lemma substhvl_TmVar_wfTerms {k11 : Hvl} {s12 : Tm} (wf : (wfTm k11 s12)) (k12 : Hvl) (ts11 : Terms) (wf2 : (wfTerms k12 ts11)) :
    (forall {d32 : (Trace TmVar)} {k15 : Hvl} ,
      (substhvl_TmVar k11 d32 k12 k15) -> (wfTerms k15 (subst_TmVar_Terms d32 s12 ts11))).
  Proof.
    apply ((substhvl_TmVar_wfTele_Tm_Terms wf k12)).
    auto .
  Qed.
End SubstWellFormed.
 Hint Resolve substhvl_TmVar_wfindex_TmVar : infra.
 Hint Resolve substhvl_TmVar_wfindex_TmVar : subst.
 Hint Resolve substhvl_TmVar_wfindex_TmVar : subst_wf.
 Hint Resolve substhvl_TmVar_wfindex_TmVar : wf.
 Hint Resolve substhvl_TmVar_wfTele substhvl_TmVar_wfTerms substhvl_TmVar_wfTm : infra.
 Hint Resolve substhvl_TmVar_wfTele substhvl_TmVar_wfTerms substhvl_TmVar_wfTm : subst.
 Hint Resolve substhvl_TmVar_wfTele substhvl_TmVar_wfTerms substhvl_TmVar_wfTm : subst_wf.
 Hint Resolve substhvl_TmVar_wfTele substhvl_TmVar_wfTerms substhvl_TmVar_wfTm : wf.
 Hint Constructors substhvl_TmVar : infra.
 Hint Constructors substhvl_TmVar : subst.
 Hint Constructors substhvl_TmVar : subst_wf.
 Hint Constructors substhvl_TmVar : wf.
Fixpoint subhvl_TmVar (k11 : Hvl) {struct k11} : Prop :=
  match k11 with
    | (H0) => True
    | (HS a k11) => match a with
      | (TmVar) => (subhvl_TmVar k11)
    end
  end.
Lemma subhvl_TmVar_append  :
  (forall (k11 : Hvl) (k12 : Hvl) ,
    (subhvl_TmVar k11) -> (subhvl_TmVar k12) -> (subhvl_TmVar (appendHvl k11 k12))).
Proof.
  needleGenericSubHvlAppend .
Qed.
 Hint Resolve subhvl_TmVar_append : infra.
 Hint Resolve subhvl_TmVar_append : wf.
Section Context.
  
End Context.
Section ContextStuff.
  Section ShiftEnvAppendEnv.
    
  End ShiftEnvAppendEnv.
  Section SubstEnvAppendEnv.
    
  End SubstEnvAppendEnv.
  Section Lookups.
    
  End Lookups.
End ContextStuff.
 Hint Constructors wfTele wfTerms wfTm : infra.
 Hint Constructors wfTele wfTerms wfTm : wf.
 Hint Extern 10 ((wfTele _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTerms _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTm _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 2 ((wfTele _ _)) => match goal with
  | H : (wfTele _ (TNil)) |- _ => inversion H; subst; clear H
  | H : (wfTele _ (TCons _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfTm _ _)) => match goal with
  | H : (wfTm _ (Var _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Abs _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Pi _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (App _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfTerms _ _)) => match goal with
  | H : (wfTerms _ (Nil)) |- _ => inversion H; subst; clear H
  | H : (wfTerms _ (Cons _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
Fixpoint size_Tele (d : Tele) {struct d} : nat :=
  match d with
    | (TNil) => 1
    | (TCons T d0) => (plus 1 (plus (size_Tm T) (size_Tele d0)))
  end
with size_Tm (s : Tm) {struct s} : nat :=
  match s with
    | (Var x) => 1
    | (Abs d t) => (plus 1 (plus (size_Tele d) (size_Tm t)))
    | (Pi d0 T) => (plus 1 (plus (size_Tele d0) (size_Tm T)))
    | (App t0 ts) => (plus 1 (plus (size_Tm t0) (size_Terms ts)))
  end
with size_Terms (ts : Terms) {struct ts} : nat :=
  match ts with
    | (Nil) => 1
    | (Cons t ts0) => (plus 1 (plus (size_Tm t) (size_Terms ts0)))
  end.
Lemma shift_TmVar__size_Tele_Tm_Terms  :
  (forall (d26 : Tele) (c2 : (Cutoff TmVar)) ,
    ((size_Tele (shift_TmVar_Tele c2 d26)) =
    (size_Tele d26))) /\
  (forall (s12 : Tm) (c2 : (Cutoff TmVar)) ,
    ((size_Tm (shift_TmVar_Tm c2 s12)) =
    (size_Tm s12))) /\
  (forall (ts9 : Terms) (c2 : (Cutoff TmVar)) ,
    ((size_Terms (shift_TmVar_Terms c2 ts9)) =
    (size_Terms ts9))).
Proof.
  apply_mutual_ind ind_Tele_Tm_Terms.
  - intros c2; simpl; apply ((f_equal S)); reflexivity.
  - intros T9 IHT9 d26 IHd26.
  intros c2; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT9 c2)).
  + exact ((IHd26 (CS c2))).
  - intros x16 c2.
  reflexivity.
  - intros d27 IHd27 t14 IHt14.
  intros c3; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHd27 c3)).
  + exact ((IHt14 (weakenCutoffTmVar c3 (bind d27)))).
  - intros d28 IHd28 T10 IHT10.
  intros c3; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHd28 c3)).
  + exact ((IHT10 (weakenCutoffTmVar c3 (bind d28)))).
  - intros t15 IHt15 ts9 IHts9.
  intros c3; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt15 c3)).
  + exact ((IHts9 c3)).
  - intros c3; simpl; apply ((f_equal S)); reflexivity.
  - intros t16 IHt16 ts10 IHts10.
  intros c3; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt16 c3)).
  + exact ((IHts10 c3)).
Qed.

Lemma shift_TmVar__size_Tele  :
  (forall (d29 : Tele) (c3 : (Cutoff TmVar)) ,
    ((size_Tele (shift_TmVar_Tele c3 d29)) =
    (size_Tele d29))).
Proof.
  pose proof (shift_TmVar__size_Tele_Tm_Terms).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TmVar__size_Tm  :
  (forall (s12 : Tm) (c3 : (Cutoff TmVar)) ,
    ((size_Tm (shift_TmVar_Tm c3 s12)) =
    (size_Tm s12))).
Proof.
  pose proof (shift_TmVar__size_Tele_Tm_Terms).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TmVar__size_Terms  :
  (forall (ts11 : Terms) (c3 : (Cutoff TmVar)) ,
    ((size_Terms (shift_TmVar_Terms c3 ts11)) =
    (size_Terms ts11))).
Proof.
  pose proof (shift_TmVar__size_Tele_Tm_Terms).
  destruct_conjs .
  easy .
Qed.
 Hint Rewrite shift_TmVar__size_Tele shift_TmVar__size_Terms shift_TmVar__size_Tm : interaction.
 Hint Rewrite shift_TmVar__size_Tele shift_TmVar__size_Terms shift_TmVar__size_Tm : shift_size.
Lemma weaken_size_Tele  :
  (forall (k11 : Hvl) (d29 : Tele) ,
    ((size_Tele (weakenTele d29 k11)) =
    (size_Tele d29))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Terms  :
  (forall (k11 : Hvl) (ts11 : Terms) ,
    ((size_Terms (weakenTerms ts11 k11)) =
    (size_Terms ts11))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Tm  :
  (forall (k11 : Hvl) (s12 : Tm) ,
    ((size_Tm (weakenTm s12 k11)) =
    (size_Tm s12))).
Proof.
  needleGenericWeakenSize .
Qed.
 Hint Rewrite weaken_size_Tele weaken_size_Terms weaken_size_Tm : interaction.
 Hint Rewrite weaken_size_Tele weaken_size_Terms weaken_size_Tm : weaken_size.