Require Import Coq.Program.Equality.
Require Import Coq.Program.Tactics.
Require Export Needle.
Section Namespace.
  Inductive Namespace : Type :=
    | TmVar 
    | TyVar .
  Lemma eq_namespace_dec (a : Namespace) (b : Namespace) :
    (sumbool (a =
    b) (not (a =
    b))).
  Proof.
    decide equality .
  Defined.
End Namespace.

Section HVarlist.
  Inductive Hvl : Type :=
    | H0 
    | HS (a : Namespace) (k : Hvl).
  
  Fixpoint appendHvl (k : Hvl) (k0 : Hvl) {struct k0} : Hvl :=
    match k0 with
      | (H0) => k
      | (HS a k0) => (HS a (appendHvl k k0))
    end.
  
  Lemma appendHvl_assoc  :
    (forall (k : Hvl) (k0 : Hvl) (k1 : Hvl) ,
      ((appendHvl (appendHvl k k0) k1) =
      (appendHvl k (appendHvl k0 k1)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End HVarlist.

Section Index.
  Inductive Index (a : Namespace) : Type :=
    | I0 
    | IS : (Index a) -> (Index a).
  
  Global Arguments I0 [a] .
  Global Arguments IS [a] _ .
  
  Lemma eq_index_dec {a : Namespace} (i : (Index a)) (j : (Index a)) :
    (sumbool (i =
    j) (not (i =
    j))).
  Proof.
    decide equality .
  Qed.
  
  Fixpoint weakenIndexTmVar (x : (Index TmVar)) (k : Hvl) {struct k} : (Index TmVar) :=
    match k with
      | (H0) => x
      | (HS a k) => match a with
        | (TmVar) => (IS (weakenIndexTmVar x k))
        | _ => (weakenIndexTmVar x k)
      end
    end.
  Fixpoint weakenIndexTyVar (X : (Index TyVar)) (k : Hvl) {struct k} : (Index TyVar) :=
    match k with
      | (H0) => X
      | (HS a k) => match a with
        | (TyVar) => (IS (weakenIndexTyVar X k))
        | _ => (weakenIndexTyVar X k)
      end
    end.
  Lemma weakenIndexTmVar_append  :
    (forall (x : (Index TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTmVar (weakenIndexTmVar x k) k0) =
      (weakenIndexTmVar x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenIndexTyVar_append  :
    (forall (X : (Index TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTyVar (weakenIndexTyVar X k) k0) =
      (weakenIndexTyVar X (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End Index.

Section Terms.
  Inductive Tm : Set :=
    | Var (x : (Index TmVar))
    | Abs (T : Ty) (t : Tm)
    | App (t1 : Tm) (t2 : Tm)
    | All (T : Ty) (t : Tm)
  with Ty : Set :=
    | TVar (X : (Index TyVar))
    | TPi (T1 : Ty) (T2 : Ty)
    | TApp (T : Ty) (t : Tm)
    | Prp 
    | Prf .
  Scheme ind_Tm := Induction for Tm Sort Prop
  with ind_Ty := Induction for Ty Sort Prop.
  Combined Scheme ind_Tm_Ty from ind_Tm, ind_Ty.
  
  Inductive Kind : Set :=
    | Star 
    | KPi (T : Ty) (K : Kind).
  Scheme ind_Kind := Induction for Kind Sort Prop.
End Terms.

Section Functions.
  
End Functions.

Section Shift.
  Inductive Cutoff (a : Namespace) : Type :=
    | C0 
    | CS : (Cutoff a) -> (Cutoff a).
  
  Global Arguments C0 {a} .
  Global Arguments CS {a} _ .
  Fixpoint weakenCutoffTmVar (c : (Cutoff TmVar)) (k : Hvl) {struct k} : (Cutoff TmVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TmVar) => (CS (weakenCutoffTmVar c k))
        | _ => (weakenCutoffTmVar c k)
      end
    end.
  Fixpoint weakenCutoffTyVar (c : (Cutoff TyVar)) (k : Hvl) {struct k} : (Cutoff TyVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TyVar) => (CS (weakenCutoffTyVar c k))
        | _ => (weakenCutoffTyVar c k)
      end
    end.
  
  Lemma weakenCutoffTmVar_append  :
    (forall (c : (Cutoff TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTmVar (weakenCutoffTmVar c k) k0) =
      (weakenCutoffTmVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenCutoffTyVar_append  :
    (forall (c : (Cutoff TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTyVar (weakenCutoffTyVar c k) k0) =
      (weakenCutoffTyVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint shiftIndex (c : (Cutoff TmVar)) (x : (Index TmVar)) {struct c} : (Index TmVar) :=
    match c with
      | (C0) => (IS x)
      | (CS c) => match x with
        | (I0) => I0
        | (IS x) => (IS (shiftIndex c x))
      end
    end.
  Fixpoint tshiftIndex (c : (Cutoff TyVar)) (X : (Index TyVar)) {struct c} : (Index TyVar) :=
    match c with
      | (C0) => (IS X)
      | (CS c) => match X with
        | (I0) => I0
        | (IS X) => (IS (tshiftIndex c X))
      end
    end.
  Fixpoint shiftTm (c : (Cutoff TmVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var (shiftIndex c x))
      | (Abs T t) => (Abs (shiftTy (weakenCutoffTmVar c H0) T) (shiftTm (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (shiftTm (weakenCutoffTmVar c H0) t1) (shiftTm (weakenCutoffTmVar c H0) t2))
      | (All T0 t0) => (All (shiftTy (weakenCutoffTmVar c H0) T0) (shiftTm (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) t0))
    end
  with shiftTy (c : (Cutoff TmVar)) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (TVar X) => (TVar X)
      | (TPi T1 T2) => (TPi (shiftTy (weakenCutoffTmVar c H0) T1) (shiftTy (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) T2))
      | (TApp T t) => (TApp (shiftTy (weakenCutoffTmVar c H0) T) (shiftTm (weakenCutoffTmVar c H0) t))
      | (Prp) => (Prp)
      | (Prf) => (Prf)
    end.
  Fixpoint tshiftTm (c : (Cutoff TyVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (Abs T t) => (Abs (tshiftTy (weakenCutoffTyVar c H0) T) (tshiftTm (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (tshiftTm (weakenCutoffTyVar c H0) t1) (tshiftTm (weakenCutoffTyVar c H0) t2))
      | (All T0 t0) => (All (tshiftTy (weakenCutoffTyVar c H0) T0) (tshiftTm (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) t0))
    end
  with tshiftTy (c : (Cutoff TyVar)) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (TVar X) => (TVar (tshiftIndex c X))
      | (TPi T1 T2) => (TPi (tshiftTy (weakenCutoffTyVar c H0) T1) (tshiftTy (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) T2))
      | (TApp T t) => (TApp (tshiftTy (weakenCutoffTyVar c H0) T) (tshiftTm (weakenCutoffTyVar c H0) t))
      | (Prp) => (Prp)
      | (Prf) => (Prf)
    end.
  Fixpoint shiftKind (c : (Cutoff TmVar)) (K : Kind) {struct K} : Kind :=
    match K with
      | (Star) => (Star)
      | (KPi T K0) => (KPi (shiftTy (weakenCutoffTmVar c H0) T) (shiftKind (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) K0))
    end.
  Fixpoint tshiftKind (c : (Cutoff TyVar)) (K : Kind) {struct K} : Kind :=
    match K with
      | (Star) => (Star)
      | (KPi T K0) => (KPi (tshiftTy (weakenCutoffTyVar c H0) T) (tshiftKind (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) K0))
    end.
End Shift.

Section Weaken.
  Fixpoint weakenTm (s : Tm) (k : Hvl) {struct k} : Tm :=
    match k with
      | (H0) => s
      | (HS TmVar k) => (shiftTm C0 (weakenTm s k))
      | (HS TyVar k) => (tshiftTm C0 (weakenTm s k))
    end.
  Fixpoint weakenTy (S0 : Ty) (k : Hvl) {struct k} : Ty :=
    match k with
      | (H0) => S0
      | (HS TmVar k) => (shiftTy C0 (weakenTy S0 k))
      | (HS TyVar k) => (tshiftTy C0 (weakenTy S0 k))
    end.
  Fixpoint weakenKind (K : Kind) (k : Hvl) {struct k} : Kind :=
    match k with
      | (H0) => K
      | (HS TmVar k) => (shiftKind C0 (weakenKind K k))
      | (HS TyVar k) => (tshiftKind C0 (weakenKind K k))
    end.
End Weaken.

Section Subst.
  Inductive Trace (a : Namespace) : Type :=
    | X0 
    | XS (b : Namespace)
        (T : (Trace a)).
  
  Global Arguments X0 [a] .
  Global Arguments XS [a] b _ .
  Fixpoint weakenTrace {a : Namespace} (x : (Trace a)) (k : Hvl) {struct k} : (Trace a) :=
    match k with
      | (H0) => x
      | (HS b k) => (XS b (weakenTrace x k))
    end.
  Lemma weakenTrace_append (a : Namespace) :
    (forall (x : (Trace a)) (k : Hvl) (k0 : Hvl) ,
      ((weakenTrace (weakenTrace x k) k0) =
      (weakenTrace x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint substIndex (d : (Trace TmVar)) (s : Tm) (x : (Index TmVar)) {struct d} : Tm :=
    match d with
      | (X0) => match x with
        | (I0) => s
        | (IS x) => (Var x)
      end
      | (XS TmVar d) => match x with
        | (I0) => (Var I0)
        | (IS x) => (shiftTm C0 (substIndex d s x))
      end
      | (XS TyVar d) => (tshiftTm C0 (substIndex d s x))
    end.
  Fixpoint tsubstIndex (d : (Trace TyVar)) (S0 : Ty) (X : (Index TyVar)) {struct d} : Ty :=
    match d with
      | (X0) => match X with
        | (I0) => S0
        | (IS X) => (TVar X)
      end
      | (XS TmVar d) => (shiftTy C0 (tsubstIndex d S0 X))
      | (XS TyVar d) => match X with
        | (I0) => (TVar I0)
        | (IS X) => (tshiftTy C0 (tsubstIndex d S0 X))
      end
    end.
  Fixpoint substTm (d : (Trace TmVar)) (s : Tm) (s0 : Tm) {struct s0} : Tm :=
    match s0 with
      | (Var x) => (substIndex d s x)
      | (Abs T t) => (Abs (substTy d s T) (substTm (XS TmVar d) s t))
      | (App t1 t2) => (App (substTm d s t1) (substTm d s t2))
      | (All T0 t0) => (All (substTy d s T0) (substTm (XS TmVar d) s t0))
    end
  with substTy (d : (Trace TmVar)) (s : Tm) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (TVar X) => (TVar X)
      | (TPi T1 T2) => (TPi (substTy d s T1) (substTy (XS TmVar d) s T2))
      | (TApp T t) => (TApp (substTy d s T) (substTm d s t))
      | (Prp) => (Prp)
      | (Prf) => (Prf)
    end.
  Fixpoint tsubstTm (d : (Trace TyVar)) (S0 : Ty) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (Abs T t) => (Abs (tsubstTy d S0 T) (tsubstTm (XS TmVar d) S0 t))
      | (App t1 t2) => (App (tsubstTm d S0 t1) (tsubstTm d S0 t2))
      | (All T0 t0) => (All (tsubstTy d S0 T0) (tsubstTm (XS TmVar d) S0 t0))
    end
  with tsubstTy (d : (Trace TyVar)) (S0 : Ty) (S1 : Ty) {struct S1} : Ty :=
    match S1 with
      | (TVar X) => (tsubstIndex d S0 X)
      | (TPi T1 T2) => (TPi (tsubstTy d S0 T1) (tsubstTy (XS TmVar d) S0 T2))
      | (TApp T t) => (TApp (tsubstTy d S0 T) (tsubstTm d S0 t))
      | (Prp) => (Prp)
      | (Prf) => (Prf)
    end.
  Fixpoint substKind (d : (Trace TmVar)) (s : Tm) (K : Kind) {struct K} : Kind :=
    match K with
      | (Star) => (Star)
      | (KPi T K0) => (KPi (substTy d s T) (substKind (XS TmVar d) s K0))
    end.
  Fixpoint tsubstKind (d : (Trace TyVar)) (S0 : Ty) (K : Kind) {struct K} : Kind :=
    match K with
      | (Star) => (Star)
      | (KPi T K0) => (KPi (tsubstTy d S0 T) (tsubstKind (XS TmVar d) S0 K0))
    end.
End Subst.

Global Hint Resolve (f_equal (shiftKind C0)) : cong_shift0.
Global Hint Resolve (f_equal (tshiftKind C0)) : cong_shift0.
Global Hint Resolve (f_equal (shiftTm C0)) : cong_shift0.
Global Hint Resolve (f_equal (tshiftTm C0)) : cong_shift0.
Global Hint Resolve (f_equal (shiftTy C0)) : cong_shift0.
Global Hint Resolve (f_equal (tshiftTy C0)) : cong_shift0.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : interaction.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : weakencutoff_append.
 Hint Rewrite weakenTrace_append : interaction.
 Hint Rewrite weakenTrace_append : weakentrace_append.
Section SubstShiftReflection.
  Lemma substIndex0_shiftIndex0_reflection_ind (s : Tm) :
    (forall (k7 : Hvl) (x23 : (Index TmVar)) ,
      ((substIndex (weakenTrace X0 k7) s (shiftIndex (weakenCutoffTmVar C0 k7) x23)) =
      (Var x23))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Lemma tsubstIndex0_tshiftIndex0_reflection_ind (S0 : Ty) :
    (forall (k7 : Hvl) (X7 : (Index TyVar)) ,
      ((tsubstIndex (weakenTrace X0 k7) S0 (tshiftIndex (weakenCutoffTyVar C0 k7) X7)) =
      (TVar X7))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Definition subst0_shift0_Tm_Ty_reflection_ind := (ind_Tm_Ty (fun (s1 : Tm) =>
    (forall (k11 : Hvl) (s2 : Tm) ,
      ((substTm (weakenTrace X0 k11) s2 (shiftTm (weakenCutoffTmVar C0 k11) s1)) =
      s1))) (fun (S2 : Ty) =>
    (forall (k11 : Hvl) (s1 : Tm) ,
      ((substTy (weakenTrace X0 k11) s1 (shiftTy (weakenCutoffTmVar C0 k11) S2)) =
      S2))) (fun (x35 : (Index TmVar)) =>
    (fun (k11 : Hvl) (s1 : Tm) =>
      (substIndex0_shiftIndex0_reflection_ind s1 k11 x35))) (fun (T : Ty) IHT35 (t : Tm) IHt29 (k11 : Hvl) (s1 : Tm) =>
    (f_equal2 Abs (IHT35 k11 s1) (eq_trans (f_equal3 substTm (weakenTrace_append TmVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHt29 (HS TmVar k11) s1)))) (fun (t1 : Tm) IHt29 (t2 : Tm) IHt30 (k11 : Hvl) (s1 : Tm) =>
    (f_equal2 App (IHt29 k11 s1) (IHt30 k11 s1))) (fun (T : Ty) IHT35 (t : Tm) IHt29 (k11 : Hvl) (s1 : Tm) =>
    (f_equal2 All (IHT35 k11 s1) (eq_trans (f_equal3 substTm (weakenTrace_append TmVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHt29 (HS TmVar k11) s1)))) (fun (X11 : (Index TyVar)) =>
    (fun (k11 : Hvl) (s1 : Tm) =>
      eq_refl)) (fun (T1 : Ty) IHT35 (T2 : Ty) IHT36 (k11 : Hvl) (s1 : Tm) =>
    (f_equal2 TPi (IHT35 k11 s1) (eq_trans (f_equal3 substTy (weakenTrace_append TmVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHT36 (HS TmVar k11) s1)))) (fun (T : Ty) IHT35 (t : Tm) IHt29 (k11 : Hvl) (s1 : Tm) =>
    (f_equal2 TApp (IHT35 k11 s1) (IHt29 k11 s1))) (fun (k11 : Hvl) (s1 : Tm) =>
    eq_refl) (fun (k11 : Hvl) (s1 : Tm) =>
    eq_refl)).
  Lemma subst0_shift0_Tm_reflection_ind  :
    (forall (s1 : Tm) ,
      (forall (k11 : Hvl) (s2 : Tm) ,
        ((substTm (weakenTrace X0 k11) s2 (shiftTm (weakenCutoffTmVar C0 k11) s1)) =
        s1))).
  Proof.
    pose proof (subst0_shift0_Tm_Ty_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst0_shift0_Ty_reflection_ind  :
    (forall (S2 : Ty) ,
      (forall (k11 : Hvl) (s2 : Tm) ,
        ((substTy (weakenTrace X0 k11) s2 (shiftTy (weakenCutoffTmVar C0 k11) S2)) =
        S2))).
  Proof.
    pose proof (subst0_shift0_Tm_Ty_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Definition tsubst0_tshift0_Tm_Ty_reflection_ind := (ind_Tm_Ty (fun (s1 : Tm) =>
    (forall (k11 : Hvl) (S2 : Ty) ,
      ((tsubstTm (weakenTrace X0 k11) S2 (tshiftTm (weakenCutoffTyVar C0 k11) s1)) =
      s1))) (fun (S2 : Ty) =>
    (forall (k11 : Hvl) (S3 : Ty) ,
      ((tsubstTy (weakenTrace X0 k11) S3 (tshiftTy (weakenCutoffTyVar C0 k11) S2)) =
      S2))) (fun (x35 : (Index TmVar)) =>
    (fun (k11 : Hvl) (S2 : Ty) =>
      eq_refl)) (fun (T : Ty) IHT35 (t : Tm) IHt29 (k11 : Hvl) (S2 : Ty) =>
    (f_equal2 Abs (IHT35 k11 S2) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHt29 (HS TmVar k11) S2)))) (fun (t1 : Tm) IHt29 (t2 : Tm) IHt30 (k11 : Hvl) (S2 : Ty) =>
    (f_equal2 App (IHt29 k11 S2) (IHt30 k11 S2))) (fun (T : Ty) IHT35 (t : Tm) IHt29 (k11 : Hvl) (S2 : Ty) =>
    (f_equal2 All (IHT35 k11 S2) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHt29 (HS TmVar k11) S2)))) (fun (X11 : (Index TyVar)) =>
    (fun (k11 : Hvl) (S2 : Ty) =>
      (tsubstIndex0_tshiftIndex0_reflection_ind S2 k11 X11))) (fun (T1 : Ty) IHT35 (T2 : Ty) IHT36 (k11 : Hvl) (S2 : Ty) =>
    (f_equal2 TPi (IHT35 k11 S2) (eq_trans (f_equal3 tsubstTy (weakenTrace_append TyVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHT36 (HS TmVar k11) S2)))) (fun (T : Ty) IHT35 (t : Tm) IHt29 (k11 : Hvl) (S2 : Ty) =>
    (f_equal2 TApp (IHT35 k11 S2) (IHt29 k11 S2))) (fun (k11 : Hvl) (S2 : Ty) =>
    eq_refl) (fun (k11 : Hvl) (S2 : Ty) =>
    eq_refl)).
  Lemma tsubst0_tshift0_Tm_reflection_ind  :
    (forall (s1 : Tm) ,
      (forall (k11 : Hvl) (S2 : Ty) ,
        ((tsubstTm (weakenTrace X0 k11) S2 (tshiftTm (weakenCutoffTyVar C0 k11) s1)) =
        s1))).
  Proof.
    pose proof (tsubst0_tshift0_Tm_Ty_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma tsubst0_tshift0_Ty_reflection_ind  :
    (forall (S2 : Ty) ,
      (forall (k11 : Hvl) (S3 : Ty) ,
        ((tsubstTy (weakenTrace X0 k11) S3 (tshiftTy (weakenCutoffTyVar C0 k11) S2)) =
        S2))).
  Proof.
    pose proof (tsubst0_tshift0_Tm_Ty_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Definition subst0_shift0_Kind_reflection_ind := (ind_Kind (fun (K5 : Kind) =>
    (forall (k11 : Hvl) (s1 : Tm) ,
      ((substKind (weakenTrace X0 k11) s1 (shiftKind (weakenCutoffTmVar C0 k11) K5)) =
      K5))) (fun (k11 : Hvl) (s1 : Tm) =>
    eq_refl) (fun (T : Ty) (K : Kind) IHK5 (k11 : Hvl) (s1 : Tm) =>
    (f_equal2 KPi (let IHT35 := (subst0_shift0_Ty_reflection_ind T) in
    (IHT35 k11 s1)) (eq_trans (f_equal3 substKind (weakenTrace_append TmVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHK5 (HS TmVar k11) s1))))).
  Definition tsubst0_tshift0_Kind_reflection_ind := (ind_Kind (fun (K5 : Kind) =>
    (forall (k11 : Hvl) (S2 : Ty) ,
      ((tsubstKind (weakenTrace X0 k11) S2 (tshiftKind (weakenCutoffTyVar C0 k11) K5)) =
      K5))) (fun (k11 : Hvl) (S2 : Ty) =>
    eq_refl) (fun (T : Ty) (K : Kind) IHK5 (k11 : Hvl) (S2 : Ty) =>
    (f_equal2 KPi (let IHT35 := (tsubst0_tshift0_Ty_reflection_ind T) in
    (IHT35 k11 S2)) (eq_trans (f_equal3 tsubstKind (weakenTrace_append TyVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHK5 (HS TmVar k11) S2))))).
  Definition substTm0_shiftTm0_reflection (s2 : Tm) : (forall (s1 : Tm) ,
    ((substTm X0 s1 (shiftTm C0 s2)) =
    s2)) := (subst0_shift0_Tm_reflection_ind s2 H0).
  Definition substTy0_shiftTy0_reflection (S2 : Ty) : (forall (s1 : Tm) ,
    ((substTy X0 s1 (shiftTy C0 S2)) =
    S2)) := (subst0_shift0_Ty_reflection_ind S2 H0).
  Definition tsubstTm0_tshiftTm0_reflection (s1 : Tm) : (forall (S2 : Ty) ,
    ((tsubstTm X0 S2 (tshiftTm C0 s1)) =
    s1)) := (tsubst0_tshift0_Tm_reflection_ind s1 H0).
  Definition tsubstTy0_tshiftTy0_reflection (S3 : Ty) : (forall (S2 : Ty) ,
    ((tsubstTy X0 S2 (tshiftTy C0 S3)) =
    S3)) := (tsubst0_tshift0_Ty_reflection_ind S3 H0).
  Definition substKind0_shiftKind0_reflection (K5 : Kind) : (forall (s1 : Tm) ,
    ((substKind X0 s1 (shiftKind C0 K5)) =
    K5)) := (subst0_shift0_Kind_reflection_ind K5 H0).
  Definition tsubstKind0_tshiftKind0_reflection (K5 : Kind) : (forall (S2 : Ty) ,
    ((tsubstKind X0 S2 (tshiftKind C0 K5)) =
    K5)) := (tsubst0_tshift0_Kind_reflection_ind K5 H0).
End SubstShiftReflection.
Section ShiftInteraction.
  Section ShiftIndexCommInd.
    Lemma shiftIndex_shiftIndex0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff TmVar)) (x : (Index TmVar)) ,
        ((shiftIndex (weakenCutoffTmVar (CS c) k) (shiftIndex (weakenCutoffTmVar C0 k) x)) =
        (shiftIndex (weakenCutoffTmVar C0 k) (shiftIndex (weakenCutoffTmVar c k) x)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
    Lemma tshiftIndex_tshiftIndex0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff TyVar)) (X : (Index TyVar)) ,
        ((tshiftIndex (weakenCutoffTyVar (CS c) k) (tshiftIndex (weakenCutoffTyVar C0 k) X)) =
        (tshiftIndex (weakenCutoffTyVar C0 k) (tshiftIndex (weakenCutoffTyVar c k) X)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
  End ShiftIndexCommInd.
  Section ShiftCommInd.
    Definition shift_shift0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s : Tm) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
        ((shiftTm (weakenCutoffTmVar (CS c7) k7) (shiftTm (weakenCutoffTmVar C0 k7) s)) =
        (shiftTm (weakenCutoffTmVar C0 k7) (shiftTm (weakenCutoffTmVar c7 k7) s))))) (fun (S0 : Ty) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
        ((shiftTy (weakenCutoffTmVar (CS c7) k7) (shiftTy (weakenCutoffTmVar C0 k7) S0)) =
        (shiftTy (weakenCutoffTmVar C0 k7) (shiftTy (weakenCutoffTmVar c7 k7) S0))))) (fun (x23 : (Index TmVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
        (f_equal Var (shiftIndex_shiftIndex0_comm_ind k7 c7 x23)))) (fun (T : Ty) IHT23 (t : Tm) IHt19 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 Abs (IHT23 k7 c7) (IHt19 (HS TmVar k7) c7))) (fun (t1 : Tm) IHt19 (t2 : Tm) IHt20 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt19 k7 c7) (IHt20 k7 c7))) (fun (T : Ty) IHT23 (t : Tm) IHt19 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 All (IHT23 k7 c7) (IHt19 (HS TmVar k7) c7))) (fun (X7 : (Index TyVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T1 : Ty) IHT23 (T2 : Ty) IHT24 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 TPi (IHT23 k7 c7) (IHT24 (HS TmVar k7) c7))) (fun (T : Ty) IHT23 (t : Tm) IHt19 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 TApp (IHT23 k7 c7) (IHt19 k7 c7))) (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      eq_refl) (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      eq_refl)).
    Lemma shift_shift0_Tm_comm_ind  :
      (forall (s : Tm) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
          ((shiftTm (weakenCutoffTmVar (CS c7) k7) (shiftTm (weakenCutoffTmVar C0 k7) s)) =
          (shiftTm (weakenCutoffTmVar C0 k7) (shiftTm (weakenCutoffTmVar c7 k7) s))))).
    Proof.
      pose proof (shift_shift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_shift0_Ty_comm_ind  :
      (forall (S0 : Ty) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
          ((shiftTy (weakenCutoffTmVar (CS c7) k7) (shiftTy (weakenCutoffTmVar C0 k7) S0)) =
          (shiftTy (weakenCutoffTmVar C0 k7) (shiftTy (weakenCutoffTmVar c7 k7) S0))))).
    Proof.
      pose proof (shift_shift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_tshift0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s : Tm) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
        ((shiftTm (weakenCutoffTmVar c7 k7) (tshiftTm (weakenCutoffTyVar C0 k7) s)) =
        (tshiftTm (weakenCutoffTyVar C0 k7) (shiftTm (weakenCutoffTmVar c7 k7) s))))) (fun (S0 : Ty) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
        ((shiftTy (weakenCutoffTmVar c7 k7) (tshiftTy (weakenCutoffTyVar C0 k7) S0)) =
        (tshiftTy (weakenCutoffTyVar C0 k7) (shiftTy (weakenCutoffTmVar c7 k7) S0))))) (fun (x23 : (Index TmVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T : Ty) IHT23 (t : Tm) IHt19 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 Abs (IHT23 k7 c7) (IHt19 (HS TmVar k7) c7))) (fun (t1 : Tm) IHt19 (t2 : Tm) IHt20 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt19 k7 c7) (IHt20 k7 c7))) (fun (T : Ty) IHT23 (t : Tm) IHt19 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 All (IHT23 k7 c7) (IHt19 (HS TmVar k7) c7))) (fun (X7 : (Index TyVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T1 : Ty) IHT23 (T2 : Ty) IHT24 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 TPi (IHT23 k7 c7) (IHT24 (HS TmVar k7) c7))) (fun (T : Ty) IHT23 (t : Tm) IHt19 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 TApp (IHT23 k7 c7) (IHt19 k7 c7))) (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      eq_refl) (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      eq_refl)).
    Lemma shift_tshift0_Tm_comm_ind  :
      (forall (s : Tm) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
          ((shiftTm (weakenCutoffTmVar c7 k7) (tshiftTm (weakenCutoffTyVar C0 k7) s)) =
          (tshiftTm (weakenCutoffTyVar C0 k7) (shiftTm (weakenCutoffTmVar c7 k7) s))))).
    Proof.
      pose proof (shift_tshift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_tshift0_Ty_comm_ind  :
      (forall (S0 : Ty) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
          ((shiftTy (weakenCutoffTmVar c7 k7) (tshiftTy (weakenCutoffTyVar C0 k7) S0)) =
          (tshiftTy (weakenCutoffTyVar C0 k7) (shiftTy (weakenCutoffTmVar c7 k7) S0))))).
    Proof.
      pose proof (shift_tshift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition tshift_shift0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s : Tm) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
        ((tshiftTm (weakenCutoffTyVar c7 k7) (shiftTm (weakenCutoffTmVar C0 k7) s)) =
        (shiftTm (weakenCutoffTmVar C0 k7) (tshiftTm (weakenCutoffTyVar c7 k7) s))))) (fun (S0 : Ty) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
        ((tshiftTy (weakenCutoffTyVar c7 k7) (shiftTy (weakenCutoffTmVar C0 k7) S0)) =
        (shiftTy (weakenCutoffTmVar C0 k7) (tshiftTy (weakenCutoffTyVar c7 k7) S0))))) (fun (x23 : (Index TmVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) IHT23 (t : Tm) IHt19 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 Abs (IHT23 k7 c7) (IHt19 (HS TmVar k7) c7))) (fun (t1 : Tm) IHt19 (t2 : Tm) IHt20 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt19 k7 c7) (IHt20 k7 c7))) (fun (T : Ty) IHT23 (t : Tm) IHt19 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 All (IHT23 k7 c7) (IHt19 (HS TmVar k7) c7))) (fun (X7 : (Index TyVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T1 : Ty) IHT23 (T2 : Ty) IHT24 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 TPi (IHT23 k7 c7) (IHT24 (HS TmVar k7) c7))) (fun (T : Ty) IHT23 (t : Tm) IHt19 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 TApp (IHT23 k7 c7) (IHt19 k7 c7))) (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      eq_refl) (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      eq_refl)).
    Lemma tshift_shift0_Tm_comm_ind  :
      (forall (s : Tm) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
          ((tshiftTm (weakenCutoffTyVar c7 k7) (shiftTm (weakenCutoffTmVar C0 k7) s)) =
          (shiftTm (weakenCutoffTmVar C0 k7) (tshiftTm (weakenCutoffTyVar c7 k7) s))))).
    Proof.
      pose proof (tshift_shift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma tshift_shift0_Ty_comm_ind  :
      (forall (S0 : Ty) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
          ((tshiftTy (weakenCutoffTyVar c7 k7) (shiftTy (weakenCutoffTmVar C0 k7) S0)) =
          (shiftTy (weakenCutoffTmVar C0 k7) (tshiftTy (weakenCutoffTyVar c7 k7) S0))))).
    Proof.
      pose proof (tshift_shift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition tshift_tshift0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s : Tm) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
        ((tshiftTm (weakenCutoffTyVar (CS c7) k7) (tshiftTm (weakenCutoffTyVar C0 k7) s)) =
        (tshiftTm (weakenCutoffTyVar C0 k7) (tshiftTm (weakenCutoffTyVar c7 k7) s))))) (fun (S0 : Ty) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
        ((tshiftTy (weakenCutoffTyVar (CS c7) k7) (tshiftTy (weakenCutoffTyVar C0 k7) S0)) =
        (tshiftTy (weakenCutoffTyVar C0 k7) (tshiftTy (weakenCutoffTyVar c7 k7) S0))))) (fun (x23 : (Index TmVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) IHT23 (t : Tm) IHt19 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 Abs (IHT23 k7 c7) (IHt19 (HS TmVar k7) c7))) (fun (t1 : Tm) IHt19 (t2 : Tm) IHt20 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt19 k7 c7) (IHt20 k7 c7))) (fun (T : Ty) IHT23 (t : Tm) IHt19 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 All (IHT23 k7 c7) (IHt19 (HS TmVar k7) c7))) (fun (X7 : (Index TyVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
        (f_equal TVar (tshiftIndex_tshiftIndex0_comm_ind k7 c7 X7)))) (fun (T1 : Ty) IHT23 (T2 : Ty) IHT24 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 TPi (IHT23 k7 c7) (IHT24 (HS TmVar k7) c7))) (fun (T : Ty) IHT23 (t : Tm) IHt19 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 TApp (IHT23 k7 c7) (IHt19 k7 c7))) (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      eq_refl) (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      eq_refl)).
    Lemma tshift_tshift0_Tm_comm_ind  :
      (forall (s : Tm) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
          ((tshiftTm (weakenCutoffTyVar (CS c7) k7) (tshiftTm (weakenCutoffTyVar C0 k7) s)) =
          (tshiftTm (weakenCutoffTyVar C0 k7) (tshiftTm (weakenCutoffTyVar c7 k7) s))))).
    Proof.
      pose proof (tshift_tshift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma tshift_tshift0_Ty_comm_ind  :
      (forall (S0 : Ty) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
          ((tshiftTy (weakenCutoffTyVar (CS c7) k7) (tshiftTy (weakenCutoffTyVar C0 k7) S0)) =
          (tshiftTy (weakenCutoffTyVar C0 k7) (tshiftTy (weakenCutoffTyVar c7 k7) S0))))).
    Proof.
      pose proof (tshift_tshift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_shift0_Kind_comm_ind := (ind_Kind (fun (K3 : Kind) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
        ((shiftKind (weakenCutoffTmVar (CS c7) k7) (shiftKind (weakenCutoffTmVar C0 k7) K3)) =
        (shiftKind (weakenCutoffTmVar C0 k7) (shiftKind (weakenCutoffTmVar c7 k7) K3))))) (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK3 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 KPi (let IHT23 := (shift_shift0_Ty_comm_ind T) in
      (IHT23 k7 c7)) (IHK3 (HS TmVar k7) c7)))).
    Definition shift_tshift0_Kind_comm_ind := (ind_Kind (fun (K3 : Kind) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
        ((shiftKind (weakenCutoffTmVar c7 k7) (tshiftKind (weakenCutoffTyVar C0 k7) K3)) =
        (tshiftKind (weakenCutoffTyVar C0 k7) (shiftKind (weakenCutoffTmVar c7 k7) K3))))) (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK3 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 KPi (let IHT23 := (shift_tshift0_Ty_comm_ind T) in
      (IHT23 k7 c7)) (IHK3 (HS TmVar k7) c7)))).
    Definition tshift_shift0_Kind_comm_ind := (ind_Kind (fun (K3 : Kind) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
        ((tshiftKind (weakenCutoffTyVar c7 k7) (shiftKind (weakenCutoffTmVar C0 k7) K3)) =
        (shiftKind (weakenCutoffTmVar C0 k7) (tshiftKind (weakenCutoffTyVar c7 k7) K3))))) (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK3 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 KPi (let IHT23 := (tshift_shift0_Ty_comm_ind T) in
      (IHT23 k7 c7)) (IHK3 (HS TmVar k7) c7)))).
    Definition tshift_tshift0_Kind_comm_ind := (ind_Kind (fun (K3 : Kind) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
        ((tshiftKind (weakenCutoffTyVar (CS c7) k7) (tshiftKind (weakenCutoffTyVar C0 k7) K3)) =
        (tshiftKind (weakenCutoffTyVar C0 k7) (tshiftKind (weakenCutoffTyVar c7 k7) K3))))) (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK3 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 KPi (let IHT23 := (tshift_tshift0_Ty_comm_ind T) in
      (IHT23 k7 c7)) (IHK3 (HS TmVar k7) c7)))).
  End ShiftCommInd.
  Section ShiftComm.
    Definition shift_shift0_Tm_comm (s : Tm) : (forall (c7 : (Cutoff TmVar)) ,
      ((shiftTm (CS c7) (shiftTm C0 s)) =
      (shiftTm C0 (shiftTm c7 s)))) := (shift_shift0_Tm_comm_ind s H0).
    Definition shift_shift0_Ty_comm (S0 : Ty) : (forall (c7 : (Cutoff TmVar)) ,
      ((shiftTy (CS c7) (shiftTy C0 S0)) =
      (shiftTy C0 (shiftTy c7 S0)))) := (shift_shift0_Ty_comm_ind S0 H0).
    Definition shift_tshift0_Tm_comm (s : Tm) : (forall (c7 : (Cutoff TmVar)) ,
      ((shiftTm c7 (tshiftTm C0 s)) =
      (tshiftTm C0 (shiftTm c7 s)))) := (shift_tshift0_Tm_comm_ind s H0).
    Definition shift_tshift0_Ty_comm (S0 : Ty) : (forall (c7 : (Cutoff TmVar)) ,
      ((shiftTy c7 (tshiftTy C0 S0)) =
      (tshiftTy C0 (shiftTy c7 S0)))) := (shift_tshift0_Ty_comm_ind S0 H0).
    Definition tshift_shift0_Tm_comm (s : Tm) : (forall (c7 : (Cutoff TyVar)) ,
      ((tshiftTm c7 (shiftTm C0 s)) =
      (shiftTm C0 (tshiftTm c7 s)))) := (tshift_shift0_Tm_comm_ind s H0).
    Definition tshift_shift0_Ty_comm (S0 : Ty) : (forall (c7 : (Cutoff TyVar)) ,
      ((tshiftTy c7 (shiftTy C0 S0)) =
      (shiftTy C0 (tshiftTy c7 S0)))) := (tshift_shift0_Ty_comm_ind S0 H0).
    Definition tshift_tshift0_Tm_comm (s : Tm) : (forall (c7 : (Cutoff TyVar)) ,
      ((tshiftTm (CS c7) (tshiftTm C0 s)) =
      (tshiftTm C0 (tshiftTm c7 s)))) := (tshift_tshift0_Tm_comm_ind s H0).
    Definition tshift_tshift0_Ty_comm (S0 : Ty) : (forall (c7 : (Cutoff TyVar)) ,
      ((tshiftTy (CS c7) (tshiftTy C0 S0)) =
      (tshiftTy C0 (tshiftTy c7 S0)))) := (tshift_tshift0_Ty_comm_ind S0 H0).
    Definition shift_shift0_Kind_comm (K3 : Kind) : (forall (c7 : (Cutoff TmVar)) ,
      ((shiftKind (CS c7) (shiftKind C0 K3)) =
      (shiftKind C0 (shiftKind c7 K3)))) := (shift_shift0_Kind_comm_ind K3 H0).
    Definition shift_tshift0_Kind_comm (K3 : Kind) : (forall (c7 : (Cutoff TmVar)) ,
      ((shiftKind c7 (tshiftKind C0 K3)) =
      (tshiftKind C0 (shiftKind c7 K3)))) := (shift_tshift0_Kind_comm_ind K3 H0).
    Definition tshift_shift0_Kind_comm (K3 : Kind) : (forall (c7 : (Cutoff TyVar)) ,
      ((tshiftKind c7 (shiftKind C0 K3)) =
      (shiftKind C0 (tshiftKind c7 K3)))) := (tshift_shift0_Kind_comm_ind K3 H0).
    Definition tshift_tshift0_Kind_comm (K3 : Kind) : (forall (c7 : (Cutoff TyVar)) ,
      ((tshiftKind (CS c7) (tshiftKind C0 K3)) =
      (tshiftKind C0 (tshiftKind c7 K3)))) := (tshift_tshift0_Kind_comm_ind K3 H0).
  End ShiftComm.
End ShiftInteraction.
 Hint Rewrite shift_shift0_Kind_comm shift_tshift0_Kind_comm tshift_shift0_Kind_comm tshift_tshift0_Kind_comm shift_shift0_Tm_comm shift_tshift0_Tm_comm tshift_shift0_Tm_comm tshift_tshift0_Tm_comm shift_shift0_Ty_comm shift_tshift0_Ty_comm tshift_shift0_Ty_comm tshift_tshift0_Ty_comm : interaction.
 Hint Rewrite shift_shift0_Kind_comm shift_tshift0_Kind_comm tshift_shift0_Kind_comm tshift_tshift0_Kind_comm shift_shift0_Tm_comm shift_tshift0_Tm_comm tshift_shift0_Tm_comm tshift_tshift0_Tm_comm shift_shift0_Ty_comm shift_tshift0_Ty_comm tshift_shift0_Ty_comm tshift_tshift0_Ty_comm : shift_shift0.
Section WeakenShift.
  Lemma weakenTm_shiftTm  :
    (forall (k67 : Hvl) (c15 : (Cutoff TmVar)) (s31 : Tm) ,
      ((weakenTm (shiftTm c15 s31) k67) =
      (shiftTm (weakenCutoffTmVar c15 k67) (weakenTm s31 k67)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTy_shiftTy  :
    (forall (k67 : Hvl) (c15 : (Cutoff TmVar)) (S32 : Ty) ,
      ((weakenTy (shiftTy c15 S32) k67) =
      (shiftTy (weakenCutoffTmVar c15 k67) (weakenTy S32 k67)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_tshiftTm  :
    (forall (k67 : Hvl) (c15 : (Cutoff TyVar)) (s31 : Tm) ,
      ((weakenTm (tshiftTm c15 s31) k67) =
      (tshiftTm (weakenCutoffTyVar c15 k67) (weakenTm s31 k67)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTy_tshiftTy  :
    (forall (k67 : Hvl) (c15 : (Cutoff TyVar)) (S32 : Ty) ,
      ((weakenTy (tshiftTy c15 S32) k67) =
      (tshiftTy (weakenCutoffTyVar c15 k67) (weakenTy S32 k67)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenKind_shiftKind  :
    (forall (k67 : Hvl) (c15 : (Cutoff TmVar)) (K15 : Kind) ,
      ((weakenKind (shiftKind c15 K15) k67) =
      (shiftKind (weakenCutoffTmVar c15 k67) (weakenKind K15 k67)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenKind_tshiftKind  :
    (forall (k67 : Hvl) (c15 : (Cutoff TyVar)) (K15 : Kind) ,
      ((weakenKind (tshiftKind c15 K15) k67) =
      (tshiftKind (weakenCutoffTyVar c15 k67) (weakenKind K15 k67)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
End WeakenShift.
Section ShiftSubstInteraction.
  Section ShiftSubstIndexCommInd.
    Lemma shiftTm_substIndex0_comm_ind (c7 : (Cutoff TmVar)) (s1 : Tm) :
      (forall (k11 : Hvl) (x35 : (Index TmVar)) ,
        ((shiftTm (weakenCutoffTmVar c7 k11) (substIndex (weakenTrace X0 k11) s1 x35)) =
        (substIndex (weakenTrace X0 k11) (shiftTm c7 s1) (shiftIndex (weakenCutoffTmVar (CS c7) k11) x35)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma tshiftTm_substIndex0_comm_ind (c7 : (Cutoff TyVar)) (s1 : Tm) :
      (forall (k11 : Hvl) (x35 : (Index TmVar)) ,
        ((tshiftTm (weakenCutoffTyVar c7 k11) (substIndex (weakenTrace X0 k11) s1 x35)) =
        (substIndex (weakenTrace X0 k11) (tshiftTm c7 s1) x35))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shiftTy_tsubstIndex0_comm_ind (c7 : (Cutoff TmVar)) (S2 : Ty) :
      (forall (k11 : Hvl) (X11 : (Index TyVar)) ,
        ((shiftTy (weakenCutoffTmVar c7 k11) (tsubstIndex (weakenTrace X0 k11) S2 X11)) =
        (tsubstIndex (weakenTrace X0 k11) (shiftTy c7 S2) X11))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma tshiftTy_tsubstIndex0_comm_ind (c7 : (Cutoff TyVar)) (S2 : Ty) :
      (forall (k11 : Hvl) (X11 : (Index TyVar)) ,
        ((tshiftTy (weakenCutoffTyVar c7 k11) (tsubstIndex (weakenTrace X0 k11) S2 X11)) =
        (tsubstIndex (weakenTrace X0 k11) (tshiftTy c7 S2) (tshiftIndex (weakenCutoffTyVar (CS c7) k11) X11)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End ShiftSubstIndexCommInd.
  Section ShiftSubstCommInd.
    Definition shift_subst0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s5 : Tm) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (s6 : Tm) ,
        ((shiftTm (weakenCutoffTmVar c15 k19) (substTm (weakenTrace X0 k19) s6 s5)) =
        (substTm (weakenTrace X0 k19) (shiftTm c15 s6) (shiftTm (weakenCutoffTmVar (CS c15) k19) s5))))) (fun (S6 : Ty) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) ,
        ((shiftTy (weakenCutoffTmVar c15 k19) (substTy (weakenTrace X0 k19) s5 S6)) =
        (substTy (weakenTrace X0 k19) (shiftTm c15 s5) (shiftTy (weakenCutoffTmVar (CS c15) k19) S6))))) (fun (x59 : (Index TmVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
        (shiftTm_substIndex0_comm_ind c15 s5 k19 x59))) (fun (T : Ty) IHT59 (t : Tm) IHt49 (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
      (f_equal2 Abs (IHT59 k19 c15 s5) (eq_trans (f_equal2 shiftTm eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt49 (HS TmVar k19) c15 s5) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt49 (t2 : Tm) IHt50 (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
      (f_equal2 App (IHt49 k19 c15 s5) (IHt50 k19 c15 s5))) (fun (T : Ty) IHT59 (t : Tm) IHt49 (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
      (f_equal2 All (IHT59 k19 c15 s5) (eq_trans (f_equal2 shiftTm eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt49 (HS TmVar k19) c15 s5) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (X19 : (Index TyVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
        eq_refl)) (fun (T1 : Ty) IHT59 (T2 : Ty) IHT60 (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
      (f_equal2 TPi (IHT59 k19 c15 s5) (eq_trans (f_equal2 shiftTy eq_refl (f_equal3 substTy (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT60 (HS TmVar k19) c15 s5) (f_equal3 substTy (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (T : Ty) IHT59 (t : Tm) IHt49 (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
      (f_equal2 TApp (IHT59 k19 c15 s5) (IHt49 k19 c15 s5))) (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
      eq_refl) (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
      eq_refl)).
    Lemma shift_subst0_Tm_comm_ind  :
      (forall (s5 : Tm) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (s6 : Tm) ,
          ((shiftTm (weakenCutoffTmVar c15 k19) (substTm (weakenTrace X0 k19) s6 s5)) =
          (substTm (weakenTrace X0 k19) (shiftTm c15 s6) (shiftTm (weakenCutoffTmVar (CS c15) k19) s5))))).
    Proof.
      pose proof (shift_subst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_subst0_Ty_comm_ind  :
      (forall (S6 : Ty) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (s6 : Tm) ,
          ((shiftTy (weakenCutoffTmVar c15 k19) (substTy (weakenTrace X0 k19) s6 S6)) =
          (substTy (weakenTrace X0 k19) (shiftTm c15 s6) (shiftTy (weakenCutoffTmVar (CS c15) k19) S6))))).
    Proof.
      pose proof (shift_subst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_tsubst0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s5 : Tm) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) ,
        ((shiftTm (weakenCutoffTmVar c15 k19) (tsubstTm (weakenTrace X0 k19) S6 s5)) =
        (tsubstTm (weakenTrace X0 k19) (shiftTy c15 S6) (shiftTm (weakenCutoffTmVar c15 k19) s5))))) (fun (S6 : Ty) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (S7 : Ty) ,
        ((shiftTy (weakenCutoffTmVar c15 k19) (tsubstTy (weakenTrace X0 k19) S7 S6)) =
        (tsubstTy (weakenTrace X0 k19) (shiftTy c15 S7) (shiftTy (weakenCutoffTmVar c15 k19) S6))))) (fun (x59 : (Index TmVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
        eq_refl)) (fun (T : Ty) IHT59 (t : Tm) IHt49 (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 Abs (IHT59 k19 c15 S6) (eq_trans (f_equal2 shiftTm eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt49 (HS TmVar k19) c15 S6) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt49 (t2 : Tm) IHt50 (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 App (IHt49 k19 c15 S6) (IHt50 k19 c15 S6))) (fun (T : Ty) IHT59 (t : Tm) IHt49 (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 All (IHT59 k19 c15 S6) (eq_trans (f_equal2 shiftTm eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt49 (HS TmVar k19) c15 S6) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (X19 : (Index TyVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
        (shiftTy_tsubstIndex0_comm_ind c15 S6 k19 X19))) (fun (T1 : Ty) IHT59 (T2 : Ty) IHT60 (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 TPi (IHT59 k19 c15 S6) (eq_trans (f_equal2 shiftTy eq_refl (f_equal3 tsubstTy (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT60 (HS TmVar k19) c15 S6) (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (T : Ty) IHT59 (t : Tm) IHt49 (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 TApp (IHT59 k19 c15 S6) (IHt49 k19 c15 S6))) (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
      eq_refl) (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
      eq_refl)).
    Lemma shift_tsubst0_Tm_comm_ind  :
      (forall (s5 : Tm) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) ,
          ((shiftTm (weakenCutoffTmVar c15 k19) (tsubstTm (weakenTrace X0 k19) S6 s5)) =
          (tsubstTm (weakenTrace X0 k19) (shiftTy c15 S6) (shiftTm (weakenCutoffTmVar c15 k19) s5))))).
    Proof.
      pose proof (shift_tsubst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_tsubst0_Ty_comm_ind  :
      (forall (S6 : Ty) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (S7 : Ty) ,
          ((shiftTy (weakenCutoffTmVar c15 k19) (tsubstTy (weakenTrace X0 k19) S7 S6)) =
          (tsubstTy (weakenTrace X0 k19) (shiftTy c15 S7) (shiftTy (weakenCutoffTmVar c15 k19) S6))))).
    Proof.
      pose proof (shift_tsubst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition tshift_subst0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s5 : Tm) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (s6 : Tm) ,
        ((tshiftTm (weakenCutoffTyVar c15 k19) (substTm (weakenTrace X0 k19) s6 s5)) =
        (substTm (weakenTrace X0 k19) (tshiftTm c15 s6) (tshiftTm (weakenCutoffTyVar c15 k19) s5))))) (fun (S6 : Ty) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) ,
        ((tshiftTy (weakenCutoffTyVar c15 k19) (substTy (weakenTrace X0 k19) s5 S6)) =
        (substTy (weakenTrace X0 k19) (tshiftTm c15 s5) (tshiftTy (weakenCutoffTyVar c15 k19) S6))))) (fun (x59 : (Index TmVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
        (tshiftTm_substIndex0_comm_ind c15 s5 k19 x59))) (fun (T : Ty) IHT59 (t : Tm) IHt49 (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
      (f_equal2 Abs (IHT59 k19 c15 s5) (eq_trans (f_equal2 tshiftTm eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt49 (HS TmVar k19) c15 s5) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt49 (t2 : Tm) IHt50 (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
      (f_equal2 App (IHt49 k19 c15 s5) (IHt50 k19 c15 s5))) (fun (T : Ty) IHT59 (t : Tm) IHt49 (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
      (f_equal2 All (IHT59 k19 c15 s5) (eq_trans (f_equal2 tshiftTm eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt49 (HS TmVar k19) c15 s5) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (X19 : (Index TyVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
        eq_refl)) (fun (T1 : Ty) IHT59 (T2 : Ty) IHT60 (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
      (f_equal2 TPi (IHT59 k19 c15 s5) (eq_trans (f_equal2 tshiftTy eq_refl (f_equal3 substTy (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT60 (HS TmVar k19) c15 s5) (f_equal3 substTy (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (T : Ty) IHT59 (t : Tm) IHt49 (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
      (f_equal2 TApp (IHT59 k19 c15 s5) (IHt49 k19 c15 s5))) (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
      eq_refl) (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
      eq_refl)).
    Lemma tshift_subst0_Tm_comm_ind  :
      (forall (s5 : Tm) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (s6 : Tm) ,
          ((tshiftTm (weakenCutoffTyVar c15 k19) (substTm (weakenTrace X0 k19) s6 s5)) =
          (substTm (weakenTrace X0 k19) (tshiftTm c15 s6) (tshiftTm (weakenCutoffTyVar c15 k19) s5))))).
    Proof.
      pose proof (tshift_subst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma tshift_subst0_Ty_comm_ind  :
      (forall (S6 : Ty) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (s6 : Tm) ,
          ((tshiftTy (weakenCutoffTyVar c15 k19) (substTy (weakenTrace X0 k19) s6 S6)) =
          (substTy (weakenTrace X0 k19) (tshiftTm c15 s6) (tshiftTy (weakenCutoffTyVar c15 k19) S6))))).
    Proof.
      pose proof (tshift_subst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition tshift_tsubst0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s5 : Tm) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) ,
        ((tshiftTm (weakenCutoffTyVar c15 k19) (tsubstTm (weakenTrace X0 k19) S6 s5)) =
        (tsubstTm (weakenTrace X0 k19) (tshiftTy c15 S6) (tshiftTm (weakenCutoffTyVar (CS c15) k19) s5))))) (fun (S6 : Ty) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (S7 : Ty) ,
        ((tshiftTy (weakenCutoffTyVar c15 k19) (tsubstTy (weakenTrace X0 k19) S7 S6)) =
        (tsubstTy (weakenTrace X0 k19) (tshiftTy c15 S7) (tshiftTy (weakenCutoffTyVar (CS c15) k19) S6))))) (fun (x59 : (Index TmVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
        eq_refl)) (fun (T : Ty) IHT59 (t : Tm) IHt49 (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 Abs (IHT59 k19 c15 S6) (eq_trans (f_equal2 tshiftTm eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt49 (HS TmVar k19) c15 S6) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt49 (t2 : Tm) IHt50 (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 App (IHt49 k19 c15 S6) (IHt50 k19 c15 S6))) (fun (T : Ty) IHT59 (t : Tm) IHt49 (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 All (IHT59 k19 c15 S6) (eq_trans (f_equal2 tshiftTm eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt49 (HS TmVar k19) c15 S6) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (X19 : (Index TyVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
        (tshiftTy_tsubstIndex0_comm_ind c15 S6 k19 X19))) (fun (T1 : Ty) IHT59 (T2 : Ty) IHT60 (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 TPi (IHT59 k19 c15 S6) (eq_trans (f_equal2 tshiftTy eq_refl (f_equal3 tsubstTy (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT60 (HS TmVar k19) c15 S6) (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (T : Ty) IHT59 (t : Tm) IHt49 (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 TApp (IHT59 k19 c15 S6) (IHt49 k19 c15 S6))) (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
      eq_refl) (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
      eq_refl)).
    Lemma tshift_tsubst0_Tm_comm_ind  :
      (forall (s5 : Tm) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) ,
          ((tshiftTm (weakenCutoffTyVar c15 k19) (tsubstTm (weakenTrace X0 k19) S6 s5)) =
          (tsubstTm (weakenTrace X0 k19) (tshiftTy c15 S6) (tshiftTm (weakenCutoffTyVar (CS c15) k19) s5))))).
    Proof.
      pose proof (tshift_tsubst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma tshift_tsubst0_Ty_comm_ind  :
      (forall (S6 : Ty) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (S7 : Ty) ,
          ((tshiftTy (weakenCutoffTyVar c15 k19) (tsubstTy (weakenTrace X0 k19) S7 S6)) =
          (tsubstTy (weakenTrace X0 k19) (tshiftTy c15 S7) (tshiftTy (weakenCutoffTyVar (CS c15) k19) S6))))).
    Proof.
      pose proof (tshift_tsubst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_subst0_Kind_comm_ind := (ind_Kind (fun (K9 : Kind) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) ,
        ((shiftKind (weakenCutoffTmVar c15 k19) (substKind (weakenTrace X0 k19) s5 K9)) =
        (substKind (weakenTrace X0 k19) (shiftTm c15 s5) (shiftKind (weakenCutoffTmVar (CS c15) k19) K9))))) (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK9 (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
      (f_equal2 KPi (let IHT59 := (shift_subst0_Ty_comm_ind T) in
      (IHT59 k19 c15 s5)) (eq_trans (f_equal2 shiftKind eq_refl (f_equal3 substKind (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK9 (HS TmVar k19) c15 s5) (f_equal3 substKind (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl)))))).
    Definition shift_tsubst0_Kind_comm_ind := (ind_Kind (fun (K9 : Kind) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) ,
        ((shiftKind (weakenCutoffTmVar c15 k19) (tsubstKind (weakenTrace X0 k19) S6 K9)) =
        (tsubstKind (weakenTrace X0 k19) (shiftTy c15 S6) (shiftKind (weakenCutoffTmVar c15 k19) K9))))) (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK9 (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 KPi (let IHT59 := (shift_tsubst0_Ty_comm_ind T) in
      (IHT59 k19 c15 S6)) (eq_trans (f_equal2 shiftKind eq_refl (f_equal3 tsubstKind (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK9 (HS TmVar k19) c15 S6) (f_equal3 tsubstKind (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl)))))).
    Definition tshift_subst0_Kind_comm_ind := (ind_Kind (fun (K9 : Kind) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) ,
        ((tshiftKind (weakenCutoffTyVar c15 k19) (substKind (weakenTrace X0 k19) s5 K9)) =
        (substKind (weakenTrace X0 k19) (tshiftTm c15 s5) (tshiftKind (weakenCutoffTyVar c15 k19) K9))))) (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK9 (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
      (f_equal2 KPi (let IHT59 := (tshift_subst0_Ty_comm_ind T) in
      (IHT59 k19 c15 s5)) (eq_trans (f_equal2 tshiftKind eq_refl (f_equal3 substKind (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK9 (HS TmVar k19) c15 s5) (f_equal3 substKind (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl)))))).
    Definition tshift_tsubst0_Kind_comm_ind := (ind_Kind (fun (K9 : Kind) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) ,
        ((tshiftKind (weakenCutoffTyVar c15 k19) (tsubstKind (weakenTrace X0 k19) S6 K9)) =
        (tsubstKind (weakenTrace X0 k19) (tshiftTy c15 S6) (tshiftKind (weakenCutoffTyVar (CS c15) k19) K9))))) (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK9 (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 KPi (let IHT59 := (tshift_tsubst0_Ty_comm_ind T) in
      (IHT59 k19 c15 S6)) (eq_trans (f_equal2 tshiftKind eq_refl (f_equal3 tsubstKind (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK9 (HS TmVar k19) c15 S6) (f_equal3 tsubstKind (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl)))))).
  End ShiftSubstCommInd.
  Section ShiftSubstComm.
    Definition shiftTm_substTm0_comm (s6 : Tm) : (forall (c15 : (Cutoff TmVar)) (s5 : Tm) ,
      ((shiftTm c15 (substTm X0 s5 s6)) =
      (substTm X0 (shiftTm c15 s5) (shiftTm (CS c15) s6)))) := (shift_subst0_Tm_comm_ind s6 H0).
    Definition shiftTy_substTy0_comm (S6 : Ty) : (forall (c15 : (Cutoff TmVar)) (s5 : Tm) ,
      ((shiftTy c15 (substTy X0 s5 S6)) =
      (substTy X0 (shiftTm c15 s5) (shiftTy (CS c15) S6)))) := (shift_subst0_Ty_comm_ind S6 H0).
    Definition shiftTm_tsubstTm0_comm (s5 : Tm) : (forall (c15 : (Cutoff TmVar)) (S6 : Ty) ,
      ((shiftTm c15 (tsubstTm X0 S6 s5)) =
      (tsubstTm X0 (shiftTy c15 S6) (shiftTm c15 s5)))) := (shift_tsubst0_Tm_comm_ind s5 H0).
    Definition shiftTy_tsubstTy0_comm (S7 : Ty) : (forall (c15 : (Cutoff TmVar)) (S6 : Ty) ,
      ((shiftTy c15 (tsubstTy X0 S6 S7)) =
      (tsubstTy X0 (shiftTy c15 S6) (shiftTy c15 S7)))) := (shift_tsubst0_Ty_comm_ind S7 H0).
    Definition tshiftTm_substTm0_comm (s6 : Tm) : (forall (c15 : (Cutoff TyVar)) (s5 : Tm) ,
      ((tshiftTm c15 (substTm X0 s5 s6)) =
      (substTm X0 (tshiftTm c15 s5) (tshiftTm c15 s6)))) := (tshift_subst0_Tm_comm_ind s6 H0).
    Definition tshiftTy_substTy0_comm (S6 : Ty) : (forall (c15 : (Cutoff TyVar)) (s5 : Tm) ,
      ((tshiftTy c15 (substTy X0 s5 S6)) =
      (substTy X0 (tshiftTm c15 s5) (tshiftTy c15 S6)))) := (tshift_subst0_Ty_comm_ind S6 H0).
    Definition tshiftTm_tsubstTm0_comm (s5 : Tm) : (forall (c15 : (Cutoff TyVar)) (S6 : Ty) ,
      ((tshiftTm c15 (tsubstTm X0 S6 s5)) =
      (tsubstTm X0 (tshiftTy c15 S6) (tshiftTm (CS c15) s5)))) := (tshift_tsubst0_Tm_comm_ind s5 H0).
    Definition tshiftTy_tsubstTy0_comm (S7 : Ty) : (forall (c15 : (Cutoff TyVar)) (S6 : Ty) ,
      ((tshiftTy c15 (tsubstTy X0 S6 S7)) =
      (tsubstTy X0 (tshiftTy c15 S6) (tshiftTy (CS c15) S7)))) := (tshift_tsubst0_Ty_comm_ind S7 H0).
    Definition shiftKind_substKind0_comm (K9 : Kind) : (forall (c15 : (Cutoff TmVar)) (s5 : Tm) ,
      ((shiftKind c15 (substKind X0 s5 K9)) =
      (substKind X0 (shiftTm c15 s5) (shiftKind (CS c15) K9)))) := (shift_subst0_Kind_comm_ind K9 H0).
    Definition shiftKind_tsubstKind0_comm (K9 : Kind) : (forall (c15 : (Cutoff TmVar)) (S6 : Ty) ,
      ((shiftKind c15 (tsubstKind X0 S6 K9)) =
      (tsubstKind X0 (shiftTy c15 S6) (shiftKind c15 K9)))) := (shift_tsubst0_Kind_comm_ind K9 H0).
    Definition tshiftKind_substKind0_comm (K9 : Kind) : (forall (c15 : (Cutoff TyVar)) (s5 : Tm) ,
      ((tshiftKind c15 (substKind X0 s5 K9)) =
      (substKind X0 (tshiftTm c15 s5) (tshiftKind c15 K9)))) := (tshift_subst0_Kind_comm_ind K9 H0).
    Definition tshiftKind_tsubstKind0_comm (K9 : Kind) : (forall (c15 : (Cutoff TyVar)) (S6 : Ty) ,
      ((tshiftKind c15 (tsubstKind X0 S6 K9)) =
      (tsubstKind X0 (tshiftTy c15 S6) (tshiftKind (CS c15) K9)))) := (tshift_tsubst0_Kind_comm_ind K9 H0).
  End ShiftSubstComm.
  Section SubstShiftIndexCommInd.
    Lemma substIndex_shiftTm0_comm_ind (d : (Trace TmVar)) (s5 : Tm) :
      (forall (k19 : Hvl) (x59 : (Index TmVar)) ,
        ((substIndex (weakenTrace (XS TmVar d) k19) s5 (shiftIndex (weakenCutoffTmVar C0 k19) x59)) =
        (shiftTm (weakenCutoffTmVar C0 k19) (substIndex (weakenTrace d k19) s5 x59)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma substIndex_tshiftTm0_comm_ind (d : (Trace TmVar)) (s5 : Tm) :
      (forall (k19 : Hvl) (x59 : (Index TmVar)) ,
        ((substIndex (weakenTrace (XS TyVar d) k19) s5 x59) =
        (tshiftTm (weakenCutoffTyVar C0 k19) (substIndex (weakenTrace d k19) s5 x59)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma tsubstIndex_shiftTy0_comm_ind (d : (Trace TyVar)) (S6 : Ty) :
      (forall (k19 : Hvl) (X19 : (Index TyVar)) ,
        ((tsubstIndex (weakenTrace (XS TmVar d) k19) S6 X19) =
        (shiftTy (weakenCutoffTmVar C0 k19) (tsubstIndex (weakenTrace d k19) S6 X19)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma tsubstIndex_tshiftTy0_comm_ind (d : (Trace TyVar)) (S6 : Ty) :
      (forall (k19 : Hvl) (X19 : (Index TyVar)) ,
        ((tsubstIndex (weakenTrace (XS TyVar d) k19) S6 (tshiftIndex (weakenCutoffTyVar C0 k19) X19)) =
        (tshiftTy (weakenCutoffTyVar C0 k19) (tsubstIndex (weakenTrace d k19) S6 X19)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End SubstShiftIndexCommInd.
  Section SubstShiftCommInd.
    Definition subst_shift0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s27 : Tm) =>
      (forall (k63 : Hvl) (d43 : (Trace TmVar)) (s28 : Tm) ,
        ((substTm (weakenTrace (XS TmVar d43) k63) s28 (shiftTm (weakenCutoffTmVar C0 k63) s27)) =
        (shiftTm (weakenCutoffTmVar C0 k63) (substTm (weakenTrace d43 k63) s28 s27))))) (fun (S28 : Ty) =>
      (forall (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) ,
        ((substTy (weakenTrace (XS TmVar d43) k63) s27 (shiftTy (weakenCutoffTmVar C0 k63) S28)) =
        (shiftTy (weakenCutoffTmVar C0 k63) (substTy (weakenTrace d43 k63) s27 S28))))) (fun (x83 : (Index TmVar)) =>
      (fun (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
        (substIndex_shiftTm0_comm_ind d43 s27 k63 x83))) (fun (T : Ty) IHT83 (t : Tm) IHt69 (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      (f_equal2 Abs (IHT83 k63 d43 s27) (eq_trans (f_equal3 substTm (weakenTrace_append TmVar (XS TmVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt69 (HS TmVar k63) d43 s27) (f_equal2 shiftTm eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d43 k63 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      (f_equal2 App (IHt69 k63 d43 s27) (IHt70 k63 d43 s27))) (fun (T : Ty) IHT83 (t : Tm) IHt69 (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      (f_equal2 All (IHT83 k63 d43 s27) (eq_trans (f_equal3 substTm (weakenTrace_append TmVar (XS TmVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt69 (HS TmVar k63) d43 s27) (f_equal2 shiftTm eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d43 k63 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (X27 : (Index TyVar)) =>
      (fun (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
        eq_refl)) (fun (T1 : Ty) IHT83 (T2 : Ty) IHT84 (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      (f_equal2 TPi (IHT83 k63 d43 s27) (eq_trans (f_equal3 substTy (weakenTrace_append TmVar (XS TmVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHT84 (HS TmVar k63) d43 s27) (f_equal2 shiftTy eq_refl (f_equal3 substTy (eq_sym (weakenTrace_append TmVar d43 k63 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT83 (t : Tm) IHt69 (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      (f_equal2 TApp (IHT83 k63 d43 s27) (IHt69 k63 d43 s27))) (fun (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      eq_refl) (fun (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      eq_refl)).
    Lemma subst_shift0_Tm_comm_ind  :
      (forall (s27 : Tm) ,
        (forall (k63 : Hvl) (d43 : (Trace TmVar)) (s28 : Tm) ,
          ((substTm (weakenTrace (XS TmVar d43) k63) s28 (shiftTm (weakenCutoffTmVar C0 k63) s27)) =
          (shiftTm (weakenCutoffTmVar C0 k63) (substTm (weakenTrace d43 k63) s28 s27))))).
    Proof.
      pose proof (subst_shift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_shift0_Ty_comm_ind  :
      (forall (S28 : Ty) ,
        (forall (k63 : Hvl) (d43 : (Trace TmVar)) (s28 : Tm) ,
          ((substTy (weakenTrace (XS TmVar d43) k63) s28 (shiftTy (weakenCutoffTmVar C0 k63) S28)) =
          (shiftTy (weakenCutoffTmVar C0 k63) (substTy (weakenTrace d43 k63) s28 S28))))).
    Proof.
      pose proof (subst_shift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_tshift0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s27 : Tm) =>
      (forall (k63 : Hvl) (d43 : (Trace TmVar)) (s28 : Tm) ,
        ((substTm (weakenTrace (XS TyVar d43) k63) s28 (tshiftTm (weakenCutoffTyVar C0 k63) s27)) =
        (tshiftTm (weakenCutoffTyVar C0 k63) (substTm (weakenTrace d43 k63) s28 s27))))) (fun (S28 : Ty) =>
      (forall (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) ,
        ((substTy (weakenTrace (XS TyVar d43) k63) s27 (tshiftTy (weakenCutoffTyVar C0 k63) S28)) =
        (tshiftTy (weakenCutoffTyVar C0 k63) (substTy (weakenTrace d43 k63) s27 S28))))) (fun (x83 : (Index TmVar)) =>
      (fun (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
        (substIndex_tshiftTm0_comm_ind d43 s27 k63 x83))) (fun (T : Ty) IHT83 (t : Tm) IHt69 (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      (f_equal2 Abs (IHT83 k63 d43 s27) (eq_trans (f_equal3 substTm (weakenTrace_append TmVar (XS TyVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt69 (HS TmVar k63) d43 s27) (f_equal2 tshiftTm eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d43 k63 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      (f_equal2 App (IHt69 k63 d43 s27) (IHt70 k63 d43 s27))) (fun (T : Ty) IHT83 (t : Tm) IHt69 (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      (f_equal2 All (IHT83 k63 d43 s27) (eq_trans (f_equal3 substTm (weakenTrace_append TmVar (XS TyVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt69 (HS TmVar k63) d43 s27) (f_equal2 tshiftTm eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d43 k63 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (X27 : (Index TyVar)) =>
      (fun (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
        eq_refl)) (fun (T1 : Ty) IHT83 (T2 : Ty) IHT84 (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      (f_equal2 TPi (IHT83 k63 d43 s27) (eq_trans (f_equal3 substTy (weakenTrace_append TmVar (XS TyVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHT84 (HS TmVar k63) d43 s27) (f_equal2 tshiftTy eq_refl (f_equal3 substTy (eq_sym (weakenTrace_append TmVar d43 k63 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT83 (t : Tm) IHt69 (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      (f_equal2 TApp (IHT83 k63 d43 s27) (IHt69 k63 d43 s27))) (fun (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      eq_refl) (fun (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      eq_refl)).
    Lemma subst_tshift0_Tm_comm_ind  :
      (forall (s27 : Tm) ,
        (forall (k63 : Hvl) (d43 : (Trace TmVar)) (s28 : Tm) ,
          ((substTm (weakenTrace (XS TyVar d43) k63) s28 (tshiftTm (weakenCutoffTyVar C0 k63) s27)) =
          (tshiftTm (weakenCutoffTyVar C0 k63) (substTm (weakenTrace d43 k63) s28 s27))))).
    Proof.
      pose proof (subst_tshift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_tshift0_Ty_comm_ind  :
      (forall (S28 : Ty) ,
        (forall (k63 : Hvl) (d43 : (Trace TmVar)) (s28 : Tm) ,
          ((substTy (weakenTrace (XS TyVar d43) k63) s28 (tshiftTy (weakenCutoffTyVar C0 k63) S28)) =
          (tshiftTy (weakenCutoffTyVar C0 k63) (substTy (weakenTrace d43 k63) s28 S28))))).
    Proof.
      pose proof (subst_tshift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition tsubst_shift0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s27 : Tm) =>
      (forall (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) ,
        ((tsubstTm (weakenTrace (XS TmVar d43) k63) S28 (shiftTm (weakenCutoffTmVar C0 k63) s27)) =
        (shiftTm (weakenCutoffTmVar C0 k63) (tsubstTm (weakenTrace d43 k63) S28 s27))))) (fun (S28 : Ty) =>
      (forall (k63 : Hvl) (d43 : (Trace TyVar)) (S29 : Ty) ,
        ((tsubstTy (weakenTrace (XS TmVar d43) k63) S29 (shiftTy (weakenCutoffTmVar C0 k63) S28)) =
        (shiftTy (weakenCutoffTmVar C0 k63) (tsubstTy (weakenTrace d43 k63) S29 S28))))) (fun (x83 : (Index TmVar)) =>
      (fun (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
        eq_refl)) (fun (T : Ty) IHT83 (t : Tm) IHt69 (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      (f_equal2 Abs (IHT83 k63 d43 S28) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar (XS TmVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt69 (HS TmVar k63) d43 S28) (f_equal2 shiftTm eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d43 k63 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      (f_equal2 App (IHt69 k63 d43 S28) (IHt70 k63 d43 S28))) (fun (T : Ty) IHT83 (t : Tm) IHt69 (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      (f_equal2 All (IHT83 k63 d43 S28) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar (XS TmVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt69 (HS TmVar k63) d43 S28) (f_equal2 shiftTm eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d43 k63 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (X27 : (Index TyVar)) =>
      (fun (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
        (tsubstIndex_shiftTy0_comm_ind d43 S28 k63 X27))) (fun (T1 : Ty) IHT83 (T2 : Ty) IHT84 (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      (f_equal2 TPi (IHT83 k63 d43 S28) (eq_trans (f_equal3 tsubstTy (weakenTrace_append TyVar (XS TmVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHT84 (HS TmVar k63) d43 S28) (f_equal2 shiftTy eq_refl (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar d43 k63 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT83 (t : Tm) IHt69 (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      (f_equal2 TApp (IHT83 k63 d43 S28) (IHt69 k63 d43 S28))) (fun (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      eq_refl) (fun (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      eq_refl)).
    Lemma tsubst_shift0_Tm_comm_ind  :
      (forall (s27 : Tm) ,
        (forall (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) ,
          ((tsubstTm (weakenTrace (XS TmVar d43) k63) S28 (shiftTm (weakenCutoffTmVar C0 k63) s27)) =
          (shiftTm (weakenCutoffTmVar C0 k63) (tsubstTm (weakenTrace d43 k63) S28 s27))))).
    Proof.
      pose proof (tsubst_shift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma tsubst_shift0_Ty_comm_ind  :
      (forall (S28 : Ty) ,
        (forall (k63 : Hvl) (d43 : (Trace TyVar)) (S29 : Ty) ,
          ((tsubstTy (weakenTrace (XS TmVar d43) k63) S29 (shiftTy (weakenCutoffTmVar C0 k63) S28)) =
          (shiftTy (weakenCutoffTmVar C0 k63) (tsubstTy (weakenTrace d43 k63) S29 S28))))).
    Proof.
      pose proof (tsubst_shift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition tsubst_tshift0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s27 : Tm) =>
      (forall (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) ,
        ((tsubstTm (weakenTrace (XS TyVar d43) k63) S28 (tshiftTm (weakenCutoffTyVar C0 k63) s27)) =
        (tshiftTm (weakenCutoffTyVar C0 k63) (tsubstTm (weakenTrace d43 k63) S28 s27))))) (fun (S28 : Ty) =>
      (forall (k63 : Hvl) (d43 : (Trace TyVar)) (S29 : Ty) ,
        ((tsubstTy (weakenTrace (XS TyVar d43) k63) S29 (tshiftTy (weakenCutoffTyVar C0 k63) S28)) =
        (tshiftTy (weakenCutoffTyVar C0 k63) (tsubstTy (weakenTrace d43 k63) S29 S28))))) (fun (x83 : (Index TmVar)) =>
      (fun (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
        eq_refl)) (fun (T : Ty) IHT83 (t : Tm) IHt69 (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      (f_equal2 Abs (IHT83 k63 d43 S28) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar (XS TyVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt69 (HS TmVar k63) d43 S28) (f_equal2 tshiftTm eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d43 k63 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      (f_equal2 App (IHt69 k63 d43 S28) (IHt70 k63 d43 S28))) (fun (T : Ty) IHT83 (t : Tm) IHt69 (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      (f_equal2 All (IHT83 k63 d43 S28) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar (XS TyVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt69 (HS TmVar k63) d43 S28) (f_equal2 tshiftTm eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d43 k63 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (X27 : (Index TyVar)) =>
      (fun (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
        (tsubstIndex_tshiftTy0_comm_ind d43 S28 k63 X27))) (fun (T1 : Ty) IHT83 (T2 : Ty) IHT84 (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      (f_equal2 TPi (IHT83 k63 d43 S28) (eq_trans (f_equal3 tsubstTy (weakenTrace_append TyVar (XS TyVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHT84 (HS TmVar k63) d43 S28) (f_equal2 tshiftTy eq_refl (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar d43 k63 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT83 (t : Tm) IHt69 (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      (f_equal2 TApp (IHT83 k63 d43 S28) (IHt69 k63 d43 S28))) (fun (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      eq_refl) (fun (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      eq_refl)).
    Lemma tsubst_tshift0_Tm_comm_ind  :
      (forall (s27 : Tm) ,
        (forall (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) ,
          ((tsubstTm (weakenTrace (XS TyVar d43) k63) S28 (tshiftTm (weakenCutoffTyVar C0 k63) s27)) =
          (tshiftTm (weakenCutoffTyVar C0 k63) (tsubstTm (weakenTrace d43 k63) S28 s27))))).
    Proof.
      pose proof (tsubst_tshift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma tsubst_tshift0_Ty_comm_ind  :
      (forall (S28 : Ty) ,
        (forall (k63 : Hvl) (d43 : (Trace TyVar)) (S29 : Ty) ,
          ((tsubstTy (weakenTrace (XS TyVar d43) k63) S29 (tshiftTy (weakenCutoffTyVar C0 k63) S28)) =
          (tshiftTy (weakenCutoffTyVar C0 k63) (tsubstTy (weakenTrace d43 k63) S29 S28))))).
    Proof.
      pose proof (tsubst_tshift0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_shift0_Kind_comm_ind := (ind_Kind (fun (K13 : Kind) =>
      (forall (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) ,
        ((substKind (weakenTrace (XS TmVar d43) k63) s27 (shiftKind (weakenCutoffTmVar C0 k63) K13)) =
        (shiftKind (weakenCutoffTmVar C0 k63) (substKind (weakenTrace d43 k63) s27 K13))))) (fun (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK13 (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      (f_equal2 KPi (let IHT83 := (subst_shift0_Ty_comm_ind T) in
      (IHT83 k63 d43 s27)) (eq_trans (f_equal3 substKind (weakenTrace_append TmVar (XS TmVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHK13 (HS TmVar k63) d43 s27) (f_equal2 shiftKind eq_refl (f_equal3 substKind (eq_sym (weakenTrace_append TmVar d43 k63 (HS TmVar H0))) eq_refl eq_refl))))))).
    Definition subst_tshift0_Kind_comm_ind := (ind_Kind (fun (K13 : Kind) =>
      (forall (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) ,
        ((substKind (weakenTrace (XS TyVar d43) k63) s27 (tshiftKind (weakenCutoffTyVar C0 k63) K13)) =
        (tshiftKind (weakenCutoffTyVar C0 k63) (substKind (weakenTrace d43 k63) s27 K13))))) (fun (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK13 (k63 : Hvl) (d43 : (Trace TmVar)) (s27 : Tm) =>
      (f_equal2 KPi (let IHT83 := (subst_tshift0_Ty_comm_ind T) in
      (IHT83 k63 d43 s27)) (eq_trans (f_equal3 substKind (weakenTrace_append TmVar (XS TyVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHK13 (HS TmVar k63) d43 s27) (f_equal2 tshiftKind eq_refl (f_equal3 substKind (eq_sym (weakenTrace_append TmVar d43 k63 (HS TmVar H0))) eq_refl eq_refl))))))).
    Definition tsubst_shift0_Kind_comm_ind := (ind_Kind (fun (K13 : Kind) =>
      (forall (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) ,
        ((tsubstKind (weakenTrace (XS TmVar d43) k63) S28 (shiftKind (weakenCutoffTmVar C0 k63) K13)) =
        (shiftKind (weakenCutoffTmVar C0 k63) (tsubstKind (weakenTrace d43 k63) S28 K13))))) (fun (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK13 (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      (f_equal2 KPi (let IHT83 := (tsubst_shift0_Ty_comm_ind T) in
      (IHT83 k63 d43 S28)) (eq_trans (f_equal3 tsubstKind (weakenTrace_append TyVar (XS TmVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHK13 (HS TmVar k63) d43 S28) (f_equal2 shiftKind eq_refl (f_equal3 tsubstKind (eq_sym (weakenTrace_append TyVar d43 k63 (HS TmVar H0))) eq_refl eq_refl))))))).
    Definition tsubst_tshift0_Kind_comm_ind := (ind_Kind (fun (K13 : Kind) =>
      (forall (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) ,
        ((tsubstKind (weakenTrace (XS TyVar d43) k63) S28 (tshiftKind (weakenCutoffTyVar C0 k63) K13)) =
        (tshiftKind (weakenCutoffTyVar C0 k63) (tsubstKind (weakenTrace d43 k63) S28 K13))))) (fun (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK13 (k63 : Hvl) (d43 : (Trace TyVar)) (S28 : Ty) =>
      (f_equal2 KPi (let IHT83 := (tsubst_tshift0_Ty_comm_ind T) in
      (IHT83 k63 d43 S28)) (eq_trans (f_equal3 tsubstKind (weakenTrace_append TyVar (XS TyVar d43) k63 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHK13 (HS TmVar k63) d43 S28) (f_equal2 tshiftKind eq_refl (f_equal3 tsubstKind (eq_sym (weakenTrace_append TyVar d43 k63 (HS TmVar H0))) eq_refl eq_refl))))))).
  End SubstShiftCommInd.
  Section SubstShiftComm.
    Definition substTm_shiftTm0_comm (s28 : Tm) : (forall (d43 : (Trace TmVar)) (s27 : Tm) ,
      ((substTm (XS TmVar d43) s27 (shiftTm C0 s28)) =
      (shiftTm C0 (substTm d43 s27 s28)))) := (subst_shift0_Tm_comm_ind s28 H0).
    Definition substTy_shiftTy0_comm (S28 : Ty) : (forall (d43 : (Trace TmVar)) (s27 : Tm) ,
      ((substTy (XS TmVar d43) s27 (shiftTy C0 S28)) =
      (shiftTy C0 (substTy d43 s27 S28)))) := (subst_shift0_Ty_comm_ind S28 H0).
    Definition substTm_tshiftTm0_comm (s28 : Tm) : (forall (d43 : (Trace TmVar)) (s27 : Tm) ,
      ((substTm (XS TyVar d43) s27 (tshiftTm C0 s28)) =
      (tshiftTm C0 (substTm d43 s27 s28)))) := (subst_tshift0_Tm_comm_ind s28 H0).
    Definition substTy_tshiftTy0_comm (S28 : Ty) : (forall (d43 : (Trace TmVar)) (s27 : Tm) ,
      ((substTy (XS TyVar d43) s27 (tshiftTy C0 S28)) =
      (tshiftTy C0 (substTy d43 s27 S28)))) := (subst_tshift0_Ty_comm_ind S28 H0).
    Definition tsubstTm_shiftTm0_comm (s27 : Tm) : (forall (d43 : (Trace TyVar)) (S28 : Ty) ,
      ((tsubstTm (XS TmVar d43) S28 (shiftTm C0 s27)) =
      (shiftTm C0 (tsubstTm d43 S28 s27)))) := (tsubst_shift0_Tm_comm_ind s27 H0).
    Definition tsubstTy_shiftTy0_comm (S29 : Ty) : (forall (d43 : (Trace TyVar)) (S28 : Ty) ,
      ((tsubstTy (XS TmVar d43) S28 (shiftTy C0 S29)) =
      (shiftTy C0 (tsubstTy d43 S28 S29)))) := (tsubst_shift0_Ty_comm_ind S29 H0).
    Definition tsubstTm_tshiftTm0_comm (s27 : Tm) : (forall (d43 : (Trace TyVar)) (S28 : Ty) ,
      ((tsubstTm (XS TyVar d43) S28 (tshiftTm C0 s27)) =
      (tshiftTm C0 (tsubstTm d43 S28 s27)))) := (tsubst_tshift0_Tm_comm_ind s27 H0).
    Definition tsubstTy_tshiftTy0_comm (S29 : Ty) : (forall (d43 : (Trace TyVar)) (S28 : Ty) ,
      ((tsubstTy (XS TyVar d43) S28 (tshiftTy C0 S29)) =
      (tshiftTy C0 (tsubstTy d43 S28 S29)))) := (tsubst_tshift0_Ty_comm_ind S29 H0).
    Definition substKind_shiftKind0_comm (K13 : Kind) : (forall (d43 : (Trace TmVar)) (s27 : Tm) ,
      ((substKind (XS TmVar d43) s27 (shiftKind C0 K13)) =
      (shiftKind C0 (substKind d43 s27 K13)))) := (subst_shift0_Kind_comm_ind K13 H0).
    Definition substKind_tshiftKind0_comm (K13 : Kind) : (forall (d43 : (Trace TmVar)) (s27 : Tm) ,
      ((substKind (XS TyVar d43) s27 (tshiftKind C0 K13)) =
      (tshiftKind C0 (substKind d43 s27 K13)))) := (subst_tshift0_Kind_comm_ind K13 H0).
    Definition tsubstKind_shiftKind0_comm (K13 : Kind) : (forall (d43 : (Trace TyVar)) (S28 : Ty) ,
      ((tsubstKind (XS TmVar d43) S28 (shiftKind C0 K13)) =
      (shiftKind C0 (tsubstKind d43 S28 K13)))) := (tsubst_shift0_Kind_comm_ind K13 H0).
    Definition tsubstKind_tshiftKind0_comm (K13 : Kind) : (forall (d43 : (Trace TyVar)) (S28 : Ty) ,
      ((tsubstKind (XS TyVar d43) S28 (tshiftKind C0 K13)) =
      (tshiftKind C0 (tsubstKind d43 S28 K13)))) := (tsubst_tshift0_Kind_comm_ind K13 H0).
  End SubstShiftComm.
  Section SubstSubordInd.
    
  End SubstSubordInd.
  Section SubstSubord.
    
  End SubstSubord.
End ShiftSubstInteraction.
 Hint Rewrite substKind0_shiftKind0_reflection tsubstKind0_tshiftKind0_reflection substTm0_shiftTm0_reflection tsubstTm0_tshiftTm0_reflection substTy0_shiftTy0_reflection tsubstTy0_tshiftTy0_reflection : interaction.
 Hint Rewrite substKind0_shiftKind0_reflection tsubstKind0_tshiftKind0_reflection substTm0_shiftTm0_reflection tsubstTm0_tshiftTm0_reflection substTy0_shiftTy0_reflection tsubstTy0_tshiftTy0_reflection : reflection.
 Hint Rewrite substKind_shiftKind0_comm substKind_tshiftKind0_comm tsubstKind_shiftKind0_comm tsubstKind_tshiftKind0_comm substTm_shiftTm0_comm substTm_tshiftTm0_comm tsubstTm_shiftTm0_comm tsubstTm_tshiftTm0_comm substTy_shiftTy0_comm substTy_tshiftTy0_comm tsubstTy_shiftTy0_comm tsubstTy_tshiftTy0_comm : interaction.
 Hint Rewrite substKind_shiftKind0_comm substKind_tshiftKind0_comm tsubstKind_shiftKind0_comm tsubstKind_tshiftKind0_comm substTm_shiftTm0_comm substTm_tshiftTm0_comm tsubstTm_shiftTm0_comm tsubstTm_tshiftTm0_comm substTy_shiftTy0_comm substTy_tshiftTy0_comm tsubstTy_shiftTy0_comm tsubstTy_tshiftTy0_comm : subst_shift0.
 Hint Rewrite shiftKind_substKind0_comm shiftKind_tsubstKind0_comm tshiftKind_substKind0_comm tshiftKind_tsubstKind0_comm shiftTm_substTm0_comm shiftTm_tsubstTm0_comm tshiftTm_substTm0_comm tshiftTm_tsubstTm0_comm shiftTy_substTy0_comm shiftTy_tsubstTy0_comm tshiftTy_substTy0_comm tshiftTy_tsubstTy0_comm : interaction.
 Hint Rewrite shiftKind_substKind0_comm shiftKind_tsubstKind0_comm tshiftKind_substKind0_comm tshiftKind_tsubstKind0_comm shiftTm_substTm0_comm shiftTm_tsubstTm0_comm tshiftTm_substTm0_comm tshiftTm_tsubstTm0_comm shiftTy_substTy0_comm shiftTy_tsubstTy0_comm tshiftTy_substTy0_comm tshiftTy_tsubstTy0_comm : shift_subst0.
Section SubstSubstInteraction.
  Section SubstSubstIndexCommInd.
    Lemma substTm_substIndex0_commright_ind (d43 : (Trace TmVar)) (s27 : Tm) (s28 : Tm) :
      (forall (k63 : Hvl) (x83 : (Index TmVar)) ,
        ((substTm (weakenTrace d43 k63) s27 (substIndex (weakenTrace X0 k63) s28 x83)) =
        (substTm (weakenTrace X0 k63) (substTm d43 s27 s28) (substIndex (weakenTrace (XS TmVar d43) k63) s27 x83)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma tsubstTm_substIndex0_commright_ind (d43 : (Trace TyVar)) (S28 : Ty) (s27 : Tm) :
      (forall (k63 : Hvl) (x83 : (Index TmVar)) ,
        ((tsubstTm (weakenTrace d43 k63) S28 (substIndex (weakenTrace X0 k63) s27 x83)) =
        (substIndex (weakenTrace X0 k63) (tsubstTm d43 S28 s27) x83))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma substTy_tsubstIndex0_commright_ind (d43 : (Trace TmVar)) (s27 : Tm) (S28 : Ty) :
      (forall (k63 : Hvl) (X27 : (Index TyVar)) ,
        ((substTy (weakenTrace d43 k63) s27 (tsubstIndex (weakenTrace X0 k63) S28 X27)) =
        (tsubstIndex (weakenTrace X0 k63) (substTy d43 s27 S28) X27))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma tsubstTy_tsubstIndex0_commright_ind (d43 : (Trace TyVar)) (S28 : Ty) (S29 : Ty) :
      (forall (k63 : Hvl) (X27 : (Index TyVar)) ,
        ((tsubstTy (weakenTrace d43 k63) S28 (tsubstIndex (weakenTrace X0 k63) S29 X27)) =
        (tsubstTy (weakenTrace X0 k63) (tsubstTy d43 S28 S29) (tsubstIndex (weakenTrace (XS TyVar d43) k63) S28 X27)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma substTy_tsubstIndex0_commleft_ind (d43 : (Trace TmVar)) (s27 : Tm) (S28 : Ty) :
      (forall (k63 : Hvl) (x83 : (Index TmVar)) ,
        ((substIndex (weakenTrace d43 k63) s27 x83) =
        (tsubstTm (weakenTrace X0 k63) (substTy d43 s27 S28) (substIndex (weakenTrace (XS TyVar d43) k63) s27 x83)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma tsubstTm_substIndex0_commleft_ind (d43 : (Trace TyVar)) (S28 : Ty) (s27 : Tm) :
      (forall (k63 : Hvl) (X27 : (Index TyVar)) ,
        ((tsubstIndex (weakenTrace d43 k63) S28 X27) =
        (substTy (weakenTrace X0 k63) (tsubstTm d43 S28 s27) (tsubstIndex (weakenTrace (XS TmVar d43) k63) S28 X27)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
  End SubstSubstIndexCommInd.
  Section SubstSubstCommInd.
    Definition subst_subst0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s31 : Tm) =>
      (forall (k67 : Hvl) (d47 : (Trace TmVar)) (s32 : Tm) (s33 : Tm) ,
        ((substTm (weakenTrace d47 k67) s32 (substTm (weakenTrace X0 k67) s33 s31)) =
        (substTm (weakenTrace X0 k67) (substTm d47 s32 s33) (substTm (weakenTrace (XS TmVar d47) k67) s32 s31))))) (fun (S32 : Ty) =>
      (forall (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) ,
        ((substTy (weakenTrace d47 k67) s31 (substTy (weakenTrace X0 k67) s32 S32)) =
        (substTy (weakenTrace X0 k67) (substTm d47 s31 s32) (substTy (weakenTrace (XS TmVar d47) k67) s31 S32))))) (fun (x95 : (Index TmVar)) =>
      (fun (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) =>
        (substTm_substIndex0_commright_ind d47 s31 s32 k67 x95))) (fun (T : Ty) IHT95 (t : Tm) IHt79 (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) =>
      (f_equal2 Abs (IHT95 k67 d47 s31 s32) (eq_trans (f_equal3 substTm (weakenTrace_append TmVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt79 (HS TmVar k67) d47 s31 s32) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TmVar d47) k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt79 (t2 : Tm) IHt80 (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) =>
      (f_equal2 App (IHt79 k67 d47 s31 s32) (IHt80 k67 d47 s31 s32))) (fun (T : Ty) IHT95 (t : Tm) IHt79 (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) =>
      (f_equal2 All (IHT95 k67 d47 s31 s32) (eq_trans (f_equal3 substTm (weakenTrace_append TmVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt79 (HS TmVar k67) d47 s31 s32) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TmVar d47) k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (X31 : (Index TyVar)) =>
      (fun (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) =>
        eq_refl)) (fun (T1 : Ty) IHT95 (T2 : Ty) IHT96 (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) =>
      (f_equal2 TPi (IHT95 k67 d47 s31 s32) (eq_trans (f_equal3 substTy (weakenTrace_append TmVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 substTy (weakenTrace_append TmVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT96 (HS TmVar k67) d47 s31 s32) (f_equal3 substTy (eq_sym (weakenTrace_append TmVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 substTy (eq_sym (weakenTrace_append TmVar (XS TmVar d47) k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT95 (t : Tm) IHt79 (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) =>
      (f_equal2 TApp (IHT95 k67 d47 s31 s32) (IHt79 k67 d47 s31 s32))) (fun (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) =>
      eq_refl) (fun (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) =>
      eq_refl)).
    Lemma subst_subst0_Tm_comm_ind  :
      (forall (s31 : Tm) ,
        (forall (k67 : Hvl) (d47 : (Trace TmVar)) (s32 : Tm) (s33 : Tm) ,
          ((substTm (weakenTrace d47 k67) s32 (substTm (weakenTrace X0 k67) s33 s31)) =
          (substTm (weakenTrace X0 k67) (substTm d47 s32 s33) (substTm (weakenTrace (XS TmVar d47) k67) s32 s31))))).
    Proof.
      pose proof (subst_subst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_subst0_Ty_comm_ind  :
      (forall (S32 : Ty) ,
        (forall (k67 : Hvl) (d47 : (Trace TmVar)) (s32 : Tm) (s33 : Tm) ,
          ((substTy (weakenTrace d47 k67) s32 (substTy (weakenTrace X0 k67) s33 S32)) =
          (substTy (weakenTrace X0 k67) (substTm d47 s32 s33) (substTy (weakenTrace (XS TmVar d47) k67) s32 S32))))).
    Proof.
      pose proof (subst_subst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_tsubst0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s31 : Tm) =>
      (forall (k67 : Hvl) (d47 : (Trace TmVar)) (s32 : Tm) (S32 : Ty) ,
        ((substTm (weakenTrace d47 k67) s32 (tsubstTm (weakenTrace X0 k67) S32 s31)) =
        (tsubstTm (weakenTrace X0 k67) (substTy d47 s32 S32) (substTm (weakenTrace (XS TyVar d47) k67) s32 s31))))) (fun (S32 : Ty) =>
      (forall (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (S33 : Ty) ,
        ((substTy (weakenTrace d47 k67) s31 (tsubstTy (weakenTrace X0 k67) S33 S32)) =
        (tsubstTy (weakenTrace X0 k67) (substTy d47 s31 S33) (substTy (weakenTrace (XS TyVar d47) k67) s31 S32))))) (fun (x95 : (Index TmVar)) =>
      (fun (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) =>
        (substTy_tsubstIndex0_commleft_ind d47 s31 S32 k67 x95))) (fun (T : Ty) IHT95 (t : Tm) IHt79 (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) =>
      (f_equal2 Abs (IHT95 k67 d47 s31 S32) (eq_trans (f_equal3 substTm (weakenTrace_append TmVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt79 (HS TmVar k67) d47 s31 S32) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TyVar d47) k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt79 (t2 : Tm) IHt80 (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) =>
      (f_equal2 App (IHt79 k67 d47 s31 S32) (IHt80 k67 d47 s31 S32))) (fun (T : Ty) IHT95 (t : Tm) IHt79 (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) =>
      (f_equal2 All (IHT95 k67 d47 s31 S32) (eq_trans (f_equal3 substTm (weakenTrace_append TmVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt79 (HS TmVar k67) d47 s31 S32) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TyVar d47) k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (X31 : (Index TyVar)) =>
      (fun (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) =>
        (substTy_tsubstIndex0_commright_ind d47 s31 S32 k67 X31))) (fun (T1 : Ty) IHT95 (T2 : Ty) IHT96 (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) =>
      (f_equal2 TPi (IHT95 k67 d47 s31 S32) (eq_trans (f_equal3 substTy (weakenTrace_append TmVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 tsubstTy (weakenTrace_append TyVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT96 (HS TmVar k67) d47 s31 S32) (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 substTy (eq_sym (weakenTrace_append TmVar (XS TyVar d47) k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT95 (t : Tm) IHt79 (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) =>
      (f_equal2 TApp (IHT95 k67 d47 s31 S32) (IHt79 k67 d47 s31 S32))) (fun (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) =>
      eq_refl) (fun (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) =>
      eq_refl)).
    Lemma subst_tsubst0_Tm_comm_ind  :
      (forall (s31 : Tm) ,
        (forall (k67 : Hvl) (d47 : (Trace TmVar)) (s32 : Tm) (S32 : Ty) ,
          ((substTm (weakenTrace d47 k67) s32 (tsubstTm (weakenTrace X0 k67) S32 s31)) =
          (tsubstTm (weakenTrace X0 k67) (substTy d47 s32 S32) (substTm (weakenTrace (XS TyVar d47) k67) s32 s31))))).
    Proof.
      pose proof (subst_tsubst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_tsubst0_Ty_comm_ind  :
      (forall (S32 : Ty) ,
        (forall (k67 : Hvl) (d47 : (Trace TmVar)) (s32 : Tm) (S33 : Ty) ,
          ((substTy (weakenTrace d47 k67) s32 (tsubstTy (weakenTrace X0 k67) S33 S32)) =
          (tsubstTy (weakenTrace X0 k67) (substTy d47 s32 S33) (substTy (weakenTrace (XS TyVar d47) k67) s32 S32))))).
    Proof.
      pose proof (subst_tsubst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition tsubst_subst0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s31 : Tm) =>
      (forall (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (s32 : Tm) ,
        ((tsubstTm (weakenTrace d47 k67) S32 (substTm (weakenTrace X0 k67) s32 s31)) =
        (substTm (weakenTrace X0 k67) (tsubstTm d47 S32 s32) (tsubstTm (weakenTrace (XS TmVar d47) k67) S32 s31))))) (fun (S32 : Ty) =>
      (forall (k67 : Hvl) (d47 : (Trace TyVar)) (S33 : Ty) (s31 : Tm) ,
        ((tsubstTy (weakenTrace d47 k67) S33 (substTy (weakenTrace X0 k67) s31 S32)) =
        (substTy (weakenTrace X0 k67) (tsubstTm d47 S33 s31) (tsubstTy (weakenTrace (XS TmVar d47) k67) S33 S32))))) (fun (x95 : (Index TmVar)) =>
      (fun (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) =>
        (tsubstTm_substIndex0_commright_ind d47 S32 s31 k67 x95))) (fun (T : Ty) IHT95 (t : Tm) IHt79 (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) =>
      (f_equal2 Abs (IHT95 k67 d47 S32 s31) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt79 (HS TmVar k67) d47 S32 s31) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar (XS TmVar d47) k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt79 (t2 : Tm) IHt80 (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) =>
      (f_equal2 App (IHt79 k67 d47 S32 s31) (IHt80 k67 d47 S32 s31))) (fun (T : Ty) IHT95 (t : Tm) IHt79 (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) =>
      (f_equal2 All (IHT95 k67 d47 S32 s31) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt79 (HS TmVar k67) d47 S32 s31) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar (XS TmVar d47) k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (X31 : (Index TyVar)) =>
      (fun (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) =>
        (tsubstTm_substIndex0_commleft_ind d47 S32 s31 k67 X31))) (fun (T1 : Ty) IHT95 (T2 : Ty) IHT96 (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) =>
      (f_equal2 TPi (IHT95 k67 d47 S32 s31) (eq_trans (f_equal3 tsubstTy (weakenTrace_append TyVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 substTy (weakenTrace_append TmVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT96 (HS TmVar k67) d47 S32 s31) (f_equal3 substTy (eq_sym (weakenTrace_append TmVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar (XS TmVar d47) k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT95 (t : Tm) IHt79 (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) =>
      (f_equal2 TApp (IHT95 k67 d47 S32 s31) (IHt79 k67 d47 S32 s31))) (fun (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) =>
      eq_refl) (fun (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) =>
      eq_refl)).
    Lemma tsubst_subst0_Tm_comm_ind  :
      (forall (s31 : Tm) ,
        (forall (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (s32 : Tm) ,
          ((tsubstTm (weakenTrace d47 k67) S32 (substTm (weakenTrace X0 k67) s32 s31)) =
          (substTm (weakenTrace X0 k67) (tsubstTm d47 S32 s32) (tsubstTm (weakenTrace (XS TmVar d47) k67) S32 s31))))).
    Proof.
      pose proof (tsubst_subst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma tsubst_subst0_Ty_comm_ind  :
      (forall (S32 : Ty) ,
        (forall (k67 : Hvl) (d47 : (Trace TyVar)) (S33 : Ty) (s32 : Tm) ,
          ((tsubstTy (weakenTrace d47 k67) S33 (substTy (weakenTrace X0 k67) s32 S32)) =
          (substTy (weakenTrace X0 k67) (tsubstTm d47 S33 s32) (tsubstTy (weakenTrace (XS TmVar d47) k67) S33 S32))))).
    Proof.
      pose proof (tsubst_subst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition tsubst_tsubst0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s31 : Tm) =>
      (forall (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) ,
        ((tsubstTm (weakenTrace d47 k67) S32 (tsubstTm (weakenTrace X0 k67) S33 s31)) =
        (tsubstTm (weakenTrace X0 k67) (tsubstTy d47 S32 S33) (tsubstTm (weakenTrace (XS TyVar d47) k67) S32 s31))))) (fun (S32 : Ty) =>
      (forall (k67 : Hvl) (d47 : (Trace TyVar)) (S33 : Ty) (S34 : Ty) ,
        ((tsubstTy (weakenTrace d47 k67) S33 (tsubstTy (weakenTrace X0 k67) S34 S32)) =
        (tsubstTy (weakenTrace X0 k67) (tsubstTy d47 S33 S34) (tsubstTy (weakenTrace (XS TyVar d47) k67) S33 S32))))) (fun (x95 : (Index TmVar)) =>
      (fun (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) =>
        eq_refl)) (fun (T : Ty) IHT95 (t : Tm) IHt79 (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) =>
      (f_equal2 Abs (IHT95 k67 d47 S32 S33) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt79 (HS TmVar k67) d47 S32 S33) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar (XS TyVar d47) k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt79 (t2 : Tm) IHt80 (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) =>
      (f_equal2 App (IHt79 k67 d47 S32 S33) (IHt80 k67 d47 S32 S33))) (fun (T : Ty) IHT95 (t : Tm) IHt79 (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) =>
      (f_equal2 All (IHT95 k67 d47 S32 S33) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt79 (HS TmVar k67) d47 S32 S33) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar (XS TyVar d47) k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (X31 : (Index TyVar)) =>
      (fun (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) =>
        (tsubstTy_tsubstIndex0_commright_ind d47 S32 S33 k67 X31))) (fun (T1 : Ty) IHT95 (T2 : Ty) IHT96 (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) =>
      (f_equal2 TPi (IHT95 k67 d47 S32 S33) (eq_trans (f_equal3 tsubstTy (weakenTrace_append TyVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 tsubstTy (weakenTrace_append TyVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT96 (HS TmVar k67) d47 S32 S33) (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar (XS TyVar d47) k67 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT95 (t : Tm) IHt79 (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) =>
      (f_equal2 TApp (IHT95 k67 d47 S32 S33) (IHt79 k67 d47 S32 S33))) (fun (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) =>
      eq_refl) (fun (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) =>
      eq_refl)).
    Lemma tsubst_tsubst0_Tm_comm_ind  :
      (forall (s31 : Tm) ,
        (forall (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) ,
          ((tsubstTm (weakenTrace d47 k67) S32 (tsubstTm (weakenTrace X0 k67) S33 s31)) =
          (tsubstTm (weakenTrace X0 k67) (tsubstTy d47 S32 S33) (tsubstTm (weakenTrace (XS TyVar d47) k67) S32 s31))))).
    Proof.
      pose proof (tsubst_tsubst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma tsubst_tsubst0_Ty_comm_ind  :
      (forall (S32 : Ty) ,
        (forall (k67 : Hvl) (d47 : (Trace TyVar)) (S33 : Ty) (S34 : Ty) ,
          ((tsubstTy (weakenTrace d47 k67) S33 (tsubstTy (weakenTrace X0 k67) S34 S32)) =
          (tsubstTy (weakenTrace X0 k67) (tsubstTy d47 S33 S34) (tsubstTy (weakenTrace (XS TyVar d47) k67) S33 S32))))).
    Proof.
      pose proof (tsubst_tsubst0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_subst0_Kind_comm_ind := (ind_Kind (fun (K15 : Kind) =>
      (forall (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) ,
        ((substKind (weakenTrace d47 k67) s31 (substKind (weakenTrace X0 k67) s32 K15)) =
        (substKind (weakenTrace X0 k67) (substTm d47 s31 s32) (substKind (weakenTrace (XS TmVar d47) k67) s31 K15))))) (fun (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK15 (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) =>
      (f_equal2 KPi (let IHT95 := (subst_subst0_Ty_comm_ind T) in
      (IHT95 k67 d47 s31 s32)) (eq_trans (f_equal3 substKind (weakenTrace_append TmVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 substKind (weakenTrace_append TmVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK15 (HS TmVar k67) d47 s31 s32) (f_equal3 substKind (eq_sym (weakenTrace_append TmVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 substKind (eq_sym (weakenTrace_append TmVar (XS TmVar d47) k67 (HS TmVar H0))) eq_refl eq_refl))))))).
    Definition subst_tsubst0_Kind_comm_ind := (ind_Kind (fun (K15 : Kind) =>
      (forall (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) ,
        ((substKind (weakenTrace d47 k67) s31 (tsubstKind (weakenTrace X0 k67) S32 K15)) =
        (tsubstKind (weakenTrace X0 k67) (substTy d47 s31 S32) (substKind (weakenTrace (XS TyVar d47) k67) s31 K15))))) (fun (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK15 (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) =>
      (f_equal2 KPi (let IHT95 := (subst_tsubst0_Ty_comm_ind T) in
      (IHT95 k67 d47 s31 S32)) (eq_trans (f_equal3 substKind (weakenTrace_append TmVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 tsubstKind (weakenTrace_append TyVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK15 (HS TmVar k67) d47 s31 S32) (f_equal3 tsubstKind (eq_sym (weakenTrace_append TyVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 substKind (eq_sym (weakenTrace_append TmVar (XS TyVar d47) k67 (HS TmVar H0))) eq_refl eq_refl))))))).
    Definition tsubst_subst0_Kind_comm_ind := (ind_Kind (fun (K15 : Kind) =>
      (forall (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) ,
        ((tsubstKind (weakenTrace d47 k67) S32 (substKind (weakenTrace X0 k67) s31 K15)) =
        (substKind (weakenTrace X0 k67) (tsubstTm d47 S32 s31) (tsubstKind (weakenTrace (XS TmVar d47) k67) S32 K15))))) (fun (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK15 (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) =>
      (f_equal2 KPi (let IHT95 := (tsubst_subst0_Ty_comm_ind T) in
      (IHT95 k67 d47 S32 s31)) (eq_trans (f_equal3 tsubstKind (weakenTrace_append TyVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 substKind (weakenTrace_append TmVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK15 (HS TmVar k67) d47 S32 s31) (f_equal3 substKind (eq_sym (weakenTrace_append TmVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 tsubstKind (eq_sym (weakenTrace_append TyVar (XS TmVar d47) k67 (HS TmVar H0))) eq_refl eq_refl))))))).
    Definition tsubst_tsubst0_Kind_comm_ind := (ind_Kind (fun (K15 : Kind) =>
      (forall (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) ,
        ((tsubstKind (weakenTrace d47 k67) S32 (tsubstKind (weakenTrace X0 k67) S33 K15)) =
        (tsubstKind (weakenTrace X0 k67) (tsubstTy d47 S32 S33) (tsubstKind (weakenTrace (XS TyVar d47) k67) S32 K15))))) (fun (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK15 (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) =>
      (f_equal2 KPi (let IHT95 := (tsubst_tsubst0_Ty_comm_ind T) in
      (IHT95 k67 d47 S32 S33)) (eq_trans (f_equal3 tsubstKind (weakenTrace_append TyVar d47 k67 (HS TmVar H0)) eq_refl (f_equal3 tsubstKind (weakenTrace_append TyVar X0 k67 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK15 (HS TmVar k67) d47 S32 S33) (f_equal3 tsubstKind (eq_sym (weakenTrace_append TyVar X0 k67 (HS TmVar H0))) eq_refl (f_equal3 tsubstKind (eq_sym (weakenTrace_append TyVar (XS TyVar d47) k67 (HS TmVar H0))) eq_refl eq_refl))))))).
  End SubstSubstCommInd.
  Section SubstSubstComm.
    Definition substTm_substTm0_comm (s33 : Tm) : (forall (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) ,
      ((substTm d47 s31 (substTm X0 s32 s33)) =
      (substTm X0 (substTm d47 s31 s32) (substTm (XS TmVar d47) s31 s33)))) := (subst_subst0_Tm_comm_ind s33 H0).
    Definition substTy_substTy0_comm (S32 : Ty) : (forall (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) ,
      ((substTy d47 s31 (substTy X0 s32 S32)) =
      (substTy X0 (substTm d47 s31 s32) (substTy (XS TmVar d47) s31 S32)))) := (subst_subst0_Ty_comm_ind S32 H0).
    Definition substTm_tsubstTm0_comm (s32 : Tm) : (forall (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) ,
      ((substTm d47 s31 (tsubstTm X0 S32 s32)) =
      (tsubstTm X0 (substTy d47 s31 S32) (substTm (XS TyVar d47) s31 s32)))) := (subst_tsubst0_Tm_comm_ind s32 H0).
    Definition substTy_tsubstTy0_comm (S33 : Ty) : (forall (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) ,
      ((substTy d47 s31 (tsubstTy X0 S32 S33)) =
      (tsubstTy X0 (substTy d47 s31 S32) (substTy (XS TyVar d47) s31 S33)))) := (subst_tsubst0_Ty_comm_ind S33 H0).
    Definition tsubstTm_substTm0_comm (s32 : Tm) : (forall (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) ,
      ((tsubstTm d47 S32 (substTm X0 s31 s32)) =
      (substTm X0 (tsubstTm d47 S32 s31) (tsubstTm (XS TmVar d47) S32 s32)))) := (tsubst_subst0_Tm_comm_ind s32 H0).
    Definition tsubstTy_substTy0_comm (S33 : Ty) : (forall (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) ,
      ((tsubstTy d47 S32 (substTy X0 s31 S33)) =
      (substTy X0 (tsubstTm d47 S32 s31) (tsubstTy (XS TmVar d47) S32 S33)))) := (tsubst_subst0_Ty_comm_ind S33 H0).
    Definition tsubstTm_tsubstTm0_comm (s31 : Tm) : (forall (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) ,
      ((tsubstTm d47 S32 (tsubstTm X0 S33 s31)) =
      (tsubstTm X0 (tsubstTy d47 S32 S33) (tsubstTm (XS TyVar d47) S32 s31)))) := (tsubst_tsubst0_Tm_comm_ind s31 H0).
    Definition tsubstTy_tsubstTy0_comm (S34 : Ty) : (forall (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) ,
      ((tsubstTy d47 S32 (tsubstTy X0 S33 S34)) =
      (tsubstTy X0 (tsubstTy d47 S32 S33) (tsubstTy (XS TyVar d47) S32 S34)))) := (tsubst_tsubst0_Ty_comm_ind S34 H0).
    Definition substKind_substKind0_comm (K15 : Kind) : (forall (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) ,
      ((substKind d47 s31 (substKind X0 s32 K15)) =
      (substKind X0 (substTm d47 s31 s32) (substKind (XS TmVar d47) s31 K15)))) := (subst_subst0_Kind_comm_ind K15 H0).
    Definition substKind_tsubstKind0_comm (K15 : Kind) : (forall (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) ,
      ((substKind d47 s31 (tsubstKind X0 S32 K15)) =
      (tsubstKind X0 (substTy d47 s31 S32) (substKind (XS TyVar d47) s31 K15)))) := (subst_tsubst0_Kind_comm_ind K15 H0).
    Definition tsubstKind_substKind0_comm (K15 : Kind) : (forall (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) ,
      ((tsubstKind d47 S32 (substKind X0 s31 K15)) =
      (substKind X0 (tsubstTm d47 S32 s31) (tsubstKind (XS TmVar d47) S32 K15)))) := (tsubst_subst0_Kind_comm_ind K15 H0).
    Definition tsubstKind_tsubstKind0_comm (K15 : Kind) : (forall (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) ,
      ((tsubstKind d47 S32 (tsubstKind X0 S33 K15)) =
      (tsubstKind X0 (tsubstTy d47 S32 S33) (tsubstKind (XS TyVar d47) S32 K15)))) := (tsubst_tsubst0_Kind_comm_ind K15 H0).
  End SubstSubstComm.
  Section WeakenSubst.
    Lemma weakenTm_substTm  :
      (forall (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (s32 : Tm) ,
        ((weakenTm (substTm d47 s31 s32) k67) =
        (substTm (weakenTrace d47 k67) s31 (weakenTm s32 k67)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTy_substTy  :
      (forall (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (S32 : Ty) ,
        ((weakenTy (substTy d47 s31 S32) k67) =
        (substTy (weakenTrace d47 k67) s31 (weakenTy S32 k67)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_tsubstTm  :
      (forall (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (s31 : Tm) ,
        ((weakenTm (tsubstTm d47 S32 s31) k67) =
        (tsubstTm (weakenTrace d47 k67) S32 (weakenTm s31 k67)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTy_tsubstTy  :
      (forall (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (S33 : Ty) ,
        ((weakenTy (tsubstTy d47 S32 S33) k67) =
        (tsubstTy (weakenTrace d47 k67) S32 (weakenTy S33 k67)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenKind_substKind  :
      (forall (k67 : Hvl) (d47 : (Trace TmVar)) (s31 : Tm) (K15 : Kind) ,
        ((weakenKind (substKind d47 s31 K15) k67) =
        (substKind (weakenTrace d47 k67) s31 (weakenKind K15 k67)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenKind_tsubstKind  :
      (forall (k67 : Hvl) (d47 : (Trace TyVar)) (S32 : Ty) (K15 : Kind) ,
        ((weakenKind (tsubstKind d47 S32 K15) k67) =
        (tsubstKind (weakenTrace d47 k67) S32 (weakenKind K15 k67)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
  End WeakenSubst.
  Section WeakenAppend.
    Lemma weakenTm_append  :
      (forall (s31 : Tm) (k67 : Hvl) (k68 : Hvl) ,
        ((weakenTm (weakenTm s31 k67) k68) =
        (weakenTm s31 (appendHvl k67 k68)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenTy_append  :
      (forall (S32 : Ty) (k67 : Hvl) (k68 : Hvl) ,
        ((weakenTy (weakenTy S32 k67) k68) =
        (weakenTy S32 (appendHvl k67 k68)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenKind_append  :
      (forall (K15 : Kind) (k67 : Hvl) (k68 : Hvl) ,
        ((weakenKind (weakenKind K15 k67) k68) =
        (weakenKind K15 (appendHvl k67 k68)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
  End WeakenAppend.
End SubstSubstInteraction.
 Hint Rewrite substKind_substKind0_comm tsubstKind_tsubstKind0_comm substTm_substTm0_comm tsubstTm_tsubstTm0_comm substTy_substTy0_comm tsubstTy_tsubstTy0_comm : interaction.
 Hint Rewrite substKind_substKind0_comm tsubstKind_tsubstKind0_comm substTm_substTm0_comm tsubstTm_tsubstTm0_comm substTy_substTy0_comm tsubstTy_tsubstTy0_comm : subst_subst0.
 Hint Rewrite weakenKind_shiftKind weakenKind_tshiftKind weakenTm_shiftTm weakenTm_tshiftTm weakenTy_shiftTy weakenTy_tshiftTy : weaken_shift.
 Hint Rewrite weakenKind_substKind weakenKind_tsubstKind weakenTm_substTm weakenTm_tsubstTm weakenTy_substTy weakenTy_tsubstTy : weaken_subst.
Section WellFormed.
  Fixpoint wfindex {a : Namespace} (k67 : Hvl) (i : (Index a)) {struct k67} : Prop :=
    match k67 with
      | (H0) => False
      | (HS b k67) => match (eq_namespace_dec a b) with
        | (left e) => match i with
          | (I0) => True
          | (IS i) => (wfindex k67 i)
        end
        | (right n) => (wfindex k67 i)
      end
    end.
  Inductive wfTm (k67 : Hvl) : Tm -> Prop :=
    | wfTm_Var
        {x107 : (Index TmVar)}
        (wfi : (wfindex k67 x107)) :
        (wfTm k67 (Var x107))
    | wfTm_Abs {T107 : Ty}
        (wf : (wfTy k67 T107))
        {t89 : Tm}
        (wf0 : (wfTm (HS TmVar k67) t89))
        : (wfTm k67 (Abs T107 t89))
    | wfTm_App {t90 : Tm}
        (wf : (wfTm k67 t90)) {t91 : Tm}
        (wf0 : (wfTm k67 t91)) :
        (wfTm k67 (App t90 t91))
    | wfTm_All {T108 : Ty}
        (wf : (wfTy k67 T108))
        {t92 : Tm}
        (wf0 : (wfTm (HS TmVar k67) t92))
        : (wfTm k67 (All T108 t92))
  with wfTy (k67 : Hvl) : Ty -> Prop :=
    | wfTy_TVar
        {X35 : (Index TyVar)}
        (wfi : (wfindex k67 X35)) :
        (wfTy k67 (TVar X35))
    | wfTy_TPi {T107 : Ty}
        (wf : (wfTy k67 T107))
        {T108 : Ty}
        (wf0 : (wfTy (HS TmVar k67) T108))
        : (wfTy k67 (TPi T107 T108))
    | wfTy_TApp {T109 : Ty}
        (wf : (wfTy k67 T109))
        {t89 : Tm}
        (wf0 : (wfTm k67 t89)) :
        (wfTy k67 (TApp T109 t89))
    | wfTy_Prp : (wfTy k67 (Prp))
    | wfTy_Prf : (wfTy k67 (Prf)).
  Inductive wfKind (k67 : Hvl) : Kind -> Prop :=
    | wfKind_Star :
        (wfKind k67 (Star))
    | wfKind_KPi {T107 : Ty}
        (wf : (wfTy k67 T107))
        {K17 : Kind}
        (wf0 : (wfKind (HS TmVar k67) K17))
        : (wfKind k67 (KPi T107 K17)).
  Scheme ind_wfTm := Induction for wfTm Sort Prop
  with ind_wfTy := Induction for wfTy Sort Prop.
  Combined Scheme ind_wfTm_Ty from ind_wfTm, ind_wfTy.
  Scheme ind_wfKind := Induction for wfKind Sort Prop.
End WellFormed.
Section ShiftWellFormed.
  Inductive shifthvl_TmVar : (forall (c19 : (Cutoff TmVar)) (k67 : Hvl) (k68 : Hvl) ,
    Prop) :=
    | shifthvl_TmVar_here
        (k67 : Hvl) :
        (shifthvl_TmVar C0 k67 (HS TmVar k67))
    | shifthvl_TmVar_there_TmVar
        (c19 : (Cutoff TmVar))
        (k67 : Hvl) (k68 : Hvl) :
        (shifthvl_TmVar c19 k67 k68) -> (shifthvl_TmVar (CS c19) (HS TmVar k67) (HS TmVar k68))
    | shifthvl_TmVar_there_TyVar
        (c19 : (Cutoff TmVar))
        (k67 : Hvl) (k68 : Hvl) :
        (shifthvl_TmVar c19 k67 k68) -> (shifthvl_TmVar c19 (HS TyVar k67) (HS TyVar k68)).
  Inductive shifthvl_TyVar : (forall (c19 : (Cutoff TyVar)) (k67 : Hvl) (k68 : Hvl) ,
    Prop) :=
    | shifthvl_TyVar_here
        (k67 : Hvl) :
        (shifthvl_TyVar C0 k67 (HS TyVar k67))
    | shifthvl_TyVar_there_TmVar
        (c19 : (Cutoff TyVar))
        (k67 : Hvl) (k68 : Hvl) :
        (shifthvl_TyVar c19 k67 k68) -> (shifthvl_TyVar c19 (HS TmVar k67) (HS TmVar k68))
    | shifthvl_TyVar_there_TyVar
        (c19 : (Cutoff TyVar))
        (k67 : Hvl) (k68 : Hvl) :
        (shifthvl_TyVar c19 k67 k68) -> (shifthvl_TyVar (CS c19) (HS TyVar k67) (HS TyVar k68)).
  Lemma weaken_shifthvl_TmVar  :
    (forall (k67 : Hvl) {c19 : (Cutoff TmVar)} {k68 : Hvl} {k69 : Hvl} ,
      (shifthvl_TmVar c19 k68 k69) -> (shifthvl_TmVar (weakenCutoffTmVar c19 k67) (appendHvl k68 k67) (appendHvl k69 k67))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma weaken_shifthvl_TyVar  :
    (forall (k67 : Hvl) {c19 : (Cutoff TyVar)} {k68 : Hvl} {k69 : Hvl} ,
      (shifthvl_TyVar c19 k68 k69) -> (shifthvl_TyVar (weakenCutoffTyVar c19 k67) (appendHvl k68 k67) (appendHvl k69 k67))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma shift_wfindex_TmVar  :
    (forall (c19 : (Cutoff TmVar)) (k67 : Hvl) (k68 : Hvl) (ins : (shifthvl_TmVar c19 k67 k68)) (x107 : (Index TmVar)) ,
      (wfindex k67 x107) -> (wfindex k68 (shiftIndex c19 x107))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_wfindex_TyVar  :
    (forall (c19 : (Cutoff TmVar)) (k67 : Hvl) (k68 : Hvl) (ins : (shifthvl_TmVar c19 k67 k68)) (X35 : (Index TyVar)) ,
      (wfindex k67 X35) -> (wfindex k68 X35)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma tshift_wfindex_TmVar  :
    (forall (c19 : (Cutoff TyVar)) (k67 : Hvl) (k68 : Hvl) (ins : (shifthvl_TyVar c19 k67 k68)) (x107 : (Index TmVar)) ,
      (wfindex k67 x107) -> (wfindex k68 x107)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma tshift_wfindex_TyVar  :
    (forall (c19 : (Cutoff TyVar)) (k67 : Hvl) (k68 : Hvl) (ins : (shifthvl_TyVar c19 k67 k68)) (X35 : (Index TyVar)) ,
      (wfindex k67 X35) -> (wfindex k68 (tshiftIndex c19 X35))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Definition shift_wfTm_Ty : (forall (k67 : Hvl) ,
    (forall (s31 : Tm) (wf : (wfTm k67 s31)) ,
      (forall (c19 : (Cutoff TmVar)) (k68 : Hvl) ,
        (shifthvl_TmVar c19 k67 k68) -> (wfTm k68 (shiftTm c19 s31)))) /\
    (forall (S32 : Ty) (wf : (wfTy k67 S32)) ,
      (forall (c19 : (Cutoff TmVar)) (k68 : Hvl) ,
        (shifthvl_TmVar c19 k67 k68) -> (wfTy k68 (shiftTy c19 S32))))) := (ind_wfTm_Ty (fun (k67 : Hvl) (s31 : Tm) (wf : (wfTm k67 s31)) =>
    (forall (c19 : (Cutoff TmVar)) (k68 : Hvl) ,
      (shifthvl_TmVar c19 k67 k68) -> (wfTm k68 (shiftTm c19 s31)))) (fun (k67 : Hvl) (S32 : Ty) (wf : (wfTy k67 S32)) =>
    (forall (c19 : (Cutoff TmVar)) (k68 : Hvl) ,
      (shifthvl_TmVar c19 k67 k68) -> (wfTy k68 (shiftTy c19 S32)))) (fun (k67 : Hvl) (x107 : (Index TmVar)) (wfi : (wfindex k67 x107)) (c19 : (Cutoff TmVar)) (k68 : Hvl) (ins : (shifthvl_TmVar c19 k67 k68)) =>
    (wfTm_Var k68 (shift_wfindex_TmVar c19 k67 k68 ins x107 wfi))) (fun (k67 : Hvl) (T : Ty) (wf : (wfTy k67 T)) IHT107 (t : Tm) (wf0 : (wfTm (HS TmVar k67) t)) IHt89 (c19 : (Cutoff TmVar)) (k68 : Hvl) (ins : (shifthvl_TmVar c19 k67 k68)) =>
    (wfTm_Abs k68 (IHT107 c19 k68 (weaken_shifthvl_TmVar H0 ins)) (IHt89 (CS c19) (HS TmVar k68) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k67 : Hvl) (t1 : Tm) (wf : (wfTm k67 t1)) IHt89 (t2 : Tm) (wf0 : (wfTm k67 t2)) IHt90 (c19 : (Cutoff TmVar)) (k68 : Hvl) (ins : (shifthvl_TmVar c19 k67 k68)) =>
    (wfTm_App k68 (IHt89 c19 k68 (weaken_shifthvl_TmVar H0 ins)) (IHt90 c19 k68 (weaken_shifthvl_TmVar H0 ins)))) (fun (k67 : Hvl) (T : Ty) (wf : (wfTy k67 T)) IHT107 (t : Tm) (wf0 : (wfTm (HS TmVar k67) t)) IHt89 (c19 : (Cutoff TmVar)) (k68 : Hvl) (ins : (shifthvl_TmVar c19 k67 k68)) =>
    (wfTm_All k68 (IHT107 c19 k68 (weaken_shifthvl_TmVar H0 ins)) (IHt89 (CS c19) (HS TmVar k68) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k67 : Hvl) (X35 : (Index TyVar)) (wfi : (wfindex k67 X35)) (c19 : (Cutoff TmVar)) (k68 : Hvl) (ins : (shifthvl_TmVar c19 k67 k68)) =>
    (wfTy_TVar k68 (shift_wfindex_TyVar c19 k67 k68 ins X35 wfi))) (fun (k67 : Hvl) (T1 : Ty) (wf : (wfTy k67 T1)) IHT107 (T2 : Ty) (wf0 : (wfTy (HS TmVar k67) T2)) IHT108 (c19 : (Cutoff TmVar)) (k68 : Hvl) (ins : (shifthvl_TmVar c19 k67 k68)) =>
    (wfTy_TPi k68 (IHT107 c19 k68 (weaken_shifthvl_TmVar H0 ins)) (IHT108 (CS c19) (HS TmVar k68) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k67 : Hvl) (T : Ty) (wf : (wfTy k67 T)) IHT107 (t : Tm) (wf0 : (wfTm k67 t)) IHt89 (c19 : (Cutoff TmVar)) (k68 : Hvl) (ins : (shifthvl_TmVar c19 k67 k68)) =>
    (wfTy_TApp k68 (IHT107 c19 k68 (weaken_shifthvl_TmVar H0 ins)) (IHt89 c19 k68 (weaken_shifthvl_TmVar H0 ins)))) (fun (k67 : Hvl) (c19 : (Cutoff TmVar)) (k68 : Hvl) (ins : (shifthvl_TmVar c19 k67 k68)) =>
    (wfTy_Prp k68)) (fun (k67 : Hvl) (c19 : (Cutoff TmVar)) (k68 : Hvl) (ins : (shifthvl_TmVar c19 k67 k68)) =>
    (wfTy_Prf k68))).
  Lemma shift_wfTm (k67 : Hvl) :
    (forall (s31 : Tm) (wf : (wfTm k67 s31)) ,
      (forall (c19 : (Cutoff TmVar)) (k68 : Hvl) ,
        (shifthvl_TmVar c19 k67 k68) -> (wfTm k68 (shiftTm c19 s31)))).
  Proof.
    pose proof ((shift_wfTm_Ty k67)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_wfTy (k67 : Hvl) :
    (forall (S32 : Ty) (wf0 : (wfTy k67 S32)) ,
      (forall (c20 : (Cutoff TmVar)) (k69 : Hvl) ,
        (shifthvl_TmVar c20 k67 k69) -> (wfTy k69 (shiftTy c20 S32)))).
  Proof.
    pose proof ((shift_wfTm_Ty k67)).
    destruct_conjs .
    easy .
  Qed.
  Definition tshift_wfTm_Ty : (forall (k67 : Hvl) ,
    (forall (s31 : Tm) (wf : (wfTm k67 s31)) ,
      (forall (c19 : (Cutoff TyVar)) (k68 : Hvl) ,
        (shifthvl_TyVar c19 k67 k68) -> (wfTm k68 (tshiftTm c19 s31)))) /\
    (forall (S32 : Ty) (wf : (wfTy k67 S32)) ,
      (forall (c19 : (Cutoff TyVar)) (k68 : Hvl) ,
        (shifthvl_TyVar c19 k67 k68) -> (wfTy k68 (tshiftTy c19 S32))))) := (ind_wfTm_Ty (fun (k67 : Hvl) (s31 : Tm) (wf : (wfTm k67 s31)) =>
    (forall (c19 : (Cutoff TyVar)) (k68 : Hvl) ,
      (shifthvl_TyVar c19 k67 k68) -> (wfTm k68 (tshiftTm c19 s31)))) (fun (k67 : Hvl) (S32 : Ty) (wf : (wfTy k67 S32)) =>
    (forall (c19 : (Cutoff TyVar)) (k68 : Hvl) ,
      (shifthvl_TyVar c19 k67 k68) -> (wfTy k68 (tshiftTy c19 S32)))) (fun (k67 : Hvl) (x107 : (Index TmVar)) (wfi : (wfindex k67 x107)) (c19 : (Cutoff TyVar)) (k68 : Hvl) (ins : (shifthvl_TyVar c19 k67 k68)) =>
    (wfTm_Var k68 (tshift_wfindex_TmVar c19 k67 k68 ins x107 wfi))) (fun (k67 : Hvl) (T : Ty) (wf : (wfTy k67 T)) IHT107 (t : Tm) (wf0 : (wfTm (HS TmVar k67) t)) IHt89 (c19 : (Cutoff TyVar)) (k68 : Hvl) (ins : (shifthvl_TyVar c19 k67 k68)) =>
    (wfTm_Abs k68 (IHT107 c19 k68 (weaken_shifthvl_TyVar H0 ins)) (IHt89 c19 (HS TmVar k68) (weaken_shifthvl_TyVar (HS TmVar H0) ins)))) (fun (k67 : Hvl) (t1 : Tm) (wf : (wfTm k67 t1)) IHt89 (t2 : Tm) (wf0 : (wfTm k67 t2)) IHt90 (c19 : (Cutoff TyVar)) (k68 : Hvl) (ins : (shifthvl_TyVar c19 k67 k68)) =>
    (wfTm_App k68 (IHt89 c19 k68 (weaken_shifthvl_TyVar H0 ins)) (IHt90 c19 k68 (weaken_shifthvl_TyVar H0 ins)))) (fun (k67 : Hvl) (T : Ty) (wf : (wfTy k67 T)) IHT107 (t : Tm) (wf0 : (wfTm (HS TmVar k67) t)) IHt89 (c19 : (Cutoff TyVar)) (k68 : Hvl) (ins : (shifthvl_TyVar c19 k67 k68)) =>
    (wfTm_All k68 (IHT107 c19 k68 (weaken_shifthvl_TyVar H0 ins)) (IHt89 c19 (HS TmVar k68) (weaken_shifthvl_TyVar (HS TmVar H0) ins)))) (fun (k67 : Hvl) (X35 : (Index TyVar)) (wfi : (wfindex k67 X35)) (c19 : (Cutoff TyVar)) (k68 : Hvl) (ins : (shifthvl_TyVar c19 k67 k68)) =>
    (wfTy_TVar k68 (tshift_wfindex_TyVar c19 k67 k68 ins X35 wfi))) (fun (k67 : Hvl) (T1 : Ty) (wf : (wfTy k67 T1)) IHT107 (T2 : Ty) (wf0 : (wfTy (HS TmVar k67) T2)) IHT108 (c19 : (Cutoff TyVar)) (k68 : Hvl) (ins : (shifthvl_TyVar c19 k67 k68)) =>
    (wfTy_TPi k68 (IHT107 c19 k68 (weaken_shifthvl_TyVar H0 ins)) (IHT108 c19 (HS TmVar k68) (weaken_shifthvl_TyVar (HS TmVar H0) ins)))) (fun (k67 : Hvl) (T : Ty) (wf : (wfTy k67 T)) IHT107 (t : Tm) (wf0 : (wfTm k67 t)) IHt89 (c19 : (Cutoff TyVar)) (k68 : Hvl) (ins : (shifthvl_TyVar c19 k67 k68)) =>
    (wfTy_TApp k68 (IHT107 c19 k68 (weaken_shifthvl_TyVar H0 ins)) (IHt89 c19 k68 (weaken_shifthvl_TyVar H0 ins)))) (fun (k67 : Hvl) (c19 : (Cutoff TyVar)) (k68 : Hvl) (ins : (shifthvl_TyVar c19 k67 k68)) =>
    (wfTy_Prp k68)) (fun (k67 : Hvl) (c19 : (Cutoff TyVar)) (k68 : Hvl) (ins : (shifthvl_TyVar c19 k67 k68)) =>
    (wfTy_Prf k68))).
  Lemma tshift_wfTm (k67 : Hvl) :
    (forall (s31 : Tm) (wf : (wfTm k67 s31)) ,
      (forall (c19 : (Cutoff TyVar)) (k68 : Hvl) ,
        (shifthvl_TyVar c19 k67 k68) -> (wfTm k68 (tshiftTm c19 s31)))).
  Proof.
    pose proof ((tshift_wfTm_Ty k67)).
    destruct_conjs .
    easy .
  Qed.
  Lemma tshift_wfTy (k67 : Hvl) :
    (forall (S32 : Ty) (wf0 : (wfTy k67 S32)) ,
      (forall (c20 : (Cutoff TyVar)) (k69 : Hvl) ,
        (shifthvl_TyVar c20 k67 k69) -> (wfTy k69 (tshiftTy c20 S32)))).
  Proof.
    pose proof ((tshift_wfTm_Ty k67)).
    destruct_conjs .
    easy .
  Qed.
  Definition shift_wfKind : (forall (k67 : Hvl) ,
    (forall (K17 : Kind) (wf : (wfKind k67 K17)) ,
      (forall (c19 : (Cutoff TmVar)) (k68 : Hvl) ,
        (shifthvl_TmVar c19 k67 k68) -> (wfKind k68 (shiftKind c19 K17))))) := (ind_wfKind (fun (k67 : Hvl) (K17 : Kind) (wf : (wfKind k67 K17)) =>
    (forall (c19 : (Cutoff TmVar)) (k68 : Hvl) ,
      (shifthvl_TmVar c19 k67 k68) -> (wfKind k68 (shiftKind c19 K17)))) (fun (k67 : Hvl) (c19 : (Cutoff TmVar)) (k68 : Hvl) (ins : (shifthvl_TmVar c19 k67 k68)) =>
    (wfKind_Star k68)) (fun (k67 : Hvl) (T : Ty) (wf : (wfTy k67 T)) (K : Kind) (wf0 : (wfKind (HS TmVar k67) K)) IHK17 (c19 : (Cutoff TmVar)) (k68 : Hvl) (ins : (shifthvl_TmVar c19 k67 k68)) =>
    (wfKind_KPi k68 (shift_wfTy k67 T wf c19 k68 (weaken_shifthvl_TmVar H0 ins)) (IHK17 (CS c19) (HS TmVar k68) (weaken_shifthvl_TmVar (HS TmVar H0) ins))))).
  Definition tshift_wfKind : (forall (k67 : Hvl) ,
    (forall (K17 : Kind) (wf : (wfKind k67 K17)) ,
      (forall (c19 : (Cutoff TyVar)) (k68 : Hvl) ,
        (shifthvl_TyVar c19 k67 k68) -> (wfKind k68 (tshiftKind c19 K17))))) := (ind_wfKind (fun (k67 : Hvl) (K17 : Kind) (wf : (wfKind k67 K17)) =>
    (forall (c19 : (Cutoff TyVar)) (k68 : Hvl) ,
      (shifthvl_TyVar c19 k67 k68) -> (wfKind k68 (tshiftKind c19 K17)))) (fun (k67 : Hvl) (c19 : (Cutoff TyVar)) (k68 : Hvl) (ins : (shifthvl_TyVar c19 k67 k68)) =>
    (wfKind_Star k68)) (fun (k67 : Hvl) (T : Ty) (wf : (wfTy k67 T)) (K : Kind) (wf0 : (wfKind (HS TmVar k67) K)) IHK17 (c19 : (Cutoff TyVar)) (k68 : Hvl) (ins : (shifthvl_TyVar c19 k67 k68)) =>
    (wfKind_KPi k68 (tshift_wfTy k67 T wf c19 k68 (weaken_shifthvl_TyVar H0 ins)) (IHK17 c19 (HS TmVar k68) (weaken_shifthvl_TyVar (HS TmVar H0) ins))))).
End ShiftWellFormed.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : infra.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : shift.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : shift_wf.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : wf.
 Hint Resolve shift_wfKind shift_wfTm shift_wfTy tshift_wfKind tshift_wfTm tshift_wfTy : infra.
 Hint Resolve shift_wfKind shift_wfTm shift_wfTy tshift_wfKind tshift_wfTm tshift_wfTy : shift.
 Hint Resolve shift_wfKind shift_wfTm shift_wfTy tshift_wfKind tshift_wfTm tshift_wfTy : shift_wf.
 Hint Resolve shift_wfKind shift_wfTm shift_wfTy tshift_wfKind tshift_wfTm tshift_wfTy : wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : infra.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift_wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : infra.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift_wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : weaken.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : wf.
Section SubstWellFormed.
  Inductive substhvl_TmVar (k67 : Hvl) : (forall (d47 : (Trace TmVar)) (k68 : Hvl) (k69 : Hvl) ,
    Prop) :=
    | substhvl_TmVar_here :
        (substhvl_TmVar k67 X0 (HS TmVar k67) k67)
    | substhvl_TmVar_there_TmVar
        {d47 : (Trace TmVar)}
        {k68 : Hvl} {k69 : Hvl} :
        (substhvl_TmVar k67 d47 k68 k69) -> (substhvl_TmVar k67 (XS TmVar d47) (HS TmVar k68) (HS TmVar k69))
    | substhvl_TmVar_there_TyVar
        {d47 : (Trace TmVar)}
        {k68 : Hvl} {k69 : Hvl} :
        (substhvl_TmVar k67 d47 k68 k69) -> (substhvl_TmVar k67 (XS TyVar d47) (HS TyVar k68) (HS TyVar k69)).
  Inductive substhvl_TyVar (k67 : Hvl) : (forall (d47 : (Trace TyVar)) (k68 : Hvl) (k69 : Hvl) ,
    Prop) :=
    | substhvl_TyVar_here :
        (substhvl_TyVar k67 X0 (HS TyVar k67) k67)
    | substhvl_TyVar_there_TmVar
        {d47 : (Trace TyVar)}
        {k68 : Hvl} {k69 : Hvl} :
        (substhvl_TyVar k67 d47 k68 k69) -> (substhvl_TyVar k67 (XS TmVar d47) (HS TmVar k68) (HS TmVar k69))
    | substhvl_TyVar_there_TyVar
        {d47 : (Trace TyVar)}
        {k68 : Hvl} {k69 : Hvl} :
        (substhvl_TyVar k67 d47 k68 k69) -> (substhvl_TyVar k67 (XS TyVar d47) (HS TyVar k68) (HS TyVar k69)).
  Lemma weaken_substhvl_TmVar  :
    (forall {k68 : Hvl} (k67 : Hvl) {d47 : (Trace TmVar)} {k69 : Hvl} {k70 : Hvl} ,
      (substhvl_TmVar k68 d47 k69 k70) -> (substhvl_TmVar k68 (weakenTrace d47 k67) (appendHvl k69 k67) (appendHvl k70 k67))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma weaken_substhvl_TyVar  :
    (forall {k68 : Hvl} (k67 : Hvl) {d47 : (Trace TyVar)} {k69 : Hvl} {k70 : Hvl} ,
      (substhvl_TyVar k68 d47 k69 k70) -> (substhvl_TyVar k68 (weakenTrace d47 k67) (appendHvl k69 k67) (appendHvl k70 k67))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma substhvl_TmVar_wfindex_TmVar {k67 : Hvl} {s31 : Tm} (wft : (wfTm k67 s31)) :
    (forall {d47 : (Trace TmVar)} {k68 : Hvl} {k69 : Hvl} ,
      (substhvl_TmVar k67 d47 k68 k69) -> (forall {x107 : (Index TmVar)} ,
        (wfindex k68 x107) -> (wfTm k69 (substIndex d47 s31 x107)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TyVar_wfindex_TyVar {k67 : Hvl} {S32 : Ty} (wft : (wfTy k67 S32)) :
    (forall {d47 : (Trace TyVar)} {k68 : Hvl} {k69 : Hvl} ,
      (substhvl_TyVar k67 d47 k68 k69) -> (forall {X35 : (Index TyVar)} ,
        (wfindex k68 X35) -> (wfTy k69 (tsubstIndex d47 S32 X35)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TmVar_wfindex_TyVar {k67 : Hvl} :
    (forall {d47 : (Trace TmVar)} {k68 : Hvl} {k69 : Hvl} ,
      (substhvl_TmVar k67 d47 k68 k69) -> (forall {X35 : (Index TyVar)} ,
        (wfindex k68 X35) -> (wfindex k69 X35))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_TyVar_wfindex_TmVar {k67 : Hvl} :
    (forall {d47 : (Trace TyVar)} {k68 : Hvl} {k69 : Hvl} ,
      (substhvl_TyVar k67 d47 k68 k69) -> (forall {x107 : (Index TmVar)} ,
        (wfindex k68 x107) -> (wfindex k69 x107))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Definition substhvl_TmVar_wfTm_Ty {k67 : Hvl} {s31 : Tm} (wf : (wfTm k67 s31)) : (forall (k68 : Hvl) ,
    (forall (s32 : Tm) (wf0 : (wfTm k68 s32)) ,
      (forall {d47 : (Trace TmVar)} {k69 : Hvl} ,
        (substhvl_TmVar k67 d47 k68 k69) -> (wfTm k69 (substTm d47 s31 s32)))) /\
    (forall (S32 : Ty) (wf0 : (wfTy k68 S32)) ,
      (forall {d47 : (Trace TmVar)} {k69 : Hvl} ,
        (substhvl_TmVar k67 d47 k68 k69) -> (wfTy k69 (substTy d47 s31 S32))))) := (ind_wfTm_Ty (fun (k68 : Hvl) (s32 : Tm) (wf0 : (wfTm k68 s32)) =>
    (forall {d47 : (Trace TmVar)} {k69 : Hvl} ,
      (substhvl_TmVar k67 d47 k68 k69) -> (wfTm k69 (substTm d47 s31 s32)))) (fun (k68 : Hvl) (S32 : Ty) (wf0 : (wfTy k68 S32)) =>
    (forall {d47 : (Trace TmVar)} {k69 : Hvl} ,
      (substhvl_TmVar k67 d47 k68 k69) -> (wfTy k69 (substTy d47 s31 S32)))) (fun (k68 : Hvl) {x107 : (Index TmVar)} (wfi : (wfindex k68 x107)) {d47 : (Trace TmVar)} {k69 : Hvl} (del : (substhvl_TmVar k67 d47 k68 k69)) =>
    (substhvl_TmVar_wfindex_TmVar wf del wfi)) (fun (k68 : Hvl) (T : Ty) (wf0 : (wfTy k68 T)) IHT107 (t : Tm) (wf1 : (wfTm (HS TmVar k68) t)) IHt89 {d47 : (Trace TmVar)} {k69 : Hvl} (del : (substhvl_TmVar k67 d47 k68 k69)) =>
    (wfTm_Abs k69 (IHT107 (weakenTrace d47 H0) k69 (weaken_substhvl_TmVar H0 del)) (IHt89 (weakenTrace d47 (HS TmVar H0)) (HS TmVar k69) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k68 : Hvl) (t1 : Tm) (wf0 : (wfTm k68 t1)) IHt89 (t2 : Tm) (wf1 : (wfTm k68 t2)) IHt90 {d47 : (Trace TmVar)} {k69 : Hvl} (del : (substhvl_TmVar k67 d47 k68 k69)) =>
    (wfTm_App k69 (IHt89 (weakenTrace d47 H0) k69 (weaken_substhvl_TmVar H0 del)) (IHt90 (weakenTrace d47 H0) k69 (weaken_substhvl_TmVar H0 del)))) (fun (k68 : Hvl) (T : Ty) (wf0 : (wfTy k68 T)) IHT107 (t : Tm) (wf1 : (wfTm (HS TmVar k68) t)) IHt89 {d47 : (Trace TmVar)} {k69 : Hvl} (del : (substhvl_TmVar k67 d47 k68 k69)) =>
    (wfTm_All k69 (IHT107 (weakenTrace d47 H0) k69 (weaken_substhvl_TmVar H0 del)) (IHt89 (weakenTrace d47 (HS TmVar H0)) (HS TmVar k69) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k68 : Hvl) {X35 : (Index TyVar)} (wfi : (wfindex k68 X35)) {d47 : (Trace TmVar)} {k69 : Hvl} (del : (substhvl_TmVar k67 d47 k68 k69)) =>
    (wfTy_TVar k69 (substhvl_TmVar_wfindex_TyVar del wfi))) (fun (k68 : Hvl) (T1 : Ty) (wf0 : (wfTy k68 T1)) IHT107 (T2 : Ty) (wf1 : (wfTy (HS TmVar k68) T2)) IHT108 {d47 : (Trace TmVar)} {k69 : Hvl} (del : (substhvl_TmVar k67 d47 k68 k69)) =>
    (wfTy_TPi k69 (IHT107 (weakenTrace d47 H0) k69 (weaken_substhvl_TmVar H0 del)) (IHT108 (weakenTrace d47 (HS TmVar H0)) (HS TmVar k69) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k68 : Hvl) (T : Ty) (wf0 : (wfTy k68 T)) IHT107 (t : Tm) (wf1 : (wfTm k68 t)) IHt89 {d47 : (Trace TmVar)} {k69 : Hvl} (del : (substhvl_TmVar k67 d47 k68 k69)) =>
    (wfTy_TApp k69 (IHT107 (weakenTrace d47 H0) k69 (weaken_substhvl_TmVar H0 del)) (IHt89 (weakenTrace d47 H0) k69 (weaken_substhvl_TmVar H0 del)))) (fun (k68 : Hvl) {d47 : (Trace TmVar)} {k69 : Hvl} (del : (substhvl_TmVar k67 d47 k68 k69)) =>
    (wfTy_Prp k69)) (fun (k68 : Hvl) {d47 : (Trace TmVar)} {k69 : Hvl} (del : (substhvl_TmVar k67 d47 k68 k69)) =>
    (wfTy_Prf k69))).
  Lemma substhvl_TmVar_wfTm {k67 : Hvl} {s31 : Tm} (wf : (wfTm k67 s31)) (k68 : Hvl) (s32 : Tm) (wf0 : (wfTm k68 s32)) :
    (forall {d47 : (Trace TmVar)} {k69 : Hvl} ,
      (substhvl_TmVar k67 d47 k68 k69) -> (wfTm k69 (substTm d47 s31 s32))).
  Proof.
    apply ((substhvl_TmVar_wfTm_Ty wf k68)).
    auto .
  Qed.
  Lemma substhvl_TmVar_wfTy {k67 : Hvl} {s31 : Tm} (wf : (wfTm k67 s31)) (k68 : Hvl) (S32 : Ty) (wf1 : (wfTy k68 S32)) :
    (forall {d48 : (Trace TmVar)} {k70 : Hvl} ,
      (substhvl_TmVar k67 d48 k68 k70) -> (wfTy k70 (substTy d48 s31 S32))).
  Proof.
    apply ((substhvl_TmVar_wfTm_Ty wf k68)).
    auto .
  Qed.
  Definition substhvl_TyVar_wfTm_Ty {k67 : Hvl} {S32 : Ty} (wf : (wfTy k67 S32)) : (forall (k68 : Hvl) ,
    (forall (s31 : Tm) (wf0 : (wfTm k68 s31)) ,
      (forall {d47 : (Trace TyVar)} {k69 : Hvl} ,
        (substhvl_TyVar k67 d47 k68 k69) -> (wfTm k69 (tsubstTm d47 S32 s31)))) /\
    (forall (S33 : Ty) (wf0 : (wfTy k68 S33)) ,
      (forall {d47 : (Trace TyVar)} {k69 : Hvl} ,
        (substhvl_TyVar k67 d47 k68 k69) -> (wfTy k69 (tsubstTy d47 S32 S33))))) := (ind_wfTm_Ty (fun (k68 : Hvl) (s31 : Tm) (wf0 : (wfTm k68 s31)) =>
    (forall {d47 : (Trace TyVar)} {k69 : Hvl} ,
      (substhvl_TyVar k67 d47 k68 k69) -> (wfTm k69 (tsubstTm d47 S32 s31)))) (fun (k68 : Hvl) (S33 : Ty) (wf0 : (wfTy k68 S33)) =>
    (forall {d47 : (Trace TyVar)} {k69 : Hvl} ,
      (substhvl_TyVar k67 d47 k68 k69) -> (wfTy k69 (tsubstTy d47 S32 S33)))) (fun (k68 : Hvl) {x107 : (Index TmVar)} (wfi : (wfindex k68 x107)) {d47 : (Trace TyVar)} {k69 : Hvl} (del : (substhvl_TyVar k67 d47 k68 k69)) =>
    (wfTm_Var k69 (substhvl_TyVar_wfindex_TmVar del wfi))) (fun (k68 : Hvl) (T : Ty) (wf0 : (wfTy k68 T)) IHT107 (t : Tm) (wf1 : (wfTm (HS TmVar k68) t)) IHt89 {d47 : (Trace TyVar)} {k69 : Hvl} (del : (substhvl_TyVar k67 d47 k68 k69)) =>
    (wfTm_Abs k69 (IHT107 (weakenTrace d47 H0) k69 (weaken_substhvl_TyVar H0 del)) (IHt89 (weakenTrace d47 (HS TmVar H0)) (HS TmVar k69) (weaken_substhvl_TyVar (HS TmVar H0) del)))) (fun (k68 : Hvl) (t1 : Tm) (wf0 : (wfTm k68 t1)) IHt89 (t2 : Tm) (wf1 : (wfTm k68 t2)) IHt90 {d47 : (Trace TyVar)} {k69 : Hvl} (del : (substhvl_TyVar k67 d47 k68 k69)) =>
    (wfTm_App k69 (IHt89 (weakenTrace d47 H0) k69 (weaken_substhvl_TyVar H0 del)) (IHt90 (weakenTrace d47 H0) k69 (weaken_substhvl_TyVar H0 del)))) (fun (k68 : Hvl) (T : Ty) (wf0 : (wfTy k68 T)) IHT107 (t : Tm) (wf1 : (wfTm (HS TmVar k68) t)) IHt89 {d47 : (Trace TyVar)} {k69 : Hvl} (del : (substhvl_TyVar k67 d47 k68 k69)) =>
    (wfTm_All k69 (IHT107 (weakenTrace d47 H0) k69 (weaken_substhvl_TyVar H0 del)) (IHt89 (weakenTrace d47 (HS TmVar H0)) (HS TmVar k69) (weaken_substhvl_TyVar (HS TmVar H0) del)))) (fun (k68 : Hvl) {X35 : (Index TyVar)} (wfi : (wfindex k68 X35)) {d47 : (Trace TyVar)} {k69 : Hvl} (del : (substhvl_TyVar k67 d47 k68 k69)) =>
    (substhvl_TyVar_wfindex_TyVar wf del wfi)) (fun (k68 : Hvl) (T1 : Ty) (wf0 : (wfTy k68 T1)) IHT107 (T2 : Ty) (wf1 : (wfTy (HS TmVar k68) T2)) IHT108 {d47 : (Trace TyVar)} {k69 : Hvl} (del : (substhvl_TyVar k67 d47 k68 k69)) =>
    (wfTy_TPi k69 (IHT107 (weakenTrace d47 H0) k69 (weaken_substhvl_TyVar H0 del)) (IHT108 (weakenTrace d47 (HS TmVar H0)) (HS TmVar k69) (weaken_substhvl_TyVar (HS TmVar H0) del)))) (fun (k68 : Hvl) (T : Ty) (wf0 : (wfTy k68 T)) IHT107 (t : Tm) (wf1 : (wfTm k68 t)) IHt89 {d47 : (Trace TyVar)} {k69 : Hvl} (del : (substhvl_TyVar k67 d47 k68 k69)) =>
    (wfTy_TApp k69 (IHT107 (weakenTrace d47 H0) k69 (weaken_substhvl_TyVar H0 del)) (IHt89 (weakenTrace d47 H0) k69 (weaken_substhvl_TyVar H0 del)))) (fun (k68 : Hvl) {d47 : (Trace TyVar)} {k69 : Hvl} (del : (substhvl_TyVar k67 d47 k68 k69)) =>
    (wfTy_Prp k69)) (fun (k68 : Hvl) {d47 : (Trace TyVar)} {k69 : Hvl} (del : (substhvl_TyVar k67 d47 k68 k69)) =>
    (wfTy_Prf k69))).
  Lemma substhvl_TyVar_wfTm {k67 : Hvl} {S32 : Ty} (wf : (wfTy k67 S32)) (k68 : Hvl) (s31 : Tm) (wf0 : (wfTm k68 s31)) :
    (forall {d47 : (Trace TyVar)} {k69 : Hvl} ,
      (substhvl_TyVar k67 d47 k68 k69) -> (wfTm k69 (tsubstTm d47 S32 s31))).
  Proof.
    apply ((substhvl_TyVar_wfTm_Ty wf k68)).
    auto .
  Qed.
  Lemma substhvl_TyVar_wfTy {k67 : Hvl} {S32 : Ty} (wf : (wfTy k67 S32)) (k68 : Hvl) (S33 : Ty) (wf1 : (wfTy k68 S33)) :
    (forall {d48 : (Trace TyVar)} {k70 : Hvl} ,
      (substhvl_TyVar k67 d48 k68 k70) -> (wfTy k70 (tsubstTy d48 S32 S33))).
  Proof.
    apply ((substhvl_TyVar_wfTm_Ty wf k68)).
    auto .
  Qed.
  Definition substhvl_TmVar_wfKind {k67 : Hvl} {s31 : Tm} (wf : (wfTm k67 s31)) : (forall (k68 : Hvl) ,
    (forall (K17 : Kind) (wf0 : (wfKind k68 K17)) ,
      (forall {d47 : (Trace TmVar)} {k69 : Hvl} ,
        (substhvl_TmVar k67 d47 k68 k69) -> (wfKind k69 (substKind d47 s31 K17))))) := (ind_wfKind (fun (k68 : Hvl) (K17 : Kind) (wf0 : (wfKind k68 K17)) =>
    (forall {d47 : (Trace TmVar)} {k69 : Hvl} ,
      (substhvl_TmVar k67 d47 k68 k69) -> (wfKind k69 (substKind d47 s31 K17)))) (fun (k68 : Hvl) {d47 : (Trace TmVar)} {k69 : Hvl} (del : (substhvl_TmVar k67 d47 k68 k69)) =>
    (wfKind_Star k69)) (fun (k68 : Hvl) (T : Ty) (wf0 : (wfTy k68 T)) (K : Kind) (wf1 : (wfKind (HS TmVar k68) K)) IHK17 {d47 : (Trace TmVar)} {k69 : Hvl} (del : (substhvl_TmVar k67 d47 k68 k69)) =>
    (wfKind_KPi k69 (substhvl_TmVar_wfTy wf k68 T wf0 (weaken_substhvl_TmVar H0 del)) (IHK17 (weakenTrace d47 (HS TmVar H0)) (HS TmVar k69) (weaken_substhvl_TmVar (HS TmVar H0) del))))).
  Definition substhvl_TyVar_wfKind {k67 : Hvl} {S32 : Ty} (wf : (wfTy k67 S32)) : (forall (k68 : Hvl) ,
    (forall (K17 : Kind) (wf0 : (wfKind k68 K17)) ,
      (forall {d47 : (Trace TyVar)} {k69 : Hvl} ,
        (substhvl_TyVar k67 d47 k68 k69) -> (wfKind k69 (tsubstKind d47 S32 K17))))) := (ind_wfKind (fun (k68 : Hvl) (K17 : Kind) (wf0 : (wfKind k68 K17)) =>
    (forall {d47 : (Trace TyVar)} {k69 : Hvl} ,
      (substhvl_TyVar k67 d47 k68 k69) -> (wfKind k69 (tsubstKind d47 S32 K17)))) (fun (k68 : Hvl) {d47 : (Trace TyVar)} {k69 : Hvl} (del : (substhvl_TyVar k67 d47 k68 k69)) =>
    (wfKind_Star k69)) (fun (k68 : Hvl) (T : Ty) (wf0 : (wfTy k68 T)) (K : Kind) (wf1 : (wfKind (HS TmVar k68) K)) IHK17 {d47 : (Trace TyVar)} {k69 : Hvl} (del : (substhvl_TyVar k67 d47 k68 k69)) =>
    (wfKind_KPi k69 (substhvl_TyVar_wfTy wf k68 T wf0 (weaken_substhvl_TyVar H0 del)) (IHK17 (weakenTrace d47 (HS TmVar H0)) (HS TmVar k69) (weaken_substhvl_TyVar (HS TmVar H0) del))))).
End SubstWellFormed.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : infra.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst_wf.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : wf.
 Hint Resolve substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : infra.
 Hint Resolve substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst.
 Hint Resolve substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst_wf.
 Hint Resolve substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : infra.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst_wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : wf.
Section Context.
  Inductive Ctx : Type :=
    | empty 
    | evar (G : Ctx) (T : Ty)
    | etvar (G : Ctx) (K : Kind).
  Fixpoint appendCtx (G : Ctx) (G0 : Ctx) : Ctx :=
    match G0 with
      | (empty) => G
      | (evar G1 T) => (evar (appendCtx G G1) T)
      | (etvar G1 K) => (etvar (appendCtx G G1) K)
    end.
  Fixpoint domainCtx (G : Ctx) : Hvl :=
    match G with
      | (empty) => H0
      | (evar G0 T) => (HS TmVar (domainCtx G0))
      | (etvar G0 K) => (HS TyVar (domainCtx G0))
    end.
  Lemma appendCtx_assoc  :
    (forall (G : Ctx) (G0 : Ctx) (G1 : Ctx) ,
      ((appendCtx G (appendCtx G0 G1)) =
      (appendCtx (appendCtx G G0) G1))).
  Proof.
    needleGenericAppendEnvAssoc .
  Qed.
  Lemma domainCtx_appendCtx  :
    (forall (G : Ctx) (G0 : Ctx) ,
      ((domainCtx (appendCtx G G0)) =
      (appendHvl (domainCtx G) (domainCtx G0)))).
  Proof.
    needleGenericDomainEnvAppendEnv .
  Qed.
  Fixpoint shiftCtx (c19 : (Cutoff TmVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shiftCtx c19 G0) (shiftTy (weakenCutoffTmVar c19 (domainCtx G0)) T))
      | (etvar G0 K) => (etvar (shiftCtx c19 G0) (shiftKind (weakenCutoffTmVar c19 (domainCtx G0)) K))
    end.
  Fixpoint tshiftCtx (c19 : (Cutoff TyVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (tshiftCtx c19 G0) (tshiftTy (weakenCutoffTyVar c19 (domainCtx G0)) T))
      | (etvar G0 K) => (etvar (tshiftCtx c19 G0) (tshiftKind (weakenCutoffTyVar c19 (domainCtx G0)) K))
    end.
  Fixpoint substCtx (d47 : (Trace TmVar)) (s31 : Tm) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (substCtx d47 s31 G0) (substTy (weakenTrace d47 (domainCtx G0)) s31 T))
      | (etvar G0 K) => (etvar (substCtx d47 s31 G0) (substKind (weakenTrace d47 (domainCtx G0)) s31 K))
    end.
  Fixpoint tsubstCtx (d47 : (Trace TyVar)) (S32 : Ty) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (tsubstCtx d47 S32 G0) (tsubstTy (weakenTrace d47 (domainCtx G0)) S32 T))
      | (etvar G0 K) => (etvar (tsubstCtx d47 S32 G0) (tsubstKind (weakenTrace d47 (domainCtx G0)) S32 K))
    end.
  Lemma domainCtx_shiftCtx  :
    (forall (c19 : (Cutoff TmVar)) (G : Ctx) ,
      ((domainCtx (shiftCtx c19 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_tshiftCtx  :
    (forall (c19 : (Cutoff TyVar)) (G : Ctx) ,
      ((domainCtx (tshiftCtx c19 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_substCtx  :
    (forall (d47 : (Trace TmVar)) (s31 : Tm) (G : Ctx) ,
      ((domainCtx (substCtx d47 s31 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
  Lemma domainCtx_tsubstCtx  :
    (forall (d47 : (Trace TyVar)) (S32 : Ty) (G : Ctx) ,
      ((domainCtx (tsubstCtx d47 S32 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
End Context.
 Hint Rewrite domainCtx_appendCtx : interaction.
 Hint Rewrite domainCtx_appendCtx : env_domain_append.
Section ContextStuff.
  Section ShiftEnvAppendEnv.
    Lemma shiftCtx_appendCtx  :
      (forall (c19 : (Cutoff TmVar)) (G : Ctx) (G0 : Ctx) ,
        ((shiftCtx c19 (appendCtx G G0)) =
        (appendCtx (shiftCtx c19 G) (shiftCtx (weakenCutoffTmVar c19 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
    Lemma tshiftCtx_appendCtx  :
      (forall (c19 : (Cutoff TyVar)) (G : Ctx) (G0 : Ctx) ,
        ((tshiftCtx c19 (appendCtx G G0)) =
        (appendCtx (tshiftCtx c19 G) (tshiftCtx (weakenCutoffTyVar c19 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
  End ShiftEnvAppendEnv.
  Section SubstEnvAppendEnv.
    Lemma substCtx_appendCtx  :
      (forall (d47 : (Trace TmVar)) (s31 : Tm) (G : Ctx) (G0 : Ctx) ,
        ((substCtx d47 s31 (appendCtx G G0)) =
        (appendCtx (substCtx d47 s31 G) (substCtx (weakenTrace d47 (domainCtx G)) s31 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
    Lemma tsubstCtx_appendCtx  :
      (forall (d47 : (Trace TyVar)) (S32 : Ty) (G : Ctx) (G0 : Ctx) ,
        ((tsubstCtx d47 S32 (appendCtx G G0)) =
        (appendCtx (tsubstCtx d47 S32 G) (tsubstCtx (weakenTrace d47 (domainCtx G)) S32 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
  End SubstEnvAppendEnv.
  Section Lookups.
    Inductive lookup_evar : Ctx -> (Index TmVar) -> Ty -> Prop :=
      | lookup_evar_here {G : Ctx}
          {T107 : Ty} :
          (wfTy (domainCtx G) T107) -> (lookup_evar (evar G T107) I0 (shiftTy C0 T107))
      | lookup_evar_there_evar
          {G : Ctx} {x107 : (Index TmVar)}
          {T107 : Ty} {T108 : Ty} :
          (lookup_evar G x107 T107) -> (lookup_evar (evar G T108) (IS x107) (shiftTy C0 T107))
      | lookup_evar_there_etvar
          {G : Ctx} {x107 : (Index TmVar)}
          {T107 : Ty} {K17 : Kind} :
          (lookup_evar G x107 T107) -> (lookup_evar (etvar G K17) x107 (tshiftTy C0 T107)).
    Inductive lookup_etvar : Ctx -> (Index TyVar) -> Kind -> Prop :=
      | lookup_etvar_here {G : Ctx}
          {K17 : Kind} :
          (wfKind (domainCtx G) K17) -> (lookup_etvar (etvar G K17) I0 (tshiftKind C0 K17))
      | lookup_etvar_there_evar
          {G : Ctx} {X35 : (Index TyVar)}
          {K17 : Kind} {T107 : Ty} :
          (lookup_etvar G X35 K17) -> (lookup_etvar (evar G T107) X35 (shiftKind C0 K17))
      | lookup_etvar_there_etvar
          {G : Ctx} {X35 : (Index TyVar)}
          {K17 : Kind} {K18 : Kind} :
          (lookup_etvar G X35 K17) -> (lookup_etvar (etvar G K18) (IS X35) (tshiftKind C0 K17)).
    Lemma lookup_evar_inversion_here  :
      (forall (G : Ctx) (S32 : Ty) (S33 : Ty) ,
        (lookup_evar (evar G S32) I0 S33) -> ((shiftTy C0 S32) =
        S33)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_etvar_inversion_here  :
      (forall (G : Ctx) (K17 : Kind) (K18 : Kind) ,
        (lookup_etvar (etvar G K17) I0 K18) -> ((tshiftKind C0 K17) =
        K18)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_evar_functional  :
      (forall {G : Ctx} {x107 : (Index TmVar)} ,
        (forall {S32 : Ty} ,
          (lookup_evar G x107 S32) -> (forall {S33 : Ty} ,
            (lookup_evar G x107 S33) -> (S32 =
            S33)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_etvar_functional  :
      (forall {G : Ctx} {X35 : (Index TyVar)} ,
        (forall {K17 : Kind} ,
          (lookup_etvar G X35 K17) -> (forall {K18 : Kind} ,
            (lookup_etvar G X35 K18) -> (K17 =
            K18)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_evar_wf  :
      (forall {G : Ctx} {x107 : (Index TmVar)} {S32 : Ty} ,
        (lookup_evar G x107 S32) -> (wfTy (domainCtx G) S32)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma lookup_etvar_wf  :
      (forall {G : Ctx} {X35 : (Index TyVar)} {K17 : Kind} ,
        (lookup_etvar G X35 K17) -> (wfKind (domainCtx G) K17)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma weaken_lookup_evar  :
      (forall {G : Ctx} (G0 : Ctx) {x107 : (Index TmVar)} {T107 : Ty} ,
        (lookup_evar G x107 T107) -> (lookup_evar (appendCtx G G0) (weakenIndexTmVar x107 (domainCtx G0)) (weakenTy T107 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma weaken_lookup_etvar  :
      (forall {G : Ctx} (G0 : Ctx) {X35 : (Index TyVar)} {K17 : Kind} ,
        (lookup_etvar G X35 K17) -> (lookup_etvar (appendCtx G G0) (weakenIndexTyVar X35 (domainCtx G0)) (weakenKind K17 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma lookup_evar_wfindex  :
      (forall {G : Ctx} {x107 : (Index TmVar)} {S32 : Ty} ,
        (lookup_evar G x107 S32) -> (wfindex (domainCtx G) x107)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
    Lemma lookup_etvar_wfindex  :
      (forall {G : Ctx} {X35 : (Index TyVar)} {K18 : Kind} ,
        (lookup_etvar G X35 K18) -> (wfindex (domainCtx G) X35)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
  End Lookups.
  Inductive shift_evar : (Cutoff TmVar) -> Ctx -> Ctx -> Prop :=
    | shift_evar_here {G : Ctx}
        {T107 : Ty} :
        (shift_evar C0 G (evar G T107))
    | shiftevar_there_evar
        {c19 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_evar c19 G G0) -> (shift_evar (CS c19) (evar G T) (evar G0 (shiftTy c19 T)))
    | shiftevar_there_etvar
        {c19 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} {K : Kind} :
        (shift_evar c19 G G0) -> (shift_evar c19 (etvar G K) (etvar G0 (shiftKind c19 K))).
  Lemma weaken_shift_evar  :
    (forall (G : Ctx) {c19 : (Cutoff TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_evar c19 G0 G1) -> (shift_evar (weakenCutoffTmVar c19 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 (shiftCtx c19 G)))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Inductive shift_etvar : (Cutoff TyVar) -> Ctx -> Ctx -> Prop :=
    | shift_etvar_here {G : Ctx}
        {K17 : Kind} :
        (shift_etvar C0 G (etvar G K17))
    | shiftetvar_there_evar
        {c19 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_etvar c19 G G0) -> (shift_etvar c19 (evar G T) (evar G0 (tshiftTy c19 T)))
    | shiftetvar_there_etvar
        {c19 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} {K : Kind} :
        (shift_etvar c19 G G0) -> (shift_etvar (CS c19) (etvar G K) (etvar G0 (tshiftKind c19 K))).
  Lemma weaken_shift_etvar  :
    (forall (G : Ctx) {c19 : (Cutoff TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_etvar c19 G0 G1) -> (shift_etvar (weakenCutoffTyVar c19 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 (tshiftCtx c19 G)))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Lemma shift_evar_shifthvl_TmVar  :
    (forall {c19 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_evar c19 G G0) -> (shifthvl_TmVar c19 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Lemma shift_etvar_shifthvl_TyVar  :
    (forall {c19 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_etvar c19 G G0) -> (shifthvl_TyVar c19 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Inductive subst_evar (G : Ctx) (T107 : Ty) (s31 : Tm) : (Trace TmVar) -> Ctx -> Ctx -> Prop :=
    | subst_evar_here :
        (subst_evar G T107 s31 X0 (evar G T107) G)
    | subst_evar_there_evar
        {d47 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_evar G T107 s31 d47 G0 G1) -> (subst_evar G T107 s31 (XS TmVar d47) (evar G0 T) (evar G1 (substTy d47 s31 T)))
    | subst_evar_there_etvar
        {d47 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} {K : Kind} :
        (subst_evar G T107 s31 d47 G0 G1) -> (subst_evar G T107 s31 (XS TyVar d47) (etvar G0 K) (etvar G1 (substKind d47 s31 K))).
  Lemma weaken_subst_evar {G : Ctx} {T107 : Ty} {s31 : Tm} :
    (forall (G0 : Ctx) {d47 : (Trace TmVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_evar G T107 s31 d47 G1 G2) -> (subst_evar G T107 s31 (weakenTrace d47 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 (substCtx d47 s31 G0)))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Inductive subst_etvar (G : Ctx) (K17 : Kind) (S32 : Ty) : (Trace TyVar) -> Ctx -> Ctx -> Prop :=
    | subst_etvar_here :
        (subst_etvar G K17 S32 X0 (etvar G K17) G)
    | subst_etvar_there_evar
        {d47 : (Trace TyVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_etvar G K17 S32 d47 G0 G1) -> (subst_etvar G K17 S32 (XS TmVar d47) (evar G0 T) (evar G1 (tsubstTy d47 S32 T)))
    | subst_etvar_there_etvar
        {d47 : (Trace TyVar)} {G0 : Ctx}
        {G1 : Ctx} {K : Kind} :
        (subst_etvar G K17 S32 d47 G0 G1) -> (subst_etvar G K17 S32 (XS TyVar d47) (etvar G0 K) (etvar G1 (tsubstKind d47 S32 K))).
  Lemma weaken_subst_etvar {G : Ctx} {K17 : Kind} {S32 : Ty} :
    (forall (G0 : Ctx) {d47 : (Trace TyVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_etvar G K17 S32 d47 G1 G2) -> (subst_etvar G K17 S32 (weakenTrace d47 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 (tsubstCtx d47 S32 G0)))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Lemma subst_evar_substhvl_TmVar {G : Ctx} {s31 : Tm} {T107 : Ty} :
    (forall {d47 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_evar G T107 s31 d47 G0 G1) -> (substhvl_TmVar (domainCtx G) d47 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
  Lemma subst_etvar_substhvl_TyVar {G : Ctx} {S32 : Ty} {K17 : Kind} :
    (forall {d47 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_etvar G K17 S32 d47 G0 G1) -> (substhvl_TyVar (domainCtx G) d47 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
End ContextStuff.
 Hint Rewrite domainCtx_shiftCtx domainCtx_tshiftCtx : interaction.
 Hint Rewrite domainCtx_shiftCtx domainCtx_tshiftCtx : env_domain_shift.
 Hint Rewrite domainCtx_substCtx domainCtx_tsubstCtx : interaction.
 Hint Rewrite domainCtx_substCtx domainCtx_tsubstCtx : env_domain_subst.
 Hint Rewrite shiftCtx_appendCtx tshiftCtx_appendCtx : interaction.
 Hint Rewrite shiftCtx_appendCtx tshiftCtx_appendCtx : env_shift_append.
 Hint Rewrite substCtx_appendCtx tsubstCtx_appendCtx : interaction.
 Hint Rewrite substCtx_appendCtx tsubstCtx_appendCtx : env_shift_append.
 Hint Constructors lookup_evar lookup_etvar : infra.
 Hint Constructors lookup_evar lookup_etvar : lookup.
 Hint Constructors lookup_evar lookup_etvar : shift.
 Hint Constructors lookup_evar lookup_etvar : subst.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : infra.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : lookup.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : shift.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : weaken.
Lemma weaken_lookup_evar_here  :
  (forall {G : Ctx} (G0 : Ctx) {T107 : Ty} (wf : (wfTy (domainCtx G) T107)) ,
    (lookup_evar (appendCtx (evar G T107) G0) (weakenIndexTmVar I0 (domainCtx G0)) (weakenTy (shiftTy C0 T107) (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
Lemma weaken_lookup_etvar_here  :
  (forall {G : Ctx} (G0 : Ctx) {K17 : Kind} (wf : (wfKind (domainCtx G) K17)) ,
    (lookup_etvar (appendCtx (etvar G K17) G0) (weakenIndexTyVar I0 (domainCtx G0)) (weakenKind (tshiftKind C0 K17) (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
 Hint Constructors wfKind wfTm wfTy : infra.
 Hint Constructors wfKind wfTm wfTy : wf.
 Hint Extern 10 ((wfKind _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTm _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTy _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 2 ((wfTm _ _)) => match goal with
  | H : (wfTm _ (Var _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Abs _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (App _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (All _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfTy _ _)) => match goal with
  | H : (wfTy _ (TVar _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TPi _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TApp _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (Prp)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (Prf)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfKind _ _)) => match goal with
  | H : (wfKind _ (Star)) |- _ => inversion H; subst; clear H
  | H : (wfKind _ (KPi _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Resolve lookup_evar_wf lookup_etvar_wf : infra.
 Hint Resolve lookup_evar_wf lookup_etvar_wf : wf.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : infra.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : lookup.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : wf.
 Hint Constructors shift_evar shift_etvar : infra.
 Hint Constructors shift_evar shift_etvar : shift.
 Hint Constructors shift_evar shift_etvar : subst.
 Hint Resolve weaken_shift_evar weaken_shift_etvar : infra.
 Hint Resolve weaken_shift_evar weaken_shift_etvar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : infra.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : shift_wf.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : wf.
 Hint Constructors subst_evar subst_etvar : infra.
 Hint Constructors subst_evar subst_etvar : subst.
 Hint Resolve weaken_subst_evar weaken_subst_etvar : infra.
 Hint Resolve weaken_subst_evar weaken_subst_etvar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : infra.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : subst_wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : substenv_substhvl.
Lemma shift_evar_lookup_evar  :
  (forall {c19 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c19 G G0)) {x107 : (Index TmVar)} {T107 : Ty} ,
    (lookup_evar G x107 T107) -> (lookup_evar G0 (shiftIndex c19 x107) (shiftTy c19 T107))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_evar_lookup_etvar  :
  (forall {c19 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c19 G G0)) {X35 : (Index TyVar)} {K17 : Kind} ,
    (lookup_etvar G X35 K17) -> (lookup_etvar G0 X35 (shiftKind c19 K17))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_evar  :
  (forall {c19 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c19 G G0)) {x107 : (Index TmVar)} {T107 : Ty} ,
    (lookup_evar G x107 T107) -> (lookup_evar G0 x107 (tshiftTy c19 T107))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_etvar  :
  (forall {c19 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c19 G G0)) {X35 : (Index TyVar)} {K17 : Kind} ,
    (lookup_etvar G X35 K17) -> (lookup_etvar G0 (tshiftIndex c19 X35) (tshiftKind c19 K17))).
Proof.
  needleGenericInsertLookup .
Qed.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : infra.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : lookup.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : shift.
Lemma subst_evar_lookup_etvar {G : Ctx} {T108 : Ty} {s31 : Tm} (wf : (wfTm (domainCtx G) s31)) :
  (forall {d47 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_evar G T108 s31 d47 G0 G1)) {X35 : (Index TyVar)} {K18 : Kind} ,
    (lookup_etvar G0 X35 K18) -> (lookup_etvar G1 X35 (substKind d47 s31 K18))).
Proof.
  needleGenericSubstEnvLookup .
Qed.
Lemma subst_etvar_lookup_evar {G : Ctx} {K18 : Kind} {S32 : Ty} (wf : (wfTy (domainCtx G) S32)) :
  (forall {d47 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_etvar G K18 S32 d47 G0 G1)) {x107 : (Index TmVar)} {T108 : Ty} ,
    (lookup_evar G0 x107 T108) -> (lookup_evar G1 x107 (tsubstTy d47 S32 T108))).
Proof.
  needleGenericSubstEnvLookup .
Qed.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : infra.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : lookup.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : subst.
Fixpoint size_Tm (s : Tm) {struct s} : nat :=
  match s with
    | (Var x) => 1
    | (Abs T t) => (plus 1 (plus (size_Ty T) (size_Tm t)))
    | (App t1 t2) => (plus 1 (plus (size_Tm t1) (size_Tm t2)))
    | (All T0 t0) => (plus 1 (plus (size_Ty T0) (size_Tm t0)))
  end
with size_Ty (S0 : Ty) {struct S0} : nat :=
  match S0 with
    | (TVar X) => 1
    | (TPi T1 T2) => (plus 1 (plus (size_Ty T1) (size_Ty T2)))
    | (TApp T t) => (plus 1 (plus (size_Ty T) (size_Tm t)))
    | (Prp) => 1
    | (Prf) => 1
  end.
Fixpoint size_Kind (K : Kind) {struct K} : nat :=
  match K with
    | (Star) => 1
    | (KPi T K0) => (plus 1 (plus (size_Ty T) (size_Kind K0)))
  end.
Lemma shift_size_Tm_Ty  :
  (forall (s31 : Tm) (c15 : (Cutoff TmVar)) ,
    ((size_Tm (shiftTm c15 s31)) =
    (size_Tm s31))) /\
  (forall (S32 : Ty) (c15 : (Cutoff TmVar)) ,
    ((size_Ty (shiftTy c15 S32)) =
    (size_Ty S32))).
Proof.
  apply_mutual_ind ind_Tm_Ty.
  - intros x96 c15.
  reflexivity.
  - intros T95 IHT95 t79 IHt79.
  intros c16; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT95 c16)).
  + exact ((IHt79 (CS c16))).
  - intros t80 IHt80 t81 IHt81.
  intros c16; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt80 c16)).
  + exact ((IHt81 c16)).
  - intros T96 IHT96 t82 IHt82.
  intros c16; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT96 c16)).
  + exact ((IHt82 (CS c16))).
  - intros x99 c16.
  reflexivity.
  - intros T97 IHT97 T98 IHT98.
  intros c17; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT97 c17)).
  + exact ((IHT98 (CS c17))).
  - intros T99 IHT99 t83 IHt83.
  intros c17; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT99 c17)).
  + exact ((IHt83 c17)).
  - intros c17; simpl; apply ((f_equal S)); reflexivity.
  - intros c17; simpl; apply ((f_equal S)); reflexivity.
Qed.

Lemma shift_size_Tm  :
  (forall (s31 : Tm) (c17 : (Cutoff TmVar)) ,
    ((size_Tm (shiftTm c17 s31)) =
    (size_Tm s31))).
Proof.
  pose proof (shift_size_Tm_Ty).
  destruct_conjs .
  easy .
Qed.
Lemma shift_size_Ty  :
  (forall (S32 : Ty) (c17 : (Cutoff TmVar)) ,
    ((size_Ty (shiftTy c17 S32)) =
    (size_Ty S32))).
Proof.
  pose proof (shift_size_Tm_Ty).
  destruct_conjs .
  easy .
Qed.
Lemma tshift_size_Tm_Ty  :
  (forall (s31 : Tm) (c17 : (Cutoff TyVar)) ,
    ((size_Tm (tshiftTm c17 s31)) =
    (size_Tm s31))) /\
  (forall (S32 : Ty) (c17 : (Cutoff TyVar)) ,
    ((size_Ty (tshiftTy c17 S32)) =
    (size_Ty S32))).
Proof.
  apply_mutual_ind ind_Tm_Ty.
  - intros X32 c17.
  reflexivity.
  - intros T100 IHT100 t84 IHt84.
  intros c18; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT100 c18)).
  + exact ((IHt84 c18)).
  - intros t85 IHt85 t86 IHt86.
  intros c18; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt85 c18)).
  + exact ((IHt86 c18)).
  - intros T101 IHT101 t87 IHt87.
  intros c18; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT101 c18)).
  + exact ((IHt87 c18)).
  - intros X34 c18.
  reflexivity.
  - intros T102 IHT102 T103 IHT103.
  intros c19; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT102 c19)).
  + exact ((IHT103 c19)).
  - intros T104 IHT104 t88 IHt88.
  intros c19; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT104 c19)).
  + exact ((IHt88 c19)).
  - intros c19; simpl; apply ((f_equal S)); reflexivity.
  - intros c19; simpl; apply ((f_equal S)); reflexivity.
Qed.

Lemma tshift_size_Tm  :
  (forall (s31 : Tm) (c19 : (Cutoff TyVar)) ,
    ((size_Tm (tshiftTm c19 s31)) =
    (size_Tm s31))).
Proof.
  pose proof (tshift_size_Tm_Ty).
  destruct_conjs .
  easy .
Qed.
Lemma tshift_size_Ty  :
  (forall (S32 : Ty) (c19 : (Cutoff TyVar)) ,
    ((size_Ty (tshiftTy c19 S32)) =
    (size_Ty S32))).
Proof.
  pose proof (tshift_size_Tm_Ty).
  destruct_conjs .
  easy .
Qed.
Lemma shift_size_Kind  :
  (forall (K15 : Kind) (c19 : (Cutoff TmVar)) ,
    ((size_Kind (shiftKind c19 K15)) =
    (size_Kind K15))).
Proof.
  apply_mutual_ind ind_Kind.
  - intros c19; simpl; apply ((f_equal S)); reflexivity.
  - intros T105 K15 IHK15.
  pose proof ((shift_size_Ty T105)) as IHT105.
  intros c19; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT105 c19)).
  + exact ((IHK15 (CS c19))).
Qed.

Lemma tshift_size_Kind  :
  (forall (K16 : Kind) (c19 : (Cutoff TyVar)) ,
    ((size_Kind (tshiftKind c19 K16)) =
    (size_Kind K16))).
Proof.
  apply_mutual_ind ind_Kind.
  - intros c19; simpl; apply ((f_equal S)); reflexivity.
  - intros T106 K16 IHK16.
  pose proof ((tshift_size_Ty T106)) as IHT106.
  intros c19; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT106 c19)).
  + exact ((IHK16 c19)).
Qed.

 Hint Rewrite shift_size_Kind tshift_size_Kind shift_size_Tm tshift_size_Tm shift_size_Ty tshift_size_Ty : interaction.
 Hint Rewrite shift_size_Kind tshift_size_Kind shift_size_Tm tshift_size_Tm shift_size_Ty tshift_size_Ty : shift_size.
Lemma weaken_size_Kind  :
  (forall (k67 : Hvl) (K17 : Kind) ,
    ((size_Kind (weakenKind K17 k67)) =
    (size_Kind K17))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Tm  :
  (forall (k67 : Hvl) (s31 : Tm) ,
    ((size_Tm (weakenTm s31 k67)) =
    (size_Tm s31))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Ty  :
  (forall (k67 : Hvl) (S32 : Ty) ,
    ((size_Ty (weakenTy S32 k67)) =
    (size_Ty S32))).
Proof.
  needleGenericWeakenSize .
Qed.
 Hint Rewrite weaken_size_Kind weaken_size_Tm weaken_size_Ty : interaction.
 Hint Rewrite weaken_size_Kind weaken_size_Tm weaken_size_Ty : weaken_size.