Require Import Coq.Program.Equality.
Require Import Coq.Program.Tactics.
Require Export Needle.
Section Namespace.
  Inductive Namespace : Type :=
    | TmVar 
    | TyVar .
  Lemma eq_namespace_dec (a : Namespace) (b : Namespace) :
    (sumbool (a =
    b) (not (a =
    b))).
  Proof.
    decide equality .
  Defined.
End Namespace.

Section HVarlist.
  Inductive Hvl : Type :=
    | H0 
    | HS (a : Namespace) (k : Hvl).
  
  Fixpoint appendHvl (k : Hvl) (k0 : Hvl) {struct k0} : Hvl :=
    match k0 with
      | (H0) => k
      | (HS a k0) => (HS a (appendHvl k k0))
    end.
  
  Lemma appendHvl_assoc  :
    (forall (k : Hvl) (k0 : Hvl) (k1 : Hvl) ,
      ((appendHvl (appendHvl k k0) k1) =
      (appendHvl k (appendHvl k0 k1)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End HVarlist.

Section Index.
  Inductive Index (a : Namespace) : Type :=
    | I0 
    | IS : (Index a) -> (Index a).
  
  Global Arguments I0 [a] .
  Global Arguments IS [a] _ .
  
  Lemma eq_index_dec {a : Namespace} (i : (Index a)) (j : (Index a)) :
    (sumbool (i =
    j) (not (i =
    j))).
  Proof.
    decide equality .
  Qed.
  
  Fixpoint weakenIndexTmVar (x : (Index TmVar)) (k : Hvl) {struct k} : (Index TmVar) :=
    match k with
      | (H0) => x
      | (HS a k) => match a with
        | (TmVar) => (IS (weakenIndexTmVar x k))
        | _ => (weakenIndexTmVar x k)
      end
    end.
  Fixpoint weakenIndexTyVar (X : (Index TyVar)) (k : Hvl) {struct k} : (Index TyVar) :=
    match k with
      | (H0) => X
      | (HS a k) => match a with
        | (TyVar) => (IS (weakenIndexTyVar X k))
        | _ => (weakenIndexTyVar X k)
      end
    end.
  Lemma weakenIndexTmVar_append  :
    (forall (x : (Index TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTmVar (weakenIndexTmVar x k) k0) =
      (weakenIndexTmVar x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenIndexTyVar_append  :
    (forall (X : (Index TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTyVar (weakenIndexTyVar X k) k0) =
      (weakenIndexTyVar X (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End Index.

Section Terms.
  Inductive Ty : Set :=
    | TVar (X : (Index TyVar))
    | TArr (T1 : Ty) (T2 : Ty)
    | TAll (T : Ty).
  Scheme ind_Ty := Induction for Ty Sort Prop.
  
  Inductive Decls : Set :=
    | DNil 
    | DCons (t : Tm) (d : Decls)
  with Tm : Set :=
    | Var (x : (Index TmVar))
    | Abs (T : Ty) (t : Tm)
    | App (t1 : Tm) (t2 : Tm)
    | TAbs (t : Tm)
    | TApp (t : Tm) (T : Ty)
    | Let (d : Decls) (t : Tm).
  Scheme ind_Decls := Induction for Decls Sort Prop
  with ind_Tm := Induction for Tm Sort Prop.
  Combined Scheme ind_Decls_Tm from ind_Decls, ind_Tm.
End Terms.

Section Functions.
  Fixpoint bind (d : Decls) {struct d} : Hvl :=
    match d with
      | (DNil) => H0
      | (DCons t d) => (appendHvl (HS TmVar H0) (appendHvl (bind d) H0))
    end.
  Scheme ind_bind_Decls := Induction for Decls Sort Prop.
End Functions.

Section Shift.
  Inductive Cutoff (a : Namespace) : Type :=
    | C0 
    | CS : (Cutoff a) -> (Cutoff a).
  
  Global Arguments C0 {a} .
  Global Arguments CS {a} _ .
  Fixpoint weakenCutoffTmVar (c : (Cutoff TmVar)) (k : Hvl) {struct k} : (Cutoff TmVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TmVar) => (CS (weakenCutoffTmVar c k))
        | _ => (weakenCutoffTmVar c k)
      end
    end.
  Fixpoint weakenCutoffTyVar (c : (Cutoff TyVar)) (k : Hvl) {struct k} : (Cutoff TyVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TyVar) => (CS (weakenCutoffTyVar c k))
        | _ => (weakenCutoffTyVar c k)
      end
    end.
  
  Lemma weakenCutoffTmVar_append  :
    (forall (c : (Cutoff TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTmVar (weakenCutoffTmVar c k) k0) =
      (weakenCutoffTmVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenCutoffTyVar_append  :
    (forall (c : (Cutoff TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTyVar (weakenCutoffTyVar c k) k0) =
      (weakenCutoffTyVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint shift_TmVar_Index (c : (Cutoff TmVar)) (x : (Index TmVar)) {struct c} : (Index TmVar) :=
    match c with
      | (C0) => (IS x)
      | (CS c) => match x with
        | (I0) => I0
        | (IS x) => (IS (shift_TmVar_Index c x))
      end
    end.
  Fixpoint shift_TyVar_Index (c : (Cutoff TyVar)) (X : (Index TyVar)) {struct c} : (Index TyVar) :=
    match c with
      | (C0) => (IS X)
      | (CS c) => match X with
        | (I0) => I0
        | (IS X) => (IS (shift_TyVar_Index c X))
      end
    end.
  Fixpoint shift_TyVar_Ty (c : (Cutoff TyVar)) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (TVar X) => (TVar (shift_TyVar_Index c X))
      | (TArr T1 T2) => (TArr (shift_TyVar_Ty (weakenCutoffTyVar c H0) T1) (shift_TyVar_Ty (weakenCutoffTyVar c H0) T2))
      | (TAll T) => (TAll (shift_TyVar_Ty (weakenCutoffTyVar c (appendHvl (HS TyVar H0) H0)) T))
    end.
  Fixpoint shift_TmVar_Decls (c : (Cutoff TmVar)) (d : Decls) {struct d} : Decls :=
    match d with
      | (DNil) => (DNil)
      | (DCons t d0) => (DCons (shift_TmVar_Tm (weakenCutoffTmVar c H0) t) (shift_TmVar_Decls (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) d0))
    end
  with shift_TmVar_Tm (c : (Cutoff TmVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var (shift_TmVar_Index c x))
      | (Abs T t) => (Abs T (shift_TmVar_Tm (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (shift_TmVar_Tm (weakenCutoffTmVar c H0) t1) (shift_TmVar_Tm (weakenCutoffTmVar c H0) t2))
      | (TAbs t0) => (TAbs (shift_TmVar_Tm (weakenCutoffTmVar c (appendHvl (HS TyVar H0) H0)) t0))
      | (TApp t3 T0) => (TApp (shift_TmVar_Tm (weakenCutoffTmVar c H0) t3) T0)
      | (Let d t4) => (Let (shift_TmVar_Decls (weakenCutoffTmVar c H0) d) (shift_TmVar_Tm (weakenCutoffTmVar c (appendHvl (bind d) H0)) t4))
    end.
  Fixpoint shift_TyVar_Decls (c : (Cutoff TyVar)) (d : Decls) {struct d} : Decls :=
    match d with
      | (DNil) => (DNil)
      | (DCons t d0) => (DCons (shift_TyVar_Tm (weakenCutoffTyVar c H0) t) (shift_TyVar_Decls (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) d0))
    end
  with shift_TyVar_Tm (c : (Cutoff TyVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (Abs T t) => (Abs (shift_TyVar_Ty (weakenCutoffTyVar c H0) T) (shift_TyVar_Tm (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (shift_TyVar_Tm (weakenCutoffTyVar c H0) t1) (shift_TyVar_Tm (weakenCutoffTyVar c H0) t2))
      | (TAbs t0) => (TAbs (shift_TyVar_Tm (weakenCutoffTyVar c (appendHvl (HS TyVar H0) H0)) t0))
      | (TApp t3 T0) => (TApp (shift_TyVar_Tm (weakenCutoffTyVar c H0) t3) (shift_TyVar_Ty (weakenCutoffTyVar c H0) T0))
      | (Let d t4) => (Let (shift_TyVar_Decls (weakenCutoffTyVar c H0) d) (shift_TyVar_Tm (weakenCutoffTyVar c (appendHvl (bind d) H0)) t4))
    end.
End Shift.

Section Weaken.
  Fixpoint weakenTy (S0 : Ty) (k : Hvl) {struct k} : Ty :=
    match k with
      | (H0) => S0
      | (HS TmVar k) => (weakenTy S0 k)
      | (HS TyVar k) => (shift_TyVar_Ty C0 (weakenTy S0 k))
    end.
  Fixpoint weakenDecls (d : Decls) (k : Hvl) {struct k} : Decls :=
    match k with
      | (H0) => d
      | (HS TmVar k) => (shift_TmVar_Decls C0 (weakenDecls d k))
      | (HS TyVar k) => (shift_TyVar_Decls C0 (weakenDecls d k))
    end.
  Fixpoint weakenTm (s : Tm) (k : Hvl) {struct k} : Tm :=
    match k with
      | (H0) => s
      | (HS TmVar k) => (shift_TmVar_Tm C0 (weakenTm s k))
      | (HS TyVar k) => (shift_TyVar_Tm C0 (weakenTm s k))
    end.
End Weaken.

Section Subst.
  Inductive Trace (a : Namespace) : Type :=
    | X0 
    | XS (b : Namespace)
        (T : (Trace a)).
  
  Global Arguments X0 [a] .
  Global Arguments XS [a] b _ .
  Fixpoint weakenTrace {a : Namespace} (x : (Trace a)) (k : Hvl) {struct k} : (Trace a) :=
    match k with
      | (H0) => x
      | (HS b k) => (XS b (weakenTrace x k))
    end.
  Lemma weakenTrace_append (a : Namespace) :
    (forall (x : (Trace a)) (k : Hvl) (k0 : Hvl) ,
      ((weakenTrace (weakenTrace x k) k0) =
      (weakenTrace x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint subst_TmVar_Index (d : (Trace TmVar)) (s : Tm) (x : (Index TmVar)) {struct d} : Tm :=
    match d with
      | (X0) => match x with
        | (I0) => s
        | (IS x) => (Var x)
      end
      | (XS TmVar d) => match x with
        | (I0) => (Var I0)
        | (IS x) => (shift_TmVar_Tm C0 (subst_TmVar_Index d s x))
      end
      | (XS TyVar d) => (shift_TyVar_Tm C0 (subst_TmVar_Index d s x))
    end.
  Fixpoint subst_TyVar_Index (d : (Trace TyVar)) (S0 : Ty) (X : (Index TyVar)) {struct d} : Ty :=
    match d with
      | (X0) => match X with
        | (I0) => S0
        | (IS X) => (TVar X)
      end
      | (XS TmVar d) => (subst_TyVar_Index d S0 X)
      | (XS TyVar d) => match X with
        | (I0) => (TVar I0)
        | (IS X) => (shift_TyVar_Ty C0 (subst_TyVar_Index d S0 X))
      end
    end.
  Fixpoint subst_TyVar_Ty (d : (Trace TyVar)) (S0 : Ty) (S1 : Ty) {struct S1} : Ty :=
    match S1 with
      | (TVar X) => (subst_TyVar_Index d S0 X)
      | (TArr T1 T2) => (TArr (subst_TyVar_Ty d S0 T1) (subst_TyVar_Ty d S0 T2))
      | (TAll T) => (TAll (subst_TyVar_Ty (XS TyVar d) S0 T))
    end.
  Fixpoint subst_TmVar_Decls (d : (Trace TmVar)) (s : Tm) (d0 : Decls) {struct d0} : Decls :=
    match d0 with
      | (DNil) => (DNil)
      | (DCons t d1) => (DCons (subst_TmVar_Tm d s t) (subst_TmVar_Decls (XS TmVar d) s d1))
    end
  with subst_TmVar_Tm (d : (Trace TmVar)) (s : Tm) (s0 : Tm) {struct s0} : Tm :=
    match s0 with
      | (Var x) => (subst_TmVar_Index d s x)
      | (Abs T t) => (Abs T (subst_TmVar_Tm (XS TmVar d) s t))
      | (App t1 t2) => (App (subst_TmVar_Tm d s t1) (subst_TmVar_Tm d s t2))
      | (TAbs t0) => (TAbs (subst_TmVar_Tm (XS TyVar d) s t0))
      | (TApp t3 T0) => (TApp (subst_TmVar_Tm d s t3) T0)
      | (Let d0 t4) => (Let (subst_TmVar_Decls d s d0) (subst_TmVar_Tm (weakenTrace d (bind d0)) s t4))
    end.
  Fixpoint subst_TyVar_Decls (d : (Trace TyVar)) (S0 : Ty) (d0 : Decls) {struct d0} : Decls :=
    match d0 with
      | (DNil) => (DNil)
      | (DCons t d1) => (DCons (subst_TyVar_Tm d S0 t) (subst_TyVar_Decls (XS TmVar d) S0 d1))
    end
  with subst_TyVar_Tm (d : (Trace TyVar)) (S0 : Ty) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (Abs T t) => (Abs (subst_TyVar_Ty d S0 T) (subst_TyVar_Tm (XS TmVar d) S0 t))
      | (App t1 t2) => (App (subst_TyVar_Tm d S0 t1) (subst_TyVar_Tm d S0 t2))
      | (TAbs t0) => (TAbs (subst_TyVar_Tm (XS TyVar d) S0 t0))
      | (TApp t3 T0) => (TApp (subst_TyVar_Tm d S0 t3) (subst_TyVar_Ty d S0 T0))
      | (Let d0 t4) => (Let (subst_TyVar_Decls d S0 d0) (subst_TyVar_Tm (weakenTrace d (bind d0)) S0 t4))
    end.
End Subst.

Global Hint Resolve (f_equal (shift_TmVar_Decls C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TyVar_Decls C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TmVar_Tm C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TyVar_Tm C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TyVar_Ty C0)) : cong_shift0.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : interaction.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : weakencutoff_append.
 Hint Rewrite weakenTrace_append : interaction.
 Hint Rewrite weakenTrace_append : weakentrace_append.
Lemma stability_shift_TmVar__bind  :
  (forall (d : Decls) ,
    (forall (c : (Cutoff TmVar)) ,
      ((bind (shift_TmVar_Decls c d)) =
      (bind d)))).
Proof.
  apply_mutual_ind (ind_bind_Decls); simpl; intros; congruence .
Qed.
Lemma stability_shift_TyVar__bind  :
  (forall (d0 : Decls) ,
    (forall (c0 : (Cutoff TyVar)) ,
      ((bind (shift_TyVar_Decls c0 d0)) =
      (bind d0)))).
Proof.
  apply_mutual_ind (ind_bind_Decls); simpl; intros; congruence .
Qed.
 Hint Rewrite stability_shift_TmVar__bind stability_shift_TyVar__bind : interaction.
 Hint Rewrite stability_shift_TmVar__bind stability_shift_TyVar__bind : stability_shift.
Lemma stability_weaken_bind  :
  (forall (k : Hvl) (d1 : Decls) ,
    ((bind (weakenDecls d1 k)) =
    (bind d1))).
Proof.
  needleGenericStabilityWeaken .
Qed.
 Hint Rewrite stability_weaken_bind : interaction.
 Hint Rewrite stability_weaken_bind : stability_weaken.
Lemma stability_subst_TmVar__bind  :
  (forall (d1 : Decls) ,
    (forall (d2 : (Trace TmVar)) (s : Tm) ,
      ((bind (subst_TmVar_Decls d2 s d1)) =
      (bind d1)))).
Proof.
  apply_mutual_ind (ind_bind_Decls); simpl; intros; congruence .
Qed.
Lemma stability_subst_TyVar__bind  :
  (forall (d3 : Decls) ,
    (forall (d4 : (Trace TyVar)) (S0 : Ty) ,
      ((bind (subst_TyVar_Decls d4 S0 d3)) =
      (bind d3)))).
Proof.
  apply_mutual_ind (ind_bind_Decls); simpl; intros; congruence .
Qed.
 Hint Rewrite stability_subst_TmVar__bind stability_subst_TyVar__bind : interaction.
 Hint Rewrite stability_subst_TmVar__bind stability_subst_TyVar__bind : stability_subst.
Section SubstShiftReflection.
  Lemma subst_TmVar_Index0_shift_TmVar_Index0_reflection_ind (s0 : Tm) :
    (forall (k4 : Hvl) (x13 : (Index TmVar)) ,
      ((subst_TmVar_Index (weakenTrace X0 k4) s0 (shift_TmVar_Index (weakenCutoffTmVar C0 k4) x13)) =
      (Var x13))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Lemma subst_TyVar_Index0_shift_TyVar_Index0_reflection_ind (S1 : Ty) :
    (forall (k4 : Hvl) (X8 : (Index TyVar)) ,
      ((subst_TyVar_Index (weakenTrace X0 k4) S1 (shift_TyVar_Index (weakenCutoffTyVar C0 k4) X8)) =
      (TVar X8))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Definition subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind := (ind_Ty (fun (S3 : Ty) =>
    (forall (k7 : Hvl) (S4 : Ty) ,
      ((subst_TyVar_Ty (weakenTrace X0 k7) S4 (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) S3)) =
      S3))) (fun (X14 : (Index TyVar)) =>
    (fun (k7 : Hvl) (S3 : Ty) =>
      (subst_TyVar_Index0_shift_TyVar_Index0_reflection_ind S3 k7 X14))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k7 : Hvl) (S3 : Ty) =>
    (f_equal2 TArr (IHT0 k7 S3) (IHT3 k7 S3))) (fun (T : Ty) IHT0 (k7 : Hvl) (S3 : Ty) =>
    (f_equal TAll (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k7 (HS TyVar H0)) eq_refl eq_refl) (IHT0 (HS TyVar k7) S3))))).
  Definition subst_TmVar_0_shift_TmVar_0_Decls_Tm_reflection_ind := (ind_Decls_Tm (fun (d17 : Decls) =>
    (forall (k7 : Hvl) (s1 : Tm) ,
      ((subst_TmVar_Decls (weakenTrace X0 k7) s1 (shift_TmVar_Decls (weakenCutoffTmVar C0 k7) d17)) =
      d17))) (fun (s1 : Tm) =>
    (forall (k7 : Hvl) (s2 : Tm) ,
      ((subst_TmVar_Tm (weakenTrace X0 k7) s2 (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) s1)) =
      s1))) (fun (k7 : Hvl) (s1 : Tm) =>
    eq_refl) (fun (t : Tm) IHt41 (d : Decls) IHd (k7 : Hvl) (s1 : Tm) =>
    (f_equal2 DCons (IHt41 k7 s1) (eq_trans (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar X0 k7 (HS TmVar H0)) eq_refl eq_refl) (IHd (HS TmVar k7) s1)))) (fun (x20 : (Index TmVar)) =>
    (fun (k7 : Hvl) (s1 : Tm) =>
      (subst_TmVar_Index0_shift_TmVar_Index0_reflection_ind s1 k7 x20))) (fun (T : Ty) (t : Tm) IHt41 (k7 : Hvl) (s1 : Tm) =>
    (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k7 (HS TmVar H0)) eq_refl eq_refl) (IHt41 (HS TmVar k7) s1)))) (fun (t1 : Tm) IHt41 (t2 : Tm) IHt42 (k7 : Hvl) (s1 : Tm) =>
    (f_equal2 App (IHt41 k7 s1) (IHt42 k7 s1))) (fun (t : Tm) IHt41 (k7 : Hvl) (s1 : Tm) =>
    (f_equal TAbs (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k7 (HS TyVar H0)) eq_refl eq_refl) (IHt41 (HS TyVar k7) s1)))) (fun (t : Tm) IHt41 (T : Ty) (k7 : Hvl) (s1 : Tm) =>
    (f_equal2 TApp (IHt41 k7 s1) eq_refl)) (fun (d : Decls) IHd (t : Tm) IHt41 (k7 : Hvl) (s1 : Tm) =>
    (f_equal2 Let (IHd k7 s1) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenTrace_append TmVar X0 k7 (bind d))) eq_refl (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k7 (bind d)) eq_refl)) (IHt41 (appendHvl k7 (bind d)) s1))))).
  Lemma subst_TmVar_0_shift_TmVar_0_Decls_reflection_ind  :
    (forall (d17 : Decls) ,
      (forall (k7 : Hvl) (s1 : Tm) ,
        ((subst_TmVar_Decls (weakenTrace X0 k7) s1 (shift_TmVar_Decls (weakenCutoffTmVar C0 k7) d17)) =
        d17))).
  Proof.
    pose proof (subst_TmVar_0_shift_TmVar_0_Decls_Tm_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_TmVar_0_shift_TmVar_0_Tm_reflection_ind  :
    (forall (s1 : Tm) ,
      (forall (k7 : Hvl) (s2 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace X0 k7) s2 (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) s1)) =
        s1))).
  Proof.
    pose proof (subst_TmVar_0_shift_TmVar_0_Decls_Tm_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Definition subst_TyVar_0_shift_TyVar_0_Decls_Tm_reflection_ind := (ind_Decls_Tm (fun (d17 : Decls) =>
    (forall (k7 : Hvl) (S3 : Ty) ,
      ((subst_TyVar_Decls (weakenTrace X0 k7) S3 (shift_TyVar_Decls (weakenCutoffTyVar C0 k7) d17)) =
      d17))) (fun (s1 : Tm) =>
    (forall (k7 : Hvl) (S3 : Ty) ,
      ((subst_TyVar_Tm (weakenTrace X0 k7) S3 (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) s1)) =
      s1))) (fun (k7 : Hvl) (S3 : Ty) =>
    eq_refl) (fun (t : Tm) IHt41 (d : Decls) IHd (k7 : Hvl) (S3 : Ty) =>
    (f_equal2 DCons (IHt41 k7 S3) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar X0 k7 (HS TmVar H0)) eq_refl eq_refl) (IHd (HS TmVar k7) S3)))) (fun (x20 : (Index TmVar)) =>
    (fun (k7 : Hvl) (S3 : Ty) =>
      eq_refl)) (fun (T : Ty) (t : Tm) IHt41 (k7 : Hvl) (S3 : Ty) =>
    (f_equal2 Abs (let IHT0 := (subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind T) in
    (IHT0 k7 S3)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k7 (HS TmVar H0)) eq_refl eq_refl) (IHt41 (HS TmVar k7) S3)))) (fun (t1 : Tm) IHt41 (t2 : Tm) IHt42 (k7 : Hvl) (S3 : Ty) =>
    (f_equal2 App (IHt41 k7 S3) (IHt42 k7 S3))) (fun (t : Tm) IHt41 (k7 : Hvl) (S3 : Ty) =>
    (f_equal TAbs (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k7 (HS TyVar H0)) eq_refl eq_refl) (IHt41 (HS TyVar k7) S3)))) (fun (t : Tm) IHt41 (T : Ty) (k7 : Hvl) (S3 : Ty) =>
    (f_equal2 TApp (IHt41 k7 S3) (let IHT0 := (subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind T) in
    (IHT0 k7 S3)))) (fun (d : Decls) IHd (t : Tm) IHt41 (k7 : Hvl) (S3 : Ty) =>
    (f_equal2 Let (IHd k7 S3) (eq_trans (f_equal3 subst_TyVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TyVar__bind _ _) (eq_refl H0))) (weakenTrace_append TyVar X0 k7 (bind d))) eq_refl (f_equal2 shift_TyVar_Tm (weakenCutoffTyVar_append C0 k7 (bind d)) eq_refl)) (IHt41 (appendHvl k7 (bind d)) S3))))).
  Lemma subst_TyVar_0_shift_TyVar_0_Decls_reflection_ind  :
    (forall (d17 : Decls) ,
      (forall (k7 : Hvl) (S3 : Ty) ,
        ((subst_TyVar_Decls (weakenTrace X0 k7) S3 (shift_TyVar_Decls (weakenCutoffTyVar C0 k7) d17)) =
        d17))).
  Proof.
    pose proof (subst_TyVar_0_shift_TyVar_0_Decls_Tm_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_TyVar_0_shift_TyVar_0_Tm_reflection_ind  :
    (forall (s1 : Tm) ,
      (forall (k7 : Hvl) (S3 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace X0 k7) S3 (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) s1)) =
        s1))).
  Proof.
    pose proof (subst_TyVar_0_shift_TyVar_0_Decls_Tm_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Definition subst_TyVar_Ty0_shift_TyVar_Ty0_reflection (S4 : Ty) : (forall (S3 : Ty) ,
    ((subst_TyVar_Ty X0 S3 (shift_TyVar_Ty C0 S4)) =
    S4)) := (subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind S4 H0).
  Definition subst_TmVar_Decls0_shift_TmVar_Decls0_reflection (d17 : Decls) : (forall (s1 : Tm) ,
    ((subst_TmVar_Decls X0 s1 (shift_TmVar_Decls C0 d17)) =
    d17)) := (subst_TmVar_0_shift_TmVar_0_Decls_reflection_ind d17 H0).
  Definition subst_TmVar_Tm0_shift_TmVar_Tm0_reflection (s2 : Tm) : (forall (s1 : Tm) ,
    ((subst_TmVar_Tm X0 s1 (shift_TmVar_Tm C0 s2)) =
    s2)) := (subst_TmVar_0_shift_TmVar_0_Tm_reflection_ind s2 H0).
  Definition subst_TyVar_Decls0_shift_TyVar_Decls0_reflection (d17 : Decls) : (forall (S3 : Ty) ,
    ((subst_TyVar_Decls X0 S3 (shift_TyVar_Decls C0 d17)) =
    d17)) := (subst_TyVar_0_shift_TyVar_0_Decls_reflection_ind d17 H0).
  Definition subst_TyVar_Tm0_shift_TyVar_Tm0_reflection (s1 : Tm) : (forall (S3 : Ty) ,
    ((subst_TyVar_Tm X0 S3 (shift_TyVar_Tm C0 s1)) =
    s1)) := (subst_TyVar_0_shift_TyVar_0_Tm_reflection_ind s1 H0).
End SubstShiftReflection.
Section ShiftInteraction.
  Section ShiftIndexCommInd.
    Lemma shift_TmVar_Index_shift_TmVar_Index0_comm_ind  :
      (forall (k : Hvl) (c1 : (Cutoff TmVar)) (x : (Index TmVar)) ,
        ((shift_TmVar_Index (weakenCutoffTmVar (CS c1) k) (shift_TmVar_Index (weakenCutoffTmVar C0 k) x)) =
        (shift_TmVar_Index (weakenCutoffTmVar C0 k) (shift_TmVar_Index (weakenCutoffTmVar c1 k) x)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
    Lemma shift_TyVar_Index_shift_TyVar_Index0_comm_ind  :
      (forall (k : Hvl) (c1 : (Cutoff TyVar)) (X : (Index TyVar)) ,
        ((shift_TyVar_Index (weakenCutoffTyVar (CS c1) k) (shift_TyVar_Index (weakenCutoffTyVar C0 k) X)) =
        (shift_TyVar_Index (weakenCutoffTyVar C0 k) (shift_TyVar_Index (weakenCutoffTyVar c1 k) X)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
  End ShiftIndexCommInd.
  Section ShiftCommInd.
    Definition shift_TyVar__shift_TyVar_0_Ty_comm_ind := (ind_Ty (fun (S1 : Ty) =>
      (forall (k4 : Hvl) (c6 : (Cutoff TyVar)) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar (CS c6) k4) (shift_TyVar_Ty (weakenCutoffTyVar C0 k4) S1)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k4) (shift_TyVar_Ty (weakenCutoffTyVar c6 k4) S1))))) (fun (X8 : (Index TyVar)) =>
      (fun (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
        (f_equal TVar (shift_TyVar_Index_shift_TyVar_Index0_comm_ind k4 c6 X8)))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      (f_equal2 TArr (IHT0 k4 c6) (IHT3 k4 c6))) (fun (T : Ty) IHT0 (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      (f_equal TAll (IHT0 (HS TyVar k4) c6)))).
    Definition shift_TmVar__shift_TmVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d13 : Decls) =>
      (forall (k4 : Hvl) (c6 : (Cutoff TmVar)) ,
        ((shift_TmVar_Decls (weakenCutoffTmVar (CS c6) k4) (shift_TmVar_Decls (weakenCutoffTmVar C0 k4) d13)) =
        (shift_TmVar_Decls (weakenCutoffTmVar C0 k4) (shift_TmVar_Decls (weakenCutoffTmVar c6 k4) d13))))) (fun (s0 : Tm) =>
      (forall (k4 : Hvl) (c6 : (Cutoff TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar (CS c6) k4) (shift_TmVar_Tm (weakenCutoffTmVar C0 k4) s0)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k4) (shift_TmVar_Tm (weakenCutoffTmVar c6 k4) s0))))) (fun (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
      eq_refl) (fun (t : Tm) IHt27 (d : Decls) IHd (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
      (f_equal2 DCons (IHt27 k4 c6) (IHd (HS TmVar k4) c6))) (fun (x13 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
        (f_equal Var (shift_TmVar_Index_shift_TmVar_Index0_comm_ind k4 c6 x13)))) (fun (T : Ty) (t : Tm) IHt27 (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
      (f_equal2 Abs eq_refl (IHt27 (HS TmVar k4) c6))) (fun (t1 : Tm) IHt27 (t2 : Tm) IHt28 (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt27 k4 c6) (IHt28 k4 c6))) (fun (t : Tm) IHt27 (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
      (f_equal TAbs (IHt27 (HS TyVar k4) c6))) (fun (t : Tm) IHt27 (T : Ty) (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
      (f_equal2 TApp (IHt27 k4 c6) eq_refl)) (fun (d : Decls) IHd (t : Tm) IHt27 (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
      (f_equal2 Let (IHd k4 c6) (eq_trans (f_equal2 shift_TmVar_Tm (eq_trans (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenCutoffTmVar_append (CS c6) k4 (bind d))) (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k4 (bind d)) eq_refl)) (eq_trans (IHt27 (appendHvl k4 (bind d)) c6) (f_equal2 shift_TmVar_Tm (eq_trans (eq_sym (weakenCutoffTmVar_append C0 k4 (bind d))) (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0)))) (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append c6 k4 (bind d))) eq_refl))))))).
    Lemma shift_TmVar__shift_TmVar_0_Decls_comm_ind  :
      (forall (d13 : Decls) ,
        (forall (k4 : Hvl) (c6 : (Cutoff TmVar)) ,
          ((shift_TmVar_Decls (weakenCutoffTmVar (CS c6) k4) (shift_TmVar_Decls (weakenCutoffTmVar C0 k4) d13)) =
          (shift_TmVar_Decls (weakenCutoffTmVar C0 k4) (shift_TmVar_Decls (weakenCutoffTmVar c6 k4) d13))))).
    Proof.
      pose proof (shift_TmVar__shift_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__shift_TmVar_0_Tm_comm_ind  :
      (forall (s0 : Tm) ,
        (forall (k4 : Hvl) (c6 : (Cutoff TmVar)) ,
          ((shift_TmVar_Tm (weakenCutoffTmVar (CS c6) k4) (shift_TmVar_Tm (weakenCutoffTmVar C0 k4) s0)) =
          (shift_TmVar_Tm (weakenCutoffTmVar C0 k4) (shift_TmVar_Tm (weakenCutoffTmVar c6 k4) s0))))).
    Proof.
      pose proof (shift_TmVar__shift_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TmVar__shift_TyVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d13 : Decls) =>
      (forall (k4 : Hvl) (c6 : (Cutoff TmVar)) ,
        ((shift_TmVar_Decls (weakenCutoffTmVar c6 k4) (shift_TyVar_Decls (weakenCutoffTyVar C0 k4) d13)) =
        (shift_TyVar_Decls (weakenCutoffTyVar C0 k4) (shift_TmVar_Decls (weakenCutoffTmVar c6 k4) d13))))) (fun (s0 : Tm) =>
      (forall (k4 : Hvl) (c6 : (Cutoff TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c6 k4) (shift_TyVar_Tm (weakenCutoffTyVar C0 k4) s0)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k4) (shift_TmVar_Tm (weakenCutoffTmVar c6 k4) s0))))) (fun (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
      eq_refl) (fun (t : Tm) IHt27 (d : Decls) IHd (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
      (f_equal2 DCons (IHt27 k4 c6) (IHd (HS TmVar k4) c6))) (fun (x13 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt27 (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
      (f_equal2 Abs eq_refl (IHt27 (HS TmVar k4) c6))) (fun (t1 : Tm) IHt27 (t2 : Tm) IHt28 (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt27 k4 c6) (IHt28 k4 c6))) (fun (t : Tm) IHt27 (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
      (f_equal TAbs (IHt27 (HS TyVar k4) c6))) (fun (t : Tm) IHt27 (T : Ty) (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
      (f_equal2 TApp (IHt27 k4 c6) eq_refl)) (fun (d : Decls) IHd (t : Tm) IHt27 (k4 : Hvl) (c6 : (Cutoff TmVar)) =>
      (f_equal2 Let (IHd k4 c6) (eq_trans (f_equal2 shift_TmVar_Tm (eq_trans (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (stability_shift_TyVar__bind _ _) (eq_refl H0))) (weakenCutoffTmVar_append c6 k4 (bind d))) (f_equal2 shift_TyVar_Tm (weakenCutoffTyVar_append C0 k4 (bind d)) eq_refl)) (eq_trans (IHt27 (appendHvl k4 (bind d)) c6) (f_equal2 shift_TyVar_Tm (eq_trans (eq_sym (weakenCutoffTyVar_append C0 k4 (bind d))) (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0)))) (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append c6 k4 (bind d))) eq_refl))))))).
    Lemma shift_TmVar__shift_TyVar_0_Decls_comm_ind  :
      (forall (d13 : Decls) ,
        (forall (k4 : Hvl) (c6 : (Cutoff TmVar)) ,
          ((shift_TmVar_Decls (weakenCutoffTmVar c6 k4) (shift_TyVar_Decls (weakenCutoffTyVar C0 k4) d13)) =
          (shift_TyVar_Decls (weakenCutoffTyVar C0 k4) (shift_TmVar_Decls (weakenCutoffTmVar c6 k4) d13))))).
    Proof.
      pose proof (shift_TmVar__shift_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__shift_TyVar_0_Tm_comm_ind  :
      (forall (s0 : Tm) ,
        (forall (k4 : Hvl) (c6 : (Cutoff TmVar)) ,
          ((shift_TmVar_Tm (weakenCutoffTmVar c6 k4) (shift_TyVar_Tm (weakenCutoffTyVar C0 k4) s0)) =
          (shift_TyVar_Tm (weakenCutoffTyVar C0 k4) (shift_TmVar_Tm (weakenCutoffTmVar c6 k4) s0))))).
    Proof.
      pose proof (shift_TmVar__shift_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TyVar__shift_TmVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d13 : Decls) =>
      (forall (k4 : Hvl) (c6 : (Cutoff TyVar)) ,
        ((shift_TyVar_Decls (weakenCutoffTyVar c6 k4) (shift_TmVar_Decls (weakenCutoffTmVar C0 k4) d13)) =
        (shift_TmVar_Decls (weakenCutoffTmVar C0 k4) (shift_TyVar_Decls (weakenCutoffTyVar c6 k4) d13))))) (fun (s0 : Tm) =>
      (forall (k4 : Hvl) (c6 : (Cutoff TyVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c6 k4) (shift_TmVar_Tm (weakenCutoffTmVar C0 k4) s0)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k4) (shift_TyVar_Tm (weakenCutoffTyVar c6 k4) s0))))) (fun (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      eq_refl) (fun (t : Tm) IHt27 (d : Decls) IHd (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      (f_equal2 DCons (IHt27 k4 c6) (IHd (HS TmVar k4) c6))) (fun (x13 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt27 (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      (f_equal2 Abs eq_refl (IHt27 (HS TmVar k4) c6))) (fun (t1 : Tm) IHt27 (t2 : Tm) IHt28 (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt27 k4 c6) (IHt28 k4 c6))) (fun (t : Tm) IHt27 (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      (f_equal TAbs (IHt27 (HS TyVar k4) c6))) (fun (t : Tm) IHt27 (T : Ty) (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      (f_equal2 TApp (IHt27 k4 c6) eq_refl)) (fun (d : Decls) IHd (t : Tm) IHt27 (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      (f_equal2 Let (IHd k4 c6) (eq_trans (f_equal2 shift_TyVar_Tm (eq_trans (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenCutoffTyVar_append c6 k4 (bind d))) (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k4 (bind d)) eq_refl)) (eq_trans (IHt27 (appendHvl k4 (bind d)) c6) (f_equal2 shift_TmVar_Tm (eq_trans (eq_sym (weakenCutoffTmVar_append C0 k4 (bind d))) (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TyVar__bind _ _)) (eq_refl H0)))) (f_equal2 shift_TyVar_Tm (eq_sym (weakenCutoffTyVar_append c6 k4 (bind d))) eq_refl))))))).
    Lemma shift_TyVar__shift_TmVar_0_Decls_comm_ind  :
      (forall (d13 : Decls) ,
        (forall (k4 : Hvl) (c6 : (Cutoff TyVar)) ,
          ((shift_TyVar_Decls (weakenCutoffTyVar c6 k4) (shift_TmVar_Decls (weakenCutoffTmVar C0 k4) d13)) =
          (shift_TmVar_Decls (weakenCutoffTmVar C0 k4) (shift_TyVar_Decls (weakenCutoffTyVar c6 k4) d13))))).
    Proof.
      pose proof (shift_TyVar__shift_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__shift_TmVar_0_Tm_comm_ind  :
      (forall (s0 : Tm) ,
        (forall (k4 : Hvl) (c6 : (Cutoff TyVar)) ,
          ((shift_TyVar_Tm (weakenCutoffTyVar c6 k4) (shift_TmVar_Tm (weakenCutoffTmVar C0 k4) s0)) =
          (shift_TmVar_Tm (weakenCutoffTmVar C0 k4) (shift_TyVar_Tm (weakenCutoffTyVar c6 k4) s0))))).
    Proof.
      pose proof (shift_TyVar__shift_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TyVar__shift_TyVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d13 : Decls) =>
      (forall (k4 : Hvl) (c6 : (Cutoff TyVar)) ,
        ((shift_TyVar_Decls (weakenCutoffTyVar (CS c6) k4) (shift_TyVar_Decls (weakenCutoffTyVar C0 k4) d13)) =
        (shift_TyVar_Decls (weakenCutoffTyVar C0 k4) (shift_TyVar_Decls (weakenCutoffTyVar c6 k4) d13))))) (fun (s0 : Tm) =>
      (forall (k4 : Hvl) (c6 : (Cutoff TyVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar (CS c6) k4) (shift_TyVar_Tm (weakenCutoffTyVar C0 k4) s0)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k4) (shift_TyVar_Tm (weakenCutoffTyVar c6 k4) s0))))) (fun (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      eq_refl) (fun (t : Tm) IHt27 (d : Decls) IHd (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      (f_equal2 DCons (IHt27 k4 c6) (IHd (HS TmVar k4) c6))) (fun (x13 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt27 (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      (f_equal2 Abs (let IHT0 := (shift_TyVar__shift_TyVar_0_Ty_comm_ind T) in
      (IHT0 k4 c6)) (IHt27 (HS TmVar k4) c6))) (fun (t1 : Tm) IHt27 (t2 : Tm) IHt28 (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt27 k4 c6) (IHt28 k4 c6))) (fun (t : Tm) IHt27 (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      (f_equal TAbs (IHt27 (HS TyVar k4) c6))) (fun (t : Tm) IHt27 (T : Ty) (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      (f_equal2 TApp (IHt27 k4 c6) (let IHT0 := (shift_TyVar__shift_TyVar_0_Ty_comm_ind T) in
      (IHT0 k4 c6)))) (fun (d : Decls) IHd (t : Tm) IHt27 (k4 : Hvl) (c6 : (Cutoff TyVar)) =>
      (f_equal2 Let (IHd k4 c6) (eq_trans (f_equal2 shift_TyVar_Tm (eq_trans (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (stability_shift_TyVar__bind _ _) (eq_refl H0))) (weakenCutoffTyVar_append (CS c6) k4 (bind d))) (f_equal2 shift_TyVar_Tm (weakenCutoffTyVar_append C0 k4 (bind d)) eq_refl)) (eq_trans (IHt27 (appendHvl k4 (bind d)) c6) (f_equal2 shift_TyVar_Tm (eq_trans (eq_sym (weakenCutoffTyVar_append C0 k4 (bind d))) (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TyVar__bind _ _)) (eq_refl H0)))) (f_equal2 shift_TyVar_Tm (eq_sym (weakenCutoffTyVar_append c6 k4 (bind d))) eq_refl))))))).
    Lemma shift_TyVar__shift_TyVar_0_Decls_comm_ind  :
      (forall (d13 : Decls) ,
        (forall (k4 : Hvl) (c6 : (Cutoff TyVar)) ,
          ((shift_TyVar_Decls (weakenCutoffTyVar (CS c6) k4) (shift_TyVar_Decls (weakenCutoffTyVar C0 k4) d13)) =
          (shift_TyVar_Decls (weakenCutoffTyVar C0 k4) (shift_TyVar_Decls (weakenCutoffTyVar c6 k4) d13))))).
    Proof.
      pose proof (shift_TyVar__shift_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__shift_TyVar_0_Tm_comm_ind  :
      (forall (s0 : Tm) ,
        (forall (k4 : Hvl) (c6 : (Cutoff TyVar)) ,
          ((shift_TyVar_Tm (weakenCutoffTyVar (CS c6) k4) (shift_TyVar_Tm (weakenCutoffTyVar C0 k4) s0)) =
          (shift_TyVar_Tm (weakenCutoffTyVar C0 k4) (shift_TyVar_Tm (weakenCutoffTyVar c6 k4) s0))))).
    Proof.
      pose proof (shift_TyVar__shift_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End ShiftCommInd.
  Section ShiftComm.
    Definition shift_TyVar__shift_TyVar_0_Ty_comm (S1 : Ty) : (forall (c6 : (Cutoff TyVar)) ,
      ((shift_TyVar_Ty (CS c6) (shift_TyVar_Ty C0 S1)) =
      (shift_TyVar_Ty C0 (shift_TyVar_Ty c6 S1)))) := (shift_TyVar__shift_TyVar_0_Ty_comm_ind S1 H0).
    Definition shift_TmVar__shift_TmVar_0_Decls_comm (d13 : Decls) : (forall (c6 : (Cutoff TmVar)) ,
      ((shift_TmVar_Decls (CS c6) (shift_TmVar_Decls C0 d13)) =
      (shift_TmVar_Decls C0 (shift_TmVar_Decls c6 d13)))) := (shift_TmVar__shift_TmVar_0_Decls_comm_ind d13 H0).
    Definition shift_TmVar__shift_TmVar_0_Tm_comm (s0 : Tm) : (forall (c6 : (Cutoff TmVar)) ,
      ((shift_TmVar_Tm (CS c6) (shift_TmVar_Tm C0 s0)) =
      (shift_TmVar_Tm C0 (shift_TmVar_Tm c6 s0)))) := (shift_TmVar__shift_TmVar_0_Tm_comm_ind s0 H0).
    Definition shift_TmVar__shift_TyVar_0_Decls_comm (d13 : Decls) : (forall (c6 : (Cutoff TmVar)) ,
      ((shift_TmVar_Decls c6 (shift_TyVar_Decls C0 d13)) =
      (shift_TyVar_Decls C0 (shift_TmVar_Decls c6 d13)))) := (shift_TmVar__shift_TyVar_0_Decls_comm_ind d13 H0).
    Definition shift_TmVar__shift_TyVar_0_Tm_comm (s0 : Tm) : (forall (c6 : (Cutoff TmVar)) ,
      ((shift_TmVar_Tm c6 (shift_TyVar_Tm C0 s0)) =
      (shift_TyVar_Tm C0 (shift_TmVar_Tm c6 s0)))) := (shift_TmVar__shift_TyVar_0_Tm_comm_ind s0 H0).
    Definition shift_TyVar__shift_TmVar_0_Decls_comm (d13 : Decls) : (forall (c6 : (Cutoff TyVar)) ,
      ((shift_TyVar_Decls c6 (shift_TmVar_Decls C0 d13)) =
      (shift_TmVar_Decls C0 (shift_TyVar_Decls c6 d13)))) := (shift_TyVar__shift_TmVar_0_Decls_comm_ind d13 H0).
    Definition shift_TyVar__shift_TmVar_0_Tm_comm (s0 : Tm) : (forall (c6 : (Cutoff TyVar)) ,
      ((shift_TyVar_Tm c6 (shift_TmVar_Tm C0 s0)) =
      (shift_TmVar_Tm C0 (shift_TyVar_Tm c6 s0)))) := (shift_TyVar__shift_TmVar_0_Tm_comm_ind s0 H0).
    Definition shift_TyVar__shift_TyVar_0_Decls_comm (d13 : Decls) : (forall (c6 : (Cutoff TyVar)) ,
      ((shift_TyVar_Decls (CS c6) (shift_TyVar_Decls C0 d13)) =
      (shift_TyVar_Decls C0 (shift_TyVar_Decls c6 d13)))) := (shift_TyVar__shift_TyVar_0_Decls_comm_ind d13 H0).
    Definition shift_TyVar__shift_TyVar_0_Tm_comm (s0 : Tm) : (forall (c6 : (Cutoff TyVar)) ,
      ((shift_TyVar_Tm (CS c6) (shift_TyVar_Tm C0 s0)) =
      (shift_TyVar_Tm C0 (shift_TyVar_Tm c6 s0)))) := (shift_TyVar__shift_TyVar_0_Tm_comm_ind s0 H0).
  End ShiftComm.
End ShiftInteraction.
 Hint Rewrite shift_TmVar__shift_TmVar_0_Decls_comm shift_TmVar__shift_TyVar_0_Decls_comm shift_TyVar__shift_TmVar_0_Decls_comm shift_TyVar__shift_TyVar_0_Decls_comm shift_TmVar__shift_TmVar_0_Tm_comm shift_TmVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TmVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Ty_comm : interaction.
 Hint Rewrite shift_TmVar__shift_TmVar_0_Decls_comm shift_TmVar__shift_TyVar_0_Decls_comm shift_TyVar__shift_TmVar_0_Decls_comm shift_TyVar__shift_TyVar_0_Decls_comm shift_TmVar__shift_TmVar_0_Tm_comm shift_TmVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TmVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Ty_comm : shift_shift0.
Section WeakenShift.
  Lemma weakenTy_shift_TyVar_Ty  :
    (forall (k61 : Hvl) (c11 : (Cutoff TyVar)) (S40 : Ty) ,
      ((weakenTy (shift_TyVar_Ty c11 S40) k61) =
      (shift_TyVar_Ty (weakenCutoffTyVar c11 k61) (weakenTy S40 k61)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenDecls_shift_TmVar_Decls  :
    (forall (k61 : Hvl) (c11 : (Cutoff TmVar)) (d88 : Decls) ,
      ((weakenDecls (shift_TmVar_Decls c11 d88) k61) =
      (shift_TmVar_Decls (weakenCutoffTmVar c11 k61) (weakenDecls d88 k61)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_shift_TmVar_Tm  :
    (forall (k61 : Hvl) (c11 : (Cutoff TmVar)) (s21 : Tm) ,
      ((weakenTm (shift_TmVar_Tm c11 s21) k61) =
      (shift_TmVar_Tm (weakenCutoffTmVar c11 k61) (weakenTm s21 k61)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenDecls_shift_TyVar_Decls  :
    (forall (k61 : Hvl) (c11 : (Cutoff TyVar)) (d88 : Decls) ,
      ((weakenDecls (shift_TyVar_Decls c11 d88) k61) =
      (shift_TyVar_Decls (weakenCutoffTyVar c11 k61) (weakenDecls d88 k61)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_shift_TyVar_Tm  :
    (forall (k61 : Hvl) (c11 : (Cutoff TyVar)) (s21 : Tm) ,
      ((weakenTm (shift_TyVar_Tm c11 s21) k61) =
      (shift_TyVar_Tm (weakenCutoffTyVar c11 k61) (weakenTm s21 k61)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
End WeakenShift.
Section ShiftSubstInteraction.
  Section ShiftSubstIndexCommInd.
    Lemma shift_TmVar_Tm_subst_TmVar_Index0_comm_ind (c6 : (Cutoff TmVar)) (s1 : Tm) :
      (forall (k7 : Hvl) (x20 : (Index TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c6 k7) (subst_TmVar_Index (weakenTrace X0 k7) s1 x20)) =
        (subst_TmVar_Index (weakenTrace X0 k7) (shift_TmVar_Tm c6 s1) (shift_TmVar_Index (weakenCutoffTmVar (CS c6) k7) x20)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TyVar_Tm_subst_TmVar_Index0_comm_ind (c6 : (Cutoff TyVar)) (s1 : Tm) :
      (forall (k7 : Hvl) (x20 : (Index TmVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c6 k7) (subst_TmVar_Index (weakenTrace X0 k7) s1 x20)) =
        (subst_TmVar_Index (weakenTrace X0 k7) (shift_TyVar_Tm c6 s1) x20))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TyVar_Ty_subst_TyVar_Index0_comm_ind (c6 : (Cutoff TyVar)) (S3 : Ty) :
      (forall (k7 : Hvl) (X14 : (Index TyVar)) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar c6 k7) (subst_TyVar_Index (weakenTrace X0 k7) S3 X14)) =
        (subst_TyVar_Index (weakenTrace X0 k7) (shift_TyVar_Ty c6 S3) (shift_TyVar_Index (weakenCutoffTyVar (CS c6) k7) X14)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End ShiftSubstIndexCommInd.
  Section ShiftSubstCommInd.
    Definition shift_TyVar__subst_TyVar_0_Ty_comm_ind := (ind_Ty (fun (S6 : Ty) =>
      (forall (k12 : Hvl) (c11 : (Cutoff TyVar)) (S7 : Ty) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar c11 k12) (subst_TyVar_Ty (weakenTrace X0 k12) S7 S6)) =
        (subst_TyVar_Ty (weakenTrace X0 k12) (shift_TyVar_Ty c11 S7) (shift_TyVar_Ty (weakenCutoffTyVar (CS c11) k12) S6))))) (fun (X21 : (Index TyVar)) =>
      (fun (k12 : Hvl) (c11 : (Cutoff TyVar)) (S6 : Ty) =>
        (shift_TyVar_Ty_subst_TyVar_Index0_comm_ind c11 S6 k12 X21))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k12 : Hvl) (c11 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 TArr (IHT0 k12 c11 S6) (IHT3 k12 c11 S6))) (fun (T : Ty) IHT0 (k12 : Hvl) (c11 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal TAll (eq_trans (f_equal2 shift_TyVar_Ty eq_refl (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k12 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHT0 (HS TyVar k12) c11 S6) (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar X0 k12 (HS TyVar H0))) eq_refl eq_refl)))))).
    Definition shift_TmVar__subst_TmVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d25 : Decls) =>
      (forall (k12 : Hvl) (c11 : (Cutoff TmVar)) (s3 : Tm) ,
        ((shift_TmVar_Decls (weakenCutoffTmVar c11 k12) (subst_TmVar_Decls (weakenTrace X0 k12) s3 d25)) =
        (subst_TmVar_Decls (weakenTrace X0 k12) (shift_TmVar_Tm c11 s3) (shift_TmVar_Decls (weakenCutoffTmVar (CS c11) k12) d25))))) (fun (s3 : Tm) =>
      (forall (k12 : Hvl) (c11 : (Cutoff TmVar)) (s4 : Tm) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c11 k12) (subst_TmVar_Tm (weakenTrace X0 k12) s4 s3)) =
        (subst_TmVar_Tm (weakenTrace X0 k12) (shift_TmVar_Tm c11 s4) (shift_TmVar_Tm (weakenCutoffTmVar (CS c11) k12) s3))))) (fun (k12 : Hvl) (c11 : (Cutoff TmVar)) (s3 : Tm) =>
      eq_refl) (fun (t : Tm) IHt69 (d : Decls) IHd (k12 : Hvl) (c11 : (Cutoff TmVar)) (s3 : Tm) =>
      (f_equal2 DCons (IHt69 k12 c11 s3) (eq_trans (f_equal2 shift_TmVar_Decls eq_refl (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k12) c11 s3) (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (x36 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c11 : (Cutoff TmVar)) (s3 : Tm) =>
        (shift_TmVar_Tm_subst_TmVar_Index0_comm_ind c11 s3 k12 x36))) (fun (T : Ty) (t : Tm) IHt69 (k12 : Hvl) (c11 : (Cutoff TmVar)) (s3 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt69 (HS TmVar k12) c11 s3) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k12 : Hvl) (c11 : (Cutoff TmVar)) (s3 : Tm) =>
      (f_equal2 App (IHt69 k12 c11 s3) (IHt70 k12 c11 s3))) (fun (t : Tm) IHt69 (k12 : Hvl) (c11 : (Cutoff TmVar)) (s3 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k12 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt69 (HS TyVar k12) c11 s3) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt69 (T : Ty) (k12 : Hvl) (c11 : (Cutoff TmVar)) (s3 : Tm) =>
      (f_equal2 TApp (IHt69 k12 c11 s3) eq_refl)) (fun (d : Decls) IHd (t : Tm) IHt69 (k12 : Hvl) (c11 : (Cutoff TmVar)) (s3 : Tm) =>
      (f_equal2 Let (IHd k12 c11 s3) (eq_trans (f_equal2 shift_TmVar_Tm (eq_trans (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (stability_subst_TmVar__bind _ _ _) (eq_refl H0))) (weakenCutoffTmVar_append c11 k12 (bind d))) (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k12 (bind d)) eq_refl eq_refl)) (eq_trans (IHt69 (appendHvl k12 (bind d)) c11 s3) (f_equal3 subst_TmVar_Tm (eq_trans (eq_sym (weakenTrace_append TmVar X0 k12 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0)))) eq_refl (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append (CS c11) k12 (bind d))) eq_refl))))))).
    Lemma shift_TmVar__subst_TmVar_0_Decls_comm_ind  :
      (forall (d25 : Decls) ,
        (forall (k12 : Hvl) (c11 : (Cutoff TmVar)) (s3 : Tm) ,
          ((shift_TmVar_Decls (weakenCutoffTmVar c11 k12) (subst_TmVar_Decls (weakenTrace X0 k12) s3 d25)) =
          (subst_TmVar_Decls (weakenTrace X0 k12) (shift_TmVar_Tm c11 s3) (shift_TmVar_Decls (weakenCutoffTmVar (CS c11) k12) d25))))).
    Proof.
      pose proof (shift_TmVar__subst_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__subst_TmVar_0_Tm_comm_ind  :
      (forall (s3 : Tm) ,
        (forall (k12 : Hvl) (c11 : (Cutoff TmVar)) (s4 : Tm) ,
          ((shift_TmVar_Tm (weakenCutoffTmVar c11 k12) (subst_TmVar_Tm (weakenTrace X0 k12) s4 s3)) =
          (subst_TmVar_Tm (weakenTrace X0 k12) (shift_TmVar_Tm c11 s4) (shift_TmVar_Tm (weakenCutoffTmVar (CS c11) k12) s3))))).
    Proof.
      pose proof (shift_TmVar__subst_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TmVar__subst_TyVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d25 : Decls) =>
      (forall (k12 : Hvl) (c11 : (Cutoff TmVar)) (S6 : Ty) ,
        ((shift_TmVar_Decls (weakenCutoffTmVar c11 k12) (subst_TyVar_Decls (weakenTrace X0 k12) S6 d25)) =
        (subst_TyVar_Decls (weakenTrace X0 k12) S6 (shift_TmVar_Decls (weakenCutoffTmVar c11 k12) d25))))) (fun (s3 : Tm) =>
      (forall (k12 : Hvl) (c11 : (Cutoff TmVar)) (S6 : Ty) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c11 k12) (subst_TyVar_Tm (weakenTrace X0 k12) S6 s3)) =
        (subst_TyVar_Tm (weakenTrace X0 k12) S6 (shift_TmVar_Tm (weakenCutoffTmVar c11 k12) s3))))) (fun (k12 : Hvl) (c11 : (Cutoff TmVar)) (S6 : Ty) =>
      eq_refl) (fun (t : Tm) IHt69 (d : Decls) IHd (k12 : Hvl) (c11 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 DCons (IHt69 k12 c11 S6) (eq_trans (f_equal2 shift_TmVar_Decls eq_refl (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k12) c11 S6) (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (x36 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c11 : (Cutoff TmVar)) (S6 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt69 (k12 : Hvl) (c11 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt69 (HS TmVar k12) c11 S6) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k12 : Hvl) (c11 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 App (IHt69 k12 c11 S6) (IHt70 k12 c11 S6))) (fun (t : Tm) IHt69 (k12 : Hvl) (c11 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k12 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt69 (HS TyVar k12) c11 S6) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt69 (T : Ty) (k12 : Hvl) (c11 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 TApp (IHt69 k12 c11 S6) eq_refl)) (fun (d : Decls) IHd (t : Tm) IHt69 (k12 : Hvl) (c11 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 Let (IHd k12 c11 S6) (eq_trans (f_equal2 shift_TmVar_Tm (eq_trans (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (stability_subst_TyVar__bind _ _ _) (eq_refl H0))) (weakenCutoffTmVar_append c11 k12 (bind d))) (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k12 (bind d)) eq_refl eq_refl)) (eq_trans (IHt69 (appendHvl k12 (bind d)) c11 S6) (f_equal3 subst_TyVar_Tm (eq_trans (eq_sym (weakenTrace_append TyVar X0 k12 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0)))) eq_refl (f_equal2 shift_TmVar_Tm (eq_sym (weakenCutoffTmVar_append c11 k12 (bind d))) eq_refl))))))).
    Lemma shift_TmVar__subst_TyVar_0_Decls_comm_ind  :
      (forall (d25 : Decls) ,
        (forall (k12 : Hvl) (c11 : (Cutoff TmVar)) (S6 : Ty) ,
          ((shift_TmVar_Decls (weakenCutoffTmVar c11 k12) (subst_TyVar_Decls (weakenTrace X0 k12) S6 d25)) =
          (subst_TyVar_Decls (weakenTrace X0 k12) S6 (shift_TmVar_Decls (weakenCutoffTmVar c11 k12) d25))))).
    Proof.
      pose proof (shift_TmVar__subst_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__subst_TyVar_0_Tm_comm_ind  :
      (forall (s3 : Tm) ,
        (forall (k12 : Hvl) (c11 : (Cutoff TmVar)) (S6 : Ty) ,
          ((shift_TmVar_Tm (weakenCutoffTmVar c11 k12) (subst_TyVar_Tm (weakenTrace X0 k12) S6 s3)) =
          (subst_TyVar_Tm (weakenTrace X0 k12) S6 (shift_TmVar_Tm (weakenCutoffTmVar c11 k12) s3))))).
    Proof.
      pose proof (shift_TmVar__subst_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TyVar__subst_TmVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d25 : Decls) =>
      (forall (k12 : Hvl) (c11 : (Cutoff TyVar)) (s3 : Tm) ,
        ((shift_TyVar_Decls (weakenCutoffTyVar c11 k12) (subst_TmVar_Decls (weakenTrace X0 k12) s3 d25)) =
        (subst_TmVar_Decls (weakenTrace X0 k12) (shift_TyVar_Tm c11 s3) (shift_TyVar_Decls (weakenCutoffTyVar c11 k12) d25))))) (fun (s3 : Tm) =>
      (forall (k12 : Hvl) (c11 : (Cutoff TyVar)) (s4 : Tm) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c11 k12) (subst_TmVar_Tm (weakenTrace X0 k12) s4 s3)) =
        (subst_TmVar_Tm (weakenTrace X0 k12) (shift_TyVar_Tm c11 s4) (shift_TyVar_Tm (weakenCutoffTyVar c11 k12) s3))))) (fun (k12 : Hvl) (c11 : (Cutoff TyVar)) (s3 : Tm) =>
      eq_refl) (fun (t : Tm) IHt69 (d : Decls) IHd (k12 : Hvl) (c11 : (Cutoff TyVar)) (s3 : Tm) =>
      (f_equal2 DCons (IHt69 k12 c11 s3) (eq_trans (f_equal2 shift_TyVar_Decls eq_refl (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k12) c11 s3) (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (x36 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c11 : (Cutoff TyVar)) (s3 : Tm) =>
        (shift_TyVar_Tm_subst_TmVar_Index0_comm_ind c11 s3 k12 x36))) (fun (T : Ty) (t : Tm) IHt69 (k12 : Hvl) (c11 : (Cutoff TyVar)) (s3 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt69 (HS TmVar k12) c11 s3) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k12 : Hvl) (c11 : (Cutoff TyVar)) (s3 : Tm) =>
      (f_equal2 App (IHt69 k12 c11 s3) (IHt70 k12 c11 s3))) (fun (t : Tm) IHt69 (k12 : Hvl) (c11 : (Cutoff TyVar)) (s3 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k12 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt69 (HS TyVar k12) c11 s3) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt69 (T : Ty) (k12 : Hvl) (c11 : (Cutoff TyVar)) (s3 : Tm) =>
      (f_equal2 TApp (IHt69 k12 c11 s3) eq_refl)) (fun (d : Decls) IHd (t : Tm) IHt69 (k12 : Hvl) (c11 : (Cutoff TyVar)) (s3 : Tm) =>
      (f_equal2 Let (IHd k12 c11 s3) (eq_trans (f_equal2 shift_TyVar_Tm (eq_trans (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (stability_subst_TmVar__bind _ _ _) (eq_refl H0))) (weakenCutoffTyVar_append c11 k12 (bind d))) (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k12 (bind d)) eq_refl eq_refl)) (eq_trans (IHt69 (appendHvl k12 (bind d)) c11 s3) (f_equal3 subst_TmVar_Tm (eq_trans (eq_sym (weakenTrace_append TmVar X0 k12 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TyVar__bind _ _)) (eq_refl H0)))) eq_refl (f_equal2 shift_TyVar_Tm (eq_sym (weakenCutoffTyVar_append c11 k12 (bind d))) eq_refl))))))).
    Lemma shift_TyVar__subst_TmVar_0_Decls_comm_ind  :
      (forall (d25 : Decls) ,
        (forall (k12 : Hvl) (c11 : (Cutoff TyVar)) (s3 : Tm) ,
          ((shift_TyVar_Decls (weakenCutoffTyVar c11 k12) (subst_TmVar_Decls (weakenTrace X0 k12) s3 d25)) =
          (subst_TmVar_Decls (weakenTrace X0 k12) (shift_TyVar_Tm c11 s3) (shift_TyVar_Decls (weakenCutoffTyVar c11 k12) d25))))).
    Proof.
      pose proof (shift_TyVar__subst_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__subst_TmVar_0_Tm_comm_ind  :
      (forall (s3 : Tm) ,
        (forall (k12 : Hvl) (c11 : (Cutoff TyVar)) (s4 : Tm) ,
          ((shift_TyVar_Tm (weakenCutoffTyVar c11 k12) (subst_TmVar_Tm (weakenTrace X0 k12) s4 s3)) =
          (subst_TmVar_Tm (weakenTrace X0 k12) (shift_TyVar_Tm c11 s4) (shift_TyVar_Tm (weakenCutoffTyVar c11 k12) s3))))).
    Proof.
      pose proof (shift_TyVar__subst_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TyVar__subst_TyVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d25 : Decls) =>
      (forall (k12 : Hvl) (c11 : (Cutoff TyVar)) (S6 : Ty) ,
        ((shift_TyVar_Decls (weakenCutoffTyVar c11 k12) (subst_TyVar_Decls (weakenTrace X0 k12) S6 d25)) =
        (subst_TyVar_Decls (weakenTrace X0 k12) (shift_TyVar_Ty c11 S6) (shift_TyVar_Decls (weakenCutoffTyVar (CS c11) k12) d25))))) (fun (s3 : Tm) =>
      (forall (k12 : Hvl) (c11 : (Cutoff TyVar)) (S6 : Ty) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c11 k12) (subst_TyVar_Tm (weakenTrace X0 k12) S6 s3)) =
        (subst_TyVar_Tm (weakenTrace X0 k12) (shift_TyVar_Ty c11 S6) (shift_TyVar_Tm (weakenCutoffTyVar (CS c11) k12) s3))))) (fun (k12 : Hvl) (c11 : (Cutoff TyVar)) (S6 : Ty) =>
      eq_refl) (fun (t : Tm) IHt69 (d : Decls) IHd (k12 : Hvl) (c11 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 DCons (IHt69 k12 c11 S6) (eq_trans (f_equal2 shift_TyVar_Decls eq_refl (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k12) c11 S6) (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (x36 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c11 : (Cutoff TyVar)) (S6 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt69 (k12 : Hvl) (c11 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 Abs (let IHT0 := (shift_TyVar__subst_TyVar_0_Ty_comm_ind T) in
      (IHT0 k12 c11 S6)) (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt69 (HS TmVar k12) c11 S6) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt69 (t2 : Tm) IHt70 (k12 : Hvl) (c11 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 App (IHt69 k12 c11 S6) (IHt70 k12 c11 S6))) (fun (t : Tm) IHt69 (k12 : Hvl) (c11 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k12 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt69 (HS TyVar k12) c11 S6) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt69 (T : Ty) (k12 : Hvl) (c11 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 TApp (IHt69 k12 c11 S6) (let IHT0 := (shift_TyVar__subst_TyVar_0_Ty_comm_ind T) in
      (IHT0 k12 c11 S6)))) (fun (d : Decls) IHd (t : Tm) IHt69 (k12 : Hvl) (c11 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 Let (IHd k12 c11 S6) (eq_trans (f_equal2 shift_TyVar_Tm (eq_trans (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (stability_subst_TyVar__bind _ _ _) (eq_refl H0))) (weakenCutoffTyVar_append c11 k12 (bind d))) (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k12 (bind d)) eq_refl eq_refl)) (eq_trans (IHt69 (appendHvl k12 (bind d)) c11 S6) (f_equal3 subst_TyVar_Tm (eq_trans (eq_sym (weakenTrace_append TyVar X0 k12 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_shift_TyVar__bind _ _)) (eq_refl H0)))) eq_refl (f_equal2 shift_TyVar_Tm (eq_sym (weakenCutoffTyVar_append (CS c11) k12 (bind d))) eq_refl))))))).
    Lemma shift_TyVar__subst_TyVar_0_Decls_comm_ind  :
      (forall (d25 : Decls) ,
        (forall (k12 : Hvl) (c11 : (Cutoff TyVar)) (S6 : Ty) ,
          ((shift_TyVar_Decls (weakenCutoffTyVar c11 k12) (subst_TyVar_Decls (weakenTrace X0 k12) S6 d25)) =
          (subst_TyVar_Decls (weakenTrace X0 k12) (shift_TyVar_Ty c11 S6) (shift_TyVar_Decls (weakenCutoffTyVar (CS c11) k12) d25))))).
    Proof.
      pose proof (shift_TyVar__subst_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__subst_TyVar_0_Tm_comm_ind  :
      (forall (s3 : Tm) ,
        (forall (k12 : Hvl) (c11 : (Cutoff TyVar)) (S6 : Ty) ,
          ((shift_TyVar_Tm (weakenCutoffTyVar c11 k12) (subst_TyVar_Tm (weakenTrace X0 k12) S6 s3)) =
          (subst_TyVar_Tm (weakenTrace X0 k12) (shift_TyVar_Ty c11 S6) (shift_TyVar_Tm (weakenCutoffTyVar (CS c11) k12) s3))))).
    Proof.
      pose proof (shift_TyVar__subst_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End ShiftSubstCommInd.
  Section ShiftSubstComm.
    Definition shift_TyVar_Ty_subst_TyVar_Ty0_comm (S7 : Ty) : (forall (c11 : (Cutoff TyVar)) (S6 : Ty) ,
      ((shift_TyVar_Ty c11 (subst_TyVar_Ty X0 S6 S7)) =
      (subst_TyVar_Ty X0 (shift_TyVar_Ty c11 S6) (shift_TyVar_Ty (CS c11) S7)))) := (shift_TyVar__subst_TyVar_0_Ty_comm_ind S7 H0).
    Definition shift_TmVar_Decls_subst_TmVar_Decls0_comm (d25 : Decls) : (forall (c11 : (Cutoff TmVar)) (s3 : Tm) ,
      ((shift_TmVar_Decls c11 (subst_TmVar_Decls X0 s3 d25)) =
      (subst_TmVar_Decls X0 (shift_TmVar_Tm c11 s3) (shift_TmVar_Decls (CS c11) d25)))) := (shift_TmVar__subst_TmVar_0_Decls_comm_ind d25 H0).
    Definition shift_TmVar_Tm_subst_TmVar_Tm0_comm (s4 : Tm) : (forall (c11 : (Cutoff TmVar)) (s3 : Tm) ,
      ((shift_TmVar_Tm c11 (subst_TmVar_Tm X0 s3 s4)) =
      (subst_TmVar_Tm X0 (shift_TmVar_Tm c11 s3) (shift_TmVar_Tm (CS c11) s4)))) := (shift_TmVar__subst_TmVar_0_Tm_comm_ind s4 H0).
    Definition shift_TmVar_Decls_subst_TyVar_Decls0_comm (d25 : Decls) : (forall (c11 : (Cutoff TmVar)) (S6 : Ty) ,
      ((shift_TmVar_Decls c11 (subst_TyVar_Decls X0 S6 d25)) =
      (subst_TyVar_Decls X0 S6 (shift_TmVar_Decls c11 d25)))) := (shift_TmVar__subst_TyVar_0_Decls_comm_ind d25 H0).
    Definition shift_TmVar_Tm_subst_TyVar_Tm0_comm (s3 : Tm) : (forall (c11 : (Cutoff TmVar)) (S6 : Ty) ,
      ((shift_TmVar_Tm c11 (subst_TyVar_Tm X0 S6 s3)) =
      (subst_TyVar_Tm X0 S6 (shift_TmVar_Tm c11 s3)))) := (shift_TmVar__subst_TyVar_0_Tm_comm_ind s3 H0).
    Definition shift_TyVar_Decls_subst_TmVar_Decls0_comm (d25 : Decls) : (forall (c11 : (Cutoff TyVar)) (s3 : Tm) ,
      ((shift_TyVar_Decls c11 (subst_TmVar_Decls X0 s3 d25)) =
      (subst_TmVar_Decls X0 (shift_TyVar_Tm c11 s3) (shift_TyVar_Decls c11 d25)))) := (shift_TyVar__subst_TmVar_0_Decls_comm_ind d25 H0).
    Definition shift_TyVar_Tm_subst_TmVar_Tm0_comm (s4 : Tm) : (forall (c11 : (Cutoff TyVar)) (s3 : Tm) ,
      ((shift_TyVar_Tm c11 (subst_TmVar_Tm X0 s3 s4)) =
      (subst_TmVar_Tm X0 (shift_TyVar_Tm c11 s3) (shift_TyVar_Tm c11 s4)))) := (shift_TyVar__subst_TmVar_0_Tm_comm_ind s4 H0).
    Definition shift_TyVar_Decls_subst_TyVar_Decls0_comm (d25 : Decls) : (forall (c11 : (Cutoff TyVar)) (S6 : Ty) ,
      ((shift_TyVar_Decls c11 (subst_TyVar_Decls X0 S6 d25)) =
      (subst_TyVar_Decls X0 (shift_TyVar_Ty c11 S6) (shift_TyVar_Decls (CS c11) d25)))) := (shift_TyVar__subst_TyVar_0_Decls_comm_ind d25 H0).
    Definition shift_TyVar_Tm_subst_TyVar_Tm0_comm (s3 : Tm) : (forall (c11 : (Cutoff TyVar)) (S6 : Ty) ,
      ((shift_TyVar_Tm c11 (subst_TyVar_Tm X0 S6 s3)) =
      (subst_TyVar_Tm X0 (shift_TyVar_Ty c11 S6) (shift_TyVar_Tm (CS c11) s3)))) := (shift_TyVar__subst_TyVar_0_Tm_comm_ind s3 H0).
  End ShiftSubstComm.
  Section SubstShiftIndexCommInd.
    Lemma subst_TmVar_Index_shift_TmVar_Tm0_comm_ind (d25 : (Trace TmVar)) (s3 : Tm) :
      (forall (k12 : Hvl) (x36 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace (XS TmVar d25) k12) s3 (shift_TmVar_Index (weakenCutoffTmVar C0 k12) x36)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k12) (subst_TmVar_Index (weakenTrace d25 k12) s3 x36)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Index_shift_TyVar_Tm0_comm_ind (d25 : (Trace TmVar)) (s3 : Tm) :
      (forall (k12 : Hvl) (x36 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace (XS TyVar d25) k12) s3 x36) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k12) (subst_TmVar_Index (weakenTrace d25 k12) s3 x36)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Index_shift_TmVar_Ty0_comm_ind (d25 : (Trace TyVar)) (S6 : Ty) :
      (forall (k12 : Hvl) (X21 : (Index TyVar)) ,
        ((subst_TyVar_Index (weakenTrace (XS TmVar d25) k12) S6 X21) =
        (subst_TyVar_Index (weakenTrace d25 k12) S6 X21))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Index_shift_TyVar_Ty0_comm_ind (d25 : (Trace TyVar)) (S6 : Ty) :
      (forall (k12 : Hvl) (X21 : (Index TyVar)) ,
        ((subst_TyVar_Index (weakenTrace (XS TyVar d25) k12) S6 (shift_TyVar_Index (weakenCutoffTyVar C0 k12) X21)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k12) (subst_TyVar_Index (weakenTrace d25 k12) S6 X21)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End SubstShiftIndexCommInd.
  Section SubstShiftCommInd.
    Definition subst_TyVar__shift_TyVar_0_Ty_comm_ind := (ind_Ty (fun (S25 : Ty) =>
      (forall (k47 : Hvl) (d68 : (Trace TyVar)) (S26 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace (XS TyVar d68) k47) S26 (shift_TyVar_Ty (weakenCutoffTyVar C0 k47) S25)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k47) (subst_TyVar_Ty (weakenTrace d68 k47) S26 S25))))) (fun (X28 : (Index TyVar)) =>
      (fun (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
        (subst_TyVar_Index_shift_TyVar_Ty0_comm_ind d68 S25 k47 X28))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      (f_equal2 TArr (IHT0 k47 d68 S25) (IHT3 k47 d68 S25))) (fun (T : Ty) IHT0 (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      (f_equal TAll (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar (XS TyVar d68) k47 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHT0 (HS TyVar k47) d68 S25) (f_equal2 shift_TyVar_Ty eq_refl (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar d68 k47 (HS TyVar H0))) eq_refl eq_refl))))))).
    Definition subst_TmVar__shift_TmVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d68 : Decls) =>
      (forall (k47 : Hvl) (d69 : (Trace TmVar)) (s19 : Tm) ,
        ((subst_TmVar_Decls (weakenTrace (XS TmVar d69) k47) s19 (shift_TmVar_Decls (weakenCutoffTmVar C0 k47) d68)) =
        (shift_TmVar_Decls (weakenCutoffTmVar C0 k47) (subst_TmVar_Decls (weakenTrace d69 k47) s19 d68))))) (fun (s19 : Tm) =>
      (forall (k47 : Hvl) (d68 : (Trace TmVar)) (s20 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace (XS TmVar d68) k47) s20 (shift_TmVar_Tm (weakenCutoffTmVar C0 k47) s19)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k47) (subst_TmVar_Tm (weakenTrace d68 k47) s20 s19))))) (fun (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
      eq_refl) (fun (t : Tm) IHt97 (d : Decls) IHd (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
      (f_equal2 DCons (IHt97 k47 d68 s19) (eq_trans (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar (XS TmVar d68) k47 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHd (HS TmVar k47) d68 s19) (f_equal2 shift_TmVar_Decls eq_refl (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar d68 k47 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (x52 : (Index TmVar)) =>
      (fun (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
        (subst_TmVar_Index_shift_TmVar_Tm0_comm_ind d68 s19 k47 x52))) (fun (T : Ty) (t : Tm) IHt97 (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TmVar d68) k47 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt97 (HS TmVar k47) d68 s19) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d68 k47 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt97 (t2 : Tm) IHt98 (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
      (f_equal2 App (IHt97 k47 d68 s19) (IHt98 k47 d68 s19))) (fun (t : Tm) IHt97 (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TmVar d68) k47 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt97 (HS TyVar k47) d68 s19) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d68 k47 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt97 (T : Ty) (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
      (f_equal2 TApp (IHt97 k47 d68 s19) eq_refl)) (fun (d : Decls) IHd (t : Tm) IHt97 (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
      (f_equal2 Let (IHd k47 d68 s19) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenTrace_append TmVar (XS TmVar d68) k47 (bind d))) eq_refl (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k47 (bind d)) eq_refl)) (eq_trans (IHt97 (appendHvl k47 (bind d)) d68 s19) (f_equal2 shift_TmVar_Tm (eq_trans (eq_sym (weakenCutoffTmVar_append C0 k47 (bind d))) (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0)))) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d68 k47 (bind d))) eq_refl eq_refl))))))).
    Lemma subst_TmVar__shift_TmVar_0_Decls_comm_ind  :
      (forall (d68 : Decls) ,
        (forall (k47 : Hvl) (d69 : (Trace TmVar)) (s19 : Tm) ,
          ((subst_TmVar_Decls (weakenTrace (XS TmVar d69) k47) s19 (shift_TmVar_Decls (weakenCutoffTmVar C0 k47) d68)) =
          (shift_TmVar_Decls (weakenCutoffTmVar C0 k47) (subst_TmVar_Decls (weakenTrace d69 k47) s19 d68))))).
    Proof.
      pose proof (subst_TmVar__shift_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__shift_TmVar_0_Tm_comm_ind  :
      (forall (s19 : Tm) ,
        (forall (k47 : Hvl) (d69 : (Trace TmVar)) (s20 : Tm) ,
          ((subst_TmVar_Tm (weakenTrace (XS TmVar d69) k47) s20 (shift_TmVar_Tm (weakenCutoffTmVar C0 k47) s19)) =
          (shift_TmVar_Tm (weakenCutoffTmVar C0 k47) (subst_TmVar_Tm (weakenTrace d69 k47) s20 s19))))).
    Proof.
      pose proof (subst_TmVar__shift_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TmVar__shift_TyVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d68 : Decls) =>
      (forall (k47 : Hvl) (d69 : (Trace TmVar)) (s19 : Tm) ,
        ((subst_TmVar_Decls (weakenTrace (XS TyVar d69) k47) s19 (shift_TyVar_Decls (weakenCutoffTyVar C0 k47) d68)) =
        (shift_TyVar_Decls (weakenCutoffTyVar C0 k47) (subst_TmVar_Decls (weakenTrace d69 k47) s19 d68))))) (fun (s19 : Tm) =>
      (forall (k47 : Hvl) (d68 : (Trace TmVar)) (s20 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace (XS TyVar d68) k47) s20 (shift_TyVar_Tm (weakenCutoffTyVar C0 k47) s19)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k47) (subst_TmVar_Tm (weakenTrace d68 k47) s20 s19))))) (fun (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
      eq_refl) (fun (t : Tm) IHt97 (d : Decls) IHd (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
      (f_equal2 DCons (IHt97 k47 d68 s19) (eq_trans (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar (XS TyVar d68) k47 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHd (HS TmVar k47) d68 s19) (f_equal2 shift_TyVar_Decls eq_refl (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar d68 k47 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (x52 : (Index TmVar)) =>
      (fun (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
        (subst_TmVar_Index_shift_TyVar_Tm0_comm_ind d68 s19 k47 x52))) (fun (T : Ty) (t : Tm) IHt97 (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TyVar d68) k47 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt97 (HS TmVar k47) d68 s19) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d68 k47 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt97 (t2 : Tm) IHt98 (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
      (f_equal2 App (IHt97 k47 d68 s19) (IHt98 k47 d68 s19))) (fun (t : Tm) IHt97 (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TyVar d68) k47 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt97 (HS TyVar k47) d68 s19) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d68 k47 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt97 (T : Ty) (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
      (f_equal2 TApp (IHt97 k47 d68 s19) eq_refl)) (fun (d : Decls) IHd (t : Tm) IHt97 (k47 : Hvl) (d68 : (Trace TmVar)) (s19 : Tm) =>
      (f_equal2 Let (IHd k47 d68 s19) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TyVar__bind _ _) (eq_refl H0))) (weakenTrace_append TmVar (XS TyVar d68) k47 (bind d))) eq_refl (f_equal2 shift_TyVar_Tm (weakenCutoffTyVar_append C0 k47 (bind d)) eq_refl)) (eq_trans (IHt97 (appendHvl k47 (bind d)) d68 s19) (f_equal2 shift_TyVar_Tm (eq_trans (eq_sym (weakenCutoffTyVar_append C0 k47 (bind d))) (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0)))) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d68 k47 (bind d))) eq_refl eq_refl))))))).
    Lemma subst_TmVar__shift_TyVar_0_Decls_comm_ind  :
      (forall (d68 : Decls) ,
        (forall (k47 : Hvl) (d69 : (Trace TmVar)) (s19 : Tm) ,
          ((subst_TmVar_Decls (weakenTrace (XS TyVar d69) k47) s19 (shift_TyVar_Decls (weakenCutoffTyVar C0 k47) d68)) =
          (shift_TyVar_Decls (weakenCutoffTyVar C0 k47) (subst_TmVar_Decls (weakenTrace d69 k47) s19 d68))))).
    Proof.
      pose proof (subst_TmVar__shift_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__shift_TyVar_0_Tm_comm_ind  :
      (forall (s19 : Tm) ,
        (forall (k47 : Hvl) (d69 : (Trace TmVar)) (s20 : Tm) ,
          ((subst_TmVar_Tm (weakenTrace (XS TyVar d69) k47) s20 (shift_TyVar_Tm (weakenCutoffTyVar C0 k47) s19)) =
          (shift_TyVar_Tm (weakenCutoffTyVar C0 k47) (subst_TmVar_Tm (weakenTrace d69 k47) s20 s19))))).
    Proof.
      pose proof (subst_TmVar__shift_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TyVar__shift_TmVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d68 : Decls) =>
      (forall (k47 : Hvl) (d69 : (Trace TyVar)) (S25 : Ty) ,
        ((subst_TyVar_Decls (weakenTrace d69 k47) S25 (shift_TmVar_Decls (weakenCutoffTmVar C0 k47) d68)) =
        (shift_TmVar_Decls (weakenCutoffTmVar C0 k47) (subst_TyVar_Decls (weakenTrace d69 k47) S25 d68))))) (fun (s19 : Tm) =>
      (forall (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace d68 k47) S25 (shift_TmVar_Tm (weakenCutoffTmVar C0 k47) s19)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k47) (subst_TyVar_Tm (weakenTrace d68 k47) S25 s19))))) (fun (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      eq_refl) (fun (t : Tm) IHt97 (d : Decls) IHd (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      (f_equal2 DCons (IHt97 k47 d68 S25) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar d68 k47 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHd (HS TmVar k47) d68 S25) (f_equal2 shift_TmVar_Decls eq_refl (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar d68 k47 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (x52 : (Index TmVar)) =>
      (fun (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt97 (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d68 k47 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt97 (HS TmVar k47) d68 S25) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d68 k47 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt97 (t2 : Tm) IHt98 (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      (f_equal2 App (IHt97 k47 d68 S25) (IHt98 k47 d68 S25))) (fun (t : Tm) IHt97 (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d68 k47 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt97 (HS TyVar k47) d68 S25) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d68 k47 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt97 (T : Ty) (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      (f_equal2 TApp (IHt97 k47 d68 S25) eq_refl)) (fun (d : Decls) IHd (t : Tm) IHt97 (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      (f_equal2 Let (IHd k47 d68 S25) (eq_trans (f_equal3 subst_TyVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TmVar__bind _ _) (eq_refl H0))) (weakenTrace_append TyVar d68 k47 (bind d))) eq_refl (f_equal2 shift_TmVar_Tm (weakenCutoffTmVar_append C0 k47 (bind d)) eq_refl)) (eq_trans (IHt97 (appendHvl k47 (bind d)) d68 S25) (f_equal2 shift_TmVar_Tm (eq_trans (eq_sym (weakenCutoffTmVar_append C0 k47 (bind d))) (f_equal2 weakenCutoffTmVar eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TyVar__bind _ _ _)) (eq_refl H0)))) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d68 k47 (bind d))) eq_refl eq_refl))))))).
    Lemma subst_TyVar__shift_TmVar_0_Decls_comm_ind  :
      (forall (d68 : Decls) ,
        (forall (k47 : Hvl) (d69 : (Trace TyVar)) (S25 : Ty) ,
          ((subst_TyVar_Decls (weakenTrace d69 k47) S25 (shift_TmVar_Decls (weakenCutoffTmVar C0 k47) d68)) =
          (shift_TmVar_Decls (weakenCutoffTmVar C0 k47) (subst_TyVar_Decls (weakenTrace d69 k47) S25 d68))))).
    Proof.
      pose proof (subst_TyVar__shift_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__shift_TmVar_0_Tm_comm_ind  :
      (forall (s19 : Tm) ,
        (forall (k47 : Hvl) (d69 : (Trace TyVar)) (S25 : Ty) ,
          ((subst_TyVar_Tm (weakenTrace d69 k47) S25 (shift_TmVar_Tm (weakenCutoffTmVar C0 k47) s19)) =
          (shift_TmVar_Tm (weakenCutoffTmVar C0 k47) (subst_TyVar_Tm (weakenTrace d69 k47) S25 s19))))).
    Proof.
      pose proof (subst_TyVar__shift_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TyVar__shift_TyVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d68 : Decls) =>
      (forall (k47 : Hvl) (d69 : (Trace TyVar)) (S25 : Ty) ,
        ((subst_TyVar_Decls (weakenTrace (XS TyVar d69) k47) S25 (shift_TyVar_Decls (weakenCutoffTyVar C0 k47) d68)) =
        (shift_TyVar_Decls (weakenCutoffTyVar C0 k47) (subst_TyVar_Decls (weakenTrace d69 k47) S25 d68))))) (fun (s19 : Tm) =>
      (forall (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace (XS TyVar d68) k47) S25 (shift_TyVar_Tm (weakenCutoffTyVar C0 k47) s19)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k47) (subst_TyVar_Tm (weakenTrace d68 k47) S25 s19))))) (fun (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      eq_refl) (fun (t : Tm) IHt97 (d : Decls) IHd (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      (f_equal2 DCons (IHt97 k47 d68 S25) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar (XS TyVar d68) k47 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHd (HS TmVar k47) d68 S25) (f_equal2 shift_TyVar_Decls eq_refl (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar d68 k47 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (x52 : (Index TmVar)) =>
      (fun (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt97 (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      (f_equal2 Abs (let IHT0 := (subst_TyVar__shift_TyVar_0_Ty_comm_ind T) in
      (IHT0 k47 d68 S25)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TyVar d68) k47 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt97 (HS TmVar k47) d68 S25) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d68 k47 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt97 (t2 : Tm) IHt98 (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      (f_equal2 App (IHt97 k47 d68 S25) (IHt98 k47 d68 S25))) (fun (t : Tm) IHt97 (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TyVar d68) k47 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt97 (HS TyVar k47) d68 S25) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d68 k47 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt97 (T : Ty) (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      (f_equal2 TApp (IHt97 k47 d68 S25) (let IHT0 := (subst_TyVar__shift_TyVar_0_Ty_comm_ind T) in
      (IHT0 k47 d68 S25)))) (fun (d : Decls) IHd (t : Tm) IHt97 (k47 : Hvl) (d68 : (Trace TyVar)) (S25 : Ty) =>
      (f_equal2 Let (IHd k47 d68 S25) (eq_trans (f_equal3 subst_TyVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_shift_TyVar__bind _ _) (eq_refl H0))) (weakenTrace_append TyVar (XS TyVar d68) k47 (bind d))) eq_refl (f_equal2 shift_TyVar_Tm (weakenCutoffTyVar_append C0 k47 (bind d)) eq_refl)) (eq_trans (IHt97 (appendHvl k47 (bind d)) d68 S25) (f_equal2 shift_TyVar_Tm (eq_trans (eq_sym (weakenCutoffTyVar_append C0 k47 (bind d))) (f_equal2 weakenCutoffTyVar eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TyVar__bind _ _ _)) (eq_refl H0)))) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d68 k47 (bind d))) eq_refl eq_refl))))))).
    Lemma subst_TyVar__shift_TyVar_0_Decls_comm_ind  :
      (forall (d68 : Decls) ,
        (forall (k47 : Hvl) (d69 : (Trace TyVar)) (S25 : Ty) ,
          ((subst_TyVar_Decls (weakenTrace (XS TyVar d69) k47) S25 (shift_TyVar_Decls (weakenCutoffTyVar C0 k47) d68)) =
          (shift_TyVar_Decls (weakenCutoffTyVar C0 k47) (subst_TyVar_Decls (weakenTrace d69 k47) S25 d68))))).
    Proof.
      pose proof (subst_TyVar__shift_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__shift_TyVar_0_Tm_comm_ind  :
      (forall (s19 : Tm) ,
        (forall (k47 : Hvl) (d69 : (Trace TyVar)) (S25 : Ty) ,
          ((subst_TyVar_Tm (weakenTrace (XS TyVar d69) k47) S25 (shift_TyVar_Tm (weakenCutoffTyVar C0 k47) s19)) =
          (shift_TyVar_Tm (weakenCutoffTyVar C0 k47) (subst_TyVar_Tm (weakenTrace d69 k47) S25 s19))))).
    Proof.
      pose proof (subst_TyVar__shift_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End SubstShiftCommInd.
  Section SubstShiftComm.
    Definition subst_TyVar_Ty_shift_TyVar_Ty0_comm (S37 : Ty) : (forall (d81 : (Trace TyVar)) (S36 : Ty) ,
      ((subst_TyVar_Ty (XS TyVar d81) S36 (shift_TyVar_Ty C0 S37)) =
      (shift_TyVar_Ty C0 (subst_TyVar_Ty d81 S36 S37)))) := (subst_TyVar__shift_TyVar_0_Ty_comm_ind S37 H0).
    Definition subst_TmVar_Decls_shift_TmVar_Decls0_comm (d82 : Decls) : (forall (d81 : (Trace TmVar)) (s19 : Tm) ,
      ((subst_TmVar_Decls (XS TmVar d81) s19 (shift_TmVar_Decls C0 d82)) =
      (shift_TmVar_Decls C0 (subst_TmVar_Decls d81 s19 d82)))) := (subst_TmVar__shift_TmVar_0_Decls_comm_ind d82 H0).
    Definition subst_TmVar_Tm_shift_TmVar_Tm0_comm (s20 : Tm) : (forall (d81 : (Trace TmVar)) (s19 : Tm) ,
      ((subst_TmVar_Tm (XS TmVar d81) s19 (shift_TmVar_Tm C0 s20)) =
      (shift_TmVar_Tm C0 (subst_TmVar_Tm d81 s19 s20)))) := (subst_TmVar__shift_TmVar_0_Tm_comm_ind s20 H0).
    Definition subst_TmVar_Decls_shift_TyVar_Decls0_comm (d82 : Decls) : (forall (d81 : (Trace TmVar)) (s19 : Tm) ,
      ((subst_TmVar_Decls (XS TyVar d81) s19 (shift_TyVar_Decls C0 d82)) =
      (shift_TyVar_Decls C0 (subst_TmVar_Decls d81 s19 d82)))) := (subst_TmVar__shift_TyVar_0_Decls_comm_ind d82 H0).
    Definition subst_TmVar_Tm_shift_TyVar_Tm0_comm (s20 : Tm) : (forall (d81 : (Trace TmVar)) (s19 : Tm) ,
      ((subst_TmVar_Tm (XS TyVar d81) s19 (shift_TyVar_Tm C0 s20)) =
      (shift_TyVar_Tm C0 (subst_TmVar_Tm d81 s19 s20)))) := (subst_TmVar__shift_TyVar_0_Tm_comm_ind s20 H0).
    Definition subst_TyVar_Decls_shift_TmVar_Decls0_comm (d82 : Decls) : (forall (d81 : (Trace TyVar)) (S36 : Ty) ,
      ((subst_TyVar_Decls d81 S36 (shift_TmVar_Decls C0 d82)) =
      (shift_TmVar_Decls C0 (subst_TyVar_Decls d81 S36 d82)))) := (subst_TyVar__shift_TmVar_0_Decls_comm_ind d82 H0).
    Definition subst_TyVar_Tm_shift_TmVar_Tm0_comm (s19 : Tm) : (forall (d81 : (Trace TyVar)) (S36 : Ty) ,
      ((subst_TyVar_Tm d81 S36 (shift_TmVar_Tm C0 s19)) =
      (shift_TmVar_Tm C0 (subst_TyVar_Tm d81 S36 s19)))) := (subst_TyVar__shift_TmVar_0_Tm_comm_ind s19 H0).
    Definition subst_TyVar_Decls_shift_TyVar_Decls0_comm (d82 : Decls) : (forall (d81 : (Trace TyVar)) (S36 : Ty) ,
      ((subst_TyVar_Decls (XS TyVar d81) S36 (shift_TyVar_Decls C0 d82)) =
      (shift_TyVar_Decls C0 (subst_TyVar_Decls d81 S36 d82)))) := (subst_TyVar__shift_TyVar_0_Decls_comm_ind d82 H0).
    Definition subst_TyVar_Tm_shift_TyVar_Tm0_comm (s19 : Tm) : (forall (d81 : (Trace TyVar)) (S36 : Ty) ,
      ((subst_TyVar_Tm (XS TyVar d81) S36 (shift_TyVar_Tm C0 s19)) =
      (shift_TyVar_Tm C0 (subst_TyVar_Tm d81 S36 s19)))) := (subst_TyVar__shift_TyVar_0_Tm_comm_ind s19 H0).
  End SubstShiftComm.
  Section SubstSubordInd.
    Definition subst_TyVar__TmVar_Ty_ind := (ind_Ty (fun (S36 : Ty) =>
      (forall (k58 : Hvl) (d81 : (Trace TyVar)) (S37 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace (XS TmVar d81) k58) S37 S36) =
        (subst_TyVar_Ty (weakenTrace d81 k58) S37 S36)))) (fun (X32 : (Index TyVar)) =>
      (fun (k58 : Hvl) (d81 : (Trace TyVar)) (S36 : Ty) =>
        (subst_TyVar_Index_shift_TmVar_Ty0_comm_ind d81 S36 k58 X32))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k58 : Hvl) (d81 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal2 TArr (IHT0 k58 d81 S36) (IHT3 k58 d81 S36))) (fun (T : Ty) IHT0 (k58 : Hvl) (d81 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal TAll (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar (XS TmVar d81) k58 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHT0 (HS TyVar k58) d81 S36) (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar d81 k58 (HS TyVar H0))) eq_refl eq_refl)))))).
    Definition subst_TyVar__TmVar_Decls_Tm_ind := (ind_Decls_Tm (fun (d81 : Decls) =>
      (forall (k58 : Hvl) (d82 : (Trace TyVar)) (S36 : Ty) ,
        ((subst_TyVar_Decls (weakenTrace (XS TmVar d82) k58) S36 d81) =
        (subst_TyVar_Decls (weakenTrace d82 k58) S36 d81)))) (fun (s19 : Tm) =>
      (forall (k58 : Hvl) (d81 : (Trace TyVar)) (S36 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace (XS TmVar d81) k58) S36 s19) =
        (subst_TyVar_Tm (weakenTrace d81 k58) S36 s19)))) (fun (k58 : Hvl) (d81 : (Trace TyVar)) (S36 : Ty) =>
      eq_refl) (fun (t : Tm) IHt104 (d : Decls) IHd (k58 : Hvl) (d81 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal2 DCons (IHt104 k58 d81 S36) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar (XS TmVar d81) k58 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHd (HS TmVar k58) d81 S36) (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar d81 k58 (HS TmVar H0))) eq_refl eq_refl))))) (fun (x56 : (Index TmVar)) =>
      (fun (k58 : Hvl) (d81 : (Trace TyVar)) (S36 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt104 (k58 : Hvl) (d81 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal2 Abs (let IHT0 := (subst_TyVar__TmVar_Ty_ind T) in
      (IHT0 k58 d81 S36)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TmVar d81) k58 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt104 (HS TmVar k58) d81 S36) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d81 k58 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt104 (t2 : Tm) IHt105 (k58 : Hvl) (d81 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal2 App (IHt104 k58 d81 S36) (IHt105 k58 d81 S36))) (fun (t : Tm) IHt104 (k58 : Hvl) (d81 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TmVar d81) k58 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt104 (HS TyVar k58) d81 S36) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d81 k58 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt104 (T : Ty) (k58 : Hvl) (d81 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal2 TApp (IHt104 k58 d81 S36) (let IHT0 := (subst_TyVar__TmVar_Ty_ind T) in
      (IHT0 k58 d81 S36)))) (fun (d : Decls) IHd (t : Tm) IHt104 (k58 : Hvl) (d81 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal2 Let (IHd k58 d81 S36) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TmVar d81) k58 (bind d)) eq_refl eq_refl) (eq_trans (IHt104 (appendHvl k58 (bind d)) d81 S36) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d81 k58 (bind d))) eq_refl eq_refl)))))).
    Lemma subst_TyVar__TmVar_Decls_ind  :
      (forall (d81 : Decls) ,
        (forall (k58 : Hvl) (d82 : (Trace TyVar)) (S36 : Ty) ,
          ((subst_TyVar_Decls (weakenTrace (XS TmVar d82) k58) S36 d81) =
          (subst_TyVar_Decls (weakenTrace d82 k58) S36 d81)))).
    Proof.
      pose proof (subst_TyVar__TmVar_Decls_Tm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__TmVar_Tm_ind  :
      (forall (s19 : Tm) ,
        (forall (k58 : Hvl) (d82 : (Trace TyVar)) (S36 : Ty) ,
          ((subst_TyVar_Tm (weakenTrace (XS TmVar d82) k58) S36 s19) =
          (subst_TyVar_Tm (weakenTrace d82 k58) S36 s19)))).
    Proof.
      pose proof (subst_TyVar__TmVar_Decls_Tm_ind).
      destruct_conjs .
      easy .
    Qed.
  End SubstSubordInd.
  Section SubstSubord.
    Definition subst_TyVar_Ty_TmVar (S37 : Ty) : (forall (d81 : (Trace TyVar)) (S36 : Ty) ,
      ((subst_TyVar_Ty (XS TmVar d81) S36 S37) =
      (subst_TyVar_Ty d81 S36 S37))) := (subst_TyVar__TmVar_Ty_ind S37 H0).
    Definition subst_TyVar_Decls_TmVar (d82 : Decls) : (forall (d81 : (Trace TyVar)) (S36 : Ty) ,
      ((subst_TyVar_Decls (XS TmVar d81) S36 d82) =
      (subst_TyVar_Decls d81 S36 d82))) := (subst_TyVar__TmVar_Decls_ind d82 H0).
    Definition subst_TyVar_Tm_TmVar (s19 : Tm) : (forall (d81 : (Trace TyVar)) (S36 : Ty) ,
      ((subst_TyVar_Tm (XS TmVar d81) S36 s19) =
      (subst_TyVar_Tm d81 S36 s19))) := (subst_TyVar__TmVar_Tm_ind s19 H0).
  End SubstSubord.
End ShiftSubstInteraction.
 Hint Rewrite subst_TmVar_Decls0_shift_TmVar_Decls0_reflection subst_TyVar_Decls0_shift_TyVar_Decls0_reflection subst_TmVar_Tm0_shift_TmVar_Tm0_reflection subst_TyVar_Tm0_shift_TyVar_Tm0_reflection subst_TyVar_Ty0_shift_TyVar_Ty0_reflection : interaction.
 Hint Rewrite subst_TmVar_Decls0_shift_TmVar_Decls0_reflection subst_TyVar_Decls0_shift_TyVar_Decls0_reflection subst_TmVar_Tm0_shift_TmVar_Tm0_reflection subst_TyVar_Tm0_shift_TyVar_Tm0_reflection subst_TyVar_Ty0_shift_TyVar_Ty0_reflection : reflection.
 Hint Rewrite subst_TmVar_Decls_shift_TmVar_Decls0_comm subst_TmVar_Decls_shift_TyVar_Decls0_comm subst_TyVar_Decls_shift_TmVar_Decls0_comm subst_TyVar_Decls_shift_TyVar_Decls0_comm subst_TmVar_Tm_shift_TmVar_Tm0_comm subst_TmVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Tm_shift_TmVar_Tm0_comm subst_TyVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Ty_shift_TyVar_Ty0_comm : interaction.
 Hint Rewrite subst_TmVar_Decls_shift_TmVar_Decls0_comm subst_TmVar_Decls_shift_TyVar_Decls0_comm subst_TyVar_Decls_shift_TmVar_Decls0_comm subst_TyVar_Decls_shift_TyVar_Decls0_comm subst_TmVar_Tm_shift_TmVar_Tm0_comm subst_TmVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Tm_shift_TmVar_Tm0_comm subst_TyVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Ty_shift_TyVar_Ty0_comm : subst_shift0.
 Hint Rewrite subst_TyVar_Decls_TmVar subst_TyVar_Tm_TmVar subst_TyVar_Ty_TmVar : interaction.
 Hint Rewrite subst_TyVar_Decls_TmVar subst_TyVar_Tm_TmVar subst_TyVar_Ty_TmVar : subst_shift0.
 Hint Rewrite shift_TmVar_Decls_subst_TmVar_Decls0_comm shift_TmVar_Decls_subst_TyVar_Decls0_comm shift_TyVar_Decls_subst_TmVar_Decls0_comm shift_TyVar_Decls_subst_TyVar_Decls0_comm shift_TmVar_Tm_subst_TmVar_Tm0_comm shift_TmVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Tm_subst_TmVar_Tm0_comm shift_TyVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Ty_subst_TyVar_Ty0_comm : interaction.
 Hint Rewrite shift_TmVar_Decls_subst_TmVar_Decls0_comm shift_TmVar_Decls_subst_TyVar_Decls0_comm shift_TyVar_Decls_subst_TmVar_Decls0_comm shift_TyVar_Decls_subst_TyVar_Decls0_comm shift_TmVar_Tm_subst_TmVar_Tm0_comm shift_TmVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Tm_subst_TmVar_Tm0_comm shift_TyVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Ty_subst_TyVar_Ty0_comm : shift_subst0.
Section SubstSubstInteraction.
  Section SubstSubstIndexCommInd.
    Lemma subst_TmVar_Tm_subst_TmVar_Index0_commright_ind (d81 : (Trace TmVar)) (s19 : Tm) (s20 : Tm) :
      (forall (k58 : Hvl) (x56 : (Index TmVar)) ,
        ((subst_TmVar_Tm (weakenTrace d81 k58) s19 (subst_TmVar_Index (weakenTrace X0 k58) s20 x56)) =
        (subst_TmVar_Tm (weakenTrace X0 k58) (subst_TmVar_Tm d81 s19 s20) (subst_TmVar_Index (weakenTrace (XS TmVar d81) k58) s19 x56)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Tm_subst_TmVar_Index0_commright_ind (d81 : (Trace TyVar)) (S36 : Ty) (s19 : Tm) :
      (forall (k58 : Hvl) (x56 : (Index TmVar)) ,
        ((subst_TyVar_Tm (weakenTrace d81 k58) S36 (subst_TmVar_Index (weakenTrace X0 k58) s19 x56)) =
        (subst_TmVar_Index (weakenTrace X0 k58) (subst_TyVar_Tm d81 S36 s19) x56))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Ty_subst_TyVar_Index0_commright_ind (d81 : (Trace TyVar)) (S36 : Ty) (S37 : Ty) :
      (forall (k58 : Hvl) (X32 : (Index TyVar)) ,
        ((subst_TyVar_Ty (weakenTrace d81 k58) S36 (subst_TyVar_Index (weakenTrace X0 k58) S37 X32)) =
        (subst_TyVar_Ty (weakenTrace X0 k58) (subst_TyVar_Ty d81 S36 S37) (subst_TyVar_Index (weakenTrace (XS TyVar d81) k58) S36 X32)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Ty_subst_TyVar_Index0_commleft_ind (d81 : (Trace TmVar)) (s19 : Tm) (S36 : Ty) :
      (forall (k58 : Hvl) (x56 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace d81 k58) s19 x56) =
        (subst_TyVar_Tm (weakenTrace X0 k58) S36 (subst_TmVar_Index (weakenTrace (XS TyVar d81) k58) s19 x56)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
  End SubstSubstIndexCommInd.
  Section SubstSubstCommInd.
    Definition subst_TyVar__subst_TyVar_0_Ty_comm_ind := (ind_Ty (fun (S40 : Ty) =>
      (forall (k61 : Hvl) (d88 : (Trace TyVar)) (S41 : Ty) (S42 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace d88 k61) S41 (subst_TyVar_Ty (weakenTrace X0 k61) S42 S40)) =
        (subst_TyVar_Ty (weakenTrace X0 k61) (subst_TyVar_Ty d88 S41 S42) (subst_TyVar_Ty (weakenTrace (XS TyVar d88) k61) S41 S40))))) (fun (X37 : (Index TyVar)) =>
      (fun (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
        (subst_TyVar_Ty_subst_TyVar_Index0_commright_ind d88 S40 S41 k61 X37))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal2 TArr (IHT0 k61 d88 S40 S41) (IHT3 k61 d88 S40 S41))) (fun (T : Ty) IHT0 (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal TAll (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar d88 k61 (HS TyVar H0)) eq_refl (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k61 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHT0 (HS TyVar k61) d88 S40 S41) (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar X0 k61 (HS TyVar H0))) eq_refl (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar (XS TyVar d88) k61 (HS TyVar H0))) eq_refl eq_refl))))))).
    Definition subst_TmVar__subst_TmVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d88 : Decls) =>
      (forall (k61 : Hvl) (d89 : (Trace TmVar)) (s21 : Tm) (s22 : Tm) ,
        ((subst_TmVar_Decls (weakenTrace d89 k61) s21 (subst_TmVar_Decls (weakenTrace X0 k61) s22 d88)) =
        (subst_TmVar_Decls (weakenTrace X0 k61) (subst_TmVar_Tm d89 s21 s22) (subst_TmVar_Decls (weakenTrace (XS TmVar d89) k61) s21 d88))))) (fun (s21 : Tm) =>
      (forall (k61 : Hvl) (d88 : (Trace TmVar)) (s22 : Tm) (s23 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace d88 k61) s22 (subst_TmVar_Tm (weakenTrace X0 k61) s23 s21)) =
        (subst_TmVar_Tm (weakenTrace X0 k61) (subst_TmVar_Tm d88 s22 s23) (subst_TmVar_Tm (weakenTrace (XS TmVar d88) k61) s22 s21))))) (fun (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (s22 : Tm) =>
      eq_refl) (fun (t : Tm) IHt118 (d : Decls) IHd (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (s22 : Tm) =>
      (f_equal2 DCons (IHt118 k61 d88 s21 s22) (eq_trans (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar d88 k61 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar X0 k61 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k61) d88 s21 s22) (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar X0 k61 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar (XS TmVar d88) k61 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (x64 : (Index TmVar)) =>
      (fun (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (s22 : Tm) =>
        (subst_TmVar_Tm_subst_TmVar_Index0_commright_ind d88 s21 s22 k61 x64))) (fun (T : Ty) (t : Tm) IHt118 (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (s22 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d88 k61 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k61 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt118 (HS TmVar k61) d88 s21 s22) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k61 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TmVar d88) k61 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt118 (t2 : Tm) IHt119 (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (s22 : Tm) =>
      (f_equal2 App (IHt118 k61 d88 s21 s22) (IHt119 k61 d88 s21 s22))) (fun (t : Tm) IHt118 (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (s22 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d88 k61 (HS TyVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k61 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt118 (HS TyVar k61) d88 s21 s22) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k61 (HS TyVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TmVar d88) k61 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt118 (T : Ty) (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (s22 : Tm) =>
      (f_equal2 TApp (IHt118 k61 d88 s21 s22) eq_refl)) (fun (d : Decls) IHd (t : Tm) IHt118 (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (s22 : Tm) =>
      (f_equal2 Let (IHd k61 d88 s21 s22) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_subst_TmVar__bind _ _ _) (eq_refl H0))) (weakenTrace_append TmVar d88 k61 (bind d))) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k61 (bind d)) eq_refl eq_refl)) (eq_trans (IHt118 (appendHvl k61 (bind d)) d88 s21 s22) (f_equal3 subst_TmVar_Tm (eq_trans (eq_sym (weakenTrace_append TmVar X0 k61 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0)))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TmVar d88) k61 (bind d))) eq_refl eq_refl))))))).
    Lemma subst_TmVar__subst_TmVar_0_Decls_comm_ind  :
      (forall (d88 : Decls) ,
        (forall (k61 : Hvl) (d89 : (Trace TmVar)) (s21 : Tm) (s22 : Tm) ,
          ((subst_TmVar_Decls (weakenTrace d89 k61) s21 (subst_TmVar_Decls (weakenTrace X0 k61) s22 d88)) =
          (subst_TmVar_Decls (weakenTrace X0 k61) (subst_TmVar_Tm d89 s21 s22) (subst_TmVar_Decls (weakenTrace (XS TmVar d89) k61) s21 d88))))).
    Proof.
      pose proof (subst_TmVar__subst_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__subst_TmVar_0_Tm_comm_ind  :
      (forall (s21 : Tm) ,
        (forall (k61 : Hvl) (d89 : (Trace TmVar)) (s22 : Tm) (s23 : Tm) ,
          ((subst_TmVar_Tm (weakenTrace d89 k61) s22 (subst_TmVar_Tm (weakenTrace X0 k61) s23 s21)) =
          (subst_TmVar_Tm (weakenTrace X0 k61) (subst_TmVar_Tm d89 s22 s23) (subst_TmVar_Tm (weakenTrace (XS TmVar d89) k61) s22 s21))))).
    Proof.
      pose proof (subst_TmVar__subst_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TmVar__subst_TyVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d88 : Decls) =>
      (forall (k61 : Hvl) (d89 : (Trace TmVar)) (s21 : Tm) (S40 : Ty) ,
        ((subst_TmVar_Decls (weakenTrace d89 k61) s21 (subst_TyVar_Decls (weakenTrace X0 k61) S40 d88)) =
        (subst_TyVar_Decls (weakenTrace X0 k61) S40 (subst_TmVar_Decls (weakenTrace (XS TyVar d89) k61) s21 d88))))) (fun (s21 : Tm) =>
      (forall (k61 : Hvl) (d88 : (Trace TmVar)) (s22 : Tm) (S40 : Ty) ,
        ((subst_TmVar_Tm (weakenTrace d88 k61) s22 (subst_TyVar_Tm (weakenTrace X0 k61) S40 s21)) =
        (subst_TyVar_Tm (weakenTrace X0 k61) S40 (subst_TmVar_Tm (weakenTrace (XS TyVar d88) k61) s22 s21))))) (fun (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (S40 : Ty) =>
      eq_refl) (fun (t : Tm) IHt118 (d : Decls) IHd (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (S40 : Ty) =>
      (f_equal2 DCons (IHt118 k61 d88 s21 S40) (eq_trans (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar d88 k61 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar X0 k61 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k61) d88 s21 S40) (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar X0 k61 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar (XS TyVar d88) k61 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (x64 : (Index TmVar)) =>
      (fun (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (S40 : Ty) =>
        (subst_TmVar_Ty_subst_TyVar_Index0_commleft_ind d88 s21 S40 k61 x64))) (fun (T : Ty) (t : Tm) IHt118 (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (S40 : Ty) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d88 k61 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k61 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt118 (HS TmVar k61) d88 s21 S40) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k61 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TyVar d88) k61 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt118 (t2 : Tm) IHt119 (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (S40 : Ty) =>
      (f_equal2 App (IHt118 k61 d88 s21 S40) (IHt119 k61 d88 s21 S40))) (fun (t : Tm) IHt118 (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (S40 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d88 k61 (HS TyVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k61 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt118 (HS TyVar k61) d88 s21 S40) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k61 (HS TyVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TyVar d88) k61 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt118 (T : Ty) (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (S40 : Ty) =>
      (f_equal2 TApp (IHt118 k61 d88 s21 S40) eq_refl)) (fun (d : Decls) IHd (t : Tm) IHt118 (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (S40 : Ty) =>
      (f_equal2 Let (IHd k61 d88 s21 S40) (eq_trans (f_equal3 subst_TmVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_subst_TyVar__bind _ _ _) (eq_refl H0))) (weakenTrace_append TmVar d88 k61 (bind d))) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k61 (bind d)) eq_refl eq_refl)) (eq_trans (IHt118 (appendHvl k61 (bind d)) d88 s21 S40) (f_equal3 subst_TyVar_Tm (eq_trans (eq_sym (weakenTrace_append TyVar X0 k61 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0)))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TyVar d88) k61 (bind d))) eq_refl eq_refl))))))).
    Lemma subst_TmVar__subst_TyVar_0_Decls_comm_ind  :
      (forall (d88 : Decls) ,
        (forall (k61 : Hvl) (d89 : (Trace TmVar)) (s21 : Tm) (S40 : Ty) ,
          ((subst_TmVar_Decls (weakenTrace d89 k61) s21 (subst_TyVar_Decls (weakenTrace X0 k61) S40 d88)) =
          (subst_TyVar_Decls (weakenTrace X0 k61) S40 (subst_TmVar_Decls (weakenTrace (XS TyVar d89) k61) s21 d88))))).
    Proof.
      pose proof (subst_TmVar__subst_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__subst_TyVar_0_Tm_comm_ind  :
      (forall (s21 : Tm) ,
        (forall (k61 : Hvl) (d89 : (Trace TmVar)) (s22 : Tm) (S40 : Ty) ,
          ((subst_TmVar_Tm (weakenTrace d89 k61) s22 (subst_TyVar_Tm (weakenTrace X0 k61) S40 s21)) =
          (subst_TyVar_Tm (weakenTrace X0 k61) S40 (subst_TmVar_Tm (weakenTrace (XS TyVar d89) k61) s22 s21))))).
    Proof.
      pose proof (subst_TmVar__subst_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TyVar__subst_TmVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d88 : Decls) =>
      (forall (k61 : Hvl) (d89 : (Trace TyVar)) (S40 : Ty) (s21 : Tm) ,
        ((subst_TyVar_Decls (weakenTrace d89 k61) S40 (subst_TmVar_Decls (weakenTrace X0 k61) s21 d88)) =
        (subst_TmVar_Decls (weakenTrace X0 k61) (subst_TyVar_Tm d89 S40 s21) (subst_TyVar_Decls (weakenTrace d89 k61) S40 d88))))) (fun (s21 : Tm) =>
      (forall (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (s22 : Tm) ,
        ((subst_TyVar_Tm (weakenTrace d88 k61) S40 (subst_TmVar_Tm (weakenTrace X0 k61) s22 s21)) =
        (subst_TmVar_Tm (weakenTrace X0 k61) (subst_TyVar_Tm d88 S40 s22) (subst_TyVar_Tm (weakenTrace d88 k61) S40 s21))))) (fun (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (s21 : Tm) =>
      eq_refl) (fun (t : Tm) IHt118 (d : Decls) IHd (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (s21 : Tm) =>
      (f_equal2 DCons (IHt118 k61 d88 S40 s21) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar d88 k61 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Decls (weakenTrace_append TmVar X0 k61 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k61) d88 S40 s21) (f_equal3 subst_TmVar_Decls (eq_sym (weakenTrace_append TmVar X0 k61 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar d88 k61 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (x64 : (Index TmVar)) =>
      (fun (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (s21 : Tm) =>
        (subst_TyVar_Tm_subst_TmVar_Index0_commright_ind d88 S40 s21 k61 x64))) (fun (T : Ty) (t : Tm) IHt118 (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (s21 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d88 k61 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k61 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt118 (HS TmVar k61) d88 S40 s21) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k61 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d88 k61 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt118 (t2 : Tm) IHt119 (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (s21 : Tm) =>
      (f_equal2 App (IHt118 k61 d88 S40 s21) (IHt119 k61 d88 S40 s21))) (fun (t : Tm) IHt118 (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (s21 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d88 k61 (HS TyVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k61 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt118 (HS TyVar k61) d88 S40 s21) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k61 (HS TyVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d88 k61 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt118 (T : Ty) (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (s21 : Tm) =>
      (f_equal2 TApp (IHt118 k61 d88 S40 s21) eq_refl)) (fun (d : Decls) IHd (t : Tm) IHt118 (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (s21 : Tm) =>
      (f_equal2 Let (IHd k61 d88 S40 s21) (eq_trans (f_equal3 subst_TyVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_subst_TmVar__bind _ _ _) (eq_refl H0))) (weakenTrace_append TyVar d88 k61 (bind d))) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k61 (bind d)) eq_refl eq_refl)) (eq_trans (IHt118 (appendHvl k61 (bind d)) d88 S40 s21) (f_equal3 subst_TmVar_Tm (eq_trans (eq_sym (weakenTrace_append TmVar X0 k61 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TyVar__bind _ _ _)) (eq_refl H0)))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d88 k61 (bind d))) eq_refl eq_refl))))))).
    Lemma subst_TyVar__subst_TmVar_0_Decls_comm_ind  :
      (forall (d88 : Decls) ,
        (forall (k61 : Hvl) (d89 : (Trace TyVar)) (S40 : Ty) (s21 : Tm) ,
          ((subst_TyVar_Decls (weakenTrace d89 k61) S40 (subst_TmVar_Decls (weakenTrace X0 k61) s21 d88)) =
          (subst_TmVar_Decls (weakenTrace X0 k61) (subst_TyVar_Tm d89 S40 s21) (subst_TyVar_Decls (weakenTrace d89 k61) S40 d88))))).
    Proof.
      pose proof (subst_TyVar__subst_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__subst_TmVar_0_Tm_comm_ind  :
      (forall (s21 : Tm) ,
        (forall (k61 : Hvl) (d89 : (Trace TyVar)) (S40 : Ty) (s22 : Tm) ,
          ((subst_TyVar_Tm (weakenTrace d89 k61) S40 (subst_TmVar_Tm (weakenTrace X0 k61) s22 s21)) =
          (subst_TmVar_Tm (weakenTrace X0 k61) (subst_TyVar_Tm d89 S40 s22) (subst_TyVar_Tm (weakenTrace d89 k61) S40 s21))))).
    Proof.
      pose proof (subst_TyVar__subst_TmVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TyVar__subst_TyVar_0_Decls_Tm_comm_ind := (ind_Decls_Tm (fun (d88 : Decls) =>
      (forall (k61 : Hvl) (d89 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) ,
        ((subst_TyVar_Decls (weakenTrace d89 k61) S40 (subst_TyVar_Decls (weakenTrace X0 k61) S41 d88)) =
        (subst_TyVar_Decls (weakenTrace X0 k61) (subst_TyVar_Ty d89 S40 S41) (subst_TyVar_Decls (weakenTrace (XS TyVar d89) k61) S40 d88))))) (fun (s21 : Tm) =>
      (forall (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace d88 k61) S40 (subst_TyVar_Tm (weakenTrace X0 k61) S41 s21)) =
        (subst_TyVar_Tm (weakenTrace X0 k61) (subst_TyVar_Ty d88 S40 S41) (subst_TyVar_Tm (weakenTrace (XS TyVar d88) k61) S40 s21))))) (fun (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      eq_refl) (fun (t : Tm) IHt118 (d : Decls) IHd (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal2 DCons (IHt118 k61 d88 S40 S41) (eq_trans (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar d88 k61 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Decls (weakenTrace_append TyVar X0 k61 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHd (HS TmVar k61) d88 S40 S41) (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar X0 k61 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Decls (eq_sym (weakenTrace_append TyVar (XS TyVar d88) k61 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (x64 : (Index TmVar)) =>
      (fun (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt118 (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal2 Abs (let IHT0 := (subst_TyVar__subst_TyVar_0_Ty_comm_ind T) in
      (IHT0 k61 d88 S40 S41)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d88 k61 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k61 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt118 (HS TmVar k61) d88 S40 S41) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k61 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar (XS TyVar d88) k61 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt118 (t2 : Tm) IHt119 (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal2 App (IHt118 k61 d88 S40 S41) (IHt119 k61 d88 S40 S41))) (fun (t : Tm) IHt118 (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d88 k61 (HS TyVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k61 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt118 (HS TyVar k61) d88 S40 S41) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k61 (HS TyVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar (XS TyVar d88) k61 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt118 (T : Ty) (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal2 TApp (IHt118 k61 d88 S40 S41) (let IHT0 := (subst_TyVar__subst_TyVar_0_Ty_comm_ind T) in
      (IHT0 k61 d88 S40 S41)))) (fun (d : Decls) IHd (t : Tm) IHt118 (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal2 Let (IHd k61 d88 S40 S41) (eq_trans (f_equal3 subst_TyVar_Tm (eq_trans (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (stability_subst_TyVar__bind _ _ _) (eq_refl H0))) (weakenTrace_append TyVar d88 k61 (bind d))) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k61 (bind d)) eq_refl eq_refl)) (eq_trans (IHt118 (appendHvl k61 (bind d)) d88 S40 S41) (f_equal3 subst_TyVar_Tm (eq_trans (eq_sym (weakenTrace_append TyVar X0 k61 (bind d))) (f_equal2 weakenTrace eq_refl (f_equal2 appendHvl (eq_sym (stability_subst_TyVar__bind _ _ _)) (eq_refl H0)))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar (XS TyVar d88) k61 (bind d))) eq_refl eq_refl))))))).
    Lemma subst_TyVar__subst_TyVar_0_Decls_comm_ind  :
      (forall (d88 : Decls) ,
        (forall (k61 : Hvl) (d89 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) ,
          ((subst_TyVar_Decls (weakenTrace d89 k61) S40 (subst_TyVar_Decls (weakenTrace X0 k61) S41 d88)) =
          (subst_TyVar_Decls (weakenTrace X0 k61) (subst_TyVar_Ty d89 S40 S41) (subst_TyVar_Decls (weakenTrace (XS TyVar d89) k61) S40 d88))))).
    Proof.
      pose proof (subst_TyVar__subst_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__subst_TyVar_0_Tm_comm_ind  :
      (forall (s21 : Tm) ,
        (forall (k61 : Hvl) (d89 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) ,
          ((subst_TyVar_Tm (weakenTrace d89 k61) S40 (subst_TyVar_Tm (weakenTrace X0 k61) S41 s21)) =
          (subst_TyVar_Tm (weakenTrace X0 k61) (subst_TyVar_Ty d89 S40 S41) (subst_TyVar_Tm (weakenTrace (XS TyVar d89) k61) S40 s21))))).
    Proof.
      pose proof (subst_TyVar__subst_TyVar_0_Decls_Tm_comm_ind).
      destruct_conjs .
      easy .
    Qed.
  End SubstSubstCommInd.
  Section SubstSubstComm.
    Definition subst_TyVar_Ty_subst_TyVar_Ty0_comm (S42 : Ty) : (forall (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) ,
      ((subst_TyVar_Ty d88 S40 (subst_TyVar_Ty X0 S41 S42)) =
      (subst_TyVar_Ty X0 (subst_TyVar_Ty d88 S40 S41) (subst_TyVar_Ty (XS TyVar d88) S40 S42)))) := (subst_TyVar__subst_TyVar_0_Ty_comm_ind S42 H0).
    Definition subst_TmVar_Decls_subst_TmVar_Decls0_comm (d89 : Decls) : (forall (d88 : (Trace TmVar)) (s21 : Tm) (s22 : Tm) ,
      ((subst_TmVar_Decls d88 s21 (subst_TmVar_Decls X0 s22 d89)) =
      (subst_TmVar_Decls X0 (subst_TmVar_Tm d88 s21 s22) (subst_TmVar_Decls (XS TmVar d88) s21 d89)))) := (subst_TmVar__subst_TmVar_0_Decls_comm_ind d89 H0).
    Definition subst_TmVar_Tm_subst_TmVar_Tm0_comm (s23 : Tm) : (forall (d88 : (Trace TmVar)) (s21 : Tm) (s22 : Tm) ,
      ((subst_TmVar_Tm d88 s21 (subst_TmVar_Tm X0 s22 s23)) =
      (subst_TmVar_Tm X0 (subst_TmVar_Tm d88 s21 s22) (subst_TmVar_Tm (XS TmVar d88) s21 s23)))) := (subst_TmVar__subst_TmVar_0_Tm_comm_ind s23 H0).
    Definition subst_TmVar_Decls_subst_TyVar_Decls0_comm (d89 : Decls) : (forall (d88 : (Trace TmVar)) (s21 : Tm) (S40 : Ty) ,
      ((subst_TmVar_Decls d88 s21 (subst_TyVar_Decls X0 S40 d89)) =
      (subst_TyVar_Decls X0 S40 (subst_TmVar_Decls (XS TyVar d88) s21 d89)))) := (subst_TmVar__subst_TyVar_0_Decls_comm_ind d89 H0).
    Definition subst_TmVar_Tm_subst_TyVar_Tm0_comm (s22 : Tm) : (forall (d88 : (Trace TmVar)) (s21 : Tm) (S40 : Ty) ,
      ((subst_TmVar_Tm d88 s21 (subst_TyVar_Tm X0 S40 s22)) =
      (subst_TyVar_Tm X0 S40 (subst_TmVar_Tm (XS TyVar d88) s21 s22)))) := (subst_TmVar__subst_TyVar_0_Tm_comm_ind s22 H0).
    Definition subst_TyVar_Decls_subst_TmVar_Decls0_comm (d89 : Decls) : (forall (d88 : (Trace TyVar)) (S40 : Ty) (s21 : Tm) ,
      ((subst_TyVar_Decls d88 S40 (subst_TmVar_Decls X0 s21 d89)) =
      (subst_TmVar_Decls X0 (subst_TyVar_Tm d88 S40 s21) (subst_TyVar_Decls d88 S40 d89)))) := (subst_TyVar__subst_TmVar_0_Decls_comm_ind d89 H0).
    Definition subst_TyVar_Tm_subst_TmVar_Tm0_comm (s22 : Tm) : (forall (d88 : (Trace TyVar)) (S40 : Ty) (s21 : Tm) ,
      ((subst_TyVar_Tm d88 S40 (subst_TmVar_Tm X0 s21 s22)) =
      (subst_TmVar_Tm X0 (subst_TyVar_Tm d88 S40 s21) (subst_TyVar_Tm d88 S40 s22)))) := (subst_TyVar__subst_TmVar_0_Tm_comm_ind s22 H0).
    Definition subst_TyVar_Decls_subst_TyVar_Decls0_comm (d89 : Decls) : (forall (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) ,
      ((subst_TyVar_Decls d88 S40 (subst_TyVar_Decls X0 S41 d89)) =
      (subst_TyVar_Decls X0 (subst_TyVar_Ty d88 S40 S41) (subst_TyVar_Decls (XS TyVar d88) S40 d89)))) := (subst_TyVar__subst_TyVar_0_Decls_comm_ind d89 H0).
    Definition subst_TyVar_Tm_subst_TyVar_Tm0_comm (s21 : Tm) : (forall (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) ,
      ((subst_TyVar_Tm d88 S40 (subst_TyVar_Tm X0 S41 s21)) =
      (subst_TyVar_Tm X0 (subst_TyVar_Ty d88 S40 S41) (subst_TyVar_Tm (XS TyVar d88) S40 s21)))) := (subst_TyVar__subst_TyVar_0_Tm_comm_ind s21 H0).
  End SubstSubstComm.
  Section WeakenSubst.
    Lemma weakenTy_subst_TyVar_Ty  :
      (forall (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) ,
        ((weakenTy (subst_TyVar_Ty d88 S40 S41) k61) =
        (subst_TyVar_Ty (weakenTrace d88 k61) S40 (weakenTy S41 k61)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenDecls_subst_TmVar_Decls  :
      (forall (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (d89 : Decls) ,
        ((weakenDecls (subst_TmVar_Decls d88 s21 d89) k61) =
        (subst_TmVar_Decls (weakenTrace d88 k61) s21 (weakenDecls d89 k61)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_subst_TmVar_Tm  :
      (forall (k61 : Hvl) (d88 : (Trace TmVar)) (s21 : Tm) (s22 : Tm) ,
        ((weakenTm (subst_TmVar_Tm d88 s21 s22) k61) =
        (subst_TmVar_Tm (weakenTrace d88 k61) s21 (weakenTm s22 k61)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenDecls_subst_TyVar_Decls  :
      (forall (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (d89 : Decls) ,
        ((weakenDecls (subst_TyVar_Decls d88 S40 d89) k61) =
        (subst_TyVar_Decls (weakenTrace d88 k61) S40 (weakenDecls d89 k61)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_subst_TyVar_Tm  :
      (forall (k61 : Hvl) (d88 : (Trace TyVar)) (S40 : Ty) (s21 : Tm) ,
        ((weakenTm (subst_TyVar_Tm d88 S40 s21) k61) =
        (subst_TyVar_Tm (weakenTrace d88 k61) S40 (weakenTm s21 k61)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
  End WeakenSubst.
  Section WeakenAppend.
    Lemma weakenTy_append  :
      (forall (S40 : Ty) (k61 : Hvl) (k62 : Hvl) ,
        ((weakenTy (weakenTy S40 k61) k62) =
        (weakenTy S40 (appendHvl k61 k62)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenDecls_append  :
      (forall (d88 : Decls) (k61 : Hvl) (k62 : Hvl) ,
        ((weakenDecls (weakenDecls d88 k61) k62) =
        (weakenDecls d88 (appendHvl k61 k62)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenTm_append  :
      (forall (s21 : Tm) (k61 : Hvl) (k62 : Hvl) ,
        ((weakenTm (weakenTm s21 k61) k62) =
        (weakenTm s21 (appendHvl k61 k62)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
  End WeakenAppend.
End SubstSubstInteraction.
 Hint Rewrite subst_TmVar_Decls_subst_TmVar_Decls0_comm subst_TyVar_Decls_subst_TyVar_Decls0_comm subst_TmVar_Tm_subst_TmVar_Tm0_comm subst_TyVar_Tm_subst_TyVar_Tm0_comm subst_TyVar_Ty_subst_TyVar_Ty0_comm : interaction.
 Hint Rewrite subst_TmVar_Decls_subst_TmVar_Decls0_comm subst_TyVar_Decls_subst_TyVar_Decls0_comm subst_TmVar_Tm_subst_TmVar_Tm0_comm subst_TyVar_Tm_subst_TyVar_Tm0_comm subst_TyVar_Ty_subst_TyVar_Ty0_comm : subst_subst0.
 Hint Rewrite weakenDecls_shift_TmVar_Decls weakenDecls_shift_TyVar_Decls weakenTm_shift_TmVar_Tm weakenTm_shift_TyVar_Tm weakenTy_shift_TyVar_Ty : weaken_shift.
 Hint Rewrite weakenDecls_subst_TmVar_Decls weakenDecls_subst_TyVar_Decls weakenTm_subst_TmVar_Tm weakenTm_subst_TyVar_Tm weakenTy_subst_TyVar_Ty : weaken_subst.
Section WellFormed.
  Fixpoint wfindex {a : Namespace} (k61 : Hvl) (i : (Index a)) {struct k61} : Prop :=
    match k61 with
      | (H0) => False
      | (HS b k61) => match (eq_namespace_dec a b) with
        | (left e) => match i with
          | (I0) => True
          | (IS i) => (wfindex k61 i)
        end
        | (right n) => (wfindex k61 i)
      end
    end.
  Inductive wfTy (k61 : Hvl) : Ty -> Prop :=
    | wfTy_TVar
        {X43 : (Index TyVar)}
        (wfi : (wfindex k61 X43)) :
        (wfTy k61 (TVar X43))
    | wfTy_TArr {T58 : Ty}
        (wf : (wfTy k61 T58)) {T59 : Ty}
        (wf0 : (wfTy k61 T59)) :
        (wfTy k61 (TArr T58 T59))
    | wfTy_TAll {T60 : Ty}
        (wf : (wfTy (HS TyVar k61) T60))
        : (wfTy k61 (TAll T60)).
  Inductive wfDecls (k61 : Hvl) : Decls -> Prop :=
    | wfDecls_DNil :
        (wfDecls k61 (DNil))
    | wfDecls_DCons {t132 : Tm}
        (wf : (wfTm k61 t132))
        {d92 : Decls}
        (wf0 : (wfDecls (HS TmVar k61) d92))
        : (wfDecls k61 (DCons t132 d92))
  with wfTm (k61 : Hvl) : Tm -> Prop :=
    | wfTm_Var
        {x71 : (Index TmVar)}
        (wfi : (wfindex k61 x71)) :
        (wfTm k61 (Var x71))
    | wfTm_Abs {T58 : Ty}
        (wf : (wfTy k61 T58))
        {t132 : Tm}
        (wf0 : (wfTm (HS TmVar k61) t132))
        : (wfTm k61 (Abs T58 t132))
    | wfTm_App {t133 : Tm}
        (wf : (wfTm k61 t133))
        {t134 : Tm}
        (wf0 : (wfTm k61 t134)) :
        (wfTm k61 (App t133 t134))
    | wfTm_TAbs {t135 : Tm}
        (wf : (wfTm (HS TyVar k61) t135))
        : (wfTm k61 (TAbs t135))
    | wfTm_TApp {t136 : Tm}
        (wf : (wfTm k61 t136))
        {T59 : Ty}
        (wf0 : (wfTy k61 T59)) :
        (wfTm k61 (TApp t136 T59))
    | wfTm_Let {d92 : Decls}
        (wf : (wfDecls k61 d92))
        {t137 : Tm}
        (wf0 : (wfTm (appendHvl k61 (bind d92)) t137))
        : (wfTm k61 (Let d92 t137)).
  Scheme ind_wfTy := Induction for wfTy Sort Prop.
  Scheme ind_wfDecls := Induction for wfDecls Sort Prop
  with ind_wfTm := Induction for wfTm Sort Prop.
  Combined Scheme ind_wfDecls_Tm from ind_wfDecls, ind_wfTm.
End WellFormed.
Section ShiftWellFormed.
  Inductive shifthvl_TmVar : (forall (c14 : (Cutoff TmVar)) (k61 : Hvl) (k62 : Hvl) ,
    Prop) :=
    | shifthvl_TmVar_here
        (k61 : Hvl) :
        (shifthvl_TmVar C0 k61 (HS TmVar k61))
    | shifthvl_TmVar_there_TmVar
        (c14 : (Cutoff TmVar))
        (k61 : Hvl) (k62 : Hvl) :
        (shifthvl_TmVar c14 k61 k62) -> (shifthvl_TmVar (CS c14) (HS TmVar k61) (HS TmVar k62))
    | shifthvl_TmVar_there_TyVar
        (c14 : (Cutoff TmVar))
        (k61 : Hvl) (k62 : Hvl) :
        (shifthvl_TmVar c14 k61 k62) -> (shifthvl_TmVar c14 (HS TyVar k61) (HS TyVar k62)).
  Inductive shifthvl_TyVar : (forall (c14 : (Cutoff TyVar)) (k61 : Hvl) (k62 : Hvl) ,
    Prop) :=
    | shifthvl_TyVar_here
        (k61 : Hvl) :
        (shifthvl_TyVar C0 k61 (HS TyVar k61))
    | shifthvl_TyVar_there_TmVar
        (c14 : (Cutoff TyVar))
        (k61 : Hvl) (k62 : Hvl) :
        (shifthvl_TyVar c14 k61 k62) -> (shifthvl_TyVar c14 (HS TmVar k61) (HS TmVar k62))
    | shifthvl_TyVar_there_TyVar
        (c14 : (Cutoff TyVar))
        (k61 : Hvl) (k62 : Hvl) :
        (shifthvl_TyVar c14 k61 k62) -> (shifthvl_TyVar (CS c14) (HS TyVar k61) (HS TyVar k62)).
  Lemma weaken_shifthvl_TmVar  :
    (forall (k61 : Hvl) {c14 : (Cutoff TmVar)} {k62 : Hvl} {k63 : Hvl} ,
      (shifthvl_TmVar c14 k62 k63) -> (shifthvl_TmVar (weakenCutoffTmVar c14 k61) (appendHvl k62 k61) (appendHvl k63 k61))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma weaken_shifthvl_TyVar  :
    (forall (k61 : Hvl) {c14 : (Cutoff TyVar)} {k62 : Hvl} {k63 : Hvl} ,
      (shifthvl_TyVar c14 k62 k63) -> (shifthvl_TyVar (weakenCutoffTyVar c14 k61) (appendHvl k62 k61) (appendHvl k63 k61))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma shift_TmVar__wfindex_TmVar  :
    (forall (c14 : (Cutoff TmVar)) (k61 : Hvl) (k62 : Hvl) (ins : (shifthvl_TmVar c14 k61 k62)) (x71 : (Index TmVar)) ,
      (wfindex k61 x71) -> (wfindex k62 (shift_TmVar_Index c14 x71))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TmVar__wfindex_TyVar  :
    (forall (c14 : (Cutoff TmVar)) (k61 : Hvl) (k62 : Hvl) (ins : (shifthvl_TmVar c14 k61 k62)) (X43 : (Index TyVar)) ,
      (wfindex k61 X43) -> (wfindex k62 X43)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TyVar__wfindex_TmVar  :
    (forall (c14 : (Cutoff TyVar)) (k61 : Hvl) (k62 : Hvl) (ins : (shifthvl_TyVar c14 k61 k62)) (x71 : (Index TmVar)) ,
      (wfindex k61 x71) -> (wfindex k62 x71)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TyVar__wfindex_TyVar  :
    (forall (c14 : (Cutoff TyVar)) (k61 : Hvl) (k62 : Hvl) (ins : (shifthvl_TyVar c14 k61 k62)) (X43 : (Index TyVar)) ,
      (wfindex k61 X43) -> (wfindex k62 (shift_TyVar_Index c14 X43))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Definition shift_TmVar__wfTy : (forall (k61 : Hvl) ,
    (forall (S40 : Ty) (wf : (wfTy k61 S40)) ,
      (forall (c14 : (Cutoff TmVar)) (k62 : Hvl) ,
        (shifthvl_TmVar c14 k61 k62) -> (wfTy k62 S40)))) := (ind_wfTy (fun (k61 : Hvl) (S40 : Ty) (wf : (wfTy k61 S40)) =>
    (forall (c14 : (Cutoff TmVar)) (k62 : Hvl) ,
      (shifthvl_TmVar c14 k61 k62) -> (wfTy k62 S40))) (fun (k61 : Hvl) (X43 : (Index TyVar)) (wfi : (wfindex k61 X43)) (c14 : (Cutoff TmVar)) (k62 : Hvl) (ins : (shifthvl_TmVar c14 k61 k62)) =>
    (wfTy_TVar k62 (shift_TmVar__wfindex_TyVar c14 k61 k62 ins X43 wfi))) (fun (k61 : Hvl) (T1 : Ty) (wf : (wfTy k61 T1)) IHT0 (T2 : Ty) (wf0 : (wfTy k61 T2)) IHT3 (c14 : (Cutoff TmVar)) (k62 : Hvl) (ins : (shifthvl_TmVar c14 k61 k62)) =>
    (wfTy_TArr k62 (IHT0 c14 k62 (weaken_shifthvl_TmVar H0 ins)) (IHT3 c14 k62 (weaken_shifthvl_TmVar H0 ins)))) (fun (k61 : Hvl) (T : Ty) (wf : (wfTy (HS TyVar k61) T)) IHT0 (c14 : (Cutoff TmVar)) (k62 : Hvl) (ins : (shifthvl_TmVar c14 k61 k62)) =>
    (wfTy_TAll k62 (IHT0 c14 (HS TyVar k62) (weaken_shifthvl_TmVar (HS TyVar H0) ins))))).
  Definition shift_TyVar__wfTy : (forall (k61 : Hvl) ,
    (forall (S40 : Ty) (wf : (wfTy k61 S40)) ,
      (forall (c14 : (Cutoff TyVar)) (k62 : Hvl) ,
        (shifthvl_TyVar c14 k61 k62) -> (wfTy k62 (shift_TyVar_Ty c14 S40))))) := (ind_wfTy (fun (k61 : Hvl) (S40 : Ty) (wf : (wfTy k61 S40)) =>
    (forall (c14 : (Cutoff TyVar)) (k62 : Hvl) ,
      (shifthvl_TyVar c14 k61 k62) -> (wfTy k62 (shift_TyVar_Ty c14 S40)))) (fun (k61 : Hvl) (X43 : (Index TyVar)) (wfi : (wfindex k61 X43)) (c14 : (Cutoff TyVar)) (k62 : Hvl) (ins : (shifthvl_TyVar c14 k61 k62)) =>
    (wfTy_TVar k62 (shift_TyVar__wfindex_TyVar c14 k61 k62 ins X43 wfi))) (fun (k61 : Hvl) (T1 : Ty) (wf : (wfTy k61 T1)) IHT0 (T2 : Ty) (wf0 : (wfTy k61 T2)) IHT3 (c14 : (Cutoff TyVar)) (k62 : Hvl) (ins : (shifthvl_TyVar c14 k61 k62)) =>
    (wfTy_TArr k62 (IHT0 c14 k62 (weaken_shifthvl_TyVar H0 ins)) (IHT3 c14 k62 (weaken_shifthvl_TyVar H0 ins)))) (fun (k61 : Hvl) (T : Ty) (wf : (wfTy (HS TyVar k61) T)) IHT0 (c14 : (Cutoff TyVar)) (k62 : Hvl) (ins : (shifthvl_TyVar c14 k61 k62)) =>
    (wfTy_TAll k62 (IHT0 (CS c14) (HS TyVar k62) (weaken_shifthvl_TyVar (HS TyVar H0) ins))))).
  Definition shift_TmVar__wfDecls_Tm : (forall (k61 : Hvl) ,
    (forall (d92 : Decls) (wf : (wfDecls k61 d92)) ,
      (forall (c14 : (Cutoff TmVar)) (k62 : Hvl) ,
        (shifthvl_TmVar c14 k61 k62) -> (wfDecls k62 (shift_TmVar_Decls c14 d92)))) /\
    (forall (s21 : Tm) (wf : (wfTm k61 s21)) ,
      (forall (c14 : (Cutoff TmVar)) (k62 : Hvl) ,
        (shifthvl_TmVar c14 k61 k62) -> (wfTm k62 (shift_TmVar_Tm c14 s21))))) := (ind_wfDecls_Tm (fun (k61 : Hvl) (d92 : Decls) (wf : (wfDecls k61 d92)) =>
    (forall (c14 : (Cutoff TmVar)) (k62 : Hvl) ,
      (shifthvl_TmVar c14 k61 k62) -> (wfDecls k62 (shift_TmVar_Decls c14 d92)))) (fun (k61 : Hvl) (s21 : Tm) (wf : (wfTm k61 s21)) =>
    (forall (c14 : (Cutoff TmVar)) (k62 : Hvl) ,
      (shifthvl_TmVar c14 k61 k62) -> (wfTm k62 (shift_TmVar_Tm c14 s21)))) (fun (k61 : Hvl) (c14 : (Cutoff TmVar)) (k62 : Hvl) (ins : (shifthvl_TmVar c14 k61 k62)) =>
    (wfDecls_DNil k62)) (fun (k61 : Hvl) (t : Tm) (wf : (wfTm k61 t)) IHt132 (d : Decls) (wf0 : (wfDecls (HS TmVar k61) d)) IHd (c14 : (Cutoff TmVar)) (k62 : Hvl) (ins : (shifthvl_TmVar c14 k61 k62)) =>
    (wfDecls_DCons k62 (IHt132 c14 k62 (weaken_shifthvl_TmVar H0 ins)) (IHd (CS c14) (HS TmVar k62) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k61 : Hvl) (x71 : (Index TmVar)) (wfi : (wfindex k61 x71)) (c14 : (Cutoff TmVar)) (k62 : Hvl) (ins : (shifthvl_TmVar c14 k61 k62)) =>
    (wfTm_Var k62 (shift_TmVar__wfindex_TmVar c14 k61 k62 ins x71 wfi))) (fun (k61 : Hvl) (T : Ty) (wf : (wfTy k61 T)) (t : Tm) (wf0 : (wfTm (HS TmVar k61) t)) IHt132 (c14 : (Cutoff TmVar)) (k62 : Hvl) (ins : (shifthvl_TmVar c14 k61 k62)) =>
    (wfTm_Abs k62 (shift_TmVar__wfTy k61 T wf c14 k62 (weaken_shifthvl_TmVar H0 ins)) (IHt132 (CS c14) (HS TmVar k62) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k61 : Hvl) (t1 : Tm) (wf : (wfTm k61 t1)) IHt132 (t2 : Tm) (wf0 : (wfTm k61 t2)) IHt133 (c14 : (Cutoff TmVar)) (k62 : Hvl) (ins : (shifthvl_TmVar c14 k61 k62)) =>
    (wfTm_App k62 (IHt132 c14 k62 (weaken_shifthvl_TmVar H0 ins)) (IHt133 c14 k62 (weaken_shifthvl_TmVar H0 ins)))) (fun (k61 : Hvl) (t : Tm) (wf : (wfTm (HS TyVar k61) t)) IHt132 (c14 : (Cutoff TmVar)) (k62 : Hvl) (ins : (shifthvl_TmVar c14 k61 k62)) =>
    (wfTm_TAbs k62 (IHt132 c14 (HS TyVar k62) (weaken_shifthvl_TmVar (HS TyVar H0) ins)))) (fun (k61 : Hvl) (t : Tm) (wf : (wfTm k61 t)) IHt132 (T : Ty) (wf0 : (wfTy k61 T)) (c14 : (Cutoff TmVar)) (k62 : Hvl) (ins : (shifthvl_TmVar c14 k61 k62)) =>
    (wfTm_TApp k62 (IHt132 c14 k62 (weaken_shifthvl_TmVar H0 ins)) (shift_TmVar__wfTy k61 T wf0 c14 k62 (weaken_shifthvl_TmVar H0 ins)))) (fun (k61 : Hvl) (d : Decls) (wf : (wfDecls k61 d)) IHd (t : Tm) (wf0 : (wfTm (appendHvl k61 (bind d)) t)) IHt132 (c14 : (Cutoff TmVar)) (k62 : Hvl) (ins : (shifthvl_TmVar c14 k61 k62)) =>
    (wfTm_Let k62 (IHd c14 k62 (weaken_shifthvl_TmVar H0 ins)) (eq_ind2 wfTm (f_equal2 appendHvl (eq_refl k62) (f_equal2 appendHvl (eq_sym (stability_shift_TmVar__bind _ _)) (eq_refl H0))) eq_refl (IHt132 (weakenCutoffTmVar c14 (bind d)) (appendHvl k62 (bind d)) (weaken_shifthvl_TmVar (bind d) ins)))))).
  Lemma shift_TmVar__wfDecls (k61 : Hvl) :
    (forall (d92 : Decls) (wf : (wfDecls k61 d92)) ,
      (forall (c14 : (Cutoff TmVar)) (k62 : Hvl) ,
        (shifthvl_TmVar c14 k61 k62) -> (wfDecls k62 (shift_TmVar_Decls c14 d92)))).
  Proof.
    pose proof ((shift_TmVar__wfDecls_Tm k61)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_TmVar__wfTm (k61 : Hvl) :
    (forall (s21 : Tm) (wf0 : (wfTm k61 s21)) ,
      (forall (c15 : (Cutoff TmVar)) (k63 : Hvl) ,
        (shifthvl_TmVar c15 k61 k63) -> (wfTm k63 (shift_TmVar_Tm c15 s21)))).
  Proof.
    pose proof ((shift_TmVar__wfDecls_Tm k61)).
    destruct_conjs .
    easy .
  Qed.
  Definition shift_TyVar__wfDecls_Tm : (forall (k61 : Hvl) ,
    (forall (d92 : Decls) (wf : (wfDecls k61 d92)) ,
      (forall (c14 : (Cutoff TyVar)) (k62 : Hvl) ,
        (shifthvl_TyVar c14 k61 k62) -> (wfDecls k62 (shift_TyVar_Decls c14 d92)))) /\
    (forall (s21 : Tm) (wf : (wfTm k61 s21)) ,
      (forall (c14 : (Cutoff TyVar)) (k62 : Hvl) ,
        (shifthvl_TyVar c14 k61 k62) -> (wfTm k62 (shift_TyVar_Tm c14 s21))))) := (ind_wfDecls_Tm (fun (k61 : Hvl) (d92 : Decls) (wf : (wfDecls k61 d92)) =>
    (forall (c14 : (Cutoff TyVar)) (k62 : Hvl) ,
      (shifthvl_TyVar c14 k61 k62) -> (wfDecls k62 (shift_TyVar_Decls c14 d92)))) (fun (k61 : Hvl) (s21 : Tm) (wf : (wfTm k61 s21)) =>
    (forall (c14 : (Cutoff TyVar)) (k62 : Hvl) ,
      (shifthvl_TyVar c14 k61 k62) -> (wfTm k62 (shift_TyVar_Tm c14 s21)))) (fun (k61 : Hvl) (c14 : (Cutoff TyVar)) (k62 : Hvl) (ins : (shifthvl_TyVar c14 k61 k62)) =>
    (wfDecls_DNil k62)) (fun (k61 : Hvl) (t : Tm) (wf : (wfTm k61 t)) IHt132 (d : Decls) (wf0 : (wfDecls (HS TmVar k61) d)) IHd (c14 : (Cutoff TyVar)) (k62 : Hvl) (ins : (shifthvl_TyVar c14 k61 k62)) =>
    (wfDecls_DCons k62 (IHt132 c14 k62 (weaken_shifthvl_TyVar H0 ins)) (IHd c14 (HS TmVar k62) (weaken_shifthvl_TyVar (HS TmVar H0) ins)))) (fun (k61 : Hvl) (x71 : (Index TmVar)) (wfi : (wfindex k61 x71)) (c14 : (Cutoff TyVar)) (k62 : Hvl) (ins : (shifthvl_TyVar c14 k61 k62)) =>
    (wfTm_Var k62 (shift_TyVar__wfindex_TmVar c14 k61 k62 ins x71 wfi))) (fun (k61 : Hvl) (T : Ty) (wf : (wfTy k61 T)) (t : Tm) (wf0 : (wfTm (HS TmVar k61) t)) IHt132 (c14 : (Cutoff TyVar)) (k62 : Hvl) (ins : (shifthvl_TyVar c14 k61 k62)) =>
    (wfTm_Abs k62 (shift_TyVar__wfTy k61 T wf c14 k62 (weaken_shifthvl_TyVar H0 ins)) (IHt132 c14 (HS TmVar k62) (weaken_shifthvl_TyVar (HS TmVar H0) ins)))) (fun (k61 : Hvl) (t1 : Tm) (wf : (wfTm k61 t1)) IHt132 (t2 : Tm) (wf0 : (wfTm k61 t2)) IHt133 (c14 : (Cutoff TyVar)) (k62 : Hvl) (ins : (shifthvl_TyVar c14 k61 k62)) =>
    (wfTm_App k62 (IHt132 c14 k62 (weaken_shifthvl_TyVar H0 ins)) (IHt133 c14 k62 (weaken_shifthvl_TyVar H0 ins)))) (fun (k61 : Hvl) (t : Tm) (wf : (wfTm (HS TyVar k61) t)) IHt132 (c14 : (Cutoff TyVar)) (k62 : Hvl) (ins : (shifthvl_TyVar c14 k61 k62)) =>
    (wfTm_TAbs k62 (IHt132 (CS c14) (HS TyVar k62) (weaken_shifthvl_TyVar (HS TyVar H0) ins)))) (fun (k61 : Hvl) (t : Tm) (wf : (wfTm k61 t)) IHt132 (T : Ty) (wf0 : (wfTy k61 T)) (c14 : (Cutoff TyVar)) (k62 : Hvl) (ins : (shifthvl_TyVar c14 k61 k62)) =>
    (wfTm_TApp k62 (IHt132 c14 k62 (weaken_shifthvl_TyVar H0 ins)) (shift_TyVar__wfTy k61 T wf0 c14 k62 (weaken_shifthvl_TyVar H0 ins)))) (fun (k61 : Hvl) (d : Decls) (wf : (wfDecls k61 d)) IHd (t : Tm) (wf0 : (wfTm (appendHvl k61 (bind d)) t)) IHt132 (c14 : (Cutoff TyVar)) (k62 : Hvl) (ins : (shifthvl_TyVar c14 k61 k62)) =>
    (wfTm_Let k62 (IHd c14 k62 (weaken_shifthvl_TyVar H0 ins)) (eq_ind2 wfTm (f_equal2 appendHvl (eq_refl k62) (f_equal2 appendHvl (eq_sym (stability_shift_TyVar__bind _ _)) (eq_refl H0))) eq_refl (IHt132 (weakenCutoffTyVar c14 (bind d)) (appendHvl k62 (bind d)) (weaken_shifthvl_TyVar (bind d) ins)))))).
  Lemma shift_TyVar__wfDecls (k61 : Hvl) :
    (forall (d92 : Decls) (wf : (wfDecls k61 d92)) ,
      (forall (c14 : (Cutoff TyVar)) (k62 : Hvl) ,
        (shifthvl_TyVar c14 k61 k62) -> (wfDecls k62 (shift_TyVar_Decls c14 d92)))).
  Proof.
    pose proof ((shift_TyVar__wfDecls_Tm k61)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_TyVar__wfTm (k61 : Hvl) :
    (forall (s21 : Tm) (wf0 : (wfTm k61 s21)) ,
      (forall (c15 : (Cutoff TyVar)) (k63 : Hvl) ,
        (shifthvl_TyVar c15 k61 k63) -> (wfTm k63 (shift_TyVar_Tm c15 s21)))).
  Proof.
    pose proof ((shift_TyVar__wfDecls_Tm k61)).
    destruct_conjs .
    easy .
  Qed.
End ShiftWellFormed.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : infra.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : shift.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : shift_wf.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : wf.
 Hint Resolve shift_TmVar__wfDecls shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfDecls shift_TyVar__wfTm shift_TyVar__wfTy : infra.
 Hint Resolve shift_TmVar__wfDecls shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfDecls shift_TyVar__wfTm shift_TyVar__wfTy : shift.
 Hint Resolve shift_TmVar__wfDecls shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfDecls shift_TyVar__wfTm shift_TyVar__wfTy : shift_wf.
 Hint Resolve shift_TmVar__wfDecls shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfDecls shift_TyVar__wfTm shift_TyVar__wfTy : wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : infra.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift_wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : infra.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift_wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : weaken.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : wf.
Section SubstWellFormed.
  Inductive substhvl_TmVar (k61 : Hvl) : (forall (d92 : (Trace TmVar)) (k62 : Hvl) (k63 : Hvl) ,
    Prop) :=
    | substhvl_TmVar_here :
        (substhvl_TmVar k61 X0 (HS TmVar k61) k61)
    | substhvl_TmVar_there_TmVar
        {d92 : (Trace TmVar)}
        {k62 : Hvl} {k63 : Hvl} :
        (substhvl_TmVar k61 d92 k62 k63) -> (substhvl_TmVar k61 (XS TmVar d92) (HS TmVar k62) (HS TmVar k63))
    | substhvl_TmVar_there_TyVar
        {d92 : (Trace TmVar)}
        {k62 : Hvl} {k63 : Hvl} :
        (substhvl_TmVar k61 d92 k62 k63) -> (substhvl_TmVar k61 (XS TyVar d92) (HS TyVar k62) (HS TyVar k63)).
  Inductive substhvl_TyVar (k61 : Hvl) : (forall (d92 : (Trace TyVar)) (k62 : Hvl) (k63 : Hvl) ,
    Prop) :=
    | substhvl_TyVar_here :
        (substhvl_TyVar k61 X0 (HS TyVar k61) k61)
    | substhvl_TyVar_there_TmVar
        {d92 : (Trace TyVar)}
        {k62 : Hvl} {k63 : Hvl} :
        (substhvl_TyVar k61 d92 k62 k63) -> (substhvl_TyVar k61 (XS TmVar d92) (HS TmVar k62) (HS TmVar k63))
    | substhvl_TyVar_there_TyVar
        {d92 : (Trace TyVar)}
        {k62 : Hvl} {k63 : Hvl} :
        (substhvl_TyVar k61 d92 k62 k63) -> (substhvl_TyVar k61 (XS TyVar d92) (HS TyVar k62) (HS TyVar k63)).
  Lemma weaken_substhvl_TmVar  :
    (forall {k62 : Hvl} (k61 : Hvl) {d92 : (Trace TmVar)} {k63 : Hvl} {k64 : Hvl} ,
      (substhvl_TmVar k62 d92 k63 k64) -> (substhvl_TmVar k62 (weakenTrace d92 k61) (appendHvl k63 k61) (appendHvl k64 k61))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma weaken_substhvl_TyVar  :
    (forall {k62 : Hvl} (k61 : Hvl) {d92 : (Trace TyVar)} {k63 : Hvl} {k64 : Hvl} ,
      (substhvl_TyVar k62 d92 k63 k64) -> (substhvl_TyVar k62 (weakenTrace d92 k61) (appendHvl k63 k61) (appendHvl k64 k61))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma substhvl_TmVar_wfindex_TmVar {k63 : Hvl} {s21 : Tm} (wft : (wfTm k63 s21)) :
    (forall {d92 : (Trace TmVar)} {k64 : Hvl} {k65 : Hvl} ,
      (substhvl_TmVar k63 d92 k64 k65) -> (forall {x71 : (Index TmVar)} ,
        (wfindex k64 x71) -> (wfTm k65 (subst_TmVar_Index d92 s21 x71)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TyVar_wfindex_TyVar {k63 : Hvl} {S41 : Ty} (wft : (wfTy k63 S41)) :
    (forall {d92 : (Trace TyVar)} {k64 : Hvl} {k65 : Hvl} ,
      (substhvl_TyVar k63 d92 k64 k65) -> (forall {X43 : (Index TyVar)} ,
        (wfindex k64 X43) -> (wfTy k65 (subst_TyVar_Index d92 S41 X43)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TmVar_wfindex_TyVar {k63 : Hvl} :
    (forall {d92 : (Trace TmVar)} {k64 : Hvl} {k65 : Hvl} ,
      (substhvl_TmVar k63 d92 k64 k65) -> (forall {X43 : (Index TyVar)} ,
        (wfindex k64 X43) -> (wfindex k65 X43))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_TyVar_wfindex_TmVar {k63 : Hvl} :
    (forall {d92 : (Trace TyVar)} {k64 : Hvl} {k65 : Hvl} ,
      (substhvl_TyVar k63 d92 k64 k65) -> (forall {x71 : (Index TmVar)} ,
        (wfindex k64 x71) -> (wfindex k65 x71))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Definition substhvl_TmVar_wfTy {k63 : Hvl} : (forall (k64 : Hvl) ,
    (forall (S41 : Ty) (wf0 : (wfTy k64 S41)) ,
      (forall {d92 : (Trace TmVar)} {k65 : Hvl} ,
        (substhvl_TmVar k63 d92 k64 k65) -> (wfTy k65 S41)))) := (ind_wfTy (fun (k64 : Hvl) (S41 : Ty) (wf0 : (wfTy k64 S41)) =>
    (forall {d92 : (Trace TmVar)} {k65 : Hvl} ,
      (substhvl_TmVar k63 d92 k64 k65) -> (wfTy k65 S41))) (fun (k64 : Hvl) {X43 : (Index TyVar)} (wfi : (wfindex k64 X43)) {d92 : (Trace TmVar)} {k65 : Hvl} (del : (substhvl_TmVar k63 d92 k64 k65)) =>
    (wfTy_TVar k65 (substhvl_TmVar_wfindex_TyVar del wfi))) (fun (k64 : Hvl) (T1 : Ty) (wf0 : (wfTy k64 T1)) IHT0 (T2 : Ty) (wf1 : (wfTy k64 T2)) IHT3 {d92 : (Trace TmVar)} {k65 : Hvl} (del : (substhvl_TmVar k63 d92 k64 k65)) =>
    (wfTy_TArr k65 (IHT0 (weakenTrace d92 H0) k65 (weaken_substhvl_TmVar H0 del)) (IHT3 (weakenTrace d92 H0) k65 (weaken_substhvl_TmVar H0 del)))) (fun (k64 : Hvl) (T : Ty) (wf0 : (wfTy (HS TyVar k64) T)) IHT0 {d92 : (Trace TmVar)} {k65 : Hvl} (del : (substhvl_TmVar k63 d92 k64 k65)) =>
    (wfTy_TAll k65 (IHT0 (weakenTrace d92 (HS TyVar H0)) (HS TyVar k65) (weaken_substhvl_TmVar (HS TyVar H0) del))))).
  Definition substhvl_TyVar_wfTy {k63 : Hvl} {S41 : Ty} (wf : (wfTy k63 S41)) : (forall (k64 : Hvl) ,
    (forall (S42 : Ty) (wf0 : (wfTy k64 S42)) ,
      (forall {d92 : (Trace TyVar)} {k65 : Hvl} ,
        (substhvl_TyVar k63 d92 k64 k65) -> (wfTy k65 (subst_TyVar_Ty d92 S41 S42))))) := (ind_wfTy (fun (k64 : Hvl) (S42 : Ty) (wf0 : (wfTy k64 S42)) =>
    (forall {d92 : (Trace TyVar)} {k65 : Hvl} ,
      (substhvl_TyVar k63 d92 k64 k65) -> (wfTy k65 (subst_TyVar_Ty d92 S41 S42)))) (fun (k64 : Hvl) {X43 : (Index TyVar)} (wfi : (wfindex k64 X43)) {d92 : (Trace TyVar)} {k65 : Hvl} (del : (substhvl_TyVar k63 d92 k64 k65)) =>
    (substhvl_TyVar_wfindex_TyVar wf del wfi)) (fun (k64 : Hvl) (T1 : Ty) (wf0 : (wfTy k64 T1)) IHT0 (T2 : Ty) (wf1 : (wfTy k64 T2)) IHT3 {d92 : (Trace TyVar)} {k65 : Hvl} (del : (substhvl_TyVar k63 d92 k64 k65)) =>
    (wfTy_TArr k65 (IHT0 (weakenTrace d92 H0) k65 (weaken_substhvl_TyVar H0 del)) (IHT3 (weakenTrace d92 H0) k65 (weaken_substhvl_TyVar H0 del)))) (fun (k64 : Hvl) (T : Ty) (wf0 : (wfTy (HS TyVar k64) T)) IHT0 {d92 : (Trace TyVar)} {k65 : Hvl} (del : (substhvl_TyVar k63 d92 k64 k65)) =>
    (wfTy_TAll k65 (IHT0 (weakenTrace d92 (HS TyVar H0)) (HS TyVar k65) (weaken_substhvl_TyVar (HS TyVar H0) del))))).
  Definition substhvl_TmVar_wfDecls_Tm {k63 : Hvl} {s21 : Tm} (wf : (wfTm k63 s21)) : (forall (k64 : Hvl) ,
    (forall (d92 : Decls) (wf0 : (wfDecls k64 d92)) ,
      (forall {d93 : (Trace TmVar)} {k65 : Hvl} ,
        (substhvl_TmVar k63 d93 k64 k65) -> (wfDecls k65 (subst_TmVar_Decls d93 s21 d92)))) /\
    (forall (s22 : Tm) (wf0 : (wfTm k64 s22)) ,
      (forall {d92 : (Trace TmVar)} {k65 : Hvl} ,
        (substhvl_TmVar k63 d92 k64 k65) -> (wfTm k65 (subst_TmVar_Tm d92 s21 s22))))) := (ind_wfDecls_Tm (fun (k64 : Hvl) (d92 : Decls) (wf0 : (wfDecls k64 d92)) =>
    (forall {d93 : (Trace TmVar)} {k65 : Hvl} ,
      (substhvl_TmVar k63 d93 k64 k65) -> (wfDecls k65 (subst_TmVar_Decls d93 s21 d92)))) (fun (k64 : Hvl) (s22 : Tm) (wf0 : (wfTm k64 s22)) =>
    (forall {d92 : (Trace TmVar)} {k65 : Hvl} ,
      (substhvl_TmVar k63 d92 k64 k65) -> (wfTm k65 (subst_TmVar_Tm d92 s21 s22)))) (fun (k64 : Hvl) {d92 : (Trace TmVar)} {k65 : Hvl} (del : (substhvl_TmVar k63 d92 k64 k65)) =>
    (wfDecls_DNil k65)) (fun (k64 : Hvl) (t : Tm) (wf0 : (wfTm k64 t)) IHt132 (d : Decls) (wf1 : (wfDecls (HS TmVar k64) d)) IHd {d92 : (Trace TmVar)} {k65 : Hvl} (del : (substhvl_TmVar k63 d92 k64 k65)) =>
    (wfDecls_DCons k65 (IHt132 (weakenTrace d92 H0) k65 (weaken_substhvl_TmVar H0 del)) (IHd (weakenTrace d92 (HS TmVar H0)) (HS TmVar k65) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k64 : Hvl) {x71 : (Index TmVar)} (wfi : (wfindex k64 x71)) {d92 : (Trace TmVar)} {k65 : Hvl} (del : (substhvl_TmVar k63 d92 k64 k65)) =>
    (substhvl_TmVar_wfindex_TmVar wf del wfi)) (fun (k64 : Hvl) (T : Ty) (wf0 : (wfTy k64 T)) (t : Tm) (wf1 : (wfTm (HS TmVar k64) t)) IHt132 {d92 : (Trace TmVar)} {k65 : Hvl} (del : (substhvl_TmVar k63 d92 k64 k65)) =>
    (wfTm_Abs k65 (substhvl_TmVar_wfTy k64 T wf0 (weaken_substhvl_TmVar H0 del)) (IHt132 (weakenTrace d92 (HS TmVar H0)) (HS TmVar k65) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k64 : Hvl) (t1 : Tm) (wf0 : (wfTm k64 t1)) IHt132 (t2 : Tm) (wf1 : (wfTm k64 t2)) IHt133 {d92 : (Trace TmVar)} {k65 : Hvl} (del : (substhvl_TmVar k63 d92 k64 k65)) =>
    (wfTm_App k65 (IHt132 (weakenTrace d92 H0) k65 (weaken_substhvl_TmVar H0 del)) (IHt133 (weakenTrace d92 H0) k65 (weaken_substhvl_TmVar H0 del)))) (fun (k64 : Hvl) (t : Tm) (wf0 : (wfTm (HS TyVar k64) t)) IHt132 {d92 : (Trace TmVar)} {k65 : Hvl} (del : (substhvl_TmVar k63 d92 k64 k65)) =>
    (wfTm_TAbs k65 (IHt132 (weakenTrace d92 (HS TyVar H0)) (HS TyVar k65) (weaken_substhvl_TmVar (HS TyVar H0) del)))) (fun (k64 : Hvl) (t : Tm) (wf0 : (wfTm k64 t)) IHt132 (T : Ty) (wf1 : (wfTy k64 T)) {d92 : (Trace TmVar)} {k65 : Hvl} (del : (substhvl_TmVar k63 d92 k64 k65)) =>
    (wfTm_TApp k65 (IHt132 (weakenTrace d92 H0) k65 (weaken_substhvl_TmVar H0 del)) (substhvl_TmVar_wfTy k64 T wf1 (weaken_substhvl_TmVar H0 del)))) (fun (k64 : Hvl) (d : Decls) (wf0 : (wfDecls k64 d)) IHd (t : Tm) (wf1 : (wfTm (appendHvl k64 (bind d)) t)) IHt132 {d92 : (Trace TmVar)} {k65 : Hvl} (del : (substhvl_TmVar k63 d92 k64 k65)) =>
    (wfTm_Let k65 (IHd (weakenTrace d92 H0) k65 (weaken_substhvl_TmVar H0 del)) (eq_ind2 wfTm (f_equal2 appendHvl (eq_refl k65) (f_equal2 appendHvl (eq_sym (stability_subst_TmVar__bind _ _ _)) (eq_refl H0))) eq_refl (IHt132 (weakenTrace d92 (bind d)) (appendHvl k65 (bind d)) (weaken_substhvl_TmVar (bind d) del)))))).
  Lemma substhvl_TmVar_wfDecls {k63 : Hvl} {s21 : Tm} (wf : (wfTm k63 s21)) (k64 : Hvl) (d92 : Decls) (wf0 : (wfDecls k64 d92)) :
    (forall {d93 : (Trace TmVar)} {k65 : Hvl} ,
      (substhvl_TmVar k63 d93 k64 k65) -> (wfDecls k65 (subst_TmVar_Decls d93 s21 d92))).
  Proof.
    apply ((substhvl_TmVar_wfDecls_Tm wf k64)).
    auto .
  Qed.
  Lemma substhvl_TmVar_wfTm {k63 : Hvl} {s21 : Tm} (wf : (wfTm k63 s21)) (k64 : Hvl) (s22 : Tm) (wf1 : (wfTm k64 s22)) :
    (forall {d94 : (Trace TmVar)} {k66 : Hvl} ,
      (substhvl_TmVar k63 d94 k64 k66) -> (wfTm k66 (subst_TmVar_Tm d94 s21 s22))).
  Proof.
    apply ((substhvl_TmVar_wfDecls_Tm wf k64)).
    auto .
  Qed.
  Definition substhvl_TyVar_wfDecls_Tm {k63 : Hvl} {S41 : Ty} (wf : (wfTy k63 S41)) : (forall (k64 : Hvl) ,
    (forall (d92 : Decls) (wf0 : (wfDecls k64 d92)) ,
      (forall {d93 : (Trace TyVar)} {k65 : Hvl} ,
        (substhvl_TyVar k63 d93 k64 k65) -> (wfDecls k65 (subst_TyVar_Decls d93 S41 d92)))) /\
    (forall (s21 : Tm) (wf0 : (wfTm k64 s21)) ,
      (forall {d92 : (Trace TyVar)} {k65 : Hvl} ,
        (substhvl_TyVar k63 d92 k64 k65) -> (wfTm k65 (subst_TyVar_Tm d92 S41 s21))))) := (ind_wfDecls_Tm (fun (k64 : Hvl) (d92 : Decls) (wf0 : (wfDecls k64 d92)) =>
    (forall {d93 : (Trace TyVar)} {k65 : Hvl} ,
      (substhvl_TyVar k63 d93 k64 k65) -> (wfDecls k65 (subst_TyVar_Decls d93 S41 d92)))) (fun (k64 : Hvl) (s21 : Tm) (wf0 : (wfTm k64 s21)) =>
    (forall {d92 : (Trace TyVar)} {k65 : Hvl} ,
      (substhvl_TyVar k63 d92 k64 k65) -> (wfTm k65 (subst_TyVar_Tm d92 S41 s21)))) (fun (k64 : Hvl) {d92 : (Trace TyVar)} {k65 : Hvl} (del : (substhvl_TyVar k63 d92 k64 k65)) =>
    (wfDecls_DNil k65)) (fun (k64 : Hvl) (t : Tm) (wf0 : (wfTm k64 t)) IHt132 (d : Decls) (wf1 : (wfDecls (HS TmVar k64) d)) IHd {d92 : (Trace TyVar)} {k65 : Hvl} (del : (substhvl_TyVar k63 d92 k64 k65)) =>
    (wfDecls_DCons k65 (IHt132 (weakenTrace d92 H0) k65 (weaken_substhvl_TyVar H0 del)) (IHd (weakenTrace d92 (HS TmVar H0)) (HS TmVar k65) (weaken_substhvl_TyVar (HS TmVar H0) del)))) (fun (k64 : Hvl) {x71 : (Index TmVar)} (wfi : (wfindex k64 x71)) {d92 : (Trace TyVar)} {k65 : Hvl} (del : (substhvl_TyVar k63 d92 k64 k65)) =>
    (wfTm_Var k65 (substhvl_TyVar_wfindex_TmVar del wfi))) (fun (k64 : Hvl) (T : Ty) (wf0 : (wfTy k64 T)) (t : Tm) (wf1 : (wfTm (HS TmVar k64) t)) IHt132 {d92 : (Trace TyVar)} {k65 : Hvl} (del : (substhvl_TyVar k63 d92 k64 k65)) =>
    (wfTm_Abs k65 (substhvl_TyVar_wfTy wf k64 T wf0 (weaken_substhvl_TyVar H0 del)) (IHt132 (weakenTrace d92 (HS TmVar H0)) (HS TmVar k65) (weaken_substhvl_TyVar (HS TmVar H0) del)))) (fun (k64 : Hvl) (t1 : Tm) (wf0 : (wfTm k64 t1)) IHt132 (t2 : Tm) (wf1 : (wfTm k64 t2)) IHt133 {d92 : (Trace TyVar)} {k65 : Hvl} (del : (substhvl_TyVar k63 d92 k64 k65)) =>
    (wfTm_App k65 (IHt132 (weakenTrace d92 H0) k65 (weaken_substhvl_TyVar H0 del)) (IHt133 (weakenTrace d92 H0) k65 (weaken_substhvl_TyVar H0 del)))) (fun (k64 : Hvl) (t : Tm) (wf0 : (wfTm (HS TyVar k64) t)) IHt132 {d92 : (Trace TyVar)} {k65 : Hvl} (del : (substhvl_TyVar k63 d92 k64 k65)) =>
    (wfTm_TAbs k65 (IHt132 (weakenTrace d92 (HS TyVar H0)) (HS TyVar k65) (weaken_substhvl_TyVar (HS TyVar H0) del)))) (fun (k64 : Hvl) (t : Tm) (wf0 : (wfTm k64 t)) IHt132 (T : Ty) (wf1 : (wfTy k64 T)) {d92 : (Trace TyVar)} {k65 : Hvl} (del : (substhvl_TyVar k63 d92 k64 k65)) =>
    (wfTm_TApp k65 (IHt132 (weakenTrace d92 H0) k65 (weaken_substhvl_TyVar H0 del)) (substhvl_TyVar_wfTy wf k64 T wf1 (weaken_substhvl_TyVar H0 del)))) (fun (k64 : Hvl) (d : Decls) (wf0 : (wfDecls k64 d)) IHd (t : Tm) (wf1 : (wfTm (appendHvl k64 (bind d)) t)) IHt132 {d92 : (Trace TyVar)} {k65 : Hvl} (del : (substhvl_TyVar k63 d92 k64 k65)) =>
    (wfTm_Let k65 (IHd (weakenTrace d92 H0) k65 (weaken_substhvl_TyVar H0 del)) (eq_ind2 wfTm (f_equal2 appendHvl (eq_refl k65) (f_equal2 appendHvl (eq_sym (stability_subst_TyVar__bind _ _ _)) (eq_refl H0))) eq_refl (IHt132 (weakenTrace d92 (bind d)) (appendHvl k65 (bind d)) (weaken_substhvl_TyVar (bind d) del)))))).
  Lemma substhvl_TyVar_wfDecls {k63 : Hvl} {S41 : Ty} (wf : (wfTy k63 S41)) (k64 : Hvl) (d92 : Decls) (wf0 : (wfDecls k64 d92)) :
    (forall {d93 : (Trace TyVar)} {k65 : Hvl} ,
      (substhvl_TyVar k63 d93 k64 k65) -> (wfDecls k65 (subst_TyVar_Decls d93 S41 d92))).
  Proof.
    apply ((substhvl_TyVar_wfDecls_Tm wf k64)).
    auto .
  Qed.
  Lemma substhvl_TyVar_wfTm {k63 : Hvl} {S41 : Ty} (wf : (wfTy k63 S41)) (k64 : Hvl) (s21 : Tm) (wf1 : (wfTm k64 s21)) :
    (forall {d94 : (Trace TyVar)} {k66 : Hvl} ,
      (substhvl_TyVar k63 d94 k64 k66) -> (wfTm k66 (subst_TyVar_Tm d94 S41 s21))).
  Proof.
    apply ((substhvl_TyVar_wfDecls_Tm wf k64)).
    auto .
  Qed.
End SubstWellFormed.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : infra.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst_wf.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : wf.
 Hint Resolve substhvl_TmVar_wfDecls substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfDecls substhvl_TyVar_wfTm substhvl_TyVar_wfTy : infra.
 Hint Resolve substhvl_TmVar_wfDecls substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfDecls substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst.
 Hint Resolve substhvl_TmVar_wfDecls substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfDecls substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst_wf.
 Hint Resolve substhvl_TmVar_wfDecls substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfDecls substhvl_TyVar_wfTm substhvl_TyVar_wfTy : wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : infra.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst_wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : wf.
Fixpoint subhvl_TmVar (k63 : Hvl) {struct k63} : Prop :=
  match k63 with
    | (H0) => True
    | (HS a k63) => match a with
      | (TmVar) => (subhvl_TmVar k63)
      | _ => False
    end
  end.
Lemma subhvl_TmVar_append  :
  (forall (k63 : Hvl) (k64 : Hvl) ,
    (subhvl_TmVar k63) -> (subhvl_TmVar k64) -> (subhvl_TmVar (appendHvl k63 k64))).
Proof.
  needleGenericSubHvlAppend .
Qed.
 Hint Resolve subhvl_TmVar_append : infra.
 Hint Resolve subhvl_TmVar_append : wf.
Lemma wfTy_strengthen_subhvl_TmVar  :
  (forall (k62 : Hvl) (k61 : Hvl) (S40 : Ty) ,
    (subhvl_TmVar k62) -> (wfTy (appendHvl k61 k62) (weakenTy S40 k62)) -> (wfTy k61 S40)).
Proof.
  needleGenericWellformedStrengthen .
Qed.
 Hint Extern 2 ((wfTy _ _)) => match goal with
  | H : (wfTy (appendHvl _ _) (weakenTy _ _)) |- _ => apply (wfTy_strengthen_subhvl_TmVar) in H
end : infra wf.
Section Context.
  Inductive Ctx : Type :=
    | empty 
    | evar (G : Ctx) (T : Ty)
    | etvar (G : Ctx).
  Fixpoint appendCtx (G : Ctx) (G0 : Ctx) : Ctx :=
    match G0 with
      | (empty) => G
      | (evar G1 T) => (evar (appendCtx G G1) T)
      | (etvar G1) => (etvar (appendCtx G G1))
    end.
  Fixpoint domainCtx (G : Ctx) : Hvl :=
    match G with
      | (empty) => H0
      | (evar G0 T) => (HS TmVar (domainCtx G0))
      | (etvar G0) => (HS TyVar (domainCtx G0))
    end.
  Lemma appendCtx_assoc  :
    (forall (G : Ctx) (G0 : Ctx) (G1 : Ctx) ,
      ((appendCtx G (appendCtx G0 G1)) =
      (appendCtx (appendCtx G G0) G1))).
  Proof.
    needleGenericAppendEnvAssoc .
  Qed.
  Lemma domainCtx_appendCtx  :
    (forall (G : Ctx) (G0 : Ctx) ,
      ((domainCtx (appendCtx G G0)) =
      (appendHvl (domainCtx G) (domainCtx G0)))).
  Proof.
    needleGenericDomainEnvAppendEnv .
  Qed.
  Fixpoint shift_TmVar_Ctx (c14 : (Cutoff TmVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shift_TmVar_Ctx c14 G0) T)
      | (etvar G0) => (etvar (shift_TmVar_Ctx c14 G0))
    end.
  Fixpoint shift_TyVar_Ctx (c14 : (Cutoff TyVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shift_TyVar_Ctx c14 G0) (shift_TyVar_Ty (weakenCutoffTyVar c14 (domainCtx G0)) T))
      | (etvar G0) => (etvar (shift_TyVar_Ctx c14 G0))
    end.
  Fixpoint subst_TmVar_Ctx (d92 : (Trace TmVar)) (s21 : Tm) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (subst_TmVar_Ctx d92 s21 G0) T)
      | (etvar G0) => (etvar (subst_TmVar_Ctx d92 s21 G0))
    end.
  Fixpoint subst_TyVar_Ctx (d92 : (Trace TyVar)) (S41 : Ty) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (subst_TyVar_Ctx d92 S41 G0) (subst_TyVar_Ty (weakenTrace d92 (domainCtx G0)) S41 T))
      | (etvar G0) => (etvar (subst_TyVar_Ctx d92 S41 G0))
    end.
  Lemma domainCtx_shift_TmVar_Ctx  :
    (forall (c14 : (Cutoff TmVar)) (G : Ctx) ,
      ((domainCtx (shift_TmVar_Ctx c14 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_shift_TyVar_Ctx  :
    (forall (c14 : (Cutoff TyVar)) (G : Ctx) ,
      ((domainCtx (shift_TyVar_Ctx c14 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_subst_TmVar_Ctx  :
    (forall (d92 : (Trace TmVar)) (s21 : Tm) (G : Ctx) ,
      ((domainCtx (subst_TmVar_Ctx d92 s21 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
  Lemma domainCtx_subst_TyVar_Ctx  :
    (forall (d92 : (Trace TyVar)) (S41 : Ty) (G : Ctx) ,
      ((domainCtx (subst_TyVar_Ctx d92 S41 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
End Context.
 Hint Rewrite domainCtx_appendCtx : interaction.
 Hint Rewrite domainCtx_appendCtx : env_domain_append.
Section ContextStuff.
  Section ShiftEnvAppendEnv.
    Lemma shift_TmVar_Ctx_appendCtx  :
      (forall (c14 : (Cutoff TmVar)) (G : Ctx) (G0 : Ctx) ,
        ((shift_TmVar_Ctx c14 (appendCtx G G0)) =
        (appendCtx (shift_TmVar_Ctx c14 G) (shift_TmVar_Ctx (weakenCutoffTmVar c14 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
    Lemma shift_TyVar_Ctx_appendCtx  :
      (forall (c14 : (Cutoff TyVar)) (G : Ctx) (G0 : Ctx) ,
        ((shift_TyVar_Ctx c14 (appendCtx G G0)) =
        (appendCtx (shift_TyVar_Ctx c14 G) (shift_TyVar_Ctx (weakenCutoffTyVar c14 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
  End ShiftEnvAppendEnv.
  Section SubstEnvAppendEnv.
    Lemma subst_TmVar_Ctx_appendCtx  :
      (forall (d92 : (Trace TmVar)) (s21 : Tm) (G : Ctx) (G0 : Ctx) ,
        ((subst_TmVar_Ctx d92 s21 (appendCtx G G0)) =
        (appendCtx (subst_TmVar_Ctx d92 s21 G) (subst_TmVar_Ctx (weakenTrace d92 (domainCtx G)) s21 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
    Lemma subst_TyVar_Ctx_appendCtx  :
      (forall (d92 : (Trace TyVar)) (S41 : Ty) (G : Ctx) (G0 : Ctx) ,
        ((subst_TyVar_Ctx d92 S41 (appendCtx G G0)) =
        (appendCtx (subst_TyVar_Ctx d92 S41 G) (subst_TyVar_Ctx (weakenTrace d92 (domainCtx G)) S41 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
  End SubstEnvAppendEnv.
  Section Lookups.
    Inductive lookup_evar : Ctx -> (Index TmVar) -> Ty -> Prop :=
      | lookup_evar_here {G : Ctx}
          {T58 : Ty} :
          (wfTy (domainCtx G) T58) -> (lookup_evar (evar G T58) I0 T58)
      | lookup_evar_there_evar
          {G : Ctx} {x71 : (Index TmVar)}
          {T58 : Ty} {T59 : Ty} :
          (lookup_evar G x71 T58) -> (lookup_evar (evar G T59) (IS x71) T58)
      | lookup_evar_there_etvar
          {G : Ctx} {x71 : (Index TmVar)}
          {T58 : Ty} :
          (lookup_evar G x71 T58) -> (lookup_evar (etvar G) x71 (shift_TyVar_Ty C0 T58)).
    Inductive lookup_etvar : Ctx -> (Index TyVar) -> Prop :=
      | lookup_etvar_here {G : Ctx}
          : (lookup_etvar (etvar G) I0)
      | lookup_etvar_there_evar
          {G : Ctx} {X43 : (Index TyVar)}
          {T58 : Ty} :
          (lookup_etvar G X43) -> (lookup_etvar (evar G T58) X43)
      | lookup_etvar_there_etvar
          {G : Ctx} {X43 : (Index TyVar)}
          :
          (lookup_etvar G X43) -> (lookup_etvar (etvar G) (IS X43)).
    Lemma lookup_evar_inversion_here  :
      (forall (G : Ctx) (S41 : Ty) (S42 : Ty) ,
        (lookup_evar (evar G S41) I0 S42) -> (S41 =
        S42)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_evar_functional  :
      (forall {G : Ctx} {x71 : (Index TmVar)} ,
        (forall {S41 : Ty} ,
          (lookup_evar G x71 S41) -> (forall {S42 : Ty} ,
            (lookup_evar G x71 S42) -> (S41 =
            S42)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_evar_wf  :
      (forall {G : Ctx} {x71 : (Index TmVar)} {S41 : Ty} ,
        (lookup_evar G x71 S41) -> (wfTy (domainCtx G) S41)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma weaken_lookup_evar  :
      (forall {G : Ctx} (G0 : Ctx) {x71 : (Index TmVar)} {T58 : Ty} ,
        (lookup_evar G x71 T58) -> (lookup_evar (appendCtx G G0) (weakenIndexTmVar x71 (domainCtx G0)) (weakenTy T58 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma weaken_lookup_etvar  :
      (forall {G : Ctx} (G0 : Ctx) {X43 : (Index TyVar)} ,
        (lookup_etvar G X43) -> (lookup_etvar (appendCtx G G0) (weakenIndexTyVar X43 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma lookup_evar_wfindex  :
      (forall {G : Ctx} {x71 : (Index TmVar)} {S41 : Ty} ,
        (lookup_evar G x71 S41) -> (wfindex (domainCtx G) x71)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
    Lemma lookup_etvar_wfindex  :
      (forall {G : Ctx} {X43 : (Index TyVar)} ,
        (lookup_etvar G X43) -> (wfindex (domainCtx G) X43)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
  End Lookups.
  Inductive shift_evar : (Cutoff TmVar) -> Ctx -> Ctx -> Prop :=
    | shift_evar_here {G : Ctx}
        {T58 : Ty} :
        (shift_evar C0 G (evar G T58))
    | shiftevar_there_evar
        {c14 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_evar c14 G G0) -> (shift_evar (CS c14) (evar G T) (evar G0 T))
    | shiftevar_there_etvar
        {c14 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} :
        (shift_evar c14 G G0) -> (shift_evar c14 (etvar G) (etvar G0)).
  Lemma weaken_shift_evar  :
    (forall (G : Ctx) {c14 : (Cutoff TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_evar c14 G0 G1) -> (shift_evar (weakenCutoffTmVar c14 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 G))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Inductive shift_etvar : (Cutoff TyVar) -> Ctx -> Ctx -> Prop :=
    | shift_etvar_here {G : Ctx} :
        (shift_etvar C0 G (etvar G))
    | shiftetvar_there_evar
        {c14 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_etvar c14 G G0) -> (shift_etvar c14 (evar G T) (evar G0 (shift_TyVar_Ty c14 T)))
    | shiftetvar_there_etvar
        {c14 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} :
        (shift_etvar c14 G G0) -> (shift_etvar (CS c14) (etvar G) (etvar G0)).
  Lemma weaken_shift_etvar  :
    (forall (G : Ctx) {c14 : (Cutoff TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_etvar c14 G0 G1) -> (shift_etvar (weakenCutoffTyVar c14 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 (shift_TyVar_Ctx c14 G)))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Lemma shift_evar_shifthvl_TmVar  :
    (forall {c14 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_evar c14 G G0) -> (shifthvl_TmVar c14 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Lemma shift_etvar_shifthvl_TyVar  :
    (forall {c14 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_etvar c14 G G0) -> (shifthvl_TyVar c14 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Inductive subst_evar (G : Ctx) (T58 : Ty) (s21 : Tm) : (Trace TmVar) -> Ctx -> Ctx -> Prop :=
    | subst_evar_here :
        (subst_evar G T58 s21 X0 (evar G T58) G)
    | subst_evar_there_evar
        {d92 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_evar G T58 s21 d92 G0 G1) -> (subst_evar G T58 s21 (XS TmVar d92) (evar G0 T) (evar G1 T))
    | subst_evar_there_etvar
        {d92 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} :
        (subst_evar G T58 s21 d92 G0 G1) -> (subst_evar G T58 s21 (XS TyVar d92) (etvar G0) (etvar G1)).
  Lemma weaken_subst_evar {G : Ctx} {T58 : Ty} {s21 : Tm} :
    (forall (G0 : Ctx) {d92 : (Trace TmVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_evar G T58 s21 d92 G1 G2) -> (subst_evar G T58 s21 (weakenTrace d92 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 G0))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Inductive subst_etvar (G : Ctx) (S41 : Ty) : (Trace TyVar) -> Ctx -> Ctx -> Prop :=
    | subst_etvar_here :
        (subst_etvar G S41 X0 (etvar G) G)
    | subst_etvar_there_evar
        {d92 : (Trace TyVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_etvar G S41 d92 G0 G1) -> (subst_etvar G S41 (XS TmVar d92) (evar G0 T) (evar G1 (subst_TyVar_Ty d92 S41 T)))
    | subst_etvar_there_etvar
        {d92 : (Trace TyVar)} {G0 : Ctx}
        {G1 : Ctx} :
        (subst_etvar G S41 d92 G0 G1) -> (subst_etvar G S41 (XS TyVar d92) (etvar G0) (etvar G1)).
  Lemma weaken_subst_etvar {G : Ctx} {S41 : Ty} :
    (forall (G0 : Ctx) {d92 : (Trace TyVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_etvar G S41 d92 G1 G2) -> (subst_etvar G S41 (weakenTrace d92 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 (subst_TyVar_Ctx d92 S41 G0)))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Lemma subst_evar_substhvl_TmVar {G : Ctx} {s21 : Tm} {T58 : Ty} :
    (forall {d92 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_evar G T58 s21 d92 G0 G1) -> (substhvl_TmVar (domainCtx G) d92 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
  Lemma subst_etvar_substhvl_TyVar {G : Ctx} {S41 : Ty} :
    (forall {d92 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_etvar G S41 d92 G0 G1) -> (substhvl_TyVar (domainCtx G) d92 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
End ContextStuff.
 Hint Rewrite domainCtx_shift_TyVar_Ctx : interaction.
 Hint Rewrite domainCtx_shift_TyVar_Ctx : env_domain_shift.
 Hint Rewrite domainCtx_subst_TyVar_Ctx : interaction.
 Hint Rewrite domainCtx_subst_TyVar_Ctx : env_domain_subst.
 Hint Rewrite shift_TyVar_Ctx_appendCtx : interaction.
 Hint Rewrite shift_TyVar_Ctx_appendCtx : env_shift_append.
 Hint Rewrite subst_TyVar_Ctx_appendCtx : interaction.
 Hint Rewrite subst_TyVar_Ctx_appendCtx : env_shift_append.
 Hint Constructors lookup_evar lookup_etvar : infra.
 Hint Constructors lookup_evar lookup_etvar : lookup.
 Hint Constructors lookup_evar lookup_etvar : shift.
 Hint Constructors lookup_evar lookup_etvar : subst.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : infra.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : lookup.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : shift.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : weaken.
Lemma weaken_lookup_evar_here  :
  (forall {G : Ctx} (G0 : Ctx) {T58 : Ty} (wf : (wfTy (domainCtx G) T58)) ,
    (lookup_evar (appendCtx (evar G T58) G0) (weakenIndexTmVar I0 (domainCtx G0)) (weakenTy T58 (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
Lemma weaken_lookup_etvar_here  :
  (forall {G : Ctx} (G0 : Ctx) ,
    (lookup_etvar (appendCtx (etvar G) G0) (weakenIndexTyVar I0 (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
 Hint Constructors wfDecls wfTm wfTy : infra.
 Hint Constructors wfDecls wfTm wfTy : wf.
 Hint Extern 10 ((wfDecls _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTm _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTy _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 2 ((wfTy _ _)) => match goal with
  | H : (wfTy _ (TVar _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TArr _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TAll _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfDecls _ _)) => match goal with
  | H : (wfDecls _ (DNil)) |- _ => inversion H; subst; clear H
  | H : (wfDecls _ (DCons _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfTm _ _)) => match goal with
  | H : (wfTm _ (Var _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Abs _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (App _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (TAbs _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (TApp _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Let _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Resolve lookup_evar_wf : infra.
 Hint Resolve lookup_evar_wf : wf.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : infra.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : lookup.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : wf.
 Hint Constructors shift_evar shift_etvar : infra.
 Hint Constructors shift_evar shift_etvar : shift.
 Hint Constructors shift_evar shift_etvar : subst.
 Hint Resolve weaken_shift_evar weaken_shift_etvar : infra.
 Hint Resolve weaken_shift_evar weaken_shift_etvar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : infra.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : shift_wf.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : wf.
 Hint Constructors subst_evar subst_etvar : infra.
 Hint Constructors subst_evar subst_etvar : subst.
 Hint Resolve weaken_subst_evar weaken_subst_etvar : infra.
 Hint Resolve weaken_subst_evar weaken_subst_etvar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : infra.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : subst_wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : substenv_substhvl.
Lemma shift_evar_lookup_evar  :
  (forall {c14 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c14 G G0)) {x71 : (Index TmVar)} {T58 : Ty} ,
    (lookup_evar G x71 T58) -> (lookup_evar G0 (shift_TmVar_Index c14 x71) T58)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_evar_lookup_etvar  :
  (forall {c14 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c14 G G0)) {X43 : (Index TyVar)} ,
    (lookup_etvar G X43) -> (lookup_etvar G0 X43)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_evar  :
  (forall {c14 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c14 G G0)) {x71 : (Index TmVar)} {T58 : Ty} ,
    (lookup_evar G x71 T58) -> (lookup_evar G0 x71 (shift_TyVar_Ty c14 T58))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_etvar  :
  (forall {c14 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c14 G G0)) {X43 : (Index TyVar)} ,
    (lookup_etvar G X43) -> (lookup_etvar G0 (shift_TyVar_Index c14 X43))).
Proof.
  needleGenericInsertLookup .
Qed.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : infra.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : lookup.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : shift.
Lemma subst_evar_lookup_etvar {G : Ctx} {T59 : Ty} {s21 : Tm} :
  (forall {d92 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_evar G T59 s21 d92 G0 G1)) {X43 : (Index TyVar)} ,
    (lookup_etvar G0 X43) -> (lookup_etvar G1 X43)).
Proof.
  needleGenericSubstEnvLookup .
Qed.
Lemma subst_etvar_lookup_evar {G : Ctx} {S41 : Ty} (wf : (wfTy (domainCtx G) S41)) :
  (forall {d92 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_etvar G S41 d92 G0 G1)) {x71 : (Index TmVar)} {T59 : Ty} ,
    (lookup_evar G0 x71 T59) -> (lookup_evar G1 x71 (subst_TyVar_Ty d92 S41 T59))).
Proof.
  needleGenericSubstEnvLookup .
Qed.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : infra.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : lookup.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : subst.
Fixpoint size_Ty (S0 : Ty) {struct S0} : nat :=
  match S0 with
    | (TVar X) => 1
    | (TArr T1 T2) => (plus 1 (plus (size_Ty T1) (size_Ty T2)))
    | (TAll T) => (plus 1 (size_Ty T))
  end.
Fixpoint size_Decls (d : Decls) {struct d} : nat :=
  match d with
    | (DNil) => 1
    | (DCons t d0) => (plus 1 (plus (size_Tm t) (size_Decls d0)))
  end
with size_Tm (s : Tm) {struct s} : nat :=
  match s with
    | (Var x) => 1
    | (Abs T t) => (plus 1 (plus (size_Ty T) (size_Tm t)))
    | (App t1 t2) => (plus 1 (plus (size_Tm t1) (size_Tm t2)))
    | (TAbs t0) => (plus 1 (size_Tm t0))
    | (TApp t3 T0) => (plus 1 (plus (size_Tm t3) (size_Ty T0)))
    | (Let d t4) => (plus 1 (plus (size_Decls d) (size_Tm t4)))
  end.
Lemma shift_TyVar__size_Ty  :
  (forall (S40 : Ty) (c11 : (Cutoff TyVar)) ,
    ((size_Ty (shift_TyVar_Ty c11 S40)) =
    (size_Ty S40))).
Proof.
  apply_mutual_ind ind_Ty.
  - intros X38 c11.
  reflexivity.
  - intros T51 IHT51 T52 IHT52.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT51 c12)).
  + exact ((IHT52 c12)).
  - intros T53 IHT53.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT53 (CS c12))).
Qed.

Lemma shift_TmVar__size_Decls_Tm  :
  (forall (d88 : Decls) (c12 : (Cutoff TmVar)) ,
    ((size_Decls (shift_TmVar_Decls c12 d88)) =
    (size_Decls d88))) /\
  (forall (s21 : Tm) (c12 : (Cutoff TmVar)) ,
    ((size_Tm (shift_TmVar_Tm c12 s21)) =
    (size_Tm s21))).
Proof.
  apply_mutual_ind ind_Decls_Tm.
  - intros c12; simpl; apply ((f_equal S)); reflexivity.
  - intros t118 IHt118 d88 IHd88.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt118 c12)).
  + exact ((IHd88 (CS c12))).
  - intros x66 c12.
  reflexivity.
  - intros T54 t119 IHt119.
  intros c13; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + reflexivity.
  + exact ((IHt119 (CS c13))).
  - intros t120 IHt120 t121 IHt121.
  intros c13; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt120 c13)).
  + exact ((IHt121 c13)).
  - intros t122 IHt122.
  intros c13; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt122 c13)).
  - intros t123 IHt123 T55.
  intros c13; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt123 c13)).
  + reflexivity.
  - intros d89 IHd89 t124 IHt124.
  intros c13; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHd89 c13)).
  + exact ((IHt124 (weakenCutoffTmVar c13 (bind d89)))).
Qed.

Lemma shift_TmVar__size_Decls  :
  (forall (d90 : Decls) (c13 : (Cutoff TmVar)) ,
    ((size_Decls (shift_TmVar_Decls c13 d90)) =
    (size_Decls d90))).
Proof.
  pose proof (shift_TmVar__size_Decls_Tm).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TmVar__size_Tm  :
  (forall (s21 : Tm) (c13 : (Cutoff TmVar)) ,
    ((size_Tm (shift_TmVar_Tm c13 s21)) =
    (size_Tm s21))).
Proof.
  pose proof (shift_TmVar__size_Decls_Tm).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TyVar__size_Decls_Tm  :
  (forall (d90 : Decls) (c13 : (Cutoff TyVar)) ,
    ((size_Decls (shift_TyVar_Decls c13 d90)) =
    (size_Decls d90))) /\
  (forall (s21 : Tm) (c13 : (Cutoff TyVar)) ,
    ((size_Tm (shift_TyVar_Tm c13 s21)) =
    (size_Tm s21))).
Proof.
  apply_mutual_ind ind_Decls_Tm.
  - intros c13; simpl; apply ((f_equal S)); reflexivity.
  - intros t125 IHt125 d90 IHd90.
  intros c13; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt125 c13)).
  + exact ((IHd90 c13)).
  - intros X41 c13.
  reflexivity.
  - intros T56 t126 IHt126.
  pose proof ((shift_TyVar__size_Ty T56)) as IHT56.
  intros c14; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT56 c14)).
  + exact ((IHt126 c14)).
  - intros t127 IHt127 t128 IHt128.
  intros c14; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt127 c14)).
  + exact ((IHt128 c14)).
  - intros t129 IHt129.
  intros c14; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt129 (CS c14))).
  - intros t130 IHt130 T57.
  pose proof ((shift_TyVar__size_Ty T57)) as IHT57.
  intros c14; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt130 c14)).
  + exact ((IHT57 c14)).
  - intros d91 IHd91 t131 IHt131.
  intros c14; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHd91 c14)).
  + exact ((IHt131 (weakenCutoffTyVar c14 (bind d91)))).
Qed.

Lemma shift_TyVar__size_Decls  :
  (forall (d92 : Decls) (c14 : (Cutoff TyVar)) ,
    ((size_Decls (shift_TyVar_Decls c14 d92)) =
    (size_Decls d92))).
Proof.
  pose proof (shift_TyVar__size_Decls_Tm).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TyVar__size_Tm  :
  (forall (s21 : Tm) (c14 : (Cutoff TyVar)) ,
    ((size_Tm (shift_TyVar_Tm c14 s21)) =
    (size_Tm s21))).
Proof.
  pose proof (shift_TyVar__size_Decls_Tm).
  destruct_conjs .
  easy .
Qed.
 Hint Rewrite shift_TmVar__size_Decls shift_TyVar__size_Decls shift_TmVar__size_Tm shift_TyVar__size_Tm shift_TyVar__size_Ty : interaction.
 Hint Rewrite shift_TmVar__size_Decls shift_TyVar__size_Decls shift_TmVar__size_Tm shift_TyVar__size_Tm shift_TyVar__size_Ty : shift_size.
Lemma weaken_size_Decls  :
  (forall (k61 : Hvl) (d92 : Decls) ,
    ((size_Decls (weakenDecls d92 k61)) =
    (size_Decls d92))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Tm  :
  (forall (k61 : Hvl) (s21 : Tm) ,
    ((size_Tm (weakenTm s21 k61)) =
    (size_Tm s21))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Ty  :
  (forall (k61 : Hvl) (S40 : Ty) ,
    ((size_Ty (weakenTy S40 k61)) =
    (size_Ty S40))).
Proof.
  needleGenericWeakenSize .
Qed.
 Hint Rewrite weaken_size_Decls weaken_size_Tm weaken_size_Ty : interaction.
 Hint Rewrite weaken_size_Decls weaken_size_Tm weaken_size_Ty : weaken_size.