Require Import Coq.Program.Equality.
Require Import Coq.Program.Tactics.
Require Export Needle.
Section Namespace.
  Inductive Namespace : Type :=
    | TmVar 
    | TyVar .
  Lemma eq_namespace_dec (a : Namespace) (b : Namespace) :
    (sumbool (a =
    b) (not (a =
    b))).
  Proof.
    decide equality .
  Defined.
End Namespace.

Section HVarlist.
  Inductive Hvl : Type :=
    | H0 
    | HS (a : Namespace) (k : Hvl).
  
  Fixpoint appendHvl (k : Hvl) (k0 : Hvl) {struct k0} : Hvl :=
    match k0 with
      | (H0) => k
      | (HS a k0) => (HS a (appendHvl k k0))
    end.
  
  Lemma appendHvl_assoc  :
    (forall (k : Hvl) (k0 : Hvl) (k1 : Hvl) ,
      ((appendHvl (appendHvl k k0) k1) =
      (appendHvl k (appendHvl k0 k1)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End HVarlist.

Section Index.
  Inductive Index (a : Namespace) : Type :=
    | I0 
    | IS : (Index a) -> (Index a).
  
  Global Arguments I0 [a] .
  Global Arguments IS [a] _ .
  
  Lemma eq_index_dec {a : Namespace} (i : (Index a)) (j : (Index a)) :
    (sumbool (i =
    j) (not (i =
    j))).
  Proof.
    decide equality .
  Qed.
  
  Fixpoint weakenIndexTmVar (x : (Index TmVar)) (k : Hvl) {struct k} : (Index TmVar) :=
    match k with
      | (H0) => x
      | (HS a k) => match a with
        | (TmVar) => (IS (weakenIndexTmVar x k))
        | _ => (weakenIndexTmVar x k)
      end
    end.
  Fixpoint weakenIndexTyVar (X : (Index TyVar)) (k : Hvl) {struct k} : (Index TyVar) :=
    match k with
      | (H0) => X
      | (HS a k) => match a with
        | (TyVar) => (IS (weakenIndexTyVar X k))
        | _ => (weakenIndexTyVar X k)
      end
    end.
  Lemma weakenIndexTmVar_append  :
    (forall (x : (Index TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTmVar (weakenIndexTmVar x k) k0) =
      (weakenIndexTmVar x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenIndexTyVar_append  :
    (forall (X : (Index TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTyVar (weakenIndexTyVar X k) k0) =
      (weakenIndexTyVar X (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End Index.

Section Terms.
  Inductive Tm : Set :=
    | Var (x : (Index TmVar))
    | Abs (T : Ty) (t : Tm)
    | App (t1 : Tm) (t2 : Tm)
  with Ty : Set :=
    | TVar (X : (Index TyVar))
    | TPi (T1 : Ty) (T2 : Ty)
    | TApp (T : Ty) (t : Tm).
  Scheme ind_Tm := Induction for Tm Sort Prop
  with ind_Ty := Induction for Ty Sort Prop.
  Combined Scheme ind_Tm_Ty from ind_Tm, ind_Ty.
  
  Inductive Kind : Set :=
    | Star 
    | KPi (T : Ty) (K : Kind).
  Scheme ind_Kind := Induction for Kind Sort Prop.
End Terms.

Section Functions.
  
End Functions.

Section Shift.
  Inductive Cutoff (a : Namespace) : Type :=
    | C0 
    | CS : (Cutoff a) -> (Cutoff a).
  
  Global Arguments C0 {a} .
  Global Arguments CS {a} _ .
  Fixpoint weakenCutoffTmVar (c : (Cutoff TmVar)) (k : Hvl) {struct k} : (Cutoff TmVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TmVar) => (CS (weakenCutoffTmVar c k))
        | _ => (weakenCutoffTmVar c k)
      end
    end.
  Fixpoint weakenCutoffTyVar (c : (Cutoff TyVar)) (k : Hvl) {struct k} : (Cutoff TyVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TyVar) => (CS (weakenCutoffTyVar c k))
        | _ => (weakenCutoffTyVar c k)
      end
    end.
  
  Lemma weakenCutoffTmVar_append  :
    (forall (c : (Cutoff TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTmVar (weakenCutoffTmVar c k) k0) =
      (weakenCutoffTmVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenCutoffTyVar_append  :
    (forall (c : (Cutoff TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTyVar (weakenCutoffTyVar c k) k0) =
      (weakenCutoffTyVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint shift_TmVar_Index (c : (Cutoff TmVar)) (x : (Index TmVar)) {struct c} : (Index TmVar) :=
    match c with
      | (C0) => (IS x)
      | (CS c) => match x with
        | (I0) => I0
        | (IS x) => (IS (shift_TmVar_Index c x))
      end
    end.
  Fixpoint shift_TyVar_Index (c : (Cutoff TyVar)) (X : (Index TyVar)) {struct c} : (Index TyVar) :=
    match c with
      | (C0) => (IS X)
      | (CS c) => match X with
        | (I0) => I0
        | (IS X) => (IS (shift_TyVar_Index c X))
      end
    end.
  Fixpoint shift_TmVar_Tm (c : (Cutoff TmVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var (shift_TmVar_Index c x))
      | (Abs T t) => (Abs (shift_TmVar_Ty (weakenCutoffTmVar c H0) T) (shift_TmVar_Tm (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (shift_TmVar_Tm (weakenCutoffTmVar c H0) t1) (shift_TmVar_Tm (weakenCutoffTmVar c H0) t2))
    end
  with shift_TmVar_Ty (c : (Cutoff TmVar)) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (TVar X) => (TVar X)
      | (TPi T1 T2) => (TPi (shift_TmVar_Ty (weakenCutoffTmVar c H0) T1) (shift_TmVar_Ty (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) T2))
      | (TApp T t) => (TApp (shift_TmVar_Ty (weakenCutoffTmVar c H0) T) (shift_TmVar_Tm (weakenCutoffTmVar c H0) t))
    end.
  Fixpoint shift_TyVar_Tm (c : (Cutoff TyVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (Abs T t) => (Abs (shift_TyVar_Ty (weakenCutoffTyVar c H0) T) (shift_TyVar_Tm (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (shift_TyVar_Tm (weakenCutoffTyVar c H0) t1) (shift_TyVar_Tm (weakenCutoffTyVar c H0) t2))
    end
  with shift_TyVar_Ty (c : (Cutoff TyVar)) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (TVar X) => (TVar (shift_TyVar_Index c X))
      | (TPi T1 T2) => (TPi (shift_TyVar_Ty (weakenCutoffTyVar c H0) T1) (shift_TyVar_Ty (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) T2))
      | (TApp T t) => (TApp (shift_TyVar_Ty (weakenCutoffTyVar c H0) T) (shift_TyVar_Tm (weakenCutoffTyVar c H0) t))
    end.
  Fixpoint shift_TmVar_Kind (c : (Cutoff TmVar)) (K : Kind) {struct K} : Kind :=
    match K with
      | (Star) => (Star)
      | (KPi T K0) => (KPi (shift_TmVar_Ty (weakenCutoffTmVar c H0) T) (shift_TmVar_Kind (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) K0))
    end.
  Fixpoint shift_TyVar_Kind (c : (Cutoff TyVar)) (K : Kind) {struct K} : Kind :=
    match K with
      | (Star) => (Star)
      | (KPi T K0) => (KPi (shift_TyVar_Ty (weakenCutoffTyVar c H0) T) (shift_TyVar_Kind (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) K0))
    end.
End Shift.

Section Weaken.
  Fixpoint weakenTm (s : Tm) (k : Hvl) {struct k} : Tm :=
    match k with
      | (H0) => s
      | (HS TmVar k) => (shift_TmVar_Tm C0 (weakenTm s k))
      | (HS TyVar k) => (shift_TyVar_Tm C0 (weakenTm s k))
    end.
  Fixpoint weakenTy (S0 : Ty) (k : Hvl) {struct k} : Ty :=
    match k with
      | (H0) => S0
      | (HS TmVar k) => (shift_TmVar_Ty C0 (weakenTy S0 k))
      | (HS TyVar k) => (shift_TyVar_Ty C0 (weakenTy S0 k))
    end.
  Fixpoint weakenKind (K : Kind) (k : Hvl) {struct k} : Kind :=
    match k with
      | (H0) => K
      | (HS TmVar k) => (shift_TmVar_Kind C0 (weakenKind K k))
      | (HS TyVar k) => (shift_TyVar_Kind C0 (weakenKind K k))
    end.
End Weaken.

Section Subst.
  Inductive Trace (a : Namespace) : Type :=
    | X0 
    | XS (b : Namespace)
        (T : (Trace a)).
  
  Global Arguments X0 [a] .
  Global Arguments XS [a] b _ .
  Fixpoint weakenTrace {a : Namespace} (x : (Trace a)) (k : Hvl) {struct k} : (Trace a) :=
    match k with
      | (H0) => x
      | (HS b k) => (XS b (weakenTrace x k))
    end.
  Lemma weakenTrace_append (a : Namespace) :
    (forall (x : (Trace a)) (k : Hvl) (k0 : Hvl) ,
      ((weakenTrace (weakenTrace x k) k0) =
      (weakenTrace x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint subst_TmVar_Index (d : (Trace TmVar)) (s : Tm) (x : (Index TmVar)) {struct d} : Tm :=
    match d with
      | (X0) => match x with
        | (I0) => s
        | (IS x) => (Var x)
      end
      | (XS TmVar d) => match x with
        | (I0) => (Var I0)
        | (IS x) => (shift_TmVar_Tm C0 (subst_TmVar_Index d s x))
      end
      | (XS TyVar d) => (shift_TyVar_Tm C0 (subst_TmVar_Index d s x))
    end.
  Fixpoint subst_TyVar_Index (d : (Trace TyVar)) (S0 : Ty) (X : (Index TyVar)) {struct d} : Ty :=
    match d with
      | (X0) => match X with
        | (I0) => S0
        | (IS X) => (TVar X)
      end
      | (XS TmVar d) => (shift_TmVar_Ty C0 (subst_TyVar_Index d S0 X))
      | (XS TyVar d) => match X with
        | (I0) => (TVar I0)
        | (IS X) => (shift_TyVar_Ty C0 (subst_TyVar_Index d S0 X))
      end
    end.
  Fixpoint subst_TmVar_Tm (d : (Trace TmVar)) (s : Tm) (s0 : Tm) {struct s0} : Tm :=
    match s0 with
      | (Var x) => (subst_TmVar_Index d s x)
      | (Abs T t) => (Abs (subst_TmVar_Ty d s T) (subst_TmVar_Tm (XS TmVar d) s t))
      | (App t1 t2) => (App (subst_TmVar_Tm d s t1) (subst_TmVar_Tm d s t2))
    end
  with subst_TmVar_Ty (d : (Trace TmVar)) (s : Tm) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (TVar X) => (TVar X)
      | (TPi T1 T2) => (TPi (subst_TmVar_Ty d s T1) (subst_TmVar_Ty (XS TmVar d) s T2))
      | (TApp T t) => (TApp (subst_TmVar_Ty d s T) (subst_TmVar_Tm d s t))
    end.
  Fixpoint subst_TyVar_Tm (d : (Trace TyVar)) (S0 : Ty) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (Abs T t) => (Abs (subst_TyVar_Ty d S0 T) (subst_TyVar_Tm (XS TmVar d) S0 t))
      | (App t1 t2) => (App (subst_TyVar_Tm d S0 t1) (subst_TyVar_Tm d S0 t2))
    end
  with subst_TyVar_Ty (d : (Trace TyVar)) (S0 : Ty) (S1 : Ty) {struct S1} : Ty :=
    match S1 with
      | (TVar X) => (subst_TyVar_Index d S0 X)
      | (TPi T1 T2) => (TPi (subst_TyVar_Ty d S0 T1) (subst_TyVar_Ty (XS TmVar d) S0 T2))
      | (TApp T t) => (TApp (subst_TyVar_Ty d S0 T) (subst_TyVar_Tm d S0 t))
    end.
  Fixpoint subst_TmVar_Kind (d : (Trace TmVar)) (s : Tm) (K : Kind) {struct K} : Kind :=
    match K with
      | (Star) => (Star)
      | (KPi T K0) => (KPi (subst_TmVar_Ty d s T) (subst_TmVar_Kind (XS TmVar d) s K0))
    end.
  Fixpoint subst_TyVar_Kind (d : (Trace TyVar)) (S0 : Ty) (K : Kind) {struct K} : Kind :=
    match K with
      | (Star) => (Star)
      | (KPi T K0) => (KPi (subst_TyVar_Ty d S0 T) (subst_TyVar_Kind (XS TmVar d) S0 K0))
    end.
End Subst.

Global Hint Resolve (f_equal (shift_TmVar_Kind C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TyVar_Kind C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TmVar_Tm C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TyVar_Tm C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TmVar_Ty C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TyVar_Ty C0)) : cong_shift0.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : interaction.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : weakencutoff_append.
 Hint Rewrite weakenTrace_append : interaction.
 Hint Rewrite weakenTrace_append : weakentrace_append.
Section SubstShiftReflection.
  Lemma subst_TmVar_Index0_shift_TmVar_Index0_reflection_ind (s : Tm) :
    (forall (k7 : Hvl) (x19 : (Index TmVar)) ,
      ((subst_TmVar_Index (weakenTrace X0 k7) s (shift_TmVar_Index (weakenCutoffTmVar C0 k7) x19)) =
      (Var x19))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Lemma subst_TyVar_Index0_shift_TyVar_Index0_reflection_ind (S0 : Ty) :
    (forall (k7 : Hvl) (X7 : (Index TyVar)) ,
      ((subst_TyVar_Index (weakenTrace X0 k7) S0 (shift_TyVar_Index (weakenCutoffTyVar C0 k7) X7)) =
      (TVar X7))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Definition subst_TmVar_0_shift_TmVar_0_Tm_Ty_reflection_ind := (ind_Tm_Ty (fun (s1 : Tm) =>
    (forall (k11 : Hvl) (s2 : Tm) ,
      ((subst_TmVar_Tm (weakenTrace X0 k11) s2 (shift_TmVar_Tm (weakenCutoffTmVar C0 k11) s1)) =
      s1))) (fun (S2 : Ty) =>
    (forall (k11 : Hvl) (s1 : Tm) ,
      ((subst_TmVar_Ty (weakenTrace X0 k11) s1 (shift_TmVar_Ty (weakenCutoffTmVar C0 k11) S2)) =
      S2))) (fun (x29 : (Index TmVar)) =>
    (fun (k11 : Hvl) (s1 : Tm) =>
      (subst_TmVar_Index0_shift_TmVar_Index0_reflection_ind s1 k11 x29))) (fun (T : Ty) IHT29 (t : Tm) IHt23 (k11 : Hvl) (s1 : Tm) =>
    (f_equal2 Abs (IHT29 k11 s1) (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHt23 (HS TmVar k11) s1)))) (fun (t1 : Tm) IHt23 (t2 : Tm) IHt24 (k11 : Hvl) (s1 : Tm) =>
    (f_equal2 App (IHt23 k11 s1) (IHt24 k11 s1))) (fun (X11 : (Index TyVar)) =>
    (fun (k11 : Hvl) (s1 : Tm) =>
      eq_refl)) (fun (T1 : Ty) IHT29 (T2 : Ty) IHT30 (k11 : Hvl) (s1 : Tm) =>
    (f_equal2 TPi (IHT29 k11 s1) (eq_trans (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHT30 (HS TmVar k11) s1)))) (fun (T : Ty) IHT29 (t : Tm) IHt23 (k11 : Hvl) (s1 : Tm) =>
    (f_equal2 TApp (IHT29 k11 s1) (IHt23 k11 s1)))).
  Lemma subst_TmVar_0_shift_TmVar_0_Tm_reflection_ind  :
    (forall (s1 : Tm) ,
      (forall (k11 : Hvl) (s2 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace X0 k11) s2 (shift_TmVar_Tm (weakenCutoffTmVar C0 k11) s1)) =
        s1))).
  Proof.
    pose proof (subst_TmVar_0_shift_TmVar_0_Tm_Ty_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_TmVar_0_shift_TmVar_0_Ty_reflection_ind  :
    (forall (S2 : Ty) ,
      (forall (k11 : Hvl) (s2 : Tm) ,
        ((subst_TmVar_Ty (weakenTrace X0 k11) s2 (shift_TmVar_Ty (weakenCutoffTmVar C0 k11) S2)) =
        S2))).
  Proof.
    pose proof (subst_TmVar_0_shift_TmVar_0_Tm_Ty_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Definition subst_TyVar_0_shift_TyVar_0_Tm_Ty_reflection_ind := (ind_Tm_Ty (fun (s1 : Tm) =>
    (forall (k11 : Hvl) (S2 : Ty) ,
      ((subst_TyVar_Tm (weakenTrace X0 k11) S2 (shift_TyVar_Tm (weakenCutoffTyVar C0 k11) s1)) =
      s1))) (fun (S2 : Ty) =>
    (forall (k11 : Hvl) (S3 : Ty) ,
      ((subst_TyVar_Ty (weakenTrace X0 k11) S3 (shift_TyVar_Ty (weakenCutoffTyVar C0 k11) S2)) =
      S2))) (fun (x29 : (Index TmVar)) =>
    (fun (k11 : Hvl) (S2 : Ty) =>
      eq_refl)) (fun (T : Ty) IHT29 (t : Tm) IHt23 (k11 : Hvl) (S2 : Ty) =>
    (f_equal2 Abs (IHT29 k11 S2) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHt23 (HS TmVar k11) S2)))) (fun (t1 : Tm) IHt23 (t2 : Tm) IHt24 (k11 : Hvl) (S2 : Ty) =>
    (f_equal2 App (IHt23 k11 S2) (IHt24 k11 S2))) (fun (X11 : (Index TyVar)) =>
    (fun (k11 : Hvl) (S2 : Ty) =>
      (subst_TyVar_Index0_shift_TyVar_Index0_reflection_ind S2 k11 X11))) (fun (T1 : Ty) IHT29 (T2 : Ty) IHT30 (k11 : Hvl) (S2 : Ty) =>
    (f_equal2 TPi (IHT29 k11 S2) (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHT30 (HS TmVar k11) S2)))) (fun (T : Ty) IHT29 (t : Tm) IHt23 (k11 : Hvl) (S2 : Ty) =>
    (f_equal2 TApp (IHT29 k11 S2) (IHt23 k11 S2)))).
  Lemma subst_TyVar_0_shift_TyVar_0_Tm_reflection_ind  :
    (forall (s1 : Tm) ,
      (forall (k11 : Hvl) (S2 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace X0 k11) S2 (shift_TyVar_Tm (weakenCutoffTyVar C0 k11) s1)) =
        s1))).
  Proof.
    pose proof (subst_TyVar_0_shift_TyVar_0_Tm_Ty_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Lemma subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind  :
    (forall (S2 : Ty) ,
      (forall (k11 : Hvl) (S3 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace X0 k11) S3 (shift_TyVar_Ty (weakenCutoffTyVar C0 k11) S2)) =
        S2))).
  Proof.
    pose proof (subst_TyVar_0_shift_TyVar_0_Tm_Ty_reflection_ind).
    destruct_conjs .
    easy .
  Qed.
  Definition subst_TmVar_0_shift_TmVar_0_Kind_reflection_ind := (ind_Kind (fun (K5 : Kind) =>
    (forall (k11 : Hvl) (s1 : Tm) ,
      ((subst_TmVar_Kind (weakenTrace X0 k11) s1 (shift_TmVar_Kind (weakenCutoffTmVar C0 k11) K5)) =
      K5))) (fun (k11 : Hvl) (s1 : Tm) =>
    eq_refl) (fun (T : Ty) (K : Kind) IHK5 (k11 : Hvl) (s1 : Tm) =>
    (f_equal2 KPi (let IHT29 := (subst_TmVar_0_shift_TmVar_0_Ty_reflection_ind T) in
    (IHT29 k11 s1)) (eq_trans (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHK5 (HS TmVar k11) s1))))).
  Definition subst_TyVar_0_shift_TyVar_0_Kind_reflection_ind := (ind_Kind (fun (K5 : Kind) =>
    (forall (k11 : Hvl) (S2 : Ty) ,
      ((subst_TyVar_Kind (weakenTrace X0 k11) S2 (shift_TyVar_Kind (weakenCutoffTyVar C0 k11) K5)) =
      K5))) (fun (k11 : Hvl) (S2 : Ty) =>
    eq_refl) (fun (T : Ty) (K : Kind) IHK5 (k11 : Hvl) (S2 : Ty) =>
    (f_equal2 KPi (let IHT29 := (subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind T) in
    (IHT29 k11 S2)) (eq_trans (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar X0 k11 (HS TmVar H0)) eq_refl eq_refl) (IHK5 (HS TmVar k11) S2))))).
  Definition subst_TmVar_Tm0_shift_TmVar_Tm0_reflection (s2 : Tm) : (forall (s1 : Tm) ,
    ((subst_TmVar_Tm X0 s1 (shift_TmVar_Tm C0 s2)) =
    s2)) := (subst_TmVar_0_shift_TmVar_0_Tm_reflection_ind s2 H0).
  Definition subst_TmVar_Ty0_shift_TmVar_Ty0_reflection (S2 : Ty) : (forall (s1 : Tm) ,
    ((subst_TmVar_Ty X0 s1 (shift_TmVar_Ty C0 S2)) =
    S2)) := (subst_TmVar_0_shift_TmVar_0_Ty_reflection_ind S2 H0).
  Definition subst_TyVar_Tm0_shift_TyVar_Tm0_reflection (s1 : Tm) : (forall (S2 : Ty) ,
    ((subst_TyVar_Tm X0 S2 (shift_TyVar_Tm C0 s1)) =
    s1)) := (subst_TyVar_0_shift_TyVar_0_Tm_reflection_ind s1 H0).
  Definition subst_TyVar_Ty0_shift_TyVar_Ty0_reflection (S3 : Ty) : (forall (S2 : Ty) ,
    ((subst_TyVar_Ty X0 S2 (shift_TyVar_Ty C0 S3)) =
    S3)) := (subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind S3 H0).
  Definition subst_TmVar_Kind0_shift_TmVar_Kind0_reflection (K5 : Kind) : (forall (s1 : Tm) ,
    ((subst_TmVar_Kind X0 s1 (shift_TmVar_Kind C0 K5)) =
    K5)) := (subst_TmVar_0_shift_TmVar_0_Kind_reflection_ind K5 H0).
  Definition subst_TyVar_Kind0_shift_TyVar_Kind0_reflection (K5 : Kind) : (forall (S2 : Ty) ,
    ((subst_TyVar_Kind X0 S2 (shift_TyVar_Kind C0 K5)) =
    K5)) := (subst_TyVar_0_shift_TyVar_0_Kind_reflection_ind K5 H0).
End SubstShiftReflection.
Section ShiftInteraction.
  Section ShiftIndexCommInd.
    Lemma shift_TmVar_Index_shift_TmVar_Index0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff TmVar)) (x : (Index TmVar)) ,
        ((shift_TmVar_Index (weakenCutoffTmVar (CS c) k) (shift_TmVar_Index (weakenCutoffTmVar C0 k) x)) =
        (shift_TmVar_Index (weakenCutoffTmVar C0 k) (shift_TmVar_Index (weakenCutoffTmVar c k) x)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
    Lemma shift_TyVar_Index_shift_TyVar_Index0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff TyVar)) (X : (Index TyVar)) ,
        ((shift_TyVar_Index (weakenCutoffTyVar (CS c) k) (shift_TyVar_Index (weakenCutoffTyVar C0 k) X)) =
        (shift_TyVar_Index (weakenCutoffTyVar C0 k) (shift_TyVar_Index (weakenCutoffTyVar c k) X)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
  End ShiftIndexCommInd.
  Section ShiftCommInd.
    Definition shift_TmVar__shift_TmVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s : Tm) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar (CS c7) k7) (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) s)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) (shift_TmVar_Tm (weakenCutoffTmVar c7 k7) s))))) (fun (S0 : Ty) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
        ((shift_TmVar_Ty (weakenCutoffTmVar (CS c7) k7) (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) S0)) =
        (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) (shift_TmVar_Ty (weakenCutoffTmVar c7 k7) S0))))) (fun (x19 : (Index TmVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
        (f_equal Var (shift_TmVar_Index_shift_TmVar_Index0_comm_ind k7 c7 x19)))) (fun (T : Ty) IHT19 (t : Tm) IHt15 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 Abs (IHT19 k7 c7) (IHt15 (HS TmVar k7) c7))) (fun (t1 : Tm) IHt15 (t2 : Tm) IHt16 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt15 k7 c7) (IHt16 k7 c7))) (fun (X7 : (Index TyVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T1 : Ty) IHT19 (T2 : Ty) IHT20 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 TPi (IHT19 k7 c7) (IHT20 (HS TmVar k7) c7))) (fun (T : Ty) IHT19 (t : Tm) IHt15 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 TApp (IHT19 k7 c7) (IHt15 k7 c7)))).
    Lemma shift_TmVar__shift_TmVar_0_Tm_comm_ind  :
      (forall (s : Tm) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
          ((shift_TmVar_Tm (weakenCutoffTmVar (CS c7) k7) (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) s)) =
          (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) (shift_TmVar_Tm (weakenCutoffTmVar c7 k7) s))))).
    Proof.
      pose proof (shift_TmVar__shift_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__shift_TmVar_0_Ty_comm_ind  :
      (forall (S0 : Ty) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
          ((shift_TmVar_Ty (weakenCutoffTmVar (CS c7) k7) (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) S0)) =
          (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) (shift_TmVar_Ty (weakenCutoffTmVar c7 k7) S0))))).
    Proof.
      pose proof (shift_TmVar__shift_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TmVar__shift_TyVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s : Tm) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c7 k7) (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) s)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) (shift_TmVar_Tm (weakenCutoffTmVar c7 k7) s))))) (fun (S0 : Ty) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
        ((shift_TmVar_Ty (weakenCutoffTmVar c7 k7) (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) S0)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) (shift_TmVar_Ty (weakenCutoffTmVar c7 k7) S0))))) (fun (x19 : (Index TmVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T : Ty) IHT19 (t : Tm) IHt15 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 Abs (IHT19 k7 c7) (IHt15 (HS TmVar k7) c7))) (fun (t1 : Tm) IHt15 (t2 : Tm) IHt16 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt15 k7 c7) (IHt16 k7 c7))) (fun (X7 : (Index TyVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T1 : Ty) IHT19 (T2 : Ty) IHT20 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 TPi (IHT19 k7 c7) (IHT20 (HS TmVar k7) c7))) (fun (T : Ty) IHT19 (t : Tm) IHt15 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 TApp (IHT19 k7 c7) (IHt15 k7 c7)))).
    Lemma shift_TmVar__shift_TyVar_0_Tm_comm_ind  :
      (forall (s : Tm) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
          ((shift_TmVar_Tm (weakenCutoffTmVar c7 k7) (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) s)) =
          (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) (shift_TmVar_Tm (weakenCutoffTmVar c7 k7) s))))).
    Proof.
      pose proof (shift_TmVar__shift_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__shift_TyVar_0_Ty_comm_ind  :
      (forall (S0 : Ty) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
          ((shift_TmVar_Ty (weakenCutoffTmVar c7 k7) (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) S0)) =
          (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) (shift_TmVar_Ty (weakenCutoffTmVar c7 k7) S0))))).
    Proof.
      pose proof (shift_TmVar__shift_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TyVar__shift_TmVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s : Tm) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c7 k7) (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) s)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) (shift_TyVar_Tm (weakenCutoffTyVar c7 k7) s))))) (fun (S0 : Ty) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar c7 k7) (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) S0)) =
        (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) (shift_TyVar_Ty (weakenCutoffTyVar c7 k7) S0))))) (fun (x19 : (Index TmVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) IHT19 (t : Tm) IHt15 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 Abs (IHT19 k7 c7) (IHt15 (HS TmVar k7) c7))) (fun (t1 : Tm) IHt15 (t2 : Tm) IHt16 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt15 k7 c7) (IHt16 k7 c7))) (fun (X7 : (Index TyVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T1 : Ty) IHT19 (T2 : Ty) IHT20 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 TPi (IHT19 k7 c7) (IHT20 (HS TmVar k7) c7))) (fun (T : Ty) IHT19 (t : Tm) IHt15 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 TApp (IHT19 k7 c7) (IHt15 k7 c7)))).
    Lemma shift_TyVar__shift_TmVar_0_Tm_comm_ind  :
      (forall (s : Tm) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
          ((shift_TyVar_Tm (weakenCutoffTyVar c7 k7) (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) s)) =
          (shift_TmVar_Tm (weakenCutoffTmVar C0 k7) (shift_TyVar_Tm (weakenCutoffTyVar c7 k7) s))))).
    Proof.
      pose proof (shift_TyVar__shift_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__shift_TmVar_0_Ty_comm_ind  :
      (forall (S0 : Ty) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
          ((shift_TyVar_Ty (weakenCutoffTyVar c7 k7) (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) S0)) =
          (shift_TmVar_Ty (weakenCutoffTmVar C0 k7) (shift_TyVar_Ty (weakenCutoffTyVar c7 k7) S0))))).
    Proof.
      pose proof (shift_TyVar__shift_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TyVar__shift_TyVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s : Tm) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar (CS c7) k7) (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) s)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) (shift_TyVar_Tm (weakenCutoffTyVar c7 k7) s))))) (fun (S0 : Ty) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar (CS c7) k7) (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) S0)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) (shift_TyVar_Ty (weakenCutoffTyVar c7 k7) S0))))) (fun (x19 : (Index TmVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) IHT19 (t : Tm) IHt15 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 Abs (IHT19 k7 c7) (IHt15 (HS TmVar k7) c7))) (fun (t1 : Tm) IHt15 (t2 : Tm) IHt16 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt15 k7 c7) (IHt16 k7 c7))) (fun (X7 : (Index TyVar)) =>
      (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
        (f_equal TVar (shift_TyVar_Index_shift_TyVar_Index0_comm_ind k7 c7 X7)))) (fun (T1 : Ty) IHT19 (T2 : Ty) IHT20 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 TPi (IHT19 k7 c7) (IHT20 (HS TmVar k7) c7))) (fun (T : Ty) IHT19 (t : Tm) IHt15 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 TApp (IHT19 k7 c7) (IHt15 k7 c7)))).
    Lemma shift_TyVar__shift_TyVar_0_Tm_comm_ind  :
      (forall (s : Tm) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
          ((shift_TyVar_Tm (weakenCutoffTyVar (CS c7) k7) (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) s)) =
          (shift_TyVar_Tm (weakenCutoffTyVar C0 k7) (shift_TyVar_Tm (weakenCutoffTyVar c7 k7) s))))).
    Proof.
      pose proof (shift_TyVar__shift_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__shift_TyVar_0_Ty_comm_ind  :
      (forall (S0 : Ty) ,
        (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
          ((shift_TyVar_Ty (weakenCutoffTyVar (CS c7) k7) (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) S0)) =
          (shift_TyVar_Ty (weakenCutoffTyVar C0 k7) (shift_TyVar_Ty (weakenCutoffTyVar c7 k7) S0))))).
    Proof.
      pose proof (shift_TyVar__shift_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TmVar__shift_TmVar_0_Kind_comm_ind := (ind_Kind (fun (K3 : Kind) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
        ((shift_TmVar_Kind (weakenCutoffTmVar (CS c7) k7) (shift_TmVar_Kind (weakenCutoffTmVar C0 k7) K3)) =
        (shift_TmVar_Kind (weakenCutoffTmVar C0 k7) (shift_TmVar_Kind (weakenCutoffTmVar c7 k7) K3))))) (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK3 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 KPi (let IHT19 := (shift_TmVar__shift_TmVar_0_Ty_comm_ind T) in
      (IHT19 k7 c7)) (IHK3 (HS TmVar k7) c7)))).
    Definition shift_TmVar__shift_TyVar_0_Kind_comm_ind := (ind_Kind (fun (K3 : Kind) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TmVar)) ,
        ((shift_TmVar_Kind (weakenCutoffTmVar c7 k7) (shift_TyVar_Kind (weakenCutoffTyVar C0 k7) K3)) =
        (shift_TyVar_Kind (weakenCutoffTyVar C0 k7) (shift_TmVar_Kind (weakenCutoffTmVar c7 k7) K3))))) (fun (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK3 (k7 : Hvl) (c7 : (Cutoff TmVar)) =>
      (f_equal2 KPi (let IHT19 := (shift_TmVar__shift_TyVar_0_Ty_comm_ind T) in
      (IHT19 k7 c7)) (IHK3 (HS TmVar k7) c7)))).
    Definition shift_TyVar__shift_TmVar_0_Kind_comm_ind := (ind_Kind (fun (K3 : Kind) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
        ((shift_TyVar_Kind (weakenCutoffTyVar c7 k7) (shift_TmVar_Kind (weakenCutoffTmVar C0 k7) K3)) =
        (shift_TmVar_Kind (weakenCutoffTmVar C0 k7) (shift_TyVar_Kind (weakenCutoffTyVar c7 k7) K3))))) (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK3 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 KPi (let IHT19 := (shift_TyVar__shift_TmVar_0_Ty_comm_ind T) in
      (IHT19 k7 c7)) (IHK3 (HS TmVar k7) c7)))).
    Definition shift_TyVar__shift_TyVar_0_Kind_comm_ind := (ind_Kind (fun (K3 : Kind) =>
      (forall (k7 : Hvl) (c7 : (Cutoff TyVar)) ,
        ((shift_TyVar_Kind (weakenCutoffTyVar (CS c7) k7) (shift_TyVar_Kind (weakenCutoffTyVar C0 k7) K3)) =
        (shift_TyVar_Kind (weakenCutoffTyVar C0 k7) (shift_TyVar_Kind (weakenCutoffTyVar c7 k7) K3))))) (fun (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK3 (k7 : Hvl) (c7 : (Cutoff TyVar)) =>
      (f_equal2 KPi (let IHT19 := (shift_TyVar__shift_TyVar_0_Ty_comm_ind T) in
      (IHT19 k7 c7)) (IHK3 (HS TmVar k7) c7)))).
  End ShiftCommInd.
  Section ShiftComm.
    Definition shift_TmVar__shift_TmVar_0_Tm_comm (s : Tm) : (forall (c7 : (Cutoff TmVar)) ,
      ((shift_TmVar_Tm (CS c7) (shift_TmVar_Tm C0 s)) =
      (shift_TmVar_Tm C0 (shift_TmVar_Tm c7 s)))) := (shift_TmVar__shift_TmVar_0_Tm_comm_ind s H0).
    Definition shift_TmVar__shift_TmVar_0_Ty_comm (S0 : Ty) : (forall (c7 : (Cutoff TmVar)) ,
      ((shift_TmVar_Ty (CS c7) (shift_TmVar_Ty C0 S0)) =
      (shift_TmVar_Ty C0 (shift_TmVar_Ty c7 S0)))) := (shift_TmVar__shift_TmVar_0_Ty_comm_ind S0 H0).
    Definition shift_TmVar__shift_TyVar_0_Tm_comm (s : Tm) : (forall (c7 : (Cutoff TmVar)) ,
      ((shift_TmVar_Tm c7 (shift_TyVar_Tm C0 s)) =
      (shift_TyVar_Tm C0 (shift_TmVar_Tm c7 s)))) := (shift_TmVar__shift_TyVar_0_Tm_comm_ind s H0).
    Definition shift_TmVar__shift_TyVar_0_Ty_comm (S0 : Ty) : (forall (c7 : (Cutoff TmVar)) ,
      ((shift_TmVar_Ty c7 (shift_TyVar_Ty C0 S0)) =
      (shift_TyVar_Ty C0 (shift_TmVar_Ty c7 S0)))) := (shift_TmVar__shift_TyVar_0_Ty_comm_ind S0 H0).
    Definition shift_TyVar__shift_TmVar_0_Tm_comm (s : Tm) : (forall (c7 : (Cutoff TyVar)) ,
      ((shift_TyVar_Tm c7 (shift_TmVar_Tm C0 s)) =
      (shift_TmVar_Tm C0 (shift_TyVar_Tm c7 s)))) := (shift_TyVar__shift_TmVar_0_Tm_comm_ind s H0).
    Definition shift_TyVar__shift_TmVar_0_Ty_comm (S0 : Ty) : (forall (c7 : (Cutoff TyVar)) ,
      ((shift_TyVar_Ty c7 (shift_TmVar_Ty C0 S0)) =
      (shift_TmVar_Ty C0 (shift_TyVar_Ty c7 S0)))) := (shift_TyVar__shift_TmVar_0_Ty_comm_ind S0 H0).
    Definition shift_TyVar__shift_TyVar_0_Tm_comm (s : Tm) : (forall (c7 : (Cutoff TyVar)) ,
      ((shift_TyVar_Tm (CS c7) (shift_TyVar_Tm C0 s)) =
      (shift_TyVar_Tm C0 (shift_TyVar_Tm c7 s)))) := (shift_TyVar__shift_TyVar_0_Tm_comm_ind s H0).
    Definition shift_TyVar__shift_TyVar_0_Ty_comm (S0 : Ty) : (forall (c7 : (Cutoff TyVar)) ,
      ((shift_TyVar_Ty (CS c7) (shift_TyVar_Ty C0 S0)) =
      (shift_TyVar_Ty C0 (shift_TyVar_Ty c7 S0)))) := (shift_TyVar__shift_TyVar_0_Ty_comm_ind S0 H0).
    Definition shift_TmVar__shift_TmVar_0_Kind_comm (K3 : Kind) : (forall (c7 : (Cutoff TmVar)) ,
      ((shift_TmVar_Kind (CS c7) (shift_TmVar_Kind C0 K3)) =
      (shift_TmVar_Kind C0 (shift_TmVar_Kind c7 K3)))) := (shift_TmVar__shift_TmVar_0_Kind_comm_ind K3 H0).
    Definition shift_TmVar__shift_TyVar_0_Kind_comm (K3 : Kind) : (forall (c7 : (Cutoff TmVar)) ,
      ((shift_TmVar_Kind c7 (shift_TyVar_Kind C0 K3)) =
      (shift_TyVar_Kind C0 (shift_TmVar_Kind c7 K3)))) := (shift_TmVar__shift_TyVar_0_Kind_comm_ind K3 H0).
    Definition shift_TyVar__shift_TmVar_0_Kind_comm (K3 : Kind) : (forall (c7 : (Cutoff TyVar)) ,
      ((shift_TyVar_Kind c7 (shift_TmVar_Kind C0 K3)) =
      (shift_TmVar_Kind C0 (shift_TyVar_Kind c7 K3)))) := (shift_TyVar__shift_TmVar_0_Kind_comm_ind K3 H0).
    Definition shift_TyVar__shift_TyVar_0_Kind_comm (K3 : Kind) : (forall (c7 : (Cutoff TyVar)) ,
      ((shift_TyVar_Kind (CS c7) (shift_TyVar_Kind C0 K3)) =
      (shift_TyVar_Kind C0 (shift_TyVar_Kind c7 K3)))) := (shift_TyVar__shift_TyVar_0_Kind_comm_ind K3 H0).
  End ShiftComm.
End ShiftInteraction.
 Hint Rewrite shift_TmVar__shift_TmVar_0_Kind_comm shift_TmVar__shift_TyVar_0_Kind_comm shift_TyVar__shift_TmVar_0_Kind_comm shift_TyVar__shift_TyVar_0_Kind_comm shift_TmVar__shift_TmVar_0_Tm_comm shift_TmVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TmVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Tm_comm shift_TmVar__shift_TmVar_0_Ty_comm shift_TmVar__shift_TyVar_0_Ty_comm shift_TyVar__shift_TmVar_0_Ty_comm shift_TyVar__shift_TyVar_0_Ty_comm : interaction.
 Hint Rewrite shift_TmVar__shift_TmVar_0_Kind_comm shift_TmVar__shift_TyVar_0_Kind_comm shift_TyVar__shift_TmVar_0_Kind_comm shift_TyVar__shift_TyVar_0_Kind_comm shift_TmVar__shift_TmVar_0_Tm_comm shift_TmVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TmVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Tm_comm shift_TmVar__shift_TmVar_0_Ty_comm shift_TmVar__shift_TyVar_0_Ty_comm shift_TyVar__shift_TmVar_0_Ty_comm shift_TyVar__shift_TyVar_0_Ty_comm : shift_shift0.
Section WeakenShift.
  Lemma weakenTm_shift_TmVar_Tm  :
    (forall (k55 : Hvl) (c15 : (Cutoff TmVar)) (s25 : Tm) ,
      ((weakenTm (shift_TmVar_Tm c15 s25) k55) =
      (shift_TmVar_Tm (weakenCutoffTmVar c15 k55) (weakenTm s25 k55)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTy_shift_TmVar_Ty  :
    (forall (k55 : Hvl) (c15 : (Cutoff TmVar)) (S26 : Ty) ,
      ((weakenTy (shift_TmVar_Ty c15 S26) k55) =
      (shift_TmVar_Ty (weakenCutoffTmVar c15 k55) (weakenTy S26 k55)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_shift_TyVar_Tm  :
    (forall (k55 : Hvl) (c15 : (Cutoff TyVar)) (s25 : Tm) ,
      ((weakenTm (shift_TyVar_Tm c15 s25) k55) =
      (shift_TyVar_Tm (weakenCutoffTyVar c15 k55) (weakenTm s25 k55)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTy_shift_TyVar_Ty  :
    (forall (k55 : Hvl) (c15 : (Cutoff TyVar)) (S26 : Ty) ,
      ((weakenTy (shift_TyVar_Ty c15 S26) k55) =
      (shift_TyVar_Ty (weakenCutoffTyVar c15 k55) (weakenTy S26 k55)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenKind_shift_TmVar_Kind  :
    (forall (k55 : Hvl) (c15 : (Cutoff TmVar)) (K15 : Kind) ,
      ((weakenKind (shift_TmVar_Kind c15 K15) k55) =
      (shift_TmVar_Kind (weakenCutoffTmVar c15 k55) (weakenKind K15 k55)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenKind_shift_TyVar_Kind  :
    (forall (k55 : Hvl) (c15 : (Cutoff TyVar)) (K15 : Kind) ,
      ((weakenKind (shift_TyVar_Kind c15 K15) k55) =
      (shift_TyVar_Kind (weakenCutoffTyVar c15 k55) (weakenKind K15 k55)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
End WeakenShift.
Section ShiftSubstInteraction.
  Section ShiftSubstIndexCommInd.
    Lemma shift_TmVar_Tm_subst_TmVar_Index0_comm_ind (c7 : (Cutoff TmVar)) (s1 : Tm) :
      (forall (k11 : Hvl) (x29 : (Index TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c7 k11) (subst_TmVar_Index (weakenTrace X0 k11) s1 x29)) =
        (subst_TmVar_Index (weakenTrace X0 k11) (shift_TmVar_Tm c7 s1) (shift_TmVar_Index (weakenCutoffTmVar (CS c7) k11) x29)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TyVar_Tm_subst_TmVar_Index0_comm_ind (c7 : (Cutoff TyVar)) (s1 : Tm) :
      (forall (k11 : Hvl) (x29 : (Index TmVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c7 k11) (subst_TmVar_Index (weakenTrace X0 k11) s1 x29)) =
        (subst_TmVar_Index (weakenTrace X0 k11) (shift_TyVar_Tm c7 s1) x29))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TmVar_Ty_subst_TyVar_Index0_comm_ind (c7 : (Cutoff TmVar)) (S2 : Ty) :
      (forall (k11 : Hvl) (X11 : (Index TyVar)) ,
        ((shift_TmVar_Ty (weakenCutoffTmVar c7 k11) (subst_TyVar_Index (weakenTrace X0 k11) S2 X11)) =
        (subst_TyVar_Index (weakenTrace X0 k11) (shift_TmVar_Ty c7 S2) X11))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TyVar_Ty_subst_TyVar_Index0_comm_ind (c7 : (Cutoff TyVar)) (S2 : Ty) :
      (forall (k11 : Hvl) (X11 : (Index TyVar)) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar c7 k11) (subst_TyVar_Index (weakenTrace X0 k11) S2 X11)) =
        (subst_TyVar_Index (weakenTrace X0 k11) (shift_TyVar_Ty c7 S2) (shift_TyVar_Index (weakenCutoffTyVar (CS c7) k11) X11)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End ShiftSubstIndexCommInd.
  Section ShiftSubstCommInd.
    Definition shift_TmVar__subst_TmVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s5 : Tm) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (s6 : Tm) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c15 k19) (subst_TmVar_Tm (weakenTrace X0 k19) s6 s5)) =
        (subst_TmVar_Tm (weakenTrace X0 k19) (shift_TmVar_Tm c15 s6) (shift_TmVar_Tm (weakenCutoffTmVar (CS c15) k19) s5))))) (fun (S6 : Ty) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) ,
        ((shift_TmVar_Ty (weakenCutoffTmVar c15 k19) (subst_TmVar_Ty (weakenTrace X0 k19) s5 S6)) =
        (subst_TmVar_Ty (weakenTrace X0 k19) (shift_TmVar_Tm c15 s5) (shift_TmVar_Ty (weakenCutoffTmVar (CS c15) k19) S6))))) (fun (x49 : (Index TmVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
        (shift_TmVar_Tm_subst_TmVar_Index0_comm_ind c15 s5 k19 x49))) (fun (T : Ty) IHT49 (t : Tm) IHt39 (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
      (f_equal2 Abs (IHT49 k19 c15 s5) (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt39 (HS TmVar k19) c15 s5) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt39 (t2 : Tm) IHt40 (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
      (f_equal2 App (IHt39 k19 c15 s5) (IHt40 k19 c15 s5))) (fun (X19 : (Index TyVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
        eq_refl)) (fun (T1 : Ty) IHT49 (T2 : Ty) IHT50 (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
      (f_equal2 TPi (IHT49 k19 c15 s5) (eq_trans (f_equal2 shift_TmVar_Ty eq_refl (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT50 (HS TmVar k19) c15 s5) (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (T : Ty) IHT49 (t : Tm) IHt39 (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
      (f_equal2 TApp (IHT49 k19 c15 s5) (IHt39 k19 c15 s5)))).
    Lemma shift_TmVar__subst_TmVar_0_Tm_comm_ind  :
      (forall (s5 : Tm) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (s6 : Tm) ,
          ((shift_TmVar_Tm (weakenCutoffTmVar c15 k19) (subst_TmVar_Tm (weakenTrace X0 k19) s6 s5)) =
          (subst_TmVar_Tm (weakenTrace X0 k19) (shift_TmVar_Tm c15 s6) (shift_TmVar_Tm (weakenCutoffTmVar (CS c15) k19) s5))))).
    Proof.
      pose proof (shift_TmVar__subst_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__subst_TmVar_0_Ty_comm_ind  :
      (forall (S6 : Ty) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (s6 : Tm) ,
          ((shift_TmVar_Ty (weakenCutoffTmVar c15 k19) (subst_TmVar_Ty (weakenTrace X0 k19) s6 S6)) =
          (subst_TmVar_Ty (weakenTrace X0 k19) (shift_TmVar_Tm c15 s6) (shift_TmVar_Ty (weakenCutoffTmVar (CS c15) k19) S6))))).
    Proof.
      pose proof (shift_TmVar__subst_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TmVar__subst_TyVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s5 : Tm) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c15 k19) (subst_TyVar_Tm (weakenTrace X0 k19) S6 s5)) =
        (subst_TyVar_Tm (weakenTrace X0 k19) (shift_TmVar_Ty c15 S6) (shift_TmVar_Tm (weakenCutoffTmVar c15 k19) s5))))) (fun (S6 : Ty) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (S7 : Ty) ,
        ((shift_TmVar_Ty (weakenCutoffTmVar c15 k19) (subst_TyVar_Ty (weakenTrace X0 k19) S7 S6)) =
        (subst_TyVar_Ty (weakenTrace X0 k19) (shift_TmVar_Ty c15 S7) (shift_TmVar_Ty (weakenCutoffTmVar c15 k19) S6))))) (fun (x49 : (Index TmVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
        eq_refl)) (fun (T : Ty) IHT49 (t : Tm) IHt39 (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 Abs (IHT49 k19 c15 S6) (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt39 (HS TmVar k19) c15 S6) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt39 (t2 : Tm) IHt40 (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 App (IHt39 k19 c15 S6) (IHt40 k19 c15 S6))) (fun (X19 : (Index TyVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
        (shift_TmVar_Ty_subst_TyVar_Index0_comm_ind c15 S6 k19 X19))) (fun (T1 : Ty) IHT49 (T2 : Ty) IHT50 (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 TPi (IHT49 k19 c15 S6) (eq_trans (f_equal2 shift_TmVar_Ty eq_refl (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT50 (HS TmVar k19) c15 S6) (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (T : Ty) IHT49 (t : Tm) IHt39 (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 TApp (IHT49 k19 c15 S6) (IHt39 k19 c15 S6)))).
    Lemma shift_TmVar__subst_TyVar_0_Tm_comm_ind  :
      (forall (s5 : Tm) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) ,
          ((shift_TmVar_Tm (weakenCutoffTmVar c15 k19) (subst_TyVar_Tm (weakenTrace X0 k19) S6 s5)) =
          (subst_TyVar_Tm (weakenTrace X0 k19) (shift_TmVar_Ty c15 S6) (shift_TmVar_Tm (weakenCutoffTmVar c15 k19) s5))))).
    Proof.
      pose proof (shift_TmVar__subst_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TmVar__subst_TyVar_0_Ty_comm_ind  :
      (forall (S6 : Ty) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (S7 : Ty) ,
          ((shift_TmVar_Ty (weakenCutoffTmVar c15 k19) (subst_TyVar_Ty (weakenTrace X0 k19) S7 S6)) =
          (subst_TyVar_Ty (weakenTrace X0 k19) (shift_TmVar_Ty c15 S7) (shift_TmVar_Ty (weakenCutoffTmVar c15 k19) S6))))).
    Proof.
      pose proof (shift_TmVar__subst_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TyVar__subst_TmVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s5 : Tm) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (s6 : Tm) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c15 k19) (subst_TmVar_Tm (weakenTrace X0 k19) s6 s5)) =
        (subst_TmVar_Tm (weakenTrace X0 k19) (shift_TyVar_Tm c15 s6) (shift_TyVar_Tm (weakenCutoffTyVar c15 k19) s5))))) (fun (S6 : Ty) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar c15 k19) (subst_TmVar_Ty (weakenTrace X0 k19) s5 S6)) =
        (subst_TmVar_Ty (weakenTrace X0 k19) (shift_TyVar_Tm c15 s5) (shift_TyVar_Ty (weakenCutoffTyVar c15 k19) S6))))) (fun (x49 : (Index TmVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
        (shift_TyVar_Tm_subst_TmVar_Index0_comm_ind c15 s5 k19 x49))) (fun (T : Ty) IHT49 (t : Tm) IHt39 (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
      (f_equal2 Abs (IHT49 k19 c15 s5) (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt39 (HS TmVar k19) c15 s5) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt39 (t2 : Tm) IHt40 (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
      (f_equal2 App (IHt39 k19 c15 s5) (IHt40 k19 c15 s5))) (fun (X19 : (Index TyVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
        eq_refl)) (fun (T1 : Ty) IHT49 (T2 : Ty) IHT50 (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
      (f_equal2 TPi (IHT49 k19 c15 s5) (eq_trans (f_equal2 shift_TyVar_Ty eq_refl (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT50 (HS TmVar k19) c15 s5) (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (T : Ty) IHT49 (t : Tm) IHt39 (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
      (f_equal2 TApp (IHT49 k19 c15 s5) (IHt39 k19 c15 s5)))).
    Lemma shift_TyVar__subst_TmVar_0_Tm_comm_ind  :
      (forall (s5 : Tm) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (s6 : Tm) ,
          ((shift_TyVar_Tm (weakenCutoffTyVar c15 k19) (subst_TmVar_Tm (weakenTrace X0 k19) s6 s5)) =
          (subst_TmVar_Tm (weakenTrace X0 k19) (shift_TyVar_Tm c15 s6) (shift_TyVar_Tm (weakenCutoffTyVar c15 k19) s5))))).
    Proof.
      pose proof (shift_TyVar__subst_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__subst_TmVar_0_Ty_comm_ind  :
      (forall (S6 : Ty) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (s6 : Tm) ,
          ((shift_TyVar_Ty (weakenCutoffTyVar c15 k19) (subst_TmVar_Ty (weakenTrace X0 k19) s6 S6)) =
          (subst_TmVar_Ty (weakenTrace X0 k19) (shift_TyVar_Tm c15 s6) (shift_TyVar_Ty (weakenCutoffTyVar c15 k19) S6))))).
    Proof.
      pose proof (shift_TyVar__subst_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TyVar__subst_TyVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s5 : Tm) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c15 k19) (subst_TyVar_Tm (weakenTrace X0 k19) S6 s5)) =
        (subst_TyVar_Tm (weakenTrace X0 k19) (shift_TyVar_Ty c15 S6) (shift_TyVar_Tm (weakenCutoffTyVar (CS c15) k19) s5))))) (fun (S6 : Ty) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (S7 : Ty) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar c15 k19) (subst_TyVar_Ty (weakenTrace X0 k19) S7 S6)) =
        (subst_TyVar_Ty (weakenTrace X0 k19) (shift_TyVar_Ty c15 S7) (shift_TyVar_Ty (weakenCutoffTyVar (CS c15) k19) S6))))) (fun (x49 : (Index TmVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
        eq_refl)) (fun (T : Ty) IHT49 (t : Tm) IHt39 (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 Abs (IHT49 k19 c15 S6) (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt39 (HS TmVar k19) c15 S6) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt39 (t2 : Tm) IHt40 (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 App (IHt39 k19 c15 S6) (IHt40 k19 c15 S6))) (fun (X19 : (Index TyVar)) =>
      (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
        (shift_TyVar_Ty_subst_TyVar_Index0_comm_ind c15 S6 k19 X19))) (fun (T1 : Ty) IHT49 (T2 : Ty) IHT50 (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 TPi (IHT49 k19 c15 S6) (eq_trans (f_equal2 shift_TyVar_Ty eq_refl (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT50 (HS TmVar k19) c15 S6) (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl))))) (fun (T : Ty) IHT49 (t : Tm) IHt39 (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 TApp (IHT49 k19 c15 S6) (IHt39 k19 c15 S6)))).
    Lemma shift_TyVar__subst_TyVar_0_Tm_comm_ind  :
      (forall (s5 : Tm) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) ,
          ((shift_TyVar_Tm (weakenCutoffTyVar c15 k19) (subst_TyVar_Tm (weakenTrace X0 k19) S6 s5)) =
          (subst_TyVar_Tm (weakenTrace X0 k19) (shift_TyVar_Ty c15 S6) (shift_TyVar_Tm (weakenCutoffTyVar (CS c15) k19) s5))))).
    Proof.
      pose proof (shift_TyVar__subst_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma shift_TyVar__subst_TyVar_0_Ty_comm_ind  :
      (forall (S6 : Ty) ,
        (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (S7 : Ty) ,
          ((shift_TyVar_Ty (weakenCutoffTyVar c15 k19) (subst_TyVar_Ty (weakenTrace X0 k19) S7 S6)) =
          (subst_TyVar_Ty (weakenTrace X0 k19) (shift_TyVar_Ty c15 S7) (shift_TyVar_Ty (weakenCutoffTyVar (CS c15) k19) S6))))).
    Proof.
      pose proof (shift_TyVar__subst_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition shift_TmVar__subst_TmVar_0_Kind_comm_ind := (ind_Kind (fun (K9 : Kind) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) ,
        ((shift_TmVar_Kind (weakenCutoffTmVar c15 k19) (subst_TmVar_Kind (weakenTrace X0 k19) s5 K9)) =
        (subst_TmVar_Kind (weakenTrace X0 k19) (shift_TmVar_Tm c15 s5) (shift_TmVar_Kind (weakenCutoffTmVar (CS c15) k19) K9))))) (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK9 (k19 : Hvl) (c15 : (Cutoff TmVar)) (s5 : Tm) =>
      (f_equal2 KPi (let IHT49 := (shift_TmVar__subst_TmVar_0_Ty_comm_ind T) in
      (IHT49 k19 c15 s5)) (eq_trans (f_equal2 shift_TmVar_Kind eq_refl (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK9 (HS TmVar k19) c15 s5) (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl)))))).
    Definition shift_TmVar__subst_TyVar_0_Kind_comm_ind := (ind_Kind (fun (K9 : Kind) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) ,
        ((shift_TmVar_Kind (weakenCutoffTmVar c15 k19) (subst_TyVar_Kind (weakenTrace X0 k19) S6 K9)) =
        (subst_TyVar_Kind (weakenTrace X0 k19) (shift_TmVar_Ty c15 S6) (shift_TmVar_Kind (weakenCutoffTmVar c15 k19) K9))))) (fun (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK9 (k19 : Hvl) (c15 : (Cutoff TmVar)) (S6 : Ty) =>
      (f_equal2 KPi (let IHT49 := (shift_TmVar__subst_TyVar_0_Ty_comm_ind T) in
      (IHT49 k19 c15 S6)) (eq_trans (f_equal2 shift_TmVar_Kind eq_refl (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK9 (HS TmVar k19) c15 S6) (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl)))))).
    Definition shift_TyVar__subst_TmVar_0_Kind_comm_ind := (ind_Kind (fun (K9 : Kind) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) ,
        ((shift_TyVar_Kind (weakenCutoffTyVar c15 k19) (subst_TmVar_Kind (weakenTrace X0 k19) s5 K9)) =
        (subst_TmVar_Kind (weakenTrace X0 k19) (shift_TyVar_Tm c15 s5) (shift_TyVar_Kind (weakenCutoffTyVar c15 k19) K9))))) (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK9 (k19 : Hvl) (c15 : (Cutoff TyVar)) (s5 : Tm) =>
      (f_equal2 KPi (let IHT49 := (shift_TyVar__subst_TmVar_0_Ty_comm_ind T) in
      (IHT49 k19 c15 s5)) (eq_trans (f_equal2 shift_TyVar_Kind eq_refl (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK9 (HS TmVar k19) c15 s5) (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar X0 k19 (HS TmVar H0))) eq_refl eq_refl)))))).
    Definition shift_TyVar__subst_TyVar_0_Kind_comm_ind := (ind_Kind (fun (K9 : Kind) =>
      (forall (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) ,
        ((shift_TyVar_Kind (weakenCutoffTyVar c15 k19) (subst_TyVar_Kind (weakenTrace X0 k19) S6 K9)) =
        (subst_TyVar_Kind (weakenTrace X0 k19) (shift_TyVar_Ty c15 S6) (shift_TyVar_Kind (weakenCutoffTyVar (CS c15) k19) K9))))) (fun (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK9 (k19 : Hvl) (c15 : (Cutoff TyVar)) (S6 : Ty) =>
      (f_equal2 KPi (let IHT49 := (shift_TyVar__subst_TyVar_0_Ty_comm_ind T) in
      (IHT49 k19 c15 S6)) (eq_trans (f_equal2 shift_TyVar_Kind eq_refl (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar X0 k19 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK9 (HS TmVar k19) c15 S6) (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar X0 k19 (HS TmVar H0))) eq_refl eq_refl)))))).
  End ShiftSubstCommInd.
  Section ShiftSubstComm.
    Definition shift_TmVar_Tm_subst_TmVar_Tm0_comm (s6 : Tm) : (forall (c15 : (Cutoff TmVar)) (s5 : Tm) ,
      ((shift_TmVar_Tm c15 (subst_TmVar_Tm X0 s5 s6)) =
      (subst_TmVar_Tm X0 (shift_TmVar_Tm c15 s5) (shift_TmVar_Tm (CS c15) s6)))) := (shift_TmVar__subst_TmVar_0_Tm_comm_ind s6 H0).
    Definition shift_TmVar_Ty_subst_TmVar_Ty0_comm (S6 : Ty) : (forall (c15 : (Cutoff TmVar)) (s5 : Tm) ,
      ((shift_TmVar_Ty c15 (subst_TmVar_Ty X0 s5 S6)) =
      (subst_TmVar_Ty X0 (shift_TmVar_Tm c15 s5) (shift_TmVar_Ty (CS c15) S6)))) := (shift_TmVar__subst_TmVar_0_Ty_comm_ind S6 H0).
    Definition shift_TmVar_Tm_subst_TyVar_Tm0_comm (s5 : Tm) : (forall (c15 : (Cutoff TmVar)) (S6 : Ty) ,
      ((shift_TmVar_Tm c15 (subst_TyVar_Tm X0 S6 s5)) =
      (subst_TyVar_Tm X0 (shift_TmVar_Ty c15 S6) (shift_TmVar_Tm c15 s5)))) := (shift_TmVar__subst_TyVar_0_Tm_comm_ind s5 H0).
    Definition shift_TmVar_Ty_subst_TyVar_Ty0_comm (S7 : Ty) : (forall (c15 : (Cutoff TmVar)) (S6 : Ty) ,
      ((shift_TmVar_Ty c15 (subst_TyVar_Ty X0 S6 S7)) =
      (subst_TyVar_Ty X0 (shift_TmVar_Ty c15 S6) (shift_TmVar_Ty c15 S7)))) := (shift_TmVar__subst_TyVar_0_Ty_comm_ind S7 H0).
    Definition shift_TyVar_Tm_subst_TmVar_Tm0_comm (s6 : Tm) : (forall (c15 : (Cutoff TyVar)) (s5 : Tm) ,
      ((shift_TyVar_Tm c15 (subst_TmVar_Tm X0 s5 s6)) =
      (subst_TmVar_Tm X0 (shift_TyVar_Tm c15 s5) (shift_TyVar_Tm c15 s6)))) := (shift_TyVar__subst_TmVar_0_Tm_comm_ind s6 H0).
    Definition shift_TyVar_Ty_subst_TmVar_Ty0_comm (S6 : Ty) : (forall (c15 : (Cutoff TyVar)) (s5 : Tm) ,
      ((shift_TyVar_Ty c15 (subst_TmVar_Ty X0 s5 S6)) =
      (subst_TmVar_Ty X0 (shift_TyVar_Tm c15 s5) (shift_TyVar_Ty c15 S6)))) := (shift_TyVar__subst_TmVar_0_Ty_comm_ind S6 H0).
    Definition shift_TyVar_Tm_subst_TyVar_Tm0_comm (s5 : Tm) : (forall (c15 : (Cutoff TyVar)) (S6 : Ty) ,
      ((shift_TyVar_Tm c15 (subst_TyVar_Tm X0 S6 s5)) =
      (subst_TyVar_Tm X0 (shift_TyVar_Ty c15 S6) (shift_TyVar_Tm (CS c15) s5)))) := (shift_TyVar__subst_TyVar_0_Tm_comm_ind s5 H0).
    Definition shift_TyVar_Ty_subst_TyVar_Ty0_comm (S7 : Ty) : (forall (c15 : (Cutoff TyVar)) (S6 : Ty) ,
      ((shift_TyVar_Ty c15 (subst_TyVar_Ty X0 S6 S7)) =
      (subst_TyVar_Ty X0 (shift_TyVar_Ty c15 S6) (shift_TyVar_Ty (CS c15) S7)))) := (shift_TyVar__subst_TyVar_0_Ty_comm_ind S7 H0).
    Definition shift_TmVar_Kind_subst_TmVar_Kind0_comm (K9 : Kind) : (forall (c15 : (Cutoff TmVar)) (s5 : Tm) ,
      ((shift_TmVar_Kind c15 (subst_TmVar_Kind X0 s5 K9)) =
      (subst_TmVar_Kind X0 (shift_TmVar_Tm c15 s5) (shift_TmVar_Kind (CS c15) K9)))) := (shift_TmVar__subst_TmVar_0_Kind_comm_ind K9 H0).
    Definition shift_TmVar_Kind_subst_TyVar_Kind0_comm (K9 : Kind) : (forall (c15 : (Cutoff TmVar)) (S6 : Ty) ,
      ((shift_TmVar_Kind c15 (subst_TyVar_Kind X0 S6 K9)) =
      (subst_TyVar_Kind X0 (shift_TmVar_Ty c15 S6) (shift_TmVar_Kind c15 K9)))) := (shift_TmVar__subst_TyVar_0_Kind_comm_ind K9 H0).
    Definition shift_TyVar_Kind_subst_TmVar_Kind0_comm (K9 : Kind) : (forall (c15 : (Cutoff TyVar)) (s5 : Tm) ,
      ((shift_TyVar_Kind c15 (subst_TmVar_Kind X0 s5 K9)) =
      (subst_TmVar_Kind X0 (shift_TyVar_Tm c15 s5) (shift_TyVar_Kind c15 K9)))) := (shift_TyVar__subst_TmVar_0_Kind_comm_ind K9 H0).
    Definition shift_TyVar_Kind_subst_TyVar_Kind0_comm (K9 : Kind) : (forall (c15 : (Cutoff TyVar)) (S6 : Ty) ,
      ((shift_TyVar_Kind c15 (subst_TyVar_Kind X0 S6 K9)) =
      (subst_TyVar_Kind X0 (shift_TyVar_Ty c15 S6) (shift_TyVar_Kind (CS c15) K9)))) := (shift_TyVar__subst_TyVar_0_Kind_comm_ind K9 H0).
  End ShiftSubstComm.
  Section SubstShiftIndexCommInd.
    Lemma subst_TmVar_Index_shift_TmVar_Tm0_comm_ind (d : (Trace TmVar)) (s5 : Tm) :
      (forall (k19 : Hvl) (x49 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace (XS TmVar d) k19) s5 (shift_TmVar_Index (weakenCutoffTmVar C0 k19) x49)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k19) (subst_TmVar_Index (weakenTrace d k19) s5 x49)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Index_shift_TyVar_Tm0_comm_ind (d : (Trace TmVar)) (s5 : Tm) :
      (forall (k19 : Hvl) (x49 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace (XS TyVar d) k19) s5 x49) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k19) (subst_TmVar_Index (weakenTrace d k19) s5 x49)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Index_shift_TmVar_Ty0_comm_ind (d : (Trace TyVar)) (S6 : Ty) :
      (forall (k19 : Hvl) (X19 : (Index TyVar)) ,
        ((subst_TyVar_Index (weakenTrace (XS TmVar d) k19) S6 X19) =
        (shift_TmVar_Ty (weakenCutoffTmVar C0 k19) (subst_TyVar_Index (weakenTrace d k19) S6 X19)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Index_shift_TyVar_Ty0_comm_ind (d : (Trace TyVar)) (S6 : Ty) :
      (forall (k19 : Hvl) (X19 : (Index TyVar)) ,
        ((subst_TyVar_Index (weakenTrace (XS TyVar d) k19) S6 (shift_TyVar_Index (weakenCutoffTyVar C0 k19) X19)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k19) (subst_TyVar_Index (weakenTrace d k19) S6 X19)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End SubstShiftIndexCommInd.
  Section SubstShiftCommInd.
    Definition subst_TmVar__shift_TmVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s21 : Tm) =>
      (forall (k51 : Hvl) (d31 : (Trace TmVar)) (s22 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace (XS TmVar d31) k51) s22 (shift_TmVar_Tm (weakenCutoffTmVar C0 k51) s21)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k51) (subst_TmVar_Tm (weakenTrace d31 k51) s22 s21))))) (fun (S22 : Ty) =>
      (forall (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) ,
        ((subst_TmVar_Ty (weakenTrace (XS TmVar d31) k51) s21 (shift_TmVar_Ty (weakenCutoffTmVar C0 k51) S22)) =
        (shift_TmVar_Ty (weakenCutoffTmVar C0 k51) (subst_TmVar_Ty (weakenTrace d31 k51) s21 S22))))) (fun (x69 : (Index TmVar)) =>
      (fun (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
        (subst_TmVar_Index_shift_TmVar_Tm0_comm_ind d31 s21 k51 x69))) (fun (T : Ty) IHT69 (t : Tm) IHt55 (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
      (f_equal2 Abs (IHT69 k51 d31 s21) (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TmVar d31) k51 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt55 (HS TmVar k51) d31 s21) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d31 k51 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt55 (t2 : Tm) IHt56 (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
      (f_equal2 App (IHt55 k51 d31 s21) (IHt56 k51 d31 s21))) (fun (X27 : (Index TyVar)) =>
      (fun (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
        eq_refl)) (fun (T1 : Ty) IHT69 (T2 : Ty) IHT70 (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
      (f_equal2 TPi (IHT69 k51 d31 s21) (eq_trans (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar (XS TmVar d31) k51 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHT70 (HS TmVar k51) d31 s21) (f_equal2 shift_TmVar_Ty eq_refl (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar d31 k51 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT69 (t : Tm) IHt55 (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
      (f_equal2 TApp (IHT69 k51 d31 s21) (IHt55 k51 d31 s21)))).
    Lemma subst_TmVar__shift_TmVar_0_Tm_comm_ind  :
      (forall (s21 : Tm) ,
        (forall (k51 : Hvl) (d31 : (Trace TmVar)) (s22 : Tm) ,
          ((subst_TmVar_Tm (weakenTrace (XS TmVar d31) k51) s22 (shift_TmVar_Tm (weakenCutoffTmVar C0 k51) s21)) =
          (shift_TmVar_Tm (weakenCutoffTmVar C0 k51) (subst_TmVar_Tm (weakenTrace d31 k51) s22 s21))))).
    Proof.
      pose proof (subst_TmVar__shift_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__shift_TmVar_0_Ty_comm_ind  :
      (forall (S22 : Ty) ,
        (forall (k51 : Hvl) (d31 : (Trace TmVar)) (s22 : Tm) ,
          ((subst_TmVar_Ty (weakenTrace (XS TmVar d31) k51) s22 (shift_TmVar_Ty (weakenCutoffTmVar C0 k51) S22)) =
          (shift_TmVar_Ty (weakenCutoffTmVar C0 k51) (subst_TmVar_Ty (weakenTrace d31 k51) s22 S22))))).
    Proof.
      pose proof (subst_TmVar__shift_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TmVar__shift_TyVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s21 : Tm) =>
      (forall (k51 : Hvl) (d31 : (Trace TmVar)) (s22 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace (XS TyVar d31) k51) s22 (shift_TyVar_Tm (weakenCutoffTyVar C0 k51) s21)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k51) (subst_TmVar_Tm (weakenTrace d31 k51) s22 s21))))) (fun (S22 : Ty) =>
      (forall (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) ,
        ((subst_TmVar_Ty (weakenTrace (XS TyVar d31) k51) s21 (shift_TyVar_Ty (weakenCutoffTyVar C0 k51) S22)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k51) (subst_TmVar_Ty (weakenTrace d31 k51) s21 S22))))) (fun (x69 : (Index TmVar)) =>
      (fun (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
        (subst_TmVar_Index_shift_TyVar_Tm0_comm_ind d31 s21 k51 x69))) (fun (T : Ty) IHT69 (t : Tm) IHt55 (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
      (f_equal2 Abs (IHT69 k51 d31 s21) (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TyVar d31) k51 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt55 (HS TmVar k51) d31 s21) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d31 k51 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt55 (t2 : Tm) IHt56 (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
      (f_equal2 App (IHt55 k51 d31 s21) (IHt56 k51 d31 s21))) (fun (X27 : (Index TyVar)) =>
      (fun (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
        eq_refl)) (fun (T1 : Ty) IHT69 (T2 : Ty) IHT70 (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
      (f_equal2 TPi (IHT69 k51 d31 s21) (eq_trans (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar (XS TyVar d31) k51 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHT70 (HS TmVar k51) d31 s21) (f_equal2 shift_TyVar_Ty eq_refl (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar d31 k51 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT69 (t : Tm) IHt55 (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
      (f_equal2 TApp (IHT69 k51 d31 s21) (IHt55 k51 d31 s21)))).
    Lemma subst_TmVar__shift_TyVar_0_Tm_comm_ind  :
      (forall (s21 : Tm) ,
        (forall (k51 : Hvl) (d31 : (Trace TmVar)) (s22 : Tm) ,
          ((subst_TmVar_Tm (weakenTrace (XS TyVar d31) k51) s22 (shift_TyVar_Tm (weakenCutoffTyVar C0 k51) s21)) =
          (shift_TyVar_Tm (weakenCutoffTyVar C0 k51) (subst_TmVar_Tm (weakenTrace d31 k51) s22 s21))))).
    Proof.
      pose proof (subst_TmVar__shift_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__shift_TyVar_0_Ty_comm_ind  :
      (forall (S22 : Ty) ,
        (forall (k51 : Hvl) (d31 : (Trace TmVar)) (s22 : Tm) ,
          ((subst_TmVar_Ty (weakenTrace (XS TyVar d31) k51) s22 (shift_TyVar_Ty (weakenCutoffTyVar C0 k51) S22)) =
          (shift_TyVar_Ty (weakenCutoffTyVar C0 k51) (subst_TmVar_Ty (weakenTrace d31 k51) s22 S22))))).
    Proof.
      pose proof (subst_TmVar__shift_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TyVar__shift_TmVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s21 : Tm) =>
      (forall (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace (XS TmVar d31) k51) S22 (shift_TmVar_Tm (weakenCutoffTmVar C0 k51) s21)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k51) (subst_TyVar_Tm (weakenTrace d31 k51) S22 s21))))) (fun (S22 : Ty) =>
      (forall (k51 : Hvl) (d31 : (Trace TyVar)) (S23 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace (XS TmVar d31) k51) S23 (shift_TmVar_Ty (weakenCutoffTmVar C0 k51) S22)) =
        (shift_TmVar_Ty (weakenCutoffTmVar C0 k51) (subst_TyVar_Ty (weakenTrace d31 k51) S23 S22))))) (fun (x69 : (Index TmVar)) =>
      (fun (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
        eq_refl)) (fun (T : Ty) IHT69 (t : Tm) IHt55 (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
      (f_equal2 Abs (IHT69 k51 d31 S22) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TmVar d31) k51 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt55 (HS TmVar k51) d31 S22) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d31 k51 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt55 (t2 : Tm) IHt56 (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
      (f_equal2 App (IHt55 k51 d31 S22) (IHt56 k51 d31 S22))) (fun (X27 : (Index TyVar)) =>
      (fun (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
        (subst_TyVar_Index_shift_TmVar_Ty0_comm_ind d31 S22 k51 X27))) (fun (T1 : Ty) IHT69 (T2 : Ty) IHT70 (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
      (f_equal2 TPi (IHT69 k51 d31 S22) (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar (XS TmVar d31) k51 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHT70 (HS TmVar k51) d31 S22) (f_equal2 shift_TmVar_Ty eq_refl (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar d31 k51 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT69 (t : Tm) IHt55 (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
      (f_equal2 TApp (IHT69 k51 d31 S22) (IHt55 k51 d31 S22)))).
    Lemma subst_TyVar__shift_TmVar_0_Tm_comm_ind  :
      (forall (s21 : Tm) ,
        (forall (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) ,
          ((subst_TyVar_Tm (weakenTrace (XS TmVar d31) k51) S22 (shift_TmVar_Tm (weakenCutoffTmVar C0 k51) s21)) =
          (shift_TmVar_Tm (weakenCutoffTmVar C0 k51) (subst_TyVar_Tm (weakenTrace d31 k51) S22 s21))))).
    Proof.
      pose proof (subst_TyVar__shift_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__shift_TmVar_0_Ty_comm_ind  :
      (forall (S22 : Ty) ,
        (forall (k51 : Hvl) (d31 : (Trace TyVar)) (S23 : Ty) ,
          ((subst_TyVar_Ty (weakenTrace (XS TmVar d31) k51) S23 (shift_TmVar_Ty (weakenCutoffTmVar C0 k51) S22)) =
          (shift_TmVar_Ty (weakenCutoffTmVar C0 k51) (subst_TyVar_Ty (weakenTrace d31 k51) S23 S22))))).
    Proof.
      pose proof (subst_TyVar__shift_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TyVar__shift_TyVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s21 : Tm) =>
      (forall (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace (XS TyVar d31) k51) S22 (shift_TyVar_Tm (weakenCutoffTyVar C0 k51) s21)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k51) (subst_TyVar_Tm (weakenTrace d31 k51) S22 s21))))) (fun (S22 : Ty) =>
      (forall (k51 : Hvl) (d31 : (Trace TyVar)) (S23 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace (XS TyVar d31) k51) S23 (shift_TyVar_Ty (weakenCutoffTyVar C0 k51) S22)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k51) (subst_TyVar_Ty (weakenTrace d31 k51) S23 S22))))) (fun (x69 : (Index TmVar)) =>
      (fun (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
        eq_refl)) (fun (T : Ty) IHT69 (t : Tm) IHt55 (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
      (f_equal2 Abs (IHT69 k51 d31 S22) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TyVar d31) k51 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt55 (HS TmVar k51) d31 S22) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d31 k51 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt55 (t2 : Tm) IHt56 (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
      (f_equal2 App (IHt55 k51 d31 S22) (IHt56 k51 d31 S22))) (fun (X27 : (Index TyVar)) =>
      (fun (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
        (subst_TyVar_Index_shift_TyVar_Ty0_comm_ind d31 S22 k51 X27))) (fun (T1 : Ty) IHT69 (T2 : Ty) IHT70 (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
      (f_equal2 TPi (IHT69 k51 d31 S22) (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar (XS TyVar d31) k51 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHT70 (HS TmVar k51) d31 S22) (f_equal2 shift_TyVar_Ty eq_refl (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar d31 k51 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT69 (t : Tm) IHt55 (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
      (f_equal2 TApp (IHT69 k51 d31 S22) (IHt55 k51 d31 S22)))).
    Lemma subst_TyVar__shift_TyVar_0_Tm_comm_ind  :
      (forall (s21 : Tm) ,
        (forall (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) ,
          ((subst_TyVar_Tm (weakenTrace (XS TyVar d31) k51) S22 (shift_TyVar_Tm (weakenCutoffTyVar C0 k51) s21)) =
          (shift_TyVar_Tm (weakenCutoffTyVar C0 k51) (subst_TyVar_Tm (weakenTrace d31 k51) S22 s21))))).
    Proof.
      pose proof (subst_TyVar__shift_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__shift_TyVar_0_Ty_comm_ind  :
      (forall (S22 : Ty) ,
        (forall (k51 : Hvl) (d31 : (Trace TyVar)) (S23 : Ty) ,
          ((subst_TyVar_Ty (weakenTrace (XS TyVar d31) k51) S23 (shift_TyVar_Ty (weakenCutoffTyVar C0 k51) S22)) =
          (shift_TyVar_Ty (weakenCutoffTyVar C0 k51) (subst_TyVar_Ty (weakenTrace d31 k51) S23 S22))))).
    Proof.
      pose proof (subst_TyVar__shift_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TmVar__shift_TmVar_0_Kind_comm_ind := (ind_Kind (fun (K13 : Kind) =>
      (forall (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) ,
        ((subst_TmVar_Kind (weakenTrace (XS TmVar d31) k51) s21 (shift_TmVar_Kind (weakenCutoffTmVar C0 k51) K13)) =
        (shift_TmVar_Kind (weakenCutoffTmVar C0 k51) (subst_TmVar_Kind (weakenTrace d31 k51) s21 K13))))) (fun (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK13 (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
      (f_equal2 KPi (let IHT69 := (subst_TmVar__shift_TmVar_0_Ty_comm_ind T) in
      (IHT69 k51 d31 s21)) (eq_trans (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar (XS TmVar d31) k51 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHK13 (HS TmVar k51) d31 s21) (f_equal2 shift_TmVar_Kind eq_refl (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar d31 k51 (HS TmVar H0))) eq_refl eq_refl))))))).
    Definition subst_TmVar__shift_TyVar_0_Kind_comm_ind := (ind_Kind (fun (K13 : Kind) =>
      (forall (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) ,
        ((subst_TmVar_Kind (weakenTrace (XS TyVar d31) k51) s21 (shift_TyVar_Kind (weakenCutoffTyVar C0 k51) K13)) =
        (shift_TyVar_Kind (weakenCutoffTyVar C0 k51) (subst_TmVar_Kind (weakenTrace d31 k51) s21 K13))))) (fun (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK13 (k51 : Hvl) (d31 : (Trace TmVar)) (s21 : Tm) =>
      (f_equal2 KPi (let IHT69 := (subst_TmVar__shift_TyVar_0_Ty_comm_ind T) in
      (IHT69 k51 d31 s21)) (eq_trans (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar (XS TyVar d31) k51 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHK13 (HS TmVar k51) d31 s21) (f_equal2 shift_TyVar_Kind eq_refl (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar d31 k51 (HS TmVar H0))) eq_refl eq_refl))))))).
    Definition subst_TyVar__shift_TmVar_0_Kind_comm_ind := (ind_Kind (fun (K13 : Kind) =>
      (forall (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) ,
        ((subst_TyVar_Kind (weakenTrace (XS TmVar d31) k51) S22 (shift_TmVar_Kind (weakenCutoffTmVar C0 k51) K13)) =
        (shift_TmVar_Kind (weakenCutoffTmVar C0 k51) (subst_TyVar_Kind (weakenTrace d31 k51) S22 K13))))) (fun (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK13 (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
      (f_equal2 KPi (let IHT69 := (subst_TyVar__shift_TmVar_0_Ty_comm_ind T) in
      (IHT69 k51 d31 S22)) (eq_trans (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar (XS TmVar d31) k51 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHK13 (HS TmVar k51) d31 S22) (f_equal2 shift_TmVar_Kind eq_refl (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar d31 k51 (HS TmVar H0))) eq_refl eq_refl))))))).
    Definition subst_TyVar__shift_TyVar_0_Kind_comm_ind := (ind_Kind (fun (K13 : Kind) =>
      (forall (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) ,
        ((subst_TyVar_Kind (weakenTrace (XS TyVar d31) k51) S22 (shift_TyVar_Kind (weakenCutoffTyVar C0 k51) K13)) =
        (shift_TyVar_Kind (weakenCutoffTyVar C0 k51) (subst_TyVar_Kind (weakenTrace d31 k51) S22 K13))))) (fun (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK13 (k51 : Hvl) (d31 : (Trace TyVar)) (S22 : Ty) =>
      (f_equal2 KPi (let IHT69 := (subst_TyVar__shift_TyVar_0_Ty_comm_ind T) in
      (IHT69 k51 d31 S22)) (eq_trans (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar (XS TyVar d31) k51 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHK13 (HS TmVar k51) d31 S22) (f_equal2 shift_TyVar_Kind eq_refl (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar d31 k51 (HS TmVar H0))) eq_refl eq_refl))))))).
  End SubstShiftCommInd.
  Section SubstShiftComm.
    Definition subst_TmVar_Tm_shift_TmVar_Tm0_comm (s22 : Tm) : (forall (d31 : (Trace TmVar)) (s21 : Tm) ,
      ((subst_TmVar_Tm (XS TmVar d31) s21 (shift_TmVar_Tm C0 s22)) =
      (shift_TmVar_Tm C0 (subst_TmVar_Tm d31 s21 s22)))) := (subst_TmVar__shift_TmVar_0_Tm_comm_ind s22 H0).
    Definition subst_TmVar_Ty_shift_TmVar_Ty0_comm (S22 : Ty) : (forall (d31 : (Trace TmVar)) (s21 : Tm) ,
      ((subst_TmVar_Ty (XS TmVar d31) s21 (shift_TmVar_Ty C0 S22)) =
      (shift_TmVar_Ty C0 (subst_TmVar_Ty d31 s21 S22)))) := (subst_TmVar__shift_TmVar_0_Ty_comm_ind S22 H0).
    Definition subst_TmVar_Tm_shift_TyVar_Tm0_comm (s22 : Tm) : (forall (d31 : (Trace TmVar)) (s21 : Tm) ,
      ((subst_TmVar_Tm (XS TyVar d31) s21 (shift_TyVar_Tm C0 s22)) =
      (shift_TyVar_Tm C0 (subst_TmVar_Tm d31 s21 s22)))) := (subst_TmVar__shift_TyVar_0_Tm_comm_ind s22 H0).
    Definition subst_TmVar_Ty_shift_TyVar_Ty0_comm (S22 : Ty) : (forall (d31 : (Trace TmVar)) (s21 : Tm) ,
      ((subst_TmVar_Ty (XS TyVar d31) s21 (shift_TyVar_Ty C0 S22)) =
      (shift_TyVar_Ty C0 (subst_TmVar_Ty d31 s21 S22)))) := (subst_TmVar__shift_TyVar_0_Ty_comm_ind S22 H0).
    Definition subst_TyVar_Tm_shift_TmVar_Tm0_comm (s21 : Tm) : (forall (d31 : (Trace TyVar)) (S22 : Ty) ,
      ((subst_TyVar_Tm (XS TmVar d31) S22 (shift_TmVar_Tm C0 s21)) =
      (shift_TmVar_Tm C0 (subst_TyVar_Tm d31 S22 s21)))) := (subst_TyVar__shift_TmVar_0_Tm_comm_ind s21 H0).
    Definition subst_TyVar_Ty_shift_TmVar_Ty0_comm (S23 : Ty) : (forall (d31 : (Trace TyVar)) (S22 : Ty) ,
      ((subst_TyVar_Ty (XS TmVar d31) S22 (shift_TmVar_Ty C0 S23)) =
      (shift_TmVar_Ty C0 (subst_TyVar_Ty d31 S22 S23)))) := (subst_TyVar__shift_TmVar_0_Ty_comm_ind S23 H0).
    Definition subst_TyVar_Tm_shift_TyVar_Tm0_comm (s21 : Tm) : (forall (d31 : (Trace TyVar)) (S22 : Ty) ,
      ((subst_TyVar_Tm (XS TyVar d31) S22 (shift_TyVar_Tm C0 s21)) =
      (shift_TyVar_Tm C0 (subst_TyVar_Tm d31 S22 s21)))) := (subst_TyVar__shift_TyVar_0_Tm_comm_ind s21 H0).
    Definition subst_TyVar_Ty_shift_TyVar_Ty0_comm (S23 : Ty) : (forall (d31 : (Trace TyVar)) (S22 : Ty) ,
      ((subst_TyVar_Ty (XS TyVar d31) S22 (shift_TyVar_Ty C0 S23)) =
      (shift_TyVar_Ty C0 (subst_TyVar_Ty d31 S22 S23)))) := (subst_TyVar__shift_TyVar_0_Ty_comm_ind S23 H0).
    Definition subst_TmVar_Kind_shift_TmVar_Kind0_comm (K13 : Kind) : (forall (d31 : (Trace TmVar)) (s21 : Tm) ,
      ((subst_TmVar_Kind (XS TmVar d31) s21 (shift_TmVar_Kind C0 K13)) =
      (shift_TmVar_Kind C0 (subst_TmVar_Kind d31 s21 K13)))) := (subst_TmVar__shift_TmVar_0_Kind_comm_ind K13 H0).
    Definition subst_TmVar_Kind_shift_TyVar_Kind0_comm (K13 : Kind) : (forall (d31 : (Trace TmVar)) (s21 : Tm) ,
      ((subst_TmVar_Kind (XS TyVar d31) s21 (shift_TyVar_Kind C0 K13)) =
      (shift_TyVar_Kind C0 (subst_TmVar_Kind d31 s21 K13)))) := (subst_TmVar__shift_TyVar_0_Kind_comm_ind K13 H0).
    Definition subst_TyVar_Kind_shift_TmVar_Kind0_comm (K13 : Kind) : (forall (d31 : (Trace TyVar)) (S22 : Ty) ,
      ((subst_TyVar_Kind (XS TmVar d31) S22 (shift_TmVar_Kind C0 K13)) =
      (shift_TmVar_Kind C0 (subst_TyVar_Kind d31 S22 K13)))) := (subst_TyVar__shift_TmVar_0_Kind_comm_ind K13 H0).
    Definition subst_TyVar_Kind_shift_TyVar_Kind0_comm (K13 : Kind) : (forall (d31 : (Trace TyVar)) (S22 : Ty) ,
      ((subst_TyVar_Kind (XS TyVar d31) S22 (shift_TyVar_Kind C0 K13)) =
      (shift_TyVar_Kind C0 (subst_TyVar_Kind d31 S22 K13)))) := (subst_TyVar__shift_TyVar_0_Kind_comm_ind K13 H0).
  End SubstShiftComm.
  Section SubstSubordInd.
    
  End SubstSubordInd.
  Section SubstSubord.
    
  End SubstSubord.
End ShiftSubstInteraction.
 Hint Rewrite subst_TmVar_Kind0_shift_TmVar_Kind0_reflection subst_TyVar_Kind0_shift_TyVar_Kind0_reflection subst_TmVar_Tm0_shift_TmVar_Tm0_reflection subst_TyVar_Tm0_shift_TyVar_Tm0_reflection subst_TmVar_Ty0_shift_TmVar_Ty0_reflection subst_TyVar_Ty0_shift_TyVar_Ty0_reflection : interaction.
 Hint Rewrite subst_TmVar_Kind0_shift_TmVar_Kind0_reflection subst_TyVar_Kind0_shift_TyVar_Kind0_reflection subst_TmVar_Tm0_shift_TmVar_Tm0_reflection subst_TyVar_Tm0_shift_TyVar_Tm0_reflection subst_TmVar_Ty0_shift_TmVar_Ty0_reflection subst_TyVar_Ty0_shift_TyVar_Ty0_reflection : reflection.
 Hint Rewrite subst_TmVar_Kind_shift_TmVar_Kind0_comm subst_TmVar_Kind_shift_TyVar_Kind0_comm subst_TyVar_Kind_shift_TmVar_Kind0_comm subst_TyVar_Kind_shift_TyVar_Kind0_comm subst_TmVar_Tm_shift_TmVar_Tm0_comm subst_TmVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Tm_shift_TmVar_Tm0_comm subst_TyVar_Tm_shift_TyVar_Tm0_comm subst_TmVar_Ty_shift_TmVar_Ty0_comm subst_TmVar_Ty_shift_TyVar_Ty0_comm subst_TyVar_Ty_shift_TmVar_Ty0_comm subst_TyVar_Ty_shift_TyVar_Ty0_comm : interaction.
 Hint Rewrite subst_TmVar_Kind_shift_TmVar_Kind0_comm subst_TmVar_Kind_shift_TyVar_Kind0_comm subst_TyVar_Kind_shift_TmVar_Kind0_comm subst_TyVar_Kind_shift_TyVar_Kind0_comm subst_TmVar_Tm_shift_TmVar_Tm0_comm subst_TmVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Tm_shift_TmVar_Tm0_comm subst_TyVar_Tm_shift_TyVar_Tm0_comm subst_TmVar_Ty_shift_TmVar_Ty0_comm subst_TmVar_Ty_shift_TyVar_Ty0_comm subst_TyVar_Ty_shift_TmVar_Ty0_comm subst_TyVar_Ty_shift_TyVar_Ty0_comm : subst_shift0.
 Hint Rewrite shift_TmVar_Kind_subst_TmVar_Kind0_comm shift_TmVar_Kind_subst_TyVar_Kind0_comm shift_TyVar_Kind_subst_TmVar_Kind0_comm shift_TyVar_Kind_subst_TyVar_Kind0_comm shift_TmVar_Tm_subst_TmVar_Tm0_comm shift_TmVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Tm_subst_TmVar_Tm0_comm shift_TyVar_Tm_subst_TyVar_Tm0_comm shift_TmVar_Ty_subst_TmVar_Ty0_comm shift_TmVar_Ty_subst_TyVar_Ty0_comm shift_TyVar_Ty_subst_TmVar_Ty0_comm shift_TyVar_Ty_subst_TyVar_Ty0_comm : interaction.
 Hint Rewrite shift_TmVar_Kind_subst_TmVar_Kind0_comm shift_TmVar_Kind_subst_TyVar_Kind0_comm shift_TyVar_Kind_subst_TmVar_Kind0_comm shift_TyVar_Kind_subst_TyVar_Kind0_comm shift_TmVar_Tm_subst_TmVar_Tm0_comm shift_TmVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Tm_subst_TmVar_Tm0_comm shift_TyVar_Tm_subst_TyVar_Tm0_comm shift_TmVar_Ty_subst_TmVar_Ty0_comm shift_TmVar_Ty_subst_TyVar_Ty0_comm shift_TyVar_Ty_subst_TmVar_Ty0_comm shift_TyVar_Ty_subst_TyVar_Ty0_comm : shift_subst0.
Section SubstSubstInteraction.
  Section SubstSubstIndexCommInd.
    Lemma subst_TmVar_Tm_subst_TmVar_Index0_commright_ind (d31 : (Trace TmVar)) (s21 : Tm) (s22 : Tm) :
      (forall (k51 : Hvl) (x69 : (Index TmVar)) ,
        ((subst_TmVar_Tm (weakenTrace d31 k51) s21 (subst_TmVar_Index (weakenTrace X0 k51) s22 x69)) =
        (subst_TmVar_Tm (weakenTrace X0 k51) (subst_TmVar_Tm d31 s21 s22) (subst_TmVar_Index (weakenTrace (XS TmVar d31) k51) s21 x69)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Tm_subst_TmVar_Index0_commright_ind (d31 : (Trace TyVar)) (S22 : Ty) (s21 : Tm) :
      (forall (k51 : Hvl) (x69 : (Index TmVar)) ,
        ((subst_TyVar_Tm (weakenTrace d31 k51) S22 (subst_TmVar_Index (weakenTrace X0 k51) s21 x69)) =
        (subst_TmVar_Index (weakenTrace X0 k51) (subst_TyVar_Tm d31 S22 s21) x69))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Ty_subst_TyVar_Index0_commright_ind (d31 : (Trace TmVar)) (s21 : Tm) (S22 : Ty) :
      (forall (k51 : Hvl) (X27 : (Index TyVar)) ,
        ((subst_TmVar_Ty (weakenTrace d31 k51) s21 (subst_TyVar_Index (weakenTrace X0 k51) S22 X27)) =
        (subst_TyVar_Index (weakenTrace X0 k51) (subst_TmVar_Ty d31 s21 S22) X27))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Ty_subst_TyVar_Index0_commright_ind (d31 : (Trace TyVar)) (S22 : Ty) (S23 : Ty) :
      (forall (k51 : Hvl) (X27 : (Index TyVar)) ,
        ((subst_TyVar_Ty (weakenTrace d31 k51) S22 (subst_TyVar_Index (weakenTrace X0 k51) S23 X27)) =
        (subst_TyVar_Ty (weakenTrace X0 k51) (subst_TyVar_Ty d31 S22 S23) (subst_TyVar_Index (weakenTrace (XS TyVar d31) k51) S22 X27)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Ty_subst_TyVar_Index0_commleft_ind (d31 : (Trace TmVar)) (s21 : Tm) (S22 : Ty) :
      (forall (k51 : Hvl) (x69 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace d31 k51) s21 x69) =
        (subst_TyVar_Tm (weakenTrace X0 k51) (subst_TmVar_Ty d31 s21 S22) (subst_TmVar_Index (weakenTrace (XS TyVar d31) k51) s21 x69)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Tm_subst_TmVar_Index0_commleft_ind (d31 : (Trace TyVar)) (S22 : Ty) (s21 : Tm) :
      (forall (k51 : Hvl) (X27 : (Index TyVar)) ,
        ((subst_TyVar_Index (weakenTrace d31 k51) S22 X27) =
        (subst_TmVar_Ty (weakenTrace X0 k51) (subst_TyVar_Tm d31 S22 s21) (subst_TyVar_Index (weakenTrace (XS TmVar d31) k51) S22 X27)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
  End SubstSubstIndexCommInd.
  Section SubstSubstCommInd.
    Definition subst_TmVar__subst_TmVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s25 : Tm) =>
      (forall (k55 : Hvl) (d35 : (Trace TmVar)) (s26 : Tm) (s27 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace d35 k55) s26 (subst_TmVar_Tm (weakenTrace X0 k55) s27 s25)) =
        (subst_TmVar_Tm (weakenTrace X0 k55) (subst_TmVar_Tm d35 s26 s27) (subst_TmVar_Tm (weakenTrace (XS TmVar d35) k55) s26 s25))))) (fun (S26 : Ty) =>
      (forall (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (s26 : Tm) ,
        ((subst_TmVar_Ty (weakenTrace d35 k55) s25 (subst_TmVar_Ty (weakenTrace X0 k55) s26 S26)) =
        (subst_TmVar_Ty (weakenTrace X0 k55) (subst_TmVar_Tm d35 s25 s26) (subst_TmVar_Ty (weakenTrace (XS TmVar d35) k55) s25 S26))))) (fun (x79 : (Index TmVar)) =>
      (fun (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (s26 : Tm) =>
        (subst_TmVar_Tm_subst_TmVar_Index0_commright_ind d35 s25 s26 k55 x79))) (fun (T : Ty) IHT79 (t : Tm) IHt63 (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (s26 : Tm) =>
      (f_equal2 Abs (IHT79 k55 d35 s25 s26) (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d35 k55 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k55 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt63 (HS TmVar k55) d35 s25 s26) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k55 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TmVar d35) k55 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt63 (t2 : Tm) IHt64 (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (s26 : Tm) =>
      (f_equal2 App (IHt63 k55 d35 s25 s26) (IHt64 k55 d35 s25 s26))) (fun (X31 : (Index TyVar)) =>
      (fun (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (s26 : Tm) =>
        eq_refl)) (fun (T1 : Ty) IHT79 (T2 : Ty) IHT80 (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (s26 : Tm) =>
      (f_equal2 TPi (IHT79 k55 d35 s25 s26) (eq_trans (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar d35 k55 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar X0 k55 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT80 (HS TmVar k55) d35 s25 s26) (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar X0 k55 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar (XS TmVar d35) k55 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT79 (t : Tm) IHt63 (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (s26 : Tm) =>
      (f_equal2 TApp (IHT79 k55 d35 s25 s26) (IHt63 k55 d35 s25 s26)))).
    Lemma subst_TmVar__subst_TmVar_0_Tm_comm_ind  :
      (forall (s25 : Tm) ,
        (forall (k55 : Hvl) (d35 : (Trace TmVar)) (s26 : Tm) (s27 : Tm) ,
          ((subst_TmVar_Tm (weakenTrace d35 k55) s26 (subst_TmVar_Tm (weakenTrace X0 k55) s27 s25)) =
          (subst_TmVar_Tm (weakenTrace X0 k55) (subst_TmVar_Tm d35 s26 s27) (subst_TmVar_Tm (weakenTrace (XS TmVar d35) k55) s26 s25))))).
    Proof.
      pose proof (subst_TmVar__subst_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__subst_TmVar_0_Ty_comm_ind  :
      (forall (S26 : Ty) ,
        (forall (k55 : Hvl) (d35 : (Trace TmVar)) (s26 : Tm) (s27 : Tm) ,
          ((subst_TmVar_Ty (weakenTrace d35 k55) s26 (subst_TmVar_Ty (weakenTrace X0 k55) s27 S26)) =
          (subst_TmVar_Ty (weakenTrace X0 k55) (subst_TmVar_Tm d35 s26 s27) (subst_TmVar_Ty (weakenTrace (XS TmVar d35) k55) s26 S26))))).
    Proof.
      pose proof (subst_TmVar__subst_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TmVar__subst_TyVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s25 : Tm) =>
      (forall (k55 : Hvl) (d35 : (Trace TmVar)) (s26 : Tm) (S26 : Ty) ,
        ((subst_TmVar_Tm (weakenTrace d35 k55) s26 (subst_TyVar_Tm (weakenTrace X0 k55) S26 s25)) =
        (subst_TyVar_Tm (weakenTrace X0 k55) (subst_TmVar_Ty d35 s26 S26) (subst_TmVar_Tm (weakenTrace (XS TyVar d35) k55) s26 s25))))) (fun (S26 : Ty) =>
      (forall (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (S27 : Ty) ,
        ((subst_TmVar_Ty (weakenTrace d35 k55) s25 (subst_TyVar_Ty (weakenTrace X0 k55) S27 S26)) =
        (subst_TyVar_Ty (weakenTrace X0 k55) (subst_TmVar_Ty d35 s25 S27) (subst_TmVar_Ty (weakenTrace (XS TyVar d35) k55) s25 S26))))) (fun (x79 : (Index TmVar)) =>
      (fun (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (S26 : Ty) =>
        (subst_TmVar_Ty_subst_TyVar_Index0_commleft_ind d35 s25 S26 k55 x79))) (fun (T : Ty) IHT79 (t : Tm) IHt63 (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (S26 : Ty) =>
      (f_equal2 Abs (IHT79 k55 d35 s25 S26) (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d35 k55 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k55 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt63 (HS TmVar k55) d35 s25 S26) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k55 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TyVar d35) k55 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt63 (t2 : Tm) IHt64 (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (S26 : Ty) =>
      (f_equal2 App (IHt63 k55 d35 s25 S26) (IHt64 k55 d35 s25 S26))) (fun (X31 : (Index TyVar)) =>
      (fun (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (S26 : Ty) =>
        (subst_TmVar_Ty_subst_TyVar_Index0_commright_ind d35 s25 S26 k55 X31))) (fun (T1 : Ty) IHT79 (T2 : Ty) IHT80 (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (S26 : Ty) =>
      (f_equal2 TPi (IHT79 k55 d35 s25 S26) (eq_trans (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar d35 k55 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k55 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT80 (HS TmVar k55) d35 s25 S26) (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar X0 k55 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar (XS TyVar d35) k55 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT79 (t : Tm) IHt63 (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (S26 : Ty) =>
      (f_equal2 TApp (IHT79 k55 d35 s25 S26) (IHt63 k55 d35 s25 S26)))).
    Lemma subst_TmVar__subst_TyVar_0_Tm_comm_ind  :
      (forall (s25 : Tm) ,
        (forall (k55 : Hvl) (d35 : (Trace TmVar)) (s26 : Tm) (S26 : Ty) ,
          ((subst_TmVar_Tm (weakenTrace d35 k55) s26 (subst_TyVar_Tm (weakenTrace X0 k55) S26 s25)) =
          (subst_TyVar_Tm (weakenTrace X0 k55) (subst_TmVar_Ty d35 s26 S26) (subst_TmVar_Tm (weakenTrace (XS TyVar d35) k55) s26 s25))))).
    Proof.
      pose proof (subst_TmVar__subst_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TmVar__subst_TyVar_0_Ty_comm_ind  :
      (forall (S26 : Ty) ,
        (forall (k55 : Hvl) (d35 : (Trace TmVar)) (s26 : Tm) (S27 : Ty) ,
          ((subst_TmVar_Ty (weakenTrace d35 k55) s26 (subst_TyVar_Ty (weakenTrace X0 k55) S27 S26)) =
          (subst_TyVar_Ty (weakenTrace X0 k55) (subst_TmVar_Ty d35 s26 S27) (subst_TmVar_Ty (weakenTrace (XS TyVar d35) k55) s26 S26))))).
    Proof.
      pose proof (subst_TmVar__subst_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TyVar__subst_TmVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s25 : Tm) =>
      (forall (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (s26 : Tm) ,
        ((subst_TyVar_Tm (weakenTrace d35 k55) S26 (subst_TmVar_Tm (weakenTrace X0 k55) s26 s25)) =
        (subst_TmVar_Tm (weakenTrace X0 k55) (subst_TyVar_Tm d35 S26 s26) (subst_TyVar_Tm (weakenTrace (XS TmVar d35) k55) S26 s25))))) (fun (S26 : Ty) =>
      (forall (k55 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) (s25 : Tm) ,
        ((subst_TyVar_Ty (weakenTrace d35 k55) S27 (subst_TmVar_Ty (weakenTrace X0 k55) s25 S26)) =
        (subst_TmVar_Ty (weakenTrace X0 k55) (subst_TyVar_Tm d35 S27 s25) (subst_TyVar_Ty (weakenTrace (XS TmVar d35) k55) S27 S26))))) (fun (x79 : (Index TmVar)) =>
      (fun (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (s25 : Tm) =>
        (subst_TyVar_Tm_subst_TmVar_Index0_commright_ind d35 S26 s25 k55 x79))) (fun (T : Ty) IHT79 (t : Tm) IHt63 (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (s25 : Tm) =>
      (f_equal2 Abs (IHT79 k55 d35 S26 s25) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d35 k55 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k55 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt63 (HS TmVar k55) d35 S26 s25) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k55 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar (XS TmVar d35) k55 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt63 (t2 : Tm) IHt64 (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (s25 : Tm) =>
      (f_equal2 App (IHt63 k55 d35 S26 s25) (IHt64 k55 d35 S26 s25))) (fun (X31 : (Index TyVar)) =>
      (fun (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (s25 : Tm) =>
        (subst_TyVar_Tm_subst_TmVar_Index0_commleft_ind d35 S26 s25 k55 X31))) (fun (T1 : Ty) IHT79 (T2 : Ty) IHT80 (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (s25 : Tm) =>
      (f_equal2 TPi (IHT79 k55 d35 S26 s25) (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar d35 k55 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Ty (weakenTrace_append TmVar X0 k55 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT80 (HS TmVar k55) d35 S26 s25) (f_equal3 subst_TmVar_Ty (eq_sym (weakenTrace_append TmVar X0 k55 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar (XS TmVar d35) k55 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT79 (t : Tm) IHt63 (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (s25 : Tm) =>
      (f_equal2 TApp (IHT79 k55 d35 S26 s25) (IHt63 k55 d35 S26 s25)))).
    Lemma subst_TyVar__subst_TmVar_0_Tm_comm_ind  :
      (forall (s25 : Tm) ,
        (forall (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (s26 : Tm) ,
          ((subst_TyVar_Tm (weakenTrace d35 k55) S26 (subst_TmVar_Tm (weakenTrace X0 k55) s26 s25)) =
          (subst_TmVar_Tm (weakenTrace X0 k55) (subst_TyVar_Tm d35 S26 s26) (subst_TyVar_Tm (weakenTrace (XS TmVar d35) k55) S26 s25))))).
    Proof.
      pose proof (subst_TyVar__subst_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__subst_TmVar_0_Ty_comm_ind  :
      (forall (S26 : Ty) ,
        (forall (k55 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) (s26 : Tm) ,
          ((subst_TyVar_Ty (weakenTrace d35 k55) S27 (subst_TmVar_Ty (weakenTrace X0 k55) s26 S26)) =
          (subst_TmVar_Ty (weakenTrace X0 k55) (subst_TyVar_Tm d35 S27 s26) (subst_TyVar_Ty (weakenTrace (XS TmVar d35) k55) S27 S26))))).
    Proof.
      pose proof (subst_TyVar__subst_TmVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TyVar__subst_TyVar_0_Tm_Ty_comm_ind := (ind_Tm_Ty (fun (s25 : Tm) =>
      (forall (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace d35 k55) S26 (subst_TyVar_Tm (weakenTrace X0 k55) S27 s25)) =
        (subst_TyVar_Tm (weakenTrace X0 k55) (subst_TyVar_Ty d35 S26 S27) (subst_TyVar_Tm (weakenTrace (XS TyVar d35) k55) S26 s25))))) (fun (S26 : Ty) =>
      (forall (k55 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) (S28 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace d35 k55) S27 (subst_TyVar_Ty (weakenTrace X0 k55) S28 S26)) =
        (subst_TyVar_Ty (weakenTrace X0 k55) (subst_TyVar_Ty d35 S27 S28) (subst_TyVar_Ty (weakenTrace (XS TyVar d35) k55) S27 S26))))) (fun (x79 : (Index TmVar)) =>
      (fun (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) =>
        eq_refl)) (fun (T : Ty) IHT79 (t : Tm) IHt63 (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) =>
      (f_equal2 Abs (IHT79 k55 d35 S26 S27) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d35 k55 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k55 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt63 (HS TmVar k55) d35 S26 S27) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k55 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar (XS TyVar d35) k55 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt63 (t2 : Tm) IHt64 (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) =>
      (f_equal2 App (IHt63 k55 d35 S26 S27) (IHt64 k55 d35 S26 S27))) (fun (X31 : (Index TyVar)) =>
      (fun (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) =>
        (subst_TyVar_Ty_subst_TyVar_Index0_commright_ind d35 S26 S27 k55 X31))) (fun (T1 : Ty) IHT79 (T2 : Ty) IHT80 (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) =>
      (f_equal2 TPi (IHT79 k55 d35 S26 S27) (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar d35 k55 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k55 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHT80 (HS TmVar k55) d35 S26 S27) (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar X0 k55 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar (XS TyVar d35) k55 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (T : Ty) IHT79 (t : Tm) IHt63 (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) =>
      (f_equal2 TApp (IHT79 k55 d35 S26 S27) (IHt63 k55 d35 S26 S27)))).
    Lemma subst_TyVar__subst_TyVar_0_Tm_comm_ind  :
      (forall (s25 : Tm) ,
        (forall (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) ,
          ((subst_TyVar_Tm (weakenTrace d35 k55) S26 (subst_TyVar_Tm (weakenTrace X0 k55) S27 s25)) =
          (subst_TyVar_Tm (weakenTrace X0 k55) (subst_TyVar_Ty d35 S26 S27) (subst_TyVar_Tm (weakenTrace (XS TyVar d35) k55) S26 s25))))).
    Proof.
      pose proof (subst_TyVar__subst_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Lemma subst_TyVar__subst_TyVar_0_Ty_comm_ind  :
      (forall (S26 : Ty) ,
        (forall (k55 : Hvl) (d35 : (Trace TyVar)) (S27 : Ty) (S28 : Ty) ,
          ((subst_TyVar_Ty (weakenTrace d35 k55) S27 (subst_TyVar_Ty (weakenTrace X0 k55) S28 S26)) =
          (subst_TyVar_Ty (weakenTrace X0 k55) (subst_TyVar_Ty d35 S27 S28) (subst_TyVar_Ty (weakenTrace (XS TyVar d35) k55) S27 S26))))).
    Proof.
      pose proof (subst_TyVar__subst_TyVar_0_Tm_Ty_comm_ind).
      destruct_conjs .
      easy .
    Qed.
    Definition subst_TmVar__subst_TmVar_0_Kind_comm_ind := (ind_Kind (fun (K15 : Kind) =>
      (forall (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (s26 : Tm) ,
        ((subst_TmVar_Kind (weakenTrace d35 k55) s25 (subst_TmVar_Kind (weakenTrace X0 k55) s26 K15)) =
        (subst_TmVar_Kind (weakenTrace X0 k55) (subst_TmVar_Tm d35 s25 s26) (subst_TmVar_Kind (weakenTrace (XS TmVar d35) k55) s25 K15))))) (fun (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (s26 : Tm) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK15 (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (s26 : Tm) =>
      (f_equal2 KPi (let IHT79 := (subst_TmVar__subst_TmVar_0_Ty_comm_ind T) in
      (IHT79 k55 d35 s25 s26)) (eq_trans (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar d35 k55 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar X0 k55 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK15 (HS TmVar k55) d35 s25 s26) (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar X0 k55 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar (XS TmVar d35) k55 (HS TmVar H0))) eq_refl eq_refl))))))).
    Definition subst_TmVar__subst_TyVar_0_Kind_comm_ind := (ind_Kind (fun (K15 : Kind) =>
      (forall (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (S26 : Ty) ,
        ((subst_TmVar_Kind (weakenTrace d35 k55) s25 (subst_TyVar_Kind (weakenTrace X0 k55) S26 K15)) =
        (subst_TyVar_Kind (weakenTrace X0 k55) (subst_TmVar_Ty d35 s25 S26) (subst_TmVar_Kind (weakenTrace (XS TyVar d35) k55) s25 K15))))) (fun (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (S26 : Ty) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK15 (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (S26 : Ty) =>
      (f_equal2 KPi (let IHT79 := (subst_TmVar__subst_TyVar_0_Ty_comm_ind T) in
      (IHT79 k55 d35 s25 S26)) (eq_trans (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar d35 k55 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar X0 k55 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK15 (HS TmVar k55) d35 s25 S26) (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar X0 k55 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar (XS TyVar d35) k55 (HS TmVar H0))) eq_refl eq_refl))))))).
    Definition subst_TyVar__subst_TmVar_0_Kind_comm_ind := (ind_Kind (fun (K15 : Kind) =>
      (forall (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (s25 : Tm) ,
        ((subst_TyVar_Kind (weakenTrace d35 k55) S26 (subst_TmVar_Kind (weakenTrace X0 k55) s25 K15)) =
        (subst_TmVar_Kind (weakenTrace X0 k55) (subst_TyVar_Tm d35 S26 s25) (subst_TyVar_Kind (weakenTrace (XS TmVar d35) k55) S26 K15))))) (fun (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (s25 : Tm) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK15 (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (s25 : Tm) =>
      (f_equal2 KPi (let IHT79 := (subst_TyVar__subst_TmVar_0_Ty_comm_ind T) in
      (IHT79 k55 d35 S26 s25)) (eq_trans (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar d35 k55 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Kind (weakenTrace_append TmVar X0 k55 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK15 (HS TmVar k55) d35 S26 s25) (f_equal3 subst_TmVar_Kind (eq_sym (weakenTrace_append TmVar X0 k55 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar (XS TmVar d35) k55 (HS TmVar H0))) eq_refl eq_refl))))))).
    Definition subst_TyVar__subst_TyVar_0_Kind_comm_ind := (ind_Kind (fun (K15 : Kind) =>
      (forall (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) ,
        ((subst_TyVar_Kind (weakenTrace d35 k55) S26 (subst_TyVar_Kind (weakenTrace X0 k55) S27 K15)) =
        (subst_TyVar_Kind (weakenTrace X0 k55) (subst_TyVar_Ty d35 S26 S27) (subst_TyVar_Kind (weakenTrace (XS TyVar d35) k55) S26 K15))))) (fun (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) =>
      eq_refl) (fun (T : Ty) (K : Kind) IHK15 (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) =>
      (f_equal2 KPi (let IHT79 := (subst_TyVar__subst_TyVar_0_Ty_comm_ind T) in
      (IHT79 k55 d35 S26 S27)) (eq_trans (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar d35 k55 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Kind (weakenTrace_append TyVar X0 k55 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHK15 (HS TmVar k55) d35 S26 S27) (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar X0 k55 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Kind (eq_sym (weakenTrace_append TyVar (XS TyVar d35) k55 (HS TmVar H0))) eq_refl eq_refl))))))).
  End SubstSubstCommInd.
  Section SubstSubstComm.
    Definition subst_TmVar_Tm_subst_TmVar_Tm0_comm (s27 : Tm) : (forall (d35 : (Trace TmVar)) (s25 : Tm) (s26 : Tm) ,
      ((subst_TmVar_Tm d35 s25 (subst_TmVar_Tm X0 s26 s27)) =
      (subst_TmVar_Tm X0 (subst_TmVar_Tm d35 s25 s26) (subst_TmVar_Tm (XS TmVar d35) s25 s27)))) := (subst_TmVar__subst_TmVar_0_Tm_comm_ind s27 H0).
    Definition subst_TmVar_Ty_subst_TmVar_Ty0_comm (S26 : Ty) : (forall (d35 : (Trace TmVar)) (s25 : Tm) (s26 : Tm) ,
      ((subst_TmVar_Ty d35 s25 (subst_TmVar_Ty X0 s26 S26)) =
      (subst_TmVar_Ty X0 (subst_TmVar_Tm d35 s25 s26) (subst_TmVar_Ty (XS TmVar d35) s25 S26)))) := (subst_TmVar__subst_TmVar_0_Ty_comm_ind S26 H0).
    Definition subst_TmVar_Tm_subst_TyVar_Tm0_comm (s26 : Tm) : (forall (d35 : (Trace TmVar)) (s25 : Tm) (S26 : Ty) ,
      ((subst_TmVar_Tm d35 s25 (subst_TyVar_Tm X0 S26 s26)) =
      (subst_TyVar_Tm X0 (subst_TmVar_Ty d35 s25 S26) (subst_TmVar_Tm (XS TyVar d35) s25 s26)))) := (subst_TmVar__subst_TyVar_0_Tm_comm_ind s26 H0).
    Definition subst_TmVar_Ty_subst_TyVar_Ty0_comm (S27 : Ty) : (forall (d35 : (Trace TmVar)) (s25 : Tm) (S26 : Ty) ,
      ((subst_TmVar_Ty d35 s25 (subst_TyVar_Ty X0 S26 S27)) =
      (subst_TyVar_Ty X0 (subst_TmVar_Ty d35 s25 S26) (subst_TmVar_Ty (XS TyVar d35) s25 S27)))) := (subst_TmVar__subst_TyVar_0_Ty_comm_ind S27 H0).
    Definition subst_TyVar_Tm_subst_TmVar_Tm0_comm (s26 : Tm) : (forall (d35 : (Trace TyVar)) (S26 : Ty) (s25 : Tm) ,
      ((subst_TyVar_Tm d35 S26 (subst_TmVar_Tm X0 s25 s26)) =
      (subst_TmVar_Tm X0 (subst_TyVar_Tm d35 S26 s25) (subst_TyVar_Tm (XS TmVar d35) S26 s26)))) := (subst_TyVar__subst_TmVar_0_Tm_comm_ind s26 H0).
    Definition subst_TyVar_Ty_subst_TmVar_Ty0_comm (S27 : Ty) : (forall (d35 : (Trace TyVar)) (S26 : Ty) (s25 : Tm) ,
      ((subst_TyVar_Ty d35 S26 (subst_TmVar_Ty X0 s25 S27)) =
      (subst_TmVar_Ty X0 (subst_TyVar_Tm d35 S26 s25) (subst_TyVar_Ty (XS TmVar d35) S26 S27)))) := (subst_TyVar__subst_TmVar_0_Ty_comm_ind S27 H0).
    Definition subst_TyVar_Tm_subst_TyVar_Tm0_comm (s25 : Tm) : (forall (d35 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) ,
      ((subst_TyVar_Tm d35 S26 (subst_TyVar_Tm X0 S27 s25)) =
      (subst_TyVar_Tm X0 (subst_TyVar_Ty d35 S26 S27) (subst_TyVar_Tm (XS TyVar d35) S26 s25)))) := (subst_TyVar__subst_TyVar_0_Tm_comm_ind s25 H0).
    Definition subst_TyVar_Ty_subst_TyVar_Ty0_comm (S28 : Ty) : (forall (d35 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) ,
      ((subst_TyVar_Ty d35 S26 (subst_TyVar_Ty X0 S27 S28)) =
      (subst_TyVar_Ty X0 (subst_TyVar_Ty d35 S26 S27) (subst_TyVar_Ty (XS TyVar d35) S26 S28)))) := (subst_TyVar__subst_TyVar_0_Ty_comm_ind S28 H0).
    Definition subst_TmVar_Kind_subst_TmVar_Kind0_comm (K15 : Kind) : (forall (d35 : (Trace TmVar)) (s25 : Tm) (s26 : Tm) ,
      ((subst_TmVar_Kind d35 s25 (subst_TmVar_Kind X0 s26 K15)) =
      (subst_TmVar_Kind X0 (subst_TmVar_Tm d35 s25 s26) (subst_TmVar_Kind (XS TmVar d35) s25 K15)))) := (subst_TmVar__subst_TmVar_0_Kind_comm_ind K15 H0).
    Definition subst_TmVar_Kind_subst_TyVar_Kind0_comm (K15 : Kind) : (forall (d35 : (Trace TmVar)) (s25 : Tm) (S26 : Ty) ,
      ((subst_TmVar_Kind d35 s25 (subst_TyVar_Kind X0 S26 K15)) =
      (subst_TyVar_Kind X0 (subst_TmVar_Ty d35 s25 S26) (subst_TmVar_Kind (XS TyVar d35) s25 K15)))) := (subst_TmVar__subst_TyVar_0_Kind_comm_ind K15 H0).
    Definition subst_TyVar_Kind_subst_TmVar_Kind0_comm (K15 : Kind) : (forall (d35 : (Trace TyVar)) (S26 : Ty) (s25 : Tm) ,
      ((subst_TyVar_Kind d35 S26 (subst_TmVar_Kind X0 s25 K15)) =
      (subst_TmVar_Kind X0 (subst_TyVar_Tm d35 S26 s25) (subst_TyVar_Kind (XS TmVar d35) S26 K15)))) := (subst_TyVar__subst_TmVar_0_Kind_comm_ind K15 H0).
    Definition subst_TyVar_Kind_subst_TyVar_Kind0_comm (K15 : Kind) : (forall (d35 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) ,
      ((subst_TyVar_Kind d35 S26 (subst_TyVar_Kind X0 S27 K15)) =
      (subst_TyVar_Kind X0 (subst_TyVar_Ty d35 S26 S27) (subst_TyVar_Kind (XS TyVar d35) S26 K15)))) := (subst_TyVar__subst_TyVar_0_Kind_comm_ind K15 H0).
  End SubstSubstComm.
  Section WeakenSubst.
    Lemma weakenTm_subst_TmVar_Tm  :
      (forall (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (s26 : Tm) ,
        ((weakenTm (subst_TmVar_Tm d35 s25 s26) k55) =
        (subst_TmVar_Tm (weakenTrace d35 k55) s25 (weakenTm s26 k55)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTy_subst_TmVar_Ty  :
      (forall (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (S26 : Ty) ,
        ((weakenTy (subst_TmVar_Ty d35 s25 S26) k55) =
        (subst_TmVar_Ty (weakenTrace d35 k55) s25 (weakenTy S26 k55)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_subst_TyVar_Tm  :
      (forall (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (s25 : Tm) ,
        ((weakenTm (subst_TyVar_Tm d35 S26 s25) k55) =
        (subst_TyVar_Tm (weakenTrace d35 k55) S26 (weakenTm s25 k55)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTy_subst_TyVar_Ty  :
      (forall (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (S27 : Ty) ,
        ((weakenTy (subst_TyVar_Ty d35 S26 S27) k55) =
        (subst_TyVar_Ty (weakenTrace d35 k55) S26 (weakenTy S27 k55)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenKind_subst_TmVar_Kind  :
      (forall (k55 : Hvl) (d35 : (Trace TmVar)) (s25 : Tm) (K15 : Kind) ,
        ((weakenKind (subst_TmVar_Kind d35 s25 K15) k55) =
        (subst_TmVar_Kind (weakenTrace d35 k55) s25 (weakenKind K15 k55)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenKind_subst_TyVar_Kind  :
      (forall (k55 : Hvl) (d35 : (Trace TyVar)) (S26 : Ty) (K15 : Kind) ,
        ((weakenKind (subst_TyVar_Kind d35 S26 K15) k55) =
        (subst_TyVar_Kind (weakenTrace d35 k55) S26 (weakenKind K15 k55)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
  End WeakenSubst.
  Section WeakenAppend.
    Lemma weakenTm_append  :
      (forall (s25 : Tm) (k55 : Hvl) (k56 : Hvl) ,
        ((weakenTm (weakenTm s25 k55) k56) =
        (weakenTm s25 (appendHvl k55 k56)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenTy_append  :
      (forall (S26 : Ty) (k55 : Hvl) (k56 : Hvl) ,
        ((weakenTy (weakenTy S26 k55) k56) =
        (weakenTy S26 (appendHvl k55 k56)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenKind_append  :
      (forall (K15 : Kind) (k55 : Hvl) (k56 : Hvl) ,
        ((weakenKind (weakenKind K15 k55) k56) =
        (weakenKind K15 (appendHvl k55 k56)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
  End WeakenAppend.
End SubstSubstInteraction.
 Hint Rewrite subst_TmVar_Kind_subst_TmVar_Kind0_comm subst_TyVar_Kind_subst_TyVar_Kind0_comm subst_TmVar_Tm_subst_TmVar_Tm0_comm subst_TyVar_Tm_subst_TyVar_Tm0_comm subst_TmVar_Ty_subst_TmVar_Ty0_comm subst_TyVar_Ty_subst_TyVar_Ty0_comm : interaction.
 Hint Rewrite subst_TmVar_Kind_subst_TmVar_Kind0_comm subst_TyVar_Kind_subst_TyVar_Kind0_comm subst_TmVar_Tm_subst_TmVar_Tm0_comm subst_TyVar_Tm_subst_TyVar_Tm0_comm subst_TmVar_Ty_subst_TmVar_Ty0_comm subst_TyVar_Ty_subst_TyVar_Ty0_comm : subst_subst0.
 Hint Rewrite weakenKind_shift_TmVar_Kind weakenKind_shift_TyVar_Kind weakenTm_shift_TmVar_Tm weakenTm_shift_TyVar_Tm weakenTy_shift_TmVar_Ty weakenTy_shift_TyVar_Ty : weaken_shift.
 Hint Rewrite weakenKind_subst_TmVar_Kind weakenKind_subst_TyVar_Kind weakenTm_subst_TmVar_Tm weakenTm_subst_TyVar_Tm weakenTy_subst_TmVar_Ty weakenTy_subst_TyVar_Ty : weaken_subst.
Section WellFormed.
  Fixpoint wfindex {a : Namespace} (k55 : Hvl) (i : (Index a)) {struct k55} : Prop :=
    match k55 with
      | (H0) => False
      | (HS b k55) => match (eq_namespace_dec a b) with
        | (left e) => match i with
          | (I0) => True
          | (IS i) => (wfindex k55 i)
        end
        | (right n) => (wfindex k55 i)
      end
    end.
  Inductive wfTm (k55 : Hvl) : Tm -> Prop :=
    | wfTm_Var
        {x89 : (Index TmVar)}
        (wfi : (wfindex k55 x89)) :
        (wfTm k55 (Var x89))
    | wfTm_Abs {T89 : Ty}
        (wf : (wfTy k55 T89)) {t71 : Tm}
        (wf0 : (wfTm (HS TmVar k55) t71))
        : (wfTm k55 (Abs T89 t71))
    | wfTm_App {t72 : Tm}
        (wf : (wfTm k55 t72)) {t73 : Tm}
        (wf0 : (wfTm k55 t73)) :
        (wfTm k55 (App t72 t73))
  with wfTy (k55 : Hvl) : Ty -> Prop :=
    | wfTy_TVar
        {X35 : (Index TyVar)}
        (wfi : (wfindex k55 X35)) :
        (wfTy k55 (TVar X35))
    | wfTy_TPi {T89 : Ty}
        (wf : (wfTy k55 T89)) {T90 : Ty}
        (wf0 : (wfTy (HS TmVar k55) T90))
        : (wfTy k55 (TPi T89 T90))
    | wfTy_TApp {T91 : Ty}
        (wf : (wfTy k55 T91)) {t71 : Tm}
        (wf0 : (wfTm k55 t71)) :
        (wfTy k55 (TApp T91 t71)).
  Inductive wfKind (k55 : Hvl) : Kind -> Prop :=
    | wfKind_Star :
        (wfKind k55 (Star))
    | wfKind_KPi {T89 : Ty}
        (wf : (wfTy k55 T89))
        {K17 : Kind}
        (wf0 : (wfKind (HS TmVar k55) K17))
        : (wfKind k55 (KPi T89 K17)).
  Scheme ind_wfTm := Induction for wfTm Sort Prop
  with ind_wfTy := Induction for wfTy Sort Prop.
  Combined Scheme ind_wfTm_Ty from ind_wfTm, ind_wfTy.
  Scheme ind_wfKind := Induction for wfKind Sort Prop.
End WellFormed.
Section ShiftWellFormed.
  Inductive shifthvl_TmVar : (forall (c19 : (Cutoff TmVar)) (k55 : Hvl) (k56 : Hvl) ,
    Prop) :=
    | shifthvl_TmVar_here
        (k55 : Hvl) :
        (shifthvl_TmVar C0 k55 (HS TmVar k55))
    | shifthvl_TmVar_there_TmVar
        (c19 : (Cutoff TmVar))
        (k55 : Hvl) (k56 : Hvl) :
        (shifthvl_TmVar c19 k55 k56) -> (shifthvl_TmVar (CS c19) (HS TmVar k55) (HS TmVar k56))
    | shifthvl_TmVar_there_TyVar
        (c19 : (Cutoff TmVar))
        (k55 : Hvl) (k56 : Hvl) :
        (shifthvl_TmVar c19 k55 k56) -> (shifthvl_TmVar c19 (HS TyVar k55) (HS TyVar k56)).
  Inductive shifthvl_TyVar : (forall (c19 : (Cutoff TyVar)) (k55 : Hvl) (k56 : Hvl) ,
    Prop) :=
    | shifthvl_TyVar_here
        (k55 : Hvl) :
        (shifthvl_TyVar C0 k55 (HS TyVar k55))
    | shifthvl_TyVar_there_TmVar
        (c19 : (Cutoff TyVar))
        (k55 : Hvl) (k56 : Hvl) :
        (shifthvl_TyVar c19 k55 k56) -> (shifthvl_TyVar c19 (HS TmVar k55) (HS TmVar k56))
    | shifthvl_TyVar_there_TyVar
        (c19 : (Cutoff TyVar))
        (k55 : Hvl) (k56 : Hvl) :
        (shifthvl_TyVar c19 k55 k56) -> (shifthvl_TyVar (CS c19) (HS TyVar k55) (HS TyVar k56)).
  Lemma weaken_shifthvl_TmVar  :
    (forall (k55 : Hvl) {c19 : (Cutoff TmVar)} {k56 : Hvl} {k57 : Hvl} ,
      (shifthvl_TmVar c19 k56 k57) -> (shifthvl_TmVar (weakenCutoffTmVar c19 k55) (appendHvl k56 k55) (appendHvl k57 k55))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma weaken_shifthvl_TyVar  :
    (forall (k55 : Hvl) {c19 : (Cutoff TyVar)} {k56 : Hvl} {k57 : Hvl} ,
      (shifthvl_TyVar c19 k56 k57) -> (shifthvl_TyVar (weakenCutoffTyVar c19 k55) (appendHvl k56 k55) (appendHvl k57 k55))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma shift_TmVar__wfindex_TmVar  :
    (forall (c19 : (Cutoff TmVar)) (k55 : Hvl) (k56 : Hvl) (ins : (shifthvl_TmVar c19 k55 k56)) (x89 : (Index TmVar)) ,
      (wfindex k55 x89) -> (wfindex k56 (shift_TmVar_Index c19 x89))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TmVar__wfindex_TyVar  :
    (forall (c19 : (Cutoff TmVar)) (k55 : Hvl) (k56 : Hvl) (ins : (shifthvl_TmVar c19 k55 k56)) (X35 : (Index TyVar)) ,
      (wfindex k55 X35) -> (wfindex k56 X35)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TyVar__wfindex_TmVar  :
    (forall (c19 : (Cutoff TyVar)) (k55 : Hvl) (k56 : Hvl) (ins : (shifthvl_TyVar c19 k55 k56)) (x89 : (Index TmVar)) ,
      (wfindex k55 x89) -> (wfindex k56 x89)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TyVar__wfindex_TyVar  :
    (forall (c19 : (Cutoff TyVar)) (k55 : Hvl) (k56 : Hvl) (ins : (shifthvl_TyVar c19 k55 k56)) (X35 : (Index TyVar)) ,
      (wfindex k55 X35) -> (wfindex k56 (shift_TyVar_Index c19 X35))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Definition shift_TmVar__wfTm_Ty : (forall (k55 : Hvl) ,
    (forall (s25 : Tm) (wf : (wfTm k55 s25)) ,
      (forall (c19 : (Cutoff TmVar)) (k56 : Hvl) ,
        (shifthvl_TmVar c19 k55 k56) -> (wfTm k56 (shift_TmVar_Tm c19 s25)))) /\
    (forall (S26 : Ty) (wf : (wfTy k55 S26)) ,
      (forall (c19 : (Cutoff TmVar)) (k56 : Hvl) ,
        (shifthvl_TmVar c19 k55 k56) -> (wfTy k56 (shift_TmVar_Ty c19 S26))))) := (ind_wfTm_Ty (fun (k55 : Hvl) (s25 : Tm) (wf : (wfTm k55 s25)) =>
    (forall (c19 : (Cutoff TmVar)) (k56 : Hvl) ,
      (shifthvl_TmVar c19 k55 k56) -> (wfTm k56 (shift_TmVar_Tm c19 s25)))) (fun (k55 : Hvl) (S26 : Ty) (wf : (wfTy k55 S26)) =>
    (forall (c19 : (Cutoff TmVar)) (k56 : Hvl) ,
      (shifthvl_TmVar c19 k55 k56) -> (wfTy k56 (shift_TmVar_Ty c19 S26)))) (fun (k55 : Hvl) (x89 : (Index TmVar)) (wfi : (wfindex k55 x89)) (c19 : (Cutoff TmVar)) (k56 : Hvl) (ins : (shifthvl_TmVar c19 k55 k56)) =>
    (wfTm_Var k56 (shift_TmVar__wfindex_TmVar c19 k55 k56 ins x89 wfi))) (fun (k55 : Hvl) (T : Ty) (wf : (wfTy k55 T)) IHT89 (t : Tm) (wf0 : (wfTm (HS TmVar k55) t)) IHt71 (c19 : (Cutoff TmVar)) (k56 : Hvl) (ins : (shifthvl_TmVar c19 k55 k56)) =>
    (wfTm_Abs k56 (IHT89 c19 k56 (weaken_shifthvl_TmVar H0 ins)) (IHt71 (CS c19) (HS TmVar k56) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k55 : Hvl) (t1 : Tm) (wf : (wfTm k55 t1)) IHt71 (t2 : Tm) (wf0 : (wfTm k55 t2)) IHt72 (c19 : (Cutoff TmVar)) (k56 : Hvl) (ins : (shifthvl_TmVar c19 k55 k56)) =>
    (wfTm_App k56 (IHt71 c19 k56 (weaken_shifthvl_TmVar H0 ins)) (IHt72 c19 k56 (weaken_shifthvl_TmVar H0 ins)))) (fun (k55 : Hvl) (X35 : (Index TyVar)) (wfi : (wfindex k55 X35)) (c19 : (Cutoff TmVar)) (k56 : Hvl) (ins : (shifthvl_TmVar c19 k55 k56)) =>
    (wfTy_TVar k56 (shift_TmVar__wfindex_TyVar c19 k55 k56 ins X35 wfi))) (fun (k55 : Hvl) (T1 : Ty) (wf : (wfTy k55 T1)) IHT89 (T2 : Ty) (wf0 : (wfTy (HS TmVar k55) T2)) IHT90 (c19 : (Cutoff TmVar)) (k56 : Hvl) (ins : (shifthvl_TmVar c19 k55 k56)) =>
    (wfTy_TPi k56 (IHT89 c19 k56 (weaken_shifthvl_TmVar H0 ins)) (IHT90 (CS c19) (HS TmVar k56) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k55 : Hvl) (T : Ty) (wf : (wfTy k55 T)) IHT89 (t : Tm) (wf0 : (wfTm k55 t)) IHt71 (c19 : (Cutoff TmVar)) (k56 : Hvl) (ins : (shifthvl_TmVar c19 k55 k56)) =>
    (wfTy_TApp k56 (IHT89 c19 k56 (weaken_shifthvl_TmVar H0 ins)) (IHt71 c19 k56 (weaken_shifthvl_TmVar H0 ins))))).
  Lemma shift_TmVar__wfTm (k55 : Hvl) :
    (forall (s25 : Tm) (wf : (wfTm k55 s25)) ,
      (forall (c19 : (Cutoff TmVar)) (k56 : Hvl) ,
        (shifthvl_TmVar c19 k55 k56) -> (wfTm k56 (shift_TmVar_Tm c19 s25)))).
  Proof.
    pose proof ((shift_TmVar__wfTm_Ty k55)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_TmVar__wfTy (k55 : Hvl) :
    (forall (S26 : Ty) (wf0 : (wfTy k55 S26)) ,
      (forall (c20 : (Cutoff TmVar)) (k57 : Hvl) ,
        (shifthvl_TmVar c20 k55 k57) -> (wfTy k57 (shift_TmVar_Ty c20 S26)))).
  Proof.
    pose proof ((shift_TmVar__wfTm_Ty k55)).
    destruct_conjs .
    easy .
  Qed.
  Definition shift_TyVar__wfTm_Ty : (forall (k55 : Hvl) ,
    (forall (s25 : Tm) (wf : (wfTm k55 s25)) ,
      (forall (c19 : (Cutoff TyVar)) (k56 : Hvl) ,
        (shifthvl_TyVar c19 k55 k56) -> (wfTm k56 (shift_TyVar_Tm c19 s25)))) /\
    (forall (S26 : Ty) (wf : (wfTy k55 S26)) ,
      (forall (c19 : (Cutoff TyVar)) (k56 : Hvl) ,
        (shifthvl_TyVar c19 k55 k56) -> (wfTy k56 (shift_TyVar_Ty c19 S26))))) := (ind_wfTm_Ty (fun (k55 : Hvl) (s25 : Tm) (wf : (wfTm k55 s25)) =>
    (forall (c19 : (Cutoff TyVar)) (k56 : Hvl) ,
      (shifthvl_TyVar c19 k55 k56) -> (wfTm k56 (shift_TyVar_Tm c19 s25)))) (fun (k55 : Hvl) (S26 : Ty) (wf : (wfTy k55 S26)) =>
    (forall (c19 : (Cutoff TyVar)) (k56 : Hvl) ,
      (shifthvl_TyVar c19 k55 k56) -> (wfTy k56 (shift_TyVar_Ty c19 S26)))) (fun (k55 : Hvl) (x89 : (Index TmVar)) (wfi : (wfindex k55 x89)) (c19 : (Cutoff TyVar)) (k56 : Hvl) (ins : (shifthvl_TyVar c19 k55 k56)) =>
    (wfTm_Var k56 (shift_TyVar__wfindex_TmVar c19 k55 k56 ins x89 wfi))) (fun (k55 : Hvl) (T : Ty) (wf : (wfTy k55 T)) IHT89 (t : Tm) (wf0 : (wfTm (HS TmVar k55) t)) IHt71 (c19 : (Cutoff TyVar)) (k56 : Hvl) (ins : (shifthvl_TyVar c19 k55 k56)) =>
    (wfTm_Abs k56 (IHT89 c19 k56 (weaken_shifthvl_TyVar H0 ins)) (IHt71 c19 (HS TmVar k56) (weaken_shifthvl_TyVar (HS TmVar H0) ins)))) (fun (k55 : Hvl) (t1 : Tm) (wf : (wfTm k55 t1)) IHt71 (t2 : Tm) (wf0 : (wfTm k55 t2)) IHt72 (c19 : (Cutoff TyVar)) (k56 : Hvl) (ins : (shifthvl_TyVar c19 k55 k56)) =>
    (wfTm_App k56 (IHt71 c19 k56 (weaken_shifthvl_TyVar H0 ins)) (IHt72 c19 k56 (weaken_shifthvl_TyVar H0 ins)))) (fun (k55 : Hvl) (X35 : (Index TyVar)) (wfi : (wfindex k55 X35)) (c19 : (Cutoff TyVar)) (k56 : Hvl) (ins : (shifthvl_TyVar c19 k55 k56)) =>
    (wfTy_TVar k56 (shift_TyVar__wfindex_TyVar c19 k55 k56 ins X35 wfi))) (fun (k55 : Hvl) (T1 : Ty) (wf : (wfTy k55 T1)) IHT89 (T2 : Ty) (wf0 : (wfTy (HS TmVar k55) T2)) IHT90 (c19 : (Cutoff TyVar)) (k56 : Hvl) (ins : (shifthvl_TyVar c19 k55 k56)) =>
    (wfTy_TPi k56 (IHT89 c19 k56 (weaken_shifthvl_TyVar H0 ins)) (IHT90 c19 (HS TmVar k56) (weaken_shifthvl_TyVar (HS TmVar H0) ins)))) (fun (k55 : Hvl) (T : Ty) (wf : (wfTy k55 T)) IHT89 (t : Tm) (wf0 : (wfTm k55 t)) IHt71 (c19 : (Cutoff TyVar)) (k56 : Hvl) (ins : (shifthvl_TyVar c19 k55 k56)) =>
    (wfTy_TApp k56 (IHT89 c19 k56 (weaken_shifthvl_TyVar H0 ins)) (IHt71 c19 k56 (weaken_shifthvl_TyVar H0 ins))))).
  Lemma shift_TyVar__wfTm (k55 : Hvl) :
    (forall (s25 : Tm) (wf : (wfTm k55 s25)) ,
      (forall (c19 : (Cutoff TyVar)) (k56 : Hvl) ,
        (shifthvl_TyVar c19 k55 k56) -> (wfTm k56 (shift_TyVar_Tm c19 s25)))).
  Proof.
    pose proof ((shift_TyVar__wfTm_Ty k55)).
    destruct_conjs .
    easy .
  Qed.
  Lemma shift_TyVar__wfTy (k55 : Hvl) :
    (forall (S26 : Ty) (wf0 : (wfTy k55 S26)) ,
      (forall (c20 : (Cutoff TyVar)) (k57 : Hvl) ,
        (shifthvl_TyVar c20 k55 k57) -> (wfTy k57 (shift_TyVar_Ty c20 S26)))).
  Proof.
    pose proof ((shift_TyVar__wfTm_Ty k55)).
    destruct_conjs .
    easy .
  Qed.
  Definition shift_TmVar__wfKind : (forall (k55 : Hvl) ,
    (forall (K17 : Kind) (wf : (wfKind k55 K17)) ,
      (forall (c19 : (Cutoff TmVar)) (k56 : Hvl) ,
        (shifthvl_TmVar c19 k55 k56) -> (wfKind k56 (shift_TmVar_Kind c19 K17))))) := (ind_wfKind (fun (k55 : Hvl) (K17 : Kind) (wf : (wfKind k55 K17)) =>
    (forall (c19 : (Cutoff TmVar)) (k56 : Hvl) ,
      (shifthvl_TmVar c19 k55 k56) -> (wfKind k56 (shift_TmVar_Kind c19 K17)))) (fun (k55 : Hvl) (c19 : (Cutoff TmVar)) (k56 : Hvl) (ins : (shifthvl_TmVar c19 k55 k56)) =>
    (wfKind_Star k56)) (fun (k55 : Hvl) (T : Ty) (wf : (wfTy k55 T)) (K : Kind) (wf0 : (wfKind (HS TmVar k55) K)) IHK17 (c19 : (Cutoff TmVar)) (k56 : Hvl) (ins : (shifthvl_TmVar c19 k55 k56)) =>
    (wfKind_KPi k56 (shift_TmVar__wfTy k55 T wf c19 k56 (weaken_shifthvl_TmVar H0 ins)) (IHK17 (CS c19) (HS TmVar k56) (weaken_shifthvl_TmVar (HS TmVar H0) ins))))).
  Definition shift_TyVar__wfKind : (forall (k55 : Hvl) ,
    (forall (K17 : Kind) (wf : (wfKind k55 K17)) ,
      (forall (c19 : (Cutoff TyVar)) (k56 : Hvl) ,
        (shifthvl_TyVar c19 k55 k56) -> (wfKind k56 (shift_TyVar_Kind c19 K17))))) := (ind_wfKind (fun (k55 : Hvl) (K17 : Kind) (wf : (wfKind k55 K17)) =>
    (forall (c19 : (Cutoff TyVar)) (k56 : Hvl) ,
      (shifthvl_TyVar c19 k55 k56) -> (wfKind k56 (shift_TyVar_Kind c19 K17)))) (fun (k55 : Hvl) (c19 : (Cutoff TyVar)) (k56 : Hvl) (ins : (shifthvl_TyVar c19 k55 k56)) =>
    (wfKind_Star k56)) (fun (k55 : Hvl) (T : Ty) (wf : (wfTy k55 T)) (K : Kind) (wf0 : (wfKind (HS TmVar k55) K)) IHK17 (c19 : (Cutoff TyVar)) (k56 : Hvl) (ins : (shifthvl_TyVar c19 k55 k56)) =>
    (wfKind_KPi k56 (shift_TyVar__wfTy k55 T wf c19 k56 (weaken_shifthvl_TyVar H0 ins)) (IHK17 c19 (HS TmVar k56) (weaken_shifthvl_TyVar (HS TmVar H0) ins))))).
End ShiftWellFormed.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : infra.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : shift.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : shift_wf.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TyVar : wf.
 Hint Resolve shift_TmVar__wfKind shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfKind shift_TyVar__wfTm shift_TyVar__wfTy : infra.
 Hint Resolve shift_TmVar__wfKind shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfKind shift_TyVar__wfTm shift_TyVar__wfTy : shift.
 Hint Resolve shift_TmVar__wfKind shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfKind shift_TyVar__wfTm shift_TyVar__wfTy : shift_wf.
 Hint Resolve shift_TmVar__wfKind shift_TmVar__wfTm shift_TmVar__wfTy shift_TyVar__wfKind shift_TyVar__wfTm shift_TyVar__wfTy : wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : infra.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift_wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : infra.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift_wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : weaken.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : wf.
Section SubstWellFormed.
  Inductive substhvl_TmVar (k55 : Hvl) : (forall (d35 : (Trace TmVar)) (k56 : Hvl) (k57 : Hvl) ,
    Prop) :=
    | substhvl_TmVar_here :
        (substhvl_TmVar k55 X0 (HS TmVar k55) k55)
    | substhvl_TmVar_there_TmVar
        {d35 : (Trace TmVar)}
        {k56 : Hvl} {k57 : Hvl} :
        (substhvl_TmVar k55 d35 k56 k57) -> (substhvl_TmVar k55 (XS TmVar d35) (HS TmVar k56) (HS TmVar k57))
    | substhvl_TmVar_there_TyVar
        {d35 : (Trace TmVar)}
        {k56 : Hvl} {k57 : Hvl} :
        (substhvl_TmVar k55 d35 k56 k57) -> (substhvl_TmVar k55 (XS TyVar d35) (HS TyVar k56) (HS TyVar k57)).
  Inductive substhvl_TyVar (k55 : Hvl) : (forall (d35 : (Trace TyVar)) (k56 : Hvl) (k57 : Hvl) ,
    Prop) :=
    | substhvl_TyVar_here :
        (substhvl_TyVar k55 X0 (HS TyVar k55) k55)
    | substhvl_TyVar_there_TmVar
        {d35 : (Trace TyVar)}
        {k56 : Hvl} {k57 : Hvl} :
        (substhvl_TyVar k55 d35 k56 k57) -> (substhvl_TyVar k55 (XS TmVar d35) (HS TmVar k56) (HS TmVar k57))
    | substhvl_TyVar_there_TyVar
        {d35 : (Trace TyVar)}
        {k56 : Hvl} {k57 : Hvl} :
        (substhvl_TyVar k55 d35 k56 k57) -> (substhvl_TyVar k55 (XS TyVar d35) (HS TyVar k56) (HS TyVar k57)).
  Lemma weaken_substhvl_TmVar  :
    (forall {k56 : Hvl} (k55 : Hvl) {d35 : (Trace TmVar)} {k57 : Hvl} {k58 : Hvl} ,
      (substhvl_TmVar k56 d35 k57 k58) -> (substhvl_TmVar k56 (weakenTrace d35 k55) (appendHvl k57 k55) (appendHvl k58 k55))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma weaken_substhvl_TyVar  :
    (forall {k56 : Hvl} (k55 : Hvl) {d35 : (Trace TyVar)} {k57 : Hvl} {k58 : Hvl} ,
      (substhvl_TyVar k56 d35 k57 k58) -> (substhvl_TyVar k56 (weakenTrace d35 k55) (appendHvl k57 k55) (appendHvl k58 k55))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma substhvl_TmVar_wfindex_TmVar {k55 : Hvl} {s25 : Tm} (wft : (wfTm k55 s25)) :
    (forall {d35 : (Trace TmVar)} {k56 : Hvl} {k57 : Hvl} ,
      (substhvl_TmVar k55 d35 k56 k57) -> (forall {x89 : (Index TmVar)} ,
        (wfindex k56 x89) -> (wfTm k57 (subst_TmVar_Index d35 s25 x89)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TyVar_wfindex_TyVar {k55 : Hvl} {S26 : Ty} (wft : (wfTy k55 S26)) :
    (forall {d35 : (Trace TyVar)} {k56 : Hvl} {k57 : Hvl} ,
      (substhvl_TyVar k55 d35 k56 k57) -> (forall {X35 : (Index TyVar)} ,
        (wfindex k56 X35) -> (wfTy k57 (subst_TyVar_Index d35 S26 X35)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TmVar_wfindex_TyVar {k55 : Hvl} :
    (forall {d35 : (Trace TmVar)} {k56 : Hvl} {k57 : Hvl} ,
      (substhvl_TmVar k55 d35 k56 k57) -> (forall {X35 : (Index TyVar)} ,
        (wfindex k56 X35) -> (wfindex k57 X35))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_TyVar_wfindex_TmVar {k55 : Hvl} :
    (forall {d35 : (Trace TyVar)} {k56 : Hvl} {k57 : Hvl} ,
      (substhvl_TyVar k55 d35 k56 k57) -> (forall {x89 : (Index TmVar)} ,
        (wfindex k56 x89) -> (wfindex k57 x89))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Definition substhvl_TmVar_wfTm_Ty {k55 : Hvl} {s25 : Tm} (wf : (wfTm k55 s25)) : (forall (k56 : Hvl) ,
    (forall (s26 : Tm) (wf0 : (wfTm k56 s26)) ,
      (forall {d35 : (Trace TmVar)} {k57 : Hvl} ,
        (substhvl_TmVar k55 d35 k56 k57) -> (wfTm k57 (subst_TmVar_Tm d35 s25 s26)))) /\
    (forall (S26 : Ty) (wf0 : (wfTy k56 S26)) ,
      (forall {d35 : (Trace TmVar)} {k57 : Hvl} ,
        (substhvl_TmVar k55 d35 k56 k57) -> (wfTy k57 (subst_TmVar_Ty d35 s25 S26))))) := (ind_wfTm_Ty (fun (k56 : Hvl) (s26 : Tm) (wf0 : (wfTm k56 s26)) =>
    (forall {d35 : (Trace TmVar)} {k57 : Hvl} ,
      (substhvl_TmVar k55 d35 k56 k57) -> (wfTm k57 (subst_TmVar_Tm d35 s25 s26)))) (fun (k56 : Hvl) (S26 : Ty) (wf0 : (wfTy k56 S26)) =>
    (forall {d35 : (Trace TmVar)} {k57 : Hvl} ,
      (substhvl_TmVar k55 d35 k56 k57) -> (wfTy k57 (subst_TmVar_Ty d35 s25 S26)))) (fun (k56 : Hvl) {x89 : (Index TmVar)} (wfi : (wfindex k56 x89)) {d35 : (Trace TmVar)} {k57 : Hvl} (del : (substhvl_TmVar k55 d35 k56 k57)) =>
    (substhvl_TmVar_wfindex_TmVar wf del wfi)) (fun (k56 : Hvl) (T : Ty) (wf0 : (wfTy k56 T)) IHT89 (t : Tm) (wf1 : (wfTm (HS TmVar k56) t)) IHt71 {d35 : (Trace TmVar)} {k57 : Hvl} (del : (substhvl_TmVar k55 d35 k56 k57)) =>
    (wfTm_Abs k57 (IHT89 (weakenTrace d35 H0) k57 (weaken_substhvl_TmVar H0 del)) (IHt71 (weakenTrace d35 (HS TmVar H0)) (HS TmVar k57) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k56 : Hvl) (t1 : Tm) (wf0 : (wfTm k56 t1)) IHt71 (t2 : Tm) (wf1 : (wfTm k56 t2)) IHt72 {d35 : (Trace TmVar)} {k57 : Hvl} (del : (substhvl_TmVar k55 d35 k56 k57)) =>
    (wfTm_App k57 (IHt71 (weakenTrace d35 H0) k57 (weaken_substhvl_TmVar H0 del)) (IHt72 (weakenTrace d35 H0) k57 (weaken_substhvl_TmVar H0 del)))) (fun (k56 : Hvl) {X35 : (Index TyVar)} (wfi : (wfindex k56 X35)) {d35 : (Trace TmVar)} {k57 : Hvl} (del : (substhvl_TmVar k55 d35 k56 k57)) =>
    (wfTy_TVar k57 (substhvl_TmVar_wfindex_TyVar del wfi))) (fun (k56 : Hvl) (T1 : Ty) (wf0 : (wfTy k56 T1)) IHT89 (T2 : Ty) (wf1 : (wfTy (HS TmVar k56) T2)) IHT90 {d35 : (Trace TmVar)} {k57 : Hvl} (del : (substhvl_TmVar k55 d35 k56 k57)) =>
    (wfTy_TPi k57 (IHT89 (weakenTrace d35 H0) k57 (weaken_substhvl_TmVar H0 del)) (IHT90 (weakenTrace d35 (HS TmVar H0)) (HS TmVar k57) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k56 : Hvl) (T : Ty) (wf0 : (wfTy k56 T)) IHT89 (t : Tm) (wf1 : (wfTm k56 t)) IHt71 {d35 : (Trace TmVar)} {k57 : Hvl} (del : (substhvl_TmVar k55 d35 k56 k57)) =>
    (wfTy_TApp k57 (IHT89 (weakenTrace d35 H0) k57 (weaken_substhvl_TmVar H0 del)) (IHt71 (weakenTrace d35 H0) k57 (weaken_substhvl_TmVar H0 del))))).
  Lemma substhvl_TmVar_wfTm {k55 : Hvl} {s25 : Tm} (wf : (wfTm k55 s25)) (k56 : Hvl) (s26 : Tm) (wf0 : (wfTm k56 s26)) :
    (forall {d35 : (Trace TmVar)} {k57 : Hvl} ,
      (substhvl_TmVar k55 d35 k56 k57) -> (wfTm k57 (subst_TmVar_Tm d35 s25 s26))).
  Proof.
    apply ((substhvl_TmVar_wfTm_Ty wf k56)).
    auto .
  Qed.
  Lemma substhvl_TmVar_wfTy {k55 : Hvl} {s25 : Tm} (wf : (wfTm k55 s25)) (k56 : Hvl) (S26 : Ty) (wf1 : (wfTy k56 S26)) :
    (forall {d36 : (Trace TmVar)} {k58 : Hvl} ,
      (substhvl_TmVar k55 d36 k56 k58) -> (wfTy k58 (subst_TmVar_Ty d36 s25 S26))).
  Proof.
    apply ((substhvl_TmVar_wfTm_Ty wf k56)).
    auto .
  Qed.
  Definition substhvl_TyVar_wfTm_Ty {k55 : Hvl} {S26 : Ty} (wf : (wfTy k55 S26)) : (forall (k56 : Hvl) ,
    (forall (s25 : Tm) (wf0 : (wfTm k56 s25)) ,
      (forall {d35 : (Trace TyVar)} {k57 : Hvl} ,
        (substhvl_TyVar k55 d35 k56 k57) -> (wfTm k57 (subst_TyVar_Tm d35 S26 s25)))) /\
    (forall (S27 : Ty) (wf0 : (wfTy k56 S27)) ,
      (forall {d35 : (Trace TyVar)} {k57 : Hvl} ,
        (substhvl_TyVar k55 d35 k56 k57) -> (wfTy k57 (subst_TyVar_Ty d35 S26 S27))))) := (ind_wfTm_Ty (fun (k56 : Hvl) (s25 : Tm) (wf0 : (wfTm k56 s25)) =>
    (forall {d35 : (Trace TyVar)} {k57 : Hvl} ,
      (substhvl_TyVar k55 d35 k56 k57) -> (wfTm k57 (subst_TyVar_Tm d35 S26 s25)))) (fun (k56 : Hvl) (S27 : Ty) (wf0 : (wfTy k56 S27)) =>
    (forall {d35 : (Trace TyVar)} {k57 : Hvl} ,
      (substhvl_TyVar k55 d35 k56 k57) -> (wfTy k57 (subst_TyVar_Ty d35 S26 S27)))) (fun (k56 : Hvl) {x89 : (Index TmVar)} (wfi : (wfindex k56 x89)) {d35 : (Trace TyVar)} {k57 : Hvl} (del : (substhvl_TyVar k55 d35 k56 k57)) =>
    (wfTm_Var k57 (substhvl_TyVar_wfindex_TmVar del wfi))) (fun (k56 : Hvl) (T : Ty) (wf0 : (wfTy k56 T)) IHT89 (t : Tm) (wf1 : (wfTm (HS TmVar k56) t)) IHt71 {d35 : (Trace TyVar)} {k57 : Hvl} (del : (substhvl_TyVar k55 d35 k56 k57)) =>
    (wfTm_Abs k57 (IHT89 (weakenTrace d35 H0) k57 (weaken_substhvl_TyVar H0 del)) (IHt71 (weakenTrace d35 (HS TmVar H0)) (HS TmVar k57) (weaken_substhvl_TyVar (HS TmVar H0) del)))) (fun (k56 : Hvl) (t1 : Tm) (wf0 : (wfTm k56 t1)) IHt71 (t2 : Tm) (wf1 : (wfTm k56 t2)) IHt72 {d35 : (Trace TyVar)} {k57 : Hvl} (del : (substhvl_TyVar k55 d35 k56 k57)) =>
    (wfTm_App k57 (IHt71 (weakenTrace d35 H0) k57 (weaken_substhvl_TyVar H0 del)) (IHt72 (weakenTrace d35 H0) k57 (weaken_substhvl_TyVar H0 del)))) (fun (k56 : Hvl) {X35 : (Index TyVar)} (wfi : (wfindex k56 X35)) {d35 : (Trace TyVar)} {k57 : Hvl} (del : (substhvl_TyVar k55 d35 k56 k57)) =>
    (substhvl_TyVar_wfindex_TyVar wf del wfi)) (fun (k56 : Hvl) (T1 : Ty) (wf0 : (wfTy k56 T1)) IHT89 (T2 : Ty) (wf1 : (wfTy (HS TmVar k56) T2)) IHT90 {d35 : (Trace TyVar)} {k57 : Hvl} (del : (substhvl_TyVar k55 d35 k56 k57)) =>
    (wfTy_TPi k57 (IHT89 (weakenTrace d35 H0) k57 (weaken_substhvl_TyVar H0 del)) (IHT90 (weakenTrace d35 (HS TmVar H0)) (HS TmVar k57) (weaken_substhvl_TyVar (HS TmVar H0) del)))) (fun (k56 : Hvl) (T : Ty) (wf0 : (wfTy k56 T)) IHT89 (t : Tm) (wf1 : (wfTm k56 t)) IHt71 {d35 : (Trace TyVar)} {k57 : Hvl} (del : (substhvl_TyVar k55 d35 k56 k57)) =>
    (wfTy_TApp k57 (IHT89 (weakenTrace d35 H0) k57 (weaken_substhvl_TyVar H0 del)) (IHt71 (weakenTrace d35 H0) k57 (weaken_substhvl_TyVar H0 del))))).
  Lemma substhvl_TyVar_wfTm {k55 : Hvl} {S26 : Ty} (wf : (wfTy k55 S26)) (k56 : Hvl) (s25 : Tm) (wf0 : (wfTm k56 s25)) :
    (forall {d35 : (Trace TyVar)} {k57 : Hvl} ,
      (substhvl_TyVar k55 d35 k56 k57) -> (wfTm k57 (subst_TyVar_Tm d35 S26 s25))).
  Proof.
    apply ((substhvl_TyVar_wfTm_Ty wf k56)).
    auto .
  Qed.
  Lemma substhvl_TyVar_wfTy {k55 : Hvl} {S26 : Ty} (wf : (wfTy k55 S26)) (k56 : Hvl) (S27 : Ty) (wf1 : (wfTy k56 S27)) :
    (forall {d36 : (Trace TyVar)} {k58 : Hvl} ,
      (substhvl_TyVar k55 d36 k56 k58) -> (wfTy k58 (subst_TyVar_Ty d36 S26 S27))).
  Proof.
    apply ((substhvl_TyVar_wfTm_Ty wf k56)).
    auto .
  Qed.
  Definition substhvl_TmVar_wfKind {k55 : Hvl} {s25 : Tm} (wf : (wfTm k55 s25)) : (forall (k56 : Hvl) ,
    (forall (K17 : Kind) (wf0 : (wfKind k56 K17)) ,
      (forall {d35 : (Trace TmVar)} {k57 : Hvl} ,
        (substhvl_TmVar k55 d35 k56 k57) -> (wfKind k57 (subst_TmVar_Kind d35 s25 K17))))) := (ind_wfKind (fun (k56 : Hvl) (K17 : Kind) (wf0 : (wfKind k56 K17)) =>
    (forall {d35 : (Trace TmVar)} {k57 : Hvl} ,
      (substhvl_TmVar k55 d35 k56 k57) -> (wfKind k57 (subst_TmVar_Kind d35 s25 K17)))) (fun (k56 : Hvl) {d35 : (Trace TmVar)} {k57 : Hvl} (del : (substhvl_TmVar k55 d35 k56 k57)) =>
    (wfKind_Star k57)) (fun (k56 : Hvl) (T : Ty) (wf0 : (wfTy k56 T)) (K : Kind) (wf1 : (wfKind (HS TmVar k56) K)) IHK17 {d35 : (Trace TmVar)} {k57 : Hvl} (del : (substhvl_TmVar k55 d35 k56 k57)) =>
    (wfKind_KPi k57 (substhvl_TmVar_wfTy wf k56 T wf0 (weaken_substhvl_TmVar H0 del)) (IHK17 (weakenTrace d35 (HS TmVar H0)) (HS TmVar k57) (weaken_substhvl_TmVar (HS TmVar H0) del))))).
  Definition substhvl_TyVar_wfKind {k55 : Hvl} {S26 : Ty} (wf : (wfTy k55 S26)) : (forall (k56 : Hvl) ,
    (forall (K17 : Kind) (wf0 : (wfKind k56 K17)) ,
      (forall {d35 : (Trace TyVar)} {k57 : Hvl} ,
        (substhvl_TyVar k55 d35 k56 k57) -> (wfKind k57 (subst_TyVar_Kind d35 S26 K17))))) := (ind_wfKind (fun (k56 : Hvl) (K17 : Kind) (wf0 : (wfKind k56 K17)) =>
    (forall {d35 : (Trace TyVar)} {k57 : Hvl} ,
      (substhvl_TyVar k55 d35 k56 k57) -> (wfKind k57 (subst_TyVar_Kind d35 S26 K17)))) (fun (k56 : Hvl) {d35 : (Trace TyVar)} {k57 : Hvl} (del : (substhvl_TyVar k55 d35 k56 k57)) =>
    (wfKind_Star k57)) (fun (k56 : Hvl) (T : Ty) (wf0 : (wfTy k56 T)) (K : Kind) (wf1 : (wfKind (HS TmVar k56) K)) IHK17 {d35 : (Trace TyVar)} {k57 : Hvl} (del : (substhvl_TyVar k55 d35 k56 k57)) =>
    (wfKind_KPi k57 (substhvl_TyVar_wfTy wf k56 T wf0 (weaken_substhvl_TyVar H0 del)) (IHK17 (weakenTrace d35 (HS TmVar H0)) (HS TmVar k57) (weaken_substhvl_TyVar (HS TmVar H0) del))))).
End SubstWellFormed.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : infra.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst_wf.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : wf.
 Hint Resolve substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : infra.
 Hint Resolve substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst.
 Hint Resolve substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst_wf.
 Hint Resolve substhvl_TmVar_wfKind substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfKind substhvl_TyVar_wfTm substhvl_TyVar_wfTy : wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : infra.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst_wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : wf.
Section Context.
  Inductive Ctx : Type :=
    | empty 
    | evar (G : Ctx) (T : Ty)
    | etvar (G : Ctx) (K : Kind).
  Fixpoint appendCtx (G : Ctx) (G0 : Ctx) : Ctx :=
    match G0 with
      | (empty) => G
      | (evar G1 T) => (evar (appendCtx G G1) T)
      | (etvar G1 K) => (etvar (appendCtx G G1) K)
    end.
  Fixpoint domainCtx (G : Ctx) : Hvl :=
    match G with
      | (empty) => H0
      | (evar G0 T) => (HS TmVar (domainCtx G0))
      | (etvar G0 K) => (HS TyVar (domainCtx G0))
    end.
  Lemma appendCtx_assoc  :
    (forall (G : Ctx) (G0 : Ctx) (G1 : Ctx) ,
      ((appendCtx G (appendCtx G0 G1)) =
      (appendCtx (appendCtx G G0) G1))).
  Proof.
    needleGenericAppendEnvAssoc .
  Qed.
  Lemma domainCtx_appendCtx  :
    (forall (G : Ctx) (G0 : Ctx) ,
      ((domainCtx (appendCtx G G0)) =
      (appendHvl (domainCtx G) (domainCtx G0)))).
  Proof.
    needleGenericDomainEnvAppendEnv .
  Qed.
  Fixpoint shift_TmVar_Ctx (c19 : (Cutoff TmVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shift_TmVar_Ctx c19 G0) (shift_TmVar_Ty (weakenCutoffTmVar c19 (domainCtx G0)) T))
      | (etvar G0 K) => (etvar (shift_TmVar_Ctx c19 G0) (shift_TmVar_Kind (weakenCutoffTmVar c19 (domainCtx G0)) K))
    end.
  Fixpoint shift_TyVar_Ctx (c19 : (Cutoff TyVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shift_TyVar_Ctx c19 G0) (shift_TyVar_Ty (weakenCutoffTyVar c19 (domainCtx G0)) T))
      | (etvar G0 K) => (etvar (shift_TyVar_Ctx c19 G0) (shift_TyVar_Kind (weakenCutoffTyVar c19 (domainCtx G0)) K))
    end.
  Fixpoint subst_TmVar_Ctx (d35 : (Trace TmVar)) (s25 : Tm) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (subst_TmVar_Ctx d35 s25 G0) (subst_TmVar_Ty (weakenTrace d35 (domainCtx G0)) s25 T))
      | (etvar G0 K) => (etvar (subst_TmVar_Ctx d35 s25 G0) (subst_TmVar_Kind (weakenTrace d35 (domainCtx G0)) s25 K))
    end.
  Fixpoint subst_TyVar_Ctx (d35 : (Trace TyVar)) (S26 : Ty) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (subst_TyVar_Ctx d35 S26 G0) (subst_TyVar_Ty (weakenTrace d35 (domainCtx G0)) S26 T))
      | (etvar G0 K) => (etvar (subst_TyVar_Ctx d35 S26 G0) (subst_TyVar_Kind (weakenTrace d35 (domainCtx G0)) S26 K))
    end.
  Lemma domainCtx_shift_TmVar_Ctx  :
    (forall (c19 : (Cutoff TmVar)) (G : Ctx) ,
      ((domainCtx (shift_TmVar_Ctx c19 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_shift_TyVar_Ctx  :
    (forall (c19 : (Cutoff TyVar)) (G : Ctx) ,
      ((domainCtx (shift_TyVar_Ctx c19 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_subst_TmVar_Ctx  :
    (forall (d35 : (Trace TmVar)) (s25 : Tm) (G : Ctx) ,
      ((domainCtx (subst_TmVar_Ctx d35 s25 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
  Lemma domainCtx_subst_TyVar_Ctx  :
    (forall (d35 : (Trace TyVar)) (S26 : Ty) (G : Ctx) ,
      ((domainCtx (subst_TyVar_Ctx d35 S26 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
End Context.
 Hint Rewrite domainCtx_appendCtx : interaction.
 Hint Rewrite domainCtx_appendCtx : env_domain_append.
Section ContextStuff.
  Section ShiftEnvAppendEnv.
    Lemma shift_TmVar_Ctx_appendCtx  :
      (forall (c19 : (Cutoff TmVar)) (G : Ctx) (G0 : Ctx) ,
        ((shift_TmVar_Ctx c19 (appendCtx G G0)) =
        (appendCtx (shift_TmVar_Ctx c19 G) (shift_TmVar_Ctx (weakenCutoffTmVar c19 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
    Lemma shift_TyVar_Ctx_appendCtx  :
      (forall (c19 : (Cutoff TyVar)) (G : Ctx) (G0 : Ctx) ,
        ((shift_TyVar_Ctx c19 (appendCtx G G0)) =
        (appendCtx (shift_TyVar_Ctx c19 G) (shift_TyVar_Ctx (weakenCutoffTyVar c19 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
  End ShiftEnvAppendEnv.
  Section SubstEnvAppendEnv.
    Lemma subst_TmVar_Ctx_appendCtx  :
      (forall (d35 : (Trace TmVar)) (s25 : Tm) (G : Ctx) (G0 : Ctx) ,
        ((subst_TmVar_Ctx d35 s25 (appendCtx G G0)) =
        (appendCtx (subst_TmVar_Ctx d35 s25 G) (subst_TmVar_Ctx (weakenTrace d35 (domainCtx G)) s25 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
    Lemma subst_TyVar_Ctx_appendCtx  :
      (forall (d35 : (Trace TyVar)) (S26 : Ty) (G : Ctx) (G0 : Ctx) ,
        ((subst_TyVar_Ctx d35 S26 (appendCtx G G0)) =
        (appendCtx (subst_TyVar_Ctx d35 S26 G) (subst_TyVar_Ctx (weakenTrace d35 (domainCtx G)) S26 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
  End SubstEnvAppendEnv.
  Section Lookups.
    Inductive lookup_evar : Ctx -> (Index TmVar) -> Ty -> Prop :=
      | lookup_evar_here {G : Ctx}
          {T89 : Ty} :
          (wfTy (domainCtx G) T89) -> (lookup_evar (evar G T89) I0 (shift_TmVar_Ty C0 T89))
      | lookup_evar_there_evar
          {G : Ctx} {x89 : (Index TmVar)}
          {T89 : Ty} {T90 : Ty} :
          (lookup_evar G x89 T89) -> (lookup_evar (evar G T90) (IS x89) (shift_TmVar_Ty C0 T89))
      | lookup_evar_there_etvar
          {G : Ctx} {x89 : (Index TmVar)}
          {T89 : Ty} {K17 : Kind} :
          (lookup_evar G x89 T89) -> (lookup_evar (etvar G K17) x89 (shift_TyVar_Ty C0 T89)).
    Inductive lookup_etvar : Ctx -> (Index TyVar) -> Kind -> Prop :=
      | lookup_etvar_here {G : Ctx}
          {K17 : Kind} :
          (wfKind (domainCtx G) K17) -> (lookup_etvar (etvar G K17) I0 (shift_TyVar_Kind C0 K17))
      | lookup_etvar_there_evar
          {G : Ctx} {X35 : (Index TyVar)}
          {K17 : Kind} {T89 : Ty} :
          (lookup_etvar G X35 K17) -> (lookup_etvar (evar G T89) X35 (shift_TmVar_Kind C0 K17))
      | lookup_etvar_there_etvar
          {G : Ctx} {X35 : (Index TyVar)}
          {K17 : Kind} {K18 : Kind} :
          (lookup_etvar G X35 K17) -> (lookup_etvar (etvar G K18) (IS X35) (shift_TyVar_Kind C0 K17)).
    Lemma lookup_evar_inversion_here  :
      (forall (G : Ctx) (S26 : Ty) (S27 : Ty) ,
        (lookup_evar (evar G S26) I0 S27) -> ((shift_TmVar_Ty C0 S26) =
        S27)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_etvar_inversion_here  :
      (forall (G : Ctx) (K17 : Kind) (K18 : Kind) ,
        (lookup_etvar (etvar G K17) I0 K18) -> ((shift_TyVar_Kind C0 K17) =
        K18)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_evar_functional  :
      (forall {G : Ctx} {x89 : (Index TmVar)} ,
        (forall {S26 : Ty} ,
          (lookup_evar G x89 S26) -> (forall {S27 : Ty} ,
            (lookup_evar G x89 S27) -> (S26 =
            S27)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_etvar_functional  :
      (forall {G : Ctx} {X35 : (Index TyVar)} ,
        (forall {K17 : Kind} ,
          (lookup_etvar G X35 K17) -> (forall {K18 : Kind} ,
            (lookup_etvar G X35 K18) -> (K17 =
            K18)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_evar_wf  :
      (forall {G : Ctx} {x89 : (Index TmVar)} {S26 : Ty} ,
        (lookup_evar G x89 S26) -> (wfTy (domainCtx G) S26)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma lookup_etvar_wf  :
      (forall {G : Ctx} {X35 : (Index TyVar)} {K17 : Kind} ,
        (lookup_etvar G X35 K17) -> (wfKind (domainCtx G) K17)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma weaken_lookup_evar  :
      (forall {G : Ctx} (G0 : Ctx) {x89 : (Index TmVar)} {T89 : Ty} ,
        (lookup_evar G x89 T89) -> (lookup_evar (appendCtx G G0) (weakenIndexTmVar x89 (domainCtx G0)) (weakenTy T89 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma weaken_lookup_etvar  :
      (forall {G : Ctx} (G0 : Ctx) {X35 : (Index TyVar)} {K17 : Kind} ,
        (lookup_etvar G X35 K17) -> (lookup_etvar (appendCtx G G0) (weakenIndexTyVar X35 (domainCtx G0)) (weakenKind K17 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma lookup_evar_wfindex  :
      (forall {G : Ctx} {x89 : (Index TmVar)} {S26 : Ty} ,
        (lookup_evar G x89 S26) -> (wfindex (domainCtx G) x89)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
    Lemma lookup_etvar_wfindex  :
      (forall {G : Ctx} {X35 : (Index TyVar)} {K18 : Kind} ,
        (lookup_etvar G X35 K18) -> (wfindex (domainCtx G) X35)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
  End Lookups.
  Inductive shift_evar : (Cutoff TmVar) -> Ctx -> Ctx -> Prop :=
    | shift_evar_here {G : Ctx}
        {T89 : Ty} :
        (shift_evar C0 G (evar G T89))
    | shiftevar_there_evar
        {c19 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_evar c19 G G0) -> (shift_evar (CS c19) (evar G T) (evar G0 (shift_TmVar_Ty c19 T)))
    | shiftevar_there_etvar
        {c19 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} {K : Kind} :
        (shift_evar c19 G G0) -> (shift_evar c19 (etvar G K) (etvar G0 (shift_TmVar_Kind c19 K))).
  Lemma weaken_shift_evar  :
    (forall (G : Ctx) {c19 : (Cutoff TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_evar c19 G0 G1) -> (shift_evar (weakenCutoffTmVar c19 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 (shift_TmVar_Ctx c19 G)))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Inductive shift_etvar : (Cutoff TyVar) -> Ctx -> Ctx -> Prop :=
    | shift_etvar_here {G : Ctx}
        {K17 : Kind} :
        (shift_etvar C0 G (etvar G K17))
    | shiftetvar_there_evar
        {c19 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_etvar c19 G G0) -> (shift_etvar c19 (evar G T) (evar G0 (shift_TyVar_Ty c19 T)))
    | shiftetvar_there_etvar
        {c19 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} {K : Kind} :
        (shift_etvar c19 G G0) -> (shift_etvar (CS c19) (etvar G K) (etvar G0 (shift_TyVar_Kind c19 K))).
  Lemma weaken_shift_etvar  :
    (forall (G : Ctx) {c19 : (Cutoff TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_etvar c19 G0 G1) -> (shift_etvar (weakenCutoffTyVar c19 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 (shift_TyVar_Ctx c19 G)))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Lemma shift_evar_shifthvl_TmVar  :
    (forall {c19 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_evar c19 G G0) -> (shifthvl_TmVar c19 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Lemma shift_etvar_shifthvl_TyVar  :
    (forall {c19 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_etvar c19 G G0) -> (shifthvl_TyVar c19 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Inductive subst_evar (G : Ctx) (T89 : Ty) (s25 : Tm) : (Trace TmVar) -> Ctx -> Ctx -> Prop :=
    | subst_evar_here :
        (subst_evar G T89 s25 X0 (evar G T89) G)
    | subst_evar_there_evar
        {d35 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_evar G T89 s25 d35 G0 G1) -> (subst_evar G T89 s25 (XS TmVar d35) (evar G0 T) (evar G1 (subst_TmVar_Ty d35 s25 T)))
    | subst_evar_there_etvar
        {d35 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} {K : Kind} :
        (subst_evar G T89 s25 d35 G0 G1) -> (subst_evar G T89 s25 (XS TyVar d35) (etvar G0 K) (etvar G1 (subst_TmVar_Kind d35 s25 K))).
  Lemma weaken_subst_evar {G : Ctx} {T89 : Ty} {s25 : Tm} :
    (forall (G0 : Ctx) {d35 : (Trace TmVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_evar G T89 s25 d35 G1 G2) -> (subst_evar G T89 s25 (weakenTrace d35 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 (subst_TmVar_Ctx d35 s25 G0)))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Inductive subst_etvar (G : Ctx) (K17 : Kind) (S26 : Ty) : (Trace TyVar) -> Ctx -> Ctx -> Prop :=
    | subst_etvar_here :
        (subst_etvar G K17 S26 X0 (etvar G K17) G)
    | subst_etvar_there_evar
        {d35 : (Trace TyVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_etvar G K17 S26 d35 G0 G1) -> (subst_etvar G K17 S26 (XS TmVar d35) (evar G0 T) (evar G1 (subst_TyVar_Ty d35 S26 T)))
    | subst_etvar_there_etvar
        {d35 : (Trace TyVar)} {G0 : Ctx}
        {G1 : Ctx} {K : Kind} :
        (subst_etvar G K17 S26 d35 G0 G1) -> (subst_etvar G K17 S26 (XS TyVar d35) (etvar G0 K) (etvar G1 (subst_TyVar_Kind d35 S26 K))).
  Lemma weaken_subst_etvar {G : Ctx} {K17 : Kind} {S26 : Ty} :
    (forall (G0 : Ctx) {d35 : (Trace TyVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_etvar G K17 S26 d35 G1 G2) -> (subst_etvar G K17 S26 (weakenTrace d35 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 (subst_TyVar_Ctx d35 S26 G0)))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Lemma subst_evar_substhvl_TmVar {G : Ctx} {s25 : Tm} {T89 : Ty} :
    (forall {d35 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_evar G T89 s25 d35 G0 G1) -> (substhvl_TmVar (domainCtx G) d35 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
  Lemma subst_etvar_substhvl_TyVar {G : Ctx} {S26 : Ty} {K17 : Kind} :
    (forall {d35 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_etvar G K17 S26 d35 G0 G1) -> (substhvl_TyVar (domainCtx G) d35 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
End ContextStuff.
 Hint Rewrite domainCtx_shift_TmVar_Ctx domainCtx_shift_TyVar_Ctx : interaction.
 Hint Rewrite domainCtx_shift_TmVar_Ctx domainCtx_shift_TyVar_Ctx : env_domain_shift.
 Hint Rewrite domainCtx_subst_TmVar_Ctx domainCtx_subst_TyVar_Ctx : interaction.
 Hint Rewrite domainCtx_subst_TmVar_Ctx domainCtx_subst_TyVar_Ctx : env_domain_subst.
 Hint Rewrite shift_TmVar_Ctx_appendCtx shift_TyVar_Ctx_appendCtx : interaction.
 Hint Rewrite shift_TmVar_Ctx_appendCtx shift_TyVar_Ctx_appendCtx : env_shift_append.
 Hint Rewrite subst_TmVar_Ctx_appendCtx subst_TyVar_Ctx_appendCtx : interaction.
 Hint Rewrite subst_TmVar_Ctx_appendCtx subst_TyVar_Ctx_appendCtx : env_shift_append.
 Hint Constructors lookup_evar lookup_etvar : infra.
 Hint Constructors lookup_evar lookup_etvar : lookup.
 Hint Constructors lookup_evar lookup_etvar : shift.
 Hint Constructors lookup_evar lookup_etvar : subst.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : infra.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : lookup.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : shift.
 Hint Resolve weaken_lookup_evar weaken_lookup_etvar : weaken.
Lemma weaken_lookup_evar_here  :
  (forall {G : Ctx} (G0 : Ctx) {T89 : Ty} (wf : (wfTy (domainCtx G) T89)) ,
    (lookup_evar (appendCtx (evar G T89) G0) (weakenIndexTmVar I0 (domainCtx G0)) (weakenTy (shift_TmVar_Ty C0 T89) (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
Lemma weaken_lookup_etvar_here  :
  (forall {G : Ctx} (G0 : Ctx) {K17 : Kind} (wf : (wfKind (domainCtx G) K17)) ,
    (lookup_etvar (appendCtx (etvar G K17) G0) (weakenIndexTyVar I0 (domainCtx G0)) (weakenKind (shift_TyVar_Kind C0 K17) (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
 Hint Constructors wfKind wfTm wfTy : infra.
 Hint Constructors wfKind wfTm wfTy : wf.
 Hint Extern 10 ((wfKind _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTm _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTy _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 2 ((wfTm _ _)) => match goal with
  | H : (wfTm _ (Var _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Abs _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (App _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfTy _ _)) => match goal with
  | H : (wfTy _ (TVar _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TPi _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TApp _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfKind _ _)) => match goal with
  | H : (wfKind _ (Star)) |- _ => inversion H; subst; clear H
  | H : (wfKind _ (KPi _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Resolve lookup_evar_wf lookup_etvar_wf : infra.
 Hint Resolve lookup_evar_wf lookup_etvar_wf : wf.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : infra.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : lookup.
 Hint Resolve lookup_evar_wfindex lookup_etvar_wfindex : wf.
 Hint Constructors shift_evar shift_etvar : infra.
 Hint Constructors shift_evar shift_etvar : shift.
 Hint Constructors shift_evar shift_etvar : subst.
 Hint Resolve weaken_shift_evar weaken_shift_etvar : infra.
 Hint Resolve weaken_shift_evar weaken_shift_etvar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : infra.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : shift_wf.
 Hint Resolve shift_evar_shifthvl_TmVar shift_etvar_shifthvl_TyVar : wf.
 Hint Constructors subst_evar subst_etvar : infra.
 Hint Constructors subst_evar subst_etvar : subst.
 Hint Resolve weaken_subst_evar weaken_subst_etvar : infra.
 Hint Resolve weaken_subst_evar weaken_subst_etvar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : infra.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : subst_wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_etvar_substhvl_TyVar : substenv_substhvl.
Lemma shift_evar_lookup_evar  :
  (forall {c19 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c19 G G0)) {x89 : (Index TmVar)} {T89 : Ty} ,
    (lookup_evar G x89 T89) -> (lookup_evar G0 (shift_TmVar_Index c19 x89) (shift_TmVar_Ty c19 T89))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_evar_lookup_etvar  :
  (forall {c19 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c19 G G0)) {X35 : (Index TyVar)} {K17 : Kind} ,
    (lookup_etvar G X35 K17) -> (lookup_etvar G0 X35 (shift_TmVar_Kind c19 K17))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_evar  :
  (forall {c19 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c19 G G0)) {x89 : (Index TmVar)} {T89 : Ty} ,
    (lookup_evar G x89 T89) -> (lookup_evar G0 x89 (shift_TyVar_Ty c19 T89))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_etvar  :
  (forall {c19 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c19 G G0)) {X35 : (Index TyVar)} {K17 : Kind} ,
    (lookup_etvar G X35 K17) -> (lookup_etvar G0 (shift_TyVar_Index c19 X35) (shift_TyVar_Kind c19 K17))).
Proof.
  needleGenericInsertLookup .
Qed.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : infra.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : lookup.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_etvar : shift.
Lemma subst_evar_lookup_etvar {G : Ctx} {T90 : Ty} {s25 : Tm} (wf : (wfTm (domainCtx G) s25)) :
  (forall {d35 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_evar G T90 s25 d35 G0 G1)) {X35 : (Index TyVar)} {K18 : Kind} ,
    (lookup_etvar G0 X35 K18) -> (lookup_etvar G1 X35 (subst_TmVar_Kind d35 s25 K18))).
Proof.
  needleGenericSubstEnvLookup .
Qed.
Lemma subst_etvar_lookup_evar {G : Ctx} {K18 : Kind} {S26 : Ty} (wf : (wfTy (domainCtx G) S26)) :
  (forall {d35 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_etvar G K18 S26 d35 G0 G1)) {x89 : (Index TmVar)} {T90 : Ty} ,
    (lookup_evar G0 x89 T90) -> (lookup_evar G1 x89 (subst_TyVar_Ty d35 S26 T90))).
Proof.
  needleGenericSubstEnvLookup .
Qed.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : infra.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : lookup.
 Hint Resolve subst_evar_lookup_etvar subst_etvar_lookup_evar : subst.
Fixpoint size_Tm (s : Tm) {struct s} : nat :=
  match s with
    | (Var x) => 1
    | (Abs T t) => (plus 1 (plus (size_Ty T) (size_Tm t)))
    | (App t1 t2) => (plus 1 (plus (size_Tm t1) (size_Tm t2)))
  end
with size_Ty (S0 : Ty) {struct S0} : nat :=
  match S0 with
    | (TVar X) => 1
    | (TPi T1 T2) => (plus 1 (plus (size_Ty T1) (size_Ty T2)))
    | (TApp T t) => (plus 1 (plus (size_Ty T) (size_Tm t)))
  end.
Fixpoint size_Kind (K : Kind) {struct K} : nat :=
  match K with
    | (Star) => 1
    | (KPi T K0) => (plus 1 (plus (size_Ty T) (size_Kind K0)))
  end.
Lemma shift_TmVar__size_Tm_Ty  :
  (forall (s25 : Tm) (c15 : (Cutoff TmVar)) ,
    ((size_Tm (shift_TmVar_Tm c15 s25)) =
    (size_Tm s25))) /\
  (forall (S26 : Ty) (c15 : (Cutoff TmVar)) ,
    ((size_Ty (shift_TmVar_Ty c15 S26)) =
    (size_Ty S26))).
Proof.
  apply_mutual_ind ind_Tm_Ty.
  - intros x80 c15.
  reflexivity.
  - intros T79 IHT79 t63 IHt63.
  intros c16; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT79 c16)).
  + exact ((IHt63 (CS c16))).
  - intros t64 IHt64 t65 IHt65.
  intros c16; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt64 c16)).
  + exact ((IHt65 c16)).
  - intros x82 c16.
  reflexivity.
  - intros T80 IHT80 T81 IHT81.
  intros c17; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT80 c17)).
  + exact ((IHT81 (CS c17))).
  - intros T82 IHT82 t66 IHt66.
  intros c17; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT82 c17)).
  + exact ((IHt66 c17)).
Qed.

Lemma shift_TmVar__size_Tm  :
  (forall (s25 : Tm) (c17 : (Cutoff TmVar)) ,
    ((size_Tm (shift_TmVar_Tm c17 s25)) =
    (size_Tm s25))).
Proof.
  pose proof (shift_TmVar__size_Tm_Ty).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TmVar__size_Ty  :
  (forall (S26 : Ty) (c17 : (Cutoff TmVar)) ,
    ((size_Ty (shift_TmVar_Ty c17 S26)) =
    (size_Ty S26))).
Proof.
  pose proof (shift_TmVar__size_Tm_Ty).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TyVar__size_Tm_Ty  :
  (forall (s25 : Tm) (c17 : (Cutoff TyVar)) ,
    ((size_Tm (shift_TyVar_Tm c17 s25)) =
    (size_Tm s25))) /\
  (forall (S26 : Ty) (c17 : (Cutoff TyVar)) ,
    ((size_Ty (shift_TyVar_Ty c17 S26)) =
    (size_Ty S26))).
Proof.
  apply_mutual_ind ind_Tm_Ty.
  - intros X32 c17.
  reflexivity.
  - intros T83 IHT83 t67 IHt67.
  intros c18; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT83 c18)).
  + exact ((IHt67 c18)).
  - intros t68 IHt68 t69 IHt69.
  intros c18; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt68 c18)).
  + exact ((IHt69 c18)).
  - intros X34 c18.
  reflexivity.
  - intros T84 IHT84 T85 IHT85.
  intros c19; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT84 c19)).
  + exact ((IHT85 c19)).
  - intros T86 IHT86 t70 IHt70.
  intros c19; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT86 c19)).
  + exact ((IHt70 c19)).
Qed.

Lemma shift_TyVar__size_Tm  :
  (forall (s25 : Tm) (c19 : (Cutoff TyVar)) ,
    ((size_Tm (shift_TyVar_Tm c19 s25)) =
    (size_Tm s25))).
Proof.
  pose proof (shift_TyVar__size_Tm_Ty).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TyVar__size_Ty  :
  (forall (S26 : Ty) (c19 : (Cutoff TyVar)) ,
    ((size_Ty (shift_TyVar_Ty c19 S26)) =
    (size_Ty S26))).
Proof.
  pose proof (shift_TyVar__size_Tm_Ty).
  destruct_conjs .
  easy .
Qed.
Lemma shift_TmVar__size_Kind  :
  (forall (K15 : Kind) (c19 : (Cutoff TmVar)) ,
    ((size_Kind (shift_TmVar_Kind c19 K15)) =
    (size_Kind K15))).
Proof.
  apply_mutual_ind ind_Kind.
  - intros c19; simpl; apply ((f_equal S)); reflexivity.
  - intros T87 K15 IHK15.
  pose proof ((shift_TmVar__size_Ty T87)) as IHT87.
  intros c19; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT87 c19)).
  + exact ((IHK15 (CS c19))).
Qed.

Lemma shift_TyVar__size_Kind  :
  (forall (K16 : Kind) (c19 : (Cutoff TyVar)) ,
    ((size_Kind (shift_TyVar_Kind c19 K16)) =
    (size_Kind K16))).
Proof.
  apply_mutual_ind ind_Kind.
  - intros c19; simpl; apply ((f_equal S)); reflexivity.
  - intros T88 K16 IHK16.
  pose proof ((shift_TyVar__size_Ty T88)) as IHT88.
  intros c19; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT88 c19)).
  + exact ((IHK16 c19)).
Qed.

 Hint Rewrite shift_TmVar__size_Kind shift_TyVar__size_Kind shift_TmVar__size_Tm shift_TyVar__size_Tm shift_TmVar__size_Ty shift_TyVar__size_Ty : interaction.
 Hint Rewrite shift_TmVar__size_Kind shift_TyVar__size_Kind shift_TmVar__size_Tm shift_TyVar__size_Tm shift_TmVar__size_Ty shift_TyVar__size_Ty : shift_size.
Lemma weaken_size_Kind  :
  (forall (k55 : Hvl) (K17 : Kind) ,
    ((size_Kind (weakenKind K17 k55)) =
    (size_Kind K17))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Tm  :
  (forall (k55 : Hvl) (s25 : Tm) ,
    ((size_Tm (weakenTm s25 k55)) =
    (size_Tm s25))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Ty  :
  (forall (k55 : Hvl) (S26 : Ty) ,
    ((size_Ty (weakenTy S26 k55)) =
    (size_Ty S26))).
Proof.
  needleGenericWeakenSize .
Qed.
 Hint Rewrite weaken_size_Kind weaken_size_Tm weaken_size_Ty : interaction.
 Hint Rewrite weaken_size_Kind weaken_size_Tm weaken_size_Ty : weaken_size.