Require Import Coq.Program.Equality.
Require Import Coq.Program.Tactics.
Require Export Needle.
Section Namespace.
  Inductive Namespace : Type :=
    | TmVar 
    | TxVar 
    | TyVar .
  Lemma eq_namespace_dec (a : Namespace) (b : Namespace) :
    (sumbool (a =
    b) (not (a =
    b))).
  Proof.
    decide equality .
  Defined.
End Namespace.

Section HVarlist.
  Inductive Hvl : Type :=
    | H0 
    | HS (a : Namespace) (k : Hvl).
  
  Fixpoint appendHvl (k : Hvl) (k0 : Hvl) {struct k0} : Hvl :=
    match k0 with
      | (H0) => k
      | (HS a k0) => (HS a (appendHvl k k0))
    end.
  
  Lemma appendHvl_assoc  :
    (forall (k : Hvl) (k0 : Hvl) (k1 : Hvl) ,
      ((appendHvl (appendHvl k k0) k1) =
      (appendHvl k (appendHvl k0 k1)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End HVarlist.

Section Index.
  Inductive Index (a : Namespace) : Type :=
    | I0 
    | IS : (Index a) -> (Index a).
  
  Global Arguments I0 [a] .
  Global Arguments IS [a] _ .
  
  Lemma eq_index_dec {a : Namespace} (i : (Index a)) (j : (Index a)) :
    (sumbool (i =
    j) (not (i =
    j))).
  Proof.
    decide equality .
  Qed.
  
  Fixpoint weakenIndexTmVar (x : (Index TmVar)) (k : Hvl) {struct k} : (Index TmVar) :=
    match k with
      | (H0) => x
      | (HS a k) => match a with
        | (TmVar) => (IS (weakenIndexTmVar x k))
        | _ => (weakenIndexTmVar x k)
      end
    end.
  Fixpoint weakenIndexTxVar (a0 : (Index TxVar)) (k : Hvl) {struct k} : (Index TxVar) :=
    match k with
      | (H0) => a0
      | (HS a k) => match a with
        | (TxVar) => (IS (weakenIndexTxVar a0 k))
        | _ => (weakenIndexTxVar a0 k)
      end
    end.
  Fixpoint weakenIndexTyVar (X : (Index TyVar)) (k : Hvl) {struct k} : (Index TyVar) :=
    match k with
      | (H0) => X
      | (HS a k) => match a with
        | (TyVar) => (IS (weakenIndexTyVar X k))
        | _ => (weakenIndexTyVar X k)
      end
    end.
  Lemma weakenIndexTmVar_append  :
    (forall (x : (Index TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTmVar (weakenIndexTmVar x k) k0) =
      (weakenIndexTmVar x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenIndexTxVar_append  :
    (forall (a : (Index TxVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTxVar (weakenIndexTxVar a k) k0) =
      (weakenIndexTxVar a (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenIndexTyVar_append  :
    (forall (X : (Index TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTyVar (weakenIndexTyVar X k) k0) =
      (weakenIndexTyVar X (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End Index.

Section Terms.
  Inductive Ty : Set :=
    | TVar (X : (Index TyVar))
    | TArr (T1 : Ty) (T2 : Ty)
    | TAll (T : Ty).
  Scheme ind_Ty := Induction for Ty Sort Prop.
  
  Inductive Tm : Set :=
    | Var (x : (Index TmVar))
    | XVar (a : (Index TxVar))
    | Abs (T : Ty) (t : Tm)
    | App (t1 : Tm) (t2 : Tm)
    | TAbs (t : Tm)
    | TApp (t : Tm) (T : Ty).
  Scheme ind_Tm := Induction for Tm Sort Prop.
End Terms.

Section Functions.
  
End Functions.

Section Shift.
  Inductive Cutoff (a : Namespace) : Type :=
    | C0 
    | CS : (Cutoff a) -> (Cutoff a).
  
  Global Arguments C0 {a} .
  Global Arguments CS {a} _ .
  Fixpoint weakenCutoffTmVar (c : (Cutoff TmVar)) (k : Hvl) {struct k} : (Cutoff TmVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TmVar) => (CS (weakenCutoffTmVar c k))
        | _ => (weakenCutoffTmVar c k)
      end
    end.
  Fixpoint weakenCutoffTxVar (c : (Cutoff TxVar)) (k : Hvl) {struct k} : (Cutoff TxVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TxVar) => (CS (weakenCutoffTxVar c k))
        | _ => (weakenCutoffTxVar c k)
      end
    end.
  Fixpoint weakenCutoffTyVar (c : (Cutoff TyVar)) (k : Hvl) {struct k} : (Cutoff TyVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TyVar) => (CS (weakenCutoffTyVar c k))
        | _ => (weakenCutoffTyVar c k)
      end
    end.
  
  Lemma weakenCutoffTmVar_append  :
    (forall (c : (Cutoff TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTmVar (weakenCutoffTmVar c k) k0) =
      (weakenCutoffTmVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenCutoffTxVar_append  :
    (forall (c : (Cutoff TxVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTxVar (weakenCutoffTxVar c k) k0) =
      (weakenCutoffTxVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenCutoffTyVar_append  :
    (forall (c : (Cutoff TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTyVar (weakenCutoffTyVar c k) k0) =
      (weakenCutoffTyVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint shift_TmVar_Index (c : (Cutoff TmVar)) (x : (Index TmVar)) {struct c} : (Index TmVar) :=
    match c with
      | (C0) => (IS x)
      | (CS c) => match x with
        | (I0) => I0
        | (IS x) => (IS (shift_TmVar_Index c x))
      end
    end.
  Fixpoint shift_TxVar_Index (c : (Cutoff TxVar)) (a : (Index TxVar)) {struct c} : (Index TxVar) :=
    match c with
      | (C0) => (IS a)
      | (CS c) => match a with
        | (I0) => I0
        | (IS a) => (IS (shift_TxVar_Index c a))
      end
    end.
  Fixpoint shift_TyVar_Index (c : (Cutoff TyVar)) (X : (Index TyVar)) {struct c} : (Index TyVar) :=
    match c with
      | (C0) => (IS X)
      | (CS c) => match X with
        | (I0) => I0
        | (IS X) => (IS (shift_TyVar_Index c X))
      end
    end.
  Fixpoint shift_TyVar_Ty (c : (Cutoff TyVar)) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (TVar X) => (TVar (shift_TyVar_Index c X))
      | (TArr T1 T2) => (TArr (shift_TyVar_Ty (weakenCutoffTyVar c H0) T1) (shift_TyVar_Ty (weakenCutoffTyVar c H0) T2))
      | (TAll T) => (TAll (shift_TyVar_Ty (weakenCutoffTyVar c (appendHvl (HS TyVar H0) H0)) T))
    end.
  Fixpoint shift_TmVar_Tm (c : (Cutoff TmVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var (shift_TmVar_Index c x))
      | (XVar a) => (XVar a)
      | (Abs T t) => (Abs T (shift_TmVar_Tm (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (shift_TmVar_Tm (weakenCutoffTmVar c H0) t1) (shift_TmVar_Tm (weakenCutoffTmVar c H0) t2))
      | (TAbs t0) => (TAbs (shift_TmVar_Tm (weakenCutoffTmVar c (appendHvl (HS TyVar H0) H0)) t0))
      | (TApp t3 T0) => (TApp (shift_TmVar_Tm (weakenCutoffTmVar c H0) t3) T0)
    end.
  Fixpoint shift_TxVar_Tm (c : (Cutoff TxVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (XVar a) => (XVar (shift_TxVar_Index c a))
      | (Abs T t) => (Abs T (shift_TxVar_Tm (weakenCutoffTxVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (shift_TxVar_Tm (weakenCutoffTxVar c H0) t1) (shift_TxVar_Tm (weakenCutoffTxVar c H0) t2))
      | (TAbs t0) => (TAbs (shift_TxVar_Tm (weakenCutoffTxVar c (appendHvl (HS TyVar H0) H0)) t0))
      | (TApp t3 T0) => (TApp (shift_TxVar_Tm (weakenCutoffTxVar c H0) t3) T0)
    end.
  Fixpoint shift_TyVar_Tm (c : (Cutoff TyVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (XVar a) => (XVar a)
      | (Abs T t) => (Abs (shift_TyVar_Ty (weakenCutoffTyVar c H0) T) (shift_TyVar_Tm (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) t))
      | (App t1 t2) => (App (shift_TyVar_Tm (weakenCutoffTyVar c H0) t1) (shift_TyVar_Tm (weakenCutoffTyVar c H0) t2))
      | (TAbs t0) => (TAbs (shift_TyVar_Tm (weakenCutoffTyVar c (appendHvl (HS TyVar H0) H0)) t0))
      | (TApp t3 T0) => (TApp (shift_TyVar_Tm (weakenCutoffTyVar c H0) t3) (shift_TyVar_Ty (weakenCutoffTyVar c H0) T0))
    end.
End Shift.

Section Weaken.
  Fixpoint weakenTy (S0 : Ty) (k : Hvl) {struct k} : Ty :=
    match k with
      | (H0) => S0
      | (HS TmVar k) => (weakenTy S0 k)
      | (HS TxVar k) => (weakenTy S0 k)
      | (HS TyVar k) => (shift_TyVar_Ty C0 (weakenTy S0 k))
    end.
  Fixpoint weakenTm (s : Tm) (k : Hvl) {struct k} : Tm :=
    match k with
      | (H0) => s
      | (HS TmVar k) => (shift_TmVar_Tm C0 (weakenTm s k))
      | (HS TxVar k) => (shift_TxVar_Tm C0 (weakenTm s k))
      | (HS TyVar k) => (shift_TyVar_Tm C0 (weakenTm s k))
    end.
End Weaken.

Section Subst.
  Inductive Trace (a : Namespace) : Type :=
    | X0 
    | XS (b : Namespace)
        (T : (Trace a)).
  
  Global Arguments X0 [a] .
  Global Arguments XS [a] b _ .
  Fixpoint weakenTrace {a : Namespace} (x : (Trace a)) (k : Hvl) {struct k} : (Trace a) :=
    match k with
      | (H0) => x
      | (HS b k) => (XS b (weakenTrace x k))
    end.
  Lemma weakenTrace_append (a : Namespace) :
    (forall (x : (Trace a)) (k : Hvl) (k0 : Hvl) ,
      ((weakenTrace (weakenTrace x k) k0) =
      (weakenTrace x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint subst_TmVar_Index (d : (Trace TmVar)) (s : Tm) (x : (Index TmVar)) {struct d} : Tm :=
    match d with
      | (X0) => match x with
        | (I0) => s
        | (IS x) => (Var x)
      end
      | (XS TmVar d) => match x with
        | (I0) => (Var I0)
        | (IS x) => (shift_TmVar_Tm C0 (subst_TmVar_Index d s x))
      end
      | (XS TxVar d) => (shift_TxVar_Tm C0 (subst_TmVar_Index d s x))
      | (XS TyVar d) => (shift_TyVar_Tm C0 (subst_TmVar_Index d s x))
    end.
  Fixpoint subst_TxVar_Index (d : (Trace TxVar)) (s : Tm) (a : (Index TxVar)) {struct d} : Tm :=
    match d with
      | (X0) => match a with
        | (I0) => s
        | (IS a) => (XVar a)
      end
      | (XS TmVar d) => (shift_TmVar_Tm C0 (subst_TxVar_Index d s a))
      | (XS TxVar d) => match a with
        | (I0) => (XVar I0)
        | (IS a) => (shift_TxVar_Tm C0 (subst_TxVar_Index d s a))
      end
      | (XS TyVar d) => (shift_TyVar_Tm C0 (subst_TxVar_Index d s a))
    end.
  Fixpoint subst_TyVar_Index (d : (Trace TyVar)) (S0 : Ty) (X : (Index TyVar)) {struct d} : Ty :=
    match d with
      | (X0) => match X with
        | (I0) => S0
        | (IS X) => (TVar X)
      end
      | (XS TmVar d) => (subst_TyVar_Index d S0 X)
      | (XS TxVar d) => (subst_TyVar_Index d S0 X)
      | (XS TyVar d) => match X with
        | (I0) => (TVar I0)
        | (IS X) => (shift_TyVar_Ty C0 (subst_TyVar_Index d S0 X))
      end
    end.
  Fixpoint subst_TyVar_Ty (d : (Trace TyVar)) (S0 : Ty) (S1 : Ty) {struct S1} : Ty :=
    match S1 with
      | (TVar X) => (subst_TyVar_Index d S0 X)
      | (TArr T1 T2) => (TArr (subst_TyVar_Ty d S0 T1) (subst_TyVar_Ty d S0 T2))
      | (TAll T) => (TAll (subst_TyVar_Ty (XS TyVar d) S0 T))
    end.
  Fixpoint subst_TmVar_Tm (d : (Trace TmVar)) (s : Tm) (s0 : Tm) {struct s0} : Tm :=
    match s0 with
      | (Var x) => (subst_TmVar_Index d s x)
      | (XVar a) => (XVar a)
      | (Abs T t) => (Abs T (subst_TmVar_Tm (XS TmVar d) s t))
      | (App t1 t2) => (App (subst_TmVar_Tm d s t1) (subst_TmVar_Tm d s t2))
      | (TAbs t0) => (TAbs (subst_TmVar_Tm (XS TyVar d) s t0))
      | (TApp t3 T0) => (TApp (subst_TmVar_Tm d s t3) T0)
    end.
  Fixpoint subst_TxVar_Tm (d : (Trace TxVar)) (s : Tm) (s0 : Tm) {struct s0} : Tm :=
    match s0 with
      | (Var x) => (Var x)
      | (XVar a) => (subst_TxVar_Index d s a)
      | (Abs T t) => (Abs T (subst_TxVar_Tm (XS TmVar d) s t))
      | (App t1 t2) => (App (subst_TxVar_Tm d s t1) (subst_TxVar_Tm d s t2))
      | (TAbs t0) => (TAbs (subst_TxVar_Tm (XS TyVar d) s t0))
      | (TApp t3 T0) => (TApp (subst_TxVar_Tm d s t3) T0)
    end.
  Fixpoint subst_TyVar_Tm (d : (Trace TyVar)) (S0 : Ty) (s : Tm) {struct s} : Tm :=
    match s with
      | (Var x) => (Var x)
      | (XVar a) => (XVar a)
      | (Abs T t) => (Abs (subst_TyVar_Ty d S0 T) (subst_TyVar_Tm (XS TmVar d) S0 t))
      | (App t1 t2) => (App (subst_TyVar_Tm d S0 t1) (subst_TyVar_Tm d S0 t2))
      | (TAbs t0) => (TAbs (subst_TyVar_Tm (XS TyVar d) S0 t0))
      | (TApp t3 T0) => (TApp (subst_TyVar_Tm d S0 t3) (subst_TyVar_Ty d S0 T0))
    end.
End Subst.

Global Hint Resolve (f_equal (shift_TmVar_Tm C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TxVar_Tm C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TyVar_Tm C0)) : cong_shift0.
Global Hint Resolve (f_equal (shift_TyVar_Ty C0)) : cong_shift0.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTxVar_append weakenCutoffTyVar_append : interaction.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTxVar_append weakenCutoffTyVar_append : weakencutoff_append.
 Hint Rewrite weakenTrace_append : interaction.
 Hint Rewrite weakenTrace_append : weakentrace_append.
Section SubstShiftReflection.
  Lemma subst_TmVar_Index0_shift_TmVar_Index0_reflection_ind (s : Tm) :
    (forall (k18 : Hvl) (x23 : (Index TmVar)) ,
      ((subst_TmVar_Index (weakenTrace X0 k18) s (shift_TmVar_Index (weakenCutoffTmVar C0 k18) x23)) =
      (Var x23))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Lemma subst_TxVar_Index0_shift_TxVar_Index0_reflection_ind (s : Tm) :
    (forall (k18 : Hvl) (a14 : (Index TxVar)) ,
      ((subst_TxVar_Index (weakenTrace X0 k18) s (shift_TxVar_Index (weakenCutoffTxVar C0 k18) a14)) =
      (XVar a14))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Lemma subst_TyVar_Index0_shift_TyVar_Index0_reflection_ind (S0 : Ty) :
    (forall (k18 : Hvl) (X17 : (Index TyVar)) ,
      ((subst_TyVar_Index (weakenTrace X0 k18) S0 (shift_TyVar_Index (weakenCutoffTyVar C0 k18) X17)) =
      (TVar X17))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Definition subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind := (ind_Ty (fun (S3 : Ty) =>
    (forall (k25 : Hvl) (S4 : Ty) ,
      ((subst_TyVar_Ty (weakenTrace X0 k25) S4 (shift_TyVar_Ty (weakenCutoffTyVar C0 k25) S3)) =
      S3))) (fun (X25 : (Index TyVar)) =>
    (fun (k25 : Hvl) (S3 : Ty) =>
      (subst_TyVar_Index0_shift_TyVar_Index0_reflection_ind S3 k25 X25))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k25 : Hvl) (S3 : Ty) =>
    (f_equal2 TArr (IHT0 k25 S3) (IHT3 k25 S3))) (fun (T : Ty) IHT0 (k25 : Hvl) (S3 : Ty) =>
    (f_equal TAll (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k25 (HS TyVar H0)) eq_refl eq_refl) (IHT0 (HS TyVar k25) S3))))).
  Definition subst_TmVar_0_shift_TmVar_0_Tm_reflection_ind := (ind_Tm (fun (s3 : Tm) =>
    (forall (k25 : Hvl) (s4 : Tm) ,
      ((subst_TmVar_Tm (weakenTrace X0 k25) s4 (shift_TmVar_Tm (weakenCutoffTmVar C0 k25) s3)) =
      s3))) (fun (x31 : (Index TmVar)) =>
    (fun (k25 : Hvl) (s3 : Tm) =>
      (subst_TmVar_Index0_shift_TmVar_Index0_reflection_ind s3 k25 x31))) (fun (a19 : (Index TxVar)) =>
    (fun (k25 : Hvl) (s3 : Tm) =>
      eq_refl)) (fun (T : Ty) (t : Tm) IHt59 (k25 : Hvl) (s3 : Tm) =>
    (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k25 (HS TmVar H0)) eq_refl eq_refl) (IHt59 (HS TmVar k25) s3)))) (fun (t1 : Tm) IHt59 (t2 : Tm) IHt60 (k25 : Hvl) (s3 : Tm) =>
    (f_equal2 App (IHt59 k25 s3) (IHt60 k25 s3))) (fun (t : Tm) IHt59 (k25 : Hvl) (s3 : Tm) =>
    (f_equal TAbs (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k25 (HS TyVar H0)) eq_refl eq_refl) (IHt59 (HS TyVar k25) s3)))) (fun (t : Tm) IHt59 (T : Ty) (k25 : Hvl) (s3 : Tm) =>
    (f_equal2 TApp (IHt59 k25 s3) eq_refl))).
  Definition subst_TxVar_0_shift_TxVar_0_Tm_reflection_ind := (ind_Tm (fun (s3 : Tm) =>
    (forall (k25 : Hvl) (s4 : Tm) ,
      ((subst_TxVar_Tm (weakenTrace X0 k25) s4 (shift_TxVar_Tm (weakenCutoffTxVar C0 k25) s3)) =
      s3))) (fun (x31 : (Index TmVar)) =>
    (fun (k25 : Hvl) (s3 : Tm) =>
      eq_refl)) (fun (a19 : (Index TxVar)) =>
    (fun (k25 : Hvl) (s3 : Tm) =>
      (subst_TxVar_Index0_shift_TxVar_Index0_reflection_ind s3 k25 a19))) (fun (T : Ty) (t : Tm) IHt59 (k25 : Hvl) (s3 : Tm) =>
    (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar X0 k25 (HS TmVar H0)) eq_refl eq_refl) (IHt59 (HS TmVar k25) s3)))) (fun (t1 : Tm) IHt59 (t2 : Tm) IHt60 (k25 : Hvl) (s3 : Tm) =>
    (f_equal2 App (IHt59 k25 s3) (IHt60 k25 s3))) (fun (t : Tm) IHt59 (k25 : Hvl) (s3 : Tm) =>
    (f_equal TAbs (eq_trans (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar X0 k25 (HS TyVar H0)) eq_refl eq_refl) (IHt59 (HS TyVar k25) s3)))) (fun (t : Tm) IHt59 (T : Ty) (k25 : Hvl) (s3 : Tm) =>
    (f_equal2 TApp (IHt59 k25 s3) eq_refl))).
  Definition subst_TyVar_0_shift_TyVar_0_Tm_reflection_ind := (ind_Tm (fun (s3 : Tm) =>
    (forall (k25 : Hvl) (S3 : Ty) ,
      ((subst_TyVar_Tm (weakenTrace X0 k25) S3 (shift_TyVar_Tm (weakenCutoffTyVar C0 k25) s3)) =
      s3))) (fun (x31 : (Index TmVar)) =>
    (fun (k25 : Hvl) (S3 : Ty) =>
      eq_refl)) (fun (a19 : (Index TxVar)) =>
    (fun (k25 : Hvl) (S3 : Ty) =>
      eq_refl)) (fun (T : Ty) (t : Tm) IHt59 (k25 : Hvl) (S3 : Ty) =>
    (f_equal2 Abs (let IHT0 := (subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind T) in
    (IHT0 k25 S3)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k25 (HS TmVar H0)) eq_refl eq_refl) (IHt59 (HS TmVar k25) S3)))) (fun (t1 : Tm) IHt59 (t2 : Tm) IHt60 (k25 : Hvl) (S3 : Ty) =>
    (f_equal2 App (IHt59 k25 S3) (IHt60 k25 S3))) (fun (t : Tm) IHt59 (k25 : Hvl) (S3 : Ty) =>
    (f_equal TAbs (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k25 (HS TyVar H0)) eq_refl eq_refl) (IHt59 (HS TyVar k25) S3)))) (fun (t : Tm) IHt59 (T : Ty) (k25 : Hvl) (S3 : Ty) =>
    (f_equal2 TApp (IHt59 k25 S3) (let IHT0 := (subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind T) in
    (IHT0 k25 S3))))).
  Definition subst_TyVar_Ty0_shift_TyVar_Ty0_reflection (S4 : Ty) : (forall (S3 : Ty) ,
    ((subst_TyVar_Ty X0 S3 (shift_TyVar_Ty C0 S4)) =
    S4)) := (subst_TyVar_0_shift_TyVar_0_Ty_reflection_ind S4 H0).
  Definition subst_TmVar_Tm0_shift_TmVar_Tm0_reflection (s4 : Tm) : (forall (s3 : Tm) ,
    ((subst_TmVar_Tm X0 s3 (shift_TmVar_Tm C0 s4)) =
    s4)) := (subst_TmVar_0_shift_TmVar_0_Tm_reflection_ind s4 H0).
  Definition subst_TxVar_Tm0_shift_TxVar_Tm0_reflection (s4 : Tm) : (forall (s3 : Tm) ,
    ((subst_TxVar_Tm X0 s3 (shift_TxVar_Tm C0 s4)) =
    s4)) := (subst_TxVar_0_shift_TxVar_0_Tm_reflection_ind s4 H0).
  Definition subst_TyVar_Tm0_shift_TyVar_Tm0_reflection (s3 : Tm) : (forall (S3 : Ty) ,
    ((subst_TyVar_Tm X0 S3 (shift_TyVar_Tm C0 s3)) =
    s3)) := (subst_TyVar_0_shift_TyVar_0_Tm_reflection_ind s3 H0).
End SubstShiftReflection.
Section ShiftInteraction.
  Section ShiftIndexCommInd.
    Lemma shift_TmVar_Index_shift_TmVar_Index0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff TmVar)) (x : (Index TmVar)) ,
        ((shift_TmVar_Index (weakenCutoffTmVar (CS c) k) (shift_TmVar_Index (weakenCutoffTmVar C0 k) x)) =
        (shift_TmVar_Index (weakenCutoffTmVar C0 k) (shift_TmVar_Index (weakenCutoffTmVar c k) x)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
    Lemma shift_TxVar_Index_shift_TxVar_Index0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff TxVar)) (a : (Index TxVar)) ,
        ((shift_TxVar_Index (weakenCutoffTxVar (CS c) k) (shift_TxVar_Index (weakenCutoffTxVar C0 k) a)) =
        (shift_TxVar_Index (weakenCutoffTxVar C0 k) (shift_TxVar_Index (weakenCutoffTxVar c k) a)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
    Lemma shift_TyVar_Index_shift_TyVar_Index0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff TyVar)) (X : (Index TyVar)) ,
        ((shift_TyVar_Index (weakenCutoffTyVar (CS c) k) (shift_TyVar_Index (weakenCutoffTyVar C0 k) X)) =
        (shift_TyVar_Index (weakenCutoffTyVar C0 k) (shift_TyVar_Index (weakenCutoffTyVar c k) X)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
  End ShiftIndexCommInd.
  Section ShiftCommInd.
    Definition shift_TyVar__shift_TyVar_0_Ty_comm_ind := (ind_Ty (fun (S0 : Ty) =>
      (forall (k18 : Hvl) (c18 : (Cutoff TyVar)) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar (CS c18) k18) (shift_TyVar_Ty (weakenCutoffTyVar C0 k18) S0)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k18) (shift_TyVar_Ty (weakenCutoffTyVar c18 k18) S0))))) (fun (X17 : (Index TyVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
        (f_equal TVar (shift_TyVar_Index_shift_TyVar_Index0_comm_ind k18 c18 X17)))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
      (f_equal2 TArr (IHT0 k18 c18) (IHT3 k18 c18))) (fun (T : Ty) IHT0 (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
      (f_equal TAll (IHT0 (HS TyVar k18) c18)))).
    Definition shift_TmVar__shift_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k18 : Hvl) (c18 : (Cutoff TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar (CS c18) k18) (shift_TmVar_Tm (weakenCutoffTmVar C0 k18) s)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k18) (shift_TmVar_Tm (weakenCutoffTmVar c18 k18) s))))) (fun (x23 : (Index TmVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
        (f_equal Var (shift_TmVar_Index_shift_TmVar_Index0_comm_ind k18 c18 x23)))) (fun (a14 : (Index TxVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
      (f_equal2 Abs eq_refl (IHt44 (HS TmVar k18) c18))) (fun (t1 : Tm) IHt44 (t2 : Tm) IHt45 (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt44 k18 c18) (IHt45 k18 c18))) (fun (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
      (f_equal TAbs (IHt44 (HS TyVar k18) c18))) (fun (t : Tm) IHt44 (T : Ty) (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
      (f_equal2 TApp (IHt44 k18 c18) eq_refl))).
    Definition shift_TmVar__shift_TxVar_0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k18 : Hvl) (c18 : (Cutoff TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c18 k18) (shift_TxVar_Tm (weakenCutoffTxVar C0 k18) s)) =
        (shift_TxVar_Tm (weakenCutoffTxVar C0 k18) (shift_TmVar_Tm (weakenCutoffTmVar c18 k18) s))))) (fun (x23 : (Index TmVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
        eq_refl)) (fun (a14 : (Index TxVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
      (f_equal2 Abs eq_refl (IHt44 (HS TmVar k18) c18))) (fun (t1 : Tm) IHt44 (t2 : Tm) IHt45 (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt44 k18 c18) (IHt45 k18 c18))) (fun (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
      (f_equal TAbs (IHt44 (HS TyVar k18) c18))) (fun (t : Tm) IHt44 (T : Ty) (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
      (f_equal2 TApp (IHt44 k18 c18) eq_refl))).
    Definition shift_TmVar__shift_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k18 : Hvl) (c18 : (Cutoff TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c18 k18) (shift_TyVar_Tm (weakenCutoffTyVar C0 k18) s)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k18) (shift_TmVar_Tm (weakenCutoffTmVar c18 k18) s))))) (fun (x23 : (Index TmVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
        eq_refl)) (fun (a14 : (Index TxVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
      (f_equal2 Abs eq_refl (IHt44 (HS TmVar k18) c18))) (fun (t1 : Tm) IHt44 (t2 : Tm) IHt45 (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
      (f_equal2 App (IHt44 k18 c18) (IHt45 k18 c18))) (fun (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
      (f_equal TAbs (IHt44 (HS TyVar k18) c18))) (fun (t : Tm) IHt44 (T : Ty) (k18 : Hvl) (c18 : (Cutoff TmVar)) =>
      (f_equal2 TApp (IHt44 k18 c18) eq_refl))).
    Definition shift_TxVar__shift_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k18 : Hvl) (c18 : (Cutoff TxVar)) ,
        ((shift_TxVar_Tm (weakenCutoffTxVar c18 k18) (shift_TmVar_Tm (weakenCutoffTmVar C0 k18) s)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k18) (shift_TxVar_Tm (weakenCutoffTxVar c18 k18) s))))) (fun (x23 : (Index TmVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
        eq_refl)) (fun (a14 : (Index TxVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
      (f_equal2 Abs eq_refl (IHt44 (HS TmVar k18) c18))) (fun (t1 : Tm) IHt44 (t2 : Tm) IHt45 (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
      (f_equal2 App (IHt44 k18 c18) (IHt45 k18 c18))) (fun (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
      (f_equal TAbs (IHt44 (HS TyVar k18) c18))) (fun (t : Tm) IHt44 (T : Ty) (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
      (f_equal2 TApp (IHt44 k18 c18) eq_refl))).
    Definition shift_TxVar__shift_TxVar_0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k18 : Hvl) (c18 : (Cutoff TxVar)) ,
        ((shift_TxVar_Tm (weakenCutoffTxVar (CS c18) k18) (shift_TxVar_Tm (weakenCutoffTxVar C0 k18) s)) =
        (shift_TxVar_Tm (weakenCutoffTxVar C0 k18) (shift_TxVar_Tm (weakenCutoffTxVar c18 k18) s))))) (fun (x23 : (Index TmVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
        eq_refl)) (fun (a14 : (Index TxVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
        (f_equal XVar (shift_TxVar_Index_shift_TxVar_Index0_comm_ind k18 c18 a14)))) (fun (T : Ty) (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
      (f_equal2 Abs eq_refl (IHt44 (HS TmVar k18) c18))) (fun (t1 : Tm) IHt44 (t2 : Tm) IHt45 (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
      (f_equal2 App (IHt44 k18 c18) (IHt45 k18 c18))) (fun (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
      (f_equal TAbs (IHt44 (HS TyVar k18) c18))) (fun (t : Tm) IHt44 (T : Ty) (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
      (f_equal2 TApp (IHt44 k18 c18) eq_refl))).
    Definition shift_TxVar__shift_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k18 : Hvl) (c18 : (Cutoff TxVar)) ,
        ((shift_TxVar_Tm (weakenCutoffTxVar c18 k18) (shift_TyVar_Tm (weakenCutoffTyVar C0 k18) s)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k18) (shift_TxVar_Tm (weakenCutoffTxVar c18 k18) s))))) (fun (x23 : (Index TmVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
        eq_refl)) (fun (a14 : (Index TxVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
      (f_equal2 Abs eq_refl (IHt44 (HS TmVar k18) c18))) (fun (t1 : Tm) IHt44 (t2 : Tm) IHt45 (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
      (f_equal2 App (IHt44 k18 c18) (IHt45 k18 c18))) (fun (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
      (f_equal TAbs (IHt44 (HS TyVar k18) c18))) (fun (t : Tm) IHt44 (T : Ty) (k18 : Hvl) (c18 : (Cutoff TxVar)) =>
      (f_equal2 TApp (IHt44 k18 c18) eq_refl))).
    Definition shift_TyVar__shift_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k18 : Hvl) (c18 : (Cutoff TyVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c18 k18) (shift_TmVar_Tm (weakenCutoffTmVar C0 k18) s)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k18) (shift_TyVar_Tm (weakenCutoffTyVar c18 k18) s))))) (fun (x23 : (Index TmVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
        eq_refl)) (fun (a14 : (Index TxVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
      (f_equal2 Abs eq_refl (IHt44 (HS TmVar k18) c18))) (fun (t1 : Tm) IHt44 (t2 : Tm) IHt45 (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt44 k18 c18) (IHt45 k18 c18))) (fun (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
      (f_equal TAbs (IHt44 (HS TyVar k18) c18))) (fun (t : Tm) IHt44 (T : Ty) (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
      (f_equal2 TApp (IHt44 k18 c18) eq_refl))).
    Definition shift_TyVar__shift_TxVar_0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k18 : Hvl) (c18 : (Cutoff TyVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c18 k18) (shift_TxVar_Tm (weakenCutoffTxVar C0 k18) s)) =
        (shift_TxVar_Tm (weakenCutoffTxVar C0 k18) (shift_TyVar_Tm (weakenCutoffTyVar c18 k18) s))))) (fun (x23 : (Index TmVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
        eq_refl)) (fun (a14 : (Index TxVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
      (f_equal2 Abs eq_refl (IHt44 (HS TmVar k18) c18))) (fun (t1 : Tm) IHt44 (t2 : Tm) IHt45 (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt44 k18 c18) (IHt45 k18 c18))) (fun (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
      (f_equal TAbs (IHt44 (HS TyVar k18) c18))) (fun (t : Tm) IHt44 (T : Ty) (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
      (f_equal2 TApp (IHt44 k18 c18) eq_refl))).
    Definition shift_TyVar__shift_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k18 : Hvl) (c18 : (Cutoff TyVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar (CS c18) k18) (shift_TyVar_Tm (weakenCutoffTyVar C0 k18) s)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k18) (shift_TyVar_Tm (weakenCutoffTyVar c18 k18) s))))) (fun (x23 : (Index TmVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
        eq_refl)) (fun (a14 : (Index TxVar)) =>
      (fun (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
      (f_equal2 Abs (let IHT0 := (shift_TyVar__shift_TyVar_0_Ty_comm_ind T) in
      (IHT0 k18 c18)) (IHt44 (HS TmVar k18) c18))) (fun (t1 : Tm) IHt44 (t2 : Tm) IHt45 (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
      (f_equal2 App (IHt44 k18 c18) (IHt45 k18 c18))) (fun (t : Tm) IHt44 (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
      (f_equal TAbs (IHt44 (HS TyVar k18) c18))) (fun (t : Tm) IHt44 (T : Ty) (k18 : Hvl) (c18 : (Cutoff TyVar)) =>
      (f_equal2 TApp (IHt44 k18 c18) (let IHT0 := (shift_TyVar__shift_TyVar_0_Ty_comm_ind T) in
      (IHT0 k18 c18))))).
  End ShiftCommInd.
  Section ShiftComm.
    Definition shift_TyVar__shift_TyVar_0_Ty_comm (S0 : Ty) : (forall (c18 : (Cutoff TyVar)) ,
      ((shift_TyVar_Ty (CS c18) (shift_TyVar_Ty C0 S0)) =
      (shift_TyVar_Ty C0 (shift_TyVar_Ty c18 S0)))) := (shift_TyVar__shift_TyVar_0_Ty_comm_ind S0 H0).
    Definition shift_TmVar__shift_TmVar_0_Tm_comm (s : Tm) : (forall (c18 : (Cutoff TmVar)) ,
      ((shift_TmVar_Tm (CS c18) (shift_TmVar_Tm C0 s)) =
      (shift_TmVar_Tm C0 (shift_TmVar_Tm c18 s)))) := (shift_TmVar__shift_TmVar_0_Tm_comm_ind s H0).
    Definition shift_TmVar__shift_TxVar_0_Tm_comm (s : Tm) : (forall (c18 : (Cutoff TmVar)) ,
      ((shift_TmVar_Tm c18 (shift_TxVar_Tm C0 s)) =
      (shift_TxVar_Tm C0 (shift_TmVar_Tm c18 s)))) := (shift_TmVar__shift_TxVar_0_Tm_comm_ind s H0).
    Definition shift_TmVar__shift_TyVar_0_Tm_comm (s : Tm) : (forall (c18 : (Cutoff TmVar)) ,
      ((shift_TmVar_Tm c18 (shift_TyVar_Tm C0 s)) =
      (shift_TyVar_Tm C0 (shift_TmVar_Tm c18 s)))) := (shift_TmVar__shift_TyVar_0_Tm_comm_ind s H0).
    Definition shift_TxVar__shift_TmVar_0_Tm_comm (s : Tm) : (forall (c18 : (Cutoff TxVar)) ,
      ((shift_TxVar_Tm c18 (shift_TmVar_Tm C0 s)) =
      (shift_TmVar_Tm C0 (shift_TxVar_Tm c18 s)))) := (shift_TxVar__shift_TmVar_0_Tm_comm_ind s H0).
    Definition shift_TxVar__shift_TxVar_0_Tm_comm (s : Tm) : (forall (c18 : (Cutoff TxVar)) ,
      ((shift_TxVar_Tm (CS c18) (shift_TxVar_Tm C0 s)) =
      (shift_TxVar_Tm C0 (shift_TxVar_Tm c18 s)))) := (shift_TxVar__shift_TxVar_0_Tm_comm_ind s H0).
    Definition shift_TxVar__shift_TyVar_0_Tm_comm (s : Tm) : (forall (c18 : (Cutoff TxVar)) ,
      ((shift_TxVar_Tm c18 (shift_TyVar_Tm C0 s)) =
      (shift_TyVar_Tm C0 (shift_TxVar_Tm c18 s)))) := (shift_TxVar__shift_TyVar_0_Tm_comm_ind s H0).
    Definition shift_TyVar__shift_TmVar_0_Tm_comm (s : Tm) : (forall (c18 : (Cutoff TyVar)) ,
      ((shift_TyVar_Tm c18 (shift_TmVar_Tm C0 s)) =
      (shift_TmVar_Tm C0 (shift_TyVar_Tm c18 s)))) := (shift_TyVar__shift_TmVar_0_Tm_comm_ind s H0).
    Definition shift_TyVar__shift_TxVar_0_Tm_comm (s : Tm) : (forall (c18 : (Cutoff TyVar)) ,
      ((shift_TyVar_Tm c18 (shift_TxVar_Tm C0 s)) =
      (shift_TxVar_Tm C0 (shift_TyVar_Tm c18 s)))) := (shift_TyVar__shift_TxVar_0_Tm_comm_ind s H0).
    Definition shift_TyVar__shift_TyVar_0_Tm_comm (s : Tm) : (forall (c18 : (Cutoff TyVar)) ,
      ((shift_TyVar_Tm (CS c18) (shift_TyVar_Tm C0 s)) =
      (shift_TyVar_Tm C0 (shift_TyVar_Tm c18 s)))) := (shift_TyVar__shift_TyVar_0_Tm_comm_ind s H0).
  End ShiftComm.
End ShiftInteraction.
 Hint Rewrite shift_TmVar__shift_TmVar_0_Tm_comm shift_TmVar__shift_TxVar_0_Tm_comm shift_TmVar__shift_TyVar_0_Tm_comm shift_TxVar__shift_TmVar_0_Tm_comm shift_TxVar__shift_TxVar_0_Tm_comm shift_TxVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TmVar_0_Tm_comm shift_TyVar__shift_TxVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Ty_comm : interaction.
 Hint Rewrite shift_TmVar__shift_TmVar_0_Tm_comm shift_TmVar__shift_TxVar_0_Tm_comm shift_TmVar__shift_TyVar_0_Tm_comm shift_TxVar__shift_TmVar_0_Tm_comm shift_TxVar__shift_TxVar_0_Tm_comm shift_TxVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TmVar_0_Tm_comm shift_TyVar__shift_TxVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Tm_comm shift_TyVar__shift_TyVar_0_Ty_comm : shift_shift0.
Section WeakenShift.
  Lemma weakenTy_shift_TyVar_Ty  :
    (forall (k126 : Hvl) (c37 : (Cutoff TyVar)) (S55 : Ty) ,
      ((weakenTy (shift_TyVar_Ty c37 S55) k126) =
      (shift_TyVar_Ty (weakenCutoffTyVar c37 k126) (weakenTy S55 k126)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_shift_TmVar_Tm  :
    (forall (k126 : Hvl) (c37 : (Cutoff TmVar)) (s59 : Tm) ,
      ((weakenTm (shift_TmVar_Tm c37 s59) k126) =
      (shift_TmVar_Tm (weakenCutoffTmVar c37 k126) (weakenTm s59 k126)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_shift_TxVar_Tm  :
    (forall (k126 : Hvl) (c37 : (Cutoff TxVar)) (s59 : Tm) ,
      ((weakenTm (shift_TxVar_Tm c37 s59) k126) =
      (shift_TxVar_Tm (weakenCutoffTxVar c37 k126) (weakenTm s59 k126)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_shift_TyVar_Tm  :
    (forall (k126 : Hvl) (c37 : (Cutoff TyVar)) (s59 : Tm) ,
      ((weakenTm (shift_TyVar_Tm c37 s59) k126) =
      (shift_TyVar_Tm (weakenCutoffTyVar c37 k126) (weakenTm s59 k126)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
End WeakenShift.
Section ShiftSubstInteraction.
  Section ShiftSubstIndexCommInd.
    Lemma shift_TmVar_Tm_subst_TmVar_Index0_comm_ind (c18 : (Cutoff TmVar)) (s3 : Tm) :
      (forall (k25 : Hvl) (x31 : (Index TmVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c18 k25) (subst_TmVar_Index (weakenTrace X0 k25) s3 x31)) =
        (subst_TmVar_Index (weakenTrace X0 k25) (shift_TmVar_Tm c18 s3) (shift_TmVar_Index (weakenCutoffTmVar (CS c18) k25) x31)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TxVar_Tm_subst_TmVar_Index0_comm_ind (c18 : (Cutoff TxVar)) (s3 : Tm) :
      (forall (k25 : Hvl) (x31 : (Index TmVar)) ,
        ((shift_TxVar_Tm (weakenCutoffTxVar c18 k25) (subst_TmVar_Index (weakenTrace X0 k25) s3 x31)) =
        (subst_TmVar_Index (weakenTrace X0 k25) (shift_TxVar_Tm c18 s3) x31))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TyVar_Tm_subst_TmVar_Index0_comm_ind (c18 : (Cutoff TyVar)) (s3 : Tm) :
      (forall (k25 : Hvl) (x31 : (Index TmVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c18 k25) (subst_TmVar_Index (weakenTrace X0 k25) s3 x31)) =
        (subst_TmVar_Index (weakenTrace X0 k25) (shift_TyVar_Tm c18 s3) x31))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TmVar_Tm_subst_TxVar_Index0_comm_ind (c18 : (Cutoff TmVar)) (s3 : Tm) :
      (forall (k25 : Hvl) (a19 : (Index TxVar)) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c18 k25) (subst_TxVar_Index (weakenTrace X0 k25) s3 a19)) =
        (subst_TxVar_Index (weakenTrace X0 k25) (shift_TmVar_Tm c18 s3) a19))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TxVar_Tm_subst_TxVar_Index0_comm_ind (c18 : (Cutoff TxVar)) (s3 : Tm) :
      (forall (k25 : Hvl) (a19 : (Index TxVar)) ,
        ((shift_TxVar_Tm (weakenCutoffTxVar c18 k25) (subst_TxVar_Index (weakenTrace X0 k25) s3 a19)) =
        (subst_TxVar_Index (weakenTrace X0 k25) (shift_TxVar_Tm c18 s3) (shift_TxVar_Index (weakenCutoffTxVar (CS c18) k25) a19)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TyVar_Tm_subst_TxVar_Index0_comm_ind (c18 : (Cutoff TyVar)) (s3 : Tm) :
      (forall (k25 : Hvl) (a19 : (Index TxVar)) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c18 k25) (subst_TxVar_Index (weakenTrace X0 k25) s3 a19)) =
        (subst_TxVar_Index (weakenTrace X0 k25) (shift_TyVar_Tm c18 s3) a19))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma shift_TyVar_Ty_subst_TyVar_Index0_comm_ind (c18 : (Cutoff TyVar)) (S3 : Ty) :
      (forall (k25 : Hvl) (X25 : (Index TyVar)) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar c18 k25) (subst_TyVar_Index (weakenTrace X0 k25) S3 X25)) =
        (subst_TyVar_Index (weakenTrace X0 k25) (shift_TyVar_Ty c18 S3) (shift_TyVar_Index (weakenCutoffTyVar (CS c18) k25) X25)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End ShiftSubstIndexCommInd.
  Section ShiftSubstCommInd.
    Definition shift_TyVar__subst_TyVar_0_Ty_comm_ind := (ind_Ty (fun (S10 : Ty) =>
      (forall (k44 : Hvl) (c37 : (Cutoff TyVar)) (S11 : Ty) ,
        ((shift_TyVar_Ty (weakenCutoffTyVar c37 k44) (subst_TyVar_Ty (weakenTrace X0 k44) S11 S10)) =
        (subst_TyVar_Ty (weakenTrace X0 k44) (shift_TyVar_Ty c37 S11) (shift_TyVar_Ty (weakenCutoffTyVar (CS c37) k44) S10))))) (fun (X37 : (Index TyVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TyVar)) (S10 : Ty) =>
        (shift_TyVar_Ty_subst_TyVar_Index0_comm_ind c37 S10 k44 X37))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k44 : Hvl) (c37 : (Cutoff TyVar)) (S10 : Ty) =>
      (f_equal2 TArr (IHT0 k44 c37 S10) (IHT3 k44 c37 S10))) (fun (T : Ty) IHT0 (k44 : Hvl) (c37 : (Cutoff TyVar)) (S10 : Ty) =>
      (f_equal TAll (eq_trans (f_equal2 shift_TyVar_Ty eq_refl (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k44 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHT0 (HS TyVar k44) c37 S10) (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar X0 k44 (HS TyVar H0))) eq_refl eq_refl)))))).
    Definition shift_TmVar__subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s15 : Tm) =>
      (forall (k44 : Hvl) (c37 : (Cutoff TmVar)) (s16 : Tm) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c37 k44) (subst_TmVar_Tm (weakenTrace X0 k44) s16 s15)) =
        (subst_TmVar_Tm (weakenTrace X0 k44) (shift_TmVar_Tm c37 s16) (shift_TmVar_Tm (weakenCutoffTmVar (CS c37) k44) s15))))) (fun (x58 : (Index TmVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TmVar)) (s15 : Tm) =>
        (shift_TmVar_Tm_subst_TmVar_Index0_comm_ind c37 s15 k44 x58))) (fun (a37 : (Index TxVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TmVar)) (s15 : Tm) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TmVar)) (s15 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k44 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TmVar k44) c37 s15) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k44 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt104 (t2 : Tm) IHt105 (k44 : Hvl) (c37 : (Cutoff TmVar)) (s15 : Tm) =>
      (f_equal2 App (IHt104 k44 c37 s15) (IHt105 k44 c37 s15))) (fun (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TmVar)) (s15 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k44 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TyVar k44) c37 s15) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k44 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt104 (T : Ty) (k44 : Hvl) (c37 : (Cutoff TmVar)) (s15 : Tm) =>
      (f_equal2 TApp (IHt104 k44 c37 s15) eq_refl))).
    Definition shift_TmVar__subst_TxVar_0_Tm_comm_ind := (ind_Tm (fun (s15 : Tm) =>
      (forall (k44 : Hvl) (c37 : (Cutoff TmVar)) (s16 : Tm) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c37 k44) (subst_TxVar_Tm (weakenTrace X0 k44) s16 s15)) =
        (subst_TxVar_Tm (weakenTrace X0 k44) (shift_TmVar_Tm c37 s16) (shift_TmVar_Tm (weakenCutoffTmVar c37 k44) s15))))) (fun (x58 : (Index TmVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TmVar)) (s15 : Tm) =>
        eq_refl)) (fun (a37 : (Index TxVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TmVar)) (s15 : Tm) =>
        (shift_TmVar_Tm_subst_TxVar_Index0_comm_ind c37 s15 k44 a37))) (fun (T : Ty) (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TmVar)) (s15 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar X0 k44 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TmVar k44) c37 s15) (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar X0 k44 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt104 (t2 : Tm) IHt105 (k44 : Hvl) (c37 : (Cutoff TmVar)) (s15 : Tm) =>
      (f_equal2 App (IHt104 k44 c37 s15) (IHt105 k44 c37 s15))) (fun (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TmVar)) (s15 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar X0 k44 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TyVar k44) c37 s15) (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar X0 k44 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt104 (T : Ty) (k44 : Hvl) (c37 : (Cutoff TmVar)) (s15 : Tm) =>
      (f_equal2 TApp (IHt104 k44 c37 s15) eq_refl))).
    Definition shift_TmVar__subst_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s15 : Tm) =>
      (forall (k44 : Hvl) (c37 : (Cutoff TmVar)) (S10 : Ty) ,
        ((shift_TmVar_Tm (weakenCutoffTmVar c37 k44) (subst_TyVar_Tm (weakenTrace X0 k44) S10 s15)) =
        (subst_TyVar_Tm (weakenTrace X0 k44) S10 (shift_TmVar_Tm (weakenCutoffTmVar c37 k44) s15))))) (fun (x58 : (Index TmVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TmVar)) (S10 : Ty) =>
        eq_refl)) (fun (a37 : (Index TxVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TmVar)) (S10 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TmVar)) (S10 : Ty) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k44 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TmVar k44) c37 S10) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k44 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt104 (t2 : Tm) IHt105 (k44 : Hvl) (c37 : (Cutoff TmVar)) (S10 : Ty) =>
      (f_equal2 App (IHt104 k44 c37 S10) (IHt105 k44 c37 S10))) (fun (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TmVar)) (S10 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k44 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TyVar k44) c37 S10) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k44 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt104 (T : Ty) (k44 : Hvl) (c37 : (Cutoff TmVar)) (S10 : Ty) =>
      (f_equal2 TApp (IHt104 k44 c37 S10) eq_refl))).
    Definition shift_TxVar__subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s15 : Tm) =>
      (forall (k44 : Hvl) (c37 : (Cutoff TxVar)) (s16 : Tm) ,
        ((shift_TxVar_Tm (weakenCutoffTxVar c37 k44) (subst_TmVar_Tm (weakenTrace X0 k44) s16 s15)) =
        (subst_TmVar_Tm (weakenTrace X0 k44) (shift_TxVar_Tm c37 s16) (shift_TxVar_Tm (weakenCutoffTxVar c37 k44) s15))))) (fun (x58 : (Index TmVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TxVar)) (s15 : Tm) =>
        (shift_TxVar_Tm_subst_TmVar_Index0_comm_ind c37 s15 k44 x58))) (fun (a37 : (Index TxVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TxVar)) (s15 : Tm) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TxVar)) (s15 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shift_TxVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k44 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TmVar k44) c37 s15) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k44 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt104 (t2 : Tm) IHt105 (k44 : Hvl) (c37 : (Cutoff TxVar)) (s15 : Tm) =>
      (f_equal2 App (IHt104 k44 c37 s15) (IHt105 k44 c37 s15))) (fun (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TxVar)) (s15 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal2 shift_TxVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k44 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TyVar k44) c37 s15) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k44 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt104 (T : Ty) (k44 : Hvl) (c37 : (Cutoff TxVar)) (s15 : Tm) =>
      (f_equal2 TApp (IHt104 k44 c37 s15) eq_refl))).
    Definition shift_TxVar__subst_TxVar_0_Tm_comm_ind := (ind_Tm (fun (s15 : Tm) =>
      (forall (k44 : Hvl) (c37 : (Cutoff TxVar)) (s16 : Tm) ,
        ((shift_TxVar_Tm (weakenCutoffTxVar c37 k44) (subst_TxVar_Tm (weakenTrace X0 k44) s16 s15)) =
        (subst_TxVar_Tm (weakenTrace X0 k44) (shift_TxVar_Tm c37 s16) (shift_TxVar_Tm (weakenCutoffTxVar (CS c37) k44) s15))))) (fun (x58 : (Index TmVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TxVar)) (s15 : Tm) =>
        eq_refl)) (fun (a37 : (Index TxVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TxVar)) (s15 : Tm) =>
        (shift_TxVar_Tm_subst_TxVar_Index0_comm_ind c37 s15 k44 a37))) (fun (T : Ty) (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TxVar)) (s15 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shift_TxVar_Tm eq_refl (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar X0 k44 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TmVar k44) c37 s15) (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar X0 k44 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt104 (t2 : Tm) IHt105 (k44 : Hvl) (c37 : (Cutoff TxVar)) (s15 : Tm) =>
      (f_equal2 App (IHt104 k44 c37 s15) (IHt105 k44 c37 s15))) (fun (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TxVar)) (s15 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal2 shift_TxVar_Tm eq_refl (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar X0 k44 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TyVar k44) c37 s15) (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar X0 k44 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt104 (T : Ty) (k44 : Hvl) (c37 : (Cutoff TxVar)) (s15 : Tm) =>
      (f_equal2 TApp (IHt104 k44 c37 s15) eq_refl))).
    Definition shift_TxVar__subst_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s15 : Tm) =>
      (forall (k44 : Hvl) (c37 : (Cutoff TxVar)) (S10 : Ty) ,
        ((shift_TxVar_Tm (weakenCutoffTxVar c37 k44) (subst_TyVar_Tm (weakenTrace X0 k44) S10 s15)) =
        (subst_TyVar_Tm (weakenTrace X0 k44) S10 (shift_TxVar_Tm (weakenCutoffTxVar c37 k44) s15))))) (fun (x58 : (Index TmVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TxVar)) (S10 : Ty) =>
        eq_refl)) (fun (a37 : (Index TxVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TxVar)) (S10 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TxVar)) (S10 : Ty) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shift_TxVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k44 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TmVar k44) c37 S10) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k44 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt104 (t2 : Tm) IHt105 (k44 : Hvl) (c37 : (Cutoff TxVar)) (S10 : Ty) =>
      (f_equal2 App (IHt104 k44 c37 S10) (IHt105 k44 c37 S10))) (fun (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TxVar)) (S10 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal2 shift_TxVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k44 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TyVar k44) c37 S10) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k44 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt104 (T : Ty) (k44 : Hvl) (c37 : (Cutoff TxVar)) (S10 : Ty) =>
      (f_equal2 TApp (IHt104 k44 c37 S10) eq_refl))).
    Definition shift_TyVar__subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s15 : Tm) =>
      (forall (k44 : Hvl) (c37 : (Cutoff TyVar)) (s16 : Tm) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c37 k44) (subst_TmVar_Tm (weakenTrace X0 k44) s16 s15)) =
        (subst_TmVar_Tm (weakenTrace X0 k44) (shift_TyVar_Tm c37 s16) (shift_TyVar_Tm (weakenCutoffTyVar c37 k44) s15))))) (fun (x58 : (Index TmVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TyVar)) (s15 : Tm) =>
        (shift_TyVar_Tm_subst_TmVar_Index0_comm_ind c37 s15 k44 x58))) (fun (a37 : (Index TxVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TyVar)) (s15 : Tm) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TyVar)) (s15 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k44 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TmVar k44) c37 s15) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k44 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt104 (t2 : Tm) IHt105 (k44 : Hvl) (c37 : (Cutoff TyVar)) (s15 : Tm) =>
      (f_equal2 App (IHt104 k44 c37 s15) (IHt105 k44 c37 s15))) (fun (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TyVar)) (s15 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k44 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TyVar k44) c37 s15) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k44 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt104 (T : Ty) (k44 : Hvl) (c37 : (Cutoff TyVar)) (s15 : Tm) =>
      (f_equal2 TApp (IHt104 k44 c37 s15) eq_refl))).
    Definition shift_TyVar__subst_TxVar_0_Tm_comm_ind := (ind_Tm (fun (s15 : Tm) =>
      (forall (k44 : Hvl) (c37 : (Cutoff TyVar)) (s16 : Tm) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c37 k44) (subst_TxVar_Tm (weakenTrace X0 k44) s16 s15)) =
        (subst_TxVar_Tm (weakenTrace X0 k44) (shift_TyVar_Tm c37 s16) (shift_TyVar_Tm (weakenCutoffTyVar c37 k44) s15))))) (fun (x58 : (Index TmVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TyVar)) (s15 : Tm) =>
        eq_refl)) (fun (a37 : (Index TxVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TyVar)) (s15 : Tm) =>
        (shift_TyVar_Tm_subst_TxVar_Index0_comm_ind c37 s15 k44 a37))) (fun (T : Ty) (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TyVar)) (s15 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar X0 k44 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TmVar k44) c37 s15) (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar X0 k44 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt104 (t2 : Tm) IHt105 (k44 : Hvl) (c37 : (Cutoff TyVar)) (s15 : Tm) =>
      (f_equal2 App (IHt104 k44 c37 s15) (IHt105 k44 c37 s15))) (fun (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TyVar)) (s15 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar X0 k44 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TyVar k44) c37 s15) (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar X0 k44 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt104 (T : Ty) (k44 : Hvl) (c37 : (Cutoff TyVar)) (s15 : Tm) =>
      (f_equal2 TApp (IHt104 k44 c37 s15) eq_refl))).
    Definition shift_TyVar__subst_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s15 : Tm) =>
      (forall (k44 : Hvl) (c37 : (Cutoff TyVar)) (S10 : Ty) ,
        ((shift_TyVar_Tm (weakenCutoffTyVar c37 k44) (subst_TyVar_Tm (weakenTrace X0 k44) S10 s15)) =
        (subst_TyVar_Tm (weakenTrace X0 k44) (shift_TyVar_Ty c37 S10) (shift_TyVar_Tm (weakenCutoffTyVar (CS c37) k44) s15))))) (fun (x58 : (Index TmVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TyVar)) (S10 : Ty) =>
        eq_refl)) (fun (a37 : (Index TxVar)) =>
      (fun (k44 : Hvl) (c37 : (Cutoff TyVar)) (S10 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TyVar)) (S10 : Ty) =>
      (f_equal2 Abs (let IHT0 := (shift_TyVar__subst_TyVar_0_Ty_comm_ind T) in
      (IHT0 k44 c37 S10)) (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k44 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TmVar k44) c37 S10) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k44 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt104 (t2 : Tm) IHt105 (k44 : Hvl) (c37 : (Cutoff TyVar)) (S10 : Ty) =>
      (f_equal2 App (IHt104 k44 c37 S10) (IHt105 k44 c37 S10))) (fun (t : Tm) IHt104 (k44 : Hvl) (c37 : (Cutoff TyVar)) (S10 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k44 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt104 (HS TyVar k44) c37 S10) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k44 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt104 (T : Ty) (k44 : Hvl) (c37 : (Cutoff TyVar)) (S10 : Ty) =>
      (f_equal2 TApp (IHt104 k44 c37 S10) (let IHT0 := (shift_TyVar__subst_TyVar_0_Ty_comm_ind T) in
      (IHT0 k44 c37 S10))))).
  End ShiftSubstCommInd.
  Section ShiftSubstComm.
    Definition shift_TyVar_Ty_subst_TyVar_Ty0_comm (S11 : Ty) : (forall (c37 : (Cutoff TyVar)) (S10 : Ty) ,
      ((shift_TyVar_Ty c37 (subst_TyVar_Ty X0 S10 S11)) =
      (subst_TyVar_Ty X0 (shift_TyVar_Ty c37 S10) (shift_TyVar_Ty (CS c37) S11)))) := (shift_TyVar__subst_TyVar_0_Ty_comm_ind S11 H0).
    Definition shift_TmVar_Tm_subst_TmVar_Tm0_comm (s16 : Tm) : (forall (c37 : (Cutoff TmVar)) (s15 : Tm) ,
      ((shift_TmVar_Tm c37 (subst_TmVar_Tm X0 s15 s16)) =
      (subst_TmVar_Tm X0 (shift_TmVar_Tm c37 s15) (shift_TmVar_Tm (CS c37) s16)))) := (shift_TmVar__subst_TmVar_0_Tm_comm_ind s16 H0).
    Definition shift_TmVar_Tm_subst_TxVar_Tm0_comm (s16 : Tm) : (forall (c37 : (Cutoff TmVar)) (s15 : Tm) ,
      ((shift_TmVar_Tm c37 (subst_TxVar_Tm X0 s15 s16)) =
      (subst_TxVar_Tm X0 (shift_TmVar_Tm c37 s15) (shift_TmVar_Tm c37 s16)))) := (shift_TmVar__subst_TxVar_0_Tm_comm_ind s16 H0).
    Definition shift_TmVar_Tm_subst_TyVar_Tm0_comm (s15 : Tm) : (forall (c37 : (Cutoff TmVar)) (S10 : Ty) ,
      ((shift_TmVar_Tm c37 (subst_TyVar_Tm X0 S10 s15)) =
      (subst_TyVar_Tm X0 S10 (shift_TmVar_Tm c37 s15)))) := (shift_TmVar__subst_TyVar_0_Tm_comm_ind s15 H0).
    Definition shift_TxVar_Tm_subst_TmVar_Tm0_comm (s16 : Tm) : (forall (c37 : (Cutoff TxVar)) (s15 : Tm) ,
      ((shift_TxVar_Tm c37 (subst_TmVar_Tm X0 s15 s16)) =
      (subst_TmVar_Tm X0 (shift_TxVar_Tm c37 s15) (shift_TxVar_Tm c37 s16)))) := (shift_TxVar__subst_TmVar_0_Tm_comm_ind s16 H0).
    Definition shift_TxVar_Tm_subst_TxVar_Tm0_comm (s16 : Tm) : (forall (c37 : (Cutoff TxVar)) (s15 : Tm) ,
      ((shift_TxVar_Tm c37 (subst_TxVar_Tm X0 s15 s16)) =
      (subst_TxVar_Tm X0 (shift_TxVar_Tm c37 s15) (shift_TxVar_Tm (CS c37) s16)))) := (shift_TxVar__subst_TxVar_0_Tm_comm_ind s16 H0).
    Definition shift_TxVar_Tm_subst_TyVar_Tm0_comm (s15 : Tm) : (forall (c37 : (Cutoff TxVar)) (S10 : Ty) ,
      ((shift_TxVar_Tm c37 (subst_TyVar_Tm X0 S10 s15)) =
      (subst_TyVar_Tm X0 S10 (shift_TxVar_Tm c37 s15)))) := (shift_TxVar__subst_TyVar_0_Tm_comm_ind s15 H0).
    Definition shift_TyVar_Tm_subst_TmVar_Tm0_comm (s16 : Tm) : (forall (c37 : (Cutoff TyVar)) (s15 : Tm) ,
      ((shift_TyVar_Tm c37 (subst_TmVar_Tm X0 s15 s16)) =
      (subst_TmVar_Tm X0 (shift_TyVar_Tm c37 s15) (shift_TyVar_Tm c37 s16)))) := (shift_TyVar__subst_TmVar_0_Tm_comm_ind s16 H0).
    Definition shift_TyVar_Tm_subst_TxVar_Tm0_comm (s16 : Tm) : (forall (c37 : (Cutoff TyVar)) (s15 : Tm) ,
      ((shift_TyVar_Tm c37 (subst_TxVar_Tm X0 s15 s16)) =
      (subst_TxVar_Tm X0 (shift_TyVar_Tm c37 s15) (shift_TyVar_Tm c37 s16)))) := (shift_TyVar__subst_TxVar_0_Tm_comm_ind s16 H0).
    Definition shift_TyVar_Tm_subst_TyVar_Tm0_comm (s15 : Tm) : (forall (c37 : (Cutoff TyVar)) (S10 : Ty) ,
      ((shift_TyVar_Tm c37 (subst_TyVar_Tm X0 S10 s15)) =
      (subst_TyVar_Tm X0 (shift_TyVar_Ty c37 S10) (shift_TyVar_Tm (CS c37) s15)))) := (shift_TyVar__subst_TyVar_0_Tm_comm_ind s15 H0).
  End ShiftSubstComm.
  Section SubstShiftIndexCommInd.
    Lemma subst_TmVar_Index_shift_TmVar_Tm0_comm_ind (d : (Trace TmVar)) (s15 : Tm) :
      (forall (k44 : Hvl) (x58 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace (XS TmVar d) k44) s15 (shift_TmVar_Index (weakenCutoffTmVar C0 k44) x58)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k44) (subst_TmVar_Index (weakenTrace d k44) s15 x58)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Index_shift_TxVar_Tm0_comm_ind (d : (Trace TmVar)) (s15 : Tm) :
      (forall (k44 : Hvl) (x58 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace (XS TxVar d) k44) s15 x58) =
        (shift_TxVar_Tm (weakenCutoffTxVar C0 k44) (subst_TmVar_Index (weakenTrace d k44) s15 x58)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Index_shift_TyVar_Tm0_comm_ind (d : (Trace TmVar)) (s15 : Tm) :
      (forall (k44 : Hvl) (x58 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace (XS TyVar d) k44) s15 x58) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k44) (subst_TmVar_Index (weakenTrace d k44) s15 x58)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TxVar_Index_shift_TmVar_Tm0_comm_ind (d : (Trace TxVar)) (s15 : Tm) :
      (forall (k44 : Hvl) (a37 : (Index TxVar)) ,
        ((subst_TxVar_Index (weakenTrace (XS TmVar d) k44) s15 a37) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k44) (subst_TxVar_Index (weakenTrace d k44) s15 a37)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TxVar_Index_shift_TxVar_Tm0_comm_ind (d : (Trace TxVar)) (s15 : Tm) :
      (forall (k44 : Hvl) (a37 : (Index TxVar)) ,
        ((subst_TxVar_Index (weakenTrace (XS TxVar d) k44) s15 (shift_TxVar_Index (weakenCutoffTxVar C0 k44) a37)) =
        (shift_TxVar_Tm (weakenCutoffTxVar C0 k44) (subst_TxVar_Index (weakenTrace d k44) s15 a37)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TxVar_Index_shift_TyVar_Tm0_comm_ind (d : (Trace TxVar)) (s15 : Tm) :
      (forall (k44 : Hvl) (a37 : (Index TxVar)) ,
        ((subst_TxVar_Index (weakenTrace (XS TyVar d) k44) s15 a37) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k44) (subst_TxVar_Index (weakenTrace d k44) s15 a37)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Index_shift_TmVar_Ty0_comm_ind (d : (Trace TyVar)) (S10 : Ty) :
      (forall (k44 : Hvl) (X37 : (Index TyVar)) ,
        ((subst_TyVar_Index (weakenTrace (XS TmVar d) k44) S10 X37) =
        (subst_TyVar_Index (weakenTrace d k44) S10 X37))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Index_shift_TxVar_Ty0_comm_ind (d : (Trace TyVar)) (S10 : Ty) :
      (forall (k44 : Hvl) (X37 : (Index TyVar)) ,
        ((subst_TyVar_Index (weakenTrace (XS TxVar d) k44) S10 X37) =
        (subst_TyVar_Index (weakenTrace d k44) S10 X37))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Index_shift_TyVar_Ty0_comm_ind (d : (Trace TyVar)) (S10 : Ty) :
      (forall (k44 : Hvl) (X37 : (Index TyVar)) ,
        ((subst_TyVar_Index (weakenTrace (XS TyVar d) k44) S10 (shift_TyVar_Index (weakenCutoffTyVar C0 k44) X37)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k44) (subst_TyVar_Index (weakenTrace d k44) S10 X37)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End SubstShiftIndexCommInd.
  Section SubstShiftCommInd.
    Definition subst_TyVar__shift_TyVar_0_Ty_comm_ind := (ind_Ty (fun (S31 : Ty) =>
      (forall (k101 : Hvl) (d56 : (Trace TyVar)) (S32 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace (XS TyVar d56) k101) S32 (shift_TyVar_Ty (weakenCutoffTyVar C0 k101) S31)) =
        (shift_TyVar_Ty (weakenCutoffTyVar C0 k101) (subst_TyVar_Ty (weakenTrace d56 k101) S32 S31))))) (fun (X49 : (Index TyVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
        (subst_TyVar_Index_shift_TyVar_Ty0_comm_ind d56 S31 k101 X49))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 TArr (IHT0 k101 d56 S31) (IHT3 k101 d56 S31))) (fun (T : Ty) IHT0 (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal TAll (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar (XS TyVar d56) k101 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHT0 (HS TyVar k101) d56 S31) (f_equal2 shift_TyVar_Ty eq_refl (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar d56 k101 (HS TyVar H0))) eq_refl eq_refl))))))).
    Definition subst_TmVar__shift_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s51 : Tm) =>
      (forall (k101 : Hvl) (d56 : (Trace TmVar)) (s52 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace (XS TmVar d56) k101) s52 (shift_TmVar_Tm (weakenCutoffTmVar C0 k101) s51)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k101) (subst_TmVar_Tm (weakenTrace d56 k101) s52 s51))))) (fun (x85 : (Index TmVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
        (subst_TmVar_Index_shift_TmVar_Tm0_comm_ind d56 s51 k101 x85))) (fun (a55 : (Index TxVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TmVar d56) k101 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TmVar k101) d56 s51) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d56 k101 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt149 (t2 : Tm) IHt150 (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
      (f_equal2 App (IHt149 k101 d56 s51) (IHt150 k101 d56 s51))) (fun (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TmVar d56) k101 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TyVar k101) d56 s51) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d56 k101 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt149 (T : Ty) (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
      (f_equal2 TApp (IHt149 k101 d56 s51) eq_refl))).
    Definition subst_TmVar__shift_TxVar_0_Tm_comm_ind := (ind_Tm (fun (s51 : Tm) =>
      (forall (k101 : Hvl) (d56 : (Trace TmVar)) (s52 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace (XS TxVar d56) k101) s52 (shift_TxVar_Tm (weakenCutoffTxVar C0 k101) s51)) =
        (shift_TxVar_Tm (weakenCutoffTxVar C0 k101) (subst_TmVar_Tm (weakenTrace d56 k101) s52 s51))))) (fun (x85 : (Index TmVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
        (subst_TmVar_Index_shift_TxVar_Tm0_comm_ind d56 s51 k101 x85))) (fun (a55 : (Index TxVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TxVar d56) k101 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TmVar k101) d56 s51) (f_equal2 shift_TxVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d56 k101 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt149 (t2 : Tm) IHt150 (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
      (f_equal2 App (IHt149 k101 d56 s51) (IHt150 k101 d56 s51))) (fun (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TxVar d56) k101 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TyVar k101) d56 s51) (f_equal2 shift_TxVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d56 k101 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt149 (T : Ty) (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
      (f_equal2 TApp (IHt149 k101 d56 s51) eq_refl))).
    Definition subst_TmVar__shift_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s51 : Tm) =>
      (forall (k101 : Hvl) (d56 : (Trace TmVar)) (s52 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace (XS TyVar d56) k101) s52 (shift_TyVar_Tm (weakenCutoffTyVar C0 k101) s51)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k101) (subst_TmVar_Tm (weakenTrace d56 k101) s52 s51))))) (fun (x85 : (Index TmVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
        (subst_TmVar_Index_shift_TyVar_Tm0_comm_ind d56 s51 k101 x85))) (fun (a55 : (Index TxVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TyVar d56) k101 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TmVar k101) d56 s51) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d56 k101 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt149 (t2 : Tm) IHt150 (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
      (f_equal2 App (IHt149 k101 d56 s51) (IHt150 k101 d56 s51))) (fun (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar (XS TyVar d56) k101 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TyVar k101) d56 s51) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar d56 k101 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt149 (T : Ty) (k101 : Hvl) (d56 : (Trace TmVar)) (s51 : Tm) =>
      (f_equal2 TApp (IHt149 k101 d56 s51) eq_refl))).
    Definition subst_TxVar__shift_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s51 : Tm) =>
      (forall (k101 : Hvl) (d56 : (Trace TxVar)) (s52 : Tm) ,
        ((subst_TxVar_Tm (weakenTrace (XS TmVar d56) k101) s52 (shift_TmVar_Tm (weakenCutoffTmVar C0 k101) s51)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k101) (subst_TxVar_Tm (weakenTrace d56 k101) s52 s51))))) (fun (x85 : (Index TmVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
        eq_refl)) (fun (a55 : (Index TxVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
        (subst_TxVar_Index_shift_TmVar_Tm0_comm_ind d56 s51 k101 a55))) (fun (T : Ty) (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar (XS TmVar d56) k101 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TmVar k101) d56 s51) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar d56 k101 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt149 (t2 : Tm) IHt150 (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
      (f_equal2 App (IHt149 k101 d56 s51) (IHt150 k101 d56 s51))) (fun (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar (XS TmVar d56) k101 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TyVar k101) d56 s51) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar d56 k101 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt149 (T : Ty) (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
      (f_equal2 TApp (IHt149 k101 d56 s51) eq_refl))).
    Definition subst_TxVar__shift_TxVar_0_Tm_comm_ind := (ind_Tm (fun (s51 : Tm) =>
      (forall (k101 : Hvl) (d56 : (Trace TxVar)) (s52 : Tm) ,
        ((subst_TxVar_Tm (weakenTrace (XS TxVar d56) k101) s52 (shift_TxVar_Tm (weakenCutoffTxVar C0 k101) s51)) =
        (shift_TxVar_Tm (weakenCutoffTxVar C0 k101) (subst_TxVar_Tm (weakenTrace d56 k101) s52 s51))))) (fun (x85 : (Index TmVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
        eq_refl)) (fun (a55 : (Index TxVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
        (subst_TxVar_Index_shift_TxVar_Tm0_comm_ind d56 s51 k101 a55))) (fun (T : Ty) (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar (XS TxVar d56) k101 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TmVar k101) d56 s51) (f_equal2 shift_TxVar_Tm eq_refl (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar d56 k101 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt149 (t2 : Tm) IHt150 (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
      (f_equal2 App (IHt149 k101 d56 s51) (IHt150 k101 d56 s51))) (fun (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar (XS TxVar d56) k101 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TyVar k101) d56 s51) (f_equal2 shift_TxVar_Tm eq_refl (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar d56 k101 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt149 (T : Ty) (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
      (f_equal2 TApp (IHt149 k101 d56 s51) eq_refl))).
    Definition subst_TxVar__shift_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s51 : Tm) =>
      (forall (k101 : Hvl) (d56 : (Trace TxVar)) (s52 : Tm) ,
        ((subst_TxVar_Tm (weakenTrace (XS TyVar d56) k101) s52 (shift_TyVar_Tm (weakenCutoffTyVar C0 k101) s51)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k101) (subst_TxVar_Tm (weakenTrace d56 k101) s52 s51))))) (fun (x85 : (Index TmVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
        eq_refl)) (fun (a55 : (Index TxVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
        (subst_TxVar_Index_shift_TyVar_Tm0_comm_ind d56 s51 k101 a55))) (fun (T : Ty) (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar (XS TyVar d56) k101 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TmVar k101) d56 s51) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar d56 k101 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt149 (t2 : Tm) IHt150 (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
      (f_equal2 App (IHt149 k101 d56 s51) (IHt150 k101 d56 s51))) (fun (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar (XS TyVar d56) k101 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TyVar k101) d56 s51) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar d56 k101 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt149 (T : Ty) (k101 : Hvl) (d56 : (Trace TxVar)) (s51 : Tm) =>
      (f_equal2 TApp (IHt149 k101 d56 s51) eq_refl))).
    Definition subst_TyVar__shift_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s51 : Tm) =>
      (forall (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace d56 k101) S31 (shift_TmVar_Tm (weakenCutoffTmVar C0 k101) s51)) =
        (shift_TmVar_Tm (weakenCutoffTmVar C0 k101) (subst_TyVar_Tm (weakenTrace d56 k101) S31 s51))))) (fun (x85 : (Index TmVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
        eq_refl)) (fun (a55 : (Index TxVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d56 k101 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TmVar k101) d56 S31) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d56 k101 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt149 (t2 : Tm) IHt150 (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 App (IHt149 k101 d56 S31) (IHt150 k101 d56 S31))) (fun (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d56 k101 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TyVar k101) d56 S31) (f_equal2 shift_TmVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d56 k101 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt149 (T : Ty) (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 TApp (IHt149 k101 d56 S31) eq_refl))).
    Definition subst_TyVar__shift_TxVar_0_Tm_comm_ind := (ind_Tm (fun (s51 : Tm) =>
      (forall (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace d56 k101) S31 (shift_TxVar_Tm (weakenCutoffTxVar C0 k101) s51)) =
        (shift_TxVar_Tm (weakenCutoffTxVar C0 k101) (subst_TyVar_Tm (weakenTrace d56 k101) S31 s51))))) (fun (x85 : (Index TmVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
        eq_refl)) (fun (a55 : (Index TxVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d56 k101 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TmVar k101) d56 S31) (f_equal2 shift_TxVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d56 k101 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt149 (t2 : Tm) IHt150 (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 App (IHt149 k101 d56 S31) (IHt150 k101 d56 S31))) (fun (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d56 k101 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TyVar k101) d56 S31) (f_equal2 shift_TxVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d56 k101 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt149 (T : Ty) (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 TApp (IHt149 k101 d56 S31) eq_refl))).
    Definition subst_TyVar__shift_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s51 : Tm) =>
      (forall (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace (XS TyVar d56) k101) S31 (shift_TyVar_Tm (weakenCutoffTyVar C0 k101) s51)) =
        (shift_TyVar_Tm (weakenCutoffTyVar C0 k101) (subst_TyVar_Tm (weakenTrace d56 k101) S31 s51))))) (fun (x85 : (Index TmVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
        eq_refl)) (fun (a55 : (Index TxVar)) =>
      (fun (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 Abs (let IHT0 := (subst_TyVar__shift_TyVar_0_Ty_comm_ind T) in
      (IHT0 k101 d56 S31)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TyVar d56) k101 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TmVar k101) d56 S31) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d56 k101 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt149 (t2 : Tm) IHt150 (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 App (IHt149 k101 d56 S31) (IHt150 k101 d56 S31))) (fun (t : Tm) IHt149 (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TyVar d56) k101 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt149 (HS TyVar k101) d56 S31) (f_equal2 shift_TyVar_Tm eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d56 k101 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt149 (T : Ty) (k101 : Hvl) (d56 : (Trace TyVar)) (S31 : Ty) =>
      (f_equal2 TApp (IHt149 k101 d56 S31) (let IHT0 := (subst_TyVar__shift_TyVar_0_Ty_comm_ind T) in
      (IHT0 k101 d56 S31))))).
  End SubstShiftCommInd.
  Section SubstShiftComm.
    Definition subst_TyVar_Ty_shift_TyVar_Ty0_comm (S50 : Ty) : (forall (d74 : (Trace TyVar)) (S49 : Ty) ,
      ((subst_TyVar_Ty (XS TyVar d74) S49 (shift_TyVar_Ty C0 S50)) =
      (shift_TyVar_Ty C0 (subst_TyVar_Ty d74 S49 S50)))) := (subst_TyVar__shift_TyVar_0_Ty_comm_ind S50 H0).
    Definition subst_TmVar_Tm_shift_TmVar_Tm0_comm (s52 : Tm) : (forall (d74 : (Trace TmVar)) (s51 : Tm) ,
      ((subst_TmVar_Tm (XS TmVar d74) s51 (shift_TmVar_Tm C0 s52)) =
      (shift_TmVar_Tm C0 (subst_TmVar_Tm d74 s51 s52)))) := (subst_TmVar__shift_TmVar_0_Tm_comm_ind s52 H0).
    Definition subst_TmVar_Tm_shift_TxVar_Tm0_comm (s52 : Tm) : (forall (d74 : (Trace TmVar)) (s51 : Tm) ,
      ((subst_TmVar_Tm (XS TxVar d74) s51 (shift_TxVar_Tm C0 s52)) =
      (shift_TxVar_Tm C0 (subst_TmVar_Tm d74 s51 s52)))) := (subst_TmVar__shift_TxVar_0_Tm_comm_ind s52 H0).
    Definition subst_TmVar_Tm_shift_TyVar_Tm0_comm (s52 : Tm) : (forall (d74 : (Trace TmVar)) (s51 : Tm) ,
      ((subst_TmVar_Tm (XS TyVar d74) s51 (shift_TyVar_Tm C0 s52)) =
      (shift_TyVar_Tm C0 (subst_TmVar_Tm d74 s51 s52)))) := (subst_TmVar__shift_TyVar_0_Tm_comm_ind s52 H0).
    Definition subst_TxVar_Tm_shift_TmVar_Tm0_comm (s52 : Tm) : (forall (d74 : (Trace TxVar)) (s51 : Tm) ,
      ((subst_TxVar_Tm (XS TmVar d74) s51 (shift_TmVar_Tm C0 s52)) =
      (shift_TmVar_Tm C0 (subst_TxVar_Tm d74 s51 s52)))) := (subst_TxVar__shift_TmVar_0_Tm_comm_ind s52 H0).
    Definition subst_TxVar_Tm_shift_TxVar_Tm0_comm (s52 : Tm) : (forall (d74 : (Trace TxVar)) (s51 : Tm) ,
      ((subst_TxVar_Tm (XS TxVar d74) s51 (shift_TxVar_Tm C0 s52)) =
      (shift_TxVar_Tm C0 (subst_TxVar_Tm d74 s51 s52)))) := (subst_TxVar__shift_TxVar_0_Tm_comm_ind s52 H0).
    Definition subst_TxVar_Tm_shift_TyVar_Tm0_comm (s52 : Tm) : (forall (d74 : (Trace TxVar)) (s51 : Tm) ,
      ((subst_TxVar_Tm (XS TyVar d74) s51 (shift_TyVar_Tm C0 s52)) =
      (shift_TyVar_Tm C0 (subst_TxVar_Tm d74 s51 s52)))) := (subst_TxVar__shift_TyVar_0_Tm_comm_ind s52 H0).
    Definition subst_TyVar_Tm_shift_TmVar_Tm0_comm (s51 : Tm) : (forall (d74 : (Trace TyVar)) (S49 : Ty) ,
      ((subst_TyVar_Tm d74 S49 (shift_TmVar_Tm C0 s51)) =
      (shift_TmVar_Tm C0 (subst_TyVar_Tm d74 S49 s51)))) := (subst_TyVar__shift_TmVar_0_Tm_comm_ind s51 H0).
    Definition subst_TyVar_Tm_shift_TxVar_Tm0_comm (s51 : Tm) : (forall (d74 : (Trace TyVar)) (S49 : Ty) ,
      ((subst_TyVar_Tm d74 S49 (shift_TxVar_Tm C0 s51)) =
      (shift_TxVar_Tm C0 (subst_TyVar_Tm d74 S49 s51)))) := (subst_TyVar__shift_TxVar_0_Tm_comm_ind s51 H0).
    Definition subst_TyVar_Tm_shift_TyVar_Tm0_comm (s51 : Tm) : (forall (d74 : (Trace TyVar)) (S49 : Ty) ,
      ((subst_TyVar_Tm (XS TyVar d74) S49 (shift_TyVar_Tm C0 s51)) =
      (shift_TyVar_Tm C0 (subst_TyVar_Tm d74 S49 s51)))) := (subst_TyVar__shift_TyVar_0_Tm_comm_ind s51 H0).
  End SubstShiftComm.
  Section SubstSubordInd.
    Definition subst_TyVar__TmVar_Ty_ind := (ind_Ty (fun (S49 : Ty) =>
      (forall (k119 : Hvl) (d74 : (Trace TyVar)) (S50 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace (XS TmVar d74) k119) S50 S49) =
        (subst_TyVar_Ty (weakenTrace d74 k119) S50 S49)))) (fun (X57 : (Index TyVar)) =>
      (fun (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
        (subst_TyVar_Index_shift_TmVar_Ty0_comm_ind d74 S49 k119 X57))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
      (f_equal2 TArr (IHT0 k119 d74 S49) (IHT3 k119 d74 S49))) (fun (T : Ty) IHT0 (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
      (f_equal TAll (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar (XS TmVar d74) k119 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHT0 (HS TyVar k119) d74 S49) (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar d74 k119 (HS TyVar H0))) eq_refl eq_refl)))))).
    Definition subst_TyVar__TxVar_Ty_ind := (ind_Ty (fun (S49 : Ty) =>
      (forall (k119 : Hvl) (d74 : (Trace TyVar)) (S50 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace (XS TxVar d74) k119) S50 S49) =
        (subst_TyVar_Ty (weakenTrace d74 k119) S50 S49)))) (fun (X57 : (Index TyVar)) =>
      (fun (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
        (subst_TyVar_Index_shift_TxVar_Ty0_comm_ind d74 S49 k119 X57))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
      (f_equal2 TArr (IHT0 k119 d74 S49) (IHT3 k119 d74 S49))) (fun (T : Ty) IHT0 (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
      (f_equal TAll (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar (XS TxVar d74) k119 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHT0 (HS TyVar k119) d74 S49) (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar d74 k119 (HS TyVar H0))) eq_refl eq_refl)))))).
    Definition subst_TyVar__TmVar_Tm_ind := (ind_Tm (fun (s51 : Tm) =>
      (forall (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace (XS TmVar d74) k119) S49 s51) =
        (subst_TyVar_Tm (weakenTrace d74 k119) S49 s51)))) (fun (x91 : (Index TmVar)) =>
      (fun (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
        eq_refl)) (fun (a59 : (Index TxVar)) =>
      (fun (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt159 (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
      (f_equal2 Abs (let IHT0 := (subst_TyVar__TmVar_Ty_ind T) in
      (IHT0 k119 d74 S49)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TmVar d74) k119 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt159 (HS TmVar k119) d74 S49) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d74 k119 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt159 (t2 : Tm) IHt160 (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
      (f_equal2 App (IHt159 k119 d74 S49) (IHt160 k119 d74 S49))) (fun (t : Tm) IHt159 (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TmVar d74) k119 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt159 (HS TyVar k119) d74 S49) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d74 k119 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt159 (T : Ty) (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
      (f_equal2 TApp (IHt159 k119 d74 S49) (let IHT0 := (subst_TyVar__TmVar_Ty_ind T) in
      (IHT0 k119 d74 S49))))).
    Definition subst_TyVar__TxVar_Tm_ind := (ind_Tm (fun (s51 : Tm) =>
      (forall (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace (XS TxVar d74) k119) S49 s51) =
        (subst_TyVar_Tm (weakenTrace d74 k119) S49 s51)))) (fun (x91 : (Index TmVar)) =>
      (fun (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
        eq_refl)) (fun (a59 : (Index TxVar)) =>
      (fun (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt159 (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
      (f_equal2 Abs (let IHT0 := (subst_TyVar__TxVar_Ty_ind T) in
      (IHT0 k119 d74 S49)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TxVar d74) k119 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt159 (HS TmVar k119) d74 S49) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d74 k119 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt159 (t2 : Tm) IHt160 (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
      (f_equal2 App (IHt159 k119 d74 S49) (IHt160 k119 d74 S49))) (fun (t : Tm) IHt159 (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar (XS TxVar d74) k119 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt159 (HS TyVar k119) d74 S49) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d74 k119 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt159 (T : Ty) (k119 : Hvl) (d74 : (Trace TyVar)) (S49 : Ty) =>
      (f_equal2 TApp (IHt159 k119 d74 S49) (let IHT0 := (subst_TyVar__TxVar_Ty_ind T) in
      (IHT0 k119 d74 S49))))).
  End SubstSubordInd.
  Section SubstSubord.
    Definition subst_TyVar_Ty_TmVar (S50 : Ty) : (forall (d74 : (Trace TyVar)) (S49 : Ty) ,
      ((subst_TyVar_Ty (XS TmVar d74) S49 S50) =
      (subst_TyVar_Ty d74 S49 S50))) := (subst_TyVar__TmVar_Ty_ind S50 H0).
    Definition subst_TyVar_Ty_TxVar (S50 : Ty) : (forall (d74 : (Trace TyVar)) (S49 : Ty) ,
      ((subst_TyVar_Ty (XS TxVar d74) S49 S50) =
      (subst_TyVar_Ty d74 S49 S50))) := (subst_TyVar__TxVar_Ty_ind S50 H0).
    Definition subst_TyVar_Tm_TmVar (s51 : Tm) : (forall (d74 : (Trace TyVar)) (S49 : Ty) ,
      ((subst_TyVar_Tm (XS TmVar d74) S49 s51) =
      (subst_TyVar_Tm d74 S49 s51))) := (subst_TyVar__TmVar_Tm_ind s51 H0).
    Definition subst_TyVar_Tm_TxVar (s51 : Tm) : (forall (d74 : (Trace TyVar)) (S49 : Ty) ,
      ((subst_TyVar_Tm (XS TxVar d74) S49 s51) =
      (subst_TyVar_Tm d74 S49 s51))) := (subst_TyVar__TxVar_Tm_ind s51 H0).
  End SubstSubord.
End ShiftSubstInteraction.
 Hint Rewrite subst_TmVar_Tm0_shift_TmVar_Tm0_reflection subst_TxVar_Tm0_shift_TxVar_Tm0_reflection subst_TyVar_Tm0_shift_TyVar_Tm0_reflection subst_TyVar_Ty0_shift_TyVar_Ty0_reflection : interaction.
 Hint Rewrite subst_TmVar_Tm0_shift_TmVar_Tm0_reflection subst_TxVar_Tm0_shift_TxVar_Tm0_reflection subst_TyVar_Tm0_shift_TyVar_Tm0_reflection subst_TyVar_Ty0_shift_TyVar_Ty0_reflection : reflection.
 Hint Rewrite subst_TmVar_Tm_shift_TmVar_Tm0_comm subst_TmVar_Tm_shift_TxVar_Tm0_comm subst_TmVar_Tm_shift_TyVar_Tm0_comm subst_TxVar_Tm_shift_TmVar_Tm0_comm subst_TxVar_Tm_shift_TxVar_Tm0_comm subst_TxVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Tm_shift_TmVar_Tm0_comm subst_TyVar_Tm_shift_TxVar_Tm0_comm subst_TyVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Ty_shift_TyVar_Ty0_comm : interaction.
 Hint Rewrite subst_TmVar_Tm_shift_TmVar_Tm0_comm subst_TmVar_Tm_shift_TxVar_Tm0_comm subst_TmVar_Tm_shift_TyVar_Tm0_comm subst_TxVar_Tm_shift_TmVar_Tm0_comm subst_TxVar_Tm_shift_TxVar_Tm0_comm subst_TxVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Tm_shift_TmVar_Tm0_comm subst_TyVar_Tm_shift_TxVar_Tm0_comm subst_TyVar_Tm_shift_TyVar_Tm0_comm subst_TyVar_Ty_shift_TyVar_Ty0_comm : subst_shift0.
 Hint Rewrite subst_TyVar_Tm_TmVar subst_TyVar_Tm_TxVar subst_TyVar_Ty_TmVar subst_TyVar_Ty_TxVar : interaction.
 Hint Rewrite subst_TyVar_Tm_TmVar subst_TyVar_Tm_TxVar subst_TyVar_Ty_TmVar subst_TyVar_Ty_TxVar : subst_shift0.
 Hint Rewrite shift_TmVar_Tm_subst_TmVar_Tm0_comm shift_TmVar_Tm_subst_TxVar_Tm0_comm shift_TmVar_Tm_subst_TyVar_Tm0_comm shift_TxVar_Tm_subst_TmVar_Tm0_comm shift_TxVar_Tm_subst_TxVar_Tm0_comm shift_TxVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Tm_subst_TmVar_Tm0_comm shift_TyVar_Tm_subst_TxVar_Tm0_comm shift_TyVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Ty_subst_TyVar_Ty0_comm : interaction.
 Hint Rewrite shift_TmVar_Tm_subst_TmVar_Tm0_comm shift_TmVar_Tm_subst_TxVar_Tm0_comm shift_TmVar_Tm_subst_TyVar_Tm0_comm shift_TxVar_Tm_subst_TmVar_Tm0_comm shift_TxVar_Tm_subst_TxVar_Tm0_comm shift_TxVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Tm_subst_TmVar_Tm0_comm shift_TyVar_Tm_subst_TxVar_Tm0_comm shift_TyVar_Tm_subst_TyVar_Tm0_comm shift_TyVar_Ty_subst_TyVar_Ty0_comm : shift_subst0.
Section SubstSubstInteraction.
  Section SubstSubstIndexCommInd.
    Lemma subst_TmVar_Tm_subst_TmVar_Index0_commright_ind (d74 : (Trace TmVar)) (s51 : Tm) (s52 : Tm) :
      (forall (k119 : Hvl) (x91 : (Index TmVar)) ,
        ((subst_TmVar_Tm (weakenTrace d74 k119) s51 (subst_TmVar_Index (weakenTrace X0 k119) s52 x91)) =
        (subst_TmVar_Tm (weakenTrace X0 k119) (subst_TmVar_Tm d74 s51 s52) (subst_TmVar_Index (weakenTrace (XS TmVar d74) k119) s51 x91)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TxVar_Tm_subst_TmVar_Index0_commright_ind (d74 : (Trace TxVar)) (s51 : Tm) (s52 : Tm) :
      (forall (k119 : Hvl) (x91 : (Index TmVar)) ,
        ((subst_TxVar_Tm (weakenTrace d74 k119) s51 (subst_TmVar_Index (weakenTrace X0 k119) s52 x91)) =
        (subst_TmVar_Index (weakenTrace X0 k119) (subst_TxVar_Tm d74 s51 s52) x91))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Tm_subst_TmVar_Index0_commright_ind (d74 : (Trace TyVar)) (S49 : Ty) (s51 : Tm) :
      (forall (k119 : Hvl) (x91 : (Index TmVar)) ,
        ((subst_TyVar_Tm (weakenTrace d74 k119) S49 (subst_TmVar_Index (weakenTrace X0 k119) s51 x91)) =
        (subst_TmVar_Index (weakenTrace X0 k119) (subst_TyVar_Tm d74 S49 s51) x91))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Tm_subst_TxVar_Index0_commright_ind (d74 : (Trace TmVar)) (s51 : Tm) (s52 : Tm) :
      (forall (k119 : Hvl) (a59 : (Index TxVar)) ,
        ((subst_TmVar_Tm (weakenTrace d74 k119) s51 (subst_TxVar_Index (weakenTrace X0 k119) s52 a59)) =
        (subst_TxVar_Index (weakenTrace X0 k119) (subst_TmVar_Tm d74 s51 s52) a59))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TxVar_Tm_subst_TxVar_Index0_commright_ind (d74 : (Trace TxVar)) (s51 : Tm) (s52 : Tm) :
      (forall (k119 : Hvl) (a59 : (Index TxVar)) ,
        ((subst_TxVar_Tm (weakenTrace d74 k119) s51 (subst_TxVar_Index (weakenTrace X0 k119) s52 a59)) =
        (subst_TxVar_Tm (weakenTrace X0 k119) (subst_TxVar_Tm d74 s51 s52) (subst_TxVar_Index (weakenTrace (XS TxVar d74) k119) s51 a59)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Tm_subst_TxVar_Index0_commright_ind (d74 : (Trace TyVar)) (S49 : Ty) (s51 : Tm) :
      (forall (k119 : Hvl) (a59 : (Index TxVar)) ,
        ((subst_TyVar_Tm (weakenTrace d74 k119) S49 (subst_TxVar_Index (weakenTrace X0 k119) s51 a59)) =
        (subst_TxVar_Index (weakenTrace X0 k119) (subst_TyVar_Tm d74 S49 s51) a59))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TyVar_Ty_subst_TyVar_Index0_commright_ind (d74 : (Trace TyVar)) (S49 : Ty) (S50 : Ty) :
      (forall (k119 : Hvl) (X57 : (Index TyVar)) ,
        ((subst_TyVar_Ty (weakenTrace d74 k119) S49 (subst_TyVar_Index (weakenTrace X0 k119) S50 X57)) =
        (subst_TyVar_Ty (weakenTrace X0 k119) (subst_TyVar_Ty d74 S49 S50) (subst_TyVar_Index (weakenTrace (XS TyVar d74) k119) S49 X57)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Tm_subst_TxVar_Index0_commleft_ind (d74 : (Trace TmVar)) (s51 : Tm) (s52 : Tm) :
      (forall (k119 : Hvl) (x91 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace d74 k119) s51 x91) =
        (subst_TxVar_Tm (weakenTrace X0 k119) (subst_TmVar_Tm d74 s51 s52) (subst_TmVar_Index (weakenTrace (XS TxVar d74) k119) s51 x91)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TmVar_Ty_subst_TyVar_Index0_commleft_ind (d74 : (Trace TmVar)) (s51 : Tm) (S49 : Ty) :
      (forall (k119 : Hvl) (x91 : (Index TmVar)) ,
        ((subst_TmVar_Index (weakenTrace d74 k119) s51 x91) =
        (subst_TyVar_Tm (weakenTrace X0 k119) S49 (subst_TmVar_Index (weakenTrace (XS TyVar d74) k119) s51 x91)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TxVar_Tm_subst_TmVar_Index0_commleft_ind (d74 : (Trace TxVar)) (s51 : Tm) (s52 : Tm) :
      (forall (k119 : Hvl) (a59 : (Index TxVar)) ,
        ((subst_TxVar_Index (weakenTrace d74 k119) s51 a59) =
        (subst_TmVar_Tm (weakenTrace X0 k119) (subst_TxVar_Tm d74 s51 s52) (subst_TxVar_Index (weakenTrace (XS TmVar d74) k119) s51 a59)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma subst_TxVar_Ty_subst_TyVar_Index0_commleft_ind (d74 : (Trace TxVar)) (s51 : Tm) (S49 : Ty) :
      (forall (k119 : Hvl) (a59 : (Index TxVar)) ,
        ((subst_TxVar_Index (weakenTrace d74 k119) s51 a59) =
        (subst_TyVar_Tm (weakenTrace X0 k119) S49 (subst_TxVar_Index (weakenTrace (XS TyVar d74) k119) s51 a59)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
  End SubstSubstIndexCommInd.
  Section SubstSubstCommInd.
    Definition subst_TyVar__subst_TyVar_0_Ty_comm_ind := (ind_Ty (fun (S55 : Ty) =>
      (forall (k126 : Hvl) (d81 : (Trace TyVar)) (S56 : Ty) (S57 : Ty) ,
        ((subst_TyVar_Ty (weakenTrace d81 k126) S56 (subst_TyVar_Ty (weakenTrace X0 k126) S57 S55)) =
        (subst_TyVar_Ty (weakenTrace X0 k126) (subst_TyVar_Ty d81 S56 S57) (subst_TyVar_Ty (weakenTrace (XS TyVar d81) k126) S56 S55))))) (fun (X63 : (Index TyVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (S56 : Ty) =>
        (subst_TyVar_Ty_subst_TyVar_Index0_commright_ind d81 S55 S56 k126 X63))) (fun (T1 : Ty) IHT0 (T2 : Ty) IHT3 (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (S56 : Ty) =>
      (f_equal2 TArr (IHT0 k126 d81 S55 S56) (IHT3 k126 d81 S55 S56))) (fun (T : Ty) IHT0 (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (S56 : Ty) =>
      (f_equal TAll (eq_trans (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar d81 k126 (HS TyVar H0)) eq_refl (f_equal3 subst_TyVar_Ty (weakenTrace_append TyVar X0 k126 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHT0 (HS TyVar k126) d81 S55 S56) (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar X0 k126 (HS TyVar H0))) eq_refl (f_equal3 subst_TyVar_Ty (eq_sym (weakenTrace_append TyVar (XS TyVar d81) k126 (HS TyVar H0))) eq_refl eq_refl))))))).
    Definition subst_TmVar__subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s59 : Tm) =>
      (forall (k126 : Hvl) (d81 : (Trace TmVar)) (s60 : Tm) (s61 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace d81 k126) s60 (subst_TmVar_Tm (weakenTrace X0 k126) s61 s59)) =
        (subst_TmVar_Tm (weakenTrace X0 k126) (subst_TmVar_Tm d81 s60 s61) (subst_TmVar_Tm (weakenTrace (XS TmVar d81) k126) s60 s59))))) (fun (x100 : (Index TmVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (s60 : Tm) =>
        (subst_TmVar_Tm_subst_TmVar_Index0_commright_ind d81 s59 s60 k126 x100))) (fun (a65 : (Index TxVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (s60 : Tm) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d81 k126 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k126 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TmVar k126) d81 s59 s60) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k126 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TmVar d81) k126 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt174 (t2 : Tm) IHt175 (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal2 App (IHt174 k126 d81 s59 s60) (IHt175 k126 d81 s59 s60))) (fun (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d81 k126 (HS TyVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k126 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TyVar k126) d81 s59 s60) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k126 (HS TyVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TmVar d81) k126 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt174 (T : Ty) (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal2 TApp (IHt174 k126 d81 s59 s60) eq_refl))).
    Definition subst_TmVar__subst_TxVar_0_Tm_comm_ind := (ind_Tm (fun (s59 : Tm) =>
      (forall (k126 : Hvl) (d81 : (Trace TmVar)) (s60 : Tm) (s61 : Tm) ,
        ((subst_TmVar_Tm (weakenTrace d81 k126) s60 (subst_TxVar_Tm (weakenTrace X0 k126) s61 s59)) =
        (subst_TxVar_Tm (weakenTrace X0 k126) (subst_TmVar_Tm d81 s60 s61) (subst_TmVar_Tm (weakenTrace (XS TxVar d81) k126) s60 s59))))) (fun (x100 : (Index TmVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (s60 : Tm) =>
        (subst_TmVar_Tm_subst_TxVar_Index0_commleft_ind d81 s59 s60 k126 x100))) (fun (a65 : (Index TxVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (s60 : Tm) =>
        (subst_TmVar_Tm_subst_TxVar_Index0_commright_ind d81 s59 s60 k126 a65))) (fun (T : Ty) (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d81 k126 (HS TmVar H0)) eq_refl (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar X0 k126 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TmVar k126) d81 s59 s60) (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar X0 k126 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TxVar d81) k126 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt174 (t2 : Tm) IHt175 (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal2 App (IHt174 k126 d81 s59 s60) (IHt175 k126 d81 s59 s60))) (fun (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d81 k126 (HS TyVar H0)) eq_refl (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar X0 k126 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TyVar k126) d81 s59 s60) (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar X0 k126 (HS TyVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TxVar d81) k126 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt174 (T : Ty) (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal2 TApp (IHt174 k126 d81 s59 s60) eq_refl))).
    Definition subst_TmVar__subst_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s59 : Tm) =>
      (forall (k126 : Hvl) (d81 : (Trace TmVar)) (s60 : Tm) (S55 : Ty) ,
        ((subst_TmVar_Tm (weakenTrace d81 k126) s60 (subst_TyVar_Tm (weakenTrace X0 k126) S55 s59)) =
        (subst_TyVar_Tm (weakenTrace X0 k126) S55 (subst_TmVar_Tm (weakenTrace (XS TyVar d81) k126) s60 s59))))) (fun (x100 : (Index TmVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (S55 : Ty) =>
        (subst_TmVar_Ty_subst_TyVar_Index0_commleft_ind d81 s59 S55 k126 x100))) (fun (a65 : (Index TxVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (S55 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (S55 : Ty) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d81 k126 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k126 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TmVar k126) d81 s59 S55) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k126 (HS TmVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TyVar d81) k126 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt174 (t2 : Tm) IHt175 (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (S55 : Ty) =>
      (f_equal2 App (IHt174 k126 d81 s59 S55) (IHt175 k126 d81 s59 S55))) (fun (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (S55 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar d81 k126 (HS TyVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k126 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TyVar k126) d81 s59 S55) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k126 (HS TyVar H0))) eq_refl (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar (XS TyVar d81) k126 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt174 (T : Ty) (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (S55 : Ty) =>
      (f_equal2 TApp (IHt174 k126 d81 s59 S55) eq_refl))).
    Definition subst_TxVar__subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s59 : Tm) =>
      (forall (k126 : Hvl) (d81 : (Trace TxVar)) (s60 : Tm) (s61 : Tm) ,
        ((subst_TxVar_Tm (weakenTrace d81 k126) s60 (subst_TmVar_Tm (weakenTrace X0 k126) s61 s59)) =
        (subst_TmVar_Tm (weakenTrace X0 k126) (subst_TxVar_Tm d81 s60 s61) (subst_TxVar_Tm (weakenTrace (XS TmVar d81) k126) s60 s59))))) (fun (x100 : (Index TmVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (s60 : Tm) =>
        (subst_TxVar_Tm_subst_TmVar_Index0_commright_ind d81 s59 s60 k126 x100))) (fun (a65 : (Index TxVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (s60 : Tm) =>
        (subst_TxVar_Tm_subst_TmVar_Index0_commleft_ind d81 s59 s60 k126 a65))) (fun (T : Ty) (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar d81 k126 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k126 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TmVar k126) d81 s59 s60) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k126 (HS TmVar H0))) eq_refl (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar (XS TmVar d81) k126 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt174 (t2 : Tm) IHt175 (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal2 App (IHt174 k126 d81 s59 s60) (IHt175 k126 d81 s59 s60))) (fun (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar d81 k126 (HS TyVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k126 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TyVar k126) d81 s59 s60) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k126 (HS TyVar H0))) eq_refl (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar (XS TmVar d81) k126 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt174 (T : Ty) (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal2 TApp (IHt174 k126 d81 s59 s60) eq_refl))).
    Definition subst_TxVar__subst_TxVar_0_Tm_comm_ind := (ind_Tm (fun (s59 : Tm) =>
      (forall (k126 : Hvl) (d81 : (Trace TxVar)) (s60 : Tm) (s61 : Tm) ,
        ((subst_TxVar_Tm (weakenTrace d81 k126) s60 (subst_TxVar_Tm (weakenTrace X0 k126) s61 s59)) =
        (subst_TxVar_Tm (weakenTrace X0 k126) (subst_TxVar_Tm d81 s60 s61) (subst_TxVar_Tm (weakenTrace (XS TxVar d81) k126) s60 s59))))) (fun (x100 : (Index TmVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (s60 : Tm) =>
        eq_refl)) (fun (a65 : (Index TxVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (s60 : Tm) =>
        (subst_TxVar_Tm_subst_TxVar_Index0_commright_ind d81 s59 s60 k126 a65))) (fun (T : Ty) (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar d81 k126 (HS TmVar H0)) eq_refl (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar X0 k126 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TmVar k126) d81 s59 s60) (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar X0 k126 (HS TmVar H0))) eq_refl (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar (XS TxVar d81) k126 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt174 (t2 : Tm) IHt175 (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal2 App (IHt174 k126 d81 s59 s60) (IHt175 k126 d81 s59 s60))) (fun (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar d81 k126 (HS TyVar H0)) eq_refl (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar X0 k126 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TyVar k126) d81 s59 s60) (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar X0 k126 (HS TyVar H0))) eq_refl (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar (XS TxVar d81) k126 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt174 (T : Ty) (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (s60 : Tm) =>
      (f_equal2 TApp (IHt174 k126 d81 s59 s60) eq_refl))).
    Definition subst_TxVar__subst_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s59 : Tm) =>
      (forall (k126 : Hvl) (d81 : (Trace TxVar)) (s60 : Tm) (S55 : Ty) ,
        ((subst_TxVar_Tm (weakenTrace d81 k126) s60 (subst_TyVar_Tm (weakenTrace X0 k126) S55 s59)) =
        (subst_TyVar_Tm (weakenTrace X0 k126) S55 (subst_TxVar_Tm (weakenTrace (XS TyVar d81) k126) s60 s59))))) (fun (x100 : (Index TmVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (S55 : Ty) =>
        eq_refl)) (fun (a65 : (Index TxVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (S55 : Ty) =>
        (subst_TxVar_Ty_subst_TyVar_Index0_commleft_ind d81 s59 S55 k126 a65))) (fun (T : Ty) (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (S55 : Ty) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar d81 k126 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k126 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TmVar k126) d81 s59 S55) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k126 (HS TmVar H0))) eq_refl (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar (XS TyVar d81) k126 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt174 (t2 : Tm) IHt175 (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (S55 : Ty) =>
      (f_equal2 App (IHt174 k126 d81 s59 S55) (IHt175 k126 d81 s59 S55))) (fun (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (S55 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar d81 k126 (HS TyVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k126 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TyVar k126) d81 s59 S55) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k126 (HS TyVar H0))) eq_refl (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar (XS TyVar d81) k126 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt174 (T : Ty) (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (S55 : Ty) =>
      (f_equal2 TApp (IHt174 k126 d81 s59 S55) eq_refl))).
    Definition subst_TyVar__subst_TmVar_0_Tm_comm_ind := (ind_Tm (fun (s59 : Tm) =>
      (forall (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (s60 : Tm) ,
        ((subst_TyVar_Tm (weakenTrace d81 k126) S55 (subst_TmVar_Tm (weakenTrace X0 k126) s60 s59)) =
        (subst_TmVar_Tm (weakenTrace X0 k126) (subst_TyVar_Tm d81 S55 s60) (subst_TyVar_Tm (weakenTrace d81 k126) S55 s59))))) (fun (x100 : (Index TmVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (s59 : Tm) =>
        (subst_TyVar_Tm_subst_TmVar_Index0_commright_ind d81 S55 s59 k126 x100))) (fun (a65 : (Index TxVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (s59 : Tm) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (s59 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d81 k126 (HS TmVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k126 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TmVar k126) d81 S55 s59) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k126 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d81 k126 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt174 (t2 : Tm) IHt175 (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (s59 : Tm) =>
      (f_equal2 App (IHt174 k126 d81 S55 s59) (IHt175 k126 d81 S55 s59))) (fun (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (s59 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d81 k126 (HS TyVar H0)) eq_refl (f_equal3 subst_TmVar_Tm (weakenTrace_append TmVar X0 k126 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TyVar k126) d81 S55 s59) (f_equal3 subst_TmVar_Tm (eq_sym (weakenTrace_append TmVar X0 k126 (HS TyVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d81 k126 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt174 (T : Ty) (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (s59 : Tm) =>
      (f_equal2 TApp (IHt174 k126 d81 S55 s59) eq_refl))).
    Definition subst_TyVar__subst_TxVar_0_Tm_comm_ind := (ind_Tm (fun (s59 : Tm) =>
      (forall (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (s60 : Tm) ,
        ((subst_TyVar_Tm (weakenTrace d81 k126) S55 (subst_TxVar_Tm (weakenTrace X0 k126) s60 s59)) =
        (subst_TxVar_Tm (weakenTrace X0 k126) (subst_TyVar_Tm d81 S55 s60) (subst_TyVar_Tm (weakenTrace d81 k126) S55 s59))))) (fun (x100 : (Index TmVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (s59 : Tm) =>
        eq_refl)) (fun (a65 : (Index TxVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (s59 : Tm) =>
        (subst_TyVar_Tm_subst_TxVar_Index0_commright_ind d81 S55 s59 k126 a65))) (fun (T : Ty) (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (s59 : Tm) =>
      (f_equal2 Abs eq_refl (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d81 k126 (HS TmVar H0)) eq_refl (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar X0 k126 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TmVar k126) d81 S55 s59) (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar X0 k126 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d81 k126 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt174 (t2 : Tm) IHt175 (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (s59 : Tm) =>
      (f_equal2 App (IHt174 k126 d81 S55 s59) (IHt175 k126 d81 S55 s59))) (fun (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (s59 : Tm) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d81 k126 (HS TyVar H0)) eq_refl (f_equal3 subst_TxVar_Tm (weakenTrace_append TxVar X0 k126 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TyVar k126) d81 S55 s59) (f_equal3 subst_TxVar_Tm (eq_sym (weakenTrace_append TxVar X0 k126 (HS TyVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar d81 k126 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt174 (T : Ty) (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (s59 : Tm) =>
      (f_equal2 TApp (IHt174 k126 d81 S55 s59) eq_refl))).
    Definition subst_TyVar__subst_TyVar_0_Tm_comm_ind := (ind_Tm (fun (s59 : Tm) =>
      (forall (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (S56 : Ty) ,
        ((subst_TyVar_Tm (weakenTrace d81 k126) S55 (subst_TyVar_Tm (weakenTrace X0 k126) S56 s59)) =
        (subst_TyVar_Tm (weakenTrace X0 k126) (subst_TyVar_Ty d81 S55 S56) (subst_TyVar_Tm (weakenTrace (XS TyVar d81) k126) S55 s59))))) (fun (x100 : (Index TmVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (S56 : Ty) =>
        eq_refl)) (fun (a65 : (Index TxVar)) =>
      (fun (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (S56 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (S56 : Ty) =>
      (f_equal2 Abs (let IHT0 := (subst_TyVar__subst_TyVar_0_Ty_comm_ind T) in
      (IHT0 k126 d81 S55 S56)) (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d81 k126 (HS TmVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k126 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TmVar k126) d81 S55 S56) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k126 (HS TmVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar (XS TyVar d81) k126 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt174 (t2 : Tm) IHt175 (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (S56 : Ty) =>
      (f_equal2 App (IHt174 k126 d81 S55 S56) (IHt175 k126 d81 S55 S56))) (fun (t : Tm) IHt174 (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (S56 : Ty) =>
      (f_equal TAbs (eq_trans (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar d81 k126 (HS TyVar H0)) eq_refl (f_equal3 subst_TyVar_Tm (weakenTrace_append TyVar X0 k126 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt174 (HS TyVar k126) d81 S55 S56) (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar X0 k126 (HS TyVar H0))) eq_refl (f_equal3 subst_TyVar_Tm (eq_sym (weakenTrace_append TyVar (XS TyVar d81) k126 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt174 (T : Ty) (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (S56 : Ty) =>
      (f_equal2 TApp (IHt174 k126 d81 S55 S56) (let IHT0 := (subst_TyVar__subst_TyVar_0_Ty_comm_ind T) in
      (IHT0 k126 d81 S55 S56))))).
  End SubstSubstCommInd.
  Section SubstSubstComm.
    Definition subst_TyVar_Ty_subst_TyVar_Ty0_comm (S57 : Ty) : (forall (d81 : (Trace TyVar)) (S55 : Ty) (S56 : Ty) ,
      ((subst_TyVar_Ty d81 S55 (subst_TyVar_Ty X0 S56 S57)) =
      (subst_TyVar_Ty X0 (subst_TyVar_Ty d81 S55 S56) (subst_TyVar_Ty (XS TyVar d81) S55 S57)))) := (subst_TyVar__subst_TyVar_0_Ty_comm_ind S57 H0).
    Definition subst_TmVar_Tm_subst_TmVar_Tm0_comm (s61 : Tm) : (forall (d81 : (Trace TmVar)) (s59 : Tm) (s60 : Tm) ,
      ((subst_TmVar_Tm d81 s59 (subst_TmVar_Tm X0 s60 s61)) =
      (subst_TmVar_Tm X0 (subst_TmVar_Tm d81 s59 s60) (subst_TmVar_Tm (XS TmVar d81) s59 s61)))) := (subst_TmVar__subst_TmVar_0_Tm_comm_ind s61 H0).
    Definition subst_TmVar_Tm_subst_TxVar_Tm0_comm (s61 : Tm) : (forall (d81 : (Trace TmVar)) (s59 : Tm) (s60 : Tm) ,
      ((subst_TmVar_Tm d81 s59 (subst_TxVar_Tm X0 s60 s61)) =
      (subst_TxVar_Tm X0 (subst_TmVar_Tm d81 s59 s60) (subst_TmVar_Tm (XS TxVar d81) s59 s61)))) := (subst_TmVar__subst_TxVar_0_Tm_comm_ind s61 H0).
    Definition subst_TmVar_Tm_subst_TyVar_Tm0_comm (s60 : Tm) : (forall (d81 : (Trace TmVar)) (s59 : Tm) (S55 : Ty) ,
      ((subst_TmVar_Tm d81 s59 (subst_TyVar_Tm X0 S55 s60)) =
      (subst_TyVar_Tm X0 S55 (subst_TmVar_Tm (XS TyVar d81) s59 s60)))) := (subst_TmVar__subst_TyVar_0_Tm_comm_ind s60 H0).
    Definition subst_TxVar_Tm_subst_TmVar_Tm0_comm (s61 : Tm) : (forall (d81 : (Trace TxVar)) (s59 : Tm) (s60 : Tm) ,
      ((subst_TxVar_Tm d81 s59 (subst_TmVar_Tm X0 s60 s61)) =
      (subst_TmVar_Tm X0 (subst_TxVar_Tm d81 s59 s60) (subst_TxVar_Tm (XS TmVar d81) s59 s61)))) := (subst_TxVar__subst_TmVar_0_Tm_comm_ind s61 H0).
    Definition subst_TxVar_Tm_subst_TxVar_Tm0_comm (s61 : Tm) : (forall (d81 : (Trace TxVar)) (s59 : Tm) (s60 : Tm) ,
      ((subst_TxVar_Tm d81 s59 (subst_TxVar_Tm X0 s60 s61)) =
      (subst_TxVar_Tm X0 (subst_TxVar_Tm d81 s59 s60) (subst_TxVar_Tm (XS TxVar d81) s59 s61)))) := (subst_TxVar__subst_TxVar_0_Tm_comm_ind s61 H0).
    Definition subst_TxVar_Tm_subst_TyVar_Tm0_comm (s60 : Tm) : (forall (d81 : (Trace TxVar)) (s59 : Tm) (S55 : Ty) ,
      ((subst_TxVar_Tm d81 s59 (subst_TyVar_Tm X0 S55 s60)) =
      (subst_TyVar_Tm X0 S55 (subst_TxVar_Tm (XS TyVar d81) s59 s60)))) := (subst_TxVar__subst_TyVar_0_Tm_comm_ind s60 H0).
    Definition subst_TyVar_Tm_subst_TmVar_Tm0_comm (s60 : Tm) : (forall (d81 : (Trace TyVar)) (S55 : Ty) (s59 : Tm) ,
      ((subst_TyVar_Tm d81 S55 (subst_TmVar_Tm X0 s59 s60)) =
      (subst_TmVar_Tm X0 (subst_TyVar_Tm d81 S55 s59) (subst_TyVar_Tm d81 S55 s60)))) := (subst_TyVar__subst_TmVar_0_Tm_comm_ind s60 H0).
    Definition subst_TyVar_Tm_subst_TxVar_Tm0_comm (s60 : Tm) : (forall (d81 : (Trace TyVar)) (S55 : Ty) (s59 : Tm) ,
      ((subst_TyVar_Tm d81 S55 (subst_TxVar_Tm X0 s59 s60)) =
      (subst_TxVar_Tm X0 (subst_TyVar_Tm d81 S55 s59) (subst_TyVar_Tm d81 S55 s60)))) := (subst_TyVar__subst_TxVar_0_Tm_comm_ind s60 H0).
    Definition subst_TyVar_Tm_subst_TyVar_Tm0_comm (s59 : Tm) : (forall (d81 : (Trace TyVar)) (S55 : Ty) (S56 : Ty) ,
      ((subst_TyVar_Tm d81 S55 (subst_TyVar_Tm X0 S56 s59)) =
      (subst_TyVar_Tm X0 (subst_TyVar_Ty d81 S55 S56) (subst_TyVar_Tm (XS TyVar d81) S55 s59)))) := (subst_TyVar__subst_TyVar_0_Tm_comm_ind s59 H0).
  End SubstSubstComm.
  Section WeakenSubst.
    Lemma weakenTy_subst_TyVar_Ty  :
      (forall (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (S56 : Ty) ,
        ((weakenTy (subst_TyVar_Ty d81 S55 S56) k126) =
        (subst_TyVar_Ty (weakenTrace d81 k126) S55 (weakenTy S56 k126)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_subst_TmVar_Tm  :
      (forall (k126 : Hvl) (d81 : (Trace TmVar)) (s59 : Tm) (s60 : Tm) ,
        ((weakenTm (subst_TmVar_Tm d81 s59 s60) k126) =
        (subst_TmVar_Tm (weakenTrace d81 k126) s59 (weakenTm s60 k126)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_subst_TxVar_Tm  :
      (forall (k126 : Hvl) (d81 : (Trace TxVar)) (s59 : Tm) (s60 : Tm) ,
        ((weakenTm (subst_TxVar_Tm d81 s59 s60) k126) =
        (subst_TxVar_Tm (weakenTrace d81 k126) s59 (weakenTm s60 k126)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_subst_TyVar_Tm  :
      (forall (k126 : Hvl) (d81 : (Trace TyVar)) (S55 : Ty) (s59 : Tm) ,
        ((weakenTm (subst_TyVar_Tm d81 S55 s59) k126) =
        (subst_TyVar_Tm (weakenTrace d81 k126) S55 (weakenTm s59 k126)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
  End WeakenSubst.
  Section WeakenAppend.
    Lemma weakenTy_append  :
      (forall (S55 : Ty) (k126 : Hvl) (k127 : Hvl) ,
        ((weakenTy (weakenTy S55 k126) k127) =
        (weakenTy S55 (appendHvl k126 k127)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenTm_append  :
      (forall (s59 : Tm) (k126 : Hvl) (k127 : Hvl) ,
        ((weakenTm (weakenTm s59 k126) k127) =
        (weakenTm s59 (appendHvl k126 k127)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
  End WeakenAppend.
End SubstSubstInteraction.
 Hint Rewrite subst_TmVar_Tm_subst_TmVar_Tm0_comm subst_TxVar_Tm_subst_TxVar_Tm0_comm subst_TyVar_Tm_subst_TyVar_Tm0_comm subst_TyVar_Ty_subst_TyVar_Ty0_comm : interaction.
 Hint Rewrite subst_TmVar_Tm_subst_TmVar_Tm0_comm subst_TxVar_Tm_subst_TxVar_Tm0_comm subst_TyVar_Tm_subst_TyVar_Tm0_comm subst_TyVar_Ty_subst_TyVar_Ty0_comm : subst_subst0.
 Hint Rewrite weakenTm_shift_TmVar_Tm weakenTm_shift_TxVar_Tm weakenTm_shift_TyVar_Tm weakenTy_shift_TyVar_Ty : weaken_shift.
 Hint Rewrite weakenTm_subst_TmVar_Tm weakenTm_subst_TxVar_Tm weakenTm_subst_TyVar_Tm weakenTy_subst_TyVar_Ty : weaken_subst.
Section WellFormed.
  Fixpoint wfindex {a70 : Namespace} (k126 : Hvl) (i : (Index a70)) {struct k126} : Prop :=
    match k126 with
      | (H0) => False
      | (HS b k126) => match (eq_namespace_dec a70 b) with
        | (left e) => match i with
          | (I0) => True
          | (IS i) => (wfindex k126 i)
        end
        | (right n) => (wfindex k126 i)
      end
    end.
  Inductive wfTy (k126 : Hvl) : Ty -> Prop :=
    | wfTy_TVar
        {X71 : (Index TyVar)}
        (wfi : (wfindex k126 X71)) :
        (wfTy k126 (TVar X71))
    | wfTy_TArr {T99 : Ty}
        (wf : (wfTy k126 T99))
        {T100 : Ty}
        (wf0 : (wfTy k126 T100)) :
        (wfTy k126 (TArr T99 T100))
    | wfTy_TAll {T101 : Ty}
        (wf : (wfTy (HS TyVar k126) T101))
        : (wfTy k126 (TAll T101)).
  Inductive wfTm (k126 : Hvl) : Tm -> Prop :=
    | wfTm_Var
        {x108 : (Index TmVar)}
        (wfi : (wfindex k126 x108)) :
        (wfTm k126 (Var x108))
    | wfTm_XVar
        {a70 : (Index TxVar)}
        (wfi : (wfindex k126 a70)) :
        (wfTm k126 (XVar a70))
    | wfTm_Abs {T99 : Ty}
        (wf : (wfTy k126 T99))
        {t189 : Tm}
        (wf0 : (wfTm (HS TmVar k126) t189))
        : (wfTm k126 (Abs T99 t189))
    | wfTm_App {t190 : Tm}
        (wf : (wfTm k126 t190))
        {t191 : Tm}
        (wf0 : (wfTm k126 t191)) :
        (wfTm k126 (App t190 t191))
    | wfTm_TAbs {t192 : Tm}
        (wf : (wfTm (HS TyVar k126) t192))
        : (wfTm k126 (TAbs t192))
    | wfTm_TApp {t193 : Tm}
        (wf : (wfTm k126 t193))
        {T100 : Ty}
        (wf0 : (wfTy k126 T100)) :
        (wfTm k126 (TApp t193 T100)).
  Scheme ind_wfTy := Induction for wfTy Sort Prop.
  Scheme ind_wfTm := Induction for wfTm Sort Prop.
End WellFormed.
Section ShiftWellFormed.
  Inductive shifthvl_TmVar : (forall (c44 : (Cutoff TmVar)) (k126 : Hvl) (k127 : Hvl) ,
    Prop) :=
    | shifthvl_TmVar_here
        (k126 : Hvl) :
        (shifthvl_TmVar C0 k126 (HS TmVar k126))
    | shifthvl_TmVar_there_TmVar
        (c44 : (Cutoff TmVar))
        (k126 : Hvl) (k127 : Hvl) :
        (shifthvl_TmVar c44 k126 k127) -> (shifthvl_TmVar (CS c44) (HS TmVar k126) (HS TmVar k127))
    | shifthvl_TmVar_there_TxVar
        (c44 : (Cutoff TmVar))
        (k126 : Hvl) (k127 : Hvl) :
        (shifthvl_TmVar c44 k126 k127) -> (shifthvl_TmVar c44 (HS TxVar k126) (HS TxVar k127))
    | shifthvl_TmVar_there_TyVar
        (c44 : (Cutoff TmVar))
        (k126 : Hvl) (k127 : Hvl) :
        (shifthvl_TmVar c44 k126 k127) -> (shifthvl_TmVar c44 (HS TyVar k126) (HS TyVar k127)).
  Inductive shifthvl_TxVar : (forall (c44 : (Cutoff TxVar)) (k126 : Hvl) (k127 : Hvl) ,
    Prop) :=
    | shifthvl_TxVar_here
        (k126 : Hvl) :
        (shifthvl_TxVar C0 k126 (HS TxVar k126))
    | shifthvl_TxVar_there_TmVar
        (c44 : (Cutoff TxVar))
        (k126 : Hvl) (k127 : Hvl) :
        (shifthvl_TxVar c44 k126 k127) -> (shifthvl_TxVar c44 (HS TmVar k126) (HS TmVar k127))
    | shifthvl_TxVar_there_TxVar
        (c44 : (Cutoff TxVar))
        (k126 : Hvl) (k127 : Hvl) :
        (shifthvl_TxVar c44 k126 k127) -> (shifthvl_TxVar (CS c44) (HS TxVar k126) (HS TxVar k127))
    | shifthvl_TxVar_there_TyVar
        (c44 : (Cutoff TxVar))
        (k126 : Hvl) (k127 : Hvl) :
        (shifthvl_TxVar c44 k126 k127) -> (shifthvl_TxVar c44 (HS TyVar k126) (HS TyVar k127)).
  Inductive shifthvl_TyVar : (forall (c44 : (Cutoff TyVar)) (k126 : Hvl) (k127 : Hvl) ,
    Prop) :=
    | shifthvl_TyVar_here
        (k126 : Hvl) :
        (shifthvl_TyVar C0 k126 (HS TyVar k126))
    | shifthvl_TyVar_there_TmVar
        (c44 : (Cutoff TyVar))
        (k126 : Hvl) (k127 : Hvl) :
        (shifthvl_TyVar c44 k126 k127) -> (shifthvl_TyVar c44 (HS TmVar k126) (HS TmVar k127))
    | shifthvl_TyVar_there_TxVar
        (c44 : (Cutoff TyVar))
        (k126 : Hvl) (k127 : Hvl) :
        (shifthvl_TyVar c44 k126 k127) -> (shifthvl_TyVar c44 (HS TxVar k126) (HS TxVar k127))
    | shifthvl_TyVar_there_TyVar
        (c44 : (Cutoff TyVar))
        (k126 : Hvl) (k127 : Hvl) :
        (shifthvl_TyVar c44 k126 k127) -> (shifthvl_TyVar (CS c44) (HS TyVar k126) (HS TyVar k127)).
  Lemma weaken_shifthvl_TmVar  :
    (forall (k126 : Hvl) {c44 : (Cutoff TmVar)} {k127 : Hvl} {k128 : Hvl} ,
      (shifthvl_TmVar c44 k127 k128) -> (shifthvl_TmVar (weakenCutoffTmVar c44 k126) (appendHvl k127 k126) (appendHvl k128 k126))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma weaken_shifthvl_TxVar  :
    (forall (k126 : Hvl) {c44 : (Cutoff TxVar)} {k127 : Hvl} {k128 : Hvl} ,
      (shifthvl_TxVar c44 k127 k128) -> (shifthvl_TxVar (weakenCutoffTxVar c44 k126) (appendHvl k127 k126) (appendHvl k128 k126))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma weaken_shifthvl_TyVar  :
    (forall (k126 : Hvl) {c44 : (Cutoff TyVar)} {k127 : Hvl} {k128 : Hvl} ,
      (shifthvl_TyVar c44 k127 k128) -> (shifthvl_TyVar (weakenCutoffTyVar c44 k126) (appendHvl k127 k126) (appendHvl k128 k126))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma shift_TmVar__wfindex_TmVar  :
    (forall (c44 : (Cutoff TmVar)) (k126 : Hvl) (k127 : Hvl) (ins : (shifthvl_TmVar c44 k126 k127)) (x108 : (Index TmVar)) ,
      (wfindex k126 x108) -> (wfindex k127 (shift_TmVar_Index c44 x108))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TmVar__wfindex_TxVar  :
    (forall (c44 : (Cutoff TmVar)) (k126 : Hvl) (k127 : Hvl) (ins : (shifthvl_TmVar c44 k126 k127)) (a70 : (Index TxVar)) ,
      (wfindex k126 a70) -> (wfindex k127 a70)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TmVar__wfindex_TyVar  :
    (forall (c44 : (Cutoff TmVar)) (k126 : Hvl) (k127 : Hvl) (ins : (shifthvl_TmVar c44 k126 k127)) (X71 : (Index TyVar)) ,
      (wfindex k126 X71) -> (wfindex k127 X71)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TxVar__wfindex_TmVar  :
    (forall (c44 : (Cutoff TxVar)) (k126 : Hvl) (k127 : Hvl) (ins : (shifthvl_TxVar c44 k126 k127)) (x108 : (Index TmVar)) ,
      (wfindex k126 x108) -> (wfindex k127 x108)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TxVar__wfindex_TxVar  :
    (forall (c44 : (Cutoff TxVar)) (k126 : Hvl) (k127 : Hvl) (ins : (shifthvl_TxVar c44 k126 k127)) (a70 : (Index TxVar)) ,
      (wfindex k126 a70) -> (wfindex k127 (shift_TxVar_Index c44 a70))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TxVar__wfindex_TyVar  :
    (forall (c44 : (Cutoff TxVar)) (k126 : Hvl) (k127 : Hvl) (ins : (shifthvl_TxVar c44 k126 k127)) (X71 : (Index TyVar)) ,
      (wfindex k126 X71) -> (wfindex k127 X71)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TyVar__wfindex_TmVar  :
    (forall (c44 : (Cutoff TyVar)) (k126 : Hvl) (k127 : Hvl) (ins : (shifthvl_TyVar c44 k126 k127)) (x108 : (Index TmVar)) ,
      (wfindex k126 x108) -> (wfindex k127 x108)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TyVar__wfindex_TxVar  :
    (forall (c44 : (Cutoff TyVar)) (k126 : Hvl) (k127 : Hvl) (ins : (shifthvl_TyVar c44 k126 k127)) (a70 : (Index TxVar)) ,
      (wfindex k126 a70) -> (wfindex k127 a70)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_TyVar__wfindex_TyVar  :
    (forall (c44 : (Cutoff TyVar)) (k126 : Hvl) (k127 : Hvl) (ins : (shifthvl_TyVar c44 k126 k127)) (X71 : (Index TyVar)) ,
      (wfindex k126 X71) -> (wfindex k127 (shift_TyVar_Index c44 X71))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Definition shift_TmVar__wfTy : (forall (k126 : Hvl) ,
    (forall (S55 : Ty) (wf : (wfTy k126 S55)) ,
      (forall (c44 : (Cutoff TmVar)) (k127 : Hvl) ,
        (shifthvl_TmVar c44 k126 k127) -> (wfTy k127 S55)))) := (ind_wfTy (fun (k126 : Hvl) (S55 : Ty) (wf : (wfTy k126 S55)) =>
    (forall (c44 : (Cutoff TmVar)) (k127 : Hvl) ,
      (shifthvl_TmVar c44 k126 k127) -> (wfTy k127 S55))) (fun (k126 : Hvl) (X71 : (Index TyVar)) (wfi : (wfindex k126 X71)) (c44 : (Cutoff TmVar)) (k127 : Hvl) (ins : (shifthvl_TmVar c44 k126 k127)) =>
    (wfTy_TVar k127 (shift_TmVar__wfindex_TyVar c44 k126 k127 ins X71 wfi))) (fun (k126 : Hvl) (T1 : Ty) (wf : (wfTy k126 T1)) IHT0 (T2 : Ty) (wf0 : (wfTy k126 T2)) IHT3 (c44 : (Cutoff TmVar)) (k127 : Hvl) (ins : (shifthvl_TmVar c44 k126 k127)) =>
    (wfTy_TArr k127 (IHT0 c44 k127 (weaken_shifthvl_TmVar H0 ins)) (IHT3 c44 k127 (weaken_shifthvl_TmVar H0 ins)))) (fun (k126 : Hvl) (T : Ty) (wf : (wfTy (HS TyVar k126) T)) IHT0 (c44 : (Cutoff TmVar)) (k127 : Hvl) (ins : (shifthvl_TmVar c44 k126 k127)) =>
    (wfTy_TAll k127 (IHT0 c44 (HS TyVar k127) (weaken_shifthvl_TmVar (HS TyVar H0) ins))))).
  Definition shift_TxVar__wfTy : (forall (k126 : Hvl) ,
    (forall (S55 : Ty) (wf : (wfTy k126 S55)) ,
      (forall (c44 : (Cutoff TxVar)) (k127 : Hvl) ,
        (shifthvl_TxVar c44 k126 k127) -> (wfTy k127 S55)))) := (ind_wfTy (fun (k126 : Hvl) (S55 : Ty) (wf : (wfTy k126 S55)) =>
    (forall (c44 : (Cutoff TxVar)) (k127 : Hvl) ,
      (shifthvl_TxVar c44 k126 k127) -> (wfTy k127 S55))) (fun (k126 : Hvl) (X71 : (Index TyVar)) (wfi : (wfindex k126 X71)) (c44 : (Cutoff TxVar)) (k127 : Hvl) (ins : (shifthvl_TxVar c44 k126 k127)) =>
    (wfTy_TVar k127 (shift_TxVar__wfindex_TyVar c44 k126 k127 ins X71 wfi))) (fun (k126 : Hvl) (T1 : Ty) (wf : (wfTy k126 T1)) IHT0 (T2 : Ty) (wf0 : (wfTy k126 T2)) IHT3 (c44 : (Cutoff TxVar)) (k127 : Hvl) (ins : (shifthvl_TxVar c44 k126 k127)) =>
    (wfTy_TArr k127 (IHT0 c44 k127 (weaken_shifthvl_TxVar H0 ins)) (IHT3 c44 k127 (weaken_shifthvl_TxVar H0 ins)))) (fun (k126 : Hvl) (T : Ty) (wf : (wfTy (HS TyVar k126) T)) IHT0 (c44 : (Cutoff TxVar)) (k127 : Hvl) (ins : (shifthvl_TxVar c44 k126 k127)) =>
    (wfTy_TAll k127 (IHT0 c44 (HS TyVar k127) (weaken_shifthvl_TxVar (HS TyVar H0) ins))))).
  Definition shift_TyVar__wfTy : (forall (k126 : Hvl) ,
    (forall (S55 : Ty) (wf : (wfTy k126 S55)) ,
      (forall (c44 : (Cutoff TyVar)) (k127 : Hvl) ,
        (shifthvl_TyVar c44 k126 k127) -> (wfTy k127 (shift_TyVar_Ty c44 S55))))) := (ind_wfTy (fun (k126 : Hvl) (S55 : Ty) (wf : (wfTy k126 S55)) =>
    (forall (c44 : (Cutoff TyVar)) (k127 : Hvl) ,
      (shifthvl_TyVar c44 k126 k127) -> (wfTy k127 (shift_TyVar_Ty c44 S55)))) (fun (k126 : Hvl) (X71 : (Index TyVar)) (wfi : (wfindex k126 X71)) (c44 : (Cutoff TyVar)) (k127 : Hvl) (ins : (shifthvl_TyVar c44 k126 k127)) =>
    (wfTy_TVar k127 (shift_TyVar__wfindex_TyVar c44 k126 k127 ins X71 wfi))) (fun (k126 : Hvl) (T1 : Ty) (wf : (wfTy k126 T1)) IHT0 (T2 : Ty) (wf0 : (wfTy k126 T2)) IHT3 (c44 : (Cutoff TyVar)) (k127 : Hvl) (ins : (shifthvl_TyVar c44 k126 k127)) =>
    (wfTy_TArr k127 (IHT0 c44 k127 (weaken_shifthvl_TyVar H0 ins)) (IHT3 c44 k127 (weaken_shifthvl_TyVar H0 ins)))) (fun (k126 : Hvl) (T : Ty) (wf : (wfTy (HS TyVar k126) T)) IHT0 (c44 : (Cutoff TyVar)) (k127 : Hvl) (ins : (shifthvl_TyVar c44 k126 k127)) =>
    (wfTy_TAll k127 (IHT0 (CS c44) (HS TyVar k127) (weaken_shifthvl_TyVar (HS TyVar H0) ins))))).
  Definition shift_TmVar__wfTm : (forall (k126 : Hvl) ,
    (forall (s59 : Tm) (wf : (wfTm k126 s59)) ,
      (forall (c44 : (Cutoff TmVar)) (k127 : Hvl) ,
        (shifthvl_TmVar c44 k126 k127) -> (wfTm k127 (shift_TmVar_Tm c44 s59))))) := (ind_wfTm (fun (k126 : Hvl) (s59 : Tm) (wf : (wfTm k126 s59)) =>
    (forall (c44 : (Cutoff TmVar)) (k127 : Hvl) ,
      (shifthvl_TmVar c44 k126 k127) -> (wfTm k127 (shift_TmVar_Tm c44 s59)))) (fun (k126 : Hvl) (x108 : (Index TmVar)) (wfi : (wfindex k126 x108)) (c44 : (Cutoff TmVar)) (k127 : Hvl) (ins : (shifthvl_TmVar c44 k126 k127)) =>
    (wfTm_Var k127 (shift_TmVar__wfindex_TmVar c44 k126 k127 ins x108 wfi))) (fun (k126 : Hvl) (a70 : (Index TxVar)) (wfi : (wfindex k126 a70)) (c44 : (Cutoff TmVar)) (k127 : Hvl) (ins : (shifthvl_TmVar c44 k126 k127)) =>
    (wfTm_XVar k127 (shift_TmVar__wfindex_TxVar c44 k126 k127 ins a70 wfi))) (fun (k126 : Hvl) (T : Ty) (wf : (wfTy k126 T)) (t : Tm) (wf0 : (wfTm (HS TmVar k126) t)) IHt189 (c44 : (Cutoff TmVar)) (k127 : Hvl) (ins : (shifthvl_TmVar c44 k126 k127)) =>
    (wfTm_Abs k127 (shift_TmVar__wfTy k126 T wf c44 k127 (weaken_shifthvl_TmVar H0 ins)) (IHt189 (CS c44) (HS TmVar k127) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k126 : Hvl) (t1 : Tm) (wf : (wfTm k126 t1)) IHt189 (t2 : Tm) (wf0 : (wfTm k126 t2)) IHt190 (c44 : (Cutoff TmVar)) (k127 : Hvl) (ins : (shifthvl_TmVar c44 k126 k127)) =>
    (wfTm_App k127 (IHt189 c44 k127 (weaken_shifthvl_TmVar H0 ins)) (IHt190 c44 k127 (weaken_shifthvl_TmVar H0 ins)))) (fun (k126 : Hvl) (t : Tm) (wf : (wfTm (HS TyVar k126) t)) IHt189 (c44 : (Cutoff TmVar)) (k127 : Hvl) (ins : (shifthvl_TmVar c44 k126 k127)) =>
    (wfTm_TAbs k127 (IHt189 c44 (HS TyVar k127) (weaken_shifthvl_TmVar (HS TyVar H0) ins)))) (fun (k126 : Hvl) (t : Tm) (wf : (wfTm k126 t)) IHt189 (T : Ty) (wf0 : (wfTy k126 T)) (c44 : (Cutoff TmVar)) (k127 : Hvl) (ins : (shifthvl_TmVar c44 k126 k127)) =>
    (wfTm_TApp k127 (IHt189 c44 k127 (weaken_shifthvl_TmVar H0 ins)) (shift_TmVar__wfTy k126 T wf0 c44 k127 (weaken_shifthvl_TmVar H0 ins))))).
  Definition shift_TxVar__wfTm : (forall (k126 : Hvl) ,
    (forall (s59 : Tm) (wf : (wfTm k126 s59)) ,
      (forall (c44 : (Cutoff TxVar)) (k127 : Hvl) ,
        (shifthvl_TxVar c44 k126 k127) -> (wfTm k127 (shift_TxVar_Tm c44 s59))))) := (ind_wfTm (fun (k126 : Hvl) (s59 : Tm) (wf : (wfTm k126 s59)) =>
    (forall (c44 : (Cutoff TxVar)) (k127 : Hvl) ,
      (shifthvl_TxVar c44 k126 k127) -> (wfTm k127 (shift_TxVar_Tm c44 s59)))) (fun (k126 : Hvl) (x108 : (Index TmVar)) (wfi : (wfindex k126 x108)) (c44 : (Cutoff TxVar)) (k127 : Hvl) (ins : (shifthvl_TxVar c44 k126 k127)) =>
    (wfTm_Var k127 (shift_TxVar__wfindex_TmVar c44 k126 k127 ins x108 wfi))) (fun (k126 : Hvl) (a70 : (Index TxVar)) (wfi : (wfindex k126 a70)) (c44 : (Cutoff TxVar)) (k127 : Hvl) (ins : (shifthvl_TxVar c44 k126 k127)) =>
    (wfTm_XVar k127 (shift_TxVar__wfindex_TxVar c44 k126 k127 ins a70 wfi))) (fun (k126 : Hvl) (T : Ty) (wf : (wfTy k126 T)) (t : Tm) (wf0 : (wfTm (HS TmVar k126) t)) IHt189 (c44 : (Cutoff TxVar)) (k127 : Hvl) (ins : (shifthvl_TxVar c44 k126 k127)) =>
    (wfTm_Abs k127 (shift_TxVar__wfTy k126 T wf c44 k127 (weaken_shifthvl_TxVar H0 ins)) (IHt189 c44 (HS TmVar k127) (weaken_shifthvl_TxVar (HS TmVar H0) ins)))) (fun (k126 : Hvl) (t1 : Tm) (wf : (wfTm k126 t1)) IHt189 (t2 : Tm) (wf0 : (wfTm k126 t2)) IHt190 (c44 : (Cutoff TxVar)) (k127 : Hvl) (ins : (shifthvl_TxVar c44 k126 k127)) =>
    (wfTm_App k127 (IHt189 c44 k127 (weaken_shifthvl_TxVar H0 ins)) (IHt190 c44 k127 (weaken_shifthvl_TxVar H0 ins)))) (fun (k126 : Hvl) (t : Tm) (wf : (wfTm (HS TyVar k126) t)) IHt189 (c44 : (Cutoff TxVar)) (k127 : Hvl) (ins : (shifthvl_TxVar c44 k126 k127)) =>
    (wfTm_TAbs k127 (IHt189 c44 (HS TyVar k127) (weaken_shifthvl_TxVar (HS TyVar H0) ins)))) (fun (k126 : Hvl) (t : Tm) (wf : (wfTm k126 t)) IHt189 (T : Ty) (wf0 : (wfTy k126 T)) (c44 : (Cutoff TxVar)) (k127 : Hvl) (ins : (shifthvl_TxVar c44 k126 k127)) =>
    (wfTm_TApp k127 (IHt189 c44 k127 (weaken_shifthvl_TxVar H0 ins)) (shift_TxVar__wfTy k126 T wf0 c44 k127 (weaken_shifthvl_TxVar H0 ins))))).
  Definition shift_TyVar__wfTm : (forall (k126 : Hvl) ,
    (forall (s59 : Tm) (wf : (wfTm k126 s59)) ,
      (forall (c44 : (Cutoff TyVar)) (k127 : Hvl) ,
        (shifthvl_TyVar c44 k126 k127) -> (wfTm k127 (shift_TyVar_Tm c44 s59))))) := (ind_wfTm (fun (k126 : Hvl) (s59 : Tm) (wf : (wfTm k126 s59)) =>
    (forall (c44 : (Cutoff TyVar)) (k127 : Hvl) ,
      (shifthvl_TyVar c44 k126 k127) -> (wfTm k127 (shift_TyVar_Tm c44 s59)))) (fun (k126 : Hvl) (x108 : (Index TmVar)) (wfi : (wfindex k126 x108)) (c44 : (Cutoff TyVar)) (k127 : Hvl) (ins : (shifthvl_TyVar c44 k126 k127)) =>
    (wfTm_Var k127 (shift_TyVar__wfindex_TmVar c44 k126 k127 ins x108 wfi))) (fun (k126 : Hvl) (a70 : (Index TxVar)) (wfi : (wfindex k126 a70)) (c44 : (Cutoff TyVar)) (k127 : Hvl) (ins : (shifthvl_TyVar c44 k126 k127)) =>
    (wfTm_XVar k127 (shift_TyVar__wfindex_TxVar c44 k126 k127 ins a70 wfi))) (fun (k126 : Hvl) (T : Ty) (wf : (wfTy k126 T)) (t : Tm) (wf0 : (wfTm (HS TmVar k126) t)) IHt189 (c44 : (Cutoff TyVar)) (k127 : Hvl) (ins : (shifthvl_TyVar c44 k126 k127)) =>
    (wfTm_Abs k127 (shift_TyVar__wfTy k126 T wf c44 k127 (weaken_shifthvl_TyVar H0 ins)) (IHt189 c44 (HS TmVar k127) (weaken_shifthvl_TyVar (HS TmVar H0) ins)))) (fun (k126 : Hvl) (t1 : Tm) (wf : (wfTm k126 t1)) IHt189 (t2 : Tm) (wf0 : (wfTm k126 t2)) IHt190 (c44 : (Cutoff TyVar)) (k127 : Hvl) (ins : (shifthvl_TyVar c44 k126 k127)) =>
    (wfTm_App k127 (IHt189 c44 k127 (weaken_shifthvl_TyVar H0 ins)) (IHt190 c44 k127 (weaken_shifthvl_TyVar H0 ins)))) (fun (k126 : Hvl) (t : Tm) (wf : (wfTm (HS TyVar k126) t)) IHt189 (c44 : (Cutoff TyVar)) (k127 : Hvl) (ins : (shifthvl_TyVar c44 k126 k127)) =>
    (wfTm_TAbs k127 (IHt189 (CS c44) (HS TyVar k127) (weaken_shifthvl_TyVar (HS TyVar H0) ins)))) (fun (k126 : Hvl) (t : Tm) (wf : (wfTm k126 t)) IHt189 (T : Ty) (wf0 : (wfTy k126 T)) (c44 : (Cutoff TyVar)) (k127 : Hvl) (ins : (shifthvl_TyVar c44 k126 k127)) =>
    (wfTm_TApp k127 (IHt189 c44 k127 (weaken_shifthvl_TyVar H0 ins)) (shift_TyVar__wfTy k126 T wf0 c44 k127 (weaken_shifthvl_TyVar H0 ins))))).
End ShiftWellFormed.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TxVar shift_TmVar__wfindex_TyVar shift_TxVar__wfindex_TmVar shift_TxVar__wfindex_TxVar shift_TxVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TxVar shift_TyVar__wfindex_TyVar : infra.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TxVar shift_TmVar__wfindex_TyVar shift_TxVar__wfindex_TmVar shift_TxVar__wfindex_TxVar shift_TxVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TxVar shift_TyVar__wfindex_TyVar : shift.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TxVar shift_TmVar__wfindex_TyVar shift_TxVar__wfindex_TmVar shift_TxVar__wfindex_TxVar shift_TxVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TxVar shift_TyVar__wfindex_TyVar : shift_wf.
 Hint Resolve shift_TmVar__wfindex_TmVar shift_TmVar__wfindex_TxVar shift_TmVar__wfindex_TyVar shift_TxVar__wfindex_TmVar shift_TxVar__wfindex_TxVar shift_TxVar__wfindex_TyVar shift_TyVar__wfindex_TmVar shift_TyVar__wfindex_TxVar shift_TyVar__wfindex_TyVar : wf.
 Hint Resolve shift_TmVar__wfTm shift_TmVar__wfTy shift_TxVar__wfTm shift_TxVar__wfTy shift_TyVar__wfTm shift_TyVar__wfTy : infra.
 Hint Resolve shift_TmVar__wfTm shift_TmVar__wfTy shift_TxVar__wfTm shift_TxVar__wfTy shift_TyVar__wfTm shift_TyVar__wfTy : shift.
 Hint Resolve shift_TmVar__wfTm shift_TmVar__wfTy shift_TxVar__wfTm shift_TxVar__wfTy shift_TyVar__wfTm shift_TyVar__wfTy : shift_wf.
 Hint Resolve shift_TmVar__wfTm shift_TmVar__wfTy shift_TxVar__wfTm shift_TxVar__wfTy shift_TyVar__wfTm shift_TyVar__wfTy : wf.
 Hint Constructors shifthvl_TmVar shifthvl_TxVar shifthvl_TyVar : infra.
 Hint Constructors shifthvl_TmVar shifthvl_TxVar shifthvl_TyVar : shift.
 Hint Constructors shifthvl_TmVar shifthvl_TxVar shifthvl_TyVar : shift_wf.
 Hint Constructors shifthvl_TmVar shifthvl_TxVar shifthvl_TyVar : wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TxVar weaken_shifthvl_TyVar : infra.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TxVar weaken_shifthvl_TyVar : shift.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TxVar weaken_shifthvl_TyVar : shift_wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TxVar weaken_shifthvl_TyVar : weaken.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TxVar weaken_shifthvl_TyVar : wf.
Section SubstWellFormed.
  Inductive substhvl_TmVar (k126 : Hvl) : (forall (d81 : (Trace TmVar)) (k127 : Hvl) (k128 : Hvl) ,
    Prop) :=
    | substhvl_TmVar_here :
        (substhvl_TmVar k126 X0 (HS TmVar k126) k126)
    | substhvl_TmVar_there_TmVar
        {d81 : (Trace TmVar)}
        {k127 : Hvl} {k128 : Hvl} :
        (substhvl_TmVar k126 d81 k127 k128) -> (substhvl_TmVar k126 (XS TmVar d81) (HS TmVar k127) (HS TmVar k128))
    | substhvl_TmVar_there_TxVar
        {d81 : (Trace TmVar)}
        {k127 : Hvl} {k128 : Hvl} :
        (substhvl_TmVar k126 d81 k127 k128) -> (substhvl_TmVar k126 (XS TxVar d81) (HS TxVar k127) (HS TxVar k128))
    | substhvl_TmVar_there_TyVar
        {d81 : (Trace TmVar)}
        {k127 : Hvl} {k128 : Hvl} :
        (substhvl_TmVar k126 d81 k127 k128) -> (substhvl_TmVar k126 (XS TyVar d81) (HS TyVar k127) (HS TyVar k128)).
  Inductive substhvl_TxVar (k126 : Hvl) : (forall (d81 : (Trace TxVar)) (k127 : Hvl) (k128 : Hvl) ,
    Prop) :=
    | substhvl_TxVar_here :
        (substhvl_TxVar k126 X0 (HS TxVar k126) k126)
    | substhvl_TxVar_there_TmVar
        {d81 : (Trace TxVar)}
        {k127 : Hvl} {k128 : Hvl} :
        (substhvl_TxVar k126 d81 k127 k128) -> (substhvl_TxVar k126 (XS TmVar d81) (HS TmVar k127) (HS TmVar k128))
    | substhvl_TxVar_there_TxVar
        {d81 : (Trace TxVar)}
        {k127 : Hvl} {k128 : Hvl} :
        (substhvl_TxVar k126 d81 k127 k128) -> (substhvl_TxVar k126 (XS TxVar d81) (HS TxVar k127) (HS TxVar k128))
    | substhvl_TxVar_there_TyVar
        {d81 : (Trace TxVar)}
        {k127 : Hvl} {k128 : Hvl} :
        (substhvl_TxVar k126 d81 k127 k128) -> (substhvl_TxVar k126 (XS TyVar d81) (HS TyVar k127) (HS TyVar k128)).
  Inductive substhvl_TyVar (k126 : Hvl) : (forall (d81 : (Trace TyVar)) (k127 : Hvl) (k128 : Hvl) ,
    Prop) :=
    | substhvl_TyVar_here :
        (substhvl_TyVar k126 X0 (HS TyVar k126) k126)
    | substhvl_TyVar_there_TmVar
        {d81 : (Trace TyVar)}
        {k127 : Hvl} {k128 : Hvl} :
        (substhvl_TyVar k126 d81 k127 k128) -> (substhvl_TyVar k126 (XS TmVar d81) (HS TmVar k127) (HS TmVar k128))
    | substhvl_TyVar_there_TxVar
        {d81 : (Trace TyVar)}
        {k127 : Hvl} {k128 : Hvl} :
        (substhvl_TyVar k126 d81 k127 k128) -> (substhvl_TyVar k126 (XS TxVar d81) (HS TxVar k127) (HS TxVar k128))
    | substhvl_TyVar_there_TyVar
        {d81 : (Trace TyVar)}
        {k127 : Hvl} {k128 : Hvl} :
        (substhvl_TyVar k126 d81 k127 k128) -> (substhvl_TyVar k126 (XS TyVar d81) (HS TyVar k127) (HS TyVar k128)).
  Lemma weaken_substhvl_TmVar  :
    (forall {k127 : Hvl} (k126 : Hvl) {d81 : (Trace TmVar)} {k128 : Hvl} {k129 : Hvl} ,
      (substhvl_TmVar k127 d81 k128 k129) -> (substhvl_TmVar k127 (weakenTrace d81 k126) (appendHvl k128 k126) (appendHvl k129 k126))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma weaken_substhvl_TxVar  :
    (forall {k127 : Hvl} (k126 : Hvl) {d81 : (Trace TxVar)} {k128 : Hvl} {k129 : Hvl} ,
      (substhvl_TxVar k127 d81 k128 k129) -> (substhvl_TxVar k127 (weakenTrace d81 k126) (appendHvl k128 k126) (appendHvl k129 k126))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma weaken_substhvl_TyVar  :
    (forall {k127 : Hvl} (k126 : Hvl) {d81 : (Trace TyVar)} {k128 : Hvl} {k129 : Hvl} ,
      (substhvl_TyVar k127 d81 k128 k129) -> (substhvl_TyVar k127 (weakenTrace d81 k126) (appendHvl k128 k126) (appendHvl k129 k126))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma substhvl_TmVar_wfindex_TmVar {k126 : Hvl} {s59 : Tm} (wft : (wfTm k126 s59)) :
    (forall {d81 : (Trace TmVar)} {k127 : Hvl} {k128 : Hvl} ,
      (substhvl_TmVar k126 d81 k127 k128) -> (forall {x108 : (Index TmVar)} ,
        (wfindex k127 x108) -> (wfTm k128 (subst_TmVar_Index d81 s59 x108)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TxVar_wfindex_TxVar {k126 : Hvl} {s59 : Tm} (wft : (wfTm k126 s59)) :
    (forall {d81 : (Trace TxVar)} {k127 : Hvl} {k128 : Hvl} ,
      (substhvl_TxVar k126 d81 k127 k128) -> (forall {a70 : (Index TxVar)} ,
        (wfindex k127 a70) -> (wfTm k128 (subst_TxVar_Index d81 s59 a70)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TyVar_wfindex_TyVar {k126 : Hvl} {S55 : Ty} (wft : (wfTy k126 S55)) :
    (forall {d81 : (Trace TyVar)} {k127 : Hvl} {k128 : Hvl} ,
      (substhvl_TyVar k126 d81 k127 k128) -> (forall {X71 : (Index TyVar)} ,
        (wfindex k127 X71) -> (wfTy k128 (subst_TyVar_Index d81 S55 X71)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TmVar_wfindex_TxVar {k126 : Hvl} :
    (forall {d81 : (Trace TmVar)} {k127 : Hvl} {k128 : Hvl} ,
      (substhvl_TmVar k126 d81 k127 k128) -> (forall {a70 : (Index TxVar)} ,
        (wfindex k127 a70) -> (wfindex k128 a70))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_TmVar_wfindex_TyVar {k126 : Hvl} :
    (forall {d81 : (Trace TmVar)} {k127 : Hvl} {k128 : Hvl} ,
      (substhvl_TmVar k126 d81 k127 k128) -> (forall {X71 : (Index TyVar)} ,
        (wfindex k127 X71) -> (wfindex k128 X71))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_TxVar_wfindex_TmVar {k126 : Hvl} :
    (forall {d81 : (Trace TxVar)} {k127 : Hvl} {k128 : Hvl} ,
      (substhvl_TxVar k126 d81 k127 k128) -> (forall {x108 : (Index TmVar)} ,
        (wfindex k127 x108) -> (wfindex k128 x108))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_TxVar_wfindex_TyVar {k126 : Hvl} :
    (forall {d81 : (Trace TxVar)} {k127 : Hvl} {k128 : Hvl} ,
      (substhvl_TxVar k126 d81 k127 k128) -> (forall {X71 : (Index TyVar)} ,
        (wfindex k127 X71) -> (wfindex k128 X71))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_TyVar_wfindex_TmVar {k126 : Hvl} :
    (forall {d81 : (Trace TyVar)} {k127 : Hvl} {k128 : Hvl} ,
      (substhvl_TyVar k126 d81 k127 k128) -> (forall {x108 : (Index TmVar)} ,
        (wfindex k127 x108) -> (wfindex k128 x108))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_TyVar_wfindex_TxVar {k126 : Hvl} :
    (forall {d81 : (Trace TyVar)} {k127 : Hvl} {k128 : Hvl} ,
      (substhvl_TyVar k126 d81 k127 k128) -> (forall {a70 : (Index TxVar)} ,
        (wfindex k127 a70) -> (wfindex k128 a70))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Definition substhvl_TmVar_wfTy {k126 : Hvl} : (forall (k127 : Hvl) ,
    (forall (S55 : Ty) (wf0 : (wfTy k127 S55)) ,
      (forall {d81 : (Trace TmVar)} {k128 : Hvl} ,
        (substhvl_TmVar k126 d81 k127 k128) -> (wfTy k128 S55)))) := (ind_wfTy (fun (k127 : Hvl) (S55 : Ty) (wf0 : (wfTy k127 S55)) =>
    (forall {d81 : (Trace TmVar)} {k128 : Hvl} ,
      (substhvl_TmVar k126 d81 k127 k128) -> (wfTy k128 S55))) (fun (k127 : Hvl) {X71 : (Index TyVar)} (wfi : (wfindex k127 X71)) {d81 : (Trace TmVar)} {k128 : Hvl} (del : (substhvl_TmVar k126 d81 k127 k128)) =>
    (wfTy_TVar k128 (substhvl_TmVar_wfindex_TyVar del wfi))) (fun (k127 : Hvl) (T1 : Ty) (wf0 : (wfTy k127 T1)) IHT0 (T2 : Ty) (wf1 : (wfTy k127 T2)) IHT3 {d81 : (Trace TmVar)} {k128 : Hvl} (del : (substhvl_TmVar k126 d81 k127 k128)) =>
    (wfTy_TArr k128 (IHT0 (weakenTrace d81 H0) k128 (weaken_substhvl_TmVar H0 del)) (IHT3 (weakenTrace d81 H0) k128 (weaken_substhvl_TmVar H0 del)))) (fun (k127 : Hvl) (T : Ty) (wf0 : (wfTy (HS TyVar k127) T)) IHT0 {d81 : (Trace TmVar)} {k128 : Hvl} (del : (substhvl_TmVar k126 d81 k127 k128)) =>
    (wfTy_TAll k128 (IHT0 (weakenTrace d81 (HS TyVar H0)) (HS TyVar k128) (weaken_substhvl_TmVar (HS TyVar H0) del))))).
  Definition substhvl_TxVar_wfTy {k126 : Hvl} : (forall (k127 : Hvl) ,
    (forall (S55 : Ty) (wf0 : (wfTy k127 S55)) ,
      (forall {d81 : (Trace TxVar)} {k128 : Hvl} ,
        (substhvl_TxVar k126 d81 k127 k128) -> (wfTy k128 S55)))) := (ind_wfTy (fun (k127 : Hvl) (S55 : Ty) (wf0 : (wfTy k127 S55)) =>
    (forall {d81 : (Trace TxVar)} {k128 : Hvl} ,
      (substhvl_TxVar k126 d81 k127 k128) -> (wfTy k128 S55))) (fun (k127 : Hvl) {X71 : (Index TyVar)} (wfi : (wfindex k127 X71)) {d81 : (Trace TxVar)} {k128 : Hvl} (del : (substhvl_TxVar k126 d81 k127 k128)) =>
    (wfTy_TVar k128 (substhvl_TxVar_wfindex_TyVar del wfi))) (fun (k127 : Hvl) (T1 : Ty) (wf0 : (wfTy k127 T1)) IHT0 (T2 : Ty) (wf1 : (wfTy k127 T2)) IHT3 {d81 : (Trace TxVar)} {k128 : Hvl} (del : (substhvl_TxVar k126 d81 k127 k128)) =>
    (wfTy_TArr k128 (IHT0 (weakenTrace d81 H0) k128 (weaken_substhvl_TxVar H0 del)) (IHT3 (weakenTrace d81 H0) k128 (weaken_substhvl_TxVar H0 del)))) (fun (k127 : Hvl) (T : Ty) (wf0 : (wfTy (HS TyVar k127) T)) IHT0 {d81 : (Trace TxVar)} {k128 : Hvl} (del : (substhvl_TxVar k126 d81 k127 k128)) =>
    (wfTy_TAll k128 (IHT0 (weakenTrace d81 (HS TyVar H0)) (HS TyVar k128) (weaken_substhvl_TxVar (HS TyVar H0) del))))).
  Definition substhvl_TyVar_wfTy {k126 : Hvl} {S55 : Ty} (wf : (wfTy k126 S55)) : (forall (k127 : Hvl) ,
    (forall (S56 : Ty) (wf0 : (wfTy k127 S56)) ,
      (forall {d81 : (Trace TyVar)} {k128 : Hvl} ,
        (substhvl_TyVar k126 d81 k127 k128) -> (wfTy k128 (subst_TyVar_Ty d81 S55 S56))))) := (ind_wfTy (fun (k127 : Hvl) (S56 : Ty) (wf0 : (wfTy k127 S56)) =>
    (forall {d81 : (Trace TyVar)} {k128 : Hvl} ,
      (substhvl_TyVar k126 d81 k127 k128) -> (wfTy k128 (subst_TyVar_Ty d81 S55 S56)))) (fun (k127 : Hvl) {X71 : (Index TyVar)} (wfi : (wfindex k127 X71)) {d81 : (Trace TyVar)} {k128 : Hvl} (del : (substhvl_TyVar k126 d81 k127 k128)) =>
    (substhvl_TyVar_wfindex_TyVar wf del wfi)) (fun (k127 : Hvl) (T1 : Ty) (wf0 : (wfTy k127 T1)) IHT0 (T2 : Ty) (wf1 : (wfTy k127 T2)) IHT3 {d81 : (Trace TyVar)} {k128 : Hvl} (del : (substhvl_TyVar k126 d81 k127 k128)) =>
    (wfTy_TArr k128 (IHT0 (weakenTrace d81 H0) k128 (weaken_substhvl_TyVar H0 del)) (IHT3 (weakenTrace d81 H0) k128 (weaken_substhvl_TyVar H0 del)))) (fun (k127 : Hvl) (T : Ty) (wf0 : (wfTy (HS TyVar k127) T)) IHT0 {d81 : (Trace TyVar)} {k128 : Hvl} (del : (substhvl_TyVar k126 d81 k127 k128)) =>
    (wfTy_TAll k128 (IHT0 (weakenTrace d81 (HS TyVar H0)) (HS TyVar k128) (weaken_substhvl_TyVar (HS TyVar H0) del))))).
  Definition substhvl_TmVar_wfTm {k126 : Hvl} {s59 : Tm} (wf : (wfTm k126 s59)) : (forall (k127 : Hvl) ,
    (forall (s60 : Tm) (wf0 : (wfTm k127 s60)) ,
      (forall {d81 : (Trace TmVar)} {k128 : Hvl} ,
        (substhvl_TmVar k126 d81 k127 k128) -> (wfTm k128 (subst_TmVar_Tm d81 s59 s60))))) := (ind_wfTm (fun (k127 : Hvl) (s60 : Tm) (wf0 : (wfTm k127 s60)) =>
    (forall {d81 : (Trace TmVar)} {k128 : Hvl} ,
      (substhvl_TmVar k126 d81 k127 k128) -> (wfTm k128 (subst_TmVar_Tm d81 s59 s60)))) (fun (k127 : Hvl) {x108 : (Index TmVar)} (wfi : (wfindex k127 x108)) {d81 : (Trace TmVar)} {k128 : Hvl} (del : (substhvl_TmVar k126 d81 k127 k128)) =>
    (substhvl_TmVar_wfindex_TmVar wf del wfi)) (fun (k127 : Hvl) {a70 : (Index TxVar)} (wfi : (wfindex k127 a70)) {d81 : (Trace TmVar)} {k128 : Hvl} (del : (substhvl_TmVar k126 d81 k127 k128)) =>
    (wfTm_XVar k128 (substhvl_TmVar_wfindex_TxVar del wfi))) (fun (k127 : Hvl) (T : Ty) (wf0 : (wfTy k127 T)) (t : Tm) (wf1 : (wfTm (HS TmVar k127) t)) IHt189 {d81 : (Trace TmVar)} {k128 : Hvl} (del : (substhvl_TmVar k126 d81 k127 k128)) =>
    (wfTm_Abs k128 (substhvl_TmVar_wfTy k127 T wf0 (weaken_substhvl_TmVar H0 del)) (IHt189 (weakenTrace d81 (HS TmVar H0)) (HS TmVar k128) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k127 : Hvl) (t1 : Tm) (wf0 : (wfTm k127 t1)) IHt189 (t2 : Tm) (wf1 : (wfTm k127 t2)) IHt190 {d81 : (Trace TmVar)} {k128 : Hvl} (del : (substhvl_TmVar k126 d81 k127 k128)) =>
    (wfTm_App k128 (IHt189 (weakenTrace d81 H0) k128 (weaken_substhvl_TmVar H0 del)) (IHt190 (weakenTrace d81 H0) k128 (weaken_substhvl_TmVar H0 del)))) (fun (k127 : Hvl) (t : Tm) (wf0 : (wfTm (HS TyVar k127) t)) IHt189 {d81 : (Trace TmVar)} {k128 : Hvl} (del : (substhvl_TmVar k126 d81 k127 k128)) =>
    (wfTm_TAbs k128 (IHt189 (weakenTrace d81 (HS TyVar H0)) (HS TyVar k128) (weaken_substhvl_TmVar (HS TyVar H0) del)))) (fun (k127 : Hvl) (t : Tm) (wf0 : (wfTm k127 t)) IHt189 (T : Ty) (wf1 : (wfTy k127 T)) {d81 : (Trace TmVar)} {k128 : Hvl} (del : (substhvl_TmVar k126 d81 k127 k128)) =>
    (wfTm_TApp k128 (IHt189 (weakenTrace d81 H0) k128 (weaken_substhvl_TmVar H0 del)) (substhvl_TmVar_wfTy k127 T wf1 (weaken_substhvl_TmVar H0 del))))).
  Definition substhvl_TxVar_wfTm {k126 : Hvl} {s59 : Tm} (wf : (wfTm k126 s59)) : (forall (k127 : Hvl) ,
    (forall (s60 : Tm) (wf0 : (wfTm k127 s60)) ,
      (forall {d81 : (Trace TxVar)} {k128 : Hvl} ,
        (substhvl_TxVar k126 d81 k127 k128) -> (wfTm k128 (subst_TxVar_Tm d81 s59 s60))))) := (ind_wfTm (fun (k127 : Hvl) (s60 : Tm) (wf0 : (wfTm k127 s60)) =>
    (forall {d81 : (Trace TxVar)} {k128 : Hvl} ,
      (substhvl_TxVar k126 d81 k127 k128) -> (wfTm k128 (subst_TxVar_Tm d81 s59 s60)))) (fun (k127 : Hvl) {x108 : (Index TmVar)} (wfi : (wfindex k127 x108)) {d81 : (Trace TxVar)} {k128 : Hvl} (del : (substhvl_TxVar k126 d81 k127 k128)) =>
    (wfTm_Var k128 (substhvl_TxVar_wfindex_TmVar del wfi))) (fun (k127 : Hvl) {a70 : (Index TxVar)} (wfi : (wfindex k127 a70)) {d81 : (Trace TxVar)} {k128 : Hvl} (del : (substhvl_TxVar k126 d81 k127 k128)) =>
    (substhvl_TxVar_wfindex_TxVar wf del wfi)) (fun (k127 : Hvl) (T : Ty) (wf0 : (wfTy k127 T)) (t : Tm) (wf1 : (wfTm (HS TmVar k127) t)) IHt189 {d81 : (Trace TxVar)} {k128 : Hvl} (del : (substhvl_TxVar k126 d81 k127 k128)) =>
    (wfTm_Abs k128 (substhvl_TxVar_wfTy k127 T wf0 (weaken_substhvl_TxVar H0 del)) (IHt189 (weakenTrace d81 (HS TmVar H0)) (HS TmVar k128) (weaken_substhvl_TxVar (HS TmVar H0) del)))) (fun (k127 : Hvl) (t1 : Tm) (wf0 : (wfTm k127 t1)) IHt189 (t2 : Tm) (wf1 : (wfTm k127 t2)) IHt190 {d81 : (Trace TxVar)} {k128 : Hvl} (del : (substhvl_TxVar k126 d81 k127 k128)) =>
    (wfTm_App k128 (IHt189 (weakenTrace d81 H0) k128 (weaken_substhvl_TxVar H0 del)) (IHt190 (weakenTrace d81 H0) k128 (weaken_substhvl_TxVar H0 del)))) (fun (k127 : Hvl) (t : Tm) (wf0 : (wfTm (HS TyVar k127) t)) IHt189 {d81 : (Trace TxVar)} {k128 : Hvl} (del : (substhvl_TxVar k126 d81 k127 k128)) =>
    (wfTm_TAbs k128 (IHt189 (weakenTrace d81 (HS TyVar H0)) (HS TyVar k128) (weaken_substhvl_TxVar (HS TyVar H0) del)))) (fun (k127 : Hvl) (t : Tm) (wf0 : (wfTm k127 t)) IHt189 (T : Ty) (wf1 : (wfTy k127 T)) {d81 : (Trace TxVar)} {k128 : Hvl} (del : (substhvl_TxVar k126 d81 k127 k128)) =>
    (wfTm_TApp k128 (IHt189 (weakenTrace d81 H0) k128 (weaken_substhvl_TxVar H0 del)) (substhvl_TxVar_wfTy k127 T wf1 (weaken_substhvl_TxVar H0 del))))).
  Definition substhvl_TyVar_wfTm {k126 : Hvl} {S55 : Ty} (wf : (wfTy k126 S55)) : (forall (k127 : Hvl) ,
    (forall (s59 : Tm) (wf0 : (wfTm k127 s59)) ,
      (forall {d81 : (Trace TyVar)} {k128 : Hvl} ,
        (substhvl_TyVar k126 d81 k127 k128) -> (wfTm k128 (subst_TyVar_Tm d81 S55 s59))))) := (ind_wfTm (fun (k127 : Hvl) (s59 : Tm) (wf0 : (wfTm k127 s59)) =>
    (forall {d81 : (Trace TyVar)} {k128 : Hvl} ,
      (substhvl_TyVar k126 d81 k127 k128) -> (wfTm k128 (subst_TyVar_Tm d81 S55 s59)))) (fun (k127 : Hvl) {x108 : (Index TmVar)} (wfi : (wfindex k127 x108)) {d81 : (Trace TyVar)} {k128 : Hvl} (del : (substhvl_TyVar k126 d81 k127 k128)) =>
    (wfTm_Var k128 (substhvl_TyVar_wfindex_TmVar del wfi))) (fun (k127 : Hvl) {a70 : (Index TxVar)} (wfi : (wfindex k127 a70)) {d81 : (Trace TyVar)} {k128 : Hvl} (del : (substhvl_TyVar k126 d81 k127 k128)) =>
    (wfTm_XVar k128 (substhvl_TyVar_wfindex_TxVar del wfi))) (fun (k127 : Hvl) (T : Ty) (wf0 : (wfTy k127 T)) (t : Tm) (wf1 : (wfTm (HS TmVar k127) t)) IHt189 {d81 : (Trace TyVar)} {k128 : Hvl} (del : (substhvl_TyVar k126 d81 k127 k128)) =>
    (wfTm_Abs k128 (substhvl_TyVar_wfTy wf k127 T wf0 (weaken_substhvl_TyVar H0 del)) (IHt189 (weakenTrace d81 (HS TmVar H0)) (HS TmVar k128) (weaken_substhvl_TyVar (HS TmVar H0) del)))) (fun (k127 : Hvl) (t1 : Tm) (wf0 : (wfTm k127 t1)) IHt189 (t2 : Tm) (wf1 : (wfTm k127 t2)) IHt190 {d81 : (Trace TyVar)} {k128 : Hvl} (del : (substhvl_TyVar k126 d81 k127 k128)) =>
    (wfTm_App k128 (IHt189 (weakenTrace d81 H0) k128 (weaken_substhvl_TyVar H0 del)) (IHt190 (weakenTrace d81 H0) k128 (weaken_substhvl_TyVar H0 del)))) (fun (k127 : Hvl) (t : Tm) (wf0 : (wfTm (HS TyVar k127) t)) IHt189 {d81 : (Trace TyVar)} {k128 : Hvl} (del : (substhvl_TyVar k126 d81 k127 k128)) =>
    (wfTm_TAbs k128 (IHt189 (weakenTrace d81 (HS TyVar H0)) (HS TyVar k128) (weaken_substhvl_TyVar (HS TyVar H0) del)))) (fun (k127 : Hvl) (t : Tm) (wf0 : (wfTm k127 t)) IHt189 (T : Ty) (wf1 : (wfTy k127 T)) {d81 : (Trace TyVar)} {k128 : Hvl} (del : (substhvl_TyVar k126 d81 k127 k128)) =>
    (wfTm_TApp k128 (IHt189 (weakenTrace d81 H0) k128 (weaken_substhvl_TyVar H0 del)) (substhvl_TyVar_wfTy wf k127 T wf1 (weaken_substhvl_TyVar H0 del))))).
End SubstWellFormed.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TxVar substhvl_TmVar_wfindex_TyVar substhvl_TxVar_wfindex_TmVar substhvl_TxVar_wfindex_TxVar substhvl_TxVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TxVar substhvl_TyVar_wfindex_TyVar : infra.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TxVar substhvl_TmVar_wfindex_TyVar substhvl_TxVar_wfindex_TmVar substhvl_TxVar_wfindex_TxVar substhvl_TxVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TxVar substhvl_TyVar_wfindex_TyVar : subst.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TxVar substhvl_TmVar_wfindex_TyVar substhvl_TxVar_wfindex_TmVar substhvl_TxVar_wfindex_TxVar substhvl_TxVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TxVar substhvl_TyVar_wfindex_TyVar : subst_wf.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TxVar substhvl_TmVar_wfindex_TyVar substhvl_TxVar_wfindex_TmVar substhvl_TxVar_wfindex_TxVar substhvl_TxVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TxVar substhvl_TyVar_wfindex_TyVar : wf.
 Hint Resolve substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TxVar_wfTm substhvl_TxVar_wfTy substhvl_TyVar_wfTm substhvl_TyVar_wfTy : infra.
 Hint Resolve substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TxVar_wfTm substhvl_TxVar_wfTy substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst.
 Hint Resolve substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TxVar_wfTm substhvl_TxVar_wfTy substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst_wf.
 Hint Resolve substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TxVar_wfTm substhvl_TxVar_wfTy substhvl_TyVar_wfTm substhvl_TyVar_wfTy : wf.
 Hint Constructors substhvl_TmVar substhvl_TxVar substhvl_TyVar : infra.
 Hint Constructors substhvl_TmVar substhvl_TxVar substhvl_TyVar : subst.
 Hint Constructors substhvl_TmVar substhvl_TxVar substhvl_TyVar : subst_wf.
 Hint Constructors substhvl_TmVar substhvl_TxVar substhvl_TyVar : wf.
Section Context.
  Inductive Ctx : Type :=
    | empty 
    | evar (G : Ctx) (T : Ty)
    | evarx (G : Ctx) (T : Ty)
    | etvar (G : Ctx).
  Fixpoint appendCtx (G : Ctx) (G0 : Ctx) : Ctx :=
    match G0 with
      | (empty) => G
      | (evar G1 T) => (evar (appendCtx G G1) T)
      | (evarx G1 T) => (evarx (appendCtx G G1) T)
      | (etvar G1) => (etvar (appendCtx G G1))
    end.
  Fixpoint domainCtx (G : Ctx) : Hvl :=
    match G with
      | (empty) => H0
      | (evar G0 T) => (HS TmVar (domainCtx G0))
      | (evarx G0 T) => (HS TxVar (domainCtx G0))
      | (etvar G0) => (HS TyVar (domainCtx G0))
    end.
  Lemma appendCtx_assoc  :
    (forall (G : Ctx) (G0 : Ctx) (G1 : Ctx) ,
      ((appendCtx G (appendCtx G0 G1)) =
      (appendCtx (appendCtx G G0) G1))).
  Proof.
    needleGenericAppendEnvAssoc .
  Qed.
  Lemma domainCtx_appendCtx  :
    (forall (G : Ctx) (G0 : Ctx) ,
      ((domainCtx (appendCtx G G0)) =
      (appendHvl (domainCtx G) (domainCtx G0)))).
  Proof.
    needleGenericDomainEnvAppendEnv .
  Qed.
  Fixpoint shift_TmVar_Ctx (c44 : (Cutoff TmVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shift_TmVar_Ctx c44 G0) T)
      | (evarx G0 T) => (evarx (shift_TmVar_Ctx c44 G0) T)
      | (etvar G0) => (etvar (shift_TmVar_Ctx c44 G0))
    end.
  Fixpoint shift_TxVar_Ctx (c44 : (Cutoff TxVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shift_TxVar_Ctx c44 G0) T)
      | (evarx G0 T) => (evarx (shift_TxVar_Ctx c44 G0) T)
      | (etvar G0) => (etvar (shift_TxVar_Ctx c44 G0))
    end.
  Fixpoint shift_TyVar_Ctx (c44 : (Cutoff TyVar)) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shift_TyVar_Ctx c44 G0) (shift_TyVar_Ty (weakenCutoffTyVar c44 (domainCtx G0)) T))
      | (evarx G0 T) => (evarx (shift_TyVar_Ctx c44 G0) (shift_TyVar_Ty (weakenCutoffTyVar c44 (domainCtx G0)) T))
      | (etvar G0) => (etvar (shift_TyVar_Ctx c44 G0))
    end.
  Fixpoint subst_TmVar_Ctx (d81 : (Trace TmVar)) (s59 : Tm) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (subst_TmVar_Ctx d81 s59 G0) T)
      | (evarx G0 T) => (evarx (subst_TmVar_Ctx d81 s59 G0) T)
      | (etvar G0) => (etvar (subst_TmVar_Ctx d81 s59 G0))
    end.
  Fixpoint subst_TxVar_Ctx (d81 : (Trace TxVar)) (s59 : Tm) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (subst_TxVar_Ctx d81 s59 G0) T)
      | (evarx G0 T) => (evarx (subst_TxVar_Ctx d81 s59 G0) T)
      | (etvar G0) => (etvar (subst_TxVar_Ctx d81 s59 G0))
    end.
  Fixpoint subst_TyVar_Ctx (d81 : (Trace TyVar)) (S55 : Ty) (G : Ctx) : Ctx :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (subst_TyVar_Ctx d81 S55 G0) (subst_TyVar_Ty (weakenTrace d81 (domainCtx G0)) S55 T))
      | (evarx G0 T) => (evarx (subst_TyVar_Ctx d81 S55 G0) (subst_TyVar_Ty (weakenTrace d81 (domainCtx G0)) S55 T))
      | (etvar G0) => (etvar (subst_TyVar_Ctx d81 S55 G0))
    end.
  Lemma domainCtx_shift_TmVar_Ctx  :
    (forall (c44 : (Cutoff TmVar)) (G : Ctx) ,
      ((domainCtx (shift_TmVar_Ctx c44 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_shift_TxVar_Ctx  :
    (forall (c44 : (Cutoff TxVar)) (G : Ctx) ,
      ((domainCtx (shift_TxVar_Ctx c44 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_shift_TyVar_Ctx  :
    (forall (c44 : (Cutoff TyVar)) (G : Ctx) ,
      ((domainCtx (shift_TyVar_Ctx c44 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainCtx_subst_TmVar_Ctx  :
    (forall (d81 : (Trace TmVar)) (s59 : Tm) (G : Ctx) ,
      ((domainCtx (subst_TmVar_Ctx d81 s59 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
  Lemma domainCtx_subst_TxVar_Ctx  :
    (forall (d81 : (Trace TxVar)) (s59 : Tm) (G : Ctx) ,
      ((domainCtx (subst_TxVar_Ctx d81 s59 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
  Lemma domainCtx_subst_TyVar_Ctx  :
    (forall (d81 : (Trace TyVar)) (S55 : Ty) (G : Ctx) ,
      ((domainCtx (subst_TyVar_Ctx d81 S55 G)) =
      (domainCtx G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
End Context.
 Hint Rewrite domainCtx_appendCtx : interaction.
 Hint Rewrite domainCtx_appendCtx : env_domain_append.
Section ContextStuff.
  Section ShiftEnvAppendEnv.
    Lemma shift_TmVar_Ctx_appendCtx  :
      (forall (c44 : (Cutoff TmVar)) (G : Ctx) (G0 : Ctx) ,
        ((shift_TmVar_Ctx c44 (appendCtx G G0)) =
        (appendCtx (shift_TmVar_Ctx c44 G) (shift_TmVar_Ctx (weakenCutoffTmVar c44 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
    Lemma shift_TxVar_Ctx_appendCtx  :
      (forall (c44 : (Cutoff TxVar)) (G : Ctx) (G0 : Ctx) ,
        ((shift_TxVar_Ctx c44 (appendCtx G G0)) =
        (appendCtx (shift_TxVar_Ctx c44 G) (shift_TxVar_Ctx (weakenCutoffTxVar c44 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
    Lemma shift_TyVar_Ctx_appendCtx  :
      (forall (c44 : (Cutoff TyVar)) (G : Ctx) (G0 : Ctx) ,
        ((shift_TyVar_Ctx c44 (appendCtx G G0)) =
        (appendCtx (shift_TyVar_Ctx c44 G) (shift_TyVar_Ctx (weakenCutoffTyVar c44 (domainCtx G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
  End ShiftEnvAppendEnv.
  Section SubstEnvAppendEnv.
    Lemma subst_TmVar_Ctx_appendCtx  :
      (forall (d81 : (Trace TmVar)) (s59 : Tm) (G : Ctx) (G0 : Ctx) ,
        ((subst_TmVar_Ctx d81 s59 (appendCtx G G0)) =
        (appendCtx (subst_TmVar_Ctx d81 s59 G) (subst_TmVar_Ctx (weakenTrace d81 (domainCtx G)) s59 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
    Lemma subst_TxVar_Ctx_appendCtx  :
      (forall (d81 : (Trace TxVar)) (s59 : Tm) (G : Ctx) (G0 : Ctx) ,
        ((subst_TxVar_Ctx d81 s59 (appendCtx G G0)) =
        (appendCtx (subst_TxVar_Ctx d81 s59 G) (subst_TxVar_Ctx (weakenTrace d81 (domainCtx G)) s59 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
    Lemma subst_TyVar_Ctx_appendCtx  :
      (forall (d81 : (Trace TyVar)) (S55 : Ty) (G : Ctx) (G0 : Ctx) ,
        ((subst_TyVar_Ctx d81 S55 (appendCtx G G0)) =
        (appendCtx (subst_TyVar_Ctx d81 S55 G) (subst_TyVar_Ctx (weakenTrace d81 (domainCtx G)) S55 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
  End SubstEnvAppendEnv.
  Section Lookups.
    Inductive lookup_evar : Ctx -> (Index TmVar) -> Ty -> Prop :=
      | lookup_evar_here {G : Ctx}
          {T99 : Ty} :
          (wfTy (domainCtx G) T99) -> (lookup_evar (evar G T99) I0 T99)
      | lookup_evar_there_evar
          {G : Ctx} {x108 : (Index TmVar)}
          {T99 : Ty} {T100 : Ty} :
          (lookup_evar G x108 T99) -> (lookup_evar (evar G T100) (IS x108) T99)
      | lookup_evar_there_evarx
          {G : Ctx} {x108 : (Index TmVar)}
          {T99 : Ty} {T100 : Ty} :
          (lookup_evar G x108 T99) -> (lookup_evar (evarx G T100) x108 T99)
      | lookup_evar_there_etvar
          {G : Ctx} {x108 : (Index TmVar)}
          {T99 : Ty} :
          (lookup_evar G x108 T99) -> (lookup_evar (etvar G) x108 (shift_TyVar_Ty C0 T99)).
    Inductive lookup_evarx : Ctx -> (Index TxVar) -> Ty -> Prop :=
      | lookup_evarx_here {G : Ctx}
          {T99 : Ty} :
          (wfTy (domainCtx G) T99) -> (lookup_evarx (evarx G T99) I0 T99)
      | lookup_evarx_there_evar
          {G : Ctx} {a70 : (Index TxVar)}
          {T99 : Ty} {T100 : Ty} :
          (lookup_evarx G a70 T99) -> (lookup_evarx (evar G T100) a70 T99)
      | lookup_evarx_there_evarx
          {G : Ctx} {a70 : (Index TxVar)}
          {T99 : Ty} {T100 : Ty} :
          (lookup_evarx G a70 T99) -> (lookup_evarx (evarx G T100) (IS a70) T99)
      | lookup_evarx_there_etvar
          {G : Ctx} {a70 : (Index TxVar)}
          {T99 : Ty} :
          (lookup_evarx G a70 T99) -> (lookup_evarx (etvar G) a70 (shift_TyVar_Ty C0 T99)).
    Inductive lookup_etvar : Ctx -> (Index TyVar) -> Prop :=
      | lookup_etvar_here {G : Ctx}
          : (lookup_etvar (etvar G) I0)
      | lookup_etvar_there_evar
          {G : Ctx} {X71 : (Index TyVar)}
          {T99 : Ty} :
          (lookup_etvar G X71) -> (lookup_etvar (evar G T99) X71)
      | lookup_etvar_there_evarx
          {G : Ctx} {X71 : (Index TyVar)}
          {T99 : Ty} :
          (lookup_etvar G X71) -> (lookup_etvar (evarx G T99) X71)
      | lookup_etvar_there_etvar
          {G : Ctx} {X71 : (Index TyVar)}
          :
          (lookup_etvar G X71) -> (lookup_etvar (etvar G) (IS X71)).
    Lemma lookup_evar_inversion_here  :
      (forall (G : Ctx) (S55 : Ty) (S56 : Ty) ,
        (lookup_evar (evar G S55) I0 S56) -> (S55 =
        S56)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_evarx_inversion_here  :
      (forall (G : Ctx) (S55 : Ty) (S56 : Ty) ,
        (lookup_evarx (evarx G S55) I0 S56) -> (S55 =
        S56)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_evar_functional  :
      (forall {G : Ctx} {x108 : (Index TmVar)} ,
        (forall {S55 : Ty} ,
          (lookup_evar G x108 S55) -> (forall {S56 : Ty} ,
            (lookup_evar G x108 S56) -> (S55 =
            S56)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_evarx_functional  :
      (forall {G : Ctx} {a70 : (Index TxVar)} ,
        (forall {S55 : Ty} ,
          (lookup_evarx G a70 S55) -> (forall {S56 : Ty} ,
            (lookup_evarx G a70 S56) -> (S55 =
            S56)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_evar_wf  :
      (forall {G : Ctx} {x108 : (Index TmVar)} {S55 : Ty} ,
        (lookup_evar G x108 S55) -> (wfTy (domainCtx G) S55)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma lookup_evarx_wf  :
      (forall {G : Ctx} {a70 : (Index TxVar)} {S55 : Ty} ,
        (lookup_evarx G a70 S55) -> (wfTy (domainCtx G) S55)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma weaken_lookup_evar  :
      (forall {G : Ctx} (G0 : Ctx) {x108 : (Index TmVar)} {T99 : Ty} ,
        (lookup_evar G x108 T99) -> (lookup_evar (appendCtx G G0) (weakenIndexTmVar x108 (domainCtx G0)) (weakenTy T99 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma weaken_lookup_evarx  :
      (forall {G : Ctx} (G0 : Ctx) {a70 : (Index TxVar)} {T99 : Ty} ,
        (lookup_evarx G a70 T99) -> (lookup_evarx (appendCtx G G0) (weakenIndexTxVar a70 (domainCtx G0)) (weakenTy T99 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma weaken_lookup_etvar  :
      (forall {G : Ctx} (G0 : Ctx) {X71 : (Index TyVar)} ,
        (lookup_etvar G X71) -> (lookup_etvar (appendCtx G G0) (weakenIndexTyVar X71 (domainCtx G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma lookup_evar_wfindex  :
      (forall {G : Ctx} {x108 : (Index TmVar)} {S55 : Ty} ,
        (lookup_evar G x108 S55) -> (wfindex (domainCtx G) x108)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
    Lemma lookup_evarx_wfindex  :
      (forall {G : Ctx} {a70 : (Index TxVar)} {S55 : Ty} ,
        (lookup_evarx G a70 S55) -> (wfindex (domainCtx G) a70)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
    Lemma lookup_etvar_wfindex  :
      (forall {G : Ctx} {X71 : (Index TyVar)} ,
        (lookup_etvar G X71) -> (wfindex (domainCtx G) X71)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
  End Lookups.
  Inductive shift_evar : (Cutoff TmVar) -> Ctx -> Ctx -> Prop :=
    | shift_evar_here {G : Ctx}
        {T99 : Ty} :
        (shift_evar C0 G (evar G T99))
    | shiftevar_there_evar
        {c44 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_evar c44 G G0) -> (shift_evar (CS c44) (evar G T) (evar G0 T))
    | shiftevar_there_evarx
        {c44 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_evar c44 G G0) -> (shift_evar c44 (evarx G T) (evarx G0 T))
    | shiftevar_there_etvar
        {c44 : (Cutoff TmVar)} {G : Ctx}
        {G0 : Ctx} :
        (shift_evar c44 G G0) -> (shift_evar c44 (etvar G) (etvar G0)).
  Lemma weaken_shift_evar  :
    (forall (G : Ctx) {c44 : (Cutoff TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_evar c44 G0 G1) -> (shift_evar (weakenCutoffTmVar c44 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 G))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Inductive shift_evarx : (Cutoff TxVar) -> Ctx -> Ctx -> Prop :=
    | shift_evarx_here {G : Ctx}
        {T99 : Ty} :
        (shift_evarx C0 G (evarx G T99))
    | shiftevarx_there_evar
        {c44 : (Cutoff TxVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_evarx c44 G G0) -> (shift_evarx c44 (evar G T) (evar G0 T))
    | shiftevarx_there_evarx
        {c44 : (Cutoff TxVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_evarx c44 G G0) -> (shift_evarx (CS c44) (evarx G T) (evarx G0 T))
    | shiftevarx_there_etvar
        {c44 : (Cutoff TxVar)} {G : Ctx}
        {G0 : Ctx} :
        (shift_evarx c44 G G0) -> (shift_evarx c44 (etvar G) (etvar G0)).
  Lemma weaken_shift_evarx  :
    (forall (G : Ctx) {c44 : (Cutoff TxVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_evarx c44 G0 G1) -> (shift_evarx (weakenCutoffTxVar c44 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 G))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Inductive shift_etvar : (Cutoff TyVar) -> Ctx -> Ctx -> Prop :=
    | shift_etvar_here {G : Ctx} :
        (shift_etvar C0 G (etvar G))
    | shiftetvar_there_evar
        {c44 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_etvar c44 G G0) -> (shift_etvar c44 (evar G T) (evar G0 (shift_TyVar_Ty c44 T)))
    | shiftetvar_there_evarx
        {c44 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} {T : Ty} :
        (shift_etvar c44 G G0) -> (shift_etvar c44 (evarx G T) (evarx G0 (shift_TyVar_Ty c44 T)))
    | shiftetvar_there_etvar
        {c44 : (Cutoff TyVar)} {G : Ctx}
        {G0 : Ctx} :
        (shift_etvar c44 G G0) -> (shift_etvar (CS c44) (etvar G) (etvar G0)).
  Lemma weaken_shift_etvar  :
    (forall (G : Ctx) {c44 : (Cutoff TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (shift_etvar c44 G0 G1) -> (shift_etvar (weakenCutoffTyVar c44 (domainCtx G)) (appendCtx G0 G) (appendCtx G1 (shift_TyVar_Ctx c44 G)))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Lemma shift_evar_shifthvl_TmVar  :
    (forall {c44 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_evar c44 G G0) -> (shifthvl_TmVar c44 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Lemma shift_evarx_shifthvl_TxVar  :
    (forall {c44 : (Cutoff TxVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_evarx c44 G G0) -> (shifthvl_TxVar c44 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Lemma shift_etvar_shifthvl_TyVar  :
    (forall {c44 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} ,
      (shift_etvar c44 G G0) -> (shifthvl_TyVar c44 (domainCtx G) (domainCtx G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Inductive subst_evar (G : Ctx) (T99 : Ty) (s59 : Tm) : (Trace TmVar) -> Ctx -> Ctx -> Prop :=
    | subst_evar_here :
        (subst_evar G T99 s59 X0 (evar G T99) G)
    | subst_evar_there_evar
        {d81 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_evar G T99 s59 d81 G0 G1) -> (subst_evar G T99 s59 (XS TmVar d81) (evar G0 T) (evar G1 T))
    | subst_evar_there_evarx
        {d81 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_evar G T99 s59 d81 G0 G1) -> (subst_evar G T99 s59 (XS TxVar d81) (evarx G0 T) (evarx G1 T))
    | subst_evar_there_etvar
        {d81 : (Trace TmVar)} {G0 : Ctx}
        {G1 : Ctx} :
        (subst_evar G T99 s59 d81 G0 G1) -> (subst_evar G T99 s59 (XS TyVar d81) (etvar G0) (etvar G1)).
  Lemma weaken_subst_evar {G : Ctx} {T99 : Ty} {s59 : Tm} :
    (forall (G0 : Ctx) {d81 : (Trace TmVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_evar G T99 s59 d81 G1 G2) -> (subst_evar G T99 s59 (weakenTrace d81 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 G0))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Inductive subst_evarx (G : Ctx) (T99 : Ty) (s59 : Tm) : (Trace TxVar) -> Ctx -> Ctx -> Prop :=
    | subst_evarx_here :
        (subst_evarx G T99 s59 X0 (evarx G T99) G)
    | subst_evarx_there_evar
        {d81 : (Trace TxVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_evarx G T99 s59 d81 G0 G1) -> (subst_evarx G T99 s59 (XS TmVar d81) (evar G0 T) (evar G1 T))
    | subst_evarx_there_evarx
        {d81 : (Trace TxVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_evarx G T99 s59 d81 G0 G1) -> (subst_evarx G T99 s59 (XS TxVar d81) (evarx G0 T) (evarx G1 T))
    | subst_evarx_there_etvar
        {d81 : (Trace TxVar)} {G0 : Ctx}
        {G1 : Ctx} :
        (subst_evarx G T99 s59 d81 G0 G1) -> (subst_evarx G T99 s59 (XS TyVar d81) (etvar G0) (etvar G1)).
  Lemma weaken_subst_evarx {G : Ctx} {T99 : Ty} {s59 : Tm} :
    (forall (G0 : Ctx) {d81 : (Trace TxVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_evarx G T99 s59 d81 G1 G2) -> (subst_evarx G T99 s59 (weakenTrace d81 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 G0))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Inductive subst_etvar (G : Ctx) (S55 : Ty) : (Trace TyVar) -> Ctx -> Ctx -> Prop :=
    | subst_etvar_here :
        (subst_etvar G S55 X0 (etvar G) G)
    | subst_etvar_there_evar
        {d81 : (Trace TyVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_etvar G S55 d81 G0 G1) -> (subst_etvar G S55 (XS TmVar d81) (evar G0 T) (evar G1 (subst_TyVar_Ty d81 S55 T)))
    | subst_etvar_there_evarx
        {d81 : (Trace TyVar)} {G0 : Ctx}
        {G1 : Ctx} {T : Ty} :
        (subst_etvar G S55 d81 G0 G1) -> (subst_etvar G S55 (XS TxVar d81) (evarx G0 T) (evarx G1 (subst_TyVar_Ty d81 S55 T)))
    | subst_etvar_there_etvar
        {d81 : (Trace TyVar)} {G0 : Ctx}
        {G1 : Ctx} :
        (subst_etvar G S55 d81 G0 G1) -> (subst_etvar G S55 (XS TyVar d81) (etvar G0) (etvar G1)).
  Lemma weaken_subst_etvar {G : Ctx} {S55 : Ty} :
    (forall (G0 : Ctx) {d81 : (Trace TyVar)} {G1 : Ctx} {G2 : Ctx} ,
      (subst_etvar G S55 d81 G1 G2) -> (subst_etvar G S55 (weakenTrace d81 (domainCtx G0)) (appendCtx G1 G0) (appendCtx G2 (subst_TyVar_Ctx d81 S55 G0)))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Lemma subst_evar_substhvl_TmVar {G : Ctx} {s59 : Tm} {T99 : Ty} :
    (forall {d81 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_evar G T99 s59 d81 G0 G1) -> (substhvl_TmVar (domainCtx G) d81 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
  Lemma subst_evarx_substhvl_TxVar {G : Ctx} {s59 : Tm} {T100 : Ty} :
    (forall {d81 : (Trace TxVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_evarx G T100 s59 d81 G0 G1) -> (substhvl_TxVar (domainCtx G) d81 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
  Lemma subst_etvar_substhvl_TyVar {G : Ctx} {S55 : Ty} :
    (forall {d81 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} ,
      (subst_etvar G S55 d81 G0 G1) -> (substhvl_TyVar (domainCtx G) d81 (domainCtx G0) (domainCtx G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
End ContextStuff.
 Hint Rewrite domainCtx_shift_TyVar_Ctx : interaction.
 Hint Rewrite domainCtx_shift_TyVar_Ctx : env_domain_shift.
 Hint Rewrite domainCtx_subst_TyVar_Ctx : interaction.
 Hint Rewrite domainCtx_subst_TyVar_Ctx : env_domain_subst.
 Hint Rewrite shift_TyVar_Ctx_appendCtx : interaction.
 Hint Rewrite shift_TyVar_Ctx_appendCtx : env_shift_append.
 Hint Rewrite subst_TyVar_Ctx_appendCtx : interaction.
 Hint Rewrite subst_TyVar_Ctx_appendCtx : env_shift_append.
 Hint Constructors lookup_evar lookup_evarx lookup_etvar : infra.
 Hint Constructors lookup_evar lookup_evarx lookup_etvar : lookup.
 Hint Constructors lookup_evar lookup_evarx lookup_etvar : shift.
 Hint Constructors lookup_evar lookup_evarx lookup_etvar : subst.
 Hint Resolve weaken_lookup_evar weaken_lookup_evarx weaken_lookup_etvar : infra.
 Hint Resolve weaken_lookup_evar weaken_lookup_evarx weaken_lookup_etvar : lookup.
 Hint Resolve weaken_lookup_evar weaken_lookup_evarx weaken_lookup_etvar : shift.
 Hint Resolve weaken_lookup_evar weaken_lookup_evarx weaken_lookup_etvar : weaken.
Lemma weaken_lookup_evar_here  :
  (forall {G : Ctx} (G0 : Ctx) {T99 : Ty} (wf : (wfTy (domainCtx G) T99)) ,
    (lookup_evar (appendCtx (evar G T99) G0) (weakenIndexTmVar I0 (domainCtx G0)) (weakenTy T99 (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
Lemma weaken_lookup_evarx_here  :
  (forall {G : Ctx} (G0 : Ctx) {T99 : Ty} (wf : (wfTy (domainCtx G) T99)) ,
    (lookup_evarx (appendCtx (evarx G T99) G0) (weakenIndexTxVar I0 (domainCtx G0)) (weakenTy T99 (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
Lemma weaken_lookup_etvar_here  :
  (forall {G : Ctx} (G0 : Ctx) ,
    (lookup_etvar (appendCtx (etvar G) G0) (weakenIndexTyVar I0 (domainCtx G0)))).
Proof.
  eauto with lookup .
Qed.
 Hint Constructors wfTm wfTy : infra.
 Hint Constructors wfTm wfTy : wf.
 Hint Extern 10 ((wfTm _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTy _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 2 ((wfTy _ _)) => match goal with
  | H : (wfTy _ (TVar _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TArr _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (TAll _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfTm _ _)) => match goal with
  | H : (wfTm _ (Var _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (XVar _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (Abs _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (App _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (TAbs _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (TApp _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Resolve lookup_evar_wf lookup_evarx_wf : infra.
 Hint Resolve lookup_evar_wf lookup_evarx_wf : wf.
 Hint Resolve lookup_evar_wfindex lookup_evarx_wfindex lookup_etvar_wfindex : infra.
 Hint Resolve lookup_evar_wfindex lookup_evarx_wfindex lookup_etvar_wfindex : lookup.
 Hint Resolve lookup_evar_wfindex lookup_evarx_wfindex lookup_etvar_wfindex : wf.
 Hint Constructors shift_evar shift_evarx shift_etvar : infra.
 Hint Constructors shift_evar shift_evarx shift_etvar : shift.
 Hint Constructors shift_evar shift_evarx shift_etvar : subst.
 Hint Resolve weaken_shift_evar weaken_shift_evarx weaken_shift_etvar : infra.
 Hint Resolve weaken_shift_evar weaken_shift_evarx weaken_shift_etvar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_evarx_shifthvl_TxVar shift_etvar_shifthvl_TyVar : infra.
 Hint Resolve shift_evar_shifthvl_TmVar shift_evarx_shifthvl_TxVar shift_etvar_shifthvl_TyVar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_evarx_shifthvl_TxVar shift_etvar_shifthvl_TyVar : shift_wf.
 Hint Resolve shift_evar_shifthvl_TmVar shift_evarx_shifthvl_TxVar shift_etvar_shifthvl_TyVar : wf.
 Hint Constructors subst_evar subst_evarx subst_etvar : infra.
 Hint Constructors subst_evar subst_evarx subst_etvar : subst.
 Hint Resolve weaken_subst_evar weaken_subst_evarx weaken_subst_etvar : infra.
 Hint Resolve weaken_subst_evar weaken_subst_evarx weaken_subst_etvar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_evarx_substhvl_TxVar subst_etvar_substhvl_TyVar : infra.
 Hint Resolve subst_evar_substhvl_TmVar subst_evarx_substhvl_TxVar subst_etvar_substhvl_TyVar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_evarx_substhvl_TxVar subst_etvar_substhvl_TyVar : subst_wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_evarx_substhvl_TxVar subst_etvar_substhvl_TyVar : wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_evarx_substhvl_TxVar subst_etvar_substhvl_TyVar : substenv_substhvl.
Lemma shift_evar_lookup_evar  :
  (forall {c44 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c44 G G0)) {x108 : (Index TmVar)} {T99 : Ty} ,
    (lookup_evar G x108 T99) -> (lookup_evar G0 (shift_TmVar_Index c44 x108) T99)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_evar_lookup_evarx  :
  (forall {c44 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c44 G G0)) {a70 : (Index TxVar)} {T99 : Ty} ,
    (lookup_evarx G a70 T99) -> (lookup_evarx G0 a70 T99)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_evar_lookup_etvar  :
  (forall {c44 : (Cutoff TmVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evar c44 G G0)) {X71 : (Index TyVar)} ,
    (lookup_etvar G X71) -> (lookup_etvar G0 X71)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_evarx_lookup_evar  :
  (forall {c44 : (Cutoff TxVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evarx c44 G G0)) {x108 : (Index TmVar)} {T99 : Ty} ,
    (lookup_evar G x108 T99) -> (lookup_evar G0 x108 T99)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_evarx_lookup_evarx  :
  (forall {c44 : (Cutoff TxVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evarx c44 G G0)) {a70 : (Index TxVar)} {T99 : Ty} ,
    (lookup_evarx G a70 T99) -> (lookup_evarx G0 (shift_TxVar_Index c44 a70) T99)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_evarx_lookup_etvar  :
  (forall {c44 : (Cutoff TxVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_evarx c44 G G0)) {X71 : (Index TyVar)} ,
    (lookup_etvar G X71) -> (lookup_etvar G0 X71)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_evar  :
  (forall {c44 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c44 G G0)) {x108 : (Index TmVar)} {T99 : Ty} ,
    (lookup_evar G x108 T99) -> (lookup_evar G0 x108 (shift_TyVar_Ty c44 T99))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_evarx  :
  (forall {c44 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c44 G G0)) {a70 : (Index TxVar)} {T99 : Ty} ,
    (lookup_evarx G a70 T99) -> (lookup_evarx G0 a70 (shift_TyVar_Ty c44 T99))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_etvar_lookup_etvar  :
  (forall {c44 : (Cutoff TyVar)} {G : Ctx} {G0 : Ctx} (ins : (shift_etvar c44 G G0)) {X71 : (Index TyVar)} ,
    (lookup_etvar G X71) -> (lookup_etvar G0 (shift_TyVar_Index c44 X71))).
Proof.
  needleGenericInsertLookup .
Qed.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_evarx shift_evar_lookup_etvar shift_evarx_lookup_evar shift_evarx_lookup_evarx shift_evarx_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_evarx shift_etvar_lookup_etvar : infra.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_evarx shift_evar_lookup_etvar shift_evarx_lookup_evar shift_evarx_lookup_evarx shift_evarx_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_evarx shift_etvar_lookup_etvar : lookup.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_evarx shift_evar_lookup_etvar shift_evarx_lookup_evar shift_evarx_lookup_evarx shift_evarx_lookup_etvar shift_etvar_lookup_evar shift_etvar_lookup_evarx shift_etvar_lookup_etvar : shift.
Lemma subst_evar_lookup_evarx {G : Ctx} {T101 : Ty} {s59 : Tm} :
  (forall {d81 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_evar G T101 s59 d81 G0 G1)) {a70 : (Index TxVar)} {T102 : Ty} ,
    (lookup_evarx G0 a70 T102) -> (lookup_evarx G1 a70 T102)).
Proof.
  needleGenericSubstEnvLookup .
Qed.
Lemma subst_evar_lookup_etvar {G : Ctx} {T101 : Ty} {s59 : Tm} :
  (forall {d81 : (Trace TmVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_evar G T101 s59 d81 G0 G1)) {X71 : (Index TyVar)} ,
    (lookup_etvar G0 X71) -> (lookup_etvar G1 X71)).
Proof.
  needleGenericSubstEnvLookup .
Qed.
Lemma subst_evarx_lookup_evar {G : Ctx} {T101 : Ty} {s59 : Tm} :
  (forall {d81 : (Trace TxVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_evarx G T101 s59 d81 G0 G1)) {x108 : (Index TmVar)} {T102 : Ty} ,
    (lookup_evar G0 x108 T102) -> (lookup_evar G1 x108 T102)).
Proof.
  needleGenericSubstEnvLookup .
Qed.
Lemma subst_evarx_lookup_etvar {G : Ctx} {T101 : Ty} {s59 : Tm} :
  (forall {d81 : (Trace TxVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_evarx G T101 s59 d81 G0 G1)) {X71 : (Index TyVar)} ,
    (lookup_etvar G0 X71) -> (lookup_etvar G1 X71)).
Proof.
  needleGenericSubstEnvLookup .
Qed.
Lemma subst_etvar_lookup_evar {G : Ctx} {S55 : Ty} (wf : (wfTy (domainCtx G) S55)) :
  (forall {d81 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_etvar G S55 d81 G0 G1)) {x108 : (Index TmVar)} {T101 : Ty} ,
    (lookup_evar G0 x108 T101) -> (lookup_evar G1 x108 (subst_TyVar_Ty d81 S55 T101))).
Proof.
  needleGenericSubstEnvLookup .
Qed.
Lemma subst_etvar_lookup_evarx {G : Ctx} {S55 : Ty} (wf : (wfTy (domainCtx G) S55)) :
  (forall {d81 : (Trace TyVar)} {G0 : Ctx} {G1 : Ctx} (sub : (subst_etvar G S55 d81 G0 G1)) {a70 : (Index TxVar)} {T101 : Ty} ,
    (lookup_evarx G0 a70 T101) -> (lookup_evarx G1 a70 (subst_TyVar_Ty d81 S55 T101))).
Proof.
  needleGenericSubstEnvLookup .
Qed.
 Hint Resolve subst_evar_lookup_evarx subst_evar_lookup_etvar subst_evarx_lookup_evar subst_evarx_lookup_etvar subst_etvar_lookup_evar subst_etvar_lookup_evarx : infra.
 Hint Resolve subst_evar_lookup_evarx subst_evar_lookup_etvar subst_evarx_lookup_evar subst_evarx_lookup_etvar subst_etvar_lookup_evar subst_etvar_lookup_evarx : lookup.
 Hint Resolve subst_evar_lookup_evarx subst_evar_lookup_etvar subst_evarx_lookup_evar subst_evarx_lookup_etvar subst_etvar_lookup_evar subst_etvar_lookup_evarx : subst.
Fixpoint size_Ty (S0 : Ty) {struct S0} : nat :=
  match S0 with
    | (TVar X) => 1
    | (TArr T1 T2) => (plus 1 (plus (size_Ty T1) (size_Ty T2)))
    | (TAll T) => (plus 1 (size_Ty T))
  end.
Fixpoint size_Tm (s : Tm) {struct s} : nat :=
  match s with
    | (Var x) => 1
    | (XVar a) => 1
    | (Abs T t) => (plus 1 (plus (size_Ty T) (size_Tm t)))
    | (App t1 t2) => (plus 1 (plus (size_Tm t1) (size_Tm t2)))
    | (TAbs t0) => (plus 1 (size_Tm t0))
    | (TApp t3 T0) => (plus 1 (plus (size_Tm t3) (size_Ty T0)))
  end.
Lemma shift_TyVar__size_Ty  :
  (forall (S55 : Ty) (c37 : (Cutoff TyVar)) ,
    ((size_Ty (shift_TyVar_Ty c37 S55)) =
    (size_Ty S55))).
Proof.
  apply_mutual_ind ind_Ty.
  - intros X64 c37.
  reflexivity.
  - intros T90 IHT90 T91 IHT91.
  intros c38; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT90 c38)).
  + exact ((IHT91 c38)).
  - intros T92 IHT92.
  intros c38; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT92 (CS c38))).
Qed.

Lemma shift_TmVar__size_Tm  :
  (forall (s59 : Tm) (c38 : (Cutoff TmVar)) ,
    ((size_Tm (shift_TmVar_Tm c38 s59)) =
    (size_Tm s59))).
Proof.
  apply_mutual_ind ind_Tm.
  - intros x101 c38.
  reflexivity.
  - intros x102 c39.
  reflexivity.
  - intros T93 t174 IHt174.
  intros c40; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + reflexivity.
  + exact ((IHt174 (CS c40))).
  - intros t175 IHt175 t176 IHt176.
  intros c40; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt175 c40)).
  + exact ((IHt176 c40)).
  - intros t177 IHt177.
  intros c40; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt177 c40)).
  - intros t178 IHt178 T94.
  intros c40; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt178 c40)).
  + reflexivity.
Qed.

Lemma shift_TxVar__size_Tm  :
  (forall (s59 : Tm) (c40 : (Cutoff TxVar)) ,
    ((size_Tm (shift_TxVar_Tm c40 s59)) =
    (size_Tm s59))).
Proof.
  apply_mutual_ind ind_Tm.
  - intros a66 c40.
  reflexivity.
  - intros a68 c41.
  reflexivity.
  - intros T95 t179 IHt179.
  intros c42; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + reflexivity.
  + exact ((IHt179 c42)).
  - intros t180 IHt180 t181 IHt181.
  intros c42; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt180 c42)).
  + exact ((IHt181 c42)).
  - intros t182 IHt182.
  intros c42; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt182 c42)).
  - intros t183 IHt183 T96.
  intros c42; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt183 c42)).
  + reflexivity.
Qed.

Lemma shift_TyVar__size_Tm  :
  (forall (s59 : Tm) (c42 : (Cutoff TyVar)) ,
    ((size_Tm (shift_TyVar_Tm c42 s59)) =
    (size_Tm s59))).
Proof.
  apply_mutual_ind ind_Tm.
  - intros X68 c42.
  reflexivity.
  - intros X69 c43.
  reflexivity.
  - intros T97 t184 IHt184.
  pose proof ((shift_TyVar__size_Ty T97)) as IHT97.
  intros c44; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT97 c44)).
  + exact ((IHt184 c44)).
  - intros t185 IHt185 t186 IHt186.
  intros c44; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt185 c44)).
  + exact ((IHt186 c44)).
  - intros t187 IHt187.
  intros c44; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt187 (CS c44))).
  - intros t188 IHt188 T98.
  pose proof ((shift_TyVar__size_Ty T98)) as IHT98.
  intros c44; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt188 c44)).
  + exact ((IHT98 c44)).
Qed.

 Hint Rewrite shift_TmVar__size_Tm shift_TxVar__size_Tm shift_TyVar__size_Tm shift_TyVar__size_Ty : interaction.
 Hint Rewrite shift_TmVar__size_Tm shift_TxVar__size_Tm shift_TyVar__size_Tm shift_TyVar__size_Ty : shift_size.
Lemma weaken_size_Tm  :
  (forall (k126 : Hvl) (s59 : Tm) ,
    ((size_Tm (weakenTm s59 k126)) =
    (size_Tm s59))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Ty  :
  (forall (k126 : Hvl) (S55 : Ty) ,
    ((size_Ty (weakenTy S55 k126)) =
    (size_Ty S55))).
Proof.
  needleGenericWeakenSize .
Qed.
 Hint Rewrite weaken_size_Tm weaken_size_Ty : interaction.
 Hint Rewrite weaken_size_Tm weaken_size_Ty : weaken_size.